Package: irace
Type: Package
Title: Iterated Racing for Automatic Algorithm Configuration
Description: Iterated race is an extension of the Iterated F-race method for
             the automatic configuration of optimization algorithms, that is,
             (offline) tuning their parameters by finding the most appropriate
             settings given a set of instances of an optimization problem.
             M. López-Ibáñez, J. Dubois-Lacoste, L. Pérez Cáceres, T. Stützle,
             and M. Birattari (2016) <doi:10.1016/j.orp.2016.09.002>.
Version: 4.4.3
Authors@R: c(person("Manuel", "López-Ibáñez", role = c("aut", "cre"),
                              email = "manuel.lopez-ibanez@manchester.ac.uk",
                              comment = c(ORCID = "0000-0001-9974-1295")),
             person("Jérémie", "Dubois-Lacoste", role = "aut"),
             person("Leslie", "Pérez Cáceres", role = "aut"),
             person("Thomas", "Stützle", role = "aut"),
             person("Mauro", "Birattari", role = "aut"),
             person("Eric", "Yuan", role = "ctb"),
             person("Prasanna", "Balaprakash", role = "ctb"),
             person("Nguyen", "Dang", role = "ctb", email = "nttd@st-andrews.ac.uk"))
Depends: R (>= 4.0.0)
Imports: R6, codetools, compiler, data.table (>= 1.15.0), fs,
        matrixStats (>= 1.4.1), spacefillr, stats, utils, withr
Suggests: Rmpi (>= 0.6.0), highr, knitr, parallel, testthat (>= 3.1.8)
VignetteBuilder: knitr
License: GPL (>= 2)
URL: https://mlopez-ibanez.github.io/irace/,
        https://github.com/MLopez-Ibanez/irace
BugReports: https://github.com/MLopez-Ibanez/irace/issues
ByteCompile: yes
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: GNU make
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-04-23 13:43:45 UTC; manu
Author: Manuel López-Ibáñez [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9974-1295>),
  Jérémie Dubois-Lacoste [aut],
  Leslie Pérez Cáceres [aut],
  Thomas Stützle [aut],
  Mauro Birattari [aut],
  Eric Yuan [ctb],
  Prasanna Balaprakash [ctb],
  Nguyen Dang [ctb]
Maintainer: Manuel López-Ibáñez <manuel.lopez-ibanez@manchester.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-23 14:20:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-04-23 17:22:02 UTC; unix

Package: irace
Type: Package
Title: Iterated Racing for Automatic Algorithm Configuration
Description: Iterated race is an extension of the Iterated F-race method for
             the automatic configuration of optimization algorithms, that is,
             (offline) tuning their parameters by finding the most appropriate
             settings given a set of instances of an optimization problem.
             M. López-Ibáñez, J. Dubois-Lacoste, L. Pérez Cáceres, T. Stützle,
             and M. Birattari (2016) <doi:10.1016/j.orp.2016.09.002>.
Version: 4.4.3
Authors@R: c(person("Manuel", "López-Ibáñez", role = c("aut", "cre"),
                              email = "manuel.lopez-ibanez@manchester.ac.uk",
                              comment = c(ORCID = "0000-0001-9974-1295")),
             person("Jérémie", "Dubois-Lacoste", role = "aut"),
             person("Leslie", "Pérez Cáceres", role = "aut"),
             person("Thomas", "Stützle", role = "aut"),
             person("Mauro", "Birattari", role = "aut"),
             person("Eric", "Yuan", role = "ctb"),
             person("Prasanna", "Balaprakash", role = "ctb"),
             person("Nguyen", "Dang", role = "ctb", email = "nttd@st-andrews.ac.uk"))
Depends: R (>= 4.0.0)
Imports: R6, codetools, compiler, data.table (>= 1.15.0), fs,
        matrixStats (>= 1.4.1), spacefillr, stats, utils, withr
Suggests: Rmpi (>= 0.6.0), highr, knitr, parallel, testthat (>= 3.1.8)
VignetteBuilder: knitr
License: GPL (>= 2)
URL: https://mlopez-ibanez.github.io/irace/,
        https://github.com/MLopez-Ibanez/irace
BugReports: https://github.com/MLopez-Ibanez/irace/issues
ByteCompile: yes
Encoding: UTF-8
RoxygenNote: 7.3.3
SystemRequirements: GNU make
Config/testthat/edition: 3
NeedsCompilation: yes
Packaged: 2026-04-23 13:43:45 UTC; manu
Author: Manuel López-Ibáñez [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9974-1295>),
  Jérémie Dubois-Lacoste [aut],
  Leslie Pérez Cáceres [aut],
  Thomas Stützle [aut],
  Mauro Birattari [aut],
  Eric Yuan [ctb],
  Prasanna Balaprakash [ctb],
  Nguyen Dang [ctb]
Maintainer: Manuel López-Ibáñez <manuel.lopez-ibanez@manchester.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-23 14:20:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 20:27:06 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(param_quantile,ParamCat)
S3method(param_quantile,ParamInt)
S3method(param_quantile,ParamOrd)
S3method(param_quantile,ParamReal)
S3method(print,ParameterSpace)
S3method(sample_model,ParamCat)
S3method(sample_model,ParamInt)
S3method(sample_model,ParamOrd)
S3method(sample_model,ParamReal)
S3method(sample_unif,ParamCat)
S3method(sample_unif,ParamInt)
S3method(sample_unif,ParamOrd)
S3method(sample_unif,ParamReal)
export(ablation)
export(ablation_cmdline)
export(buildCommandLine)
export(checkIraceScenario)
export(checkParameters)
export(checkScenario)
export(check_output_target_runner)
export(configurations_print)
export(configurations_print_command)
export(defaultScenario)
export(getConfigurationById)
export(getConfigurationByIteration)
export(getFinalElites)
export(get_instanceID_seed_pairs)
export(get_random_seed)
export(has_testing_data)
export(irace)
export(irace.cmdline)
export(irace_cmdline)
export(irace_license)
export(irace_main)
export(irace_summarise)
export(irace_version)
export(multi_irace)
export(param_cat)
export(param_int)
export(param_ord)
export(param_real)
export(parametersNew)
export(path_rel2abs)
export(plotAblation)
export(printParameters)
export(printScenario)
export(psRace)
export(readConfigurationsFile)
export(readParameters)
export(readScenario)
export(read_ablogfile)
export(read_logfile)
export(read_pcs_file)
export(removeConfigurationsMetaData)
export(restore_random_seed)
export(scenario_update_paths)
export(set_random_seed)
export(target_evaluator_default)
export(target_runner_default)
export(testConfigurations)
export(testing_fromfile)
export(testing_fromlog)
import(data.table)
import(matrixStats)
import(stats)
import(withr)
importFrom(R6,R6Class)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,boxplot)
importFrom(graphics,bxp)
importFrom(graphics,grid)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(graphics,strwidth)
importFrom(spacefillr,generate_sobol_set)
importFrom(utils,str)

**irace**: Iterated Racing for Automatic Algorithm Configuration
================================================================

<!-- badges: start -->
[![CRAN version](https://www.r-pkg.org/badges/version-last-release/irace)](https://cran.r-project.org/package=irace)
[![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/irace)](https://CRAN.R-project.org/package=irace)
[![CRAN status](https://badges.cranchecks.info/worst/irace.svg)](https://cran.r-project.org/web/checks/check_results_irace.html)
[![R build
status](https://github.com/MLopez-Ibanez/irace/workflows/R-CMD-check/badge.svg)](https://github.com/MLopez-Ibanez/irace/actions)
[![Codecov test coverage](https://codecov.io/gh/MLopez-Ibanez/irace/branch/master/graph/badge.svg)](https://app.codecov.io/gh/MLopez-Ibanez/irace?branch=master)
[![r-universe version](https://mlopez-ibanez.r-universe.dev/badges/irace)](https://mlopez-ibanez.r-universe.dev/irace)
[![r-universe build status](https://github.com/r-universe/mlopez-ibanez/actions/workflows/build.yml/badge.svg)](https://github.com/r-universe/mlopez-ibanez/actions/workflows/build.yml)
<!-- badges: end -->

[ [**Homepage**][irace-homepage] ] [ [**User Guide (PDF)**][user-guide] ] [ [**Tutorial**](https://lopez-ibanez.eu/2024-redheur/) ]

**Maintainers:** [Manuel López-Ibáñez](https://lopez-ibanez.eu/), Leslie Pérez Cáceres

**Creators:** [Manuel López-Ibáñez](https://lopez-ibanez.eu/), Jérémie Dubois-Lacoste

**Contributors:** Jérémie Dubois-Lacoste, Thomas Stützle, Mauro Birattari, Eric
  Yuan and Prasanna Balaprakash.

**Contact:** <https://groups.google.com/d/forum/irace-package>

---------------------------------------

Introduction
------------

The **irace** package implements the Iterated Race method, which is a
generalization of the Iterated F-race method for the automatic configuration of
optimization algorithms, that is, the tuning of their parameters by finding the
most appropriate settings given a set of instances of an optimization problem.
It builds upon the race package by Birattari and it is implemented in R.

You may also find the [**iraceplot**](https://auto-optimization.github.io/iraceplot/) package useful for analyzing the output of irace.

**Keywords:** automatic configuration, offline tuning, parameter tuning, racing, F-race.

**Relevant literature:**

 1. M. López-Ibáñez, J. Dubois-Lacoste, L. Pérez Cáceres, T. Stützle, and
    M. Birattari. [The irace package: Iterated Racing for Automatic Algorithm Configuration](http://dx.doi.org/10.1016/j.orp.2016.09.002).
*Operations Research Perspectives*, 3:43–58, 2016.<br>
    [ [bibtex](https://iridia-ulb.github.io/references/index_bib.html#LopDubPerStuBir2016irace)
    | doi: [10.1016/j.orp.2016.09.002](http://dx.doi.org/10.1016/j.orp.2016.09.002) ]

 2. Leslie Perez Cáceres, Manuel López-Ibáñez, Holger Hoos, and Thomas Stützle. [An Experimental Study of Adaptive Capping in irace](https://doi.org/10.1007/978-3-319-69404-7_17). *Learning and Intelligent Optimization. LION 2017*. Lecture Notes in Computer Science, vol 10556. Springer, Cham.<br>
    [ [bibtex](https://iridia-ulb.github.io/references/index_bib.html#PerLopHooStu2017:lion)
    | doi: [10.1007/978-3-319-69404-7_17](https://doi.org/10.1007/978-3-319-69404-7_17) ]

 3. Thomas Stützle and Manuel López-Ibáñez. [Tutorial: Automated algorithm
    configuration and design](https://doi.org/10.1145/3449726.3461404). GECCO
    '21: Proceedings of the Genetic and Evolutionary Computation Conference
    Companion, July 2021.
    [doi:10.1145/3449726.3461404](https://doi.org/10.1145/3449726.3461404)


### Requisites ###

 * R (<https://www.r-project.org>) is required for running irace, but
   you don't need to know the R language to use it.


### User guide ###

A complete [user guide][user-guide]
comes with the package. You can access it online or, after installing the irace
package, invoking from the R console the following command:

```R
        vignette("irace-package")
```

The following is a quick-start guide. The user guide gives more detailed
instructions.

Quick Start
-----------

1. Install R (with your favourite package manager, and see more details below).
2. Install irace. This command works on Bash shell (Linux and MacOS) and Powershell (Windows) with R added to PATH (see detailed instructions below).
```bash
    Rscript -e "install.packages('irace', repos='https://cloud.r-project.org')"
```

3. Add irace to the `PATH` environment variable. This command works in Bash shell (Linux and MacOS). For Windows user, this step is unfortunately more involved, so please see more detailed instructions below.
```bash
    export PATH="$(Rscript -e "cat(paste0(system.file(package='irace', 'bin', mustWork=TRUE), ':'))" 2> /dev/null)${PATH}"
```
Consider adding this line to your `~/.bashrc`, `~/.zshrc`, or `~/.profile` for it to persist between sessions.

4. You can open the user guide with the following command. This command works on Bash shell (Linux and MacOS) and and Powershell (Windows) with R added to PATH (see detailed instructions below).
```bash
    Rscript -e "vignette('irace-package')"
```

Installing R
------------

The official instructions are available at
<https://cran.r-project.org/doc/manuals/r-release/R-admin.html>. We give below
a quick R installation guide that will work in most cases.

### GNU/Linux ###

You should install R from your package manager. On a Debian/Ubuntu system, you will run in the Bash shell:
```bash
    sudo apt-get install r-base
```

Once R is installed, you can launch R from the Terminal and from the R
prompt install the irace package. See instructions below.


### OS X ###

You can install R directly from a CRAN mirror (<https://cran.r-project.org/bin/macosx/>).

Alternatively, if you use homebrew, you can just run from the Terminal (Bash shell):
```bash
    brew install --cask r
```

(Using `brew install r` is not recommended because that will build R from source and you will not be able to use any CRAN binary, possibly resulting in annoying build failures).

Once R is installed, you can launch R from the Terminal (or from your
Applications), and from the R prompt install the irace package. See
instructions below.

### Windows ###

You can install R from a CRAN mirror
(<https://cran.r-project.org/bin/windows/>). Once R is installed, you can
launch the R console and install the irace package from it. See instructions
below.

In addition to using the R console, it might be very useful to add R to PATH so you can run R commands in CMD or Powershell. Usually, R is installed in `C:\Program Files\R\R-4.1.3` (the version number depends on your installation).

You should add the following line [to PATH](https://www.architectryan.com/2018/03/17/add-to-the-path-on-windows-10/) (if you want to use the 64-bit version)
```
C:\Program Files\R\R-4.1.3\bin\x64
```

Or, if you are on a 32-bit version
```
C:\Program Files\R\R-4.1.3\bin\i386
```

Installing the irace package
----------------------------

There are two methods for installing the
[irace][irace-homepage] R package on your computer:

1. Install within R (automatic download). Run the following from the R console:
```R
        install.packages("irace", repos = "https://cloud.r-project.org")
```
   and test the installation with:
```R
        library(irace)
        irace_cmdline("--version")
        q()
```

2. Manually
   [download the package from CRAN](https://cran.r-project.org/package=irace)
   and invoke at the Bash shell, Terminal or Powershell:
```bash
        R CMD INSTALL <package>
```
   where `<package>` is one of the three versions available: `.tar.gz`
   (Unix/BSD/GNU/Linux), `.tgz` (MacOS X), or `.zip` (Windows).


If the package fails to install because of insufficient permissions,
you need to force a *local installation* by typing in the Bash shell:
```bash
    mkdir ~/R
    R CMD INSTALL --library=~/R irace.tar.gz
    export R_LIBS=~/R:${R_LIBS}
```

Once installed, test that it is working by typing in the R console (not in the bash shell):
```R
    library(irace)
    irace_cmdline("--version")
    cat(system.file(package="irace", "bin", mustWork=TRUE), "\n")
```
The last command gives you the installation folder of `irace`, for example, `/home/user/R/irace/bin`.

### GNU/Linux and OS X ###

Save the installation directory of `irace` to a variable, and add it to your
`.bash_profile`, `.bashrc` or `.profile`:

```bash
    export IRACE_HOME=/home/user/R/irace/bin/ # Path given by system.file(package="irace", "bin", mustWork=TRUE)
    export PATH=${IRACE_HOME}:$PATH
    # export R_LIBS=~/R:${R_LIBS} # Only if local installation was forced
```

After adding this and opening a new terminal, you should be able to
invoke `irace` as follows:

```bash
    irace --help
```


### Windows ###

You can find out where the irace binary is installed by running the following in Powershell or CMD:

```Powershell
    C:\> Rscript -e "cat(gsub('/', '\\\\', system.file(package='irace', 'bin', 'x64', mustWork=TRUE)))"
```

It will output a path, such as `C:\Program Files\R\R-4.1.3\library\irace\bin\x64` (replace `x64` with `i386` if you are on a 32-bit system), which can you [add to PATH](https://www.architectryan.com/2018/03/17/add-to-the-path-on-windows-10/).

Then running the following should work:
```Powershell
    C:\> irace --help
```

You can also launch irace by opening the R console and executing:

```R
    library(irace)
    irace_cmdline("--help")
```

### GitHub (Development version) ###

If you wish to try the development version, you can install it by executing the
following command within the R console:

```R
    install.packages('irace', repos = c('https://mlopez-ibanez.r-universe.dev', 'https://cloud.r-project.org'))
```

### Python ###

You can use the irace R package from Python using `rpy2`. There are actually two different implementations of this idea:  [iracepy](https://github.com/auto-optimization/iracepy) and [iracepy-tiny](https://github.com/Saethox/iracepy-tiny).

### Rust ###

You can use `irace` from Rust by using [irace-rs](https://github.com/Saethox/irace-rs).

Usage
-----

1. Create a directory for storing the tuning scenario setup (Bash shell):
```bash
        mkdir ./tuning
        cd ./tuning
```

2. Initialize your tuning directory with template config files (Bash shell):
```bash
        $IRACE_HOME/bin/irace --init
```

3. Modify the generated files following the instructions found within each file. In particular,
    * The scripts `target-runner` and `target-evaluator` (if you need it at all)
      should be executable. The output of `target-runner` (or
      `target-evaluator` if you use a separate evaluation step) is minimized by
      default. If you wish to maximize it, just multiply the value by `-1`
      within the script.
    * In `scenario.txt`, uncomment and assign only the parameters for which
      you need a value different than the default one. For example, you may need to set
      `trainInstancesDir="./Instances/"`.

    There are examples in `$IRACE_HOME/examples/`.

4. Put the instances in `./tuning/Instances/`. In addition, you can
   create a file that specifies which instances from that directory
   should be run and which instance-specific parameters to use. See
   `scenario.txt` and `instances-list.txt` for examples. The command
   irace will not attempt to create the execution directory (`execDir`),
   so it must exist before calling irace. The default `execDir` is the
   current directory.

5. Calling the command in the Bash shell:
```bash
        cd ./tuning/ && $IRACE_HOME/bin/irace
```
   performs one run of Iterated Race. See the output of `irace --help` for
   additional irace parameters. Command-line parameters override the
   scenario setup specified in the `scenario.txt` file.


### Many tuning runs in parallel ###

For executing several repetitions of irace in parallel, call the
program `parallel-irace` from the Bash shell:

```bash
    cd ./tuning/ && $IRACE_HOME/bin/parallel-irace N
```

where N is the number of repetitions. By default, the execution directory of
each run of irace will be set to `./execdir-dd`, where `dd` is a number padded
with zeroes.

**Be careful**, `parallel-irace` will create these directories from
scratch, deleting them first if they already exist.

Check the help of `parallel-irace` by running it without parameters.


### Parallelize one tuning ###

A single run of irace can be done much faster by executing the calls
to `targetRunner` (the runs of the algorithm being tuned) in
parallel. See the [user guide][user-guide] for the details.

License
-------

**irace** is Copyright (C) 2010-2020 Manuel López-Ibáñez, Jérémie Dubois-Lacoste
and Leslie Pérez-Cáceres.

This program is free software (software libre); you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the [GNU General Public License](https://cran.r-project.org/web/licenses/GPL-3) for more details.

**IMPORTANT NOTE:** Please be aware that the fact that this program is released as
Free Software does not excuse you from scientific propriety, which obligates
you to give appropriate credit! If you write a scientific paper describing
research that made substantive use of this program, it is your obligation as a
scientist to (a) mention the fashion in which this software was used in the
Methods section; (b) mention the algorithm in the References section. The
appropriate citation is:

 * Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres, Thomas
  Stützle, and Mauro Birattari.  [**The irace package: Iterated Racing for
  Automatic Algorithm
  Configuration.**](http://dx.doi.org/10.1016/j.orp.2016.09.002) _Operations
  Research Perspectives_, 2016. doi: 10.1016/j.orp.2016.09.002

The **irace** package uses code under the GPL from the [race
package](https://CRAN.R-project.org/package=race) is Copyright (C) 2003 Mauro
Birattari.


Building an irace standalone container
--------------------------------------

Thanks to [Singularity](https://sylabs.io/singularity/), you can build a
standalone container of `irace` using the file `irace.sindef` which is
available in the directory `inst/` in the source tarball and github repository
or, after installing the irace R package, in the installation directory given
by the R expression `system.file(package="irace")`. After installing
SingularityCE, the container may be build using:

    sudo singularity build irace.sindef irace.sif

and run with:

    singularity run irace.sif <arguments>



Frequently Asked Questions
--------------------------

The [user guide][user-guide] contains a list of frequently asked questions.


[user-guide]: https://mlopez-ibanez.github.io/irace/irace-package.pdf
[irace-homepage]: https://mlopez-ibanez.github.io/irace/

<!--
Local Variables:
mode: markdown
End:
-->

**irace**: Iterated Racing for Automatic Algorithm Configuration
================================================================

<!-- badges: start -->
[![CRAN version](https://www.r-pkg.org/badges/version-last-release/irace)](https://cran.r-project.org/package=irace)
[![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/irace)](https://CRAN.R-project.org/package=irace)
[![CRAN status](https://badges.cranchecks.info/worst/irace.svg)](https://cran.r-project.org/web/checks/check_results_irace.html)
[![R build
status](https://github.com/MLopez-Ibanez/irace/workflows/R-CMD-check/badge.svg)](https://github.com/MLopez-Ibanez/irace/actions) 
[![Codecov test coverage](https://codecov.io/gh/MLopez-Ibanez/irace/branch/master/graph/badge.svg)](https://app.codecov.io/gh/MLopez-Ibanez/irace?branch=master)
[![r-universe version](https://mlopez-ibanez.r-universe.dev/badges/irace)](https://mlopez-ibanez.r-universe.dev/irace)
[![r-universe build status](https://github.com/r-universe/mlopez-ibanez/actions/workflows/build.yml/badge.svg)](https://github.com/r-universe/mlopez-ibanez/actions/workflows/build.yml)
<!-- badges: end -->

[ [**Homepage**](https://mlopez-ibanez.github.io/irace/) ] [ [**User Guide (PDF)**](https://cran.r-project.org/package=irace/vignettes/irace-package.pdf) ]

**Maintainers:** [Manuel López-Ibáñez](https://lopez-ibanez.eu/), Leslie Pérez Cáceres

**Creators:** [Manuel López-Ibáñez](https://lopez-ibanez.eu/), Jérémie Dubois-Lacoste
  
**Contributors:** Jérémie Dubois-Lacoste, Thomas Stützle, Mauro Birattari, Eric
  Yuan and Prasanna Balaprakash.

**Contact:** <https://groups.google.com/d/forum/irace-package>

---------------------------------------

Introduction
------------

The **irace** package implements the Iterated Race method, which is a
generalization of the Iterated F-race method for the automatic configuration of
optimization algorithms, that is, the tuning of their parameters by finding the
most appropriate settings given a set of instances of an optimization problem.
It builds upon the race package by Birattari and it is implemented in R.

You may also find the [**iraceplot**](https://auto-optimization.github.io/iraceplot/) package useful for analyzing the output of irace.

**Keywords:** automatic configuration, offline tuning, parameter tuning, racing, F-race.

**Relevant literature:**

 1. M. López-Ibáñez, J. Dubois-Lacoste, L. Pérez Cáceres, T. Stützle, and
    M. Birattari. [The irace package: Iterated Racing for Automatic Algorithm Configuration](http://dx.doi.org/10.1016/j.orp.2016.09.002).
*Operations Research Perspectives*, 3:43–58, 2016.<br>
    [ [bibtex](https://iridia-ulb.github.io/references/index_bib.html#LopDubPerStuBir2016irace)
    | doi: [10.1016/j.orp.2016.09.002](http://dx.doi.org/10.1016/j.orp.2016.09.002) ]

 2. Leslie Perez Cáceres, Manuel López-Ibáñez, Holger Hoos, and Thomas Stützle. [An Experimental Study of Adaptive Capping in irace](https://doi.org/10.1007/978-3-319-69404-7_17). *Learning and Intelligent Optimization. LION 2017*. Lecture Notes in Computer Science, vol 10556. Springer, Cham.<br>
    [ [bibtex](https://iridia-ulb.github.io/references/index_bib.html#PerLopHooStu2017:lion)
    | doi: [10.1007/978-3-319-69404-7_17](https://doi.org/10.1007/978-3-319-69404-7_17) ]

 3. Thomas Stützle and Manuel López-Ibáñez. [Tutorial: Automated algorithm
    configuration and design](https://doi.org/10.1145/3449726.3461404). GECCO
    '21: Proceedings of the Genetic and Evolutionary Computation Conference
    Companion, July 2021.
    [doi:10.1145/3449726.3461404](https://doi.org/10.1145/3449726.3461404)


### Requisites ###

 * R (<https://www.r-project.org>) is required for running irace, but
   you don't need to know the R language to use it.


### User guide ###

A complete [user guide](https://cran.r-project.org/package=irace/vignettes/irace-package.pdf)
comes with the package. You can access it online or, after installing the irace
package, invoking from the R console the following command:

```R
        vignette("irace-package")
```

The following is a quick-start guide. The user guide gives more detailed
instructions.

Quick Start
-----------

1. Install R (with your favourite package manager, and see more details below).
2. Install irace. This command works on Bash shell (Linux and MacOS) and Powershell (Windows) with R added to PATH (see detailed instructions below).
```bash
    Rscript -e "install.packages('irace', repos='https://cloud.r-project.org')"
```

3. Add irace to the `PATH` environment variable. This command works in Bash shell (Linux and MacOS). For Windows user, this step is unfortunately more involved, so please see more detailed instructions below.
```bash
    export PATH="$(Rscript -e "cat(paste0(system.file(package='irace', 'bin', mustWork=TRUE), ':'))" 2> /dev/null)${PATH}"
```
Consider adding this line to your `~/.bashrc`, `~/.zshrc`, or `~/.profile` for it to persist between sessions. 

4. You can open the user guide with the following command. This command works on Bash shell (Linux and MacOS) and and Powershell (Windows) with R added to PATH (see detailed instructions below).
```bash
    Rscript -e "vignette('irace-package')"
```

Installing R
------------

The official instructions are available at
<https://cran.r-project.org/doc/manuals/r-release/R-admin.html>. We give below
a quick R installation guide that will work in most cases.

### GNU/Linux ###

You should install R from your package manager. On a Debian/Ubuntu system, you will run in the Bash shell:
```bash
    sudo apt-get install r-base
```

Once R is installed, you can launch R from the Terminal and from the R
prompt install the irace package. See instructions below.


### OS X ###

You can install R directly from a CRAN mirror (<https://cran.r-project.org/bin/macosx/>).

Alternatively, if you use homebrew, you can just run from the Terminal (Bash shell):
```bash
    brew install --cask r
```

(Using `brew install r` is not recommended because that will build R from source and you will not be able to use any CRAN binary, possibly resulting in annoying build failures). 

Once R is installed, you can launch R from the Terminal (or from your
Applications), and from the R prompt install the irace package. See
instructions below.

### Windows ###

You can install R from a CRAN mirror
(<https://cran.r-project.org/bin/windows/>). Once R is installed, you can
launch the R console and install the irace package from it. See instructions
below.

In addition to using the R console, it might be very useful to add R to PATH so you can run R commands in CMD or Powershell. Usually, R is installed in `C:\Program Files\R\R-4.1.3` (the version number depends on your installation).

You should add the following line [to PATH](https://www.architectryan.com/2018/03/17/add-to-the-path-on-windows-10/) (if you want to use the 64-bit version)
```
C:\Program Files\R\R-4.1.3\bin\x64
```

Or, if you are on a 32-bit version 
```
C:\Program Files\R\R-4.1.3\bin\i386
```

Installing the irace package
----------------------------

There are two methods for installing the
[irace](https://mlopez-ibanez.github.io/irace/) R package on your computer:

1. Install within R (automatic download). Run the following from the R console:
```R
        install.packages("irace", repos = "https://cloud.r-project.org")
```
   and test the installation with:
```R
        library(irace)
        irace_cmdline("--version")
        q()
```

2. Manually
   [download the package from CRAN](https://cran.r-project.org/package=irace)
   and invoke at the Bash shell, Terminal or Powershell:
```bash
        R CMD INSTALL <package>
```
   where `<package>` is one of the three versions available: `.tar.gz`
   (Unix/BSD/GNU/Linux), `.tgz` (MacOS X), or `.zip` (Windows).


If the package fails to install because of insufficient permissions,
you need to force a *local installation* by typing in the Bash shell:
```bash
    mkdir ~/R
    R CMD INSTALL --library=~/R irace.tar.gz
    export R_LIBS=~/R:${R_LIBS}
```

Once installed, test that it is working by typing in the R console (not in the bash shell):
```R
    library(irace)
    irace_cmdline("--version")
    cat(system.file(package="irace", "bin", mustWork=TRUE), "\n")
```
The last command gives you the installation folder of `irace`, for example, `/home/user/R/irace/bin`.

### GNU/Linux and OS X ###

Save the installation directory of `irace` to a variable, and add it to your
`.bash_profile`, `.bashrc` or `.profile`:

```bash
    export IRACE_HOME=/home/user/R/irace/bin/ # Path given by system.file(package="irace", "bin", mustWork=TRUE)
    export PATH=${IRACE_HOME}:$PATH
    # export R_LIBS=~/R:${R_LIBS} # Only if local installation was forced
```

After adding this and opening a new terminal, you should be able to
invoke `irace` as follows:

```bash
    irace --help
```


### Windows ###

You can find out where the irace binary is installed by running the following in Powershell or CMD:

```Powershell
    C:\> Rscript -e "cat(gsub('/', '\\\\', system.file(package='irace', 'bin', 'x64', mustWork=TRUE)))"
```

It will output a path, such as `C:\Program Files\R\R-4.1.3\library\irace\bin\x64` (replace `x64` with `i386` if you are on a 32-bit system), which can you [add to PATH](https://www.architectryan.com/2018/03/17/add-to-the-path-on-windows-10/). 

Then running the following should work:
```Powershell 
    C:\> irace --help
```

You can also launch irace by opening the R console and executing:

```R
    library(irace)
    irace_cmdline("--help")
```

### GitHub (Development version) ###

If you wish to try the development version, you can install it by executing the
following command within the R console:

```R
    install.packages('irace', repos = c('https://mlopez-ibanez.r-universe.dev', 'https://cloud.r-project.org'))
```

### Python ###

You can use the irace R package from Python using `rpy2`. There are actually two different implementations of this idea:  [iracepy](https://github.com/auto-optimization/iracepy) and [iracepy-tiny](https://github.com/Saethox/iracepy-tiny).

### Rust ###

You can use `irace` from Rust by using [irace-rs](https://github.com/Saethox/irace-rs).

Usage
-----

1. Create a directory for storing the tuning scenario setup (Bash shell):
```bash
        mkdir ./tuning
        cd ./tuning
```

2. Initialize your tuning directory with template config files (Bash shell):
```bash
        $IRACE_HOME/bin/irace --init
```

3. Modify the generated files following the instructions found within each file. In particular,
    * The scripts `target-runner` and `target-evaluator` (if you need it at all)
      should be executable. The output of `target-runner` (or
      `target-evaluator` if you use a separate evaluation step) is minimized by
      default. If you wish to maximize it, just multiply the value by `-1`
      within the script.
    * In `scenario.txt`, uncomment and assign only the parameters for which
      you need a value different than the default one. For example, you may need to set
      `trainInstancesDir="./Instances/"`.
   
    There are examples in `$IRACE_HOME/examples/`.

4. Put the instances in `./tuning/Instances/`. In addition, you can
   create a file that specifies which instances from that directory
   should be run and which instance-specific parameters to use. See
   `scenario.txt` and `instances-list.txt` for examples. The command
   irace will not attempt to create the execution directory (`execDir`),
   so it must exist before calling irace. The default `execDir` is the
   current directory.

5. Calling the command in the Bash shell:
```bash
        cd ./tuning/ && $IRACE_HOME/bin/irace
```
   performs one run of Iterated Race. See the output of `irace --help` for
   additional irace parameters. Command-line parameters override the
   scenario setup specified in the `scenario.txt` file.


### Many tuning runs in parallel ###

For executing several repetitions of irace in parallel, call the
program `parallel-irace` from the Bash shell:

```bash
    cd ./tuning/ && $IRACE_HOME/bin/parallel-irace N
```

where N is the number of repetitions. By default, the execution directory of
each run of irace will be set to `./execdir-dd`, where `dd` is a number padded
with zeroes.

**Be careful**, `parallel-irace` will create these directories from
scratch, deleting them first if they already exist.

Check the help of `parallel-irace` by running it without parameters.


### Parallelize one tuning ###

A single run of irace can be done much faster by executing the calls
to `targetRunner` (the runs of the algorithm being tuned) in
parallel. See the [user guide](https://cran.r-project.org/package=irace/vignettes/irace-package.pdf) for the details.

License
-------

**irace** is Copyright (C) 2010-2020 Manuel López-Ibáñez, Jérémie Dubois-Lacoste
and Leslie Pérez-Cáceres.

This program is free software (software libre); you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your option)
any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the [GNU General Public License](https://cran.r-project.org/web/licenses/GPL-3) for more details.

**IMPORTANT NOTE:** Please be aware that the fact that this program is released as
Free Software does not excuse you from scientific propriety, which obligates
you to give appropriate credit! If you write a scientific paper describing
research that made substantive use of this program, it is your obligation as a
scientist to (a) mention the fashion in which this software was used in the
Methods section; (b) mention the algorithm in the References section. The
appropriate citation is:

 * Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres, Thomas
  Stützle, and Mauro Birattari.  [**The irace package: Iterated Racing for
  Automatic Algorithm
  Configuration.**](http://dx.doi.org/10.1016/j.orp.2016.09.002) _Operations
  Research Perspectives_, 2016. doi: 10.1016/j.orp.2016.09.002

The **irace** package uses code under the GPL from the [race
package](https://CRAN.R-project.org/package=race) is Copyright (C) 2003 Mauro
Birattari.


Building an irace standalone container
--------------------------------------

Thanks to [Singularity](https://sylabs.io/singularity/), you can build a
standalone container of `irace` using the file `irace.sindef` which is
available in the directory `inst/` in the source tarball and github repository
or, after installing the irace R package, in the installation directory given
by the R expression `system.file(package="irace")`. After installing
SingularityCE, the container may be build using:

    sudo singularity build irace.sindef irace.sif
    
and run with:

    singularity run irace.sif <arguments>
    


Frequently Asked Questions
--------------------------

The
[user guide](https://cran.r-project.org/package=irace/vignettes/irace-package.pdf)
contains a list of frequently asked questions.

<!--
Local Variables:
mode: markdown
End:
-->

Additional examples of scenarios can be found in:

Capping Methods for the Automatic Configuration of Optimization Algorithms
Marcelo de Souza, Marcus Ritt and Manuel López-Ibáñez
https://github.com/souzamarcelo/supp-cor-capopt

 * ACOTSP: ant colony optimization algorithms for the symmetric traveling salesperson problem.
 * HEACOL: hybrid evolutionary algorithm for graph coloring.
 * TSBPP: tabu search for the bin packing problem.
 * HHBQP: hybrid heuristic for unconstrained binary quadratic programming.
 * LKH: a heuristic algorithm for solving the symmetric traveling salesperson problem.
 * SCIP: an exact solver for mixed integer programs for solving the
   combinatorial auction winner determination problem.


https://aclib.net/
Frank Hutter, Manuel López-Ibáñez, Chris Fawcett, Marius Thomas Lindauer,
Holger H. Hoos, Kevin Leyton-Brown, and Thomas Stützle. AClib: a Benchmark
Library for Algorithm Configuration. In P. M. Pardalos, M. G. C. Resende,
C. Vogiatzis, and J. L. Walteros, editors, Learning and Intelligent
Optimization, 8th International Conference, LION 8, volume 8426 of Lecture
Notes in Computer Science, pages 36–40. Springer, 2014.

Example configuration scenario: Spear
=====================================

This example scenario was inspired by ParamILS. The Spear program can
be downloaded from: 

 http://www.cs.ubc.ca/labs/beta/Projects/ParamILS/algorithms.html

WARNING: Please note that no source code is available and we do not know what
this executable is actually doing. We do NOT recommend running this executable
in your computer.  If Spear doesn't work in your computer, there is nothing we
can do to fix it.  Unless you absolutely must run Spear, we suggest to use a
different scenario for which source code is available.

* Tuning instances:

 http://www.cs.ubc.ca/labs/beta/Projects/ParamILS/benchmark_instances/SpearSWV/SWV-scrambled-first302.tar.gz

* Testing instances:

 http://www.cs.ubc.ca/labs/beta/Projects/ParamILS/benchmark_instances/SpearSWV/SWV-scrambled-last302.tar.gz


Two parameter files are provided: 

 * parameters-cat.txt only contains categorical parameters.

 * parameters-mixed.txt contains a mix of numerical and categorical parameters.

These are example hooks to tune a multi-objective algorithm (using the
hypervolume indicator). You need to set targetRunner and targetEvaluator,
either in the scenario file of irace (usually scenario.txt), or in
the command-line (see irace --help), so irace is able to find these
scripts.

Update target-runner with the path, name and fixed parameters of your
program.

Update target-evaluator with the path and parameters for computing the
hypervolume. The example uses the hypervolume code from
< http://lopez-ibanez.eu/hypervolume >. Additional
transformations of the input data may be applied before computing the
hypervolume, such as normalization.

Example configuration scenario: MOACO
=====================================

This example scenario shows how irace can be used to tune a
multi-objective optimization framework. It uses the multi-objective
ant colony optimization (MOACO) framework proposed here for the bi-objective
traveling salesperson problem (bTSP):

    Manuel López-Ibáñez and Thomas Stützle. The Automatic Design of
    Multi-Objective Ant Colony Optimization Algorithms. IEEE
    Transactions on Evolutionary Computation, 2012. 
    doi:10.1109/TEVC.2011.2182651
    (PDF preprint) https://iridia.ulb.ac.be/IridiaTrSeries/link/IridiaTr2011-003.pdf

The MOACO framework software can be download from:

    http://lopez-ibanez.eu/moaco

Instances of the bTSP can be generated by concatenating two single-objective TSP
instances:

    $ cat euclidA100.tsp euclidB100.tsp > euclidAB100.tsp

TSP instances and more information is available at:

https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/index.html

Example configuration scenario: ACOTSP
======================================

This example scenario shows how irace can be used to tune
an optimization algorithm.

ACOTSP is a software that implements various Ant Colony algorithms
for the symmetric Traveling Salesman Problem.

We will use the version provided here:

https://github.com/MLopez-Ibanez/ACOTSPQAP/archive/refs/heads/master.zip

TSP instances and more information is available at:

https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/index.html

You should use the instances from this file:

https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/scenarios/acotsp/instances.tar.gz

Example configuration scenario: ACOTSP
=====================================

This example scenario shows how irace can be used to tune
an optimization algorithm.

ACOTSP is a software that implements various Ant Colony algorithms
for the symmetric Traveling Salesman Problem. 
The ACOTSP software is available at:

http://www.aco-metaheuristic.org/aco-code/

TSP instances and more information is available at:

https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/index.html

Additional examples of scenarios can be found in:

Capping Methods for the Automatic Configuration of Optimization Algorithms
Marcelo de Souza, Marcus Ritt and Manuel López-Ibáñez
https://github.com/souzamarcelo/supp-cor-capopt

 * ACOTSP: ant colony optimization algorithms for the symmetric traveling salesperson problem.
 * HEACOL: hybrid evolutionary algorithm for graph coloring.
 * TSBPP: tabu search for the bin packing problem.
 * HHBQP: hybrid heuristic for unconstrained binary quadratic programming.
 * LKH: a heuristic algorithm for solving the symmetric traveling salesperson problem.
 * SCIP: an exact solver for mixed integer programs for solving the
   combinatorial auction winner determination problem.
   

http://aclib.net/
Frank Hutter, Manuel López-Ibáñez, Chris Fawcett, Marius Thomas Lindauer,
Holger H. Hoos, Kevin Leyton-Brown, and Thomas Stützle. AClib: a Benchmark
Library for Algorithm Configuration. In P. M. Pardalos, M. G. C. Resende,
C. Vogiatzis, and J. L. Walteros, editors, Learning and Intelligent
Optimization, 8th International Conference, LION 8, volume 8426 of Lecture
Notes in Computer Science, pages 36–40. Springer, 2014. 

## ----setup, include=FALSE-----------------------------------------------------
library(knitr)
knit_hooks$set(inline = function(x) {
  if (is.numeric(x)) return(knitr:::format_sci(x, 'latex'))
  highr::hi_latex(x)
})

## ----include=FALSE------------------------------------------------------------
library(knitr)

## ----exampleload,eval=TRUE,include=FALSE----------------------------
library("irace")
load("examples.Rdata") # loads "experiment" and "output"
iraceResults <- irace::read_logfile(system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE))
log_ablation_file <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork = TRUE)
load(log_ablation_file)
options(width = 70)

## ----R_irace_install, prompt=FALSE, eval=FALSE----------------------
# install.packages("irace")

## ----R_irace_launch,eval=FALSE, prompt=FALSE------------------------
# library("irace")
# q() # To exit R

## ----install_win1,eval=FALSE, prompt=FALSE--------------------------
# install.packages("<package>", repos = NULL)

## ----install_win2,eval=FALSE, prompt=FALSE--------------------------
# # Replace <package> with the path to the downloaded file.
# # Replace <R_LIBS_USER> with the path used for installation.
# install.packages("<package>", repos = NULL, lib = "<R_LIBS_USER>")
# # Tell R where to find R_LIBS_USER.
# # This must be executed for every new session.
# .libPaths(c("<R_LIBS_USER>", .libPaths()))

## ----R_irace_test1, prompt=FALSE, eval=FALSE------------------------
# # Load the package
# library("irace")
# # Obtain the installation path
# system.file(package = "irace")

## ----windows_irace_help ,eval=FALSE, prompt=FALSE-------------------
# library("irace")
# irace_cmdline("--help")

## ----irace_R_check, eval=FALSE, prompt=FALSE------------------------
# library("irace")
# scenario <- readScenario(filename = "scenario.txt",
#                          scenario = defaultScenario())
# checkIraceScenario(scenario = scenario)

## ----irace_R_exe, eval=FALSE, prompt=FALSE--------------------------
# library("irace")
# # Go to the directory containing the scenario files
# setwd("./tuning")
# scenario <- readScenario(filename = "scenario.txt",
#                          scenario = defaultScenario())
# irace_main(scenario = scenario)

## ----runexample2,prompt=FALSE,eval=FALSE----------------------------
# library("irace")
# setwd("./tuning/")
# irace_cmdline()

## ----readParameters,prompt=FALSE, eval=FALSE------------------------
# parameters <- readParameters(file = "parameters.txt")

## ----parametersetup,eval=TRUE,include=FALSE-------------------------
# Setup example
parameters <- readParameters(text='
algorithm       "--"                 c		(as,mmas,eas,ras,acs)
ants            "--ants "            i	        (5, 100)
q0              "--q0 "              r          (0.0, 1.0)              | algorithm %in% c("acs")
')

## ----parameterlist,eval=TRUE, prompt=TRUE, size='normalsize', comment=""----
str(parameters, vec.len = 10)

## ----targetRunner,prompt=FALSE, eval=FALSE--------------------------
# targetRunner(experiment, scenario)

## ----experimentlist,eval=TRUE,size='normalsize', prompt=TRUE,  comment=""----
print(experiment)

## ----targetEvaluator, eval=FALSE------------------------------------
# targetEvaluator(experiment, num_configurations, all_conf_id, scenario,
#                 target_runner_call)

## ----instance1, prompt=FALSE, eval=FALSE----------------------------
# scenario$instances <- c("rosenbrock_20 --function=12 --nvar 20",
#                         "rosenbrock_40 --function=12 --nvar 30",
#                         "rastrigin_20 --function=15 --nvar 20",
#                         "rastrigin_40 --function=15 --nvar 30")

## ----repairEx,prompt=FALSE, eval=FALSE------------------------------
# repairConfiguration = function(configuration, parameters)
# {
#   isreal <- names(which(parameters$types[colnames(configuration)] == "r"))
#   # This ignores 'digits'
#   c_real <- unlist(configuration[isreal])
#   c_real <- c_real / sum(c_real)
#   configuration[isreal] <- c_real
#   return(configuration)
# }

## ----repairEx2,prompt=FALSE, eval=FALSE-----------------------------
# repairConfiguration = function(configuration, parameters)
# {
#  columns <- c("p1","p2","p3")
#  # cat("Before"); print(configuration)
#  configuration[columns] <- sort(unlist(configuration[columns], use.names=FALSE))
#  # cat("After"); print(configuration)
#  return(configuration)
# }

## ----targetRunnerParallel,prompt=FALSE, eval=FALSE------------------
# targetRunnerParallel(experiments, exec_target_runner, scenario, target_runner)

## ----targetRunnerParallel2,prompt=FALSE, eval=FALSE-----------------
# targetRunnerParallel <- function(experiments, exec_target_runner, scenario,
#                                  target_runner)
# {
#   lapply(experiments, exec_target_runner, scenario = scenario,
#          target_runner = target_runner)
# }

## ----targetRunnerParallel3,prompt=FALSE, eval=TRUE------------------
   print(output)

## ----testing_r, prompt=FALSE, eval=FALSE----------------------------
# testing_fromlog(logFile = "./irace.Rdata", testNbElites = 1)

## ----change_recover, prompt=TRUE, eval=FALSE------------------------
# iraceResults <- read_logfile("./tuning/irace.Rdata")
# new_path <- "./experiments/tuning/instances/"
# iraceResults$scenario$instances <-
#    paste0(new_path, basename(iraceResults$scenario$instances))
# save(iraceResults, file="./tuning/irace.Rdata")

## ----load_rdata, prompt=FALSE, eval=FALSE---------------------------
# logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
# iraceResults <- read_logfile(logfile)

## ----show_version, prompt=TRUE, eval=TRUE, comment=""---------------
iraceResults$irace_version

## ----show_configurations, prompt=TRUE, eval=TRUE, comment=""--------
head(iraceResults$allConfigurations)

## ----show_idelites, prompt=TRUE, eval=TRUE, comment=""--------------
print(iraceResults$allElites)

## ----get_elites, prompt=TRUE, eval=TRUE, comment=""-----------------
logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
getFinalElites(logfile, n = 0)

## ----show_iditelites, prompt=TRUE, eval=TRUE, comment=""------------
print(iraceResults$iterationElites)

## ----get_elite, prompt=TRUE, eval=TRUE, comment=""------------------
last <- length(iraceResults$iterationElites)
id <- iraceResults$iterationElites[last]
getConfigurationById(iraceResults, ids = id)

## ----get_experiments, prompt=TRUE, eval=TRUE, comment=""------------
# As an example, we use the best configuration found
best_config <- getFinalElites(iraceResults, n = 1)
best_id <- as.character(best_config$.ID.)
# Obtain the results of the best configuration
all_exp <- iraceResults$experiments[, best_id]
# all_exp is a vector and names(all_exp) is the (instance,seed) index.
all_exp
# Obtain the results of the first and best configurations
all_exp <- iraceResults$experiments[, c("1", best_id)]
# all_exp is a matrix: colnames(all_exp) is configurationID and
# rownames(all_exp) is the (instance,seed) index.
all_exp

## ----get_instance_seed, prompt=TRUE, eval=TRUE, comment=""----------
# As an example, we get instanceID, seeds and instances of the experiments
# of the best configuration.
# We could get the indexes of the instances on which at least one
# configuration was executed:
pair_index <- which(apply(!is.na(all_exp), 1L, any))
# or the instances on which all configurations were executed:
pair_index <- which(apply(!is.na(all_exp), 1L, all))
# but in this example we get the indexes of the instances executed for
# the best configuration.
pair_index <- which(!is.na(all_exp[, best_id]))
instanceID <- get_instanceID_seed_pairs(iraceResults)[["instanceID"]][pair_index]
# or get the seeds
get_instanceID_seed_pairs(iraceResults)[["seed"]][pair_index]
# or obtain the actual instances.
iraceResults$scenario$instances[instanceID]
# If the instances are of atomic type (integers, floating-point numbers or
# character strings), the above is similar to:
get_instanceID_seed_pairs(iraceResults, index = pair_index, instances=TRUE)

## ----get_model, prompt=TRUE, eval=TRUE, comment=""------------------
# As an example, we get the model probabilities for the
# localsearch parameter.
iraceResults$state$model["localsearch"]
# The order of the probabilities corresponds to:
iraceResults$scenario$parameters$domains$localsearch

## ----get_test_exp, prompt=TRUE, eval=TRUE, comment=""---------------
# Get the results of the testing
iraceResults$testing$experiments

## ----get_test_seeds, prompt=TRUE, eval=TRUE, comment=""-------------
# Get the seeds used for testing
iraceResults$testing$seeds

## ----wilcox_test,prompt=TRUE, eval=TRUE, comment=""-----------------
results <- iraceResults$testing$experiments
# Wilcoxon paired test
conf <- gl(ncol(results), # number of configurations
           nrow(results), # number of instances
           labels = colnames(results))
pairwise.wilcox.test (as.vector(results), conf, paired = TRUE, p.adj = "bonf")

## ----conc, prompt=TRUE, eval=TRUE, comment=""-----------------------
irace:::concordance(iraceResults$testing$experiments)

## ----testEvo, fig.pos="tb", fig.align="center", out.width='0.7\\textwidth', fig.cap="Testing set performance of the best-so-far configuration over number of experiments. Label of each point is the configuration ID.", prompt=FALSE, eval=TRUE, comment=""----
# Get summary data from the logfile.
irs <- irace_summarise(iraceResults)
# Get number of iterations
iters <- irs$n_iterations
# Get number of experiments (runs of target-runner) up to each iteration
fes <- cumsum(table(iraceResults$state$experiment_log[["iteration"]]))
# Get the mean value of all experiments executed up to each iteration
# for the best configuration of that iteration.
elites <- as.character(iraceResults$iterationElites)
values <- colMeans(iraceResults$testing$experiments[, elites])
stderr <- function(x) sqrt(var(x)/length(x))
err <- apply(iraceResults$testing$experiments[, elites], 2L, stderr)
plot(fes, values, type = "s",
     xlab = "Number of runs of the target algorithm",
     ylab = "Mean value over testing set", ylim=c(min(values-err),max(values+err)))
points(fes, values, pch=19)
arrows(fes, values - err, fes, values + err, length=0.05, angle=90, code=3)
text(fes, values, elites, pos = 1)

## ----ablation, prompt=FALSE, eval=FALSE-----------------------------
# ablog <- ablation("irace.Rdata", src = 1)
# plotAblation(ablog)

## ----testAb, fig.pos="htb!", fig.align="center", out.width="0.75\\textwidth", fig.cap="Example of plot generated by \\code{plotAblation()}.", prompt=FALSE, eval=TRUE, echo=FALSE----
logfile <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork=TRUE)
plotAblation(logfile)

## ----ablation_cmdline, prompt=FALSE, eval=TRUE,echo=FALSE, comment=""----
ablation_cmdline("--help")

## ----postsel, prompt=FALSE, eval=FALSE------------------------------
# # Execute all elite configurations in the iterations
# psRace("irace.Rdata", max_experiments = 0.5, iteration_elites=TRUE)
# # Execute a set of configurations IDs providing budget
# psRace("irace.Rdata", conf_ids = c(34, 87, 102, 172, 293), max_experiments = 500)

## ----targetCmdline, prompt=FALSE, eval=FALSE------------------------
# targetRunner="./real_target_runner.py"
# targetRunnerLauncher="python"
# targetCmdLine="-m {targetRunner} {configurationID} {instanceID}\
#  --seed {seed} -i {instance} --cutoff {bound} {targetRunnerArgs}"

## ----faq3, eval=FALSE-----------------------------------------------
# library(Rmpi)
# mpi.spawn.Rslaves(nslaves = 10)
# paths <- mpi.applyLB(1:10, function(x) {
#   library(irace); return(path.package("irace")) })
# print(paths)

## ----R_irace_home2, prompt=FALSE, eval=FALSE------------------------
# system.file(package = "irace")

# This is an example of launching irace from an R script.
library(irace)
# Initial set up. Please check in the documentation of your cluster whether
# $SLURM_NTASKS_PER_NODE is the correct environment variable to get the number
# of tasks that can be run in parallel.
ncores <- as.numeric(Sys.getenv("SLURM_NTASKS_PER_NODE", parallel::detectCores()))

# This needs to be customized by the user.  Instead of defining
# 'parameter_table' you could use a file 'parameters.txt'.
parameter_table <- '
param1 "--param1 " i (1,100)
param2 "--param2=" c (0,1)
'
parameters <- readParameters(text = parameter_table)

# This needs to be customized by the user.  Instead of defining targetRunner as
# an R function 'my_target_runner' you could use an executable file
# 'my_target_runner'.
my_target_runner <- function(experiment, scenario)
{
  instance <- experiment$instance
  configuration <- experiment$configuration
  seed <- experiment$seed
  cost <- my_algorithm(instance, configuration, seed)
  # You could also return 'time' if you are using 'maxTime' in the scenario.
  return(list(cost = cost))
}


# Some function that loads or defines the instances. You could also use
# trainInstancesFile or trainInstancesDir in the scenario.
instances <- load_instances() 
scenario <- list(targetRunner = my_target_runner,
                 instances = instances,
                 maxExperiments = 200,
                 parallel = ncores, parameters = parameters)

# check if the scenario is valid (comment out if taking too much time)
checkIraceScenario(scenario)

# run irace
irace_main(scenario = scenario)

files <- c(file.path("iracebin", c("irace","ablation")),
  file.path("dummy", "target-runner-dummy"))

file_move <- function(from, to) {
  file.copy(from = from, to = to, overwrite = TRUE)
  file.remove(from)
}

if (WINDOWS) files <- paste0(files, ".exe")
if (any(file.exists(files))) {
  dest <- file.path(R_PACKAGE_DIR,  paste0('bin', R_ARCH))
  dir.create(dest, recursive = TRUE, showWarnings = FALSE)
  file_move(files, dest)
} else {
  stop(files, " not found!")
}

### Submit/wait for jobs in batch clusters.
sge.job.finished <- function(jobid)
  system (paste0("qstat -j ", jobid), ignore.stdout = TRUE, ignore.stderr = TRUE,
          intern = FALSE, wait = TRUE)

pbs.job.finished <- function(jobid)
  system (paste0("qstat ", jobid), ignore.stdout = TRUE, ignore.stderr = TRUE,
          intern = FALSE, wait = TRUE)

torque.job.finished <- function(jobid)
{
  output <- suppressWarnings(system2("qstat", jobid, stdout = TRUE, stderr = TRUE))
  ## 1. If the return code of qstat in not 0, then no such job is in the queue
  ## anymore. That means that the job was (assumption) successfully submitted,
  ## executed and completed. Then, for some time the job id was marked as
  ## completed in the jobs queue and the job status removed from the queue list
  ## after some time. No job completed status of successfully submitted and
  ## completed jobs in the queue list happens if a small task was executed
  ## first and a task of some hours execution time finishes much later. The job
  ## status of the small task will not be in the queue anymore after some
  ## minutes. So: If no job id is in the queue list anymore and torq's qstat
  ## returns an error (return code > 0), then the job has been successfully
  ## executed (if it has been started successfully before).
  if (!is.null(attr(output, "status"))) return(TRUE)
  # 2. If qstat returns OK (return code ==0), then one has to parse qstat's
  # output. If the 5th token in the last line is a 'C', then the job has
  # terminated and its output files can be processed. Otherwise the job is not
  # completed (queued, running, exiting...)
  any(grepl(paste0(jobid, ".*\\sC\\s"), output))
}

slurm.job.finished <- function(jobid)
{
  output <- suppressWarnings(system2("squeue", c("-j", jobid, "--noheader"),
                                     stdout = TRUE, stderr = TRUE))
  # If the above returns non-zero, either the job terminated or it never
  # existed.
  if (!is.null(attr(output, "status"))) return(TRUE)
  # If may return zero, but the job is not in the system anymore because it
  # completed. This is different from the Torque case.
  !any(grepl(paste0("\\s", jobid, "\\s"), output))
}

htcondor.job.finished <- function(jobid)
{
  output <- suppressWarnings(system2("condor_q", jobid, stdout = TRUE, stderr = TRUE))
  # Check if job is still in the queue, otherwise it is considered finished
  !any(grepl(paste0("ID:\\s", jobid), output))
}

## Launch a job with qsub and return its jobID. This function does not
## call qsub directly, but instead targetRunner should be a script that
## invokes qsub and returns a jobID.
target_runner_qsub <- function(experiment, scenario)
{
  debugLevel   <- scenario$debugLevel
  res <- run_target_runner(experiment, scenario)
  cmd <- res$cmd
  output <- res$output
  args <- res$args

  jobID <- NULL
  outputRaw <- output$output
  err.msg <- output$error
  if (is.null(err.msg)) {
    # We cannot use parse.output because that tries to convert to numeric.
    if (debugLevel >= 2) { cat (outputRaw, sep = "\n") }
    # Initialize output as raw. If it is empty stays like this.
    # strsplit crashes if outputRaw == character(0)
    if (length(outputRaw) > 0) {
      jobID <- strsplit(trim(outputRaw), "[[:space:]]+")[[1]]
    }
    if (length(jobID) != 1) {
      err.msg <- paste0("The output of targetRunner should be only the jobID!")
      jobID <- NULL
    }
  }
  list(jobID = jobID, error = err.msg, outputRaw = outputRaw, call = paste(cmd, args))
}

cluster_lapply <- function(X, scenario, poll.time = 2)
{
  debugLevel <- scenario$debugLevel

  cluster.job.finished <-
    switch(scenario$batchmode,
           sge = sge.job.finished,
           pbs = pbs.job.finished,
           torque = torque.job.finished,
           slurm = slurm.job.finished,
           htcondor = htcondor.job.finished,
           irace_error ("Invalid value of scenario$batchmode = ", scenario$batchmode))

  # Parallel controls how many jobs we send at once. Some clusters have low
  # limits.
  ## FIXME: It would be better to submit up to the limit, then one by one as jobs finish.
  chunksize <- scenario$parallel
  if (chunksize < 0L) {
    chunksize <- length(X)
  }
  chunks <- split(X, ceiling(seq_along(X) / chunksize))
  for (chunk in chunks) {
    if (debugLevel >= 1) {
      irace_note ("Sending ", length(chunk), " / ", length(X), " jobs\n")
    }
    output <- lapply(chunk, exec_target_runner, scenario = scenario, target_runner = target_runner_qsub)
    jobIDs <- sapply(output, "[[", "jobID")
  
    ## Wait for cluster jobs to finish.
    if (length(jobIDs) > 0L && debugLevel >= 1L) {
      irace_note("Waiting for jobs ('.' == ", poll.time, " s) ")
    }
    for (jobID in jobIDs) {
      while (!cluster.job.finished(jobID)) {
        if (debugLevel >= 1) { cat(".") }
        Sys.sleep(poll.time)
      }
      if (debugLevel >= 1) {
        cat("\n")
        irace_note ("DONE (", jobID, ")\n")
      }
    }
  }
  output
}

# Checks that variables in the expressions are within the parameters names.
check_domain_dependencies <- function (params, depends, types)
{
  allnames <- names(params)
  allowed_fx <- c("+", "-", "*", "/", "%%", "min", "max", "round", "floor", "ceiling", "trunc")
  depends <- Filter(length, depends)
  for (p in names(depends)) {
    parameter <- params[[p]]
    domain <- parameter[["domain"]]
    vars <- depends[[p]]
    flag <- vars %in% allnames
    if (!all(flag)) {
      irace_error ("Domain (", paste0(domain, collapse=", "),
                   ") of parameter '", p, "' is not valid: '",
                   paste0(vars[!flag], collapse=", "),
                   "' cannot be found in the scenario parameters: ",
                   paste0(allnames, collapse=", ")," .")
    }
    flag <- types[vars] %in% c("i", "r")
    if (!all(flag)) {
      irace_error ("Domain of parameter '", p, "' depends on non-numerical",
        " parameters: ", paste0(vars[!flag], collapse=", "), " .")
    }

    # Supported operations for dependent domains
    fx <- setdiff(all.names(domain, unique=TRUE), all.vars(domain, unique=TRUE))
    flag <- fx %in% allowed_fx
    if (!all(flag)) {
      irace_error ("Domain of parameter '", p, "' uses function(s) ",
                   "not yet supported by irace: ",
                   paste0(fx[!flag], collapse=", "), " .")
    }
  }
  invisible(TRUE)
}

check_forbidden_params <- function(x, pnames, filename = NULL)
{
  if (length(x) == 0L) return(invisible())
  if (any(unique(unlist(lapply(x, all.names))) %in% c("&&", "||"))) {
    for (ex in x) {
      if (any(all.names(ex) %in% c("&&", "||")))
        irace_error("Please use '&' and '|' instead of '&&' and '|' in: ", deparse(ex), " .\n")
    }
  }
  if (all(all.vars(x) %in% pnames)) return(invisible())
  for (ex in x) {
    v <- setdiff(all.vars(ex), pnames)
    if (length(v)) {
      v <- paste0(v, collapse=", ")
      if (is.null(filename)) {
        irace_error("Forbidden expression '", deparse(ex), "' contains unknown parameter(s): ", v)
      } else if (is.na(filename)) {
        irace_error("Expression '", deparse(ex), "' after [forbidden] contains unknown parameter(s): ", v)
      } else {
        irace_error("Expression '", deparse(ex), "' after [forbidden] in '", filename, "' contains unknown parameter(s): ", v)
      }
    }
  }
}


# ---------------------------------------------------------------------------
# Ordering of the parameters according to conditions hierarchy
# *  The conditions hierarchy is an acyclic directed graph.
#    Function treeLevel() computes an order on vertex s.t:
#    level(A) > level(B)  <=>  There is an arc A ---> B
#    (A depends on B to be activated)
# *  If a cycle is detected, execution is stopped
# *  If a parameter depends on another one not defined, execution is stopped
param_level <- function(paramName, varsTree, rootParam = paramName)
{
  # The last parameter is used to record the root parameter of the
  # recursive call in order to detect the presence of cycles.
  vars <- varsTree[[paramName]]
  if (length(vars) == 0L) return(1L) # This parameter does not have conditions

  # This parameter has some conditions
  # Recursive call: level <- MAX( level(m) : m in children )
  maxChildLevel <- 0L
  for (child in vars) {
    # The following line detects cycles
    if (child == rootParam)
      irace_error("Cycle detected in subordinate parameters! ",
                  "Check definition of conditions and/or dependent domains.\n",
                  "One parameter of this cycle is '", rootParam, "'")

    # The following line detects a missing definition
    if (child %not_in% names(varsTree))
      irace_error("A parameter definition is missing! ",
                  "Check definition of parameters.\n",
                  "Parameter '", paramName,
                  "' depends on '", child, "' which is not defined.")

    level <- param_level(child, varsTree, rootParam)
    if (level > maxChildLevel)
      maxChildLevel <- level
  }
  maxChildLevel + 1L
}


transform_domain <- function(transf, domain, type)
{
  # We do not support transformation of dependent parameters, yet
  # TODO: think about dependent domain transformation
  if (is.expression(domain))
    stop("Parameter domain transformations are not yet available for",
         " dependent parameter domains.")

  lower <- domain[1L]
  upper <- domain[2L]

  if (transf == "log") {
    # Reject log if domain contains zero or negative values
    if (any(domain <= 0))
      stop("Domain (", lower, ", ", upper, ") of parameter of type 'log' contains non-positive values")

    trLower <- log(lower)
    # +1 to adjust before floor()
    trUpper <- if (type == "i") log(upper + 1) else log(upper)

    irace_assert(is.finite(trLower))
    irace_assert(is.finite(trUpper))
    attr(transf, "lower") <- trLower
    attr(transf, "upper") <- trUpper
    return(transf)
  }
  stop("unrecognized transformation type '", transf, "'")
}

named_stop <- function(name, ...)
  stop(errorCondition(paste0(..., collapse=""), class = name, call = sys.call(-1)))

param_parse_condition <- function(x, type)
{
  if (is.language(x))
    x <- as.expression(x)
  else if (is.character(x)) {
    x <- tryCatch(str2expression(x), error = function(e) {
      msg <- sub("<text>:[0-9]+:[0-9]+: ", "", conditionMessage(e), perl=TRUE)
      msg <- sub("\n1:", "\n ", msg, perl=TRUE)
      stop("Invalid condition '", x, "': ", msg, "\n")
    })
  }
  if (!is.expression(x))
      stop("Invalid condition '", x, "'")
  if (any(all.names(x) %in% c("&&", "||")))
    stop("Please use '&' and '|' instead of '&&' and '|' in: ", deparse(x[[1L]]), " .\n")
  x
}

valid_real_bound <- function(x, digits)
{
  if (is.na(x) || x == 0) return(TRUE)
  rx <- round(x, digits = digits)
  ((abs(rx - x) <= .irace_tolerance * max(1, abs(x))) && digits >= -log10(abs(x)))
}

parse_exp_domain <- function(domain)
{
  # Convert to expression if string
  domain <- sapply(domain, USE.NAMES=FALSE, function(x) if (is.character(x)) str2expression(x) else x)
  # If is an expression or language without variables, then evaluate to convert it to numeric.
  domain <- sapply(domain, USE.NAMES=FALSE, function(x) if (is.numeric(x) || length(all.vars(x, unique=FALSE))) x else eval(x))
  domain
}

Parameter <- function(name, type, domain, label, condition, transf,
                      digits = if (type=='r') 15L else NULL)
{
  cls <- switch(type, i="ParamInt", r="ParamReal", c="ParamCat", o="ParamOrd",
    stop("Unknown parameter type '", type, "'."))

  if (!isTRUE(condition)) {
    condition <- param_parse_condition(condition)
  }

  if (type %in% c("i", "r")) {
    exp_domain <- parse_exp_domain(domain)
    domain <- sapply(exp_domain, USE.NAMES=FALSE, function(x) if (is.numeric(x)) x else NA)
    if (type == "r") {
      # digits >= 15 is almost infinite-precision so we do not complain.
      digits <- as.integer(digits)
      if (digits < 15L &&
          (!valid_real_bound(domain[1L], digits)
            || !valid_real_bound(domain[2L], digits))) {
        for (i in seq.int(digits+1L, 15L)) {
          if (valid_real_bound(domain[1L], i) && valid_real_bound(domain[2L], i))
            break
        }
        stop("Domain bounds (", domain[1L], ", ", domain[2L],
             ") of parameter '", name, "' of type 'real' must be representable within the given 'digits=",
             digits, "'; you would need at least 'digits=", i, "' or adjust the domain")
      }
      if (anyNA(domain))
        domain <- as.expression(exp_domain)
    } else { # type == "i"
      if (any(!is.wholenumber(domain[!is.na(domain)])))
        named_stop("invalid_domain", "for parameter '", name, "' of type 'integer', values must be integers")
      if (anyNA(domain))
        domain <- as.expression(sapply(exp_domain, USE.NAMES=FALSE, function(x) if (is.numeric(x)) as.integer(x) else x))
      else
        domain <- as.integer(domain)
    }
    if (!is.expression(domain) && domain[1L] >= domain[2L])
      named_stop("invalid_range", "lower bound must be smaller than upper bound in numeric range")

  } else { # type %in% c("c", "o")
    if (anyDuplicated(domain)) {
      dups <- duplicated(domain)
      stop("duplicated values (", paste0('\"', domain[dups], "\"", collapse = ', '), ") for parameter '", name, "'")
    }
    if (type == "o") {
      tmp <- suppressWarnings(as.numeric(domain))
      if (!anyNA(tmp) && !identical(order(tmp), seq_along(tmp)))
        stop("the domain of parameter '", name, "' appears to be a discretization of a numerical range, but the values are not in increasing order: ",
          paste0(domain, collapse = ', '))
    }
  }

  if (transf != "")
    transf <- transform_domain(transf, domain, type)

  isFixed <- (type == "c" || type == "o") && (length(domain) == 1L)
  param <- list(name = name, type = type, domain = domain, label = label, is_dependent = is.expression(domain),
    isFixed = isFixed, transform = transf, condition = condition)
  if (type == "r")
    param$digits <- digits

  class(param) <- c(cls, "Parameter", class(param))
  param
}

#' Create a parameter space to be tuned.
#'
#' @description
#'  - `param_cat()` creates a categorical parameter.
#'  - `param_ord()` creates an ordinal parameter.
#'  - `param_real()` creates a real-valued parameter.
#'  - `param_int()` creates an integer parameter.
#'
#' @param ... one or more parameters created by `param_int()`, `param_real()`, `param_ord()`, or `param_cat()`.
#' @param forbidden `expression()|character(1)`\cr String or list of expressions that define forbidden parameter configurations.
#' @inheritParams readParameters
#'
#' @return `ParameterSpace`
#' @name parameters
#' @examples
#' digits <- 4L
#' parametersNew(
#'    param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--"),
#'    param_ord(name = "localsearch", values = c("0", "1", "2", "3"), label = "--localsearch "),
#'    param_real(name = "alpha", lower = 0.0, upper=5.0, label = "--alpha ", digits = digits),
#'    param_real(name = "beta", lower = 0.0, upper = 10.0, label = "--beta ", digits = digits),
#'    param_real(name = "rho", lower = 0.01, upper = 1.00, label = "--rho ", digits = digits),
#'    param_int(name = "ants", lower = 5, upper = 100, transf = "log", label = "--ants "),
#'    param_real(name = "q0", label = "--q0 ", lower=0.0, upper=1.0,
#'               condition = expression(algorithm == "acs")),
#'    param_int(name = "rasrank", label = "--rasranks ", lower=1, upper=quote(min(ants, 10)),
#'              condition = 'algorithm == "ras"'),
#'    param_int(name = "elitistants", label = "--elitistants ", lower=1, upper=expression(ants),
#'              condition = 'algorithm == "eas"'),
#'    param_int(name = "nnls", label = "--nnls ", lower = 5, upper = 50,
#'              condition = expression(localsearch %in% c(1,2,3))),
#'    param_cat(name = "dlb",  label = "--dlb ", values = c(0,1),
#'              condition = "localsearch %in% c(1,2,3)"),
#'    forbidden = "(alpha == 0) & (beta == 0)")
#'
#' @export
parametersNew <- function(..., forbidden = NULL, debugLevel = 0L)
  ParameterSpace$new(..., forbidden = forbidden, verbose = debugLevel)

ParameterSpace <- R6::R6Class("ParameterSpace", cloneable = TRUE, lock_class = TRUE, lock_objects = TRUE, public = list(
  .params = NULL,
  conditions = NULL,
  depends = NULL,
  domains = NULL,
  forbidden = NULL,
  hierarchy = NULL,
  isFixed = NULL,
  names = NULL,
  names_fixed = NULL,
  names_variable = NULL,
  names_numeric = NULL,
  nbParameters = NULL,
  nbFixed = NULL,
  nbVariable = NULL,
  switches = NULL,
  types = NULL,
  initialize = function(..., forbidden = NULL, verbose = 0L) {
    .params <- list(...)
    names(.params) <- sapply(.params, "[[", "name")
    self$names <- names(.params)
    self$types <- sapply(.params, "[[", "type")
    self$switches <- sapply(.params, "[[", "label")
    # This has to be a list because each element is a vector.
    self$domains <- lapply(.params, "[[", "domain")
    self$conditions <- lapply(.params, "[[", "condition")
    self$isFixed <- sapply(.params, "[[", "isFixed")
    self$names_fixed <- self$names[self$isFixed]
    self$names_variable <- self$names[!self$isFixed]
    self$names_numeric <- self$names[self$types %in% c("r", "i")]
    self$nbParameters <- length(.params)
    self$nbFixed <- sum(self$isFixed)
    self$nbVariable <- sum(!self$isFixed)

    # Obtain the variables in each condition
    self$depends <- lapply(self$domains, all.vars)
    # Check that domain dependencies are OK.
    check_domain_dependencies(.params, self$depends, self$types)
    # Merge dependencies and conditions
    self$depends <- Map(union, self$depends, lapply(self$conditions, all.vars))
    # Sort parameters in 'conditions' in the proper order according to conditions
    hierarchyLevel <- sapply(names(.params), param_level, varsTree = self$depends)
    self$hierarchy <- hierarchyLevel
    self$conditions <- self$conditions[order(hierarchyLevel)]
    names(self$hierarchy) <- names(.params)
    stopifnot(length(self$conditions) == length(.params))

    if (!is.null(forbidden)) {
      if (is.language(forbidden))
        forbidden <- as.expression(forbidden)
      else if (is.character(forbidden))
        forbidden <- str2expression(forbidden)
      if (!is.expression(forbidden))
        stop ("Invalid forbidden expression.")
      check_forbidden_params(forbidden, names(.params))
      # FIXME: Instead of a list, we should generate a single expression that is
      # the logical-OR of all elements of the list.
      # First we would need to handle the "is.na(x) | !(x)" case here.
      # Maybe: sapply(forbiddenExps, function(x) substitute(is.na(x) | !(x), list(x=x)))
      # x <- parse(text=paste0("(", paste0(forbiddenExps,collapse=")||("), ")"))
      self$forbidden <- sapply(forbidden, compile_forbidden)
    }

    # Optimize always TRUE conditions.
    cond_names <- names(which(!unlist(lapply(self$conditions, is.logical))))
    for (p in cond_names) {
      deps <- self$depends[[p]]
      # p depends on deps and they are both fixed and always active.
      if (all(self$isFixed[deps]) && all(sapply(self$conditions[deps], isTRUE, USE.NAMES=FALSE))) {
        self$conditions[[p]] <- eval(self$conditions[[p]], envir = self$domains[deps], enclos = NULL)
      }
    }

    # Print the hierarchy vector:
    if (verbose >= 1L) {
      cat ("# --- Parameters Hierarchy ---\n")
      print(data.frame(Parameter = paste0(names(self$hierarchy)),
        Level = self$hierarchy,
        "Depends on" = sapply(names(self$hierarchy), function(x) paste0("", x, collapse=", ")),
        row.names=NULL))
      cat("# ------------------------\n")
    }
    self$.params <- .params
  },
  get = function(x) if (missing(x)) self$.params else self$.params[[x]],
  get_ordered = function() self$.params[names(self$conditions)],

  forbid_configurations = function(x) {
    # FIXME: This copies the whole list. Avoid the copy and just append.
    self$forbidden <- c(self$forbidden,
      buildForbiddenExp(configurations = x[, self$names, drop=FALSE]))
  },
  as_character = function() {
    names_len <- max(nchar(names(self$.params)))
    label_len <- max(nchar(sapply(self$.params, "[[", "label"))) + 2L
    output <- ""
    digits <- c()
    for (p in self$.params) {
      type <- p[["type"]]
      domain <- p[["domain"]]
      if (is.expression(domain)) {
        domain <- sapply(domain, USE.NAMES=FALSE, function(x) {
          if (is.numeric(x)) {
            if (type == "r")
              x <- formatC(x, digits = p[["digits"]], format="f", drop0trailing=TRUE)
          } else {
            x <- paste0("\"", deparse(x, width.cutoff = 500L), "\"")
          }
          x
        })
      }
      domain <- paste0('(', paste0(domain, collapse=","), ')')
      condition <- p[["condition"]]
      condition <- if (isTRUE(condition)) "" else paste0(" | ", condition)
      transf <- p[["transform"]]
      if (!is.null(transf) && transf != "")
        type <- paste0(type, ",", transf)
      label <- paste0('"', p[["label"]], '"')
      output <- paste0(output, sprintf('%*s %*s %s %-15s%s\n', -names_len, p[["name"]], -label_len, label, type, domain, condition))
      if (type == "r")
        digits <- c(digits, p[["digits"]])
    }
    if (!is.null(self$forbidden)) {
      output <- paste0(output, "\n[forbidden]\n",
        paste0(collapse="\n", sapply(self$forbidden, attr, "source", USE.NAMES=FALSE)),
        "\n")
    }
    digits <- unique(digits)
    n <- length(digits)
    if (n > 1L)
      stop("Different digits per parameter is not supported yet")
    if (n == 1L)
      output <- paste0(output, "\n[global]\ndigits = ", digits[[1L]], "\n")
    output
  }),
)

#' @param name  `character(1)`\cr Parameter name (must be alphanumeric).
#' @param values  `character()`\cr Domain as a vector of strings.
#' @param label `character(1)`\cr Label associated to the parameter. Often used to encode a command-line switch that activates the parameter.
#' @param condition `expression(1)|character(1)`\cr Expression that defines when the parameter is active according to the value of other parameters.
#' @rdname parameters
#' @export
param_cat <- function(name = name, values, label = "", condition = TRUE)
  Parameter(name = name, type = "c", domain = as.character(values), label = label, condition = condition, transf = "")

#' @rdname parameters
#' @export
param_ord <- function(name, values, label = "", condition = TRUE)
  Parameter(name = name, type = "o", domain = as.character(values), label = label, condition = condition, transf = "")

#' @param lower,upper  Lower and upper limits of the valid domain.
#' @param transf `character(1)`\cr If `"log"`, then the parameter is sampled in a logarithmic scale.
#' @inheritParams readParameters
#' @rdname parameters
#' @export
param_real <- function(name, lower, upper, label = "", condition = TRUE, transf = "", digits = 15L)
  Parameter(name = name, type = "r", domain = c(lower, upper), label = label, condition = condition, transf = transf, digits = digits)

#' @rdname parameters
#' @export
param_int <- function(name, lower, upper, label = "", condition = TRUE, transf = "")
  Parameter(name = name, type = "i", domain = c(lower, upper), label = label, condition = condition, transf = transf)


transform_from_log <- function(x, transf, lower, upper)
{
  trLower <- attr(transf, "lower")
  trUpper <- attr(transf, "upper")
  x <- exp(trLower + (trUpper - trLower) * x)
  clamp(x, lower, upper)
}

transform_to_log <- function(x, transf)
{
  trLower <- attr(transf, "lower")
  trUpper <- attr(transf, "upper")
  (log(x) - trLower)/(trUpper - trLower)
}

## How to sample integer values?
#
# The problem: If we have an integer with domain [1,3] and we sample a real value
# and round, then there are more chances of getting 2 than 1 or 3:
# [1, 1,5) -> 1
# [1.5, 2,5) -> 2
# [2.5, 3) -> 3
#
# The solution: Sample in [lowerbound, upperbound + 1], that is, [1, 4], then floor():
# [1, 2) -> 1
# [2, 3) -> 2
# [3, 4) -> 3
#
# Why floor() and not trunc()?
# Because trunc(-1.5) -> -1, while floor(-1.5) -> -2, so for a domain [-3,-1]:
#
# [-3, -2) -> -3
# [-2, -1) -> -2
# [-1, 0)  -> -1
#
# Issue 1: We can sample 4 (upperbound + 1). In that case, we return 3.
#
# Issue 2: When sampling from a truncated normal distribution, the extremes are
# not symmetric.
#
# nsamples <- 100000
# table(floor(rtnorm(nsamples, mean=1, sd=1, lower=1,upper=4)))/nsamples
# table(floor(rtnorm(nsamples, mean=3, sd=1, lower=1,upper=4)))/nsamples
#
# To make them symmetric, we translate by 0.5, so that the mean is at the
# actual center of the interval that will produce the same value after
# truncation, e.g., given an integer value of 1, then mean=1.5, which is at the
# center of [1,2).
#
# nsamples <- 100000
# table(floor(rtnorm(nsamples, mean=1.5, sd=1, lower=1,upper=4)))/nsamples
# table(floor(rtnorm(nsamples, mean=3.5, sd=1, lower=1,upper=4)))/nsamples
#
# The above reasoning also works for log-transformed domains, because
# floor() happens in the original domain, not in the log-transformed one,
# except for the case of log-transformed negative domains, where we have to
# translate by -0.5.
#
integer_round <- function(x, lower, upper)
{
  x <- floor(x)
  # The probability of this happening is very small, but it happens.
  x <- pmin.int(upper, x)
  irace_assert(all(x >= lower))
  as.integer(x)
}

sample_numeric_unif <- function(n, lower, upper, transf)
{
  if (transf == "log") {
    value <- runif(n, min = 0, max = 1)
    return(transform_from_log(value, transf, lower, upper))
  }
  runif(n, min = lower, max = upper)
}

sample_numeric_norm <- function(n, mean, sd, lower, upper, transf)
{
  if (transf == "log") {
    x <- transform_to_log(mean, transf)
    x <- rtnorm(n, x, sd, lower = 0, upper = 1)
    return(transform_from_log(x, transf, lower, upper))
  }
  rtnorm(n, mean, sd, lower, upper)
}

sample_model <- function(param, n, model, domain)
  UseMethod("sample_model")

sample_unif <- function(param, n, domain)
  UseMethod("sample_unif")

param_quantile <- function(param, probs, domain)
  UseMethod("param_quantile")

#' @exportS3Method
sample_model.ParamCat <- function(param, n, model, domain = param[["domain"]])
  sample(domain, size = n, replace = TRUE, prob = model)

#' @exportS3Method
sample_unif.ParamCat <- function(param, n, domain = param[["domain"]])
  sample(domain, n, replace=TRUE)

#' @exportS3Method
param_quantile.ParamCat <- function(param, probs, domain = param[["domain"]])
{
  n <- length(domain)
  z <- integer_round(as.numeric(probs) * n + 1L, 1L, n)
  domain[z]
}

#' @exportS3Method
sample_model.ParamOrd <- function(param, n, model, domain = param[["domain"]])
  domain[floor(sample_numeric_norm(n, mean = model[[2L]], sd = model[[1L]],
    lower = 1L, upper = length(domain), transf = ""))]

#' @exportS3Method
sample_unif.ParamOrd <- function(param, n, domain = param[["domain"]])
  sample(domain, n, replace=TRUE)

#' @exportS3Method
param_quantile.ParamOrd <- param_quantile.ParamCat

#' @exportS3Method
sample_model.ParamInt <- function(param, n, model, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_integer_)
  lower <- domain[1L]
  upper <- domain[2L]
  transf <- param[["transform"]]
  mean <- model[[2L]]
  if (is.na(mean))
    # +1 for correct rounding before floor()
    value <- sample_numeric_unif(n, lower, 1L + upper, transf)
  else
    # + 0.5 because negative domains are log-transformed to positive domains.
    value <- sample_numeric_norm(n, mean + 0.5, sd = model[[1L]], lower = lower,
      # +1 for correct rounding before floor()
      upper = 1L + upper, transf = transf)
  # We use original upper, not the +1L for 'i'.
  integer_round(value, lower, upper)
}

#' @exportS3Method
sample_unif.ParamInt <- function(param, n, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_integer_)
  lower <- domain[1L]
  upper <- domain[2L]
  # +1 for correct rounding before floor()
  value <- sample_numeric_unif(n, lower, 1L + upper, transf = param[["transform"]])
  # We use original upperBound, not the +1L for 'i'.
  integer_round(value, lower, upper)
}

#' @exportS3Method
param_quantile.ParamInt <- function(param, probs, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_integer_)
  lower <- domain[1L]
  upper <- domain[2L]
  probs <- as.numeric(probs)
  transf <- param[["transform"]]
  if (transf == "log") {
    # +1 for correct rounding before floor()
    probs <- transform_from_log(probs, transf, lower, upper + 1L)
  } else {
    probs <- probs * (upper + 1L - lower) + lower
  }
  integer_round(probs, lower, upper)
}
#' @exportS3Method
sample_model.ParamReal <- function(param, n, model, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_real_)
  lower <- domain[[1L]]
  upper <- domain[[2L]]
  transf <- param[["transform"]]
  mean <- model[[2L]]
  if (is.na(mean))
    x <- sample_numeric_unif(n, lower, upper, transf)
  else
    x <- sample_numeric_norm(n, mean, sd = model[[1L]], lower, upper, transf)
  x <- round(x, digits = param[["digits"]])
  irace_assert(all(x >= lower) && all(x <= upper))
  x
}

#' @exportS3Method
sample_unif.ParamReal <- function(param, n, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_real_)
  lower <- domain[[1L]]
  upper <- domain[[2L]]
  x <- sample_numeric_unif(n, lower, upper, transf = param[["transform"]])
  x <- round(x, digits = param[["digits"]])
  irace_assert(all(x >= lower) && all(x <= upper))
  x
}

#' @exportS3Method
param_quantile.ParamReal <- function(param, probs, domain = param[["domain"]])
{
  # Dependent domains could be not available because of inactivity of parameters
  # on which they are depedent. In this case, the dependent parameter becomes
  # not active and we return NA.
  if (anyNA(domain))
    return(NA_real_)
  lower <- domain[1L]
  upper <- domain[2L]
  probs <- as.numeric(probs)
  digits <- param[["digits"]]
  transf <- param[["transform"]]
  if (transf == "log")
    return(round(transform_from_log(probs, transf, lower, upper), digits))
  probs <- probs * (upper - lower) + lower
  clamp(round(probs, digits), lower, upper)
}

#' Print parameter space in the textual format accepted by irace.
#'
#' @param parameters `ParameterSpace`\cr Data structure containing the parameter
#'   space definition. The data structure has to similar to the one returned by the
#'   function [`readParameters`].
#'
#'
#' @return `character()`
#'
#' @seealso [readParameters()]
#' @examples
#' parameters_table <- '
#'  # name       switch           type  values               [conditions (using R syntax)]
#'  algorithm    "--"             c     (as,mmas,eas,ras,acs)
#'  localsearch  "--localsearch " c     (0, 1, 2, 3)
#'  alpha        "--alpha "       r     (0.00, 5.00)
#'  beta         "--beta "        r     (0.00, 10.00)
#'  rho          "--rho  "        r     (0.01, 1.00)
#'  ants         "--ants "        i,log (5, 100)
#'  q0           "--q0 "          r     (0.0, 1.0)           | algorithm == "acs"
#'  q0dep       "--q0 "           r     (0.0, q0)            | algorithm != "acs"
#'  rasrank      "--rasranks "    i     (1, "min(ants, 10)") | algorithm == "ras"
#'  elitistants  "--elitistants " i     (1, ants)            | algorithm == "eas"
#'  nnls         "--nnls "        i     (5, 50)              | localsearch %in% c(1,2,3)
#'  dlb          "--dlb "         c     (0, 1)               | localsearch %in% c(1,2,3)
#'
#'  [forbidden]
#'  (alpha == 0.0) & (beta == 0.0)
#' '
#' parameters <- readParameters(text=parameters_table)
#' printParameters(parameters)
#' @export
printParameters <- function(parameters)
  cat(parameters$as_character())

#' @export
print.ParameterSpace <- function(x, digits = 15L, ...)
{
  utils::str(x$.params, digits.d = digits)
  cat("Forbidden:\n")
  print(x$forbidden, digits = 15L)
}

## digits <- 4L
## x <- parametersNew(param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--"),
##                    param_ord(name = "localsearch", values = c("0", "1", "2", "3"), label = "--localsearch "),
##                    param_real(name = "alpha", lower = 0.0, upper=5.0, label = "--alpha ", digits = digits),
##                    param_real(name = "beta", lower = 0.0, upper = 10.0, label = "--beta ", digits = digits),
##                    param_real(name = "rho", lower = 0.01, upper = 1.00, label = "--rho ", digits = digits),
##                    param_int(name = "ants", lower = 5, upper = 100, transf = "log", label = "--ants "),
##                    param_real(name = "q0", label = "--q0 ", lower=0.0, upper=1.0, condition = expression(algorithm == "acs")),
##                    param_int(name = "rasrank", label = "--rasranks ", lower=1, upper=quote(min(ants, 10)), condition = 'algorithm == "ras"'),
##                    param_int(name = "elitistants", label = "--elitistants ", lower=1, upper=expression(ants), condition = 'algorithm == "eas"'),
##                    param_int(name = "nnls", label = "--nnls ", lower = 5, upper = 50, condition = expression(localsearch %in% c(1,2,3))),
##                    param_cat(name = "dlb",  label = "--dlb ", values = c(0,1), condition = "localsearch %in% c(1,2,3)"),
##                    param_cat(name = "fixed1",  label = "--fixed1 ", values = "1"),
##                    param_cat(name = "fixed2",  label = "--fixed2 ", values = "0"),
##                    forbidden = "(alpha == 0) & (beta == 0)")


## printParameters(x)
## print(x)

withr::with_output_sink("test-sobol.Rout", {

  test_that("bug with conditional dependent", {
    parameters  <- parametersNew(param_cat(name = "algorithm", values = c("as", "mmas", "ras", "acs")),
      param_real(name = "alpha", lower = 0.0, upper=5.0),
      param_real(name = "beta", lower = 0.0, upper = 10.0),
      param_int(name = "ants", lower = 2, upper = 100),
      param_real(name = "q0", lower=0.0, upper=1.0, condition = expression(algorithm == "acs")),
      param_int(name = "rasrank", lower=1, upper=quote(min(ants, 10)), condition = 'algorithm == "ras"'),
      param_int(name = "eants", lower=0, upper=expression(rasrank)),
      param_cat(name = "dlb",  values = c(0,1), condition = "localsearch == 1"),
      param_int(name = "nnls", lower = 5, upper = 50, condition = expression(dlb == 1)),
      param_ord(name = "localsearch", values = c("0", "1")),
      param_cat(name = "fixed",  values = "0"))

    confs <- irace:::sampleSobol(parameters, 1000L)
    expect_equal(nrow(confs), 1000L)
    expect_valid_configurations(confs, parameters)
  })

  test_that("bug with dependent fixed", {
    params <- readParameters(text='
ROOT		"ROOT="		c		("SimpleAlgorithm")
ROOT_SimpleAlgorithm.constructive		"ROOT_SimpleAlgorithm.constructive="		c		("FasterInvertedConstructive", "SlowConstructive")		| ROOT %in% c("SimpleAlgorithm")
ROOT_SimpleAlgorithm.constructive_FasterInvertedConstructive.sumThis		"ROOT_SimpleAlgorithm.constructive_FasterInvertedConstructive.sumThis="		r		(-50.0, 50.0)		| ROOT_SimpleAlgorithm.constructive %in% c("FasterInvertedConstructive")
ROOT_SimpleAlgorithm.constructive_SlowConstructive.sumThis		"ROOT_SimpleAlgorithm.constructive_SlowConstructive.sumThis="		i		(-10, 10)		| ROOT_SimpleAlgorithm.constructive %in% c("SlowConstructive")
ROOT_SimpleAlgorithm.improver		"ROOT_SimpleAlgorithm.improver="		c		("FlippyFlopImprover")		| ROOT %in% c("SimpleAlgorithm")
ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.enabled		"ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.enabled="		c		("true", "false")		| ROOT_SimpleAlgorithm.improver %in% c("FlippyFlopImprover")
ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.sleepy		"ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.sleepy="		c		("8", "6", "12", "11", "7", "5", "4", "10", "1", "9", "2", "3", "13")		| ROOT_SimpleAlgorithm.improver %in% c("FlippyFlopImprover")
')
    confs <- irace:::sampleSobol(params, 10L)
    expect_equal(nrow(confs), 10L)
    expect_valid_configurations(confs, params)
  })
  test_that("bug with dependent fixed #2", {
    params <- readParameters(text='
ROOT		"ROOT="		c		("SimpleAlgorithm", "ComplexAlgorithms")
ROOT_SimpleAlgorithm.constructive		"ROOT_SimpleAlgorithm.constructive="		c		("FasterInvertedConstructive", "SlowConstructive")		| ROOT %in% c("SimpleAlgorithm")
ROOT_SimpleAlgorithm.constructive_FasterInvertedConstructive.sumThis		"ROOT_SimpleAlgorithm.constructive_FasterInvertedConstructive.sumThis="		r		(-50.0, 50.0)		| ROOT_SimpleAlgorithm.constructive %in% c("FasterInvertedConstructive")
ROOT_SimpleAlgorithm.constructive_SlowConstructive.sumThis		"ROOT_SimpleAlgorithm.constructive_SlowConstructive.sumThis="		i		(-10, 10)		| ROOT_SimpleAlgorithm.constructive %in% c("SlowConstructive")
ROOT_SimpleAlgorithm.improver		"ROOT_SimpleAlgorithm.improver="		c		("FlippyFlopImprover")		| ROOT %in% c("SimpleAlgorithm")
ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.enabled		"ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.enabled="		c		("true", "false")		| ROOT_SimpleAlgorithm.improver %in% c("FlippyFlopImprover")
ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.sleepy		"ROOT_SimpleAlgorithm.improver_FlippyFlopImprover.sleepy="		c		("8", "6", "12", "11", "7", "5", "4", "10", "1", "9", "2", "3", "13")		| ROOT_SimpleAlgorithm.improver %in% c("FlippyFlopImprover")
')
    confs <- irace:::sampleSobol(params, 10)
    expect_equal(nrow(confs), 10)
    expect_valid_configurations(confs, params)
  })

})

withr::with_output_sink("test-bug-87.Rout", {

test_that("bug-87", {
  parameters <- parametersNew(
    param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--"),
    param_ord(name = "localsearch", values = c("0", "1", "2", "3"), label = "--localsearch "),
    param_real(name = "alpha", lower = 0.0, upper = 5.0, label = "--alpha "),
    param_real(name = "beta", lower = 0.0, upper = 10.0, label = "--beta "),
    param_real(name = "rho", lower = 0.01, upper = 1.00, label = "--rho "),
    param_int(name = "ants", lower = 5, upper = 100, transf = "log", label = "--ants "),
    param_real(name = "q0", label = "--q0 ", lower = 0.0, upper = 1.0,
      condition = expression(algorithm == "acs")),
    param_int(name = "rasrank", label = "--rasranks ", lower = 1, upper = quote(min(ants, 10)),
      condition = 'algorithm == "ras"'),
    param_int(name = "elitistants", label = "--elitistants ", lower = 1, upper = expression(ants),
      condition = 'algorithm == "eas"'),
    param_int(name = "nnls", label = "--nnls ", lower = 5, upper = 50,
      condition = expression(localsearch %in% c("1", "2", "3"))),
    param_cat(name = "dlb", label = "--dlb ", values = c("0", "1"),
      condition = "localsearch %in% c('1','2','3')"),
    forbidden = "(alpha == 0) & (beta == 0)"
  )

  logFile <- withr::local_tempfile(fileext=".Rdata")
  scenario <- defaultScenario(list(parameters = parameters, instances = 1,
    maxExperiments = 300, logFile = logFile,
    targetRunner = function(experiment, scenario) {
      list(cost = experiment$configuration$alpha * experiment$configuration$beta)
    }))
  confs <- irace(scenario)
  best_conf <- getFinalElites(scenario$logFile, n = 1L, drop.metadata = TRUE)
  expect_identical(removeConfigurationsMetaData(confs[1L, , drop = FALSE]), best_conf)
})
}) # withr::with_output_sink()

withr::with_output_sink("test-ablation.Rout", {
  skip_on_cran()
  withr::local_options(warn=2)

  test_that("generateAblation", {
    parameters  <- parametersNew(param_cat(name = "algorithm", values = c("as", "mmas", "ras", "acs")),
      param_real(name = "alpha", lower = 0.0, upper=5.0),
      param_real(name = "beta", lower = 0.0, upper = 10.0),
      param_int(name = "ants", lower = 2, upper = 100),
      param_real(name = "q0", lower=0.0, upper=1.0, condition = expression(algorithm == "acs")),
      param_int(name = "rasrank", lower=1, upper=quote(min(ants, 10)), condition = 'algorithm == "ras"'),
      param_int(name = "eants", lower=0, upper=expression(rasrank)),
      param_cat(name = "dlb",  values = c(0,1), condition = "localsearch == 1"),
      param_int(name = "nnls", lower = 5, upper = 50, condition = expression(dlb == 1)),
      param_ord(name = "localsearch", values = c("0", "1")),
      param_cat(name = "fixed",  values = "0"),
      forbidden = "(alpha == 0) & (beta == 0)")

    confs <- readConfigurationsFile(parameters = parameters, text ='
algorithm alpha beta ants q0  rasrank eants dlb nnls localsearch
ras       1     0    5    NA  5       5     1   5    1
ras       0.5   0    5    NA  2       2     1   5    1
ras       0.5   0    4    NA  2       2     1   5    1
acs       0     1    2    0.5 NA      NA    NA  NA   0
mmas      1     1    6    NA  NA      NA    0   NA   1
')
    confs[[".ID."]] <- seq_len(nrow(confs))
    colClasses <- c(localsearch="character", q0="numeric",
      rasrank="integer", eants="integer", dlb="character", nnls="integer")

    check_generate_ablation <- function(src, target, configurations_table, changed) {
      aux <- irace:::generate_ablation(confs[src, , drop=FALSE], confs[target, , drop=FALSE],
        parameters, param_names = parameters$names_variable)
      expect_valid_configurations(aux$configurations, parameters)
      configurations <- read.table(header=TRUE, colClasses=colClasses, text=configurations_table)
      configurations[["fixed"]] <- "0"
      configurations[[".ID."]] <- src
      configurations[[".PARENT."]] <- src
      expect_equal(aux, list(configurations=configurations, changed_params = changed))
    }

    check_generate_ablation(1L, 2L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      ras   0.5    0    5 NA       5     5   1    5           1
      ras   1.0    0    5 NA       2     2   1    5           1
      ras   1.0    0    5 NA       5     2   1    5           1
', changed = list("alpha", c("rasrank", "eants"), "eants"))

    check_generate_ablation(1L, 3L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      ras   0.5    0    5 NA       5     5   1    5           1
      ras   1.0    0    4 NA       2     2   1    5           1
      ras   1.0    0    5 NA       2     2   1    5           1
      ras   1.0    0    5 NA       5     2   1    5           1
', changed = list("alpha", c("ants", "rasrank", "eants"), c("rasrank", "eants"),"eants"))

    check_generate_ablation(1L, 4L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      acs   1.0    0    5 0.5     NA    NA   1    5           1
      ras   1.0   1.0   5 NA       5     5   1    5           1
      ras   1.0    0    5 NA       5     5  NA   NA           0
', changed = list(c("algorithm", "q0", "rasrank", "eants"), "beta",
  c("localsearch", "dlb", "nnls")))

    check_generate_ablation(1L, 5L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
     mmas   1.0    0    5 NA      NA    NA   1    5           1
      ras   1.0   1.0   5 NA       5     5   1    5           1
      ras     1    0    6 NA       5     5   1    5           1
      ras   1.0    0    5 NA       5     5   0   NA           1
', changed = list(c("algorithm", "rasrank", "eants"), "beta", "ants", c("dlb", "nnls")))

    check_generate_ablation(2L, 1L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      ras   1.0    0    5 NA       2     2   1    5           1
      ras   0.5    0    5 NA       5     2   1    5           1
', changed = list("alpha", c("rasrank")))

    check_generate_ablation(3L, 1L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      ras   1.0    0    4 NA       2     2   1    5           1
      ras   0.5    0    5 NA       2     2   1    5           1
', changed = list("alpha", c("ants")))

    check_generate_ablation(4L, 1L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      acs     1    1    2 0.5     NA    NA  NA   NA           0
      acs     0    1    5 0.5     NA    NA  NA   NA           0
      acs     0    1    2 0.5     NA    NA    1   5           1
', changed = list("alpha", "ants", c("localsearch", "dlb", "nnls")))

    check_generate_ablation(5L, 1L, '
algorithm alpha beta ants q0 rasrank eants dlb nnls localsearch
      ras     1    1    6 NA       5     5   0   NA           1
     mmas     1    0    6 NA      NA    NA   0   NA           1
     mmas     1    1    5 NA      NA    NA   0   NA           1
     mmas     1    1    6 NA      NA    NA   1    5           1
', changed = list(c("algorithm", "rasrank", "eants"), "beta", "ants", c("dlb", "nnls")))

  })

  test_that("--help", {
    expect_output(ablation_cmdline("--help"))
  })

  outfile <- withr::local_tempfile(pattern = "log-ablation", fileext = ".Rdata")
  logfile <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")

  parameters  <- parametersNew(
    param_cat("cat", values = c("0", "1", "2", "3", "4")),
    param_real("real", lower = 0.0, upper=1.0),
    param_int("int", lower = 100, upper = 500),
    param_cat("bool", values = c("0", "1")))

  default <- data.frame(cat="4", real=1.0, int=500L, bool = "1")

  target_runner <- function(experiment, scenario) {
    conf <- experiment$configuration
    instance <- experiment$instance
    seed <- experiment$seed
    k <- if (as.logical(as.integer(conf[["bool"]]))) 1000 else 100
    list(cost = instance + 1/seed + k * (conf[["int"]] + as.integer(conf[["cat"]]) + (conf[["real"]]-0.5)^2))
  }

  check_log <- function(log) {
    instances_log <- log$state$instances_log
    instances_log[, instance:=.I]
    experiment_log <- log$state$experiment_log[instances_log, on="instance"]
    experiment_log[["instance_value"]] <- log$scenario$instances[experiment_log[["instanceID"]]]
    experiment_log <- experiment_log[log$allConfigurations, on = c(configuration=".ID.")]
    experiments <- log$experiments
    experiments = data.table(
      instance = rep(seq_len(nrow(experiments)), ncol(experiments)),
      configuration = rep(seq_len(ncol(experiments)), each = nrow(experiments)),
      cost3 = c(experiments)
    )
    experiments <- experiments[!is.na(experiments$cost3),]
    experiment_log <- experiment_log[experiments, on=.NATURAL]
    experiment_log[, cost2:=instance_value + 1/seed + fifelse(as.logical(as.integer(bool)), 1000, 100) * (int + as.integer(cat) + (real - 0.5)^2)]
    if ("bound" %in% colnames(experiment_log)) {
      experiment_log[, cost2 := pmin.int(cost2, bound)]
      experiment_log[, cost3 := pmin.int(cost3, bound)]
    }
    expect_equal(experiment_log[["cost"]], experiment_log[["cost2"]])
    expect_equal(experiment_log[["cost"]], experiment_log[["cost3"]])
  }

  src_file <- withr::local_tempfile(pattern="src", fileext=".txt",
    lines=c("cat real int bool", "4 1.0 500 1"))
  target_file <- withr::local_tempfile(pattern="target", fileext=".txt",
    lines=c("cat real int bool", "0 0.0 100 0"))

  test_that("ablation maxTime", {
    target_runner_time <- function(experiment, scenario)
      list(cost = target_runner(experiment, scenario)$cost,
        time = runif(1, min=0.1, max=1))

    scenario <- list(targetRunner = target_runner_time,
      instances = seq(1000, 10000, 1000),
      seed = 42,
      maxTime = 1000,
      initConfigurations = default,
      logFile = logfile,
      parameters = parameters)
    scenario <- checkScenario (scenario)
    irace(scenario = scenario)
    check_log(read_logfile(logfile))
    res <- ablation(logfile, ablationLogFile = outfile)
    check_log(res)
    expect_true(res$complete)
    res <- ablation(logfile, ablationLogFile = outfile, type = "racing")
    check_log(res)
    expect_true(res$complete)

    res <- ablation(logfile, ablationLogFile = outfile, src = src_file, target = target_file)
    check_log(res)
    expect_true(res$complete)

  })

  test_that("ablation capping", {
    target_runner_capping <- function(experiment, scenario) {
      cost <- min(experiment$bound, target_runner(experiment, scenario)$cost)
      list(cost = cost, time = cost)
    }

    boundMax <- 1000 + (1000 * 505)
    scenario <- list(targetRunner = target_runner_capping,
      instances = seq(1000, 10000, 1000),
      seed = 42,
      maxTime = 100 * boundMax, boundMax = boundMax,
      initConfigurations = default,
      logFile = logfile,
      parameters = parameters)
    scenario <- checkScenario(scenario)
    expect_warning(irace(scenario = scenario), "is too large")
    check_log(read_logfile(logfile))

    res <- ablation(logfile, ablationLogFile = outfile)
    check_log(res)
    expect_true(res$complete)

    res <- ablation(logfile, ablationLogFile = outfile, type = "racing")
    check_log(res)
    expect_true(res$complete)
  })

  test_that("ablation maxExperiments", {
    scenario <- list(targetRunner = target_runner,
      instances = seq(1000, 10000, 1000),
      maxExperiments = 1000, seed = 42,
      initConfigurations = default,
      logFile = logfile,
      parameters = parameters)
    scenario <- checkScenario (scenario)
    irace(scenario = scenario)
    check_log(read_logfile(logfile))

    res <- ablation(logfile, ablationLogFile = outfile)
    check_log(res)
    expect_true(res$complete)

    res <- ablation(logfile, ablationLogFile = outfile, type = "racing")
    check_log(res)
    expect_true(res$complete)

    res <- ablation(logfile, ablationLogFile = outfile, src = src_file, target = target_file)
    check_log(res)
    expect_true(res$complete)

    plotfile <- withr::local_tempfile(pattern = "ablation", fileext = ".pdf")
    res <- ablation_cmdline(paste0("--log-file=", logfile, " -o ", outfile, " -p ", plotfile))
    check_log(res)
    expect_true(res$complete)
  })

}) # withr::with_output_sink()

withr::with_output_sink("test-readParameters.Rout", {

  test_that("optimize conditions", {
    p <- readParameters(text='
ROOT		"" c ("S")
ROOT_T		"" c ("FIC")		| ROOT %in% c("S")
ROOT_T_FIC.sum	"" r (-50.0, 50.0)	| ROOT_T %in% c("FIC")
ROOT_T.FFI	"" c ("true")		| ROOT_T == "FIC"
ROOT_T.im	"" c ("FFI", "FIC")	| ROOT == "S" & ROOT_T.FFI == "true"
ROOT_E.FFI	"" c ("false")          | ROOT_T_FIC.sum < 0
ROOT_E.FIC	"" i (0,100)            | ROOT_E.FFI == "false"
')
    expect_identical(p$conditions,
      list(ROOT=TRUE, ROOT_T=TRUE, ROOT_T_FIC.sum=TRUE,
        ROOT_T.FFI=TRUE, ROOT_T.im=TRUE,
        ROOT_E.FFI=expression(ROOT_T_FIC.sum < 0),
        ROOT_E.FIC=expression(ROOT_E.FFI == "false")))
  })
  
  test_that("error checking", {

    ref <- parametersNew(param_real(name = "tmp", lower = 0, upper=1,
      label = "", digits = 5), forbidden = "tmp == 0")

    expect_identical(ref, readParameters(text='
tmp "" r (0, 1)

[global]
digits = 5
[forbidden]
tmp == 0
'))

    expect_identical(ref, readParameters(text='
tmp "" r (0, 1)
[forbidden]
tmp == 0
[global]
digits = 5
'))

    expect_identical(ref, readParameters(text='
tmp "" r (0, 1)

[forbidden]
tmp == 0

[global]

digits = 5
'))

expect_error(readParameters(text = '
rest_t "--restart_temp_ratio="  r,log (0.00001, 0.9)
[global]
digits = 4
'), "Domain bounds")

expect_no_error(readParameters(text = '
rest_t "--restart_temp_ratio="  r,log (0.00001, 0.9)
[global]
digits = 5
'))

expect_error(readParameters(text = '
temp "" r (0, 10)
t-max "" i (1, 50)
'), "name must be alphanumeric")

expect_error(readParameters(text = '
temp "" r (0, 10)
[forbidden]
tmp == 0
'), "contains unknown parameter(s): tmp", fixed = TRUE)

expect_error(readParameters(text = '
temp "" r (0, 10)
temp "" i (1, 50)
'), "Duplicated parameter name")

expect_error(readParameters(text = '
temp "" c ("a", "b", "a")
'), "duplicated values")

expect_error(readParameters(text = '
temp --temp i (0, 10)
'), "Parameter label (switch) must be a double-quoted string", fixed = TRUE)

expect_error(readParameters(text = '
temp "--temp" i,lag (0, 10)
'), "Parameter type must be a single character in .'c','i','r','o'.")

expect_error(readParameters(text = '
temp "--temp" i c(0, 10)
'), "Allowed values must be a list within parenthesis at line 2")

expect_error(readParameters(text = 'param1 "--param1 " c "a,b,c"'),
             "Allowed values must be a list within parenthesis at line 1")

expect_error(readParameters(text = 'param1 "--param1 " i (1,2,3 )'),
             "Incorrect numeric range")
expect_error(readParameters(text = 'param1 "--param1 " i ( 1,10) |'),
             "Expected condition before '|'")
expect_error(readParameters(text = 'param1 "--param1 " i ( 1, 10) param1 < param2'),
             "Expected condition after '|'")

expect_error(readParameters(text = '
param1 "--param1 " i (0,2)
param2 "--param2 " r (0,1) | param1 <> 1'),
"Invalid condition 'param1 <> 1': ")

expect_error(readParameters(text = '
param1 "--param1 " i (0,2)
param2 "--param2 " r (0,2) | param1 < param2'),
"Cycle detected.+param2")

expect_error(readParameters(text = '
param1 "--param1 " i (0,2)
param2 "--param2 " r (0,param2)'),
"Cycle detected.+param2")

expect_error(readParameters(text = '
param1 "--param1 " i (0,2)
param2 "--param2 " r (0,2) | param1 < param3'),
"Parameter 'param2' depends on 'param3' which is not defined")

expect_error(readParameters(text = '
param1 "--param1 " i (0,2)
param2 "--param2 " r (0,param3)'),
"'param3' cannot be found in the scenario parameters")

expect_error(readParameters(text = '
param1 "--param1 " c (a,b)
param2 "--param2 " r (0,param1)'),
"Domain of parameter 'param2' depends on non-numerical parameters: param1")

expect_error(readParameters(text = 'param1 "--param1 " i (1.1,2.0)'),
"For parameter 'param1' of type 'i' values must be integers")

expect_error(readParameters(text = 'param1 "--param1 " r (0.01, 0.001)'),
"Lower bound must be smaller than upper bound in numeric range")

expect_error(readParameters(text = 'param1 "--param1 " i (1, 1)'),
"Lower bound must be smaller than upper bound in numeric range")

expect_error(readParameters(text = '
#

#
'), "No parameter definition found")

test_that("ordinal out of order", {
expect_error(readParameters(text = '
param "" o (0, 2, 1)
'), "the values are not in increasing order")
})

expect_error(readParameters(text = 'param1 "--param1 " r,log (0, 100)'),
             "of parameter of type 'log' contains non-positive values")

expect_error(readParameters(text = 'param1 "--param1 " r,log (0.0001, 0.99999)', digits=3),
             "must be representable within the given 'digits=3'; you would need at least 'digits=5'")

expect_error(param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--", condition = "a == 0 && b == 2"),
             "Please use '&' and '|' instead of '&&' and '|' in: a == 0 && b == 2", fixed = TRUE)

expect_error(param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--", condition = quote(a == 0 || b == 2)),
             "Please use '&' and '|' instead of '&&' and '|' in: a == 0 || b == 2", fixed = TRUE)

expect_error(param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--", condition = 0),
             "Invalid condition", fixed = TRUE)

expect_error(parametersNew(param_real(name = "alpha", lower = 0.0, upper=5.0),
                           param_real(name = "beta", lower = 0.0, upper = 10.0),
                           forbidden = expression((alpha == 0) || (beta == 0))),
             "Please use '&' and '|' instead of '&&' and '|' in: (alpha == 0) || (beta == 0)", fixed = TRUE)

expect_error(parametersNew(param_real(name = "alpha", lower = 0.0, upper=5.0),
                           param_real(name = "beta", lower = 0.0, upper = 10.0),
                           forbidden = quote((alpha == 0) && (beta == 0))),
             "Please use '&' and '|' instead of '&&' and '|' in: (alpha == 0) && (beta == 0)", fixed = TRUE)

expect_error(parametersNew(param_real(name = "alpha", lower = 0.0, upper=5.0),
                           param_real(name = "beta", lower = 0.0, upper = 10.0),
                           forbidden = TRUE),
             "Invalid forbidden expression", fixed = TRUE)

expect_identical(param_real(name = "p", lower = "-1", upper = "10")$domain, c(-1,10))
expect_identical(param_int(name = "p", lower = "-1", upper = "10")$domain, c(-1L,10L))
expect_identical(param_int(name = "p", lower = -1, upper = 10)$domain, c(-1L,10L))
expect_identical(param_int(name = "p", lower = "-1", upper = expression(a))$domain[[1L]], -1L)
expect_identical(
  param_int(name = "p", lower = "-1", upper= "a"),
  param_int(name = "p", lower = "-1", upper= expression(a)))
expect_identical(
  param_int(name = "p", lower = -1, upper= "a"),
  param_int(name = "p", lower = "-1", upper= "a"))

})
})

withr::with_output_sink("test-repair.Rout", {

repair_irace <- function(targetRunner, repair)
{
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  parameters <- readParameters(text = '
   p1 "" r (0,1)
   p2 "" r (0,1)
   p3 "" r (0,1)
   dummy "" c ("d1", "d2")
   ')
  scenario <- list(targetRunner = targetRunner,
                   repairConfiguration = repair,
                   instances = weights,
                   maxExperiments=180,
                   seed = 1234567,
                   parameters = parameters)
  scenario <- checkScenario (scenario)
  expect_true(irace:::checkTargetFiles(scenario = scenario))
  confs <- irace(scenario = scenario)
  expect_gt(nrow(confs), 0L)
  confs
}

target_sum2one <- function(experiment, scenario)
{
  configuration <- experiment$configuration
  p1 <-  configuration[["p1"]]
  p2 <-  configuration[["p2"]]
  p3 <-  configuration[["p3"]]
  stopifnot(isTRUE(all.equal(p1+p2+p3, 1.0)))
  list(cost = -p1, call = toString(experiment))
}

repair_sum2one <- function(configuration, parameters)
{
  isreal <- parameters$names[parameters$types == "r"]
  digits <- sapply(isreal, function(x) parameters$get(x)[["digits"]])
  c_real <- unlist(configuration[isreal])
  c_real <- c_real / sum(c_real)
  c_real[-1] <- round(c_real[-1], digits[-1])
  c_real[1L] <- 1 - sum(c_real[-1])
  configuration[isreal] <- c_real
  configuration
}

target_order <- function(experiment, scenario)
{
  configuration <- experiment$configuration
  p1 <- configuration[["p1"]]
  p2 <- configuration[["p2"]]
  p3 <- configuration[["p3"]]
  stopifnot(p1 <= p2 && p2 <= p3)
  list(cost = -p1, call = toString(experiment))
}

repair_order <- function(configuration, parameters)
{
 columns <- c("p1","p2","p3")
 #cat("Before"); print(configuration)
 configuration[columns] <- sort(unlist(configuration[columns], use.names=FALSE))
 #cat("After"); print(configuration)
 configuration
}

test_that("repair: sum to one", {
  generate_set_seed()
  confs <- repair_irace(target_sum2one, repair_sum2one)
  expect_equal(unique(apply(confs[, c("p1", "p2", "p3")], 1L, sum)), 1)
})

test_that("repair: increasing order", {
  generate_set_seed()
  confs <- repair_irace(target_order, repair_order)
  expect_true(all(apply(confs[, c("p1", "p2", "p3")], 1L, diff) >= 0))
})

}) # withr::with_output_sink()

#######################################
## GENERATE CONFIGURATIONS
#######################################

## When called with an unconditional parameter, it must return TRUE.
conditionsSatisfied <- function(condition, partialConfiguration)
{
  # If there is no condition, do not waste time evaluating it.
  if (isTRUE(condition)) return(TRUE)

  v <- eval(condition, as.list(partialConfiguration))
  # Return TRUE if TRUE, FALSE if FALSE or NA
  ## FIXME: If we byte-compile the condition, then we should incorporate the
  ## following into the condition directly.
  !is.na(v) && v
}

repairConfigurations <- function(x, parameters, repair)
{
  if (!is.null(repair)) {
    # FIXME: Pass the whole newConfigurations to repair and let it handle each row.
    j <- parameters$names
    df <- as.data.frame(x)
    for (i in seq_nrow(x))
      set(x, i = i, j = j, value = repair(df[i, parameters$names], parameters))
  }
  x
}

is_within_dependent_bound <- function(param, configuration, value)
{
  domain <- unlist(lapply(param[["domain"]], eval, configuration))
  # Value gets truncated (defined from robotics initial requirements)
  if (param[["type"]] == "i")
    domain <- as.integer(domain)
  domain[[1L]] <= value && value <= domain[[2L]]
}

## Calculates the parameter bounds when parameters domain is dependent
getDependentBound <- function(param, configuration)
{
  domain <- param[["domain"]]
  if (is.expression(domain)) {
    deps <- all.vars(domain)
    # If it depends on a parameter that is disabled, then this is disabled.
    if (anyNA(configuration[deps])) return(NA)
    domain <- sapply(domain, eval, configuration)
    irace_assert(all(is.finite(domain)))
    # Value gets truncated (defined from robotics initial requirements)
    if (param[["type"]] == "i") domain <- as.integer(domain)
    if (domain[[1L]] > domain[[2L]]) {
      irace_error ("Invalid domain (", paste0(domain, collapse=", "),
                   ") generated for parameter '", param[["name"]],
                   "' that depends on parameters (", paste0(deps, collapse=", "),
                   "). This is NOT a bug in irace. Check the definition of these parameters.")
    }
  }
  domain
}

## Calculates the parameter bounds when the parameter is dependent.
get_dependent_domain <- function(param, configuration)
{
  # FIXME: Make this function handle a data.table and return a list of domains.
  configuration <- as.list(configuration)
  domain <- param[["domain"]]
  deps <- all.vars(domain)
  # FIXME: This function should not be called if the parent is disabled.
  # If it depends on a parameter that is disabled, then this is disabled.
  if (anyNA(configuration[deps]))
    return(NA)

  irace_assert(is.expression(domain))
  domain <- sapply(domain, eval, configuration, USE.NAMES=FALSE)
  irace_assert(all(is.finite(domain)))
  # Value gets truncated (defined from robotics initial requirements)
  if (param[["type"]] == "i")
    domain <- as.integer(domain)
  if (domain[[1L]] > domain[[2L]]) {
    # FIXME: Add test for this error.
    irace_error ("Invalid domain (", paste0(domain, collapse=", "),
                 ") generated for parameter '", param[["name"]],
                 "' that depends on parameters (", paste0(deps, collapse=", "),
                 "). This is NOT a bug in irace. Check the definition of these parameters.")
  }
  domain
}


generate_sobol <- function(parameters, n, repair = NULL)
{
  confs <- spacefillr::generate_sobol_set(n, dim = parameters$nbVariable,
    seed = runif_integer(size = 1L))
  confs <- data.table(confs)
  setnames(confs, parameters$names_variable)
  hierarchy <- parameters$hierarchy
  nodep_names <- names(hierarchy)[(hierarchy == 1L) & !parameters$isFixed]
  # FIXME: How to do this faster using data.table?
  for (x in nodep_names)
    set(confs, j = x, value = param_quantile(parameters$get(x), confs[[x]]))

  for (x in parameters$names_fixed)
    set(confs, j = x, value = parameters$domains[[x]])

  setcolorder(confs, parameters$names)
  max_level <- max(hierarchy)
  if (max_level > 1L) {
    .NEWVALUE <- .DOMAIN <- NULL # To silence CRAN warnings.
    for (k in seq_len(max_level - 1L)) {
      prev_names <- names(hierarchy)[hierarchy <= k]
      dep_names <- names(hierarchy)[hierarchy == k+1L]
      for (p in dep_names) {
        param <- parameters$get(p)
        if (param$isFixed) {
          if (!isTRUE(param[["condition"]]))
            # If somehow this fixed parameter was not satisfied sometimes, just set its value to NA.
            set(confs, which(!eval(param[["condition"]], confs)), j = p, value = NA_character_)
          next
        }
        idx_satisfied <- which_satisfied(confs, param[["condition"]])
        if (length(idx_satisfied)) {
          if (param[["is_dependent"]]) {
            confs[idx_satisfied, let(.DOMAIN = list(get_dependent_domain(param, .SD))), by=.I, .SDcols=prev_names]
            confs[idx_satisfied, .NEWVALUE := param_quantile(param, .SD, domain = unlist(.DOMAIN)), by=.I, .SDcols=p]
            set(confs, j = ".DOMAIN", value = NULL)
          } else {
            confs[idx_satisfied, .NEWVALUE := param_quantile(param, unlist(.SD)), .SDcols=p]
          }
          set(confs, j = c(p, ".NEWVALUE"), value = list(confs[[".NEWVALUE"]], NULL))
        } else {
          set(confs, j = p, value = .param_na_value_type[[ param[["type"]] ]])
        }
      }
    }
  }
  repairConfigurations(confs, parameters, repair)
  confs
}

sampleSobol <- function(parameters, n, repair = NULL)
{
  newConfigurations <- generate_sobol(parameters, n, repair)
  newConfigurations <- unique(newConfigurations)
  forbidden <- parameters$forbidden
  newConfigurations <- filter_forbidden(newConfigurations, forbidden)
  have <- nrow(newConfigurations)
  if (have < n) {
    needed <- max(ceiling(n + (n - have) * (2 - have / n)), min(parameters$nbVariable * 5L, 100L))
    newConfigurations <- generate_sobol(parameters, needed, repair)
    newConfigurations <- unique(newConfigurations)
    newConfigurations <- filter_forbidden(newConfigurations, forbidden)
    if (nrow(newConfigurations) == 0L) {
      irace_error("irace tried to sample a configuration not forbidden without success, perhaps your constraints are too strict?")
    }
    newConfigurations <- truncate_rows(newConfigurations, n)
  }
  set(newConfigurations, j = ".PARENT.", value = NA_integer_)
  newConfigurations
}


generate_uniform <- function(parameters, nbConfigurations, repair = NULL)
{
  newConfigurations <- configurations_alloc(parameters[["names"]],
    nrow = nbConfigurations, parameters = parameters)
  # We sample in the order of the conditions.
  for (param in parameters$get_ordered()) {
    pname <- param[["name"]]
    idx <- which_satisfied(newConfigurations, param[["condition"]])
    if (length(idx) == 0L)
      next
    if (param[["isFixed"]]) {
      # We don't need to sample, there is only one value.
      set(newConfigurations, i = idx, j = pname, value = param[["domain"]])
      next
    }
    if (param[["is_dependent"]]) {
      sample_dep_unif <- function(x)
        sample_unif(param, n = 1L, domain = get_dependent_domain(param, x))
      newConfigurations[idx, c(pname) := sample_dep_unif(.SD), by=.I, .SDcols=parameters$depends[[pname]]]
      ## newVals <- sapply(idx, function(i) {
      ##   domain <- get_dependent_domain(param, newConfigurations[i,])
      ##   runif(param, n = 1L, domain = domain)
      ## })
    } else {
      set(newConfigurations, i = idx, j = pname, value = sample_unif(param, n = length(idx)))
    }
  }
  repairConfigurations(newConfigurations, parameters, repair)
  newConfigurations
}


### Uniform sampling for the initial generation
sampleUniform <- function(parameters, nbConfigurations, repair = NULL)
{
  newConfigurations <- generate_uniform(parameters, nbConfigurations, repair)
  forbidden <- parameters$forbidden
  if (!is.null(forbidden)) {
    retries <- 100L
    repeat {
      newConfigurations <- filter_forbidden(newConfigurations, forbidden)
      needed <- nbConfigurations - nrow(newConfigurations)
      if (needed == 0L)
        break
      newConfigurations <- rbindlist(list(newConfigurations,
        generate_uniform(parameters, needed, repair = repair)))
      retries <- retries - 1L
      if (retries == 0L) {
        irace_error("irace tried 100 times to sample uniformly a configuration not forbidden without success, perhaps your constraints are too strict?")
      }
    }
  }
  set(newConfigurations, j = ".PARENT.", value = NA_integer_)
  newConfigurations
}

sample_from_model <- function(parameters, eliteConfigurations, model,
                              nbNewConfigurations, repair = NULL)
{
  # FIXME: We only need .WEIGHT. from eliteConfigurations.
  ids_elites <- names(model[[1L]])
  irace_assert(identical(as.integer(ids_elites), as.integer(eliteConfigurations[[".ID."]])), {
    print(utils::str(ids_elites))
    print(utils::str(eliteConfigurations[[".ID."]]))
  })

  newConfigurations  <- configurations_alloc(parameters$names, nrow = nbNewConfigurations, parameters)
  idx_elites <- sample.int(n = length(ids_elites), size = nbNewConfigurations,
                           prob = eliteConfigurations[[".WEIGHT."]], replace = TRUE)
  .PARENT. <- NULL # Silence CRAN warning.
  set(newConfigurations, j = ".PARENT.", value = ids_elites[idx_elites])
  # Sample a value for every parameter of the new configuration.
  for (param in parameters$get_ordered()) {
    idx_satisfied <- which_satisfied(newConfigurations, param[["condition"]])
    if (length(idx_satisfied) == 0L)
      next
    if (param[["isFixed"]]) {
      # We don't need to sample, there is only one value.
      set(newConfigurations, i = idx_satisfied, j = param[["name"]], value = param[["domain"]])
      next
    }
    pname <- param[["name"]]
    this_model <- model[[pname]]
    if (param[["is_dependent"]]) {
      sample_model <- if (param[["type"]] == "i") sample_model.ParamInt else sample_model.ParamReal
      dep_rmodel <- function(x, sd_mean) {
        domain <- get_dependent_domain(param, x)
        if (is.na(domain[[1L]])) return(NA)
        # If parameters are dependent standard deviation must be computed
        # based on the current domain
        sd <- (domain[[2L]] - domain[[1L]]) * sd_mean[[1L]]
        if (sd < .Machine$double.eps) return(domain[[1L]])
        sample_model(param, n = 1L, model = c(sd, sd_mean[[2L]]), domain = domain)
      }
      newConfigurations[idx_satisfied, c(pname) := dep_rmodel(.SD, this_model[[.PARENT.]]), by=.I, .SDcols=parameters$depends[[pname]]]
      next # We are done with this parameter.
    }
    # .BY is a list, so take the first argument.
    newConfigurations[idx_satisfied,
      c(pname) := list(sample_model(param, .N, this_model[[ .BY[[1L]] ]])),
      by = .PARENT.]
  }
  set(newConfigurations, j = ".PARENT.", value = as.integer(newConfigurations[[".PARENT."]]))
  repairConfigurations(newConfigurations, parameters, repair)
  newConfigurations
}

sampleModel <- function(parameters, eliteConfigurations, model,
                        nbNewConfigurations, repair = NULL)
{
  if (nbNewConfigurations <= 0)
    irace_error ("The number of configurations to generate appears to be negative or zero.")
  newConfigurations <- sample_from_model(parameters, eliteConfigurations,
                                         model, nbNewConfigurations, repair)
  forbidden <- parameters$forbidden
  if (!is.null(forbidden)) {
    retries <- 100L
    repeat {
      newConfigurations <- filter_forbidden(newConfigurations, forbidden)
      needed <- nbNewConfigurations - nrow(newConfigurations)
      if (needed == 0L)
        break
      tmp <- sample_from_model(parameters, eliteConfigurations, model, needed, repair = repair)
      newConfigurations <- rbindlist(list(newConfigurations, tmp))
      retries <- retries - 1L
      if (retries == 0L) {
        irace_error("irace tried 100 times to sample from the model a configuration not forbidden without success, perhaps your constraints are too strict?")
      }
    }
  }
  newConfigurations
}

withr::with_output_sink("test-dependencies.Rout", {

test_that("param depend error checking", {
  expect_error_readParameters <- function(text, error)
    expect_error(readParameters(text=text), error)

  expect_error_readParameters('
p1 "" r (0, 1)
p2 "" r (p3, 1)
', "parameter 'p2' is not valid: 'p3' cannot be found")
  
  expect_error_readParameters('
p1 "" c (0, 1)
p2 "" r (1, p1)
', "parameter 'p2' depends on non-numerical parameters")
  
  expect_error_readParameters('
p1 "" r (0, 1)
p2 "" r (1, foobar(p1))
', "parameter 'p2' uses function")

  expect_error_readParameters('
p1 "" r (0, p3)
p2 "" r (1, p1)
p3 "" r (p2, 1)
', "Cycle detected")

  expect_error_readParameters('
p1 "" r (0, 1)
p2 "" i (0.1, p1)
', "values must be integers")

})


checkConditionalAndDependency <- function(configuration, parameters)
{
  for (param in parameters$get_ordered()) {
    p <- param[["name"]]
    if (!irace:::conditionsSatisfied(param[["condition"]], configuration)) {
      expect(is.na(configuration[[p]]),
             paste0("Conditional parameter '", p, 
                    "' is not active but it has a value '", configuration[[p]], "' assigned."))
    } else if (param[["is_dependent"]]) {
      bounds <- irace:::getDependentBound(param, configuration)
      if (anyNA(bounds)) {
        expect(is.na(configuration[[p]]),
               paste0("Dependent parameter '", p, 
                      "' has a value '", configuration[[p]], "' but it should be inactive."))
        expect_true(anyNA(configuration[parameters$depends[[p]]]))
      } else {
        expect (configuration[[p]] >= bounds[1],
                paste0("Parameter '", p, "=", configuration[[p]], 
                       "' does not comply with dependency: ", parameters$depends[[p]],
                       " and lower bound: ", bounds[1]))
        expect (configuration[[p]] <= bounds[2],
                paste0("Parameter '", p, " = ", configuration[[p]], 
                       "' does not comply with dependency: ", parameters$depends[[p]],
                       " and upper bound: ", bounds[2]))
      }
    }
  }
}


test_that("test inactive dependent", {
  parameters <- readParameters(text='
p1 "" r (0,1)
p2 "" r (0, p1) | p1 < 0.5
p3 "" r (0, p2)
', digits = 2)
  confs <- irace:::sampleUniform(parameters, 50)
  confs <- as.data.frame(confs)
  for (i in seq_len(nrow(confs))) {
    checkConditionalAndDependency(confs[i,], parameters)
  }
  expect_error(readConfigurationsFile(parameters = parameters, text = '
p1  p2  p3
0.4 0.4 0.4
0.4 0.5 0.4
'), "Configuration number 2 is invalid because the value")

  expect_error(readConfigurationsFile(parameters = parameters, text = '
p1  p2  p3
0.4 0.3 0.2
0.4 0.3 0.4
'), "Configuration number 2 is invalid because the value")

    readConfigurationsFile(parameters = parameters, text = '
p1  p2  p3
0.4 0.3 0.2
0.5 NA  NA
0.4 0.4 0.4
')
})

test_checkDependencies <- function(parameterFile, ...)
{
  args <- list(...)
  
  target_runner <- function(experiment, scenario) {
    configuration <- experiment$configuration
    tmax <- configuration[["real"]]
    stopifnot(is.numeric(tmax))
    if (configuration[["mode"]] %in% c("x1", "x2"))
      temp <-  configuration[["param1"]]
    else
      temp <- 1
    stopifnot(is.numeric(temp))
    time <- max(1, abs(rnorm(1, mean=(tmax+temp)/10)))
    list(cost = time, time = time, call = toString(experiment))
  }
  
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  parameters <- readParameters(parameterFile)
  scenario <- list(targetRunner = target_runner, instances = weights,
    seed = 1234567, maxExperiments = 200, parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)
 
  nconf <- 100
  conf <- irace:::sampleUniform(parameters, nconf)
  expect_equal(nconf, nrow(conf))
  conf$.ID. <- seq_len(nconf)
  conf <- as.data.frame(conf)
  for (i in seq_len(nconf))
    checkConditionalAndDependency(conf[i,], parameters)
 
  model <- irace:::initialiseModel(parameters, conf)
  conf2 <- irace:::sampleModel(parameters, conf, model, nconf)
  conf2 <- as.data.frame(conf2)
  for (i in seq_len(nconf))
    checkConditionalAndDependency(conf2[i,], parameters)

  confs <- irace(scenario = scenario)
  for (i in seq_len(nrow(confs))) {
    checkConditionalAndDependency(confs[i,], parameters)
  }
}

test_that("checkDependencies", {
  test_checkDependencies(parameterFile="dependencies.txt")
})
test_that("checkDependencies2", {
  test_checkDependencies(parameterFile="dependencies2.txt")
})

}) # withr::with_output_sink()

context("bugs")

withr::with_output_sink("test-bugs.Rout", {

run_irace_from_rds <- function(rds_filename)
{
  # FIXME: For some unknown reason, this test doesn't work in R 3.6
  skip_if(grepl("^3.6", getRversion()))
  skip_if(grepl("^4", getRversion()))

  scenario <- readRDS(rds_filename)$scenario
  parameters <- readRDS(rds_filename)$parameters

  scenario$targetRunner <- function(experiment, scenario, filename = rds_filename) {
    saved <- readRDS(filename)
    row <- which(saved$instancesList[, "instance"] == experiment[["id.instance"]]
                 & saved$instancesList[, "seed"] == experiment[["seed"]])

    cost <- saved$experiments[row, experiment[["id.configuration"]] ]
    if (length(cost) == 0 || is.na(cost)) {
      print(row)
      print(experiment)
    }
    expect_false(is.na(cost))
    return(list(cost = cost))
  }
  confs <- irace(scenario = scenario, parameters = parameters)
  expect_gt(nrow(confs), 0L)
}

test_that("bug_large_new_instances", {
  skip_on_cran()
  # FIXME: For some unknown reason, this test doesn't work in R 3.6
  skip_if(grepl("^3.6", getRversion()))
  skip_if(grepl("^4", getRversion()))
  # FIXME: Convert this test to use run_irace_from_rds()
  load("bug_large_new_instances.Rdata", verbose = TRUE)

  scenario$targetRunner <- function(experiment, scenario) {
    saved_instances_list <- dynGet("saved_instances_list", inherits = TRUE)
    saved_experiments <- dynGet("saved_experiments", inherits = TRUE)
    row <- which(saved_instances_list[, "instance"] == experiment[["id.instance"]]
                 & saved_instances_list[, "seed"] == experiment[["seed"]])
    cost <- saved_experiments[row, experiment[["id.configuration"]] ]
    if (length(cost) == 0 || is.na(cost)) {
      print(row)
      print(experiment)
    }
    expect_false(is.na(cost))
    return(list(cost = cost))
  }
  confs <- irace(scenario = scenario, parameters = parameters)
  expect_gt(nrow(confs), 0L)
})

test_that("target.runner as string", {

  target.runner.local <- function(experiment, scenario) return(list(cost=1L))

  expect_true(irace:::is.function.name("target.runner.local"))

  # Test that a function can be given as a string.
  scenario <- list(targetRunner = "target.runner.local",
                   instances = 1:10, maxExperiments = 1000)
  scenario <- checkScenario (scenario)

  expect_equal(scenario$targetRunner, target.runner.local)
  expect_is(scenario$targetRunner, "function")
})

target.runner.global <- function(experiment, scenario) return(list(cost=1L))

test_that("target.runner as string (global)", {

  expect_true(irace:::is.function.name("target.runner.global"))

  # Test that a function can be given as a string.
  scenario <- list(targetRunner = "target.runner.global",
                   instances = 1:10, maxExperiments = 1000)
  scenario <- checkScenario (scenario)

  expect_equal(scenario$targetRunner, target.runner.global)
  expect_is(scenario$targetRunner, "function")
})


test_that("ordered assert", {
  library(irace)
  parameters <- readParameters(text='
x "" o (a,b,c,d)
')
  confs <- irace:::sampleUniform(parameters, 1, 0)
  confs$.ID. <- 1
  model <- irace:::initialiseModel(parameters, confs, 0)
  confs <- irace:::sampleModel(parameters, confs, model, 1, 0)
  expect_true(confs$x %in% parameters$domain$x)
})

test_that("maxim_bug", {
  skip_on_cran()
  run_irace_from_rds("saved_maxim_bug.rds")
})

test_that("maxim_bug2", {
  skip_on_cran()
  run_irace_from_rds("saved_maxim_bug2.rds")
})

test_that("non-normal", {
  skip_on_cran()
  df <- as.matrix(read.table(text='
0.007858276 0.007934570 0.007949829
0.009384155 0.009521484 0.009719849
0.011520386 0.012100220 0.012176514
0.214721680 0.219146729 0.219512939
0.005462646 0.005462646 0.005523682
0.005035400 0.005035400 0.005035400
0.011001587 0.011367798 0.011962891
0.153244019 0.155212402 1.157546997
0.008636475 0.008789062 0.008666992
0.008239746 0.008499146 0.008560181
0.010848999 0.011077881 0.011703491
0.191589355 0.191879272 0.191589355
0.011245728 0.011245728 0.011245728
0.006454468 0.006454468 0.006561279
0.010330200 0.010894775 0.011032104
0.160491943 0.164169312 0.167068481', header=FALSE))
  ## colnames(df) <- 1:3
  ## #df <- log(df)
  ## library(ggplot2)
  ## library(tidyr)
  ## df_long <- as.data.frame(df)
  ## df_long$run <- 1:16

  ## df_long <- df_long %>% gather("config", "y", as.character(1:3))
  ## ggplot(df_long, aes(df=y)) +
  ##   geom_histogram(aes(y=..density..), colour="black", fill="white", binwidth=0.01)+
  ##   geom_density(alpha=.2, fill="#FF6666") +
  ##   facet_grid(config ~ .)

  ## ggplot(df_long, aes(x=config, y=y)) + geom_boxplot() + geom_jitter()
  ## aggregate(y ~ config, data = df_long, FUN = function(x) c(mean = mean(x), sd = sd(x), median=median(x),
  ##                                                           shapiro.test = shapiro.test(x)$p.value) )

  ## t.test(x[,1], x[,2],paired=TRUE)
  ## t.test(x[,1], x[,3],paired=TRUE)

  ## wilcox.test(x[,1], x[,2],paired=TRUE)
  ## wilcox.test(x[,1], x[,3],paired=TRUE)

  ## print(all(x[,1] <= x[,2]))
  ## print(all(x[,1] <= x[,3]))
  parameters <- readParameters(text='p "" c (1,2,3,4,5)')
  scenario <- list(instances = 1:16, maxExperiments = 50,
                   testType="t-test")
  scenario$targetRunner <- function(experiment, scenario, this_df = df) {
    if (experiment[["id.configuration"]] > 3)
      cost <- as.numeric(experiment[["id.configuration"]])
    else
      cost <- this_df[as.numeric(experiment[["id.instance"]]),
                      as.numeric(experiment[["id.configuration"]]) ]
    return(list(cost=cost))
  }
  confs <- irace(scenario = scenario, parameters = parameters)
  print(confs)
  expect_equal(confs$.ID., 1L)
})

}) # withr::with_output_sink()

#' Post-selection race
#'
#' \code{psRace} performs a post-selection race of a set of configurations.
#'
#' @inheritParams has_testing_data
#'
#' @param max_experiments `numeric(1)`\cr Number of experiments for the
#'   post-selection race. If it is equal to or smaller than 1, then it is a
#'   fraction of the total budget given by
#'   `iraceResults$scenario$maxExperiments` or `iraceResults$scenario$maxTime /
#'   iraceResults$state$boundEstimate`.
#' @param conf_ids  IDs of the configurations in `iraceResults$allConfigurations` to be used for the post-selection.
#' If `NULL`, then the configurations are automatically selected.
#' @param iteration_elites  If `FALSE`, give priority to selecting the configurations that were elite in the last iteration.
#' If `TRUE`, select from all elite configurations of all iterations. This parameter only has an effect when `conf_ids` is not `NULL`.
#' @param psrace_logFile  `character(1)`\cr Log file to save the post-selection race log. If `NULL`, the log is saved in `iraceResults$scenario$logFile`.
#'
#' @return The elite configurations after the post-selection. In addition, if `iraceResults$scenario$logFile` is defined,
#' it saves an updated copy of `iraceResults` in that file, where `iraceResults$psrace_log` is a list with the following elements:
#'  \describe{
#'    \item{configurations}{Configurations used in the post-selection race.}
#'    \item{instances}{Data frame with the instances used in the post-selection race. First column has the
#'    instances IDs from `iraceResults$scenario$instances`, second column the seed assigned to the instance.}
#'    \item{max_experiments}{Configuration budget assigned to the post-selection race.}
#'    \item{experiments}{Matrix of results generated by the post-selection  race, in the same format as the matrix \code{iraceResults$experiments}. Column names are the configuration IDs and  row names are the instance IDs.}
#'    \item{elites}{Best configurations found in the experiments.}
#'  }
#'
#' @examples
#' \donttest{
#'   irace_log <- read_logfile(system.file(package="irace", "exdata", "sann.rda"))
#'   # Use a temporary file to not change the original "sann.rda".
#'   psrace_logFile <- withr::local_tempfile(fileext = ".Rdata")
#'   # Execute the post-selection after the execution of irace. Use 10% of the total budget.
#'   psRace(irace_log, max_experiments=0.1, psrace_logFile = psrace_logFile)
#'   # Print psrace_log
#'   irace_log <- read_logfile(psrace_logFile)
#'   str(irace_log$psrace_log)
#' }
#' @author Leslie Pérez Cáceres and Manuel López-Ibáñez
#' @export
psRace <- function(iraceResults, max_experiments, conf_ids = NULL, iteration_elites = FALSE,
                   psrace_logFile = NULL)
{
  irace_note("Starting post-selection:\n")
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing.")
  iraceResults <- read_logfile(iraceResults)
  scenario <- iraceResults$scenario
  if (!is.null(psrace_logFile))
    scenario$logFile <- psrace_logFile

  race_state <- iraceResults$state
  race_state$initialize(scenario, new = FALSE) # Restores the random seed
  if (max_experiments <= 0) stop("'max_experiments' must be positive.")
  if (max_experiments <= 1) {
    budget <- if (scenario$maxTime == 0L)
                scenario$maxExperiments else scenario$maxTime / race_state$boundEstimate
    max_experiments <- as.integer(ceiling(max_experiments * budget))
    if (scenario$maxTime == 0L)
      cat(sep="", "# scenario maxExperiments:", scenario$maxExperiments, "\n")
    else
      cat(sep="", "# scenario maxTime:", scenario$maxTime, "\n")
  }
  # Get selected configurations.
  if (is.null(conf_ids)) {
    # FIXME: Handle scenario$maxTime > 0
    irace_assert(scenario$maxTime == 0)

    which_max_last <- function(x) 1L + length(x) - which.max(rev(x))

    # We want to race at least two configurations, so we generate all
    # integers between 3 and 2^max_confs - 1L that are not powers of 2, then convert
    # it to a bit-string then to a logical vector.
    generate_combs_2 <- function(max_confs) {
      max_pow <- as.integer(2^max_confs)
      pow_2 <- as.integer(2L^(0L:(max_confs-1L)))
      s <- seq.int(3L, max_pow - 1L, 2L)
      if (max_pow >= 8L)
        s <- c(s, setdiff(seq.int(6L, max_pow - 2L, 2L), pow_2[-1L:-3L]))
      lapply(s, function(z) as.logical((z %/% pow_2) %% 2L))
    }
    # A version of the above that is not limited to two configurations.
    generate_combs_1 <- function(max_confs) {
      max_pow <- as.integer(2^max_confs)
      pow_2 <- as.integer(2L^(0L:(max_confs-1L)))
      lapply(seq.int(1L, max_pow - 1L, 1L),
        function(z) as.logical((z %/% pow_2) %% 2L))
    }
    # FIXME: Is this really faster than head(x, n=max_len) ?
    truncate_conf_needs <- function(x, max_len) {
      if (length(x) <= max_len) return(x)
      x[seq_len(max_len)]
    }

    get_confs_for_psrace <- function(iraceResults, iteration_elites, max_experiments, conf_ids, rejected_ids) {
      scenario <- iraceResults$scenario
      report_selected <-
        if (scenario$quiet) do_nothing
        else function(conf_ids, conf_needs)
          cat(sep="", "# Configurations selected: ", paste0(collapse=", ", conf_ids),
            ".\n# Pending instances: ", paste0(collapse=", ", conf_needs), ".\n")

      allElites <- iraceResults$allElites
      experiments <- iraceResults$experiments
      conf_ids <- if (iteration_elites) unlist(rev(allElites)) else allElites[[length(allElites)]]
      conf_ids <- unique(c(conf_ids, iraceResults$allConfigurations[[".ID."]]))
      # NA may be generated if we skipped iterations.
      if (anyNA(conf_ids))
        conf_ids <- conf_ids[!is.na(conf_ids)]
      # Remove rejected configurations.
      if (length(rejected_ids))
        conf_ids <- setdiff(conf_ids, rejected_ids)
      experiments <- experiments[, conf_ids, drop = FALSE]
      conf_needs <- matrixStats::colCounts(experiments, value = NA, useNames = TRUE)
      n_done <- nrow(experiments) - min(conf_needs)
      # Remove any configuration that needs more than max_experiments.
      conf_needs <- conf_needs[conf_needs <= max_experiments]
      if (length(conf_needs) == 1L) {
        if (!scenario$quiet)
          cat("# Insufficient budget to race more than one configuration!\n")
        return(names(conf_needs))
      }
      if (!scenario$deterministic || n_done < length(scenario$instances)) {
        # We want to evaluate in at least n_new instances more.
        n_new <- scenario$eachTest * scenario$blockSize
        n_confs <- floor(max_experiments / n_new)
        # If we have n_confs that have been evaluated in all instances, select those.
        if (n_confs > 1L && sum(conf_needs == 0L) >= n_confs) {
          conf_needs <- conf_needs[conf_needs == 0L][1:n_confs]
          conf_ids <- names(conf_needs)
          report_selected(conf_ids, conf_needs)
          irace_assert(length(conf_ids) > 1L, eval_after = {
            cat("n_confs: ", n_confs, "\nn_new:", n_new, "\n")
            print(conf_needs)
            save(iraceResults, file="bug-conf_ids.Rdata")
          })
          return(conf_ids)
        }
        # Let's try first to evaluate on new instances.
        conf_needs_new <- truncate_conf_needs(conf_needs + n_new, max_len = 16L)
        irace_assert(length(conf_needs_new) > 1L, eval_after={
          cat("max_experiments: ", max_experiments, "\nn_new: ", n_new, "\nconf_needs:")
          print(conf_needs)
          save(iraceResults, file="bug-conf_ids.Rdata")
        })
        combs <- generate_combs_2(length(conf_needs_new))
        left <- sapply(combs, function(x) max_experiments - sum(conf_needs_new[x]), USE.NAMES=FALSE)
        irace_assert(!is.null(left) && length(left) > 0L && !anyNA(left),
          eval_after= {
            cat("max_experiments: ", max_experiments, "\n")
            cat("length(left): ", length(left), "\n")
            cat("sum(is.na(left)): ", sum(is.na(left)), "\n")
            save(iraceResults, file="bug-conf_ids.Rdata")
          })

        if (any(left >= 0L)) { # We have enough budget to evaluate on a new instance.
          # Select the combination that will allow to evaluate the most configurations.
          combs <- combs[left >= 0]
          n <- sapply(combs, sum, USE.NAMES=FALSE)
          winner <- which.max(n)
          conf_needs_new <- conf_needs_new[combs[[winner]]]
          conf_ids <- names(conf_needs_new)
          report_selected(conf_ids, conf_needs_new)
          irace_assert(length(conf_ids) > 1L, eval_after = {
            cat("max_experiments: ", max_experiments, "\n")
            cat("length(left): ", length(left), "\n")
            cat("sum(is.na(left)): ", sum(is.na(left)), "\n")
            cat("winner: ", winner, "\n")
            cat("combs[[winner]]: ", paste0(collapse=",", combs[[winner]]), "\n")
            save(iraceResults, file="bug-conf_ids.Rdata")
          })
          return(conf_ids)
        }
      }
      # We do not have enough budget to see new instances, so we need consider
      # at least 1 configuration that has NA values, but we still include the
      # configurations that have been evaluated on all instances.
      conf_needs_zero <- conf_needs[conf_needs == 0L]
      conf_needs <- truncate_conf_needs(conf_needs[conf_needs > 0L], 16L)
      combs <- generate_combs_1(length(conf_needs))
      left <- sapply(combs, function(x) max_experiments - sum(conf_needs[x]), USE.NAMES=FALSE)
      irace_assert(any(left >= 0), eval_after=save(iraceResults, file="bug-conf_ids.Rdata"))
      # Select the combination that will allow us to evaluate the most configurations.
      combs <- combs[left >= 0]
      n <- sapply(combs, sum, USE.NAMES=FALSE)
      winner <- which.max(n)
      conf_needs <- c(conf_needs_zero, conf_needs[combs[[winner]]])
      conf_ids <- names(conf_needs)
      report_selected(conf_ids, conf_needs)
      irace_assert(length(conf_ids) > 1L, eval_after = {
        print(conf_needs)
        cat("winner: ", winner, "\n")
        cat("combs[[winner]]: ", paste0(collapse=",", combs[[winner]]), "\n")
        save(iraceResults, file="bug-conf_ids.Rdata")
      })
      return(conf_ids)
    }
    conf_ids <- get_confs_for_psrace(iraceResults, iteration_elites, max_experiments,
      conf_ids = conf_ids, rejected_ids = race_state$rejected_ids)
    irace_assert(length(conf_ids) >= 1L, eval_after = {
      # Debug what happened if the assert failed.
      rejected_ids <- race_state$rejected_ids
      cat("blockSize:", scenario$blockSize, "\niteration_elites: ", as.character(iteration_elites),
        "\nrejected: ", paste0(collapse=", ", rejected_ids), "\n")
      allElites <- iraceResults$allElites
      experiments <- iraceResults$experiments
      conf_ids <- if (iteration_elites) unlist(rev(allElites)) else allElites[[length(allElites)]]
      cat("conf_ids1:", paste0(collapse=", ", conf_ids), "\n")
      conf_ids <- unique(c(conf_ids, iraceResults$allConfigurations[[".ID."]]))
      cat("conf_ids2:", paste0(collapse=", ", conf_ids), "\n")
      # NA may be generated if we skipped iterations.
      if (anyNA(conf_ids))
        conf_ids <- conf_ids[!is.na(conf_ids)]
      cat("conf_ids3:", paste0(collapse=", ", conf_ids), "\n")
      # Remove rejected configurations.
      if (length(rejected_ids))
        conf_ids <- setdiff(conf_ids, rejected_ids)
      cat("conf_ids4:", paste0(collapse=", ", conf_ids), "\n")
      conf_needs <- matrixStats::colCounts(experiments[, conf_ids, drop = FALSE], value = NA)
      cat("conf_needs:", paste0(collapse=", ", conf_needs), "\n")
    })
    if (length(conf_ids) == 1L) {
      # If we cannot run post-selection, we just return the elite configurations.
      allElites <- iraceResults$allElites
      return(allElites[[length(allElites)]])
    }

  } else if (length(conf_ids) <= 1L) {
    irace_error ("The number configurations provided should be larger than 1.")
  } else if (length(race_state$rejected_ids) && any(conf_ids %in% race_state$rejected_ids)) {
    irace_error ("Some configuration IDs provided were rejected in the previous run: ",
      paste0(collapse=", ", intersect(conf_ids, race_state$rejected_ids)), ".")
  }
  if (!all(conf_ids %in% iraceResults$allConfigurations[[".ID."]])) {
    irace_error("Some configuration IDs provided cannot be found in the configurations: ",
      paste0(collapse=", ", setdiff(conf_ids, iraceResults$allConfigurations[[".ID."]])), ".")
  }
  elite_configurations <- iraceResults$allConfigurations[iraceResults$allConfigurations[[".ID."]] %in% conf_ids, , drop=FALSE]

  # Generate new instances.
  generateInstances(race_state, scenario, max_experiments / nrow(elite_configurations), update = TRUE)
  elite_data <- iraceResults$experiments[, as.character(elite_configurations[[".ID."]]), drop=FALSE]
  race_state$next_instance <- nrow(elite_data) + 1L

  irace_note("seed: ", race_state$seed,
    "\n# Configurations: ", nrow(elite_configurations),
    "\n# Available experiments: ", max_experiments,
    "\n# minSurvival: 1\n")

  # We do not want to stop because of this limit.
  scenario$elitistLimit <- 0L
  # FIXME: elitist_race should not require setting this, but it currently does.
  scenario$elitist <- TRUE
  raceResults <- elitist_race(race_state,
    maxExp = max_experiments,
    minSurvival = 1L,
    configurations = elite_configurations,
    scenario = scenario,
    elite_data = elite_data,
    elitist_new_instances = 0L,
    # FIXME: This should be nrow(elite_data) + 1L if we have budget to evaluate on new instances.
    firstTest = nrow(elite_data) / scenario$blockSize)

  elite_configurations <- extractElites(raceResults$configurations,
    nbElites = race_state$minSurvival, debugLevel = scenario$debugLevel)
  irace_note("Elite configurations (first number is the configuration ID;",
               " listed from best to worst according to the ",
               test.type.order.str(scenario$testType), "):\n")
  if (!scenario$quiet)
    configurations_print(elite_configurations, metadata = scenario$debugLevel >= 1L)

  if (!is.null(scenario$logFile)) {
    elapsed <- race_state$time_elapsed()
    if (!scenario$quiet)
      cat("# Total CPU user time: ", elapsed["user"], ", CPU sys time: ", elapsed["system"],
        ", Wall-clock time: ", elapsed["wallclock"], "\n", sep="")
    indexIteration <- 1L + max(race_state$experiment_log[["iteration"]])
    # We add indexIteration as an additional column.
    set(raceResults$experiment_log, j = "iteration", value = indexIteration)
    race_state$experiment_log <- rbindlist(list(race_state$experiment_log, raceResults$experiment_log), fill=TRUE, use.names=TRUE)
    # Merge new results.
    iraceResults$experiments <- merge_matrix(iraceResults$experiments, raceResults$experiments)
    iraceResults$iterationElites[indexIteration] <- elite_configurations[[".ID."]][1L]
    iraceResults$allElites[[indexIteration]] <- elite_configurations[[".ID."]]
    iraceResults$scenario <- scenario
    iraceResults$state <- race_state
    # FIXME: This log should contain only information of what was done in the
    # psRace and avoid duplicating info from iraceResults.
    iraceResults$psrace_log <- list(configurations = elite_configurations,
      instances = race_state$instances_log[seq_len(nrow(raceResults$experiments)), , drop = FALSE],
      max_experiments = max_experiments,
      experiments = raceResults$experiments,
      elites = elite_configurations[[".ID."]])
    save_irace_logfile(iraceResults, logfile = scenario$logFile)
  }
  elite_configurations
}

withr::with_output_sink("test-psrace.Rout", {

test_that("max_experiments=0.1", {
  skip_on_cran()
  generate_set_seed()
  irace_log <- read_logfile(system.file(package="irace", "exdata", "sann.rda"))
  # Use a temporary file to not change the original "sann.rda".
  psrace_logFile <- withr::local_tempfile(fileext = ".Rdata")
  # Execute the post-selection after the execution of irace. Use 10% of the total budget.
  psRace(irace_log, max_experiments=0.1, psrace_logFile = psrace_logFile)
  irace_log <- read_logfile(psrace_logFile)
  budget <- nrow(irace_log$state$experiment_log[iteration == max(iteration)])
  expect_gt(budget, 50L)
  # It should be equal but elitist_race sometimes fails to consume all the budget.
  expect_lte(budget, irace_log$psrace_log$max_experiments)
})

test_that("max_experiments=101", {
  skip_on_cran()
  generate_set_seed()
  irace_log <- read_logfile(system.file(package="irace", "exdata", "sann.rda"))
  conf_ids <- unlist(tail(irace_log$allElites, n=1L))
  # Use a temporary file to not change the original "sann.rda".
  psrace_logFile <- withr::local_tempfile(fileext = ".Rdata")
  # Execute the post-selection after the execution of irace. 
  psRace(irace_log, max_experiments=101, conf_ids = conf_ids, psrace_logFile = psrace_logFile)
  irace_log <- read_logfile(psrace_logFile)
  budget <- nrow(irace_log$state$experiment_log[iteration == max(iteration)])
  expect_gt(budget, 10L)
  # It should be equal but elitist_race sometimes fails to consume all the budget.
  expect_lte(budget, irace_log$psrace_log$max_experiments)
})

}) # withr::with_output_sink()

#' psRace
#'
#' \code{psRace} performs a postselection race a set of configurations. 
#' 
#' @param iraceLogFile NULL Log file created by \pkg{irace}, this file must contain the 
#' \code{iraceResults} object.
#' @param iraceResults NULL Object created by \pkg{irace} and saved in \code{scenario$logFile}.
#' @param conf.ids NULL IDs of the configurations in iraceResults$allConfigurations to be used for ablation.
#' If NULL, the elites argument will be used.
#' @param postselection NULL Percentage of the maxExperiments provided in the scenario to be used in the race.
#' @param max.experiments NULL Number of experiments available for the race. If NULL budget for the race is set
#'  by the parameter scenario$postselection, which defines the percentage of the total budget of \pkg{irace}
#' (iraceResults$scenario$maxExperiments or iraceResults$scenario$maxTime/iraceResults$state$timeEstimate) to use
#'  for the postselection.
#' @param elites FALSE Flag for selecting configurations. If FALSE, the best configurations of each
#' iteration are used for the race. If TRUE, the elite configurtions of each iteration are used for the race. 
#' @param seed 1234567 Numerical value to use as seed for the random number generation.
#' 
#' @return If iraceLogFile is NULL, it returns a list with the following elements:
#'  \describe{
#'    \item{configurations}{Configurations used in the race.}
#'    \item{instances}{A matrix with the instances used in the experiments. First column has the 
#'    instances ids from iraceResults$scenario$instances, second column the seed assigned to the instance.}
#'    \item{maxExperiments}{Maximum number of experiments set for the race.}
#'    \item{experiments}{A matrix with the results of the experiments (columns are configurations, rows are instances).}
#'    \item{elites}{Best configurations found in the experiments.}
#'  }
#' If \code{iraceLogFile} is provided this list object will be saved in \code{iraceResults$psrace.log}.
#'
#' @examples
#' \dontrun{
#'   # Execute the postselection automatically after irace
#'   scenario <- readScenario(filename="scenario.txt")
#'   parameters <- readParameters("parameters.txt")
#'   # Use 10% of the total budget
#'   scenario$postselection <- 0.1
#'   irace(scenario=scenario, parameters=parameters)
#'   # Execute the postselection after the execution of \pkg{irace}.
#'   psRace(iraceLogFile="irace.Rdata", max.experiments=120)
#' }
#'
#' @author Leslie Pérez Cáceres
#' @export
psRace <- function(iraceLogFile=NULL, iraceResults=NULL, conf.ids=NULL,
                   postselection=NULL, max.experiments=NULL, elites=FALSE, seed=1234567)
{
  # Input check
  if (is.null(iraceLogFile) && is.null(iraceResults)) 
    irace.error("You must provide a Rdata file or an iraceResults object.")
    
  irace.note ("Starting post-selection:\n# Seed:", seed, "\n")
  if (!is.null(iraceLogFile))
    cat("# Log file:",iraceLogFile,"\n")
  
  # Load the data of the log file
  if (!is.null(iraceLogFile)) 
    iraceResults <- read_logfile(iraceLogFile)
 
  parameters <- iraceResults$parameters
  scenario   <- iraceResults$scenario
  
  # Get selected configurations
  if (!is.null(conf.ids)) {
    if (!all(conf.ids %in% iraceResults$allConfigurations$.ID.)) 
      irace.error("Configuration ids provided", conf.ids,"cannot be found in the configurations.")
    configurations <- iraceResults$allConfigurations[iraceResults$allConfigurations$.ID.%in% conf.ids,,drop=FALSE]
  } else {
    which.elites <- if (elites) unlist(iraceResults$allElites) else iraceResults$iterationElites
    which.elites <- unique(which.elites)
    configurations <- iraceResults$allConfigurations[which.elites, ]
  }
  
  if (nrow(configurations) <= 1)
    irace.error ("The number configurations should be > 1.")

  # LESLIE: Should we use testing instances?
  # Generate new instances
  instances <- generateInstances(scenario, 1000)
  .irace$instancesList <- instances
  .irace$next.instance <- 1
  # MANUEL: Does this really work? It looks very strange.
  if (!is.null(scenario$boundMax))
    scenario$instances <- paste(scenario$instances, scenario$boundMax, sep=" ")
  
  scenario$elitist <- scenario$capping <- FALSE
  
  if (is.null(postselection)) postselection <- scenario$postselection
  # Calculate available budget
  # FIXME: add numerical checks
  if (is.null(max.experiments)) {
    budget <- if (scenario$maxExperiments > 0)
                scenario$maxExperiments else (scenario$maxTime/iraceResults$state$timeEstimate)
    max.experiments <- ceiling(postselection *  budget)
  }
  
  cat("# configurations:", nrow(configurations), "\n")
  cat("# postselection %", postselection, "\n")
  cat("# scenario experiments:",scenario$maxExperiments,"\n" )
  cat("# available experiments:",max.experiments,"\n" )
  cat("# minSurvival: 1\n")
  
  if (!is.null(seed)) set.seed(seed)
  # Should we fix the paramenters for the race?
  race.output <- race(maxExp = max.experiments,
                      minSurvival = 1,
                      elite.data = NULL,
                      configurations = configurations,
                      parameters = parameters,
                      scenario = scenario,
                      elitistNewInstances = 0)
  experiments <- race.output$experiments
  
  elite.configurations <- extractElites(scenario, parameters,
                                        race.output$configurations,
                                        min(race.output$nbAlive, 1))
  irace.note("Elite configurations (first number is the configuration ID;",
               " listed from best to worst according to the ",
               test.type.order.str(scenario$testType), "):\n")
  configurations.print(elite.configurations, metadata = scenario$debugLevel >= 1)

  psrace.log <-  list(configurations = configurations,
                      instances = instances[1:nrow(experiments),],
                      maxExperiments = max.experiments,
                      experiments = experiments,
                      elites = elite.configurations$.ID.)   
  iraceResults$psrace.log <- psrace.log        
        
  if (!is.null(iraceLogFile))
    irace_save_logfile(iraceResults, scenario)
  return(psrace.log)
}

## This file was generated by scripts/generate-options.R
# Non-variable options (such as --help and --version) have names starting with "."
# Variables that do not have a command-line option have description == ""
# Types are b(oolean), i(nteger), s(tring), r(eal), p(ath), x (R object or no value)
# FIXME: Add special type for R functions.
# FIXME: For i and r add their domain.
.irace.params.def <- structure(list(name = c(".help", ".version", ".check", ".init",
".onlytest", "scenarioFile", "execDir", "parameterFile", "parameters",
"initConfigurations", "configurationsFile", "logFile", "recoveryFile",
"instances", "trainInstancesDir", "trainInstancesFile", "sampleInstances",
"testInstancesDir", "testInstancesFile", "testInstances", "testNbElites",
"testIterationElites", "testType", "firstTest", "blockSize",
"eachTest", "targetRunner", "targetRunnerLauncher", "targetCmdline",
"targetRunnerRetries", "targetRunnerTimeout", "targetRunnerData",
"targetRunnerParallel", "targetEvaluator", "deterministic", "maxExperiments",
"minExperiments", "maxTime", "budgetEstimation", "minMeasurableTime",
"parallel", "loadBalancing", "mpi", "batchmode", "quiet", "debugLevel",
"seed", "softRestart", "softRestartThreshold", "elitist", "elitistNewInstances",
"elitistLimit", "repairConfiguration", "capping", "cappingAfterFirstTest",
"cappingType", "boundType", "boundMax", "boundDigits", "boundPar",
"boundAsTimeout", "postselection", "aclib", "nbIterations", "nbExperimentsPerIteration",
"minNbSurvival", "nbConfigurations", "mu", "confidence"), type = c("x",
"x", "x", "x", "p", "p", "p", "p", "x", "x", "p", "p", "p", "s",
"p", "p", "b", "p", "p", "x", "i", "b", "s", "i", "i", "i", "p",
"p", "s", "i", "i", "x", "x", "p", "b", "i", "i", "i", "r", "r",
"i", "b", "b", "s", "b", "i", "i", "b", "r", "b", "i", "i", "x",
"b", "b", "s", "s", "i", "i", "i", "b", "b", "b", "i", "i", "i",
"i", "i", "r"), short = c("-h", "-v", "-c", "-i", "", "-s", "",
"-p", "", "", "", "-l", "", "", "", "", "", "", "", "", "", "",
"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "",
"", "", "", "", "", "", "-q", "", "", "", "", "-e", "", "", "",
"", "", "", "", "", "", "", "", "", "", "", "", "", "", "", ""
), long = c("--help", "--version", "--check", "--init", "--only-test",
"--scenario", "--exec-dir", "--parameter-file", "", "", "--configurations-file",
"--log-file", "--recovery-file", "", "--train-instances-dir",
"--train-instances-file", "--sample-instances", "--test-instances-dir",
"--test-instances-file", "", "--test-num-elites", "--test-iteration-elites",
"--test-type", "--first-test", "--block-size", "--each-test",
"--target-runner", "--target-runner-launcher", "--target-cmdline",
"--target-runner-retries", "--target-runner-timeout", "", "",
"--target-evaluator", "--deterministic", "--max-experiments",
"--min-experiments", "--max-time", "--budget-estimation", "--min-measurable-time",
"--parallel", "--load-balancing", "--mpi", "--batchmode", "--quiet",
"--debug-level", "--seed", "--soft-restart", "--soft-restart-threshold",
"--elitist", "--elitist-new-instances", "--elitist-limit", "",
"--capping", "--capping-after-first-test", "--capping-type",
"--bound-type", "--bound-max", "--bound-digits", "--bound-par",
"--bound-as-timeout", "--postselection", "--aclib", "--iterations",
"--experiments-per-iteration", "--min-survival", "--num-configurations",
"--mu", "--confidence"), default = c(NA, NA, NA, "", "", "./scenario.txt",
"./", "./parameters.txt", "", "", "", "./irace.Rdata", "", "",
"", "", "1", "", "", "", "1", "0", "", "5", "1", "1", "./target-runner",
"", "{configurationID} {instanceID} {seed} {instance} {bound} {targetRunnerArgs}",
"0", "0", "", "", "", "0", "0", NA, "0", "0.05", "0.01", "0",
"1", "0", "0", "0", "0", NA, "1", "1e-04", "1", "1", "2", "",
NA, "0", "median", "candidate", "0", "0", "1", "1", "1", "0",
"0", "0", "0", "0", "5", "0.95"), domain = c(NA, NA, NA, NA,
NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
NA, NA, "F-test,t-test,t-test-holm,t-test-bonferroni", NA, NA,
NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA,
NA, NA, "sge,pbs,torque,slurm,htcondor", NA, NA, NA, NA, NA,
NA, NA, NA, NA, NA, NA, "median,mean,worst,best", "instance,candidate",
NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA, NA), description = c("Show this help.",
"Show irace package version.", "Check scenario.", "Initialize the working directory with template config files.",
"Only test the configurations given in the file passed as argument.",
"File that describes the configuration scenario setup and other irace settings.",
"Directory where the programs will be run.", "File that contains the description of the parameters of the target algorithm.",
"", "", "File that contains a table of initial configurations. If empty or `NULL`, all initial configurations are randomly generated.",
"File to save tuning results as an R dataset, either absolute path or relative to execDir.",
"Previously saved log file to recover the execution of `irace`, either absolute path or relative to the current directory.  If empty or `NULL`, recovery is not performed.",
"", "Directory where training instances are located; either absolute path or relative to current directory. If no `trainInstancesFiles` is provided, all the files in `trainInstancesDir` will be listed as instances.",
"File that contains a list of training instances and optionally additional parameters for them. If `trainInstancesDir` is provided, `irace` will search for the files in this folder.",
"Randomly sample the training instances or use them in the order given.",
"Directory where testing instances are located, either absolute or relative to current directory.",
"File containing a list of test instances and optionally additional parameters for them.",
"", "Number of elite configurations returned by irace that will be tested if test instances are provided.",
"Enable/disable testing the elite configurations found at each iteration.",
"Statistical test used for elimination. The default value selects `t-test` if `capping` is enabled or `F-test`, otherwise. Valid values are: F-test (Friedman test), t-test (pairwise t-tests with no correction), t-test-bonferroni (t-test with Bonferroni's correction for multiple comparisons), t-test-holm (t-test with Holm's correction for multiple comparisons).",
"Number of instances evaluated before the first elimination test. It must be a multiple of `eachTest`.",
"Number of training instances, that make up a 'block' in `trainInstancesFile`. Elimination of configurations will only be performed after evaluating a complete block and never in the middle of a block. Each block typically contains one instance from each instance class (type or family) and the block size is the number of classes. The value of `blockSize` will multiply `firstTest`, `eachTest` and `elitistNewInstances`.",
"Number of instances evaluated between elimination tests.", "Executable called for each configuration that executes the target algorithm to be tuned. See the templates and examples provided.",
"Executable that will be used to launch the target runner, when `targetRunner` cannot be executed directly (e.g., a Python script in Windows).",
"Command-line arguments provided to `targetRunner` (or `targetRunnerLauncher` if defined). The substrings `\\{configurationID\\}`, `\\{instanceID\\}`,  `\\{seed\\}`,  `\\{instance\\}`, and `\\{bound\\}` will be replaced by their corresponding values. The substring `\\{targetRunnerArgs\\}` will be replaced by the concatenation of the switch and value of all active parameters of the particular configuration being evaluated.  The substring `\\{targetRunner\\}`, if present, will be replaced by the value of `targetRunner` (useful when using `targetRunnerLauncher`).",
"Number of times to retry a call to `targetRunner` if the call failed.",
"Timeout in seconds of any `targetRunner` call (only applies to `target-runner` executables not to R functions), ignored if 0.",
"Optional data passed to `targetRunner`. This is ignored by the default `targetRunner` function, but it may be used by custom `targetRunner` functions to pass persistent data around.",
"Optional R function to provide custom parallelization of `targetRunner`.",
"Optional script or R function that provides a numeric value for each configuration. See templates/target-evaluator.tmpl",
"If the target algorithm is deterministic, configurations will be evaluated only once per instance.",
"Maximum number of runs (invocations of `targetRunner`) that will be performed. It determines the maximum budget of experiments for the tuning.",
"Minimum number of runs (invocations of `targetRunner`) that will be performed. It determines the minimum budget of experiments for the tuning. The actual budget depends on the number of parameters and `minSurvival`.",
"Maximum total execution time for the executions of `targetRunner`. `targetRunner` must return two values: cost and time. This value and the one returned by `targetRunner` must use the same units (seconds, minutes, iterations, evaluations, ...).",
"Fraction (smaller than 1) of the budget used to estimate the mean computation time of a configuration. Only used when `maxTime` > 0",
"Minimum time unit that is still (significantly) measureable.",
"Number of calls to `targetRunner` to execute in parallel. Values `0` or `1` mean no parallelization.",
"Enable/disable load-balancing when executing experiments in parallel. Load-balancing makes better use of computing resources, but increases communication overhead. If this overhead is large, disabling load-balancing may be faster.",
"Enable/disable MPI. Use `Rmpi` to execute `targetRunner` in parallel (parameter `parallel` is the number of slaves).",
"Specify how irace waits for jobs to finish when `targetRunner` submits jobs to a batch cluster: sge, pbs, torque, slurm or htcondor. `targetRunner` must submit jobs to the cluster using, for example, `qsub`.",
"Reduce the output generated by irace to a minimum.", "Debug level of the output of `irace`. Set this to 0 to silence all debug messages. Higher values provide more verbose debug messages.",
"Seed of the random number generator (by default, generate a random seed).",
"Enable/disable the soft restart strategy that avoids premature convergence of the probabilistic model.",
"Soft restart threshold value for numerical parameters.", "Enable/disable elitist irace.",
"Number of instances added to the execution list before previous instances in elitist irace.",
"In elitist irace, maximum number per race of elimination tests that do not eliminate a configuration. Use 0 for no limit.",
"User-defined R function that takes a configuration generated by irace and repairs it.",
"Enable the use of adaptive capping, a technique designed for minimizing the computation time of configurations. Capping is enabled by default if `elitist` is active, `maxTime > 0` and  `boundMax > 0`.",
"If set to 1, elimination due to capping only happens after `firstTest` instances are seen.",
"Measure used to obtain the execution bound from the performance of the elite configurations: median, mean, worst, best.",
"Method to calculate the mean performance of elite configurations: candidate or instance.",
"Maximum execution bound for `targetRunner`. It must be specified when capping is enabled.",
"Precision used for calculating the execution time. It must be specified when capping is enabled.",
"Penalization constant for timed out executions (executions that reach `boundMax` execution time).",
"Replace the configuration cost of bounded executions with `boundMax`.",
"Perform a postselection race after the execution of irace to consume all remaining budget. Value 0 disables the postselection race.",
"Enable/disable AClib mode. This option enables compatibility with GenericWrapper4AC as targetRunner script.",
"Maximum number of iterations.", "Number of runs of the target algorithm per iteration.",
"Minimum number of configurations needed to continue the execution of each race (iteration).",
"Number of configurations to be sampled and evaluated at each iteration.",
"Parameter used to define the number of configurations sampled and evaluated at each iteration.",
"Confidence level for the elimination test.")), row.names = c(".help",
".version", ".check", ".init", ".onlytest", "scenarioFile", "execDir",
"parameterFile", "parameters", "initConfigurations", "configurationsFile",
"logFile", "recoveryFile", "instances", "trainInstancesDir",
"trainInstancesFile", "sampleInstances", "testInstancesDir",
"testInstancesFile", "testInstances", "testNbElites", "testIterationElites",
"testType", "firstTest", "blockSize", "eachTest", "targetRunner",
"targetRunnerLauncher", "targetCmdline", "targetRunnerRetries",
"targetRunnerTimeout", "targetRunnerData", "targetRunnerParallel",
"targetEvaluator", "deterministic", "maxExperiments", "minExperiments",
"maxTime", "budgetEstimation", "minMeasurableTime", "parallel",
"loadBalancing", "mpi", "batchmode", "quiet", "debugLevel", "seed",
"softRestart", "softRestartThreshold", "elitist", "elitistNewInstances",
"elitistLimit", "repairConfiguration", "capping", "cappingAfterFirstTest",
"cappingType", "boundType", "boundMax", "boundDigits", "boundPar",
"boundAsTimeout", "postselection", "aclib", "nbIterations", "nbExperimentsPerIteration",
"minNbSurvival", "nbConfigurations", "mu", "confidence"), class = "data.frame")
.irace.params.names <- c("scenarioFile", "execDir", "parameterFile", "parameters", "initConfigurations",
"configurationsFile", "logFile", "recoveryFile", "instances",
"trainInstancesDir", "trainInstancesFile", "sampleInstances",
"testInstancesDir", "testInstancesFile", "testInstances", "testNbElites",
"testIterationElites", "testType", "firstTest", "blockSize",
"eachTest", "targetRunner", "targetRunnerLauncher", "targetCmdline",
"targetRunnerRetries", "targetRunnerTimeout", "targetRunnerData",
"targetRunnerParallel", "targetEvaluator", "deterministic", "maxExperiments",
"minExperiments", "maxTime", "budgetEstimation", "minMeasurableTime",
"parallel", "loadBalancing", "mpi", "batchmode", "quiet", "debugLevel",
"seed", "softRestart", "softRestartThreshold", "elitist", "elitistNewInstances",
"elitistLimit", "repairConfiguration", "capping", "cappingAfterFirstTest",
"cappingType", "boundType", "boundMax", "boundDigits", "boundPar",
"boundAsTimeout", "postselection", "aclib", "nbIterations", "nbExperimentsPerIteration",
"minNbSurvival", "nbConfigurations", "mu", "confidence")
## FIXME: If these values are special perhaps they should be saved in $state ?
.irace.params.recover <- c("instances", "seed", "testInstances",
                           # We need this because this data may mutate
                           "targetRunnerData", "elitist", "deterministic")

.param_na_value_type <- list(i = NA_integer_, r = NA_real_, c = NA_character_, o = NA_character_)

# Returns a data.table
configurations_alloc <- function(colnames, nrow, parameters)
{
  column_type <- function(x, n, types) {
    what <- switch(x,
      .ID. = NA_integer_,
      .PARENT. = NA_integer_,
      .WEIGHT. = NA_real_,
      .param_na_value_type[[ types[x] ]])
    rep_len(what, n)
  }
  x <- sapply(colnames, column_type, n=nrow, types = parameters[["types"]],
              simplify=FALSE, USE.NAMES=TRUE)
  setDT(x)
  x
}

# FIXME: It may be faster to create a single expression that concatenates all
# the elements of forbidden using '|'
checkForbidden <- function(configurations, forbidden)
{
  # We have to use a variable name that will never appear in
  # configurations, so .FORBIDDEN .
  for (.FORBIDDEN in forbidden) {
    #print(.FORBIDDEN)
    configurations <- subset(configurations, eval(.FORBIDDEN))
    #print(configurations)
    #print(str(configurations))
    ## FIXME: This is normally called with a single configuration. Thus, it
    ## would be faster to break as soon as nrow(configurations) < 1
  }
  #print(nrow(configurations))
  configurations
}

# FIXME: It may be faster to create a single expression that concatenates all
# the elements of forbidden using '|'
filter_forbidden <- function(configurations, forbidden)
{
  # We have to use a variable name that will never appear in
  # configurations, so .FORBIDDEN
  for (.FORBIDDEN in forbidden) {
    configurations <- configurations[eval(.FORBIDDEN)]
    #print(configurations)
    #print(str(configurations))
    if (nrow(configurations) == 0L) return(configurations)
  }
  #print(nrow(configurations))
  configurations
}

which_satisfied <- function(configurations, condition)
{
  # If there is no condition, do not waste time evaluating it.
  if (isTRUE(condition))
    return(seq_nrow(configurations))
  r <- eval(condition, configurations)
  # Return TRUE if TRUE, FALSE if FALSE or NA
  ## FIXME: If we byte-compile the condition, then we should incorporate the
  ## following into the condition directly.
  # r & !is.na(r)
  # Return indexes where r is TRUE.
  which(r)
}

#' removeConfigurationsMetaData
#'
#' Remove the columns with "metadata" of a data frame containing
#' configurations. Currently, metadata corresponds to column names starting
#' with a period.  This function should be used before printing the
#' configurations to output only the values for the parameters of the
#' configuration without metadata possibly useless to the user.
#'
#' @param configurations `data.frame`\cr Parameter configurations of the
#'   target algorithm (one per row).
#'
#' @return The same data frame without "metadata".
#'
#' @seealso
#'   [configurations_print_command()] to print the configurations as command lines.
#'   [configurations_print()] to print the configurations as a data frame.
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
removeConfigurationsMetaData <- function(configurations)
  configurations[, !startsWith(colnames(configurations), "."), drop = FALSE]

#' Print configurations as a data frame
#'
#' @inheritParams removeConfigurationsMetaData
#'
#' @param metadata `logical(1)`\cr whether to print the metadata or
#' not. The metadata are data for the configurations (additionally to the
#' value of each parameter) used by \pkg{irace}.
#'
#' @return None.
#'
#' @seealso
#'  [configurations_print_command()] to print the configurations as command-line strings.
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
configurations_print <- function(configurations, metadata = FALSE)
{
  if (is.data.table(configurations) || !is.data.frame(configurations))
    configurations <- as.data.frame(configurations, stringsAsFactors = FALSE)

  rownames(configurations) <- configurations[[".ID."]]
  if (!metadata)
    configurations <- removeConfigurationsMetaData(configurations)

  print.data.frame(configurations, digits = 15L)
}

#' Print configurations as command-line strings.
#'
#' Prints configurations after converting them into a representation for the
#' command-line.
#'
#' @inheritParams removeConfigurationsMetaData
#' @inheritParams printParameters
#'
#' @return None.
#'
#' @seealso
#'  [configurations_print()] to print the configurations as a data frame.
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
configurations_print_command <- function(configurations, parameters)
{
  if (nrow(configurations) <= 0L) return(invisible())
  ids <- as.numeric(configurations$.ID.)
  configurations <- removeConfigurationsMetaData(configurations)
  # Re-sort the columns
  configurations <- configurations[, parameters$names, drop = FALSE]
  # A better way to do this? We cannot use apply() because that coerces
  # to a character matrix thus messing up numerical values.
  len <- nchar(max(ids))
  for (i in seq_nrow(configurations)) {
    cat(sprintf("%-*d %s\n", len, ids[i],
                buildCommandLine(configurations[i, , drop=FALSE], parameters$switches)))
  }
}

withr::with_output_sink("test-bug-90.Rout", {

  test_that("bug-90", {
    parameters <- readParameters(text='
a  "" c    (1,2,3,4)
b  "" c    (0, 1, 2, 3)
c  "" r    (0,1)
d  "" r    (0,1)
e  "" r    (0,1)
f  "" i    (1,10)
')
    logFile <- withr::local_tempfile(fileext=".Rdata")
    scenario <- defaultScenario(list(
      parameters = parameters,
      instances = 1:20,
      elitist = FALSE,
      postselection = FALSE,
      maxExperiments = 200,
      logFile = logFile,
      targetRunner = function(experiment, scenario)
        list(cost = experiment$configuration$c * experiment$configuration$d * experiment$configuration$e)
    ))
    confs <- irace(scenario)
    best_conf <- getFinalElites(scenario$logFile, n = 1L, drop.metadata = TRUE)
    expect_identical(removeConfigurationsMetaData(confs[1L, , drop = FALSE]), best_conf)
  })
}) # withr::with_output_sink()

withr::with_output_sink("test-forbidden.Rout", {

test_that("checkForbidden", {
  
test.checkForbidden <- function(param.file)
{
  params <- readParameters(param.file)
  confs <- readConfigurationsFile("configurations.txt", params)
  exp.confs <- readConfigurationsFile(text='
param1 param2 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
5        6     "x1"   3.5   "low"
NA        NA   "x3"   4.5   "low"
', parameters = params)
  confs <- irace:::checkForbidden(confs, params$forbidden)
  rownames(confs) <- rownames(exp.confs) <- NULL
  expect_equal(confs, exp.confs)
}
  test.checkForbidden(test_path("parameters.txt"))
  test.checkForbidden(test_path("logparameters.txt"))
})

test_that("filter_forbidden", {

test_filter_forbidden <- function(param_file)
{
  params <- readParameters(param_file)
  confs <- readConfigurationsFile("configurations.txt", params)
  exp_confs <- data.table::as.data.table(readConfigurationsFile(text='
param1 param2 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
5        6     "x1"   3.5   "low"
NA        NA   "x3"   4.5   "low"
', parameters = params))
  confs <- irace:::filter_forbidden(data.table::as.data.table(confs), params$forbidden)
  expect_equal(confs, exp_confs)
}
  test_filter_forbidden(test_path("parameters.txt"))
  test_filter_forbidden(test_path("logparameters.txt"))
})

test_that("retries", {
  params <- readParameters(text='
p0 "" r (0,1)
[forbidden]
p0 > 0.9
')
  confs <- irace:::sampleUniform(params, 50)
  expect_equal(nrow(confs), 50L)
})

test_that("max retries", {
  params <- readParameters(text='
p0 "" r (0,1)
[forbidden]
p0 <= 0.5
p0 > 0.5
')
  expect_error(irace:::sampleUniform(params, 1), "perhaps your constraints are too strict")
})

})

withr::with_output_sink("test-raceconfs.Rout", {

parameters_txt <- '
launch_method "--launch_method=" c (L-BFGS-B,SLSQP)
visit_method "--visit_method=" c (DA, DE)
global_method "--global_method=" c (CMAES,DE)
'
confs_txt <- '
launch_method visit_method global_method
L-BFGS-B    DA             CMAES
SLSQP       DA             CMAES
L-BFGS-B    DE             CMAES
SLSQP       DE             CMAES
L-BFGS-B    DA             DE
SLSQP       DA             DE
L-BFGS-B    DE             DE
SLSQP       DE             DE
'

target_runner <- function(experiment, scenario)
  list(cost = 100, call = toString(experiment))

withr::with_options(list(warning=2), {
  parameters <- readParameters(text=parameters_txt)
  initconfs <- readConfigurationsFile(text=confs_txt, parameters=parameters)
  scenario <- list(targetRunner = target_runner,
    instances = 1:10,
    nbConfigurations = 8,
    maxExperiments = 96, logFile = "",
    initConfigurations = initconfs,
    parameters = parameters)
  scenario <- checkScenario (scenario)
  confs <- irace(scenario = scenario)
  confs <- data.table::as.data.table(removeConfigurationsMetaData(confs))
  initconfs <- data.table::as.data.table(initconfs)
  expect_equal(nrow(data.table::fsetdiff(confs, initconfs)), 0L)
})

}) # withr::with_output_sink()

# This file is loaded automatically by testthat.
generate_set_seed <- function()
{
  seed <- sample.int(min(2147483647L, .Machine$integer.max), size = 1L, replace = TRUE)
  cat("Seed: ", seed, "\n")
  set.seed(seed)
}

test_irace_detectCores <- function()
{
  if (!identical(Sys.getenv("NOT_CRAN"), "true")) return(1L)
  # FIXME: covr takes a very long time on github actions otherwise.
  if (identical(Sys.getenv("COVR_COVERAGE"), "true")) return(1L)
  x <- Sys.getenv("_R_CHECK_LIMIT_CORES_", "")
  if (nzchar(x) && x == "TRUE") return(2L)
  parallel::detectCores()
}

skip_on_coverage <- function() {
  if (identical(Sys.getenv("COVR_COVERAGE"), "true"))
    skip("Not run during coverage.")
  else
    invisible()
}

system_os_is_windows <- function() .Platform$OS.type == "windows"

expect_valid_configurations <- function(configurations, parameters)
{
  configurations <- irace:::removeConfigurationsMetaData(as.data.frame(configurations))
  output <- capture.output(print(configurations, width=5000L, row.names = FALSE))
  output <- paste0(output, "\n", collapse="")
  confs2 <- readConfigurationsFile(text=output, parameters=parameters)
  expect_equal(configurations, confs2)
}

## Functions ##########################################################
f_ackley <- function (x,y, nsize = 0.01) {
  # Transformation of parameter values
  # from [0,1] to [vmin,vmax]
  vmin <- -5
  vmax <- 5
  x <- (x*(vmax-vmin)) + vmin
  y <- (y*(vmax-vmin)) + vmin

  a <- -20 * exp (-0.2 * sqrt(0.5 * (x^2 + y^2)))
  b <- exp(0.5 * (cos(2*pi*x) + cos(2*pi*y)))
  f <- a - b + exp(1) + 20
  # Simulating stochasticity
  noise <- runif(1, min = (1-nsize), max = (1+nsize))
  f <- f*noise

  # Transform result from [fmin,fmax]
  # to [0,100]
  fmin <- 0
  fmax <- 15*(1+nsize)
  ((f - fmin) / (fmax-fmin)) * (100-0) + 0
}

f_goldestein_price <- function (x,y, nsize = 0.01) {
  # Transformation of parameter values
  # from [0,1] to [vmin,vmax]
  vmin <- -2
  vmax <- 2
  x <- (x*(vmax-vmin)) + vmin
  y <- (y*(vmax-vmin)) + vmin

  a <- 1 + ((x + y + 1)^2) * (19 - 14*x + 3*x^2 - 14*y + 6*x*y + 3*y^2)
  b <- 30 + ((2*x - 3*y)^2) * (18 - 32*x + 12*x^2 + 48*y - 36*x*y + 27*y^2)
  f <- a*b
  # Simulating stochasticity
  noise <- runif(1, min = (1-nsize), max = (1+nsize))
  f <- f*noise

  # Transform result from [fmin,fmax]
  # to [0,100]
  fmin <- 0
  fmax <- 1000000*(1+nsize)
  ((f - fmin) / (fmax-fmin)) * (100-0) + 0
}

f_matyas <- function (x,y, nsize = 0.01) {
  # Trasfomation of parameter values
  # from [0,1] to [vmin,vmax]
  vmin <- -10
  vmax <- 10
  x <- (x*(vmax-vmin)) + vmin
  y <- (y*(vmax-vmin)) + vmin

  f <- 0.26 * (x^2 + y^2) - (0.48*x*y)
  # Simulating stochasticity
  noise <- runif(1, min = (1-nsize), max = (1+nsize))
  f <- f*noise

  # Transform result from [fmin,fmax]
  # to [0,100]
  fmin <- 0
  fmax <- 100*(1+nsize)
  ((f - fmin) / (fmax-fmin)) * (100-0) + 0
}

f_himmelblau <- function (x,y, nsize = 0.01) {
  # Trasfomation of parameter values
  # from [0,1] to [vmin,vmax]
  vmin <- -5
  vmax <- 5
  x <- (x*(vmax-vmin)) + vmin
  y <- (y*(vmax-vmin)) + vmin

  f <- (x^2 + y - 11)^2 + (x + y^2 - 7)^2
  # Simulating stochasticity
  noise <- runif(1, min = (1-nsize), max = (1+nsize))
  f <- f*noise

  # Transform result from [fmin,fmax]
  # to [0,100]
  fmin <- 0
  fmax <- 2000*(1+nsize)
  ((f - fmin) / (fmax-fmin)) * (100-0) + 0
}

target_runner_capping_xy <- function(experiment, scenario)
{
  configuration <- experiment$configuration
  instance      <- experiment$instance
  bound         <- experiment$bound

  x <- configuration[["x"]]
  y <- configuration[["y"]]

  value <- switch(instance,
                  ackley     = f_ackley(x, y),
                  goldestein = f_goldestein_price(x, y),
                  matyas     = f_matyas(x, y),
                  himmelblau  = f_himmelblau(x, y))

  # Simulate execution bound
  list(cost = value, time=min(value + 0.1, bound))
}

irace_capping_xy <- function(..., targetRunner = force(target_runner_capping_xy),
                             parallel = test_irace_detectCores())
{
  # Silence Error in `save(iraceResults, file = logfile, version = 3L)`: (converted from warning) 'package:irace' may not be available when loading
  # See https://github.com/r-lib/testthat/issues/2044
  if (!is.null(attr(environment(targetRunner), "name", exact=TRUE))) {
    environment(targetRunner) <- globalenv()
  }

  args <- list(...)
  parameters_table <- '
   x "" r (0, 1.00)
   y "" r (0, 1.00)
   reject "" c (0,1)'

  parameters <- readParameters(text = parameters_table)
  logFile <- withr::local_tempfile(fileext=".Rdata")
  scenario <- list(instances = c("ackley", "goldestein", "matyas", "himmelblau"),
                   targetRunner = targetRunner,
                   capping = TRUE,
                   boundMax = 80,
                   testType = "t-test",
                   logFile = logFile,
                   parallel = parallel,
                   parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)

  expect_true(irace:::checkTargetFiles(scenario = scenario))

  confs <- irace(scenario = scenario)
  best_conf <- getFinalElites(scenario$logFile, n = 1L, drop.metadata = TRUE)
  expect_identical(removeConfigurationsMetaData(confs[1L, , drop = FALSE]),
                   best_conf)
}

# Useful for testing recovery.
wrap_target_runner_error <- function(target_runner, limit)
{
  counter <- force(limit)
  target_runner <- force(target_runner)

  fun <- function(experiment, scenario) {
    counter <<- counter - 1L
    if (counter <= 0L)
      return(list(cost=NA))
    target_runner(experiment, scenario)
  }
  parent.env(environment(fun)) <- globalenv()
  fun
}

wrap_target_runner_counter <- function(target_runner)
{
  counter <- 0L
  target_runner <- force(target_runner)

  fun <- function(experiment, scenario) {
    counter <<- counter + 1L
    target_runner(experiment, scenario)
  }
  parent.env(environment(fun)) <- globalenv()
  fun
}

withr::with_output_sink("test-sann-irace.Rout", {

## Functions ##########################################################
f_rosenbrock <- function (x) {
  d  <- length(x)
  z  <- x + 1
  hz <- z[1:(d - 1)]
  tz <- z[2:d]
  sum(100 * (hz^2 - tz)^2 + (hz - 1)^2)
}

f_rastrigin <- function (x) sum(x * x - 10 * cos(2 * pi * x) + 10)

## target runner ###########################################################
target_runner <- function(experiment, scenario)
{
  debugLevel    <- scenario$debugLevel
  configuration_id  <- experiment$id_configuration
  instance_id   <- experiment$id_instance
  seed          <- experiment$seed
  configuration <- experiment$configuration
  instance      <- experiment$instance

  D <- 3
  par <- runif(D, min = -1, max = 1)
  fn <- function(x) (instance * f_rastrigin(x) + (1 - instance) * f_rosenbrock(x))
  
  tmax = 1 + configuration[["tmax"]]
  temp = 11.0 + configuration[["temp"]]
  stopifnot(tmax > 0)
  stopifnot(temp > 0)
  res <- withr::with_seed(seed,
                  optim(par, fn, method = "SANN", 
                        control = list(maxit = 10, tmax = tmax, temp = temp))
                )
  list(cost = res$value, call = toString(experiment))
}

## target runner ###########################################################
target_runner_reject <- function(experiment, scenario)
{
  if (runif(1) <= 0.05) return (list(cost = -Inf, call = toString(experiment)))
  target_runner(experiment, scenario)
}

## Run function ########################################################
sann_irace <- function(log.param=FALSE, ...)
{
  args <- list(...)

  # tmax and temp must be > 0
  if (log.param)
     parameters_table <- '
       tmax "" i,log (1, 5000)
       temp "" r,log (1, 100)
       '      
  else
     parameters_table <- '
       tmax "" i (1, 5000)
       temp "" r (1, 100)
     '  
  parameters <- readParameters(text = parameters_table)

  scenario <- list(targetRunner = target_runner,
    maxExperiments = 1000, seed = 1234567,
    parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)

  confs <- irace(scenario = scenario)
  best.conf <- getFinalElites(scenario$logFile, n = 1, drop.metadata = TRUE)
  expect_identical(removeConfigurationsMetaData(confs[1, , drop = FALSE]),
                   best.conf)
}

test_that("parallel", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  sann_irace(instances = weights, parallel = test_irace_detectCores())
})

test_that("parallel reject", {
  # Reproducible results
  generate_set_seed()
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  sann_irace(instances = weights, parallel = test_irace_detectCores(), targetRunner = target_runner_reject)
})

test_that("deterministic", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  sann_irace(deterministic = TRUE, instances = weights[1:7])
})

test_that("log", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  sann_irace(log.param=TRUE, instances = weights)
})

test_that("large newInstances", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()
  weights <- rnorm(200, mean = 0.9, sd = 0.02)
  sann_irace(instances = weights, elitistNewInstances = 6, elitistLimit = 2)
})

}) # withr::with_output_sink()

#' Reads the parameters to be tuned by \pkg{irace} from a file or from a
#' character string.
#'
#' @param file `character(1)`\cr Filename containing the definitions of
#'   the parameters to be tuned.
#' @param digits `integer(1)`\cr The number of decimal places to be considered for real-valued parameters.
#' @param debugLevel `integer(1)`\cr Larger values produce more verbose output.
#' @param text `character(1)`\cr If \code{file} is not supplied and this is,
#'  then parameters are read from the value of \code{text} via a text connection.
#'
#' @return A list containing the definitions of the parameters read. The list is
#'  structured as follows:
#'   \describe{
#'     \item{`names`}{Vector that contains the names of the parameters.}
#'     \item{`types`}{Vector that contains the type of each parameter 'i', 'c', 'r', 'o'.
#'       Numerical parameters can be sampled in a log-scale with 'i,log' and 'r,log'
#'       (no spaces).}
#'     \item{`switches`}{Vector that contains the switches to be used for the
#'       parameters on the command line.}
#'     \item{`domain`}{List of vectors, where each vector may contain two
#'       values (minimum, maximum) for real and integer parameters, or
#'       possibly more for categorical parameters.}
#'     \item{`conditions`}{List of R logical expressions, with variables
#'       corresponding to parameter names.}
#'     \item{`isFixed`}{Logical vector that specifies which parameter is fixed
#'       and, thus, it does not need to be tuned.}
#'     \item{`nbParameters`}{An integer, the total number of parameters.}
#'     \item{`nbFixed`}{An integer, the number of parameters with a fixed value.}
#'     \item{`nbVariable`}{Number of variable (to be tuned) parameters.}
#'     \item{`depends`}{List of character vectors, each vector specifies
#'     which parameters depend on this one.}
#'     \item{`is_dependent`}{Logical vector that specifies which parameter has
#'       a dependent domain.}
#'     \item{`digits`}{Integer vector that specifies the number of digits per parameter.}
#'     \item{`forbidden`}{List of expressions that define which parameter configurations are forbidden.}
#'   }
#'
#' @details Either `file` or `text` must be given. If `file` is given, the
#'  parameters are read from the file `file`. If `text` is given instead,
#'  the parameters are read directly from the `text` character string.
#'  In both cases, the parameters must be given (in `text` or in the file
#'  whose name is `file`) in the expected form.  See the documentation
#'  for details.  If none of these parameters is given, \pkg{irace}
#'  will stop with an error.
#'
#'  A fixed parameter is a parameter that should not be sampled but
#'  instead should be always set to the only value of its domain.  In this
#'  function we set `isFixed` to TRUE only if the parameter is a categorical
#'  and has only one possible value.  If it is an integer and the minimum
#'  and maximum are equal, or it is a real and the minimum and maximum
#'  values satisfy `round(minimum, digits) == round(maximum, digits)`,
#'  then the parameter description is rejected as invalid to identify
#'  potential user errors.
#'
#' The order of the parameters determines the order in which parameters are
#' given to `targetRunner`. Changing the order may also change the results
#' produced by `irace`, even with the same random seed.
#'
#' @examples
#'  ## Read the parameters directly from text
#'  parameters_table <- '
#'  # name       switch           type  values               [conditions (using R syntax)]
#'  algorithm    "--"             c     (as,mmas,eas,ras,acs)
#'  localsearch  "--localsearch " o     (0, 1, 2, 3)
#'  alpha        "--alpha "       r     (0.00, 5.00)
#'  beta         "--beta "        r     (0.00, 10.00)
#'  rho          "--rho  "        r     (0.01, 1.00)
#'  ants         "--ants "        i,log (5, 100)
#'  q0           "--q0 "          r     (0.0, 1.0)           | algorithm == "acs"
#'  rasrank      "--rasranks "    i     (1, "min(ants, 10)") | algorithm == "ras"
#'  elitistants  "--elitistants " i     (1, ants)            | algorithm == "eas"
#'  nnls         "--nnls "        i     (5, 50)              | localsearch %in% c(1,2,3)
#'  dlb          "--dlb "         c     (0, 1)               | localsearch %in% c(1,2,3)
#'
#'  [forbidden]
#'  (alpha == 0.0) & (beta == 0.0)
#'  [global]
#'  digits = 4
#'  '
#'  parameters <- readParameters(text=parameters_table)
#'  str(parameters)
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
readParameters <- function (file, digits = 4L, debugLevel = 0L, text)
{
  if (missing(file) && !missing(text)) {
    filename <- NA_character_
    file <- textConnection(text)
    on.exit(close(file))
  } else if (is.character(file)) {
    filename <- file
    file.check (file, readable = TRUE, text = "readParameter: parameter file")
  } else {
    irace_error("'file' must be a character string")
  }

  digits <- as.integer(digits)

  field.match <- function (line, pattern, delimited = FALSE, sep = "[[:space:]]")
  {
    #cat ("pattern:", pattern, "\n")
    positions <- lapply(seq_along(pattern), function(x) regexpr (paste0("^", pattern[x], sep), line))
    if (all(sapply(positions, "[[", 1L) == -1L)) {
      #cat("no match: NULL\n")
      return (list(match = NULL, line = line))
    }
    pos_matched <- lapply(seq_along(pattern), function(x) regexpr (paste0("^", pattern[x]), line))
    #cat("pos.matched:", pos.matched, "\n")
    if (all(sapply(pos_matched, "[[", 1L) == -1L)) {
      #cat(line)
      return (list(match = NULL, line = line))
    }
    position <- which(sapply(pos_matched, `[[`,1L) != -1L)
    if (length(position) > 1L) {
      position <- position[1L]
    }
    pos_matched <- pos_matched[[position]]
    delimited <- as.integer(delimited)
    match <- substr(line, pos_matched[1L] + delimited,
                    attr(pos_matched, "match.length") - delimited)
    #cat("match:",match, "\n")
    line <- substr(line, pos_matched[1L] + attr(pos_matched, "match.length"), nchar(line))
    line <- trim_leading (line)
    #cat(line)
    list(match = match, line = line)
  }

  string2vector <- function(str) {
    v <- c()
    str <- trim(str)
    #cat("string2vector:", str, "\n")
    while (str != "") {
      result <- field.match (str, "\"[^\"]*\"", delimited = TRUE, sep="")
      #cat("result.match: ", result$match,"\n")
      if (is.null(result$match)) {
        result <- field.match (str, "[^,]+", sep="")
        #cat("result.match: ", result$match,"\n")
      }
      v <- c(v, result$match)
      #print(v)
      str <- sub(",[[:space:]]*", "", result$line)
      #print(str)
    }
    v
  }

  errReadParameters <- function(filename, line, context, ...)
  {
    context <- if (is.null (context)) "" else paste0(" when reading: \"", context, "\"")
    fileloc <- if (is.na(filename)) "" else paste0("'", filename, "'," )
    irace_error(paste0(..., collapse = ""), " at ", fileloc, "line ", line, context)
  }

  warnReadParameters <- function(filename, line, context, ...)
  {
    context <- if (is.null (context)) "" else paste0(" when reading: \"", context, "\"")
    fileloc <- if (is.na(filename)) "" else paste0("'", filename, "'," )
    irace_warning(paste0(..., collapse = ""), " at ", fileloc, "line ", line, context)
  }

  parse_condition <- function(s, filename, nbLines, line, context) {
    if (grepl("||", s, fixed=TRUE) || grepl("&&", s, fixed=TRUE)) {
      warnReadParameters (filename, nbLines, line,
                          paste0("Please use '&' and '|' instead of '&&' and '|'", context))
      s <- gsub("||", "|", fixed=TRUE, gsub("&&", "&", fixed=TRUE, s))
    }
    str2expression(s)
  }

  params <- list()
  pnames <- c()
  lines <- readLines(con = file)
  # Delete comments
  lines <- trim(sub("#.*$", "", lines))
  within_global <- FALSE
  nbLines <- 0L
  # Parse [global] first.
  for (line in lines) {
    nbLines <- nbLines + 1L
    if (line == "") next
    if (grepl("^[[:space:]]*\\[forbidden\\]", line)) {
      if (within_global)
        break
      next
    }
    if (within_global) {
      if (grepl("^[[:space:]]*digits[[:space:]]*=[[:space:]]*[0-9]+[[:space:]]*$", line)) {
        eval(parse(text=line))
        if (!is.wholenumber(digits) || digits > 15 || digits < 1)
          errReadParameters(filename, nbLines, line, "'digits' must be an integer within [1, 15]")
        digits <- as.integer(digits)
      } else
        errReadParameters(filename, nbLines, line, "Unknown global option")
      lines[nbLines] <- "" # Do not parse it again.
      next
    }
    if (grepl("^[[:space:]]*\\[global\\]", line)) {
      within_global <- TRUE
      lines[nbLines] <- "" # Do not parse it again.
      next
    }
  }
  forbidden <- NULL
  within_forbidden <- FALSE
  nbLines <- 0L
  for (line in lines) {
    nbLines <- nbLines + 1L
    if (line == "") next

    if (within_forbidden) {
      # FIXME: Better error reporting.
      exp <- parse_condition(line, filename, nbLines, line, " for forbidden expressions")
      forbidden <- c(forbidden, exp)
      next
    }
    if (grepl("^[[:space:]]*\\[forbidden\\]", line)) {
      within_forbidden <- TRUE
      next
    }
    ## Match name (unquoted alphanumeric string)
    result <- field.match (line, "[._[:alnum:]]+")
    name <- result$match
    line <- result$line
    if (is.null(result$match)) {
      errReadParameters (filename, nbLines, line,
                         "Parameter name must be alphanumeric")
    }

    if (name %in% pnames) {
      errReadParameters (filename, nbLines, NULL,
                         "Duplicated parameter name '", name, "'")
    }

    ## Match p_switch (quoted string)
    result <- field.match (line, "\"[^\"]*\"", delimited = TRUE)
    p_label <- result$match
    line <- result$line
    if (is.null(p_label)) {
      errReadParameters (filename, nbLines, line,
                         "Parameter label (switch) must be a double-quoted string")
    }

    ## Match param.type (longer matches must precede shorter ones)
    result <- field.match (line, c("i,log", "r,log", "c", "i", "r", "o"))
    param.type <- result$match
    line <- result$line
    if (is.null (param.type)) {
      errReadParameters(filename, nbLines, line,
                        "Parameter type must be a single character in {'c','i','r','o'}, ",
                        "with 'i', 'r' optionally followed by ',log' (no spaces in between) ",
                        "to sample using a logarithmic scale")
    } else if (param.type == "i,log") {
      param.type <- "i"
      param.transform <- "log"
    } else if (param.type == "r,log") {
      param.type <- "r"
      param.transform <- "log"
    } else {
      param.transform <- ""
    }

    ## Match domain (delimited by parenthesis)
    # Regexp to detect dependent domains of the type ("min(p1)", 100)
    result <- field.match (line, "\\([^|]+\\)", delimited = TRUE, sep = "")
    domain_str <- result$match
    line <- result$line
    if (is.null (domain_str)) {
      errReadParameters (filename, nbLines, line,
                         "Allowed values must be a list within parenthesis")
    }
    # For numerical parameters domains could be dependent
    # thus, we keep the string values in a variable
    # for example (10, param1+2)
    if (param.type %in% c("r","i")) {
      domain <- eval(parse(text=paste0("expression(", domain_str, ")"), keep.source=FALSE))
      # Some expressions like -10 may need to be evaluated again.
      domain <- sapply(domain, function(x) if (!is.expression(x) || length(all.vars(x, unique=FALSE))) x else eval(x), USE.NAMES=FALSE)
      if (is.list(domain)) domain <- as.expression(domain)
      # For dependent domains domain will be NA (we will parse it later)
      if (length(suppressWarnings(as.numeric(sapply(domain, function(x) if(is.language(x)) NA else x)))) != 2L) {
        errReadParameters (filename, nbLines, NULL,
                           "Incorrect numeric range (", result$match,
                           ") for parameter '", name, "'")
      }
    } else { # type %in% c("c", "o")
      domain <- string2vector(domain_str)
    }
    ## Match start of conditions
    result <- field.match (line, "\\|", sep="")
    line <- result$line
    if (!is.null(result$match) && result$match != "") {
      result <- field.match (line, ".*$", sep="")
      condition <- result$match
      if (is.null(result$match) || result$match == "")
        errReadParameters (filename, nbLines, line,
                           "Expected condition after '|'")
      line <- result$line
    } else if (!is.null(result$line) && result$line != "") {
      errReadParameters (filename, nbLines, line,
                         "Expected '|' before condition")
    } else {
      condition <- TRUE
    }
    # *****************************************************************
    p <- tryCatch(
      Parameter(name = name, type = param.type, domain = domain, label = p_label, condition = condition,
        transf = param.transform, digits = digits),
      invalid_domain = function(c) structure(paste0("For parameter '", name, "' of type 'i' values must be integers (",
                                             domain_str, ")"), class = "try-error"),
      invalid_range = function(c) structure(paste0("Lower bound must be smaller than upper bound in numeric range (",
                                                   domain_str, ") for parameter '", name, "'"), class="try-error"),

      error = function(c) structure(conditionMessage(c), class="try-error")
    )
    if (inherits(p, "try-error")) errReadParameters(filename, nbLines, NULL, p)
    params <- c(params, list(p))
    pnames <- c(pnames, name)
  } # end loop on lines

  # Check that we have read at least one parameter
  if (length(params) == 0) {
    if (is.na(filename))
      irace_error("No parameter definition found in the input text")
    else
      irace_error("No parameter definition found, check that the parameter file '",  filename, "' is not empty.")
  }

  if (length(forbidden)) {
    irace_note(length(forbidden), " expression(s) specifying forbidden configurations read.\n")
    check_forbidden_params(forbidden, pnames, filename = filename)
  }
  parameters <- do.call(parametersNew, c(params, list(forbidden=forbidden, debugLevel = debugLevel)))
  if (debugLevel >= 2) {
    print(parameters, digits = 15L)
    irace_note("Parameters have been read\n")
  }
  parameters
}

#' Read parameters in PCS (AClib) format and write them in irace format.
#'
#' @inheritParams readParameters
#'
#' @return A string representing the parameters in irace format.
#'
#' @details Either `file` or `text` must be given. If `file` is given, the
#'  parameters are read from the file `file`. If `text` is given instead,
#'  the parameters are read directly from the `text` character string.
#'  In both cases, the parameters must be given (in `text` or in the file
#'  whose name is `file`) in the expected form.  See the documentation
#'  for details.  If none of these parameters is given, \pkg{irace}
#'  will stop with an error.
#'
#' **FIXME:** Multiple conditions and default configuration are currently ignored. See <https://github.com/MLopez-Ibanez/irace/issues/31>
#'
#' @references
#' Frank Hutter, Manuel López-Ibáñez, Chris Fawcett, Marius Thomas Lindauer, Holger H. Hoos, Kevin Leyton-Brown, and Thomas Stützle. **AClib: A Benchmark Library for Algorithm Configuration**. In P. M. Pardalos, M. G. C. Resende, C. Vogiatzis, and J. L. Walteros, editors, _Learning and Intelligent Optimization, 8th International Conference, LION 8_, volume 8426 of Lecture Notes in Computer Science, pages 36–40. Springer, Heidelberg, 2014.
#'
#' @seealso [readParameters()]
#' @examples
#'  ## Read the parameters directly from text
#'  pcs_table <- '
#'  # name       domain
#'  algorithm    {as,mmas,eas,ras,acs}[as]
#'  localsearch  {0, 1, 2, 3}[0]
#'  alpha        [0.00, 5.00][1]
#'  beta         [0.00, 10.00][1]
#'  rho          [0.01, 1.00][0.95]
#'  ants         [1, 100][10]il
#'  q0           [0.0, 1.0][0]
#'  rasrank      [1, 100][1]i
#'  elitistants  [1, 750][1]i
#'  nnls         [5, 50][5]i
#'  dlb          {0, 1}[1]
#'  Conditionals:
#'  q0 | algorithm in {acs}
#'  rasrank | algorithm in {ras}
#'  elitistants | algorithm in {eas}
#'  nnls | localsearch in {1,2,3}
#'  dlb | localsearch in {1,2,3}
#'  {alpha=0, beta=0}'
#'  parameters_table <- read_pcs_file(text=pcs_table)
#'  cat(parameters_table)
#'  parameters <- readParameters(text=parameters_table)
#'  str(parameters)
#'
#' @author Manuel López-Ibáñez
#' @export
read_pcs_file <- function(file, digits = 4L, debugLevel = 0L, text)
{
  if (missing(file) && !missing(text)) {
    filename <- paste0("text=", deparse(substitute(text)))
    file <- textConnection(text)
    on.exit(close(file))
  } else if (is.character(file)) {
    filename <- file
    file.check (file, readable = TRUE, text = "read_pcs_file: parameter file")
  } else {
    irace_error("'file' must be a character string")
  }
  lines <- readLines(con = file)
  lines <- trim(lines) # Remove leading and trailing whitespace
  lines <- lines[!grepl("Conditionals:", lines, fixed=TRUE)] # useless line
  conditions <- list()
  forbidden <- NULL
  regex_cond <- "^([^[:space:]]+)[[:space:]]+\\|[[:space:]]+(.+)$"
  regex_forbidden <- "^{(.+)}$"
  for (k in seq_along(lines)) {
    if (grepl(regex_cond, lines[k], perl=TRUE)) {
      matches <- regmatches(lines[k],
                            regexec(regex_cond, lines[k], perl=TRUE))[[1L]]
      stopifnot(length(matches) > 0)
      lines[k] <- NA_character_
      conditions[[matches[[2L]]]] <- matches[[3L]]
    } else if (grepl(regex_forbidden, lines[k], perl=TRUE)) {
      forbidden <- c(forbidden, sub(regex_forbidden, "\\1", lines[k], perl=TRUE))
      lines[k] <- NA_character_
    }
  }

  parse_pcs_condition <- function(x, types) {
    if (is.null(x)) return ("")
    matches <- regmatches(x, regexec("([^[:space:]]+)[[:space:]]+in[[:space:]]+\\{([^}]+)\\}$", x, perl=TRUE))[[1L]]
    if (length(matches) == 0L) irace_error("unknown condition ", x)
    param <- matches[[2L]]
    type <- types[[param]]
    if (is.null(type)) irace_error("unknown type for ", param, " in condition: ", x)
    cond <- matches[[3L]]
    if (type == "c" || type == "o") {
      cond <- strsplit(cond, ",[[:space:]]*")[[1L]]
      equal <- (length(cond) == 1L)
      cond <- paste0('"', cond, '"', collapse=',')
    } else {
      equal <- grepl(",", cond, fixed=TRUE)
    }
    if (equal)
      return(paste0(" | ", param, ' == ', cond))
    return(paste0(" | ", param, " %in% c(", cond, ")"))
  }
  param_types <- list()
  param_domains <- list()
  param_comments <- list()
  lines <- lines[!is.na(lines)]
  for (line in lines) {
    if (startsWith(line, "#") || line == "") next
    # match a parameter
    matches <- regmatches(line, regexec("^([^[:space:]]+)[[:space:]]+\\[([^,]+),[[:space:]]*([^]]+)\\][[:space:]]*\\[[^]]+\\](i?l?i?)(.*)$", line, perl=TRUE))[[1]]
    if (length(matches) > 0L) {
      param_name <- matches[[2L]]

      param_type <- paste0(if(grepl("i", matches[5L], fixed=TRUE)) "i" else "r",
                           if(grepl("l", matches[5L], fixed=TRUE)) ",log" else "")
      param_types[[param_name]] <- param_type
      param_domains[[param_name]] <- paste0("(", matches[3L], ", ", matches[4L], ")")
      param_comments[[param_name]] <- matches[6L]
      next
    }
    matches <- regmatches(line, regexec("^([^[:space:]]+)[[:space:]]+\\{([^}]+)\\}[[:space:]]*\\[[^]]+\\](.*)$", line, perl=TRUE))[[1L]]
    if (length(matches) > 0L) {
      param_name <- matches[[2L]]
      param_type <- "c"
      param_types[[param_name]] <- param_type
      param_types[[param_name]] <- param_type
      param_domains[[param_name]] <- paste0("(", matches[3L], ")")
      param_comments[[param_name]] <- matches[4L]
      next
    }
  }
  output <- ""
  for (line in lines) {
    if (startsWith(line, "#") || line == "") {
      output <- paste0(output, line, "\n")
      next
    }
    # match a parameter
    matches <- regmatches(line, regexec("^([^[:space:]]+)[[:space:]]+", line, perl=TRUE))[[1L]]
    if (length(matches) > 0L) {
      param_name <- matches[[2L]]
      cond <- parse_pcs_condition(conditions[[param_name]], param_types)
      output <- paste0(output,
                       sprintf('%s "%s" %s %s%s%s\n',
                               param_name, param_name, param_types[[param_name]], param_domains[[param_name]], cond, param_comments[[param_name]]))
      next
    }
    irace_error("unrecognized line: ", line)
  }
  if (length(forbidden) > 0L) {
    exp <- sapply(forbidden, function(x) {
      # FIXME: this will break if there are "," within the values.
      x <- strsplit(x, ",[[:space:]]*")[[1L]]
      paste0(collapse=" & ",
             sapply(regmatches(x, regexec("^([^=]+)=(.+)$", x, perl=TRUE)), function(matches) {
               rhs <- trim(matches[[3L]])
               if (!any(startsWith(rhs, c("'", "\""))) && suppressWarnings(is.na(as.numeric(rhs))))
                 rhs <- paste0('"', rhs, '"')
               paste0("(", trim(matches[[2L]]), " == ", rhs, ")")
             }, USE.NAMES=FALSE))
    }, USE.NAMES=FALSE)
    output <- paste0(output, "\n[forbidden]\n", paste0(collapse="\n", exp), "\n")
  }
  output
}

#' checkParameters
#'
#' FIXME: This is incomplete, for now we only repair inputs from previous irace
#' versions.
#'
#' @inheritParams printParameters
#' @export
checkParameters <- function(parameters)
{
  ## if (is.null(parameters$isDependent)) {
  ##   parameters$isDependent <- sapply(parameters$domains, is.expression)
  ##   names(parameters$isDependent) <- parameters$names
  ## }
  if (!inherits(parameters, "ParameterSpace")) {
    irace_error("parameters must be an object of class 'ParameterSpace'")
  }
  parameters
}

withr::with_output_sink("test-readconfs.Rout", {

params <- readParameters("parameters.txt")

expect_error_readconfs <- function(params, table, exp)
  expect_error(irace::readConfigurationsFile(text=table, parameters = params), exp)

test_that("no error", {
  x <- irace::readConfigurationsFile(text='
param1 param2 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
NA       NA   "x3"   4.5   "low"', parameters=params)
  expect_equal(nrow(x), 3L)
  expect_equal(ncol(x), 6L) # It adds the fixed column
})

test_that("wrong fixed", {
  expect_error_readconfs(params, '
param1 param2 mode   real mutation fixed
5        NA    "x2"   4.0   "low"  fixed
1        NA    "x2"   4.0   "low"  fixed
NA       NA   "x3"   4.5   "low"   wrong', "is not among the valid values")
})

test_that("checkDuplicates", {
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
5        NA    "x2"   4.0   "low"
NA        NA   "x3"   4.5   "low"
', "Duplicated")

  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
', "Duplicated")

})

test_that("parameter values", {
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5       NA    "x2"   4.0   "low"
1       NA    "x2"   4.0   "low"
11      NA    "x2"   4.0   "low"
', "is not within the valid range")
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5       NA    "x2"   4.0   "low"
1       NA    "x2"   4.5001 "low"
', "is not within the valid range")
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5       NA    "x2"   4.0   "low"
1.1     NA    "x2"   4.5   "low"
', "is not an integer")
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5       NA    "x2"   4.0   "low"
1       NA    "x2"   4.5   "lower"
', "is not among the valid values")
  expect_error_readconfs(params, '
param1 param2 mode   real mutation
5       NA    "x3"   4.0   "low"
1       NA    "x2"   4.5   "low"
', "is not enabled")
})

test_that("parameter names", {
  expect_error_readconfs(params, '
param1 param0 mode   real mutation
5        NA    "x2"   4.0   "low"
1        NA    "x2"   4.0   "low"
', "do not match")
  expect_error_readconfs(params, '
param1 mode   real mutation
5       "x2"   4.0   "low"
1       "x2"   4.0   "low"
', "are missing")
  expect_error_readconfs(params, '
param1 param2 mode   real mutation param3
5        NA    "x2"   4.0   "low"  NA
1        NA    "x2"   4.0   "low"  NA
', "do not match")
})


test_that("conditional fixed", {

  params <- readParameters(text='
param1 "" c ("fixed")
param2 "" c ("cond_fixed") | param3 == "on"
param3 "" c ("on", "off")
')

  x <- irace::readConfigurationsFile(text='
param1 param3
fixed  off
', parameters=params)

  expect_equal(x[["param2"]], NA_character_)

  x <- irace::readConfigurationsFile(text='
param1 param2 param3
fixed  NA     off
', parameters=params)

  x <- irace::readConfigurationsFile(text='
param1 param3
fixed  on
', parameters=params)

  expect_equal(x[["param2"]], "cond_fixed")

  expect_error_readconfs(params, '
param1 param2      param3
fixed  cond_fixed     off
', "is not enabled because of condition")
})

}) # withr::with_output_sink

# This requires the installation of https://github.com/automl/GenericWrapper4AC
test_that("GenericWrapper4AC", {
  skip_on_cran()
  skip_on_coverage()
  skip_on_ci()
  skip_on_travis()
  skip_on_os("mac")
  # FIXME: how to check that a python package is installed?
  parameters <- readParameters(text = '
   cost "cost" r (0.1, 1.00)
   runtime "runtime" r (0.1, 1.00)', digits = 15L)

  configurations <- data.frame(.ID. = 1L, cost = 0.5, runtime = 0.8)

  instances = c("time", "cost", "cost+time")
  names(instances) = instances
  dummy_wrapper <- test_path("dummy_wrapper.py")
  scenario <- list(targetRunner = dummy_wrapper, instances = instances,
    maxExperiments = 1000, aclib = TRUE,
    parameters = parameters)

  scenario <- checkScenario (scenario)
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost", 2),
                                              seeds = 0:1,
                                              bounds = NULL)
  race_state <- irace:::RaceState$new(scenario)
  output <- irace:::execute_experiments(race_state, experiments, scenario)
  expect_equal(output[[1]]$status, "SUCCESS")
  expect_equal(output[[1]]$cost, 0.5)
  expect_true(is.null(output[[1]]$error))
  expect_equal(output[[2]]$status, "TIMEOUT")
  expect_equal(output[[2]]$cost, 0.5)
  expect_true(is.null(output[[2]]$error))
  
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost", 2),
                                              seeds = 2, bounds = NULL)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "CRASHED")
               
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost", 2),
                                              seeds = 3, bounds = NULL)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "ABORT")
  
  
  scenario <- modifyList(scenario,
                         list(capping = TRUE, boundMax = 1, maxTime = 10,
                              maxExperiments = NULL))
  scenario <- checkScenario (scenario)
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("time", 2),
                                              seeds = 0:1,
                                              bounds = scenario$boundMax)
  output <- irace:::execute_experiments(race_state, experiments, scenario)

  expect_equal(output[[1]]$status, "SUCCESS")
  expect_equal(output[[1]]$cost, 0.8)
  expect_equal(output[[1]]$time, 0.8)
  expect_match(output[[1]]$call, "--cutoff")
  expect_true(is.null(output[[1]]$error))
  expect_equal(output[[2]]$status, "TIMEOUT")
  expect_equal(output[[2]]$cost, 0.8)
  expect_equal(output[[2]]$time, 0.8)
  expect_match(output[[2]]$call, "--cutoff")
  expect_true(is.null(output[[2]]$error))

  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("time", 2),
                                              seeds = 2,
                                              bounds = scenario$boundMax)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "CRASHED")
               
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("time", 2),
                                              seeds = 3,
                                              bounds = scenario$boundMax)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "ABORT")

  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost+time", 2),
                                              seeds = 0:1,
                                              bounds = scenario$boundMax)
  output <- irace:::execute_experiments(race_state, experiments, scenario)

  expect_equal(output[[1]]$status, "SUCCESS")
  expect_equal(output[[1]]$cost, 0.5)
  expect_equal(output[[1]]$time, 0.8)
  expect_match(output[[1]]$call, "--cutoff")
  expect_true(is.null(output[[1]]$error))
  expect_equal(output[[2]]$status, "TIMEOUT")
  expect_equal(output[[2]]$cost, 0.5)
  expect_equal(output[[2]]$time, 0.8)
  expect_match(output[[2]]$call, "--cutoff")
  expect_true(is.null(output[[2]]$error))

  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost+time", 2),
                                              seeds = 2,
                                              bounds = scenario$boundMax)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "CRASHED")
               
  experiments <- irace:::createExperimentList(configurations,
                                              parameters = parameters,
                                              instances = scenario$instances,
                                              instances_ID = rep("cost+time", 2),
                                              seeds = 3,
                                              bounds = scenario$boundMax)
  expect_error(irace:::execute_experiments(race_state, experiments, scenario), "ABORT")
})

context("bug")
skip_on_cran()
# https://github.com/MLopez-Ibanez/irace/issues/13
withr::with_output_sink("test-bug-13.Rout", {

parameters.txt <- '
algorithm    "--"             c    (as,mmas,eas,ras,acs)
localsearch  "--localsearch " c    (0, 1, 2, 3)
alpha        "--alpha "       r    (0.00, 5.00)
beta         "--beta "        r    (0.00, 10.00)
rho          "--rho "         r    (0.01, 1.00)
ants         "--ants "        i    (5, 100)
q0           "--q0 "          r    (0.0, 1.0)           | algorithm == "acs"
rasrank      "--rasranks "    i    (1, 100)             | algorithm == "ras"
elitistants  "--elitistants " i    (1, 750)             | algorithm == "eas"
nnls         "--nnls "        i    (5, 50)              | localsearch %in% c(1,2,3)
dlb          "--dlb "         c    (0, 1)               | localsearch %in% c(1,2,3)
'
parameters <- irace:::readParameters(text=parameters.txt)
  
lookup <- readRDS("bug-13-lookup.rds")
counter <- 0
target.runner <- function(experiment, scenario) {
  configuration <- experiment$configuration
  instance <- experiment$instance
  switches <- experiment$switches
  debugLevel <- scenario$debugLevel
  args <- paste(instance, trimws(irace:::buildCommandLine(configuration, switches)))
  pos <- pmatch(args, lookup[,1])
  if(!is.na(pos)) {
    cost <- as.numeric(lookup[pos, 2])
    counter <<- counter + 1
  } else {
    cost <- 10
  }
  if(debugLevel > 0) {
    cat("# ", args, "\n")
    cat(cost, "\n")
  }
  return(list(cost = cost, call = toString(experiment)))
}

withr::with_options(list(warning=2), {
  scenario <- list(instances = scan(what="", quiet=TRUE, text="
2000-121.tsp
2000-122.tsp
2000-123.tsp
2000-124.tsp
2000-125.tsp
2000-126.tsp
2000-127.tsp
2000-128.tsp
2000-129.tsp
2000-130.tsp"),
trainInstancesDir = "",
trainInstancesFile = "",
maxExperiments = 3000,
digits = 3,
seed = 2357,
deterministic = 1,
elitist = 0,
targetRunner = target.runner,
debugLevel=3)
  scenario <- checkScenario (scenario)
  confs <- irace(scenario = scenario, parameters = parameters)
  expect_equal(counter, 548)
})

})

withr::with_output_sink("test-capping.Rout", {


## target runner ###########################################################
target_runner_reject <- function(experiment, scenario)
{
  if (experiment$configuration[["reject"]] == "1" && runif(1) <= 0.01)
    return(list(cost = -Inf, time = experiment$bound))
  target_runner_capping_xy(experiment, scenario)
}


test_that("irace_capping_xy maxExperiments = 1000", {
  generate_set_seed()
  irace_capping_xy(maxExperiments = 1000)
})

test_that("irace_capping_xy maxExperiments = 1000 cappingAfterFirstTest", {
  generate_set_seed()
  irace_capping_xy(maxExperiments = 1000, cappingAfterFirstTest=1)
})

test_that("irace_capping_xy maxTime = 1000", {
  generate_set_seed()
  expect_warning(irace_capping_xy(maxTime = 1000),
                 "boundMax=80 is too large, using 5 instead")
})

test_that("irace_capping_xy targetRunner = target_runner_reject, maxTime = 1000", {
  skip_on_cran()
  skip_on_coverage() # This test sometimes fails randomly
  generate_set_seed()
  irace_capping_xy(targetRunner = target_runner_reject, maxTime = 1000, boundMax = 5, debugLevel = 3,
    # FIXME: target_runner_reject does not work in parallel on Windows.
    parallel = if (system_os_is_windows()) 1L else test_irace_detectCores())
})

test_that("capping default value", {
  parameters_table <- '
   x "" r (0, 1.00)
   y "" r (0, 1.00)
   reject "" c (0,1)'

  parameters <- readParameters(text = parameters_table)

  def_scenario <- list(instances = c("ackley", "goldestein", "matyas", "himmelblau"),
                   targetRunner = target_runner_reject,
                   maxTime = 1200,
                   parameters = parameters)
  scenario <- checkScenario(def_scenario)
  expect_false(scenario$capping)
  def_scenario$boundMax <- 80
  scenario <- checkScenario(def_scenario)
  expect_true(scenario$capping)
})

}) # withr::with_output_sink()

withr::with_output_sink("test-bug1col.Rout", {

test_that("bug1col", {

  target_runner <- function(experiment, scenario)
    list(cost = runif(1))

    parameters <- readParameters(text = '
c_1 "c_1" c (1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20)
')
    scenario <- list(targetRunner = target_runner,
      instances = 1:10,
      configurationsFile="confs1col.txt",
      maxExperiments = 200,
      testType = "F-test",
      nbIterations = 1,
      postselection = 0,
      minNbSurvival = 1,
      elitistLimit = 0,
      logFile = "",
      parameters = parameters)
    confs <- irace(scenario = scenario)
    expect_false(is.null(confs))
})

}) # withr::with_output_sink()

withr::with_output_sink("test-multi_irace.Rout", {

# FIXME: Use temporary files and directories.
make.target.runner <- function(parameters) {
  function(experiment, scenario) {
    cost <- max(1, abs(rnorm(1, mean=sum(unlist(experiment$configuration[parameters])))))
    list(cost = cost, call = toString(experiment))
  }
}

make.parameters.table <- function(parameters)
  readParameters(text = sprintf('%s "" r (0, 10)', parameters))


make.scenario <- function(targetRunner, maxExperiments = 1000) {
  list(targetRunner = targetRunner, maxExperiments = maxExperiments, instances = rnorm(200, mean = 0.9, sd = 0.02))
}

target.runner.1 <- make.target.runner(c("x1", "x2"))
target.runner.2 <- make.target.runner(c("x2", "x3"))
target.runner.3 <- make.target.runner(c("x3", "x1"))

parameters.table.1 <- make.parameters.table(c("x1", "x2"))
parameters.table.2 <- make.parameters.table(c("x2", "x3"))
parameters.table.3 <- make.parameters.table(c("x3", "x1"))

check.default.logFiles <- function(n) {
  dir <- sprintf("run_%02d", 1:n)
  expect_equal(dir.exists(dir), rep_len(TRUE, n))
  expect_equal(file.exists(file.path(dir, "irace.Rdata")), rep_len(TRUE, n))
}

test_that("multiple scenarios, multiple parameters", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  scenarios <- lapply(list(target.runner.1, target.runner.2, target.runner.3), make.scenario)
  parameters <- list(parameters.table.1, parameters.table.2, parameters.table.3)

  runs <- multi_irace(scenarios, parameters)

  expect_length(runs, 3)
  check.default.logFiles(3)
})

test_that("one scenario, multiple parameters", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  scenarios <- list(make.scenario(target.runner.1))
  dummy_parameters_names <- c("dummy1", "dummy2", "dummy3")
  parameter_names <- lapply(dummy_parameters_names, function(dummy) c("x1", "x2", dummy))
  parameters <- lapply(parameter_names, make.parameters.table)

  runs <- multi_irace(scenarios, parameters)
  expect_length(runs, 3)
  check.default.logFiles(3)
})

test_that("multiple scenarios, one parameters", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  scenarios <- lapply(list(100, 500, 1000), function(maxExperiments) make.scenario(target.runner.1, maxExperiments) )
  parameters <- list(parameters.table.1)

  runs <- multi_irace(scenarios, parameters)
  expect_length(runs, 3)
  check.default.logFiles(3)
})

test_that("one scenario, one parameters, multiple n", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  scenarios <- list(make.scenario(target.runner.1))
  parameters <- list(parameters.table.1)
  runs <- multi_irace(scenarios, parameters, n = 3)
  expect_length(runs, 3)
  check.default.logFiles(3)
})

test_that("multiple scenarios, multiple parameters, multiple n", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  scenarios <- lapply(list(target.runner.1, target.runner.2, target.runner.3), make.scenario)
  parameters <- list(parameters.table.1, parameters.table.2, parameters.table.3)

  runs <- multi_irace(scenarios, parameters, n = 3)
  expect_length(runs, 9)
  check.default.logFiles(9)
})

test_that("logFile not in execDir", {
  skip_on_cran()
  # Reproducible results
  generate_set_seed()

  execDir <- path_rel2abs("./multi_irace_execDir")
  logFileDir <- path_rel2abs("./multi_irace_logFileDir")

  dir.create(execDir, showWarnings = FALSE)
  dir.create(logFileDir, showWarnings = FALSE)

  scenarios <- lapply(list(target.runner.1, target.runner.2, target.runner.3), make.scenario)
  for (i in seq_along(scenarios)) {
    scenarios[[i]]$logFile <- file.path(logFileDir, "irace.Rdata")
    scenarios[[i]]$execDir <- execDir
  }

  parameters <- list(parameters.table.1, parameters.table.2, parameters.table.3)

  expect_error(multi_irace(scenarios, parameters))

  # expect_length(runs, 3)
  # for (i in 1:3) {
  #   logFile <- file.path(logFileDir, sprintf("irace_%02d.Rdata", i))
  #   expect_true(file.exists(logFile))
  # }
})

test_that("global seed", {
  skip_on_cran()

  scenarios <- lapply(list(target.runner.1, target.runner.2, target.runner.3), make.scenario)
  parameters <- list(parameters.table.1, parameters.table.2, parameters.table.3)

  runs.1 <- multi_irace(scenarios, parameters, global_seed = 42)
  runs.2 <- multi_irace(scenarios, parameters, global_seed = 42)

  expect_equal(runs.1, runs.2)
})

test_that("sequential and parallel identical", {
  skip_on_cran()
  skip_on_os("windows")
  ncores <- test_irace_detectCores()
  skip_if(ncores <= 1L,
          message = "This test only makes sense if multiple cores are available")
  scenarios <- lapply(list(target.runner.1, target.runner.2, target.runner.3), make.scenario)
  parameters <- list(parameters.table.1, parameters.table.2, parameters.table.3)

  runs.sequential <- multi_irace(scenarios, parameters, global_seed = 42)
  runs.parallel <- multi_irace(scenarios, parameters, global_seed = 42, parallel = ncores)

  expect_equal(runs.sequential, runs.parallel)
})

}) # withr::with_output_sink()

# https://github.com/MLopez-Ibanez/irace/issues/55
withr::with_output_sink("test-bug-55.Rout", {
test_that("bug-55", {
    
  parameters_txt <- '
foo "--foo " i (0, 1)
foo2 "--foo2 " c (true, false) | foo == 0
'
  params <- readParameters(text=parameters_txt)
  configurations_txt <- '
  foo  foo2
1   0 false
2   1  <NA>
'
  confs <- readConfigurationsFile(text=configurations_txt, parameters = params)
  expect_equal(confs, data.frame(foo=c(0L,1L), foo2=c("false",NA), row.names=c("1", "2"), stringsAsFactors=FALSE))
})
})

withr::with_output_sink("test-targeteval.Rout", {

  target_evaluator <- function(experiment, num_configurations, all_conf_id,
                               scenario, target_runner_call) {
    withr::local_seed(experiment$seed)
    list(cost = runif(1), call = deparse1(experiment))
  }

  parameters <- readParameters(text = '
algorithm       "--"                 c          (as,mmas,eas,ras,acs)
')

test_that("target_evaluator", {

  target_runner <- function(experiment, scenario) {
    list(call = deparse1(experiment))
  }
  seed <- sample.int(min(2147483647L, .Machine$integer.max), size = 1L, replace = TRUE)
  instances <- 1:10
  limit <- 58L
  maxExperiments <- 200
  logFile <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  
  scenario <- checkScenario(list(
    targetRunner = target_runner, targetEvaluator = target_evaluator,
    maxExperiments = maxExperiments, instances = instances,
    logFile = logFile,
    seed = seed,
    quiet = TRUE,
    parameters = parameters))

  expect_true(irace:::checkTargetFiles(scenario = scenario))

  scenario <- checkScenario(list(
    targetRunner = target_runner, targetEvaluator = target_evaluator,
    maxExperiments = maxExperiments, instances = instances,
    logFile = logFile,
    seed = seed,
    quiet = TRUE,
    parameters = parameters))
  
  expect_silent(confs <- irace(scenario = scenario))
  expect_gt(nrow(confs), 0L)
  scenario$targetRunner <- wrap_target_runner_error(target_runner, limit)
  # Otherwise, the tests are too fast.
  with_mocked_bindings({
    expect_error(irace(scenario = scenario), "== irace == The cost returned by targetRunner is not numeric")
  },
  .irace_minimum_saving_time = 0
  )

  logFile_new <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  scenario <- modifyList(scenario, list(
    targetRunner = wrap_target_runner_counter(target_runner),
    recoveryFile = logFile,
    seed = NA,
    quiet = FALSE,
    logFile = logFile_new))
  recover_confs <- irace(scenario = scenario)
  expect_identical(confs, recover_confs)
  expect_equal(environment(scenario$targetRunner)$counter, maxExperiments - limit + 1L)
})

test_that("target_evaluator maxTime", {

  target_runner <- function(experiment, scenario) {
    withr::local_seed(experiment$seed)
    list(time = min(experiment$bound, as.integer(1 + 10*runif(1))))
  }
  
  seed <- sample.int(min(2147483647L, .Machine$integer.max), size = 1L, replace = TRUE)
  instances <- 1:10
  limit <- 100L
  logFile <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  scenario <- checkScenario(list(
     targetRunner = target_runner, targetEvaluator = target_evaluator,
     maxTime = 2000, boundMax = 10, instances = instances,
     logFile = logFile, seed = seed,
     quiet = TRUE,
     parameters = parameters))
  expect_true(scenario$capping)
  expect_silent(confs <- irace(scenario = scenario))
  expect_gt(nrow(confs), 0L)

  scenario$targetRunner <- wrap_target_runner_error(target_runner, limit)
  # Otherwise, the tests are too fast.
  with_mocked_bindings({
    expect_error(irace(scenario = scenario), "== irace == The cost returned by targetRunner is not numeric")
  },
  .irace_minimum_saving_time = 0
  )

  logFile_new <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  scenario <- modifyList(scenario, list(
    targetRunner = wrap_target_runner_counter(target_runner),
    recoveryFile = logFile,
    seed = NA,
    quiet = FALSE,
    logFile = logFile_new))
  recover_confs <- irace(scenario = scenario)
  expect_identical(confs, recover_confs)
  expect_lt(environment(scenario$targetRunner)$counter,
    sum(!is.na(read_logfile(logFile_new)$experiments)) - 10L)
})

}) # withr::with_output_sink()

withr::with_output_sink("test-read_pcs_file.Rout", {
test_that("read_pcs_file", {
  pcs_table <- '
# name       domain
algorithm    {as,mmas,eas,ras,acs}[as]
localsearch  {0, 1, 2, 3}[0]
alpha        [0.00, 5.00][1]
beta         [0.00, 10.00][1]
rho          [0.01, 1.00][0.95]l
ants         [1, 100][10]il
q0           [0.0, 1.0][0]
rasrank      [1, 100][1]i
elitistants  [1, 750][1]i
nnls         [5, 50][5]i
dlb          {0, 1}[1]
Conditionals:
q0 | algorithm in {acs}
rasrank | algorithm in {ras}
elitistants | algorithm in {eas}
nnls | localsearch in {1,2,3}
dlb | localsearch in {1,2,3}
 {alpha=0, beta= 0.0} 
{ants=1,algorithm=eas}
 {ants=1,algorithm="ras"}'
  
  parameters_table <- '
# name       domain
algorithm "algorithm" c (as,mmas,eas,ras,acs)
localsearch "localsearch" c (0, 1, 2, 3)
alpha "alpha" r (0.00, 5.00)
beta "beta" r (0.00, 10.00)
rho "rho" r,log (0.01, 1.00)
ants "ants" i,log (1, 100)
q0 "q0" r (0.0, 1.0) | algorithm == "acs"
rasrank "rasrank" i (1, 100) | algorithm == "ras"
elitistants "elitistants" i (1, 750) | algorithm == "eas"
nnls "nnls" i (5, 50) | localsearch %in% c("1","2","3")
dlb "dlb" c (0, 1) | localsearch %in% c("1","2","3")

[forbidden]
(alpha == 0) & (beta == 0.0)
(ants == 1) & (algorithm == "eas")
(ants == 1) & (algorithm == "ras")
'
  expect_equal(parameters_table, read_pcs_file(text=pcs_table))
})
})

withr::with_output_sink("test-recovery.Rout", {

test_that("recovery works", {

target_runner_xy <- function(experiment, scenario)
{
  configuration <- experiment$configuration
  instance      <- experiment$instance

  x <- configuration[["x"]]
  y <- configuration[["y"]]
  value <- switch(instance,
                  ackley     = f_ackley(x, y),
                  goldestein = f_goldestein_price(x, y),
                  matyas     = f_matyas(x, y),
                  himmelblau  = f_himmelblau(x, y))
  list(cost = value)
}

parameters_table <- '
   x "" r (0, 1.00)
   y "" r (0, 1.00)
   '

  parameters <- readParameters(text = parameters_table)
  logFile <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")

  seed <- sample.int(min(2147483647L, .Machine$integer.max), size = 1, replace = TRUE)
  maxExperiments <- 500L
  limit <- 100L

  scenario <- list(
    instances = c("ackley", "goldestein", "matyas", "himmelblau"),
    parameters = parameters,
    targetRunner = target_runner_xy,
    logFile = logFile,
    seed = seed,
    quiet = TRUE,
    maxExperiments = maxExperiments)

  expect_silent(confs <- irace(scenario = scenario))

  scenario$targetRunner <- wrap_target_runner_error(target_runner_xy, limit)
  # Otherwise, the tests are too fast.
  with_mocked_bindings({
    expect_error(irace(scenario = scenario), "== irace == The cost returned by targetRunner is not numeric")
  },
  .irace_minimum_saving_time = 0
  )

  logFile_new <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  scenario <- modifyList(scenario, list(
    targetRunner = wrap_target_runner_counter(target_runner_xy),
    recoveryFile = logFile,
    seed = NA,
    quiet = FALSE,
    logFile = logFile_new))
  recover_confs <- irace(scenario = scenario)
  expect_identical(confs, recover_confs)
  expect_lt(environment(scenario$targetRunner)$counter, maxExperiments - 50)
})

test_that("recovery maxTime", {
  target_runner <- function(experiment, scenario) {
    configuration     <- experiment$configuration
    tmax <-  configuration[["tmax"]]
    temp <-  configuration[["temp"]]
    time <- max(1, abs(rnorm(1, mean=(tmax+temp)/10)))
    list(cost = time, time = time)
  }
  parameters <- readParameters(text = '
   tmax "" i (1, 50)
   temp "" r (0, 10)
   ')
  logFile <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  seed <- 1234567

  scenario <- list(targetRunner = target_runner,
    instances = 1:10,
    seed = seed,
    maxTime = 500,
    logFile = logFile,
    quiet = TRUE,
    parameters = parameters)
  expect_silent(confs <- irace(scenario = scenario))

  scenario$targetRunner <- wrap_target_runner_error(target_runner, 200L)
  # Otherwise, the tests are too fast.
  with_mocked_bindings({
    expect_error(irace(scenario = scenario), "== irace == The cost returned by targetRunner is not numeric")
  },
  .irace_minimum_saving_time = 0
  )
  
  logFile_new <- withr::local_tempfile(pattern = "irace", fileext = ".Rdata")
  scenario <- modifyList(scenario, list(
    targetRunner = wrap_target_runner_counter(target_runner),
    recoveryFile = logFile,
    seed = NA,
    quiet = FALSE,
    logFile = logFile_new))
  recover_confs <- irace(scenario = scenario)
  expect_identical(confs, recover_confs)
  expect_lt(environment(scenario$targetRunner)$counter,
    nrow(read_logfile(logFile_new)$state$experiment_log) - 150L)
})
  
}) # withr::with_output_sink()

test_that("similarConfigurations", {

  parameters <- readParameters(text = '
n1 "" r (0,1)
n2 "" r (0,1)
n3 "" r (0,1)
c1 "" c ("a","b")
c2 "" c ("a","b")
c3 "" c ("a","b")
')
  confs <- read.table(header = TRUE, stringsAsFactors = FALSE, text = '
.ID. n1 n2 n3 c1 c2 c3
1    0.5  0.5  0.5      "a" "a" "a"
2    NA  0.5  0.5       "a" "a" "a"
3    0.5  0.5  0.5      "a" "a" "b"
4    0.5  0.501  0.5    "a" "a" "a"
5    0.5  0.5    0.499  "a" "a" "a"
6    0.5  0.5  0.5      "a" "a" "a"
7    0.5  0.5  0.5      "a"  NA "a"
8    0.5  0.1  0.5      "a" "a" "a"
9    0.5  0.49  0.5     "a" "a" "a"
10   0.5  0.5  0.5      "a" "a"  NA
11   0.5  0.5  0.5      "a" "a"  NA
12    NA  0.5  0.5      "a" "a" "a"
')
  expect_identical(
    irace:::similarConfigurations(confs[1:10,], parameters, threshold = 0.001),
    as.integer(c(1,6)))
  expect_identical(
    irace:::similarConfigurations(confs[1:10,], parameters, threshold = 0.01),
    as.integer(c(1,4,5,6)))
  expect_identical(
    irace:::similarConfigurations(confs[1:10,], parameters, threshold = 0.1),
    as.integer(c(1,4,5,6,9)))
  expect_identical(
    # FIXME: The output should be already sorted
    sort(irace:::similarConfigurations(confs, parameters, threshold = 0.001)),
    as.integer(c(1,2,6,10,11,12)))
})

withr::with_output_sink("test-targetRunnerParallel.Rout", {

test_that("targetRunnerParallel", {

setClasses <- function(x, classes)
{
  class(x) = classes
  x
}

# test that targetRunnerParallel works with arbitrary instance objects.
targetRunnerParallel <- function(experiment, exec_target_runner, scenario, target_runner)
{
  # get our param settings that irace should try
  cands = lapply(experiment, "[[", "configuration")
  # the instance is always the same for all different param setting
  theinst = experiment[[1L]]$instance
  # we check that we have instances of correct class
  expect_s3_class(theinst, "foo", exact = TRUE)
  # fabricate some random fitness vals
  ys = rnorm(length(cands))
  lapply(ys, function(y) list(cost = y, time = NA_real_))
}

   n.inst = 7L
   instances = replicate(n.inst, setClasses(list(x=123), classes = "foo"), simplify = FALSE)
   parameters = readParameters(text='x1 "" r (0,1)')
   log.file = tempfile()
   tuner_config = list(maxExperiments = 40L, nbIterations = 1L, minNbSurvival = 1L,
                       targetRunnerParallel = targetRunnerParallel,
                       instances = instances, logFile = log.file)
   confs <- irace(scenario = tuner_config)
   expect_gt(nrow(confs), 0L)
})

test_that("targetRunnerData", {
  targetRunnerParallel <- function(experiments, exec_target_runner, scenario, target_runner) {
    cat("a = ", scenario$targetRunnerData$a,
        ", b = ", scenario$targetRunnerData$b, "\n", sep = "")
    # get our param settings that irace should try
    cands = lapply(experiments, "[[", "configuration")
    # fabricate some random fitness vals
    ys = rnorm(length(cands))
    ys = lapply(ys, function(y) list(cost = y, time = NA_real_))
    return(ys)
  }
  parameters = readParameters(text='
x "x" r (1,2)
')
  expect_output(
    irace(scenario = list(targetRunnerParallel = targetRunnerParallel,
                          instances = replicate(5, list(10)),
                          targetRunnerData = list(a=1, b=2),
                          maxExperiments = 42L, parameters = parameters)),
    "a = 1, b = 2")
})

}) # with_output_sink()

withr::with_output_sink("test-badoutputraw.Rout", {

  test_that("outputRaw is vector", {

    target_runner <- function(experiment, scenario)
      list(cost = 100 + rnorm(1, 0, 0.1), outputRaw=c("a", "b"))

      parameters <- readParameters(text='
algorithm    "--"             c    (as,mmas,eas,ras,acs)
alpha        "--alpha "       r    (0.00, 5.00)
')
      scenario <- list(targetRunner = target_runner,
        instances=1:10,
        maxExperiments = 500, logFile = "",
        parameters = parameters)
      confs <- irace(scenario = scenario)
      expect_false(is.null(confs))
  })
})

withr::with_output_sink("test-bug-10.Rout", {

parameters_txt <- '
algorithm    "--"             c    (as,mmas,eas,ras,acs)
localsearch  "--localsearch " c    (0, 1, 2, 3)
alpha        "--alpha "       r    (0.00, 5.00)
beta         "--beta "        r    (0.00, 10.00)
rho          "--rho "         r    (0.01, 1.00)
ants         "--ants "        i    (5, 100)
q0           "--q0 "          r    (0.0, 1.0)           | algorithm == "acs"
rasrank      "--rasranks "    i    (1, 100)             | algorithm == "ras"
elitistants  "--elitistants " i    (1, 750)             | algorithm == "eas"
nnls         "--nnls "        i    (5, 50)              | localsearch %in% c(1,2,3)
dlb          "--dlb "         c    (0, 1)               | localsearch %in% c(1,2,3)
'

parameters <- readParameters(text=parameters_txt)

test_that("bug blocksize", {
  skip_on_cran()

  target_runner <- function(experiment, scenario)
    list(cost = 100 + rnorm(1, 0, 0.1))

withr::with_options(list(warning=2), {
  scenario <- list(targetRunner = target_runner,
                   instances=1:5, firstTest=5*5, eachTest=5,
                   sampleInstances=FALSE,
                   maxExperiments = 5000, logFile = "",
                   elitistNewInstances = 1,
                   elitist = TRUE,
                   parameters = parameters)
  scenario <- checkScenario (scenario)
  confs <- irace(scenario = scenario)
  expect_false(is.null(confs))
})
})

# https://github.com/MLopez-Ibanez/irace/issues/10
test_that("bug 10", {
  skip_on_cran()
  
  target_runner <- function(experiment, scenario)
    list(cost = 100, call = toString(experiment))

withr::with_options(list(warning=2), {
  scenario <- list(targetRunner = target_runner,
                   instances=1:10,
                   maxExperiments = 5000, logFile = "",
                   deterministic = TRUE,
                   elitistNewInstances = 0,
                   elitistLimit = 0,
                   elitist = 0,
                   parameters = parameters)
  scenario <- checkScenario (scenario)
  confs <- irace(scenario = scenario)
  expect_false(is.null(confs))
})
})

}) # withr::with_output_sink

# https://github.com/MLopez-Ibanez/irace/issues/44
withr::with_output_sink("test-bug-44.Rout", {
test_that("bug 44", {
  readParameters(text='a "a" r (0.01, 0.99)', digits=2)
  readParameters(text='a "a" r (0, 1)', digits=1)
  expect_error(readParameters(text='a "a" r (0.01, 0.99)', digits=1), "must be representable")
  expect_error(readParameters(text='a "a" r,log (1e-8, 1)', digits=4), "must be representable")
})
})

withr::with_output_sink("test-bug71.Rout", {

test_that("bug71.evaluator", {

  target.runner <- function(experiment, scenario)
    list(cost = runif(1), call = toString(experiment))

    parameters <- readParameters(text = '
algorithm       "--"                 c          (as,mmas,eas,ras,acs)
')
    scenario <- checkScenario(list(
      targetRunner = target.runner,
      maxExperiments = 200, instances = runif(10),
      initConfigurations = data.frame(algorithm="as"),
     parameters = parameters))
    expect_true(irace:::checkTargetFiles(scenario = scenario))
})

}) # withr::with_output_sink()

# Uses 10 decimal places at most so that there is space for '-', '.', and
# 'e-NN' within the 16 spaces.
.irace.format.perf <- "%#16.10g"

.onLoad <- function(libname, pkgname) {
  # FIXME:  We would like to use %#16.10g but this causes problems with
  # https://github.com/oracle/fastr/issues/191
  R_engine <- R.version$engine
  if (!is.null(R_engine) && R_engine == "FastR")
    .irace.format.perf <<- "%16.10g"  
  invisible()
}

#' Reads from a file the scenario settings to be used by \pkg{irace}.
#'
#' The scenario argument is an initial scenario that is overwritten for every
#' setting specified in the file to be read.
#'
#' @param filename `character(1)`\cr Filename from which the scenario will
#'   be read. If empty, the default `scenarioFile` is used.  An example
#'   scenario file is provided in `system.file(package="irace", "templates/scenario.txt.tmpl")`.
#' @inheritParams defaultScenario
#'
#' @return The scenario list read from the file. The scenario settings not
#'   present in the file are not present in the list, i.e., they are `NULL`.
#'
#' @seealso
#'  \describe{
#'  \item{[printScenario()]}{prints the given scenario.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#'  \item{[checkScenario()]}{to check that the scenario is valid.}
#' }
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
readScenario <- function(filename = "", scenario = list(),
                         params_def = .irace.params.def)
{
  # This function allows recursively including scenario files.
  scenario_env <- new.env()
  include.scenario <- function(rfilename, topfile = filename, envir. = scenario_env)
  {
    if (!file.exists (rfilename)) {
      irace_error ("The scenario file ", shQuote(rfilename), " included from ",
                   shQuote(topfile), " does not exist.")
    }
    handle.source.error <- function(e) {
      irace_error("Reading scenario file ", shQuote(rfilename),
                  " included from ", shQuote(topfile),
                  " produced the following errors or warnings:\n",
                  paste0(conditionMessage(e), collapse="\n"))
      return(NULL)
    }
    withCallingHandlers(
      tryCatch(source(rfilename, local = envir., chdir = TRUE),
               error = handle.source.error, warning = handle.source.error))
  }

  # First find out which file...
  filename_given <- filename != ""
  if (!filename_given) {
    filename <- path_rel2abs(params_def["scenarioFile","default"])
    if (file.exists(filename)) {
      irace_warning("A default scenario file ", shQuote(filename),
                    " has been found and will be read.")
    } else {
      irace_warning ("No scenario file given (use ",
                   params_def["scenarioFile", "short"], " or ",
                   params_def["scenarioFile", "long"],
                   ") and no default scenario file ", shQuote(filename),
                   " has been found.")
      scenario[["scenarioFile"]] <- filename
      return(scenario)
    }
  } else {
    filename <- path_rel2abs(filename)
  }

  if (file.exists (filename)) {
    debug.level <- getOption(".irace.debug.level", default = 0)
    if (debug.level >= 1)
      cat("# Reading scenario file", shQuote(filename), ".......")
    # chdir = TRUE to allow recursive sourcing.
    handle.source.error <- function(e) {
      irace_error("Reading scenario file ", shQuote(filename),
                  " produced the following errors or warnings:\n",
                  paste0(conditionMessage(e), collapse="\n"))
      return(NULL)
    }
    withCallingHandlers(
      tryCatch(source(filename, local = scenario_env, chdir = TRUE),
               error = handle.source.error, warning = handle.source.error))
    if (debug.level >= 1) cat (" done!\n")
  } else if (filename_given) {
    irace_error ("The scenario file ", shQuote(filename), " does not exist.")
  }

  ## Read scenario file variables.
  scenario[["scenarioFile"]] <- filename
  # If these are given and relative, they should be relative to the
  # scenario file (except logFile, which is relative to execDir).
  pathParams <- setdiff(params_def[params_def[, "type"] == "p",
                                          "name"], "logFile")
  params_names <- params_def[!startsWith(params_def[,"name"], "."), "name"]

  for (param in params_names) {
    if (exists (param, envir = scenario_env, inherits = FALSE)) {
      value <- get(param, envir = scenario_env, inherits = FALSE)
      if (filename_given && !is.null.or.empty(value) && is.character(value) && (param %in% pathParams)) {
        value <- path_rel2abs(value, cwd = dirname(filename))
      }
      scenario[[param]] <- value
    }
  }
  unknown_scenario_vars <- setdiff(ls(scenario_env), params_names)
  if (length(unknown_scenario_vars)) {
    # We only accept variables that match irace.params.names and if the user
    # wants to define their own, they should use names starting with ".", which
    # are ignored by ls()
    irace_error("Scenario file ", shQuote(filename), " contains unknown variables: ",
                paste0(unknown_scenario_vars, collapse=", "),
                "\nMAKE SURE NO VARIABLE NAME IS MISSPELL (for example, 'parameterFile' is correct, while 'parametersFile' is not)",
                "\nIf you wish to use your own variables in the scenario file, use names beginning with a dot `.'")
  }
  scenario
}

setup_test_instances <- function(scenario)
{
  testInstances <- scenario[["testInstances"]]
  if (is.null.or.empty(testInstances)) {
    if (!is.null.or.empty(scenario$testInstancesDir) ||
        !is.null.or.empty(scenario$testInstancesFile)) {
      scenario$testInstancesDir <- path_rel2abs(scenario$testInstancesDir)
      if (!is.null.or.empty(scenario$testInstancesFile)) {
        scenario$testInstancesFile <- path_rel2abs(scenario$testInstancesFile)
      }
      testInstances <- readInstances(instancesDir = scenario$testInstancesDir,
                                     instancesFile = scenario$testInstancesFile)
    } else {
      testInstances <- NULL
    }
  }
  if (!is.null(testInstances)) {
    if (!is.null(dim(testInstances))) {
      if (length(dim(testInstances)) == 1L ||
          (length(dim(testInstances)) == 2L && dim(testInstances)[1L] == 1L)) {
        # Remove useless dimensions
        testInstances <- c(testInstances)
      } else {
        irace_error("testInstances must be a one-dimensional vector or a list. If your instances are matrices or datasets in R, you can use scenario(testInstances=list(data1, data2, data3))")
      }
    }
    if (is.null(names(testInstances))) {
      # Create unique IDs for testInstances
      names(testInstances) <- paste0(seq_along(testInstances), "t")
    }
  }
  scenario[["testInstances"]] <- testInstances
  scenario
}

#' Check and correct the given scenario
#'
#' Checks for errors a (possibly incomplete) scenario setup of
#' \pkg{irace}  and transforms it into a valid scenario.
#'
#' @inheritParams defaultScenario
#'
#' @return The scenario received as a parameter, possibly corrected. Unset
#' scenario settings are set to their default values.
#'
#' @details   This function checks that the directories and the file names
#' provided and required by the \pkg{irace} exist. It also checks that the
#' settings are of the proper type, e.g. that settings expected to be integers
#' are really integers. Finally, it also checks that there is no inconsistency
#' between settings.  If an error is found that prevents \pkg{irace} from
#' running properly, it will stop with an error.
#'
#' @seealso
#'  \describe{
#'  \item{[readScenario()]}{for reading a configuration scenario from a file.}
#'  \item{[printScenario()]}{prints the given scenario.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#'  \item{[checkScenario()]}{to check that the scenario is valid.}
#' }
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
## FIXME: This function should only do checks and return TRUE/FALSE. There
## should be other function that does the various transformations.
checkScenario <- function(scenario = defaultScenario())
{
  quote.param <- function(name)
  {
    if (.irace.params.def[name, "long"] != "") {
      return(paste0("'", name, "' (", .irace.params.def[name, "long"], ")"))
    }
    paste0("'", name, "'")
  }

  as.boolean.param <- function(x, name)
  {
    if (is.logical(x)) return(x)
    if (is.na(x)) return(as.logical(NA))
    # We handle "0" and "1" but not "TRUE" and "FALSE".
    x <- suppressWarnings(as.integer(x))
    if (is.na(x) || (x != 0L && x != 1L)) {
      irace_error (quote.param(name), " must be either 0 or 1.")
    }
    as.logical(x)
  }

  check.valid.param <- function(x)
  {
    valid <- trimws(strsplit(.irace.params.def[x, "domain"],",",fixed=TRUE)[[1L]])
    if (scenario[[x]] %not_in% valid) {
      irace_error ("Invalid value '", scenario[[x]], "' of ",
                   quote.param(x), ", valid values are: ",
                   paste0(valid, collapse = ", "))
    }
  }

  # Fill possible unset (NULL) with default settings.
  scenario <- defaultScenario(scenario)

  # Duplicated entries will cause confusion.
  dups <- anyDuplicated(names(scenario))
  if (dups)
    irace_error("scenario contains duplicated entries: ", names(scenario)[dups])

  # We have characters everywhere, set to the right types to avoid problems
  # later.

  # Boolean control parameters.
  boolParams <- .irace.params.def[.irace.params.def[, "type"] == "b", "name"]
  for (p in boolParams) {
    scenario[[p]] <- as.boolean.param(scenario[[p]], p)
  }
  # Real [0, 1] control parameters.
  realParams <- .irace.params.def[.irace.params.def[, "type"] == "r", "name"]
  for (param in realParams) {
    if (is.null.or.empty(scenario[[param]]))
      scenario[[param]] <- NA_real_
    if (is.na(scenario[[param]]))
      next # Allow NA default values
    scenario[[param]] <- suppressWarnings(as.numeric(scenario[[param]]))
    if (is.null(scenario[[param]])
        || is.na (scenario[[param]])
        || scenario[[param]] < 0.0 || scenario[[param]] > 1.0)
      irace_error (quote.param(param), " must be a real value within [0, 1].")
  }

  # Integer control parameters
  intParams <- .irace.params.def[.irace.params.def[, "type"] == "i", "name"]
  for (param in intParams) {
    p <- scenario[[param]]
    if (is.null.or.empty(p))
      scenario[[param]] <- NA_integer_
    if (is.na(scenario[[param]]))
      next # Allow NA default values
    p <- suppressWarnings(as.numeric(p))
    if (is.null(p) || is.na (p) || !is.wholenumber(p) || p < 0)
      irace_error (quote.param (param), " must be a non-negative integer.")
    scenario[[param]] <- as.integer(p)
  }

  check_positive <- function(scenario, what) {
    if (any(scenario[what] <= 0L)) {
      for (op in what) {
        if (scenario[[op]] <= 0L)
          irace_error(quote.param (op), " = ", scenario[[op]], " must be larger than 0.")
      }
    }
  }
  check_positive(scenario, c("firstTest", "eachTest", "blockSize"))

  options(.irace.quiet = scenario$quiet)
  ## Check that everything is fine with external parameters
  # Check that the files exist and are readable.
  scenario$parameterFile <- path_rel2abs(scenario$parameterFile)
  if (is.null.or.empty(scenario$parameters)) {
    irace_note("Reading parameter file '", scenario$parameterFile, "'.\n")
    scenario$parameters <- readParameters(file = scenario$parameterFile,
      # AClib benchmarks use 15 digits
      digits = if (scenario$aclib) 15L else 4L, debugLevel = scenario$debugLevel)
  }
  scenario$parameters <- checkParameters(scenario$parameters)

  scenario$execDir <- path_rel2abs(scenario$execDir)
  file.check (scenario$execDir, isdir = TRUE,
              text = paste0("execution directory ", quote.param("execDir")))
  options(.irace.execdir = scenario$execDir)
  if (is.null.or.empty(scenario$logFile)) {
    # We cannot use NULL because defaultScenario() would override it.
    scenario$logFile <- ""
  } else {
    scenario$logFile <- path_rel2abs(scenario$logFile, cwd = scenario$execDir)
    file.check(scenario$logFile, writeable = TRUE, text = quote.param('logFile'))
  }

  if (is.null.or.empty(scenario$recoveryFile)) {
    # We cannot use NULL because defaultScenario() would override it.
    scenario$recoveryFile <- ""
  } else {
    scenario$recoveryFile <- path_rel2abs(scenario$recoveryFile)
    file.check(scenario$recoveryFile, readable = TRUE,
               text = paste0("recovery file ", quote.param("recoveryFile")))

    if (!is.null.or.empty(scenario$logFile) # Must have been set "" above.
        && scenario$recoveryFile == scenario$logFile) {
      irace_error("log file and recovery file should be different '",
                  scenario$logFile, "'")
    }
  }

  if (is.null.or.empty(scenario$targetRunnerParallel)) {
    scenario$targetRunnerParallel <- NULL
  } else if (is.function.name(scenario$targetRunnerParallel)) {
    scenario$targetRunnerParallel <- get.function(scenario$targetRunnerParallel)
  } else {
    irace_error("'targetRunnerParallel' must be a function")
  }

  if (is.null.or.empty(scenario$repairConfiguration))
    scenario$repairConfiguration <- NULL
  else if (is.function.name(scenario$repairConfiguration))    # Byte-compile it.
    scenario$repairConfiguration <- bytecompile(get.function(scenario$repairConfiguration))
  else
    irace_error("'repairConfiguration' must be a function")

  if (is.na(scenario$capping))
    scenario$capping <- (scenario$elitist && scenario$maxTime > 0 &&
                           !is.na(scenario$boundMax) && scenario$boundMax > 0)
  if (scenario$capping) {
    if (!scenario$elitist)
      irace_error("When capping == TRUE, elitist must be enabled.")
    if (scenario$boundMax <= 0)
      irace_error("When capping == TRUE, boundMax (", scenario$boundMax, ") must be > 0")
    check.valid.param("cappingType")
    check.valid.param("boundType")
    if (scenario$boundPar < 1)
      irace_error("Invalid value (", scenario$boundPar, ") ",
                  quote.param("boundPar"), " must be >= 1")
  } else if (scenario$boundMax <= 0 || is.na(scenario$boundMax)) { # no capping
    scenario$boundMax <- NULL
  }

  if (is.function.name(scenario$targetRunner)) {
    scenario$targetRunner <- get.function(scenario$targetRunner)
    irace_assert(is.function(scenario$targetRunner))
  } else if (is.null(scenario$targetRunnerParallel)) {
    if (is.character(scenario$targetRunner)) {
      scenario$targetRunner <- path_rel2abs(scenario$targetRunner)
      if (is.null.or.empty(scenario$targetRunnerLauncher)) {
        file.check (scenario$targetRunner, executable = TRUE,
                    text = paste0("target runner ", quote.param("targetRunner")))
      } else {
        scenario$targetRunnerLauncher <- path_rel2abs(scenario$targetRunnerLauncher)
        file.check (scenario$targetRunnerLauncher, executable = TRUE,
                    text = paste0("target runner launcher ", quote.param("targetRunnerLauncher")))
        file.check (scenario$targetRunner, readable = TRUE,
                    text = paste0("target runner ", quote.param("targetRunner")))
        if (!grepl("{targetRunner}", scenario$targetCmdline, fixed=TRUE)) {
          # If missing, we add it to the beginning for compatibility with irace 3.5.
          scenario$targetCmdline <- paste0("{targetRunner} ", scenario$targetCmdline)
        }
      }
      check_target_cmdline(scenario$targetCmdline, capping = scenario$capping)
    } else {
      irace_error(quote.param ('targetRunner'), " must be a function or an executable program")
    }
  }

  if (is.null.or.empty(scenario$targetEvaluator)) {
    scenario$targetEvaluator <- NULL
  } else if (is.function.name(scenario$targetEvaluator)) {
    scenario$targetEvaluator <- get.function(scenario$targetEvaluator)
    irace_assert(is.function(scenario$targetEvaluator))
  } else if (is.character(scenario$targetEvaluator)) {
    scenario$targetEvaluator <- path_rel2abs(scenario$targetEvaluator)
    file.check (scenario$targetEvaluator, executable = TRUE,
                text = "target evaluator")
  } else {
    irace_error(quote.param('targetEvaluator'), " must be a function or an executable program")
  }

  # Training instances
  if (is.null.or.empty(scenario$instances)) {
    scenario$trainInstancesDir <- path_rel2abs(scenario$trainInstancesDir)
    if (!is.null.or.empty(scenario$trainInstancesFile)) {
      scenario$trainInstancesFile <- path_rel2abs(scenario$trainInstancesFile)
    }
    if (is.null.or.empty(scenario$trainInstancesDir)
        && is.null.or.empty(scenario$trainInstancesFile))
      irace_error("Both ", quote.param ("trainInstancesDir"), " and ",
                  quote.param ("trainInstancesFile"),
                  " are empty: No instances provided")

    scenario$instances <-
      readInstances(instancesDir = scenario$trainInstancesDir,
                    instancesFile = scenario$trainInstancesFile)
  }
  if (length(scenario$instances) == 0L) {
    irace_error("No instances found in `scenario$instances`")
  } else if (!is.null(dim(scenario$instances))) {
    if (length(dim(scenario$instances)) == 1L ||
        (length(dim(scenario$instances)) == 2L && dim(scenario$instances)[1] == 1L)) {
      # Remove useless dimensions
      scenario$instances <- c(scenario$instances)
    } else {
      irace_error("Instances must be a one-dimensional vector or a list. If your instances are matrices or datasets in R, you can use scenario(instances=list(data1, data2, data3))")
    }
  }
  # Testing instances
  scenario <- setup_test_instances(scenario)

  # Configurations file
  if (!is.null.or.empty(scenario$configurationsFile)) {
    scenario$configurationsFile <- path_rel2abs(scenario$configurationsFile)
    file.check (scenario$configurationsFile, readable = TRUE,
                text = "configurations file")
    # We cannot read the configurations here because we need the parameters.
    # FIXME: We should have the parameters inside scenario.
  }

  if (is.null.or.empty(scenario$initConfigurations)) {
    scenario$initConfigurations <- NULL
  } else if (!is.data.frame(scenario$initConfigurations) && !is.matrix(scenario$initConfigurations)) {
    irace_error("if given, 'initConfigurations' must be a matrix or data.frame.")
  }

  if (length(scenario$instances) %% scenario$blockSize != 0) {
    irace_error("The number of instances (", length(scenario$instances), ") must be a multiple of ",
                quote.param("blockSize"), ".")
  }

  if (scenario$firstTest %% scenario$eachTest != 0) {
    irace_error(quote.param("firstTest"), " must be a multiple of ",
                quote.param("eachTest"), ".")
  }

  if (scenario$mu < scenario$firstTest * scenario$blockSize) {
    if (scenario$debugLevel >= 1) {
      irace_warning("Assuming 'mu = firstTest * blockSize' because 'mu' cannot be smaller.\n")
    }
    scenario$mu <- scenario$firstTest * scenario$blockSize
  } else if (scenario$mu %% scenario$blockSize != 0) {
    irace_error(quote.param("mu"), " must be a multiple of ",
                quote.param("blockSize"), ".")
  }

  if (!is.na(scenario$minExperiments))
    scenario$maxExperiments <- max(scenario$maxExperiments, scenario$minExperiments)

  ## Only maxExperiments or maxTime should be set. Negative values are not
  ## allowed.
  if (scenario$maxExperiments == 0 && scenario$maxTime == 0) {
    irace_error("Tuning budget was not provided. Set ",
                quote.param("maxExperiments"), " or ", quote.param("maxTime"), ".")
  } else if (scenario$maxExperiments > 0 && scenario$maxTime > 0) {
    irace_error("Two different tuning budgets provided, please set only ",
                quote.param("maxExperiments"), " or only ", quote.param ("maxTime"), ".")
  } else if (scenario$maxExperiments < 0 ) {
    irace_error("Negative budget provided, ", quote.param("maxExperiments"),
                "must be >= 0." )
  } else if (scenario$maxTime < 0) {
    irace_error("Negative budget provided, ", quote.param("maxTime"), " must be >= 0.")
  }

  if (scenario$maxTime > 0 && (scenario$budgetEstimation <= 0 || scenario$budgetEstimation >= 1))
    irace_error(quote.param("budgetEstimation"), " must be within (0,1).")

  if (scenario$maxTime > 0 && !scenario$elitist)
    irace_error(quote.param("maxTime"), " requires using 'elitist=1'")

  if (scenario$deterministic &&
      scenario$firstTest * scenario$blockSize > length(scenario$instances)) {
    irace_error("When deterministic == TRUE, the number of instances (",
                length(scenario$instances),
                ") cannot be smaller than firstTest (", scenario$firstTest, ") * blockSize (", scenario$blockSize, ")")
  }

  if (scenario$mpi && scenario$parallel < 2)
    irace_error (quote.param("parallel"), " must be larger than 1 when mpi is enabled.")

  if (is.null.or.empty(scenario$batchmode))
    scenario$batchmode <- 0
  if (scenario$batchmode != 0) {
    scenario$batchmode <- tolower(scenario$batchmode)
    check.valid.param("batchmode")
  }
  # Currently batchmode requires a targetEvaluator
  if (scenario$batchmode != 0 && is.null(scenario$targetEvaluator)) {
    irace_error(quote.param("batchmode"), " requires using ",
                quote.param("targetEvaluator"), ".")
  }

  if (scenario$batchmode != 0 && scenario$mpi) {
    irace_error(quote.param("mpi"), " and ", quote.param("batchmode"),
                " cannot be enabled at the same time.")
  }

  if (is_null_or_empty_or_na(scenario$testType)) {
    if (scenario$capping) scenario$testType <- "t-test"
    else scenario$testType <- "f-test"
  }

  scenario$testType <-
    switch(tolower(scenario$testType),
           "f-test" =, # Fall-through
           "friedman" = "friedman",
           "t-test" =, # Fall-through
           "t.none" = "t.none",
           "t-test-holm" =, # Fall-through,
           "t.holm" = "t.holm",
           "t-test-bonferroni" =, # Fall-through,
           "t.bonferroni" = "t.bonferroni",
           check.valid.param("testType"))
  scenario
}

#' Prints the given scenario
#'
#' @inheritParams defaultScenario
#'
#' @seealso
#'  \describe{
#'  \item{[readScenario()]}{for reading a configuration scenario from a file.}
#'  \item{[printScenario()]}{prints the given scenario.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#'  \item{[checkScenario()]}{to check that the scenario is valid.}
#' }
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
printScenario <- function(scenario)
{
  scenario_names <- .irace.params.names
  cat("## irace scenario:\n")
  for (name in scenario_names) {
    cat(name, " = ", deparse(scenario[[name]]), "\n", sep = "")
  }
  cat("## end of irace scenario\n")
}

#' Default scenario settings
#'
#' Return scenario object with default values.
#'
#' @param scenario `list()`\cr Data structure containing \pkg{irace}
#'   settings. The data structure has to be the one returned by the function
#'   [defaultScenario()] or [readScenario()].
#'
#' @param params_def `data.frame()`\cr Definition of the options accepted by
#'   the scenario. This should only be modified by packages that wish to extend
#'   \pkg{irace}.
#'
#'
#' @return A list indexed by the \pkg{irace} parameter names,
#' containing the default values for each parameter, except for those
#' already present in the scenario passed as argument.
#' The scenario list contains the following elements:
# __IRACE_OPTIONS__BEGIN__
#' \itemize{
#'  \item General options:
#'    \describe{
#'      \item{`scenarioFile`}{Path of the file that describes the configuration scenario setup and other irace settings. (Default: `"./scenario.txt"`)}
#'      \item{`execDir`}{Directory where the programs will be run. (Default: `"./"`)}
#'      \item{`logFile`}{File to save tuning results as an R dataset, either absolute path or relative to execDir. (Default: `"./irace.Rdata"`)}
#'      \item{`quiet`}{Reduce the output generated by irace to a minimum. (Default: `0`)}
#'      \item{`debugLevel`}{Debug level of the output of `irace`. Set this to 0 to silence all debug messages. Higher values provide more verbose debug messages. (Default: `0`)}
#'      \item{`seed`}{Seed of the random number generator (by default, generate a random seed). (Default: `NA`)}
#'      \item{`repairConfiguration`}{User-defined R function that takes a configuration generated by irace and repairs it. (Default: `""`)}
#'      \item{`postselection`}{Perform a postselection race after the execution of irace to consume all remaining budget. Value 0 disables the postselection race. (Default: `1`)}
#'      \item{`aclib`}{Enable/disable AClib mode. This option enables compatibility with GenericWrapper4AC as targetRunner script. (Default: `0`)}
#'    }
#'  \item Elitist `irace`:
#'    \describe{
#'      \item{`elitist`}{Enable/disable elitist irace. (Default: `1`)}
#'      \item{`elitistNewInstances`}{Number of instances added to the execution list before previous instances in elitist irace. (Default: `1`)}
#'      \item{`elitistLimit`}{In elitist irace, maximum number per race of elimination tests that do not eliminate a configuration. Use 0 for no limit. (Default: `2`)}
#'    }
#'  \item Internal `irace` options:
#'    \describe{
#'      \item{`sampleInstances`}{Randomly sample the training instances or use them in the order given. (Default: `1`)}
#'      \item{`softRestart`}{Enable/disable the soft restart strategy that avoids premature convergence of the probabilistic model. (Default: `1`)}
#'      \item{`softRestartThreshold`}{Soft restart threshold value for numerical parameters. (Default: `1e-04`)}
#'      \item{`nbIterations`}{Maximum number of iterations. (Default: `0`)}
#'      \item{`nbExperimentsPerIteration`}{Number of runs of the target algorithm per iteration. (Default: `0`)}
#'      \item{`minNbSurvival`}{Minimum number of configurations needed to continue the execution of each race (iteration). (Default: `0`)}
#'      \item{`nbConfigurations`}{Number of configurations to be sampled and evaluated at each iteration. (Default: `0`)}
#'      \item{`mu`}{Parameter used to define the number of configurations sampled and evaluated at each iteration. (Default: `5`)}
#'    }
#'  \item Target algorithm parameters:
#'    \describe{
#'      \item{`parameterFile`}{File that contains the description of the parameters of the target algorithm. (Default: `"./parameters.txt"`)}
#'      \item{`parameters`}{Parameters space object (usually read from a file using \code{readParameters}). (Default: `""`)}
#'    }
#'  \item Target algorithm execution:
#'    \describe{
#'      \item{`targetRunner`}{Executable called for each configuration that executes the target algorithm to be tuned. See the templates and examples provided. (Default: `"./target-runner"`)}
#'      \item{`targetRunnerLauncher`}{Executable that will be used to launch the target runner, when `targetRunner` cannot be executed directly (e.g., a Python script in Windows). (Default: `""`)}
#'      \item{`targetCmdline`}{Command-line arguments provided to `targetRunner` (or `targetRunnerLauncher` if defined). The substrings `\{configurationID\}`, `\{instanceID\}`,  `\{seed\}`,  `\{instance\}`, and `\{bound\}` will be replaced by their corresponding values. The substring `\{targetRunnerArgs\}` will be replaced by the concatenation of the switch and value of all active parameters of the particular configuration being evaluated.  The substring `\{targetRunner\}`, if present, will be replaced by the value of `targetRunner` (useful when using `targetRunnerLauncher`). (Default: `"{configurationID} {instanceID} {seed} {instance} {bound} {targetRunnerArgs}"`)}
#'      \item{`targetRunnerRetries`}{Number of times to retry a call to `targetRunner` if the call failed. (Default: `0`)}
#'      \item{`targetRunnerTimeout`}{Timeout in seconds of any `targetRunner` call (only applies to `target-runner` executables not to R functions), ignored if 0. (Default: `0`)}
#'      \item{`targetRunnerData`}{Optional data passed to `targetRunner`. This is ignored by the default `targetRunner` function, but it may be used by custom `targetRunner` functions to pass persistent data around. (Default: `""`)}
#'      \item{`targetRunnerParallel`}{Optional R function to provide custom parallelization of `targetRunner`. (Default: `""`)}
#'      \item{`targetEvaluator`}{Optional script or R function that provides a numeric value for each configuration. See templates/target-evaluator.tmpl (Default: `""`)}
#'      \item{`deterministic`}{If the target algorithm is deterministic, configurations will be evaluated only once per instance. (Default: `0`)}
#'      \item{`parallel`}{Number of calls to `targetRunner` to execute in parallel. Values `0` or `1` mean no parallelization. (Default: `0`)}
#'      \item{`loadBalancing`}{Enable/disable load-balancing when executing experiments in parallel. Load-balancing makes better use of computing resources, but increases communication overhead. If this overhead is large, disabling load-balancing may be faster. (Default: `1`)}
#'      \item{`mpi`}{Enable/disable MPI. Use `Rmpi` to execute `targetRunner` in parallel (parameter `parallel` is the number of slaves). (Default: `0`)}
#'      \item{`batchmode`}{Specify how irace waits for jobs to finish when `targetRunner` submits jobs to a batch cluster: sge, pbs, torque, slurm or htcondor. `targetRunner` must submit jobs to the cluster using, for example, `qsub`. (Default: `0`)}
#'    }
#'  \item Initial configurations:
#'    \describe{
#'      \item{`initConfigurations`}{Data frame describing initial configurations (usually read from a file using \code{readConfigurations}). (Default: `""`)}
#'      \item{`configurationsFile`}{File that contains a table of initial configurations. If empty or `NULL`, all initial configurations are randomly generated. (Default: `""`)}
#'    }
#'  \item Training instances:
#'    \describe{
#'      \item{`instances`}{Character vector of the instances to be used in the \code{targetRunner}. (Default: `""`)}
#'      \item{`trainInstancesDir`}{Directory where training instances are located; either absolute path or relative to current directory. If no `trainInstancesFiles` is provided, all the files in `trainInstancesDir` will be listed as instances. (Default: `""`)}
#'      \item{`trainInstancesFile`}{File that contains a list of training instances and optionally additional parameters for them. If `trainInstancesDir` is provided, `irace` will search for the files in this folder. (Default: `""`)}
#'      \item{`blockSize`}{Number of training instances, that make up a 'block' in `trainInstancesFile`. Elimination of configurations will only be performed after evaluating a complete block and never in the middle of a block. Each block typically contains one instance from each instance class (type or family) and the block size is the number of classes. The value of `blockSize` will multiply `firstTest`, `eachTest` and `elitistNewInstances`. (Default: `1`)}
#'    }
#'  \item Tuning budget:
#'    \describe{
#'      \item{`maxExperiments`}{Maximum number of runs (invocations of `targetRunner`) that will be performed. It determines the maximum budget of experiments for the tuning. (Default: `0`)}
#'      \item{`minExperiments`}{Minimum number of runs (invocations of `targetRunner`) that will be performed. It determines the minimum budget of experiments for the tuning. The actual budget depends on the number of parameters and `minSurvival`. (Default: `NA`)}
#'      \item{`maxTime`}{Maximum total execution time for the executions of `targetRunner`. `targetRunner` must return two values: cost and time. This value and the one returned by `targetRunner` must use the same units (seconds, minutes, iterations, evaluations, ...). (Default: `0`)}
#'      \item{`budgetEstimation`}{Fraction (smaller than 1) of the budget used to estimate the mean computation time of a configuration. Only used when `maxTime` > 0 (Default: `0.05`)}
#'      \item{`minMeasurableTime`}{Minimum time unit that is still (significantly) measureable. (Default: `0.01`)}
#'    }
#'  \item Statistical test:
#'    \describe{
#'      \item{`testType`}{Statistical test used for elimination. The default value selects `t-test` if `capping` is enabled or `F-test`, otherwise. Valid values are: F-test (Friedman test), t-test (pairwise t-tests with no correction), t-test-bonferroni (t-test with Bonferroni's correction for multiple comparisons), t-test-holm (t-test with Holm's correction for multiple comparisons). (Default: `""`)}
#'      \item{`firstTest`}{Number of instances evaluated before the first elimination test. It must be a multiple of `eachTest`. (Default: `5`)}
#'      \item{`eachTest`}{Number of instances evaluated between elimination tests. (Default: `1`)}
#'      \item{`confidence`}{Confidence level for the elimination test. (Default: `0.95`)}
#'    }
#'  \item Adaptive capping:
#'    \describe{
#'      \item{`capping`}{Enable the use of adaptive capping, a technique designed for minimizing the computation time of configurations. Capping is enabled by default if `elitist` is active, `maxTime > 0` and  `boundMax > 0`. (Default: `NA`)}
#'      \item{`cappingAfterFirstTest`}{If set to 1, elimination due to capping only happens after `firstTest` instances are seen. (Default: `0`)}
#'      \item{`cappingType`}{Measure used to obtain the execution bound from the performance of the elite configurations.\itemize{\item median: Median performance of the elite configurations.\item mean: Mean performance of the elite configurations.\item best: Best performance of the elite configurations.\item worst: Worst performance of the elite configurations.} (Default: `"median"`)}
#'      \item{`boundType`}{Method to calculate the mean performance of elite configurations.\itemize{\item candidate: Mean execution times across the executed instances and the current one.\item instance: Execution time of the current instance.} (Default: `"candidate"`)}
#'      \item{`boundMax`}{Maximum execution bound for `targetRunner`. It must be specified when capping is enabled. (Default: `0`)}
#'      \item{`boundDigits`}{Precision used for calculating the execution time. It must be specified when capping is enabled. (Default: `0`)}
#'      \item{`boundPar`}{Penalization constant for timed out executions (executions that reach `boundMax` execution time). (Default: `1`)}
#'      \item{`boundAsTimeout`}{Replace the configuration cost of bounded executions with `boundMax`. (Default: `1`)}
#'    }
#'  \item Recovery:
#'    \describe{
#'      \item{`recoveryFile`}{Previously saved log file to recover the execution of `irace`, either absolute path or relative to the current directory.  If empty or `NULL`, recovery is not performed. (Default: `""`)}
#'    }
#'  \item Testing:
#'    \describe{
#'      \item{`testInstancesDir`}{Directory where testing instances are located, either absolute or relative to current directory. (Default: `""`)}
#'      \item{`testInstancesFile`}{File containing a list of test instances and optionally additional parameters for them. (Default: `""`)}
#'      \item{`testInstances`}{Character vector of the instances to be used in the \code{targetRunner} when executing the testing. (Default: `""`)}
#'      \item{`testNbElites`}{Number of elite configurations returned by irace that will be tested if test instances are provided. (Default: `1`)}
#'      \item{`testIterationElites`}{Enable/disable testing the elite configurations found at each iteration. (Default: `0`)}
#'    }
#' }
# __IRACE_OPTIONS__END__
#'
#' @seealso
#'  \describe{
#'  \item{[readScenario()]}{for reading a configuration scenario from a file.}
#'  \item{[printScenario()]}{prints the given scenario.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#'  \item{[checkScenario()]}{to check that the scenario is valid.}
#' }
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
defaultScenario <- function(scenario = list(),
                            params_def = .irace.params.def)
{
  params_names <- params_def[!startsWith(params_def[,"name"], "."), "name"]
  if (is.null(names(scenario))) {
    scenario <- setNames(as.list(params_def[params_names,"default"]), params_names)
  } else if (!all(names(scenario) %in% params_names)) {
    irace_error("Unknown scenario settings: ",
                paste(names(scenario)[!(names(scenario) %in% params_names)],
                      collapse = ", "))
  } else {
    for (k in params_names) {
      if (is.null.or.na(scenario[[k]])) {
        scenario[[k]] <- params_def[k, "default"]
      }
    }
  }
  scenario
}

readInstances <- function(instancesDir = NULL, instancesFile = NULL)
{
  if (is.null.or.empty(instancesDir) && is.null.or.empty(instancesFile))
    irace_error("Both instancesDir and instancesFile are empty: No instances provided")

  instances <- NULL

  if (is.null.or.empty(instancesFile)) {
    file.check (instancesDir, isdir = TRUE, notempty = TRUE,
                text = "instances directory")
    # The files are sorted in alphabetical order, on the full path if
    # 'full.names = TRUE'.
    instances <- list.files (path = instancesDir, full.names = TRUE,
                             recursive = TRUE)
    if (length (instances) == 0)
      irace_error("No instances found in `", instancesDir, "'")
  } else {
    file.check (instancesFile, readable = TRUE, text = "instance file")
    # We do not warn if the last line does not finish with a newline.
    instances <- readLines (instancesFile, warn = FALSE)
    instances <- sub("#.*$", "", instances) # Remove comments
    instances <- sub("^[[:space:]]+", "", instances) # Remove leading whitespace
    instances <- instances[instances != ""] # Delete empty lines
    if (is.null.or.empty(instances))
      irace_error("No instances found in '", instancesFile,
                  "' (whitespace and comments starting with '#' are ignored)")
    if (!is.null.or.empty(instancesDir))
       instances <- paste0 (instancesDir, "/", instances)
  }
  instances
}

update_scenario <- function(scenario, ...)
{
  scenario_args <- list(...)
  if (length(scenario_args) == 0L)
    return(scenario)
  unknown_scenario_args <- setdiff(names(scenario_args), names(scenario))
  if (length(unknown_scenario_args))
    irace_error("Unknown scenario settings given: ", paste0(unknown_scenario_args, collapse=", "))
  utils::modifyList(scenario, scenario_args)
}

# R6 Class for parsing command-line arguments
CommandArgsParser <- R6::R6Class("CommandArgsParser", cloneable = FALSE, lock_class = TRUE, list(
  argv = NULL,
  argsdef = NULL,
  initialize = function(argv, argsdef) {
    # Handle the case where we are given a single character string like a
    # command-line.
    if (!missing(argv) && length(argv) == 1) {
      # strsplit does not respect quoted strings.
      argv <- scan(text=argv, what='character', quiet=TRUE)
    }
    self$argv <- argv
    required_colnames <- c("name", "short", "long", "type", "default")
    if (any(required_colnames %not_in% colnames(argsdef))) {
      stop("argsdef must contain the column names: ", paste0(required_colnames, collapse=", "))
    }
    self$argsdef <- argsdef
    rownames(self$argsdef) <- argsdef$name
    self
  },
  readCmdLineParameter = function (paramName, default = NULL) {
    short <- self$argsdef[paramName, "short"]
    long <- self$argsdef[paramName,"long"]
    value <- self$readArg(short = short, long = long)
    if (is.null(value)) {
      value <- if (is.null(default)) self$argsdef[paramName, "default"] else default
    } else if (is.na(value) && self$argsdef[paramName,"type"] != 'x') {
      stop("option '", long, "' requires an argument", call. = FALSE)
    }
    return(value)
  },
  # Function to read command-line arguments.
  ## FIXME: This function always consumes two arguments. This is problematic
  ## for flags that have no arguments, like --check.
  readArg = function(short = "", long = "") {
    if (length(short) == 0) short <- ""
    if (length(long) == 0) long <- ""
    if (short == "" && long == "") return(NULL)
    argv <- self$argv
    pos <- c()
    pattern <- ""
    if (short != "") {
      pattern_equal <- paste0("^", short, "=")
      pattern <- paste0("^", short, "$|", pattern_equal)
    }
    if (long != "") {
      pattern_long_equal <- paste0("^", long, "=")
      pattern_long <- paste0("^", long, "$|", pattern_long_equal)
      if (short != "") {
        pattern <-  paste0(pattern, "|", pattern_long)
        pattern_equal <- paste0(pattern_equal, "|", pattern_long_equal)
      } else {
        pattern <- pattern_long
        pattern_equal <- pattern_long_equal
      }
    }
    pos <- grep(pattern, argv)
    if (length(pos) == 0) {
      return(NULL) # Not found
    } else if (length(pos) > 0) {
      # Allow repeated parameters
      pos <- max(pos)
    }
    if (grepl(pattern_equal, argv[pos])) {
      value <- unlist(strsplit(argv[pos], '=', fixed = TRUE))[2]
      if (is.null (value) || is.na(value))
        value <- ""
    } else {
      value <- argv[pos + 1]
      self$argv <- argv[-(pos + 1)]
    }
    self$argv <- self$argv[-pos]
    return (value)
  },
  readAll = function() {
    params <- list()
    for (param in self$argsdef$name[self$argsdef$type != 'x']) {
      value <- self$readCmdLineParameter(paramName = param)
      if (is.na(value) || (length(value) > 0 && value == "")) value <- NULL
      params[[param]] <- value
    }
    params
  },
  cmdline_usage = function() {
    cmdline_usage(self$argsdef)
  })
)

# `cmdline_usage()` prints the output of `--help`
#
# @param cmdline_args Definition of the command-line arguments.
# 
# @export
cmdline_usage <- function(cmdline_args)
{
  for (i in seq_len(nrow(cmdline_args))) {
    short <- cmdline_args[i,"short"]
    long <- cmdline_args[i,"long"]
    desc <- cmdline_args[i,"description"]
    if (desc == "" || (short == "" && long == "")) next
    if (short != "") short <- paste0(short,",")
    default <- cmdline_args[i,"default"]
    if (!is_null_or_empty_or_na(default)) {
      desc <- paste0(desc, " Default: ", default, ".")
    }
    cat(sep = "\n", strwrap(desc, width = 80,
                            initial = sprintf("%3s%-20s  ", short, long),
                            exdent = 25))
  }
}

## ----include=FALSE------------------------------------------------------------
library(knitr)

## ----exampleload,eval=TRUE,include=FALSE----------------------------
library("irace")
load("examples.Rdata")
iraceResults <- irace::read_logfile("irace-acotsp.Rdata")
log_ablation_file <- file.path(system.file(package="irace"), "exdata", "log-ablation.Rdata")
load(log_ablation_file)
options(width = 70)

## ----R_irace_install, prompt=FALSE, eval=FALSE----------------------
#  install.packages("irace")

## ----R_irace_launch,eval=FALSE, prompt=FALSE------------------------
#  library("irace")
#  q() # To exit R

## ----install_win1,eval=FALSE, prompt=FALSE--------------------------
#  install.packages("<package>", repos = NULL)

## ----install_win2,eval=FALSE, prompt=FALSE--------------------------
#  # Replace <package> with the path to the downloaded file.
#  # Replace <R_LIBS_USER> with the path used for installation.
#  install.packages("<package>", repos = NULL, lib = "<R_LIBS_USER>")
#  # Tell R where to find R_LIBS_USER.
#  # This must be executed for every new session.
#  .libPaths(c("<R_LIBS_USER>", .libPaths()))

## ----R_irace_test1, prompt=FALSE, eval=FALSE------------------------
#  # Load the package
#  library("irace")
#  # Obtain the installation path
#  system.file(package = "irace")

## ----windows_irace_help ,eval=FALSE, prompt=FALSE-------------------
#  library("irace")
#  irace.cmdline("--help")

## ----irace_R_check, eval=FALSE, prompt=FALSE------------------------
#  library("irace")
#  scenario <- readScenario(filename = "scenario.txt",
#                           scenario = defaultScenario())
#  checkIraceScenario(scenario = scenario)

## ----irace_R_exe, eval=FALSE, prompt=FALSE--------------------------
#  library("irace")
#  # Go to the directory containing the scenario files
#  setwd("~/tuning")
#  scenario <- readScenario(filename = "scenario.txt",
#                           scenario = defaultScenario())
#  irace.main(scenario = scenario)

## ----runexample2,prompt=FALSE,eval=FALSE----------------------------
#  library("irace")
#  setwd("~/tuning/")
#  irace.cmdline()

## ----readParameters,prompt=FALSE, eval=FALSE------------------------
#  parameters <- readParameters(file = "parameters.txt")

## ----parametersetup,eval=TRUE,include=FALSE-------------------------
# Setup example
parameters <- irace::readParameters(text='
algorithm       "--"                 c		(as,mmas,eas,ras,acs)
ants            "--ants "            i	  	(5, 100)
q0              "--q0 "              r  	(0.0, 1.0) 		| algorithm %in% c("acs")
')

## ----parameterlist,eval=TRUE, prompt=TRUE, size='normalsize', comment=""----
str(parameters, vec.len = 10)

## ----targetRunner,prompt=FALSE, eval=FALSE--------------------------
#  targetRunner(experiment, scenario)

## ----experimentlist,eval=TRUE,size='normalsize', prompt=TRUE,  comment=""----
print(experiment)

## ----targetEvaluator, prompt=FALSE, eval=FALSE----------------------
#  targetEvaluator(experiment, num.configurations, all.conf.id,
#                  scenario, target.runner.call)

## ----instance1, prompt=FALSE, eval=FALSE----------------------------
#  scenario$instances <- c("rosenbrock_20 --function=12 --nvar 20",
#                          "rosenbrock_40 --function=12 --nvar 30",
#                          "rastrigin_20 --function=15 --nvar 20",
#                          "rastrigin_40 --function=15 --nvar 30")

## ----repairEx,prompt=FALSE, eval=FALSE------------------------------
#  repairConfiguration = function (configuration, parameters, digits)
#  {
#    isreal <- parameters$type[colnames(configuration)] %in% "r"
#    configuration[isreal] <- configuration[isreal] / sum(configuration[isreal])
#    configurations[isreal] <- round(configuration[isreal], digits)
#    return(configuration)
#  }

## ----repairEx2,prompt=FALSE, eval=FALSE-----------------------------
#  repairConfiguration = function (configuration, parameters, digits)
#  {
#   columns <- c("p1","p2","p3")
#   # cat("Before"); print(configuration)
#   configuration[columns] <- sort(configuration[columns])
#   # cat("After"); print(configuration)
#   return(configuration)
#  }

## ----targetRunnerParallel,prompt=FALSE, eval=FALSE------------------
#  targetRunnerParallel(experiments, exec.target.runner, scenario, target.runner)

## ----targetRunnerParallel2,prompt=FALSE, eval=FALSE-----------------
#  targetRunnerParallel <- function(experiments, exec.target.runner, scenario)
#  {
#    return (lapply(experiments, exec.target.runner, scenario = scenario,
#                   target.runner = target.runner))
#  }

## ----targetRunnerParallel3,prompt=FALSE, eval=TRUE------------------
   print(output)

## ----testing_r, prompt=FALSE, eval=FALSE----------------------------
#  testing_fromlog(logFile = "./irace.Rdata", testNbElites = 1)

## ----change_recover, prompt=TRUE, eval=FALSE------------------------
#  load ("~/tuning/irace.Rdata")
#  new.path <- "~/experiments/tuning/instances/"
#  iraceResults$scenario$instances <-
#     paste (new.path,
#           basename(iraceResults$scenario$instances),
#           sep="")
#  save (iraceResults, file="~/tuning/irace.Rdata")

## ----load_rdata, prompt=FALSE, eval=FALSE---------------------------
#  load("irace-acotsp.Rdata")

## ----show_version, prompt=TRUE, eval=TRUE, comment=""---------------
iraceResults$irace.version
iraceResults[["irace.version"]]

## ----show_configurations, prompt=TRUE, eval=TRUE, comment=""--------
head(iraceResults$allConfigurations)

## ----show_idelites, prompt=TRUE, eval=TRUE, comment=""--------------
print(iraceResults$allElites)

## ----get_elites, prompt=TRUE, eval=TRUE, comment=""-----------------
getFinalElites("irace-acotsp.Rdata", n = 0)

## ----show_iditelites, prompt=TRUE, eval=TRUE, comment=""------------
print(iraceResults$iterationElites)

## ----get_elite, prompt=TRUE, eval=TRUE, comment=""------------------
last <- length(iraceResults$iterationElites)
id <- iraceResults$iterationElites[last]
getConfigurationById(iraceResults, ids = id)

## ----get_experiments, prompt=TRUE, eval=TRUE, comment=""------------
# As an example, we use the best configuration found
best.config <- getFinalElites(iraceResults = iraceResults, n = 1)
id <- best.config$.ID.
# Obtain the configurations using the identifier
# of the best configuration
all.exp <- iraceResults$experiments[,as.character(id)]
all.exp[!is.na(all.exp)]

## ----get_instance_seed, prompt=TRUE, eval=TRUE, comment=""----------
# As an example, we get seed and instance of the experiments
# of the best candidate.
# Get index of the instances
pair.id <- which(!is.na(all.exp))
index <- iraceResults$state$.irace$instancesList[pair.id, "instance"]
# Obtain the instance names
iraceResults$scenario$instances[index]
# Get the seeds
iraceResults$state$.irace$instancesList[pair.id, "seed"]

## ----get_model, prompt=TRUE, eval=TRUE, comment=""------------------
# As an example, we get the model probabilities for the
# localsearch parameter.
iraceResults$state$model["localsearch"]
# The order of the probabilities corresponds to:
iraceResults$parameters$domain$localsearch

## ----get_test_exp, prompt=TRUE, eval=TRUE, comment=""---------------
# Get the results of the testing
iraceResults$testing$experiments

## ----get_test_seeds, prompt=TRUE, eval=TRUE, comment=""-------------
# Get the seeds used for testing
iraceResults$testing$seeds

## ----wilcox_test,prompt=TRUE, eval=TRUE, comment=""-----------------
results <- iraceResults$testing$experiments
# Wilcoxon paired test
conf <- gl(ncol(results), # number of configurations
           nrow(results), # number of instances
           labels = colnames(results))
pairwise.wilcox.test (as.vector(results), conf, paired = TRUE, p.adj = "bonf")

## ----conc, prompt=TRUE, eval=TRUE, comment=""-----------------------
irace:::concordance(iraceResults$testing$experiments)

## ----testEvo, fig.pos="tb", fig.align="center", out.width='0.7\\textwidth', fig.cap="Testing set performance of the best-so-far configuration over number of experiments. Label of each point is the configuration ID.", prompt=FALSE, eval=TRUE, comment=""----
# Get number of iterations
iters <- unique(iraceResults$experimentLog[, "iteration"])
# Get number of experiments (runs of target-runner) up to each iteration
fes <- cumsum(table(iraceResults$experimentLog[,"iteration"]))
# Get the mean value of all experiments executed up to each iteration
# for the best configuration of that iteration.
elites <- as.character(iraceResults$iterationElites)
values <- colMeans(iraceResults$testing$experiments[, elites])
stderr <- function(x) sqrt(var(x)/length(x))
err <- apply(iraceResults$testing$experiments[, elites], 2, stderr)
plot(fes, values, type = "s",
     xlab = "Number of runs of the target algorithm",
     ylab = "Mean value over testing set", ylim=c(23000000,23500000))
points(fes, values, pch=19)
arrows(fes, values - err, fes, values + err, length=0.05, angle=90, code=3)
text(fes, values, elites, pos = 1)

## ----ablation, prompt=FALSE, eval=FALSE-----------------------------
#  ablog <- ablation("irace.Rdata", src = 1, target = 60)
#  plotAblation(ablog)

## ----testAb, fig.pos="htb!", fig.align="center", out.width="0.75\\textwidth", fig.cap="Example of plot generated by \\code{plotAblation()}.", prompt=FALSE, eval=TRUE, echo=FALSE----
logfile <- file.path(system.file(package="irace"), "exdata", "log-ablation.Rdata")
plotAblation(logfile)

## ----ablation_cmdline, prompt=FALSE, eval=TRUE,echo=FALSE-----------
ablation_cmdline("--help")

## ----postsel, prompt=FALSE, eval=FALSE------------------------------
#  # Execute all elite configurations in the iterations
#  psRace(iraceLogFile="irace.Rdata", elites=TRUE)
#  # Execute a set of configurations IDs providing budget
#  psRace(iraceLogFile="irace.Rdata",
#               conf.ids=c(34, 87, 102, 172, 293),
#               max.experiments=500)

## ----faq3, eval=FALSE-----------------------------------------------
#  library(Rmpi)
#  mpi.spawn.Rslaves(nslaves = 10)
#  paths <- mpi.applyLB(1:10, function(x) {
#    library(irace); return(path.package("irace")) })
#  print(paths)

## ----R_irace_home2, prompt=FALSE, eval=FALSE------------------------
#  system.file(package = "irace")

#' Read parameter configurations from a file
#'
#' Reads a set of target-algorithm configurations from a file and puts them in
#' \pkg{irace} format. The configurations are checked to match the parameters
#' description provided.
#'
#' @param filename `character(1)`\cr Filename from which the configurations should be read.
#' The contents should be readable by `read.table( , header=TRUE)`.
#' @param text `character(1)`\cr If \code{file} is not supplied and this is,
#'  then configurations are read from the value of \code{text} via a text connection.
#' @inheritParams readParameters
#' @inheritParams printParameters
#'
#' @return A data frame containing the obtained configurations.
#'   Each row of the data frame is a candidate configuration,
#'   the columns correspond to the parameter names in `parameters`.
#'
#' @details
#' Example of an input file:
#' ```
#' # This is a comment line
#' param_1    param_2
#'     0.5  "value_1"
#'     1.0         NA
#'     1.2  "value_3"
#' ```
#' The order of the columns does not necessarily have to be the same
#' as in the file containing the definition of the parameters.
#'
#' @seealso
#'   [readParameters()] to obtain a valid parameter structure from a parameters file.
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
readConfigurationsFile <- function(filename, parameters, debugLevel = 0L, text)
{
  if (missing(filename) && !missing(text)) {
    filename <- NULL
    configurationTable <- utils::read.table(text = text, header = TRUE,
                                     na.strings = c("NA", "<NA>"),
                                     colClasses = "character",
                                     stringsAsFactors = FALSE)
  } else {
    # Read the file.
    configurationTable <- utils::read.table(filename, header = TRUE,
                                     na.strings = c("NA", "<NA>"),
                                     colClasses = "character",
                                     stringsAsFactors = FALSE)
  }
  irace_assert(is.data.frame(configurationTable))
  irace_note("Read ", nrow(configurationTable), " configuration(s)",
             if (is.null(filename)) "\n" else paste0(" from file '", filename, "'\n"))
  fix_configurations(configurationTable, parameters, debugLevel = debugLevel,
                     filename = filename)
}

fix_configurations <- function(configurations, parameters, debugLevel = 0L, filename = NULL)
{
  conf_error <- function(k, ...)
    irace_error("Configuration number ", k,
      if (is.null(filename)) "" else paste0(" from file '", filename, "'"),
      ...)

  if (debugLevel >= 2L)
    print(configurations, digits=15L)
  nbConfigurations <- nrow(configurations)
  namesParameters <- parameters[["names"]]
  # This ignores fixed parameters unless they are given with a different value.
  if (ncol(configurations) != length(namesParameters)
    || !setequal(colnames(configurations), namesParameters)) {
    # Column names must match a parameter, including fixed ones.
    missing <- setdiff(colnames(configurations), namesParameters)
    if (length(missing) > 0L) {
      if (is.null(filename)) {
        irace_error("The parameter names (", strlimit(paste(missing, collapse=", ")),
          ") do not match the parameter names: ", paste(namesParameters, collapse=", "))
      } else {
        irace_error("The parameter names (",
          strlimit(paste(missing, collapse=", ")),
          ") given in the first row of file ", filename,
          " do not match the parameter names: ",
          paste(namesParameters, collapse=", "))
      }
      return(NULL)
    }
    # All non-fixed parameters must appear in column names.
    varParameters <- parameters$names_variable
    missing <- setdiff (varParameters, colnames(configurations))
    if (length(missing) > 0) {
      if (is.null(filename)) {
        irace_error("The parameter names (",
          strlimit(paste(missing, collapse=", ")),
          ") are missing from the configurations provided.")
      } else {
        irace_error("The parameter names (",
          strlimit(paste(missing, collapse=", ")),
          ") are missing from the first row of file ", filename)
      }
      return(NULL)
    }
    # Add any missing fixed parameters.
    missing <- setdiff (namesParameters, colnames(configurations))
    if (length(missing) > 0L) {
      irace_assert (all(parameters$isFixed[missing]))
      configurations <- cbind.data.frame(configurations, parameters$domains[missing],
        stringsAsFactors = FALSE)
    }
  }
  # Reorder columns.
  configurations <- configurations[, namesParameters, drop = FALSE]

  # Loop over all parameters.
  for (param in parameters$get()) {
    pname <- param[["name"]]
    type <- param[["type"]]
    domain <- param[["domain"]]
    is_dep_param <- param[["is_dependent"]]
    condition <- param[["condition"]]
    # Fix up numeric columns.
    if (type == "i") {
      # For integers, only accept an integer.
      configurations[[pname]] <-
        suppressWarnings(as.numeric(configurations[[pname]]))
      # Remove NAs for this check.
      values <- configurations[[pname]]
      values[is.na(values)] <- 0L
      if (any(as.integer(values) != values)) {
        k <- which(as.integer(values) != values)[1L]
        conf_error (k, " is invalid because parameter ", pname,
          " is of type integer but its value ", values[k], " is not an integer")
        return(NULL)
      }
    } else if (type == "r") {
      configurations[[pname]] <- round(
        suppressWarnings(as.numeric(configurations[[pname]])),
        digits = parameters$get(pname)[["digits"]])
    }
    # Loop over all configurations.
    # FIXME: Vectorize this loop
    values <- configurations[[pname]]
    for (k in seq_len(nbConfigurations)) {
      currentValue <- values[k]
      # Check the status of the conditions for this parameter to know whether
      # it must be enabled.
      if (conditionsSatisfied(condition, configurations[k, ])) {
        # Check that the value is among the valid ones.
        if (is_dep_param) {
          dep_domain <- getDependentBound(param, configurations[k, ])
          if (is.na(dep_domain[1L])) {
            # Dependencies are not satisfied, so skip
            if (is.na(currentValue)) next
            conf_error (k, " is invalid because parameter \"", pname,
              "\" is not enabled, because its domain ",
              sub("expression", "", deparse(domain)),
              " depends on parameters that are not enabled, but its value is \"",
              currentValue, "\" instead of NA")
            return(NULL)
          }
          lower <- dep_domain[1L]
          upper <- dep_domain[2L]
          if (is.na(currentValue) || currentValue < lower || currentValue > upper) {
            conf_error (k, " is invalid because the value \"",
                        configurations[k, pname], "\" for the parameter ",
                        pname, " is not within the valid range ",
                        sub("expression", "", deparse(domain)),
                        ", that is, [", lower,", ", upper,"]")
            return(NULL)
          }
        } else if (type == "i" || type == "r") {
          lower <- domain[[1L]]
          upper <- domain[[2L]]
          if (is.na(currentValue) || currentValue < lower || currentValue > upper) {
            conf_error (k, " is invalid because the value \"",
                        configurations[k, pname], "\" for the parameter ",
                        pname, " is not within the valid range [",
                        lower,", ", upper,"]")
            return(NULL)
          }
          # type == "o" or "c"
        } else if (currentValue %not_in% domain) {
          conf_error (k, " is invalid because the value \"",
                      currentValue, "\" for the parameter \"",
                      pname, "\" is not among the valid values: (\"",
                      paste0(domain, collapse="\", \""), "\")")
          return(NULL)
        }
      } else if (!is.na(currentValue)) {
        conf_error (k, " is invalid because parameter \"", pname,
                    "\" is not enabled because of condition \"",
                    param[["condition"]], "\" but its value is \"",
                    currentValue, "\" instead of NA")
        return(NULL)
      }
    }
  }
  if (anyDuplicated(configurations)) {
    irace_error("Duplicated configurations",
                if (is.null(filename)) "" else paste0(" in file '", filename, "'"),
                ":\n",
                paste0(utils::capture.output(
                  configurations[duplicated(configurations), , drop=FALSE]), "\n"))
  }
  configurations
}

compile_forbidden <- function(x)
{
  if (is.null(x) || is.bytecode(x)) return(x)
  # If we are given an expression, it must be a single one.
  irace_assert(is.language(x) && (!is.expression(x) || length(x) == 1L))
  if (is.expression(x)) x <- x[[1L]]
  # When a is NA and we check a == 5, we would get NA, which is
  # always FALSE, when we actually want to be TRUE, so we test
  # is.na() first below.

  # We expect that there will be undefined variables, since the expressions
  # will be evaluated within a data.frame later.
  expr <- compiler::compile(substitute(is.na(x) | !(x), list(x = x)),
                           options = list(suppressUndefined=TRUE))
  attr(expr, "source") <- as.character(as.expression(x))
  expr
}

buildForbiddenExp <- function(configurations)
{
  if (is.null(configurations) || nrow(configurations) == 0L)
    return(NULL)
  pnames <- colnames(configurations)
  lines <- c()
  # We cannot use apply() because it converts numeric to character.
  for (k in seq_nrow(configurations)) {
    values <- as.list(configurations[k, ])
    has.value <- !is.na(values)
    values <- lapply(values[has.value], function(x) deparse(substitute(x, list(x=x))))
    lines <- c(lines,
               paste0("(", pnames[has.value]," == ", values, ")", collapse = "&"))
  }
  exps <- parse(text = lines)
  lapply(exps, compile_forbidden)
}

RaceState <- R6Class("RaceState", lock_class = TRUE,
 public = list(
   # This may be dynamically adjusted
   cluster = NULL,
   completed = "Incomplete",
   elapsed = 0L,
   elapsed_recovered = 0L,
   elitist_new_instances = 0L,
   experiment_log = NULL,
   instances_log = NULL,
   minSurvival = NULL,
   next_instance = -1L,
   race_experiment_log = NULL,
   recovery_info = NULL,
   recovery_mode = FALSE,
   rejected_ids = NULL,
   rng = NULL,
   seed = NULL,
   session_info = NULL,
   target_evaluator = NULL,
   target_runner = NULL,
   timeUsed = 0,
   time_next_save = 0,
   timer = NULL,
   # Methods.
   initialize = function(scenario, new = TRUE, recover = FALSE) {
     self$timer <- Timer$new()
     self$target_runner <- if (is.function(scenario$targetRunner))
                             bytecompile(scenario$targetRunner)
                           else if (scenario$aclib)
                             target_runner_aclib
                           else target_runner_default

     if (!is.null(scenario$targetEvaluator)) {
       self$target_evaluator <- if (is.function(scenario$targetEvaluator))
                                  bytecompile(scenario$targetEvaluator)
                                else
                                  target_evaluator_default
     }
     irace_assert(new || !recover)
     if (new) {
       # elitist_new_instances must be a multiple of blockSize.
       self$elitist_new_instances <- scenario$elitistNewInstances * scenario$blockSize
       # We cannot recover if we did not get to initialize self$rng.
       if (recover && !is.null(self$rng)) {
         restore_random_seed(self$rng)
         self$recovery_mode <- TRUE
         set(self$experiment_log, j = "iteration", value = NULL)
         self$recovery_info <- rbindlist(c(list(self$experiment_log), self$race_experiment_log), use.names = TRUE)
         # Reinitialize some state.
         self$completed = "Incomplete"
         self$elapsed = 0L
         self$elapsed_recovered = 0L
         self$experiment_log = NULL
         self$next_instance = -1L
         self$race_experiment_log = NULL
         self$rejected_ids = NULL
         self$timeUsed = 0
         self$time_next_save = 0
         # Just in case anything is still running.
         self$stop_parallel()
       } else {
         seed <- scenario$seed
         if (is.na(seed))
           seed <- trunc(runif(1, 1, .Machine$integer.max))
         set_random_seed(seed)
         self$seed <- seed
         self$rng <- get_random_seed()
       }
     } else { # !new
       self$elapsed_recovered <- self$elapsed
       restore_random_seed(self$rng)
     }

     if (is.null(self$experiment_log)) {
       self$experiment_log <- data.table(iteration=integer(0), instance=integer(0),
         configuration=integer(0), cost = numeric(0), time = numeric(0),
         bound = if (is.null(scenario$boundMax)) NULL else numeric(0))
     }

     if (scenario$debugLevel >= 3L) {
       irace_note("RNGkind: ", paste0(self$rng$rng_kind, collapse = " "), "\n",
                  "# .Random.seed: ", paste0(self$rng$random_seed, collapse = ", "), "\n")
     }
     # We do this here, so it is available even if we crash.
     self$session_info <- utils::sessionInfo()
     invisible(self)
   },

   update_experiment_log = function(output, instances, scenario) {
     # FIXME: The instances parameter is not needed.
     irace_assert(all.equal(rep(instances, each = length(unique(output[["configuration"]]))), output$instance))
     # Extract results
     self$experiment_log <- rbindlist(list(self$experiment_log, output), use.names=TRUE)
     experiments_output_to_matrix(output, scenario)
   },

   save_recovery = function(iraceResults, logfile) {
     now <- self$timer$wallclock()
     # Do not save to disk too frequently.
     if (now >= self$time_next_save) {
       # irace_note("Saving recovery info.\n")
       iraceResults$state <- self
       save_irace_logfile(iraceResults, logfile)
       self$time_next_save <- now + .irace_minimum_saving_time
     }
   },

   update_race_experiment_log = function(experiment_log, scenario) {
     self$race_experiment_log <- c(self$race_experiment_log, list(experiment_log))
     now <- self$timer$wallclock()
     # Do not save to disk too frequently.
     if (now >= self$time_next_save) {
       # irace_note("Saving recovery info.\n")
       iraceResults <- list(
         scenario = scenario,
         irace_version = irace_version,
         state = self)
       save_irace_logfile(iraceResults, logfile = scenario$logFile)
       self$time_next_save <- now + .irace_minimum_saving_time
     }
     invisible()
   },

   reset_race_experiment_log = function() {
     res <- rbindlist(self$race_experiment_log, use.names=TRUE)
     self$race_experiment_log <- NULL
     res
   },

   recover_output = function(instance_idx, configuration_id) {
     search <- data.table(instance = instance_idx, configuration = configuration_id)
     res <- self$recovery_info[search, on = .(instance,configuration), mult="first", nomatch=NULL, which=TRUE]
     irace_assert(length(res) == 0L || length(res) == nrow(search))
     if (length(res) == 0L) {
       irace_note("Cannot find the following in recovery info:")
       print(search[!self$recovery_info, on = .(instance,configuration)])
       irace_error("Recovery terminated.")
     }
     # Get the rows.
     output <- self$recovery_info[res]
     # Delete those rows.
     self$recovery_info <- self$recovery_info[-res]
     if (nrow(self$recovery_info) == 0L) {
       irace_note("Recovery completed.\n")
       self$recovery_mode <- FALSE
       self$recovery_info <- NULL
     }
     output
   },

   update_rejected = function(rejected_ids, configurations) {
     if (length(rejected_ids) == 0L) return(NULL)
     self$rejected_ids <- c(self$rejected_ids, rejected_ids)
     configurations[configurations[[".ID."]] %in% rejected_ids, , drop = FALSE]
   },

   time_elapsed = function() {
     self$elapsed <- self$timer$elapsed() + self$elapsed_recovered
     self$elapsed
   },

   start_parallel = function(scenario) {
     parallel <- scenario$parallel
     data.table::setDTthreads(if (parallel <= 1L) 1L else min(4L, parallel))
     if (!is.null(scenario$targetRunnerParallel) || parallel <= 1L)
       return(invisible(self))

     # Starting the parallel environment may set some logs to the current
     # directory, so switch to execDir momentarily.
     withr::local_dir(scenario$execDir) # setwd()
     if (scenario$mpi) {
       mpiInit(parallel, scenario$debugLevel)
     } else {
       requireNamespace("parallel", quietly = TRUE)
       if (.Platform$OS.type == 'windows' && is.null(self$cluster)) {
         # FIXME: makeCluster does not print the output generated by the workers
         # on Windows. We need to use the future package for that:
         # https://stackoverflow.com/questions/56501937/how-to-print-from-clusterapply
         self$cluster <- parallel::makeCluster(parallel)
         if (scenario$debugLevel >= 1L)
           irace_note("makeCluster initialized for ", parallel, " jobs.\n")
         # We export the global environment because the user may have defined
         # stuff there. There must be a better way to do this, but I cannot
         # figure it out. R sucks sometimes.
         parallel::clusterExport(self$cluster, ls(envir=.GlobalEnv))
         # In Windows, this needs to be exported, or we get:
         ## Error in checkForRemoteErrors(val) :
         ##  2 nodes produced errors; first error: could not find function "target_runner"
         parallel::clusterExport(self$cluster, list("target_runner"), envir=self)
         if (is.function(scenario$targetRunner)
           && !identical(environment(scenario$targetRunner), globalenv())) {
           env_target_runner <- environment(scenario$targetRunner)
           funglobs <- codetools::findGlobals(self$target_runner, merge=TRUE)
           common <- intersect(funglobs, ls(envir=env_target_runner))
           if (length(common))
             parallel::clusterExport(self$cluster, common, envir=env_target_runner)
         }
       }
     }
     invisible(self)
   },

   stop_parallel = function() {
     if (!is.null(self$cluster)) {
       try(parallel::stopCluster(self$cluster), silent=TRUE)
       self$cluster <- NULL
     }
     invisible(self)
   },

   print_mem_used = function(objects) {
     object_size_kb <- function (name, envir)
       utils::object.size(get(name, envir = envir)) / 1024

     envir <- parent.frame()
     if (missing(objects))
       objects <- ls(envir = envir, all.names = TRUE)

     x <- sapply(objects, object_size_kb, envir = envir)
     y <- sapply(names(get(class(self)[1L])$public_fields), object_size_kb, envir = self)
     names(y) <- paste0("RaceState$", names(y))
     x <- c(x, y)
     # Do not print anything that is smaller than 32 Kb
     x <- x[x > 32]
     cat(sep="", sprintf("%30s : %17.1f Kb\n", names(x), x),
       sprintf("%30s : %17.1f Mb\n", "Total", sum(x) / 1024),
       # This does garbage collection and also prints memory used by R.
       sprintf("%30s : %17.1f Mb\n", "gc", sum(gc()[,2L])))
     invisible(self)
   }
))

no_elitist_init_instances <- function(self, deterministic)
{
  max_instances <- nrow(self$instances_log)
  # if next.instance == 1 then this is the first iteration.
  # If deterministic consider all (do not resample).
  if (self$next_instance == 1L || deterministic) return(seq_len(max_instances))
  irace_assert(self$next_instance < max_instances)
  self$next_instance : max_instances
}

elitist_init_instances <- function(self, deterministic, sampleInstances, elitist_new_instances)
{
  max_instances <- nrow(self$instances_log)
  # if next_instance == 1 then this is the first iteration.
  next_instance <- self$next_instance
  if (next_instance == 1L) return(seq_len(max_instances)) # Consider all

  new_instances <- NULL
  last_new <- next_instance - 1L + elitist_new_instances
  # Do we need to add new instances?
  if (elitist_new_instances > 0L) {
    if (last_new > max_instances) {
      # This may happen if the scenario is deterministic and we would need
      # more instances than what we have.
      irace_assert(deterministic)
      if (next_instance <= max_instances) {
        # Add all instances that we have not seen yet as new ones.
        last_new <- max_instances
        new_instances <- next_instance : last_new
      } # else new_instances remains NULL and last_new remains > number of instances.
         # We need to update this because the value is used below and now there
      # may be fewer than expected, even zero.
      self$elitist_new_instances <- length(new_instances)
    } else {
      new_instances <- next_instance : last_new
    }
  }
  past_instances <- if (sampleInstances) sample.int(next_instance - 1L)
                    else seq_len(next_instance - 1L)

  # new_instances + past_instances + future_instances
  if (last_new + 1L <= max_instances) {
    future_instances <- (last_new + 1L) : max_instances
    return(c(new_instances, past_instances, future_instances))
  }
  c(new_instances, past_instances)
}

## The naive method would be:
##
## rtnorm.naive <- function(n, mean = 0, sd = 1, lower, upper)
## {
##   lower <- (lower - mean) / sd ## Algorithm works on mean 0, sd 1 scale
##   upper <- (upper - mean) / sd

##   x <- rnorm(n)
##   repeat {
##     w <- which( x > upper | x < lower)
##     if (length(w) == 0) break
##     x[w] <- rnorm(length(w))
##   }
##   return(x * sd + mean)
## }
##
## There is also https://cran.r-project.org/web/packages/truncnorm/

## Function to sample according to a truncated normal distribution 
## This function comes from the R package 'msm' maintained by 
## Christopher Jackson <chris.jackson at mrc-bsu.cam.ac.uk>
# 
# Package: msm
# Version: 1.7.1
# Date: 2023-03-23
# Title: Multi-State Markov and Hidden Markov Models in Continuous Time
# Author: Christopher Jackson <chris.jackson@mrc-bsu.cam.ac.uk>
# Maintainer: Christopher Jackson <chris.jackson@mrc-bsu.cam.ac.uk>
# Description: Functions for fitting continuous-time Markov and hidden
#     Markov multi-state models to longitudinal data.  Designed for
#     processes observed at arbitrary times in continuous time (panel data)
#     but some other observation schemes are supported. Both Markov
#     transition rates and the hidden Markov output process can be modelled
#     in terms of covariates, which may be constant or piecewise-constant
#     in time.
# License: GPL (>= 2)
# Imports:
#     survival,mvtnorm,expm
# Suggests:
#     mstate,minqa,doParallel,foreach,numDeriv,testthat,flexsurv
# URL: https://github.com/chjackson/msm
# # Repository: CRAN

## Rejection sampling algorithm by Robert (Stat. Comp (1995), 5, 121-5)
## for simulating from the truncated normal distribution.

rtnorm <- function (n, mean = 0, sd = 1, lower, upper) {
  stopifnot(length(n) == 1L && length(mean) == 1L
            && length(sd) == 1L && length(lower) == 1L && length(upper) == 1L)
  stopifnot(is.finite(lower) && is.finite(upper))
  lower <- (lower - mean) / sd # Algorithm works on mean 0, sd 1 scale
  upper <- (upper - mean) / sd
  sdzero <- (sd < .Machine$double.eps)
  nas <- is.na(mean) || is.na(sd) || (sdzero && (lower > mean || mean > upper))
  ret <- rep_len(NaN, n)
  if (nas) {
    warning("NAs produced")
    return(ret) # return NaN
  } else if (lower > upper) {
    return(ret) # return NaN
  } else if (sdzero) {
    return(rep_len(mean, n)) # SD zero, so set the sampled value to the mean. 
  } else if ((lower < 0) && (upper > 0) && (upper - lower > sqrt(2*pi))) {
    # standard "simulate from normal and reject if outside limits" method. Use if bounds are wide.
    ind.no <- seq_len(n)
    while (length(ind.no) > 0) {
      y <- rnorm(length(ind.no))
      done <- (y >= lower & y <= upper)
      ret[ind.no[done]] <- y[done]
      ind.no <- ind.no[!done]
    }
  } else if (lower >= 0 && (upper > lower + 2*sqrt(exp(1)) /
                            (lower + sqrt(lower^2 + 4)) * exp((lower*2 - lower*sqrt(lower^2 + 4)) / 4))) {
    # rejection sampling with exponential proposal. Use if lower >> mean
    ind.expl <- seq_len(n)
    a <- (sqrt(lower^2 + 4) + lower) / 2
    while (length(ind.expl) > 0) {
      z <- rexp(length(ind.expl), a) + lower
      u <- runif(length(ind.expl))
      done <- (u <= exp(-(z - a)^2 / 2)) & (z <= upper)
      ret[ind.expl[done]] <- z[done]
      ind.expl <- ind.expl[!done]
    }
  } else if (upper <= 0 && (-lower > -upper + 2*sqrt(exp(1)) /
                            (-upper + sqrt(upper^2 + 4)) * exp((upper*2 - -upper*sqrt(upper^2 + 4)) / 4))) {
    # rejection sampling with exponential proposal. Use if upper << mean.
    ind.expu <- seq_len(n)
    a <- (sqrt(upper^2 +4) - upper) / 2
    while (length(ind.expu) > 0) {
      z <- rexp(length(ind.expu), a) - upper
      u <- runif(length(ind.expu))
      done <- (u <= exp(-(z - a)^2 / 2)) & (z <= -lower)
      ret[ind.expu[done]] <- -z[done]
      ind.expu <- ind.expu[!done]
    }
  } else {
    # rejection sampling with uniform proposal. Use if bounds are narrow and central.
    ind.u <- seq_len(n)
    K <- if (lower > 0) lower^2 else if (upper < 0) upper^2 else 0
    while (length(ind.u) > 0) {
      z <- runif(length(ind.u), lower, upper)
      rho <- exp((K - z^2) / 2)
      u <- runif(length(ind.u))
      done <- (u <= rho)
      ret[ind.u[done]] <- z[done]
      ind.u <- ind.u[!done]
    }
  }        
  ret*sd + mean
}

# ---------------------------------------- -*- mode: r; mode: font-lock -*- #
# race.R                       Racing methods for the selection of the best #
# ------------------------------------------------------------------------- #

# ========================================================================= #
# Racing methods for the selection of the best                              #
# ------------------------------------------------------------------------- #
# Copyright (C) 2003 Mauro Birattari                                        #
# ========================================================================= #
# This program is free software; you can redistribute it and/or modify it   #
# under the terms of the GNU General Public License as published by the     #
# Free Software Foundation; either version 2 of the License, or (at your    #
# option) any later version.                                                #
#                                                                           #
# This program is distributed in the hope that it will be useful, but       #
# WITHOUT ANY WARRANTY; without even the implied warranty of                #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         #
# General Public License for more details.                                  #
#                                                                           #
# You should have received a copy of the GNU General Public License along   #
# with this program; if not, write to the Free Software Foundation, Inc.,   #
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.                  #
# ========================================================================= #

# ========================================================================= #
# Mauro BIRATTARI                                                           #
# IRIDIA - ULB, CP 194/6                                                    #
# Av. F. D. Roosevelt 50                                    mbiro@ulb.ac.be #
# 1050 Brussels, Belgium                     http://iridia.ulb.ac.be/~mbiro #
# ========================================================================= #
# $Id: race.R,v 1.54 2005/03/30 12:40:42 mbiro Exp $ #

createExperimentList <- function(configurations, parameters,
                                 instances, instances_ID, seeds, bounds)
{
  n_configurations <- nrow(configurations)
  configurations_id <- configurations[[".ID."]]
  pnames <- parameters$names
  configurations <- as.list(configurations[, pnames, drop=FALSE])
  # FIXME: How to do this faster?
  # - maybe purrr::transpose() ?
  # - mlr3misc::transpose_list() uses .mapply(list, configurations, list(), which is slower.
  configurations <- lapply(seq_len(n_configurations), function(i) lapply(configurations, `[`, i))
  dots <- list(id_configuration = configurations_id, configuration = configurations)
  if (!is.null(bounds)) {
    # There must be a bound for each configuration.
    # FIXME: There must be a bound for each configuration AND each instance.
    irace_assert(length(bounds) == n_configurations)
    dots$bound <- bounds
  }
  configurations <- .mapply(list, dots = dots, MoreArgs = NULL)
  instances <- instances[instances_ID]
  n_instances <- length(instances_ID)
  instances <- .mapply(list,
    dots = list(id_instance = instances_ID, seed = seeds, instance = instances),
    MoreArgs = NULL)
  .mapply(c, dots = list(
    rep(instances, each = n_configurations),
    rep(configurations, times = n_instances)), MoreArgs = NULL)
}


race_wrapper_helper <- function(race_state, configurations, instance_idx, bounds,
                         is_exe, scenario)
{
  # Experiment list to execute
  experiments <- createExperimentList(configurations, parameters = scenario$parameters,
                                      instances = scenario$instances,
                                      instances_ID = race_state$instances_log[["instanceID"]][instance_idx],
                                      seeds = race_state$instances_log[["seed"]][instance_idx],
                                      bounds = bounds)

  irace_assert(length(is_exe) == length(experiments))
  instance_idx <- rep(instance_idx, each = nrow(configurations))
  if (race_state$recovery_mode) {
    # With targetEvaluator or if everything is executed, we recover everything.
    if (!is.null(scenario$targetEvaluator) || all(is_exe)) {
      configuration_id <- unlist_element(experiments, "id_configuration")
      target_output <- race_state$recover_output(instance_idx, configuration_id)
    } else {
      irace_assert(any(is_exe))
      configuration_id <- unlist_element(experiments[is_exe], "id_configuration")
      target_output <- race_state$recover_output(instance_idx[is_exe], configuration_id)
    }
  } else { # !recovery_mode
    # We cannot let targetRunner or targetEvaluator modify our random seed, so we save it.
    withr::local_preserve_seed()
    target_output <- vector("list", length(experiments))
    # Execute experiments for which is_exe is TRUE:
    if (any(is_exe))
      target_output[is_exe] <- execute_experiments(race_state, experiments[is_exe], scenario)

    # If targetEvaluator is NULL, then target_output must contain the right
    # output already.  Otherwise, targetEvaluator considers all experiments.
    if (!is.null(scenario$targetEvaluator)) {
      target_output <- execute_evaluator(race_state$target_evaluator, experiments, scenario, target_output)
    } else if (any(!is_exe))  {
      experiments <- experiments[is_exe]
      instance_idx <- instance_idx[is_exe]
    }
    target_output <- rbindlist(target_output, fill=TRUE, use.names=TRUE)
    set(target_output, j = setdiff(colnames(target_output), c("cost", "time")), value = NULL)
    if ("time" %notin% colnames(target_output))
      set(target_output, j = "time", value = NA)
    set(target_output, j = "configuration", value = unlist_element(experiments, "id_configuration"))
    set(target_output, j = "instance", value = instance_idx)
    if (!is.null(bounds))
      set(target_output, j = "bound", value = unlist_element(experiments, "bound"))
  }
  target_output
}

## Executes a list of configurations in a particular instance
## configurations: description having the id of the configuration
## instance.idx: index of the instance,seed pair in race_state$instances_log
## bounds: execution bounds (NULL if not needed).
## is_exe: Boolean vector that determines which experiments need to executed.
race_wrapper <- function(race_state, configurations, instance_idx, bounds,
                         is_exe, scenario)
{
  target_output <- race_wrapper_helper(race_state, configurations, instance_idx,
    bounds, is_exe, scenario)
  race_state$update_race_experiment_log(target_output, scenario)
  target_output
}

experiments_output_to_matrix <- function(output, scenario)
{
  if (scenario$capping)
    output[["cost"]] <- applyPAR(output[["cost"]], boundMax = scenario$boundMax, boundPar = scenario$boundPar)
  as.matrix(dcast(output[, c("instance", "configuration", "cost")], instance ~ configuration, value.var = "cost"),
    rownames = "instance")
}

aux2.friedman <- function(y, I, alive, conf.level = 0.95)
{
  dropped.any <- FALSE
  n <- nrow(y)
  k <- length(I)
  r <- rowRanks(y, cols = I, ties.method = "average", useNames = FALSE)
  R <- colSums2(r, useNames = FALSE)
  o <- order(R)
  best <- I[o[1L]]
  TIES <- tapply(c(r), row(r), table)
  STATISTIC <- ((12 * sum((R - n * (k + 1L) / 2)^2)) /
                (n * k * (k + 1L)
                  - (sum(unlist(lapply(TIES, function (u) {u^3 - u}))) /
                     (k - 1L))))
  PARAMETER <- k - 1L
  PVAL      <- pchisq(STATISTIC, PARAMETER, lower.tail = FALSE)
  #names(STATISTIC) <- "Friedman chi-squared"
  #names(PARAMETER) <- "df"

  alpha <- 1 - conf.level
  if (!is.nan(PVAL) && PVAL < alpha) {
    # This formula for multiple comparisons comes from Conover, "Practical
    # Nonparametric Statistics", 1999, pages 369-371.
    A <- sum(as.vector(r)^2)
    t <- qt(1 - alpha / 2, df = (n - 1L) * (k - 1L)) *
      (2 * (n * A - sum(R^2)) / ((n - 1L) * (k - 1L)))^(1 / 2)
    J <- best
    for (j in 2L:k) {
      if (abs(R[o[j]] - R[o[1L]]) > t) {
        break
      } else {
        J <- c(J, I[o[j]])
      }
    }
    alive[-J] <- FALSE
    dropped.any <- TRUE
  }
  irace_assert(I[which.min(R)] == best)
  list(ranks = R, alive = alive, dropped.any = dropped.any, p.value = PVAL)
}

aux_friedman <- function(results, alive, which_alive, conf.level)
{
  no.alive <- length(which_alive)
  if (no.alive > 2L) {
    # If more then 2 configurations are left, use Friedman
    return(aux2.friedman(results, which_alive, alive, conf.level = conf.level))
  }

  ranks <- NULL
  dropped.any <- TRUE
  PVAL <- 0
  # If only 2 configurations are left, switch to Wilcoxon
  V1 <- results[, which_alive[1L]]
  V2 <- results[, which_alive[2L]]
  diffs <- V1 - V2
  # Avoid the test if the answer is obvious
  if (all(diffs <= 0)) {
    ranks <- c(1L,2L)
  } else if (all(diffs >= 0)) {
    ranks <- c(2L,1L)
  } else {
    ZEROES <- any(diffs == 0)
    if (ZEROES)
      diffs <- diffs[diffs != 0]
    r <- rank(abs(diffs))
    TIES <- length(r) != length(unique(r))
    diffs <- outer(diffs, diffs, "+")
    diffs <- sort(diffs[!lower.tri(diffs)])/2
    PVAL <- wilcox.test(V1, V2, paired = TRUE, exact = if (ZEROES || TIES) FALSE else NULL)$p.value
    irace_assert(!is.nan(PVAL) & !is.na(PVAL))
    if (PVAL >= 1 - conf.level) dropped.any <- FALSE
    # We use the pseudo median (see wilcox.test.default)
    ranks <- if (median(diffs) <= 0) c(1L,2L) else c(2L,1L)
  }
  if (dropped.any)
    alive[which_alive[ranks[2L]]] <- FALSE

  list(ranks = ranks, alive = alive, dropped.any = dropped.any, p.value = PVAL)
}

aux.one_ttest <- function(results, alive, which_alive, conf.level,
                          adjust = c("none","bonferroni","holm"))
{
  adjust <- match.arg(adjust)
  irace_assert(sum(alive) == length(which_alive))
  results <- results[, which_alive]
  means <- colMeans2(results)
  best <- which.min(means)
  mean_best <- means[best]
  pvals <- sapply(means, function(x) as.numeric(isTRUE(
                                       all.equal.numeric(mean_best[[1L]], x[[1L]], check.attributes = FALSE))))
  results_best <- results[, best]
  var_best <- var(results_best)
  which_test <- which(pvals < 1.0)
  for (j in which_test) {
    PVAL <- pvals[j]
    if (PVAL == 1.0) next
    results_j <- results[, j]
    # t.test may fail if the data in each group is almost constant. Hence, we
    # surround the call in a try() and we initialize p with 1 if the means are
    # equal or zero if they are different
    if (min(var(results_best), var(results_j)) < 10 * .Machine$double.eps) next
    # The t.test may fail if the data are not normal despite one configuration
    # clearly dominating the other.
    if (all(results_best <= results_j)) next
    try(PVAL <- t.test(results_best, results_j, alternative = "less", paired = TRUE)$p.value)
    irace_assert(!is.nan(PVAL) & !is.na(PVAL))
    pvals[j] <- PVAL
  }
  pvals <- p.adjust(pvals, method = adjust)
  dropj <- which_alive[pvals < 1.0 - conf.level]
  dropped_any <- length(dropj) > 0
  irace_assert(all(alive[dropj]))
  alive[dropj] <- FALSE
  list(ranks = means, alive = alive, dropped.any = dropped_any, p.value = min(pvals))
}

aux.ttest <- function(results, alive, which_alive, conf.level,
                      adjust = c("none","bonferroni","holm"))
{
  adjust <- match.arg(adjust)
  irace_assert(sum(alive) == length(which_alive))

  means <- colMeans2(results, cols = which_alive, useNames = FALSE)
  # FIXME: break ties using median or ranks?
  best <- which.min(means)
  mean_best <- means[best]
  pvals <- sapply(means, function(x) as.numeric(isTRUE(
                                       all.equal.numeric(mean_best[[1L]], x[[1L]], check.attributes = FALSE))))
  results <- results[, which_alive]
  results_best <- results[, best]
  var_best <- var(results_best)
  which_test <- which(pvals < 1.0)
  for (j in which_test) {
    PVAL <- pvals[j]
    if (PVAL == 1.0) next
    results_j <- results[, j]
    # t.test may fail if the data in each group is almost constant. Hence, we
    # surround the call in a try() and we initialize p with 1 if the means are
    # equal or 0 if they are different.
    if (min(var(results_best), var(results_j)) < 10 * .Machine$double.eps) next
    # The t.test may fail if the data are not normal despite one configuration
    # clearly dominating the other.
    if (all(results_best <= results_j)) next
    try(PVAL <- t.test(results_best, results_j, paired = TRUE)$p.value)
    irace_assert(!is.nan(PVAL) & !is.na(PVAL))
    pvals[j] <- PVAL
  }
  pvals <- p.adjust(pvals, method = adjust)
  dropj <- which_alive[pvals < 1.0 - conf.level]
  dropped_any <- length(dropj) > 0L
  irace_assert(all(alive[dropj]))
  alive[dropj] <- FALSE
  list(ranks = means, alive = alive, dropped.any = dropped_any, p.value = min(pvals))
}

table_hline <- function(widths)
  paste0(collapse="+", c("",strrep("-", widths), "\n"))

table_sprintf <- function(text, widths)
  paste0(collapse="|", c("", sprintf("%*s",widths, text), "\n"))

.nocap_table_fields_width <- as.integer(c(1, 11, 11, 11, 16, 11, 8, 5, 4, 6))
.nocap_colum_names <- c(" ", "Instance", "Alive", "Best", "Mean best", "Exp so far",
                        "W time", "rho", "KenW", "Qvar")
.capping_table_fields_width <- as.integer(c(1, 11, 8, 11, 11, 16, 11, 8, 5, 4, 6))
.capping_colum_names <- c(" ", "Instance", "Bound", "Alive", "Best", "Mean best", "Exp so far",
                        "W time", "rho", "KenW", "Qvar")
.capping_hline <- table_hline(.capping_table_fields_width)
.nocap_hline <- table_hline(.nocap_table_fields_width)
.race_common_header <-  "# Markers:
     x No test is performed.
     c Configurations are discarded only due to capping.
     - The test is performed and some configurations are discarded.
     = The test is performed but no configuration is discarded.
     ! The test is performed and configurations could be discarded but elite configurations are preserved.
     . All alive configurations are elite and nothing is discarded.\n\n"

.nocap_header <- paste0(.race_common_header, .nocap_hline,
  table_sprintf(.nocap_colum_names, .nocap_table_fields_width), .nocap_hline)
.capping_header <- paste0(.race_common_header, .capping_hline,
  table_sprintf(.capping_colum_names, .capping_table_fields_width), .capping_hline)

race_print_header_nocap <- function()
  cat(sep = "", .nocap_header)

race_print_header_cap <- function()
  cat(sep = "", .capping_header)

elapsed_wctime_str <- function(now, start) {
  if (now <= start) return("00:00:00")
  elapsed <- difftime(now, start, units = "secs")
  # FIXME: Maybe better and faster if we only print seconds?
  format(.POSIXct(elapsed, tz="GMT"), "%H:%M:%S")
}

race_print_task <- function(res.symb, Results,
                            instance,
                            current_task,
                            which_alive,
                            id_best,
                            best,
                            experiments_used,
                            start_time,
                            bound, capping)
{
  cat(sprintf("|%s|%11d|", res.symb, instance))
  if (capping) {
    if (is.null(bound))
      cat("      NA|")
    else
      cat(sprintf("%8.2f|", bound))
  }
  # FIXME: This is the mean of the best, but perhaps it should
  # be the sum of ranks in the case of test == friedman?
  mean_best <- mean(Results[, best])
  time_str <- elapsed_wctime_str(Sys.time(), start_time)
  cat(sprintf(paste0("%11d|%11d|", .irace.format.perf, "|%11d|%s"),
              length(which_alive), id_best, mean_best, experiments_used, time_str))

  if (current_task > 1L && length(which_alive) > 1L) {
    res <- Results[, which_alive, drop = FALSE]
    conc <- concordance(res)
    qvar <- dataVariance(res)
    # FIXME: We would like to use %+#4.2f but this causes problems with
    # https://github.com/oracle/fastr/issues/191
    cat(sprintf("|%+4.2f|%.2f|%.4f|\n", conc$spearman.rho, conc$kendall.w, qvar))
  } else {
    cat("|   NA|  NA|    NA|\n")
  }
}

race_print_footer <- function(bestconf, mean_best, break_msg, debug_level, capping = FALSE,
                              old_best_id)
{
  cat(sep = "",
      if (capping) .capping_hline else .nocap_hline,
      if (debug_level >= 1L) paste0("# Stopped because ", break_msg, "\n"),
      if (!is.null(old_best_id)) paste0("Best configuration for the instances in this race: ", old_best_id, "\n"),
      sprintf("Best-so-far configuration: %11d", bestconf[[".ID."]][1L]),
      "    mean value: ",
      sprintf(.irace.format.perf, mean_best), "\n",
      "Description of the best-so-far configuration:\n")
  configurations_print(bestconf, metadata = TRUE)
  cat("\n")
}

## This function calculates an execution bound
## data: matrix columns as configurations and rows instances
## type:
##   median: bound based on the configuration's mean median
##   mean:   bound based on the configuration's mean mean
##   worst:  bound based on the worst configuration's mean
##   best:  bound based on the best configurations's mean
executionBound <- function(data, type = "median")
{
  irace_assert (ncol(data) >= 1L)

  if (ncol(data) == 1L) {
    return (mean(data[,1L], na.rm = TRUE))
  }
  # This should never happen because the data used to obtain the execution
  # bound should be complete, that is, the bounding configurations should have
  # been executed on all previous instances.
  irace_assert (!anyNA(data))
  colmeans <- colMeans2(data)
  bound <- switch (type,
                   median = median(colmeans),
                   mean   = mean(colmeans),
                   worst  = max(colmeans),
                   # default:
                   min(colmeans))
  bound
}

## This function calculates an execution bound per instance
## data: array of the execution times on the current instance
## type:
##   median: bound based on the median time
##   mean:   bound based on the mean time
##   worst:  bound based on the worst time
##   best:  bound based on the best time
instanceBound <- function(data, type="median")
{
  irace_assert (!anyNA(data))
  switch (type,
          median = median(data),
          mean   = mean(data),
          worst  = max(data),
          # default:
          min(data))
}

## This function returns survivors obtained after applying the dominance elimination
## criterion.
##  results: matrix of experiments results (all configurations)
##  elites: index of elite configurations
##  alive: bolean array of alive configurations
##  eps: constant added to the bound to account for the measuring precision
dom_elim <- function(results, elites, alive, scenario, minSurvival, eps = 1e-5)
{
  which_alive <- which(alive)
  irace_assert(length(which_alive) >= minSurvival)
  cmeans <- colMeans2(results, cols = which_alive, useNames = FALSE)
  irace_assert(!all(is.na(cmeans))) # Only NA values when calculating mean for dominance.

  # When there are no protected elites left, select the best configurations to
  # calculate the bounds. This is quite aggressive and another alternative
  # would be to disable dom_elim when elites == 0.
  if (length(elites) == 0L) {
    # In the case we have only two alive configurations only one can be elite.
    if (length(which_alive) <= 2L)
      elites <- which_alive[which.min(cmeans)]
    else
      elites <- which_alive[order(cmeans, na.last = TRUE, decreasing = FALSE)[seq_len(minSurvival)]]
  }

  bound <- executionBound(results[, elites, drop = FALSE], type = scenario$cappingType)
  alive[which_alive] <- ((bound + eps) >= cmeans)
  alive
}

## This function applies PARX (X=boundPar) to all experiments
## that exceed the maximum execution time (boundMax)
applyPAR <- function(results, boundMax, boundPar)
{
  # We do not want to change Inf or -Inf because those represent rejection.
  if (boundPar != 1)
    results[is.finite(results) & results >= boundMax] <- boundMax * boundPar
  results
}

## This function calculates the execution time allowed for the executions
## of configurations based on previous execution times.
## It returns a list with two elements:
## * elite_bound : value (mean execution time or per instance execution time)
## used to calculate the maximum execution time (final_bounds) for each configuration.
## * final_bounds[i] : maximum execution time for candidate i on the current instance.
final_execution_bound <- function(experimentsTime, elites, current_task,
                                  which_alive, which_exe, scenario)
{
  boundMax <- scenario$boundMax
  final_bounds <- rep(boundMax, length(which_alive))

  # Elite candidates can have NA values due to the rejection.
  if (length(elites))
    elites <- elites[!is.na(experimentsTime[current_task,elites])]

  # Only apply bounds when there is previous data
  if (length(elites) == 0L || length(which_exe) == 0L) {
    # FIXME: should we use an adjusted boundMax ?
    return(list(final_bounds = final_bounds, elite_bound = boundMax))
  }
  minMeasurableTime <- scenario$minMeasurableTime

  # The elite membership is updated before calling this function to know
  # the configurations used for calculating the bounds we need to do this.
  # Note that some of these configurations could be not elite anymore for the
  # elimination phase, given that all their evaluations have been used.
  if (scenario$boundType == "instance") {
    elite_bound <- instanceBound(experimentsTime[current_task, elites],
      type = scenario$cappingType)
    final_bounds_exe <- min(elite_bound + minMeasurableTime, boundMax)
    final_bounds_exe <- ceiling_digits(final_bounds_exe, scenario$boundDigits)
  } else {
    elite_bound <- executionBound(experimentsTime[seq_len(current_task), elites, drop = FALSE],
      type = scenario$cappingType)
    elite_bound <- min(elite_bound, boundMax)
    total_time <- (current_task * elite_bound) + minMeasurableTime
    final_bounds_exe <- total_time - colSums2(experimentsTime, rows = seq_len(current_task), cols = which_exe, na.rm = TRUE, useNames = FALSE)
    # There are cases in which a small negative budget is used. For example:
    # assuming candidates 1 and 2 are elite:
    # maxBound <- 80
    # executionTime <- matrix(c(0.010,0.0170,0.010, 24,28,27, 0.010,0.017,NA),
    #        byrow=TRUE, ncol=3, nrow=3, dimnames=list(c(1,2,3),c(1,2,3)))
    # current_task <- 3; boundDigits = 0
    # elite_bound  <- irace:::executionBound(executionTime[1:current_task,1:2], "median")
    # total_time   <- elite_bound * current_task + 0.01
    # final_bounds_exe <- total_time - colSums2(executionTime[1:current_task,3,drop=FALSE], na.rm=TRUE)
    #
    # We set the execution time to the elite_bound, this should be enough
    # to eliminate a bad candidate for the next task.
    final_bounds_exe[final_bounds_exe <= 0] <- elite_bound
    # We round up the bounds up to the specified number of digits. This may
    # be necessary if the target-algorithm does not support higher precision.
    final_bounds_exe <- ceiling_digits(final_bounds_exe, scenario$boundDigits)
    final_bounds_exe[final_bounds_exe > boundMax] <- boundMax
    irace_assert(all(final_bounds_exe > 0))
  }
  final_bounds[which(which_alive %in% which_exe)] <- final_bounds_exe
  list(final_bounds = final_bounds, elite_bound = elite_bound)
}

get_ranks <- function(x, test)
  if (test == "friedman") colSums2(rowRanks(x, ties.method="average")) else rank(colMeans2(x))

# Recompute the best as follows. Given two configurations, the one evaluated
# on more instances is ranked better. Otherwise, break ties according to the
# criteria of the stat test.
overall_ranks <- function(x, test)
{
  if (ncol(x) == 1L) return(1L)

  ninstances <- colSums2(!is.na(x))
  uniq_ninstances <- sort(unique(ninstances), decreasing = TRUE)
  last_r <- 0L
  ranks <- rep_len(Inf, ncol(x))
  # Iterate from the largest to the lowest number of instances.
  for (k in uniq_ninstances) {
    confs <- which(ninstances == k)
    irace_assert(all(is.infinite(ranks[confs])))
    r <- 1L
    if (length(confs) > 1L) {
      # Select only non-NA rows
      y <- x[, confs, drop = FALSE]
      y <- y[complete.cases(y), , drop = FALSE]
      r <- get_ranks(y, test = test)
    }
    r <- r + last_r
    last_r <- max(r)
    ranks[confs] <- r
  }
  ranks
}

# Remove one elite count from every configuration not executed.
update_is_elite <- function(is_elite, which_exe)
{
  which_notexecuted <- setdiff(which(is_elite > 0L), which_exe)
  is_elite[which_notexecuted] <- is_elite[which_notexecuted] - 1L
  irace_assert (all(is_elite >= 0L))
  is_elite
}

update_elite_safe <- function(Results, is_elite)
{
  elites <- is_elite > 0L
  if (!any(elites)) return(0L) # All elites rejected.
  max(which(rowAnyNotNAs(Results, cols = elites)))
}

generateTimeMatrix <- function(elite_ids, experiment_log)
{
  # Silence CRAN warnings
  configuration <- bound <- NULL
  # Remove everything that we don't need.
  experiment_log <- experiment_log[(configuration %in% elite_ids) & !is.na(time),
    c("configuration", "instance", "time", "bound")]
  experiment_log[, time := pmin.int(time, bound)]
  experiment_log[, bound := NULL]
  time_matrix <- dcast(experiment_log, instance ~ configuration, value.var = "time")
  setcolorder(time_matrix, neworder = as.character(elite_ids))
  as.matrix(time_matrix, rownames = "instance")
}

race <- function(race_state, maxExp,
                 minSurvival = 1L,
                 configurations,
                 scenario)
  elitist_race(race_state = race_state, maxExp = maxExp,
               minSurvival = minSurvival,
               elite_data = NULL,
               configurations = configurations,
               scenario = scenario,
               elitist_new_instances = 0L)

elitist_race <- function(race_state, maxExp,
                 minSurvival = 1L,
                 elite_data = NULL,
                 configurations,
                 scenario,
                 elitist_new_instances,
                 firstTest = scenario$firstTest)
{
  blockSize <- scenario$blockSize
  conf.level <- scenario$confidence
  firstTest <- blockSize * firstTest
  eachTest <- blockSize * scenario$eachTest
  elitist <- scenario$elitist
  capping <- scenario$capping
  n_configurations <- nrow(configurations)
  alive <- rep_len(TRUE, n_configurations)
  is_rejected <- logical(n_configurations)

  ## FIXME: Remove argument checking. This must have been done by the caller.
  irace_assert(maxExp > 0L)

  if (n_configurations <= minSurvival) {
    irace_error("Not enough configurations (", n_configurations,
                ") for a race (minSurvival=", minSurvival, ")")
  }

  # Check argument: conf.level
  if (!missing(conf.level) &&
      (!is.numeric(conf.level) || length(conf.level)!=1 ||
       !is.finite(conf.level) || conf.level < 0 || conf.level > 1))
    stop("conf.level must be a single number between 0 and 1")

  if (scenario$quiet) {
    print_header <- print_task <- print_footer <- do_nothing
  } else {
    print_header <- if (capping) race_print_header_cap else race_print_header_nocap
    print_task <- race_print_task
    print_footer <- race_print_footer
  }

  stat_test <- scenario$testType
  do_test <- switch(stat_test,
    friedman = function(results, alive, which_alive, conf.level = scenario$confidence)
      aux_friedman(results, alive, which_alive, conf.level = conf.level),
    t.none = function(results, alive, which_alive, conf.level = scenario$confidence)
      aux.ttest(results, alive, which_alive, conf.level = conf.level, adjust = "none"),
    t.holm = function(results, alive, which_alive, conf.level = scenario$confidence)
      aux.ttest(results, alive, which_alive, conf.level = conf.level, adjust = "holm"),
    t.bonferroni = function(results, alive, which_alive, conf.level = scenario$confidence)
      aux.ttest(results, alive, which_alive, conf.level = conf.level, adjust = "bonferroni"))
  do_test <- bytecompile(do_test)

  # Create the instance list according to the algorithm selected.
  if (elitist) {
    # FIXME: we should sample taking into account the block-size, so we sample blocks, not instances.
    irace_assert((race_state$next_instance - 1L) %% blockSize == 0,
      eval_after={cat("next_instance:", race_state$next_instance, ", block_size:", blockSize, "\n")})
    race_instances <- elitist_init_instances(race_state, deterministic = scenario$deterministic,
      sampleInstances = scenario$sampleInstances, elitist_new_instances = elitist_new_instances)
    # It may be reduced further by elitist_init_instances()
    elitist_new_instances <- min(elitist_new_instances, race_state$elitist_new_instances)
    all_elite_instances_evaluated <- function() {
      if (race_state$next_instance == 1L) return(TRUE)
      evaluated <- !is.na(Results[, alive, drop=FALSE])
      # All instances that have been previously seen have been evaluated by at
      # least one configuration
      all(rowAnys(evaluated)) &&
        # and the number of instances evaluated per configuration is a multiple
        # of eachTest (which is scenario$blockSize * scenario$eachTest).
        all(colSums2(evaluated) %% eachTest == 0L)
    }
  } else {
    race_instances <- no_elitist_init_instances(race_state, deterministic = scenario$deterministic)
    all_elite_instances_evaluated <- function() TRUE
  }

  irace_assert(!anyDuplicated(race_instances))
  irace_assert(identical(sort(race_instances), seq_along(race_instances)))
  nb_tasks <- length(race_instances)

  # Initialize some variables...
  experiments_used <- 0L

  ## FIXME: Probably, instead of this, we should keep elite_safe in the race_state.
  if (is.null(elite_data)) {
    elite_safe <- 0L
    elite_instances_ID <- NULL
  } else {
    irace_assert(race_state$next_instance - 1L == nrow(elite_data))
    # There must be a non-NA entry for each instance.
    irace_assert(all(rowAnyNotNAs(elite_data)),
                 eval_after = { print(elite_data)})
    # There must be a non-NA entry for each configuration.
    irace_assert(all(colAnyNotNAs(elite_data)),
                 eval_after = {
                   cat("elite_data:\n")
                   print(elite_data)
                   cat("experiment_log:\n")
                   print(race_state$experiment_log)
                 })

    # elite_safe: maximum instance number for which any configuration may be
    # considered elite. After evaluating this instance, no configuration is
    # elite.
    elite_safe <- elitist_new_instances + nrow(elite_data)
    elite_instances_ID <- as.character(race_instances[seq_len(elite_safe)])
  }

  configurations_ID <- as.character(configurations[[".ID."]])
  Results <- matrix(NA_real_,
                    nrow = elite_safe,
                    ncol = n_configurations,
                    dimnames = list(elite_instances_ID, configurations_ID))
  if (capping)
    experimentsTime <- matrix(NA_real_,
                              nrow = elite_safe,
                              ncol = n_configurations,
                              dimnames = list(elite_instances_ID, configurations_ID))

  if (is.null(elite_data)) {
    # is_elite[i] : number of instances to be seen in this race on which i has
    # been previously evaluated.
    is_elite <- integer(n_configurations)
  } else {
    Results[rownames(elite_data), colnames(elite_data)] <- elite_data
    irace_assert(all(colnames(elite_data) %chin% configurations_ID))
    if (capping) {
      # Temporarily use only 0 or 1, we will calculate the real value below.
      is_elite <- as.integer(configurations_ID %chin% colnames(elite_data))
      tmp <- generateTimeMatrix(elite_ids = colnames(elite_data),
                                experiment_log = race_state$experiment_log)
      experimentsTime[rownames(tmp), colnames(tmp)] <- tmp
      # Preliminary execution of elite configurations to calculate
      # the execution bound of initial configurations (capping only).
      if (elitist_new_instances > 0L) {
        irace_assert(elitist_new_instances %% blockSize == 0L)
        # FIXME: This should go into its own function.
        n_elite <- ncol(elite_data)
        which_elites <- which(is_elite > 0L, useNames=FALSE)
        irace_assert(identical(which_elites, seq_len(n_elite)))
        irace_note("Preliminary execution of ", n_elite,
          " elite configuration(s) over ", elitist_new_instances, " instance(s).\n")
        # FIXME: Launch all seq_len(elitist_new_instances) experiments in parallel.
        for (k in seq_len(elitist_new_instances)) {
          output <- race_wrapper (race_state,
            configurations = configurations[which_elites, , drop = FALSE],
            instance_idx = race_instances[k],
            bounds = rep(scenario$boundMax, n_elite),
            # which_exe values are within 1:nbConfigurations, whereas experiments
            # indices are within 1:length(which_alive). The following line converts
            # from one to the other.
            is_exe = rep_len(TRUE, n_elite), scenario = scenario)
          # Extract results:
          irace_assert(length(output[["cost"]]) == n_elite)
          Results[k, which_elites] <- applyPAR(output[["cost"]], boundMax = scenario$boundMax, boundPar = scenario$boundPar)
          irace_assert(!anyNA(output[["time"]]))
          experimentsTime[k, which_elites] <- output[["time"]] # capping is enabled
          irace_assert(all.equal(configurations[[".ID."]][which_elites], output[["configuration"]]))
          irace_assert(all.equal(output[["bound"]], rep(scenario$boundMax, n_elite)))
          irace_assert(all.equal(unique(output[["instance"]]), race_instances[k]))
          experiments_used <- experiments_used + n_elite

          # We remove elite configurations that are rejected given that
          # is not possible to calculate the bounds.
          rejected <- is.infinite(output[["cost"]])
          irace_assert(all.equal(as.vector(is.infinite(Results[k, which_elites])), rejected), eval_after={
            cat("rejected:\n")
            print(output[["cost"]])
            print(rejected)
            cat("Results:\n")
            print(Results[k, which_elites])
            print(is.infinite(Results[k, which_elites]))
          })
          is_rejected[which_elites] <- rejected
          is_elite[rejected] <- 0L
          which_elites <- which_elites[!rejected]
          n_elite <- length(which_elites)

          # If all elite are eliminated we stop execution of initial instances.
          if (n_elite == 0L) {
            irace_note ("All elite configurations are rejected. Execution of non-elites will be not bounded.\n")
            break
          }
        }
        if (any(is_rejected)) {
          irace_note ("Immediately rejected configurations: ",
            paste0(configurations[[".ID."]][is_rejected], collapse = ", ") , "\n")
          alive[is_rejected] <- FALSE
          # Calculate the maximum instance that has any non-NA value.
          # FIXME: Use update_elite_safe()
          if (n_elite > 0L)
            elite_safe <- max(which(rowAnys(!is.na(Results[, which_elites, drop=FALSE]))))
          else
            elite_safe <- 0L
          irace_assert(identical(update_elite_safe(Results, is_elite), elite_safe))
          # FIXME: If sum(alive) <= minSurvival, we stop later but we will have
          # elites that have not been evaluated in any instance, which is
          # bad. Ideally we would sample new configurations here but that is
          # too hard to do in iterated-racing.
        }
      }
    } # end if(capping)
    # Compute the elite membership.
    is_elite <- colSums2(!is.na(Results))
    # Remove rejected configurations.
    is_elite[is_rejected] <- 0L
  }


  no_elimination <- 0L # number of tasks without elimination.
  print_header()
  # Start main loop.
  break_msg <- NULL
  best <- NA_integer_
  which_alive <- which(alive)
  nb_alive    <- length(which_alive)

  for (current_task in seq_len(nb_tasks)) {
    which_exe   <- which_alive
    if (elitist && any(is_elite > 0L)) {
      # Filter configurations that do not need to be executed (elites).
      # This is valid only for previous iteration instances.
      irace_assert(current_task <= elite_safe)
      # Execute everything that is alive and not yet executed.
      which_exe <- which(alive & is.na(Results[current_task, ]))
      if (length(which_exe) == 0L) {
        is_elite <- update_is_elite(is_elite, which_exe)
        # LESLIE: This is the case in which there are only elite configurations alive
        # and we are still in the previous instances execution, but we can still
        # continue with the race. (This is only possible because the early termination
        # criterion is disabled)
        ## MANUEL: So what is the reason to not immediately terminate here? Is
        ## there a reason to continue?
        if (current_task == 1L) {
          # We may reach this point in the first iteration so we need to calculate best.
          if (nb_alive == 1L) {
            best <- which_alive
          } else  {
            tmpResults <- Results[1L, which_alive, drop = FALSE]
            irace_assert(!anyNA(tmpResults))
            # which.min returns only the first minimum.
            best <- which_alive[which.min(get_ranks(tmpResults, test = stat_test))]
          }
        }
        if (is.na(best)) {
          utils::dump.frames(dumpto = "best_crash", to.file = TRUE,
                      include.GlobalEnv = TRUE)
          irace_assert(!is.na(best))
        }
        id_best <- configurations[[".ID."]][best]
        print_task(".", Results[seq_len(current_task), , drop = FALSE],
                   race_instances[current_task],
                   current_task, which_alive = which_alive,
                   id_best = id_best,
                   best = best, experiments_used, start_time = Sys.time(),
                   # FIXME: Why do we pass NA as bound? Why not pass the actual bound if any?
                   bound = NA_real_, capping = capping)
        next
      }
    }

    if (all_elite_instances_evaluated()) {
      # We stop when we have less configurations than required.
      if (nb_alive <= minSurvival) {
        # Stop race if we have less or equal than the minimum number of
        # configurations.
        break_msg <- paste0("number of alive configurations (", nb_alive,
          ") <= minimum number of configurations (", minSurvival, ")")
        break
      }
      ## We continue running if (1) we have not reached the firstTest or (2)
      ## there are instances previously seen that have not been evaluated on any
      ## alive configuration.  If we just did a test, check that we have enough
      ## budget to reach the next test.
      # FIXME: In post-selection racing, we want to consume all budget, so we
      # should discard configurations until we have 2.
      if (current_task > firstTest && ( (current_task - 1L) %% eachTest) == 0L
        && experiments_used + length(which_exe) * eachTest > maxExp) {
        break_msg <- paste0("experiments for next test (",
          experiments_used + length(which_exe) * eachTest,
          ") > max experiments (", maxExp, ")")
        break
      }

      if (elitist && scenario$elitistLimit != 0L && no_elimination >= scenario$elitistLimit) {
        break_msg <- paste0("tests without elimination (", no_elimination,
          ") >= elitistLimit (", scenario$elitistLimit, ")")
        break
      }
    }

    if (nrow(Results) < current_task) {
      Results <- rbind(Results, rep_len(NA_real_, ncol(Results)))
      rownames(Results) <- race_instances[seq_nrow(Results)]
      if (capping) {
        experimentsTime <- rbind(experimentsTime, rep_len(NA_real_, ncol(experimentsTime)))
        rownames(experimentsTime) <- race_instances[seq_nrow(experimentsTime)]
      }
    }

    start_time <- Sys.time()

    # Calculate bounds for executing if needed.
    which_elite_exe <- intersect(which_exe, which(is_elite > 0L))
    irace_assert(setequal(which_elite_exe, which(is_elite & is.na(Results[current_task,]))))
    if (capping) {
      # Pre-execute elite configurations that are not yet executed in the current instance.
      if (length(which_elite_exe)) {
        # FIXME: This should go into its own function
        output <- race_wrapper(race_state,
          configurations = configurations[which_elite_exe, , drop = FALSE],
          instance_idx = race_instances[current_task],
          # FIXME: What if we already have a bound for this instance?
          bounds = rep(scenario$boundMax, length(which_elite_exe)),
          is_exe = rep_len(TRUE, length(which_elite_exe)), scenario = scenario)
        # Extract results
        irace_assert(length(output[["cost"]]) == length(which_elite_exe))
        Results[current_task, which_elite_exe] <- applyPAR(output[["cost"]], boundMax = scenario$boundMax, boundPar = scenario$boundPar)
        irace_assert(!anyNA(output[["time"]]))
        irace_assert(all.equal(configurations[which_elite_exe, ".ID."], output[["configuration"]]))
        experimentsTime[current_task, which_elite_exe] <- output[["time"]]
        irace_assert(all.equal(unique(output[["instance"]]), race_instances[current_task]))
        experiments_used <- experiments_used + length(which_elite_exe)

        # We remove elite configurations that are rejected given that
        # is not possible to calculate the bounds
        rejected <- is.infinite(output[["cost"]])
        if (any(rejected)) {
          irace_note("Immediately rejected configurations: ",
            paste0(configurations[[".ID."]][which_elite_exe[rejected]],
              collapse = ", ") , "\n")
          is_rejected[which_elite_exe] <- rejected
          is_elite[is_rejected] <- 0L
          alive[which_elite_exe] <- !rejected
          if (!any(alive)) {
            # FIXME: Only report this error if (all(is_rejected)); otherwise
            # restore non-rejected non-alive ones. Restoring a non-alive
            # configuration is difficult. We need to evaluate it in all the
            # instances that it has missed.
            irace_error("All configurations have been immediately rejected (all of them returned Inf) !")
          }
          which_alive <- which(alive)
          nb_alive    <- length(which_alive)
          elite_safe <- update_elite_safe(Results, is_elite)
        }
        which_exe <- setdiff(which_exe, which_elite_exe)
        # FIXME: There is similar code above. Can we merge these code paths?
        if (length(which_exe) == 0L) {
          is_elite <- update_is_elite(is_elite, which_elite_exe)
          if (current_task == 1L) {
            # We may reach this point in the first iteration so we need to calculate best.
            if (nb_alive == 1L) {
              best <- which_alive
            } else  {
              tmpResults <- Results[1L, which_alive, drop = FALSE]
              irace_assert(!anyNA(tmpResults))
              # which.min returns only the first minimum.
              best <- which_alive[which.min(get_ranks(tmpResults, test = stat_test))]
            }
          }
          if (is.na(best)) {
            utils::dump.frames(dumpto = "best_crash", to.file = TRUE,
                        include.GlobalEnv = TRUE)
            irace_assert(!is.na(best))
          }
          id_best <- configurations[[".ID."]][best]
          print_task(".", Results[seq_len(current_task), , drop = FALSE],
                     race_instances[current_task],
                     current_task, which_alive = which_alive,
                     id_best = id_best,
                     best = best, experiments_used, start_time = start_time,
                     # FIXME: Why do we pass NA as bound? Why not pass the actual bound if any?
                     bound = if (is.null(scenario$boundMax)) NA_real_ else scenario$boundMax, capping)
          next
        }
      }
      all_bounds <- final_execution_bound(experimentsTime,
                                          elites = which(is_elite > 0L),
                                          current_task, which_alive,
                                          which_exe, scenario)
      final_bounds <- all_bounds$final_bounds
      elite_bound <- all_bounds$elite_bound
    } else {
      final_bounds <- rep(scenario$boundMax, length(which_alive))
      elite_bound <- NULL
    }

    # Execute experiments.
    output <- race_wrapper(race_state, configurations = configurations[which_alive, , drop = FALSE],
                           instance_idx = race_instances[current_task],
                           bounds = final_bounds,
                           is_exe = which_alive %in% which_exe,
                           scenario = scenario)
    # Extract results
    # Set max execution bound to timed out executions which have execution
    # times smaller than boundMax and implement parX if required.
    vcost <- output[["cost"]]
    # Output is not indexed in the same way as configurations.
    which_has_time <- which(which_alive %in% which_exe)
    # With !is.null(scenario$targetEvaluator) we will have duplicated (instance, configuration) in output.
    irace_assert(all.equal(output[["bound"]], if (is.null(scenario$targetEvaluator)) final_bounds[which_has_time]
                                              else final_bounds))
    if (capping) {
      vcost <- applyPAR(vcost, boundMax = scenario$boundMax, boundPar = scenario$boundPar)
      if (scenario$boundAsTimeout) {
        timeout_bounds <- if (is.null(scenario$targetEvaluator)) final_bounds[which_has_time]
                          else final_bounds
        irace_assert(all.equal(output[["bound"]], timeout_bounds))
        # We do not want to change Inf or -Inf because those represent rejection.
        vcost[is.finite(vcost) & (vcost >= timeout_bounds) & (vcost < scenario$boundMax)] <- scenario$boundMax
      }

      # If targetEvaluator was used, we do not update the times because no
      # evaluation actually happened, only the cost values possibly changed.
      vtimes <- if (is.null(scenario$targetEvaluator)) output[["time"]] else output[["time"]][which_has_time]
      # Correct higher execution times.
      irace_assert(all.equal(if (is.null(scenario$targetEvaluator)) output[["bound"]] else output[["bound"]][which_has_time], final_bounds[which_has_time]))
      experimentsTime[current_task, which_has_time] <- pmin(vtimes, final_bounds[which_has_time])
    }
    ## Currently, targetEvaluator always re-evaluates, which implies that the
    ## value may change without counting as an evaluation. We do this to allow online normalization.
    which_has_cost <- if (is.null(scenario$targetEvaluator)) which_exe else which_alive
    irace_assert(length(output[["cost"]]) == length(which_has_cost))
    Results[current_task, which_has_cost] <- vcost
    experiments_used <- experiments_used + length(which_exe)

    # We update the elites that have been executed.
    is_elite <- update_is_elite(is_elite, which_elite_exe)

    ## Drop bad configurations.
    ## Infinite values denote immediate rejection of a configuration.
    # FIXME: Should this be which_has_cost?
    rejected <- is.infinite(Results[current_task, which_exe])
    if (any(rejected)) {
      irace_note ("Immediately rejected configurations: ",
                  paste0(configurations[which_exe[rejected], ".ID."],
                         collapse = ", ") , "\n")
      is_rejected[which_exe] <- rejected
      is_elite[is_rejected] <- 0L
      alive[is_rejected] <- FALSE
      if (!any(alive)) {
        # FIXME: Only report this error if (all(is_rejected)); otherwise
        # restore non-rejected non-alive ones. Restoring a non-alive
        # configuration is difficult. We need to evaluate it in all the
        # instances that it has missed.
        irace_error("All configurations have been immediately rejected (all of them returned Inf) !")
      }
      which_alive <- which(alive)
      nb_alive    <- length(which_alive)
      # FIXME: Should we stop  if (nbAlive <= minSurvival) ???
      elite_safe <- update_elite_safe(Results, is_elite)
    }
    irace_assert(!anyNA(Results[seq_len(current_task), alive, drop=FALSE]))
    irace_assert(!any(is.infinite(Results[, alive, drop=FALSE])))

    # Variables required to produce output of elimination test.
    cap_done     <- FALSE # if dominance elimination was performed
    test.done    <- FALSE # if statistical test elimination was performed
    cap_dropped  <- FALSE # if capping has drop any configuration
    test.dropped <- FALSE # if any candidates has been eliminated by testing
    cap_alive    <- test.alive <- alive

    ## Dominance elimination (Capping only).
    # The second condition can be false if we eliminated via immediate
    # rejection.  The third condition ensures that we see the block before capping.
    if (capping && nb_alive > minSurvival && (current_task %% blockSize) == 0L
      && (!scenario$cappingAfterFirstTest || current_task >= firstTest)) {
      irace_assert(!any(is_elite > 0L) == (current_task >= elite_safe))
      cap_alive <- dom_elim(Results[seq_len(current_task), , drop = FALSE],
                            # Get current elite configurations.
                            elites = which(is_elite > 0L),
                            alive, scenario, minSurvival)
      cap_dropped <- nb_alive > sum(cap_alive)
      cap_done    <- TRUE
    }

    # We assume that firstTest is a multiple of eachTest.  In any
    # case, this will only do the first test after the first multiple
    # of eachTest that is larger than firstTest.
    if (current_task >= firstTest && (current_task %% eachTest) == 0L
        && nb_alive > 1L) {
      irace_assert(sum(alive) == nb_alive)
      test_res <- do_test(Results[seq_len(current_task), ], alive, which_alive)
      # FIXME: This race_ranks is unused. We should check if it matches the one computed below.
      race_ranks <- test_res$ranks
      test.alive <- test_res$alive
      test.dropped <- nb_alive > sum(test.alive)
      test.done   <- TRUE
    }

    # Merge the result of both eliminations.
    prev_nb_alive <- nb_alive
    prev_which_alive  <- which_alive
    alive <- cap_alive & test.alive

    # Handle elites when elimination is performed.  The elite configurations
    # can be removed only when they have no more previously-executed instances.
    irace_assert(!any(is_elite > 0L) == (current_task >= elite_safe))
    if (!is.null(elite_data) && any(is_elite > 0L)) {
      irace_assert (length(alive) == length(is_elite))
      alive <- alive | (is_elite > 0L)
    }

    # It may happen that the capping and the test eliminate together all
    # configurations. In that case, we only trust the capping elimination.
    if (capping && !any(alive)) {
      if (scenario$debugLevel >= 2L) {
        irace_warning("Elimination tests have eliminated all configurations, keeping the capping results.\n")
        irace_note("Alive according to capping:", which(cap_alive), "\n")
        irace_note("Alive according to test:", which(test.alive), "\n")
      }
      alive <- cap_alive
    }
    which_alive <- which(alive)
    nb_alive <- length(which_alive)
    # Output the result of the elimination test.
    res.symb <- if (cap_dropped && !test.dropped && prev_nb_alive != nb_alive) {
                  "c" # Removed just by capping.
                } else if (cap_dropped || test.dropped) {
                  if (prev_nb_alive != nb_alive) "-" else "!"
                } else if (cap_done || test.done) "=" else "x"

    # Rank alive configurations: order all configurations (eliminated or not)
    # LESLIE: we have to make the ranking outside: we can have configurations eliminated by capping
    # that are not eliminated by the test.
    # MANUEL: I don't understand the above comment.
    if (length(prev_which_alive) == 1L) {
      race_ranks <- 1L
      best <- prev_which_alive
    } else  {
      tmpResults <- Results[seq_len(current_task), prev_which_alive, drop = FALSE]
      irace_assert(!anyNA(tmpResults))
      race_ranks <- get_ranks(tmpResults, test = stat_test)
      # which.min() returns only the first minimum.
      best <- prev_which_alive[which.min(race_ranks)]
    }

    irace_assert(best == prev_which_alive[order(race_ranks)][1L])
    irace_assert(length(race_ranks) == length(prev_which_alive))

    # Remove the ranks of those that are not alive anymore
    race_ranks <- race_ranks[which_alive]
    irace_assert(length(race_ranks) == nb_alive)
    id_best <- configurations[[".ID."]][best]
    print_task(res.symb, Results[seq_len(current_task), , drop = FALSE],
               race_instances[current_task],
               current_task, which_alive = which_alive,
               id_best = id_best, best = best, experiments_used, start_time = start_time,
               bound = elite_bound, capping)

    if (elitist) {
      # Compute number of statistical tests without eliminations.
      irace_assert(!any(is_elite > 0L) == (current_task >= elite_safe))
      if (!any(is_elite > 0L)
          && current_task > firstTest && (current_task %% eachTest) == 0L) {
        no_elimination <- if (nb_alive == prev_nb_alive) no_elimination + 1L else 0L
      }
    }
  }

  if (is.null(break_msg))
    break_msg <- paste0("all instances (", nb_tasks, ") evaluated")

  # Adding this given that when ncandidates = minsurvival+1
  # and there one elite configuration that gets discarded in the new instances
  # execution the race is finished with no executions.
  # FIXME: we should handle this better, maybe allowing irace to handle no elite
  # in irace()
  # MANUEL: Leslie, how can we reach this error in normal circumstances?
  # Can we handle this better?
  if (current_task == 1L && all(is_elite == 0L))
    irace_error ("Maximum number configurations immediately rejected reached!")

  # All instances that are not new in this race must have been evaluated by at
  # least one configuration.
  irace_assert(all_elite_instances_evaluated(),
               eval_after = { print(Results[,alive, drop=FALSE])})
  # If we stop the loop before we see all new instances, there may be new
  # instances that have not been executed by any configuration.
  Results <- Results[rowAnyNotNAs(Results), , drop = FALSE]
  # If we reject configurations so that sum(alive) <= minSurvival, we may stop
  # before we evaluate some configurations in any instance.
  if (any(is_rejected)) {
    alive <- alive & colAnyNotNAs(Results)
    which_alive <- which(alive)
    if (!any(alive)) {
      # FIXME: Only report this error if (all(is_rejected)); otherwise
      # restore non-rejected non-alive ones. Restoring a non-alive
      # configuration is difficult. We need to evaluate it in all the
      # instances that it has missed.
      irace_error("All configurations have been immediately rejected (all of them returned Inf) !")
    }
  }

  race_ranks <- overall_ranks(Results[, alive, drop = FALSE], test = stat_test)
  if (!scenario$quiet) {
    old_best <- best # old_best could be NA.
    best <- which_alive[which.min(race_ranks)]
    mean_best <- mean(Results[, best])
    print_footer(bestconf = configurations[best, , drop = FALSE],
                 # FIXME: This is the mean of the best, but perhaps it
                 # should be the sum of ranks in the case of test == friedman?
                 mean_best = mean_best,
                 break_msg = break_msg, debug_level = scenario$debugLevel,
                 capping = capping,
                 old_best_id  = if (old_best == best || is.na(old_best)) NULL else id_best)
  }
  rejected_ids <- configurations[is_rejected, ".ID."]
  scenario$parameters$forbid_configurations(
    race_state$update_rejected(rejected_ids, configurations)
  )
  # Only return alive ones.
  configurations <- configurations[alive, , drop=FALSE]
  irace_assert(all(configurations[[".ID."]] %not_in% rejected_ids)) # No rejected is alive.
  # Assign the proper ranks in the configurations data.frame.
  configurations[[".RANK."]] <- race_ranks
  # Now we can sort the data.frame by the rank.
  configurations <- configurations[order(configurations[[".RANK."]]), , drop=FALSE]
  if (scenario$debugLevel >= 3L) {
    irace_note ("Memory used in race():\n")
    race_state$print_mem_used()
  }
  local_experiment_log <- race_state$reset_race_experiment_log()

  # nrow(Results) may be smaller, equal or larger than current_task.
  if (is.null(scenario$targetEvaluator)) {
    # With targetEvaluator, we may have the recorded a new cost value but not
    # counted it as an experiment used if targetRunner was not called.
    irace_assert(anyDuplicated(local_experiment_log[, c("instance", "configuration")]) == 0L,
      eval_after = {
        print(local_experiment_log)
        print(mget(ls()))
      })
    irace_assert(nrow(local_experiment_log) == experiments_used)
  }
  list(experiments = Results,
       experiment_log = local_experiment_log,
       experimentsUsed = experiments_used,
       configurations = configurations)
}

#' Summarise the results of a run of irace
#' 
#' @inheritParams has_testing_data
#'
#' @return `list()`
#' 
#' @examples
#' irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata",
#'                                           package="irace", mustWork=TRUE))
#' irace_summarise(irace_results)
#' 
#' @author Manuel López-Ibáñez
#' @concept analysis
#' @export
irace_summarise <- function(iraceResults)
{
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing")
  iraceResults <- read_logfile(iraceResults)

  if (is.null(iraceResults$state$elapsed)) {
    time_cpu_user <- time_cpu_sys <- time_cpu_total <- time_wallclock <- NA
  } else {
    time_cpu_user <- iraceResults$state$elapsed[["user"]]
    time_cpu_sys <- iraceResults$state$elapsed[["system"]]
    time_cpu_total <- time_cpu_user + time_cpu_sys
    time_wallclock <- iraceResults$state$elapsed[["wallclock"]]
  }

  n_iterations <- length(iraceResults$allElites)
  experiment_log <- iraceResults$state$experiment_log
  if (is.null(experiment_log)) {
    experiment_log <- iraceResults$experimentLog
  } else if (is.null(experiment_log))
    stop("Experiment log is NULL")
  
  version <- iraceResults$irace_version
  if (is.null(version))
    version <- iraceResults$irace.version

  # Here to support older versions of irace.
  time_targetrunner <- iraceResults$state$recovery_info$timeUsed
  if (is.null(time_targetrunner))
    time_targetrunner <- iraceResults$state$timeUsed

  rejected_ids <- iraceResults$state$rejected_ids
  if (is.null(rejected_ids))
    rejected_ids <- iraceResults$state$rejectedIDs
  
  list(
    version = version,
    n_iterations = n_iterations,
    n_configurations = nrow(iraceResults$allConfigurations),
    n_initial_configurations = if (is.null(iraceResults$scenario$initConfigurations)) 0L else nrow(iraceResults$scenario$initConfigurations),
    n_instances = nrow(iraceResults$experiments),
    n_experiments = if (is.null(iraceResults$scenario$targetEvaluator)) nrow(experiment_log) else sum(!is.na(iraceResults$experiments)),
    n_elites = length(iraceResults$allElites[[n_iterations]]),
    n_soft_restarts = sum(iraceResults$softRestart),
    n_rejected = length(rejected_ids),
    time_targetrunner = time_targetrunner,
    time_cpu_user = time_cpu_user,
    time_cpu_sys = time_cpu_sys,
    time_cpu_total = time_cpu_total,
    time_wallclock = time_wallclock,
    termination_reason = if (is.null(iraceResults$state$completed)) "Missing" else iraceResults$state$completed
  )
}

#' Execute [irace()] multiple times with the same or different scenarios and parameter space definitions.
#'
#' There are three modes of operation:
#' \itemize{
#'  \item One `scenarios` and `k` `parameters`: `k` runs with the same scenario and each parameter space definition.
#'  \item One `parameters` and `k` `scenarios`: `k` runs with the same parameter space definition and each scenario.
#'  \item `k` `parameters` and `k` scenarios: `k` runs with each scenario and parameter space definition.
#' }
#' Each of the `k` runs can be repeated `n` times by supplying a value for `n`.
#'
#' @param scenarios `list()`\cr A list of scenarios.
#' If only a single scenario is supplied, it is used for all parameters.
#' @param parameters `list()`\cr A list of parameter space definitions.
#' If only a single definition is supplied, it is used for all scenarios.
#' @param n `integer(1)`\cr The number of repetitions.
#' @param parallel `integer(1)`\cr The number of workers to use.
#' A value of `1` means sequential execution. Note that `parallel > 1` is not supported on Windows.
#' @param split_output `logical(1)`\cr If `TRUE`, the output of [irace()] is written to `{execDir}/run_{i}/irace.out`
#'  instead of the standard output.
#' @param global_seed `integer(1)`\cr The global seed used to seed the individual runs.
#'
#' @return A list of the outputs of [irace()].
#'
#' @seealso
#' \describe{
#'  \item{[irace()]}{the main interface for single irace runs.}
#' }
#'
#' @concept running
#' @export
multi_irace <- function(scenarios, parameters, n = 1L, parallel = 1L, split_output = parallel > 1L, global_seed = NULL)
{
  # Parallel execution is not available on Windows.
  if (.Platform$OS.type == 'windows' && parallel > 1L) {
    irace_error("multi_irace() does not yet support parallel > 1 on Windows")
  }

  # Allow either the same number of scenarios and parameters, or a single scenario or parameter space definition.
  if (length(scenarios) != length(parameters)) {
    if (length(scenarios) == 1L) {
      scenarios <- rep(scenarios, each = length(parameters))
    } else if (length(parameters) == 1L) {
      parameters <- rep(parameters, each = length(scenarios))
    } else {
      irace_error("Invalid arguments: ",
                  "Cannot execute 'irace' with", length(scenarios),
                  "scenarios and", length(parameters), "parameters.",
                  "Either supply the same number of scenarios and parameters,
                  or one scenario or one parameter space definition.")
    }
  }

  # Repeat the existing scenarios and parameters 'n' times.
  if (n > 1L) {
    scenarios <- rep(scenarios, each = n)
    parameters <- rep(parameters, each = n)
  }

  # Overwrite scenario seeds.
  seeds <- gen_random_seeds(length(scenarios), global_seed = global_seed)

  for (i in seq_along(scenarios)) {
    # FIXME: We should store this seed in state not in scenario. We should not modify scenario.
    scenarios[[i]]$seed <- seeds[[i]]
    scenarios[[i]]$parameters <- parameters[[i]]
    # Check each scenario.
    scenarios[[i]] <- checkScenario(scenarios[[i]])

    # Modify 'logFile' and 'execDir' with the index of the run.
    # Paths are guaranteed to be absolute because of 'checkScenario'.
    logFile_old <- scenarios[[i]]$logFile
    execDir_old <- scenarios[[i]]$execDir

    # 'path/to/execDir' -> 'path/to/execDir/run_{i}'.
    execDir <- file.path(execDir_old, sprintf("run_%02d", i))
    scenarios[[i]]$execDir <- execDir
    fs::dir_create(execDir)

    if (logFile_old != "") {
      logFile <- if (is.sub.path(logFile_old, execDir_old)) {
        # 'logFile' is located in the old 'execDir', so move it into the new 'execDir'.
        # 'path/to/execDir/logFile.rdata' -> 'path/to/execDir/run_{i}/logFile.rdata'.
        sub(execDir_old, execDir, logFile_old)
      } else {
        # 'logFile' is located outside, so adapt the file name.
        # 'path/to/logFile.rdata' -> 'path/to/logFile_{i}.rdata'.
        # pathWithoutExt <- tools::file_path_sans_ext(logFile_old)
        # pathExt <- tools::file_ext(logFile_old)
        # pathWithoutExtWithIndex <- sprintf("%s_%02d", pathWithoutExt, i)
        # paste(pathWithoutExtWithIndex, pathExt, sep = ".")
        irace_error("Invalid 'logFile' path (", logFile_old, "): ",
          "The 'logFile' must be located inside the 'execDir' (", execDir_old, ").")
      }
      scenarios[[i]]$logFile <- logFile
    }
  }


  irace_run <- function(scenario) {
    if (scenario$quiet || !split_output) {
      irace(scenario)
    } else {
      withr::with_output_sink(file.path(scenario$execDir, "irace.out"), {
        irace(scenario)
      })
    }
  }

  if (parallel > 1L) {
    runs <- parallel::mcmapply(irace_run, scenarios, mc.cores = parallel, SIMPLIFY = FALSE)
    # FIXME: if stop() is called from mcmapply, it does not
    # terminate the execution of the parent process, so it will
    # continue and give more errors later. We have to terminate
    # here, but is there a nicer way to detect this and terminate?
    if (any(sapply(runs, inherits, "try-error")) || any(sapply(runs, is.null))) {
      # FIXME: mcmapply has some bugs in case of error. In that
      # case, each element of the list does not keep the output of
      # each configuration and repetitions may occur.
      errors <- unique(unlist(runs[sapply(runs, inherits, "try-error")]))
      cat(errors, file = stderr())
      irace_error("A child process triggered a fatal error")
    }
  } else {
    runs <- mapply(irace_run, scenarios, SIMPLIFY = FALSE)
  }
  runs
}

# Print a user-level warning message, when the calling context
# cannot help the user to understand why the program failed.
irace_warning <- function(...)
{
  if (getOption(".irace.quiet", default=FALSE)) return()
  warning(paste0(.irace_msg_prefix, ..., collapse=""),
          call. = FALSE, immediate. = TRUE)
}

# Print a user-level fatal error message, when the calling context
# cannot help the user to understand why the program failed.
irace_error <- function(...)
{
  # The default is only 1000, which is too small. 8170 is the maximum
  # value allowed up to R 3.0.2
  withr::local_options(list(warning.length = 8170))
  stop (.irace_msg_prefix, ..., call. = FALSE)
}

## When irace crashes, it generates a file "iracedump.rda". To debug the crash use:
## R> load("iracedump.rda")
## R> debugger(iracedump)
##
## See help(dump.frames) for more details.
irace_dump_frames <- function()
{
  execDir <- getOption(".irace.execdir")
  if (!is.null(execDir)) cwd <- setwd(execDir)
  utils::dump.frames(dumpto = "iracedump", to.file = TRUE, include.GlobalEnv = TRUE)
  # FIXME: We want to use on.exit(setwd(cwd)) but q() does not run on.exit.
  if (!is.null(execDir)) setwd(cwd)
  # We need this to signal an error in R CMD check. See help(dump.frames)
  if (!interactive()) quit("no", status = 1)
}

# Print an internal fatal error message that signals a bug in irace.
irace_internal_error <- function(...)
{
  .irace.bug.report <-
    paste0(.irace_msg_prefix, "An unexpected condition occurred. ",
           "Please report this bug to the authors of the irace package <https://github.com/MLopez-Ibanez/irace/issues>")

  op <- list(warning.length = 8170L)
  if (!base::interactive())
    op <- c(op, list(error = irace_dump_frames))
  withr::local_options(op)
  # 6 to not show anything below irace_assert()
  bt <- utils::capture.output(traceback(5))
  warnings()
  stop (.irace_msg_prefix, paste0(..., collapse = "\n"), "\n",
        paste0(bt, collapse= "\n"), "\n",
        .irace.bug.report, call. = FALSE)
  invisible()
}

irace_assert <- function(exp, eval_after = NULL)
{
  # FIXME: It would be great if we could save into a file the state of
  # the function that called this one.
  if (isTRUE(as.logical(exp))) return(invisible())
  mc <- sys.call()[[2L]]
  msg <- paste0("'", deparse(mc), "' is not TRUE")
  if (!is.null(eval_after)) {
    msg_after <- eval.parent(utils::capture.output(eval_after))
    msg <- paste0(msg, "\n", paste0(msg_after, collapse="\n"))
  }
  irace_internal_error(msg)
  invisible()
}

irace_note <- function(...)
{
  # FIXME: If this was a function within an irace object, we could replace it
  # when using quiet.
  if (getOption(".irace.quiet", default=FALSE)) return()
  cat ("# ", format(Sys.time(), usetz=TRUE), ": ",
       paste0(..., collapse = ""), sep = "")
}

file.check <- function (file, executable = FALSE, readable = executable,
                        writeable = FALSE,
                        isdir = FALSE, notempty = FALSE, text = NULL)
{
  EXEC <- 1 # See documentation of the function file.access()
  WRITE <- 2
  READ <- 4

  if (!is.character(file) || is.null.or.empty(file)) {
    irace_error (text, " ", shQuote(file), " is not a vaild filename")
  }
  file <- path_rel2abs(file)
  ## The above should remove the trailing separator if present for windows OS
  ## compatibility, except when we have just C:/, where the trailing separator
  ## must remain.

  if (!file.exists(file)) {
    if (writeable) {
      if (tryCatch({ suppressWarnings(file.create(file) && file.remove(file)) },
                   error=function(e) FALSE))
        return(TRUE)
      irace_error("cannot create ", text, " ", shQuote(file))
      return (FALSE)
    }
    irace_error (text, " '", file, "' does not exist")
    return(FALSE)
  }

  if (writeable && (file.access(file, mode = WRITE) != 0)) {
    irace_error(text, " '", file, "' cannot be written into")
    return(FALSE)
  }

  if (readable && (file.access(file, mode = READ) != 0)) {
    irace_error(text, " '", file, "' is not readable")
    return (FALSE)
  }
  if (executable && file.access(file, mode = EXEC) != 0) {
    irace_error(text, " '", file, "' is not executable")
    return (FALSE)
  }

  if (isdir) {
    if (!file.info(file)$isdir) {
      irace_error(text, " '", file, "' is not a directory")
      return (FALSE)
    }
    if (notempty && length(list.files (file, recursive=TRUE)) == 0) {
      irace_error(text, " '", file, "' does not contain any file")
      return (FALSE)
    }
  } else if (file.info(file)$isdir) {
    irace_error(text, " '", file, "' is a directory, not a file")
    return (FALSE)
  }
  return (TRUE)
}

# Returns the smallest multiple of d that is higher than or equal to x.
round_to_next_multiple <- function(x, d) (x + d - 1L - (x - 1L) %% d)

# This returns FALSE for Inf/-Inf/NA
is.wholenumber <- function(x, tol = .Machine$double.eps^0.5)
  is.finite(x) & (abs(x - round(x)) < tol)


is_na_nowarn <- function(x)
  length(x) == 1L && suppressWarnings(is.na(x))

is_na_or_empty <- function(x)
  (length(x) == 0L) || is_na_nowarn(x)

is.null.or.na <- function(x)
  is.null(x) || is_na_nowarn(x)

is.null.or.empty <- function(x)
  (length(x) == 0L) || (length(x) == 1L && !suppressWarnings(is.na(x)) && is.character(x) && x == "")

is_null_or_empty_or_na <- function(x)
  (length(x) == 0L) || is_na_nowarn(x) || (length(x) == 1L && !suppressWarnings(is.na(x)) && is.character(x) && x == "")

is.function.name <- function(FUN)
{
  # FIXME: Is there a simpler way to do this check?
  is.function(FUN) ||
    (!is.null(FUN) && !is.na(FUN) && as.character(FUN) != "" &&
     !is.null(get.function(FUN)))
}

get.function <- function(FUN)
{
  if (is.function(FUN)) return(FUN)
  FUN <- dynGet(as.character(FUN), ifnotfound = NULL, inherits = TRUE)
  if (is.function(FUN)) return(FUN)
  NULL
}

is.bytecode <- function(x) typeof(x) == "bytecode"

bytecompile <- function(x)
{
  if (is.bytecode(x)) return(x)
  compiler::cmpfun(x)
}

# This function is used to trim potentially large strings for printing, since
# the maximum error/warning length is 8170 characters (R 3.0.2)
strlimit <- function(s, limit = 5000L)
{
  if (nchar(s) <= limit) return(s)
  paste0(substr(s, 1L, limit - 3L), "...")
}

#' Update filesystem paths of a scenario consistently.
#'
#' This function should be used to change the filesystem paths stored in a
#' scenario object. Useful when moving a scenario from one computer to another.
#'
#' @inheritParams defaultScenario
#' @param from `character(1)`\cr Character string containing a regular expression (or character
#'   string for `fixed = TRUE`) to be matched.
#' @param to `character(1)`\cr  The replacement string.character string. For `fixed = FALSE`
#'   this can include backreferences `"\1"` to `"\9"` to
#'   parenthesized subexpressions of `from`.
#' @param fixed `logical(1)`\cr If `TRUE`, `from` is a string to be matched
#'   as is.
#' @return The updated scenario
#' @examples
#' \dontrun{
#' scenario <- readScenario(filename = "scenario.txt")
#' scenario <- scenario_update_paths(scenario, from = "/home/manuel/", to = "/home/leslie")
#' }
#' @seealso [base::grep()]
#' @export
scenario_update_paths <- function(scenario, from, to, fixed = TRUE)
{
  pathParams <- .irace.params.def[.irace.params.def[, "type"] == "p", "name"]
  # Only consider the ones that actually appear in scenario.
  pathParams <- intersect(pathParams, names(scenario))
  scenario[pathParams] <- lapply(scenario[pathParams], sub, pattern = from, replacement = to, fixed = fixed)
  scenario
}

test.type.order.str <- function(test.type)
  switch(test.type,
         friedman = "sum of ranks",
         t.none =, # Fall-throught
         t.holm =, # Fall-throught
         t.bonferroni = "mean value",
         irace_internal_error ("test.type.order.str() Invalid value '",
                               test.type, "' of test.type"))


trim_leading <- function(str)
  sub('^[[:space:]]+', '', str, perl = TRUE) ## white space, POSIX-style

trim_trailing <- function(str)
  sub('[[:space:]]+$', '', str, perl = TRUE) ## white space, POSIX-style

# remove leading and trailing white space characters
trim <- function(str) trim_trailing(trim_leading(str))

## This function takes two matrices x and y and merges them such that the
## resulting matrix z has:
# rownames(z) <- setunion(rownames(x), rownames(y)) and
# rownames(z) <- setunion(rownames(x), rownames(y)) and
# z[rownames(x), colnames(x)] <- x and z[rownames(y), colnames(y)] <- y, and
# z[i, j] <- NA for all i,j not in x nor y.
merge_matrix <- function(x, y)
{
  rownames_x <- rownames(x)
  colnames_x <- colnames(x)
  rownames_y <- rownames(y)
  colnames_y <- colnames(y)

  if (is.null(rownames_x) || is.null(colnames_x))
    return(y)

  if (is.null(rownames_y) || is.null(colnames_y))
    return(x)

  row_union <- union(rownames_x, rownames_y)
  col_union <- union(colnames_x, colnames_y)

  z <- matrix(NA_real_, nrow = length(row_union), ncol = length(col_union),
              dimnames = list(row_union, col_union))

  # Map row and column names to indices for efficient assignment.
  row_idx_x <- chmatch(rownames_x, row_union)
  col_idx_x <- chmatch(colnames_x, col_union)
  row_idx_y <- chmatch(rownames_y, row_union)
  col_idx_y <- chmatch(colnames_y, col_union)

  z[row_idx_x, col_idx_x] <- x
  z[row_idx_y, col_idx_y] <- y

  # There must be a non-NA entry for each instance.
  irace_assert(all(rowAnyNotNAs(z)))
  return(z)
}

# FIXME: This may not work when working interactively. For example,
# one cannot change the number of slaves.  A more robust function
# would try to close any open slaves, and then re-spawn a different
# number.
##
# FIXME2: Slaves will load the irace namespace independently of how the master
# loaded it. This can be seen by doing in the slaves:
#
## print(loadedNamespaces())
## try(print(as.list(get(".__NAMESPACE__.", envir = asNamespace("irace", base.OK = FALSE),
##                       inherits = FALSE))$path))
## try(print(path.package("irace")))
#
# That is, neither R_LIBS, .libPaths or whether library was called with lib.loc
# will affect the slaves. It also happens before we can set those variables on
# the slaves. Thus, one may end up running a different version of irace on the
# slaves than on the master. I wasted more than 12 hours trying to find a
# work-around but nothing seems to work.
mpiInit <- function(nslaves, debugLevel = 0)
{
  # Load the Rmpi package if it is not already loaded.
  if ("Rmpi" %not_in% loadedNamespaces()) {
    if (!suppressPackageStartupMessages(requireNamespace("Rmpi", quietly = TRUE)))
      irace_error("The 'Rmpi' package is required for using MPI")

    # When R exits, finalize MPI.
    reg.finalizer(environment(Rmpi::mpi.exit), function(e) {
      # Rmpi already prints a message, so we don't need this.
      # cat("# Finalize MPI...\n")
      if (Rmpi::mpi.comm.size(1) > 0L)
        # FIXME: dellog == TRUE tries to delete log files, but it does
        # not take into account that we may have changed directory and
        # it does not fail gracefully but produces an annoying:
        # Warning message: running command 'ls *.30577+1.*.log 2>/dev/null' had status 2
        Rmpi::mpi.close.Rslaves(dellog = FALSE)
      # We would like to use .Call("mpi_finalize", PACKAGE = "Rmpi"), which is
      # what mpi.finalize does, minus the annoying message: "Exiting Rmpi. Rmpi
      # cannot be used unless relaunching R", which we do not care about
      # because this finalizer should only be called when exiting R.
      utils::capture.output(Rmpi::mpi.finalize(),
        file = if (.Platform$OS.type == 'windows') 'NUL' else '/dev/null')
    }, onexit = TRUE)

    # Create slaves
    #  needlog: a logical. If TRUE, R BATCH outputs will be saved in log
    #           files.  If FALSE, the outputs will send to /dev/null.
    #  quiet: a logical. If TRUE, do not print anything unless an error occurs.
    #         If FALSE, prints to stdio how many slaves are successfully
    #         spawned and where they are running.
    Rmpi::mpi.spawn.Rslaves(nslaves = nslaves, quiet = (debugLevel < 2),
                            needlog = (debugLevel > 0))
  }
}

## FIXME: Move this to the manual page.
## FIXME: Export this function.
# Computes:
# * Kendall's W (also known as Kendall's coefficient of concordance)
#   If 1, all configurations have ranked in the same order in all instances.
#   If 0, the ranking of each configuration on each instance is essentially random.
#   W = Friedman / (m * (k-1))
#
# * Spearman's rho: average (Spearman) correlation coefficient computed on the
#   ranks of all pairs of raters. If there are no repeated data values, a
#   perfect Spearman correlation of +1 or −1 occurs when each of the variables
#   is a perfect monotone function of the other.
#
# data: matrix with the data, instances in rows (judges), configurations in
# columns.
concordance <- function(data)
{
  irace_assert (is.matrix(data) && is.numeric(data))

  n <- nrow(data) #judges
  k <- ncol(data) #objects
  if (n <= 1L || k <= 1L)
    return(list(kendall.w = NA_real_, spearman.rho = NA_real_))

  # Get rankings by rows (per instance)
  r <- rowRanks(data, ties.method = "average", useNames=FALSE)
  TIES <- c(table(r,row(r)))
  # If everything is tied, then W=1, perfect homogeneity.
  if (all(TIES == k)) {
    W <- 1
  } else {
    # FIXME: This formula seems slightly different from the one in
    # friedman.test. Why?
    TIES <- sum(TIES^3L - TIES)
    R <- colSums2(r, useNames=FALSE)
    W <- ((12 * sum((R - n * (k + 1) / 2)^2L)) /
            ((n^2L * (k^3L - k)) - (n * TIES)))
  }
  # Spearman's rho
  rho <- (n * W - 1) / (n - 1L)

  ## Same as in friedman test
  #STATISTIC <- n * (k - 1) * W
  #PARAMETER <- k - 1
  #pvalue <- pchisq(PARAMETER, df = PARAMETER, lower.tail = FALSE)
  list(kendall.w = W, spearman.rho = rho)
}

## FIXME: Move this to the manual page.
## FIXME: Reference! Explain a bit what is computed!
# Calculates Performance similarity of instances
#       data: matrix with the data, instances in rows (judges), configurations
#             in columns.
# Returns: variance value [0,1], where 0 is a homogeneous set of instances and
#          1 is a heterogeneous set.
# FIXME: How to handle missing values?
dataVariance <- function(data)
{
  irace_assert (is.matrix(data) && is.numeric(data))
  # LESLIE: should we rank data??
  # MANUEL: We should add the option.
  if (nrow(data) <= 1L || ncol(data) <= 1L) return(NA_real_)

  # Normalize
  #datamin <- apply(data,1,min,na.rm=TRUE)
  #datamax <- apply(data,1,max,na.rm=TRUE)
  #normdata <- (data - datamin) / (datamax-datamin)

  #standardize
  meandata <- rowMeans2(data, useNames=FALSE)
  stddata  <- rowSds(data, useNames=FALSE)
  # If stddata == 0, then data is constant and it doesn't matter as long as it
  # is non-zero.
  stddata[stddata == 0] <- 1
  zscoredata <- (data - meandata) / stddata

  # FIXME: We could log-tranform if needed
  # Variance of configurations
  mean(colVars(zscoredata))
}

runcommand <- function(command, args, id, debugLevel, timeout = 0)
{
  if (debugLevel >= 2L) {
    irace_note (command, " ", args, "\n")
    elapsed <- proc.time()["elapsed"]
  }
  err <- NULL
  output <- withCallingHandlers(
    tryCatch(system2(command, args, stdout = TRUE, stderr = TRUE, timeout = timeout),
             error = function(e) {
               err <<- c(err, paste(conditionMessage(e), collapse="\n"))
               NULL
             }), warning = function(w) {
               err <<- c(err, paste(conditionMessage(w), collapse="\n"))
               invokeRestart("muffleWarning")
             })
  if (is.null(output))
    output <- ""
  # If the command could not be run an R error is generated.  If ‘command’
  # runs but gives a non-zero exit status this will be reported with a
  # warning and in the attribute ‘"status"’ of the result: an attribute
  # ‘"errmsg"’ may also be available.
  if (!is.null(err)) {
    err <- paste(err, collapse = "\n")
    if (!is.null(attr(output, "errmsg")))
      err <- paste(sep = "\n", err, attr(output, "errmsg"))
    if (debugLevel >= 2L)
      irace_note ("ERROR (", id, "): ", err, "\n")
    return(list(output = output, error = err))
  }
  if (debugLevel >= 2L) {
    irace_note ("DONE (", id, ") Elapsed wall-clock seconds: ",
                formatC(proc.time()["elapsed"] - elapsed,
                        format = "f", digits = 2), "\n")
  }
  # TODO: Return elapsed time so that we can report at the end the total
  # elapsed time taken by irace vs the time taken by the target-runner.
  list(output = output, error = NULL)
}

# Safe sampling of vector:
resample <- function(x, ...) x[sample.int(length(x), ...)]

# Rounds up the number x to the specified number of decimal places 'digits'.
ceiling_digits <- function(x, digits)
{
   multiple <- 10^-digits
   div <- x / multiple
   int_div <- trunc(div)
   int_div * multiple + ceiling(div - int_div) * multiple
}

# ceil.decimal <- function(x, d) {
  # # get the significant digits in the integer part.
  # ssd <- x * 10^(d)
  # # get the non significant digits
  # nsd <- ssd - floor(ssd)

  # ssd <- trunc(ssd)
  # sel <- nsd > 0 | ssd==0
  # ssd[sel] <- ssd[sel] + 1
  # x2 <- ssd/10^(d)
  # return(x2)
# }

is.file.extension <- function(filename, ext)
  substring(filename, nchar(filename) + 1L - nchar(ext)) == ext

is.sub.path <- function(x, dir, n = nchar(dir)) substr(x, 1L, n) == dir

# Same as !(x %in% table). Package data.table has %notin%.
"%not_in%" <- function(x, table) is.na(match(x, table))

#' Save the log generated by \pkg{{irace}} to a file (by default `irace.Rdata`).
#'
#' This function may be useful if you are manually editing the log data generated by a run of \pkg{irace}.
#'
#' @param iraceResults `list()`\cr Object created by \pkg{irace} and typically saved in the log file `irace.Rdata`.
#'
#' @param logfile `character(1)`\cr Filename to save `iraceResults`. Usually, this is given by `scenario$logFile`. If `NULL` or `""`, no data is saved.
#'
#' @seealso [read_logfile()]
#' @concept analysis
save_irace_logfile <- function(iraceResults, logfile)
{
  # FIXME: Raul Santa Maria proposed to only save if sufficient time (>= 1
  # minute) has passed since the last save. We would need an option to force
  # saving the last one.
  if (is.null.or.empty(logfile)) return(invisible())
  # Files produced by `saveRDS` (or `serialize` to a file connection) are not
  # suitable as an interchange format between machines, for example to download
  # from a website. The files produced by `save` have a header identifying the
  # file type and so are better protected against erroneous use.
  save(iraceResults, file = logfile, version = 3L)
}

valid_iracelog <- function(x)
{
  is.list(x) && ("scenario" %in% names(x))
}

#' Read the log file produced by irace (`irace.Rdata`).
#'
#' @param filename Filename that contains the log file saved by irace. Example: `irace.Rdata`.
#'
#' @param name Optional argument that allows overriding the default name of the object in the file.
#'
#' @return (`list()`)
#' @examples
#' irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata", package="irace",
#'                                           mustWork=TRUE))
#' str(irace_results)
#' @concept analysis
#' @export
read_logfile <- function(filename, name = "iraceResults")
{
  if (is_na_or_empty(filename))
    irace_error("read_logfile: 'filename' is NULL or NA.")
  # If filename is already the iraceResults object, just return it.
  if (valid_iracelog(filename))
    return(filename)

  if (file.access(filename, mode = 4) != 0)
    irace_error("read_logfile: Cannot read file '", filename, "'.")

  load(filename)
  iraceResults <- get0(name, inherits=FALSE)
  if (!valid_iracelog(iraceResults))
    irace_error("read_logfile: The file '", filename, "' does not contain the '", name, "' object.")

  # data.table recommends doing this after loading a data.table from a file.
  setDT(iraceResults$state$experiment_log)
  setDT(iraceResults$state$instances_log)
  iraceResults
}

get_log_clean_version <- function(iraceResults)
{
  log_version <- iraceResults$irace_version
  if (is.null(log_version))
    log_version <- iraceResults$irace.version
  if (is.null(log_version))
    return(package_version("0.0.0"))
  if (log_version == "unknown") # This may happen during development.
    return(package_version("1000.0.0"))
  if (length(gregexpr("\\.", log_version)[[1L]]) > 3L
    || grepl("[a-z]", log_version))
    log_version <- sub("\\.[^.]*$", "", log_version)
  package_version(log_version)
}

#' Check if the results object generated by irace has data about the testing phase.
#'
#' @param iraceResults `list()`|`character(1)`\cr Object created by \pkg{irace} and typically saved in the log file `irace.Rdata`. If a character string is given, then it is interpreted as the path to the log file from which the `iraceResults` object will be loaded.
#'
#' @return `logical(1)`
#' @examples
#' irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata", package="irace",
#'                                           mustWork=TRUE))
#' print(has_testing_data(irace_results))
#' @export
has_testing_data <- function(iraceResults)
{
  ins <- iraceResults$scenario$testInstances
  exp <- iraceResults$testing$experiments
  !(length(ins) == 0L
    || (length(ins) == 1L && (is.na(ins) || ins == ""))
    || length(exp) == 0L || !(is.matrix(exp) || is.data.frame(exp)))
}

do_nothing <- function(...) invisible()

seq_nrow <- function(x) seq_len(nrow(x))

clamp <- function(x, lower, upper) pmax.int(lower, pmin.int(x, upper))

truncate_rows <- function(x, n)
{
  nx <- nrow(x)
  if (nx <= n) return(x)
  x[-seq.int(n + 1L, nx), ]
}


vlast <- function(x)
{
  stopifnot(is.null(dim(x)))
  lx <- length(x)
  if (!lx)
    x
  else
    x[[lx]]
}

# 2147483647 is the maximum value for a 32-bit signed integer.
# We use replace = TRUE, because replace = FALSE allocates memory for each possible number.
# Do not use .Machine$integer.max unless it is smaller, to minimize differences
# between machines.
runif_integer <- function(size)
  sample.int(min(2147483647L, .Machine$integer.max), size = size, replace = TRUE)

unlist_element <- function(x, element)
  unlist(lapply(x, "[[", element, exact=TRUE), recursive=FALSE, use.names=FALSE)

# Extensions of matrixStats
rowAnyNotNAs <- function(x, rows = NULL, cols = NULL, ..., useNames = FALSE)
  !rowAlls(x, rows = rows, cols = cols, value = NA, ..., useNames = useNames)

colAnyNotNAs <- function(x, rows = NULL, cols = NULL, ..., useNames = FALSE)
  !colAlls(x, rows = rows, cols = cols, value = NA, ..., useNames = useNames)

withr::with_output_sink("test-get-functions.Rout", {

  test_that("getConfigurationById()", {
    log <- read_logfile(system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE))
    ids <- sample(log$allConfigurations[[".ID."]], size=3)
    ids <- c(ids, rev(ids))
    sel_ids <- getConfigurationById(log, ids = ids)[[".ID."]]
    expect_identical(sel_ids, ids)
})

})

####################################################
## INITIALISE AND UPDATE THE MODEL
####################################################

# Initial standard deviation for numerical sampling model.
init_sd_numeric <- function(param)
{
  # Dependent parameters define the standard deviation as
  # a portion of the size of the domain interval. In this case,
  # 0.5 indicates half of the interval, equivalent  to
  # (domain[2] - domain[1]) * 0.5
  if (param[["is_dependent"]] || param[["transform"]] == "log")
    return(0.5)
  domain <- param[["domain"]]
  (domain[[2L]] - domain[[1L]]) * 0.5
}

## Initialisation of the model after the first iteration
##
# IN: configurations matrix, parameters datastructure
##
# OUTPUT: A list of list of vectors. The higher-level list contains
# one element per categorical parameter.  Each categorical parameter
# contains a list of vector. This list contains elements which are the
# .ID. of the configuration. 
initialiseModel <- function (parameters, configurations)
{
  nbConfigurations <- nrow(configurations)
  ids <- as.character(configurations[[".ID."]])
  param_names <- parameters$names_variable
  model <- setNames(vector("list", length(param_names)), param_names)
                    
  for (currentParameter in param_names) {
    param <- parameters$get(currentParameter)
    type <- param[["type"]]
    if (type == "c") {
      nbValues <- length(param[["domain"]])
      value <- rep_len(1. / nbValues, nbValues)
      param <- rep_len(list(value), nbConfigurations)
    } else {
      if (type == "r" || type == "i") {
        sd <- init_sd_numeric(param)
        values <- configurations[[currentParameter]]
      } else if (type == "o") {
        domain <- param[["domain"]]
        sd <- (length(domain) - 1L) * 0.5
        values <- match(configurations[[currentParameter]], domain)
      } else {
        irace_internal_error("Unknown parameter type '", type, "'")
      }
      # Assign current parameter value to model.
      param <- mapply(c, sd, values, SIMPLIFY=FALSE, USE.NAMES=FALSE)
    }
    names(param) <- ids
    model[[currentParameter]] <- param
  }
  model
}

updateModel <- function(parameters, eliteConfigurations, oldModel,
                        indexIteration, nbIterations, nbNewConfigurations, elitist)
{
  dec_factor <- (1 - ((indexIteration - 1) / nbIterations))
  add_factor <- ((indexIteration - 1) / nbIterations)
  num_factor <- ((1 / nbNewConfigurations)^(1 / parameters$nbVariable))
  prob_max <- 0.2^(1. / parameters$nbVariable)
  update_prob <- if (elitist) function(p, idx) {
    # Decrease first all values in the vector:
    p <- p * dec_factor
    p[idx] <- p[idx] + add_factor
    # Normalize probabilities.
    p <- p / sum(p)
    # Prevent probabilities from growing too much.
    p <- pmin.int(p, prob_max)
    p <- p / sum(p)
    p / sum(p)
  } else function(p, idx) {
    # Decrease first all values in the vector:
    p <- p * dec_factor
    p[idx] <- p[idx] + add_factor
    # Normalize probabilities.
    p <- p / sum(p)
  }
  
  param_names <- parameters$names_variable
  model_ids <- names(oldModel[[1L]])
  # If the elite is older than the current iteration, it has its own model
  # that has evolved with time. If the elite is new (generated in the current
  # iteration), it does not have any, and we have to copy the one from its
  # parent. The condition of the IF statement is for checking whether the
  # configuration already has its model or not.
  elite_ids <- as.character(eliteConfigurations[[".ID."]])
  not_in <- elite_ids %not_in% model_ids
  ids_in_model <- elite_ids
  # If a configuration does not have any entry, copy the parent one.
  ids_in_model[not_in] <- as.character(eliteConfigurations[[".PARENT."]][not_in])
  newModel <- setNames(vector("list", length(param_names)), param_names)  
  
  for (currentParameter in param_names) {
    param <- parameters$get(currentParameter)
    irace_assert(all(ids_in_model %in% names(oldModel[[currentParameter]])))
    this_model <- oldModel[[currentParameter]][ids_in_model]
    values <- eliteConfigurations[[currentParameter]]
    values_not_na <- !is.na(values)
    values <- values[values_not_na]
    type <- param[["type"]]
    if (type == "c") {
      # Find the value that has been "chosen" to increase its probability.
      values <- match(values, param[["domain"]])
      this_model[values_not_na] <- mapply(update_prob, this_model[values_not_na], values, SIMPLIFY=FALSE)
    } else {
      irace_assert(type %in% c("i", "r", "o"))
      if (type == "o") 
        values <- match(values, param[["domain"]])
      this_model[values_not_na] <- mapply(function(p, value) c(p[[1L]] * num_factor, value),
                                          this_model[values_not_na], values, SIMPLIFY=FALSE)
    }
    names(this_model) <- elite_ids
    newModel[[currentParameter]] <- this_model
  }
  newModel
}

printModel <- function (model)
{
  cat("# Model:\n")
  print(model)
}

restartModel <- function(model, configurations, restart_ids, parameters,
                         nbConfigurations)
{
  back_factor <- nbConfigurations^(2 / parameters$nbVariable)
  second_factor <- (1 / nbConfigurations)^(1 / parameters$nbVariable)

  model_ids <- names(model[[1L]])
  restart_ids <- as.character(sort.int(as.integer(restart_ids)))
  not_in <- restart_ids %not_in% model_ids
  configurations <- configurations[configurations[[".ID."]] %in% restart_ids, c(".ID.", ".PARENT.")]
  restart_ids[not_in] <- configurations[[".PARENT."]][order(as.integer(configurations[[".ID."]]))][not_in]
  restart_ids <- as.character(unique(restart_ids))
  restart_ids <- restart_ids[!is.na(restart_ids)]
  
  for (pname in parameters$names_variable) {
    model_param <- model[[pname]]
    irace_assert (all(restart_ids %in% names(model_param)), {
      cat("Param:", pname, "\n")
      print(restart_ids)
      print(model)
      print(configurations[, c(".ID.", ".PARENT.")])
    })
    param <- parameters$get(pname)
    type <- param[["type"]]
    if (type == "c") {
      model[[pname]][restart_ids] <- sapply(model_param[restart_ids],
                                            function(p) {
                                              p <- 0.9 * p + 0.1 * max(p)
                                              p / sum(p)
                                            }, simplify=FALSE)
    } else {
      if (type == "i" || type == "r") {
        value <- init_sd_numeric(param)
      } else {
        irace_assert(type == "o")
        value <- (length(param[["domain"]]) - 1L) * 0.5
      }
      # Bring back the value 2 iterations or to the second iteration value.
      value <- value * second_factor
      model[[pname]][restart_ids] <- sapply(
        model_param[restart_ids],
        function(x) c(min(x[[1L]] * back_factor, value), x[[2L]]),
        simplify=FALSE)
    }
  }
  model
}

#' Get, set and restore the state of the random number generator state.
#'
#' @details These functions originate from the [`withr`][withr::with_seed()] package.
#' 
#' @param seed (`list()`|`integer(1)`)\cr Either an integer or the list returned gy [get_random_seed()].

#' @return [get_random_seed()] returns a list with two components `random_seed` and `rng_kind` or NULL if no seed was set; [set_random_seed()] and [restore_random_seed()] do not return anything.
#' @name random_seed
#' @examples
#' old_seed <- get_random_seed()
#' on.exit(restore_random_seed(old_seed))
#' set_random_seed(42)
#' value1 <- runif(1)
#' set_random_seed(42)
#' value2 <- runif(1)
#' stopifnot(all.equal(value1,value2))
#' 
#' @export
get_random_seed <- function()
{
  if (has_random_seed()) {
    return(list(random_seed = get(".Random.seed", globalenv(), mode = "integer", 
                                  inherits = FALSE), rng_kind = RNGkind()))
  }
  NULL
}

#' @rdname random_seed
#' @export
set_random_seed <- function(seed)
{
  if (is.list(seed)) {
    RNGkind(seed$rng_kind[[1L]], normal.kind = seed$rng_kind[[2L]],
            sample.kind = seed$rng_kind[[3L]])
    assign(".Random.seed", seed$random_seed, globalenv())
  } else {
    set.seed(seed)
  }
}

#' @rdname random_seed
#' @export
restore_random_seed <- function(seed)
{
  if (is.null(seed)) {
    rm_random_seed()
  } else {
    set_random_seed(seed)
  }
}

rm_random_seed <- function ()
{
  if (has_random_seed()) {
    set.seed(seed = NULL)
    rm(".Random.seed", envir = globalenv())
  }
}

has_random_seed <- function() 
{
  exists(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
}

# @return [gen_random_seeds()] returns a list of `n` seeds that were generated from the `global_seed`.
# The generated seeds can then e.g. be used to seed thread-local RNGs.
gen_random_seeds <- function(n, global_seed = NULL)
{
  # Use a random global seed if not set.
  if (!is.null(global_seed))
    set_random_seed(global_seed)

  # Generate 'n' seeds using the 'global_seed'.
  trunc(runif(n, 1, .Machine$integer.max))
}

# =========================================================================
# irace: An implementation in R of Iterated Race.
# -------------------------------------------------------------------------
#
#  Copyright (C) 2010-2025
#  Manuel López-Ibáñez     <manuel.lopez-ibanez@manchester.ac.uk>
#  Jérémie Dubois-Lacoste  <jeremie.dubois-lacoste@ulb.ac.be>
#  Leslie Perez Caceres    <leslie.perez.caceres@ulb.ac.be>
#
# -------------------------------------------------------------------------
#  This program is free software (software libre); you can redistribute
#  it and/or modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, you can obtain a copy of the GNU
#  General Public License at:
#                  http://www.gnu.org/copyleft/gpl.html
#  or by writing to the  Free Software Foundation, Inc., 59 Temple Place,
#                  Suite 330, Boston, MA 02111-1307 USA
# -------------------------------------------------------------------------
# $Revision$
# =========================================================================

#' irace_license
#'
#' A character string containing the license information of \pkg{irace}.
#'
#' @export
## __VERSION__ below will be replaced by the version defined in R/version.R
## This avoids constant conflicts within this file.
irace_license <-
'#------------------------------------------------------------------------------
# irace: An implementation in R of (Elitist) Iterated Racing
# Version: __VERSION__
# Copyright (C) 2010-2025
# Manuel Lopez-Ibanez     <manuel.lopez-ibanez@manchester.ac.uk>
# Jeremie Dubois-Lacoste
# Leslie Perez Caceres    <leslie.perez.caceres@ulb.ac.be>
#
# This is free software, and you are welcome to redistribute it under certain
# conditions.  See the GNU General Public License for details. There is NO
# WARRANTY; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# irace builds upon previous code from the race package:
#     race: Racing methods for the selection of the best
#     Copyright (C) 2003 Mauro Birattari
#------------------------------------------------------------------------------
'
cat_irace_license <- function()
  cat(sub("__VERSION__", irace_version, irace_license, fixed=TRUE))

#' Higher-level interface to launch irace.
#'
#' @inheritParams defaultScenario
#'
#' @param output.width `integer(1)`\cr The width used for the screen
#'   output.
#'
#' @details This function checks the correctness of the scenario, reads the
#'   parameter space from \code{scenario$parameterFile}, invokes [irace()],
#'   prints its results in various formatted ways, (optionally) calls
#'   [psRace()] and, finally, evaluates the best configurations on the test
#'   instances (if provided). If you want a lower-level interface that just
#'   runs irace, please see function [irace()].
#'
#' @templateVar return_invisible TRUE
#' @template return_irace
#' @seealso
#'  \describe{
#'  \item{[irace_cmdline()]}{a command-line interface to [irace()].}
#'  \item{[readScenario()]}{for reading a configuration scenario from a file.}
#'  \item{[readParameters()]}{read the target algorithm parameters from a file.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#' }
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @concept running
#' @export
irace_main <- function(scenario, output.width = 9999L)
  irace_common(scenario = scenario, simple=FALSE, output.width = output.width)

#' Test configurations given in the logfile (typically `irace.Rdata`) produced by \pkg{irace}.
#'
#' `testing_fromlog` executes the testing of the target algorithm configurations
#' found by an \pkg{irace} execution.
#'
#' @param logFile `character(1)`\cr Path to the logfile (typically `irace.Rdata`) produced by \pkg{irace}.
#'
#' @param testNbElites Number of (final) elite configurations to test. Overrides
#'   the value found in `logFile`.
#'
#' @param testIterationElites `logical(1)`\cr If `FALSE`, only the final
#'   `testNbElites` configurations are tested; otherwise, also test the best
#'   configurations of each iteration. Overrides the value found in `logFile`.
#'
#' @param testInstancesDir  Directory where testing instances are located, either absolute or relative to current directory.
#'
#' @param testInstancesFile File containing a list of test instances and optionally additional parameters for them.
#'
#' @param testInstances Character vector of the instances to be used in the `targetRunner` when executing the testing.
#'
#' @return `logical(1)`\cr `TRUE` if the testing ended successfully otherwise, `FALSE`.
#'
#' @details The function `testing_fromlog` loads the `logFile` and obtains the
#'   testing setup and configurations to be tested.  Within the `logFile`, the
#'   variable `scenario$testNbElites` specifies how many final elite
#'   configurations to test and `scenario$testIterationElites` indicates
#'   whether test the best configuration of each iteration. The values may be
#'   overridden by setting the corresponding arguments in this function.  The
#'   set of testing instances must appear in `scenario[["testInstances"]]`.
#'
#' @seealso [defaultScenario()] to provide a default scenario for \pkg{irace}.
#' [testing_fromfile()] provides a different interface for testing.
#'
#' @author Manuel López-Ibáñez and Leslie Pérez Cáceres
#' @concept running
#' @export
testing_fromlog <- function(logFile, testNbElites, testIterationElites,
                            testInstancesDir, testInstancesFile, testInstances)
{
  if (is.null.or.empty(logFile)) {
    irace_note("No logFile provided to perform the testing of configurations. Skipping testing.\n")
    return(FALSE)
  }
  iraceResults <- read_logfile(logFile)
  scenario <- iraceResults[["scenario"]]
  instances_changed <- FALSE

  if (!missing(testNbElites))
    scenario[["testNbElites"]] <- testNbElites
  if (!missing(testIterationElites))
    scenario$testIterationElites <- testIterationElites

  if (!missing(testInstances))
    scenario[["testInstances"]] <- testInstances

  if (!missing(testInstancesDir)) {
    scenario$testInstancesDir <- testInstancesDir
    instances_changed <- TRUE
  }
  if (!missing(testInstancesFile)) {
    scenario$testInstancesFile <- testInstancesFile
    instances_changed <- TRUE
  }

  cat("\n\n# Testing of elite configurations:", scenario$testNbElites,
      "\n# Testing iteration configurations:", scenario$testIterationElites,"\n")
  if (scenario$testNbElites <= 0)
    return (FALSE)

  # If they are already setup, don't change them.
  if (instances_changed || is.null.or.empty(scenario[["testInstances"]])) {
    scenario <- setup_test_instances(scenario)
    if (is.null.or.empty(scenario[["testInstances"]])) {
      irace_note("No test instances, skip testing\n")
      return(FALSE)
    }
  }

  # Get configurations that will be tested
  if (scenario$testIterationElites)
    testing_id <- sapply(iraceResults$allElites,
                         function(x) x[seq_len(min(length(x), scenario$testNbElites))])
  else {
    tmp <- iraceResults$allElites[[length(iraceResults$allElites)]]
    testing_id <- tmp[seq_len(min(length(tmp), scenario$testNbElites))]
  }
  testing_id <- unique.default(unlist(testing_id))
  configurations <- iraceResults$allConfigurations[testing_id, , drop=FALSE]

  irace_note ("Testing configurations (in no particular order): ", paste(testing_id, collapse=" "), "\n")
  testing_common(configurations, scenario, iraceResults)
  return(TRUE)
}

#' Test configurations given an explicit table of configurations and a scenario file
#'
#' Executes the testing of an explicit list of configurations given in
#' `filename` (same format as in [readConfigurationsFile()]). A `logFile` is
#' created unless disabled in `scenario`. This may overwrite an existing one!
#'
#' @param filename `character(1)`\cr Path to a file containing configurations: one configuration
#'   per line, one parameter per column, parameter names in header.
#'
#' @inheritParams defaultScenario
#'
#' @return iraceResults
#'
#' @seealso [testing_fromlog()] provides a different interface for testing.
#'
#' @author Manuel López-Ibáñez
#' @concept running
#' @export
testing_fromfile <- function(filename, scenario)
{
  irace_note ("Checking scenario.\n")
  scenario <- checkScenario(scenario)
  if (!scenario$quiet) printScenario(scenario)

  configurations <- readConfigurationsFile(filename, scenario$parameters)
  configurations <- cbind(.ID. = seq_nrow(configurations), configurations, .PARENT. = NA_integer_)
  rownames(configurations) <- configurations[[".ID."]]
  num <- nrow(configurations)
  configurations <- checkForbidden(configurations, scenario$parameters$forbidden)
  if (nrow(configurations) < num) {
    irace_warning("Some of the configurations in the configurations file were forbidden",
                  "and, thus, discarded.")
  }
  # To save the logs
  iraceResults <- list(scenario = scenario,
                       irace_version = irace_version,
                       allConfigurations = configurations)

  irace_note ("Testing configurations (in the order given as input): \n")
  testing_common(configurations, scenario, iraceResults)
}

testing_common <- function(configurations, scenario, iraceResults)
{
  verbose <- !scenario$quiet
  if (verbose) configurations_print(configurations)
  iraceResults$testing <- testConfigurations(configurations, scenario)
  save_irace_logfile(iraceResults, logfile = scenario$logFile)
  irace_note ("Testing results (column number is configuration ID in no particular order):\n")
  if (verbose) print(cbind(seeds = iraceResults$testing$seeds,
                           as.data.frame(iraceResults$testing$experiments)))
  irace_note ("Finished testing\n")
  iraceResults
}

#' Test that the given irace scenario can be run.
#'
#' Test that the given irace scenario can be run by checking the scenario
#' settings provided and trying to run the target-algorithm.
#'
#' @inheritParams defaultScenario
#'
#' @return returns `TRUE` if successful and gives an error and returns `FALSE`
#'   otherwise.
#'
#' @details If the `parameters` argument is missing, then the parameters
#'   will be read from the file `parameterFile`  given by `scenario`. If
#'   `parameters` is provided, then `parameterFile` will not be read.  This function will
#'   try to execute the target-algorithm.
#'
#' @seealso
#'  \describe{
#'  \item{\code{\link{readScenario}}}{for reading a configuration scenario from a file.}
#'  \item{\code{\link{printScenario}}}{prints the given scenario.}
#'  \item{\code{\link{defaultScenario}}}{returns the default scenario settings of \pkg{irace}.}
#'  \item{\code{\link{checkScenario}}}{to check that the scenario is valid.}
#' }
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
checkIraceScenario <- function(scenario)
{
  irace_note ("Checking scenario\n")
  scenario$debugLevel <- 2L
  scenario <- checkScenario(scenario)
  if (!scenario$quiet) printScenario(scenario)

  irace_note("Checking target runner.\n")
  if (checkTargetFiles(scenario = scenario)) {
    irace_note("Check successful.\n")
    return(TRUE)
  }
  irace_error("Check unsuccessful.\n")
  return(FALSE)
}

init <- function()
{
  irace_note("Initializing working directory...\n")
  libPath <- system.file(package = "irace")
  tmplFiles <- list.files(file.path(libPath, "templates"))
  for (file in tmplFiles) {
    if (grepl(".tmpl", file) && (file != "target-evaluator.tmpl")) {
      newFile <- gsub(".tmpl", "", file)
      if (file == "target-runner.tmpl" && .Platform$OS.type == 'windows') {
        file.copy(file.path(libPath, "templates", "windows", "target-runner.bat"), file.path(getwd(), "target-runner.bat"), overwrite = FALSE)
      } else {
        file.copy(file.path(libPath, "templates", file), file.path(getwd(), newFile), overwrite = FALSE)
      }
    }
  }
}


#' Launch `irace` with command-line options.
#'
#' Calls [irace_main()] using command-line options, maybe parsed from the
#' command line used to invoke R.
#'
#' @param argv `character()`\cr The arguments
#' provided on the R command line as a character vector, e.g.,
#' `c("--scenario", "scenario.txt", "-p", "parameters.txt")`.
#' Using the  default value (not providing the parameter) is the
#' easiest way to call [irace_cmdline()].
#'
#' @details The function reads the parameters given on the command line
#' used to invoke R, finds the name of the scenario file,
#'  initializes the scenario from the file (with the function
#'  [readScenario()]) and possibly from parameters passed in
#'  the command line. It finally starts \pkg{irace} by calling
#'  [irace_main()].
#'
#' List of command-line options:
#' ```{r echo=FALSE,comment=NA}
#' cmdline_usage(.irace.params.def)
#' ```
#'
#' @templateVar return_invisible TRUE
#' @template return_irace
#'
#' @seealso
#'  [irace_main()] to start \pkg{irace} with a given scenario.
#' @examples
#' irace_cmdline("--version")
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @concept running
#' @export
irace_cmdline <- function(argv = commandArgs(trailingOnly = TRUE))
{
  parser <- CommandArgsParser$new(argv = argv, argsdef = .irace.params.def)
  quiet <- !is.null(parser$readArg (short = "-q", long = "--quiet"))
  if (quiet) {
    op <- options(.irace.quiet = TRUE)
    on.exit(options(op))
  } else {
    cat_irace_license()
    cat("# installed at: ", system.file(package="irace"), "\n",
        "# called with: ", paste(argv, collapse = " "), "\n", sep = "")
  }
  if (!is.null(parser$readArg(short = "-h", long = "--help"))) {
    parser$cmdline_usage()
    return(invisible(NULL))
  }
  if (!is.null(parser$readArg(short = "-v", long = "--version"))) {
    print(utils::citation(package="irace"))
    return(invisible(NULL))
  }

  if (!is.null(parser$readArg(short = "-i", long = "--init"))) {
    init()
    return(invisible(NULL))
  }

  # Read the scenario file and the command line
  scenarioFile <- parser$readCmdLineParameter ("scenarioFile", default = "")
  scenario <- readScenario(scenarioFile)
  for (param in .irace.params.names) {
    scenario[[param]] <-
      parser$readCmdLineParameter(paramName = param, default = scenario[[param]])
  }
  if (quiet) scenario$quiet <- TRUE

  # Check scenario
  if (!is.null(parser$readArg (short = "-c", long = "--check"))) {
    checkIraceScenario(scenario)
    return(invisible(NULL))
  }

  # Only do testing
  testFile <- parser$readArg (long = "--only-test")
  if (!is.null(testFile)) {
    return(invisible(testing_fromfile(testFile, scenario)))
  }

  if (length(parser$argv))
    irace_error ("Unknown command-line options: ", paste(parser$argv, collapse = " "))

  irace_common(scenario = scenario, simple=FALSE)
}

#' @rdname irace_cmdline
#' @export
irace.cmdline <- function(argv = commandArgs(trailingOnly = TRUE))
{
  .Deprecated("irace.cmdline")
  irace_cmdline(argv = argv)
}

## Check targetRunner execution
checkTargetFiles <- function(scenario)
{
  ## Create two random configurations
  configurations <- sampleUniform(scenario$parameters, 2L,
    repair = scenario$repairConfiguration)
  set(configurations, j = ".ID.", value = seq_nrow(configurations))
  setcolorder(configurations, ".ID.", before = 1L)

  # Read initial configurations provided by the user.
  initConfigurations <- allConfigurationsInit(scenario)
  setDT(initConfigurations)
  if (nrow(initConfigurations) > 0L) {
    irace_assert(all(colnames(configurations) == colnames(initConfigurations)))
    configurations <- rbindlist(list(initConfigurations, configurations))
    set(configurations, j = ".ID.", value = seq_nrow(configurations))
  }

  bounds <- rep(scenario$boundMax, nrow(configurations))
  instances_ID <- if (scenario$sampleInstances)
                    sample.int(length(scenario$instances), 1L) else 1L
  setDF(configurations)
  experiments <- createExperimentList(
    configurations, scenario$parameters, instances = scenario$instances,
    instances_ID = instances_ID, seeds = 1234567L, bounds = bounds)

  race_state <- RaceState$new(scenario)
  race_state$start_parallel(scenario)
  on.exit(race_state$stop_parallel(), add = TRUE)
  # FIXME: Create a function try.call(err.msg,warn.msg, fun, ...)
  # Executing targetRunner
  irace_note("Executing targetRunner (", nrow(configurations), " times)...\n")
  result <- TRUE
  # We cannot let targetRunner or targetEvaluator modify our random seed, so we save it.
  withr::local_preserve_seed()
  output <-  withCallingHandlers(
    tryCatch(execute_experiments(race_state, experiments, scenario),
             error = function(e) {
               cat(sep = "\n",
                   "\n# Error occurred while executing targetRunner:",
                   paste0(conditionMessage(e), collapse="\n"))
               result <<- FALSE
               NULL
             }), warning = function(w) {
               cat(sep = "\n",
                   "\n# Warning occurred while executing targetRunner:",
                   paste0(conditionMessage(w), collapse="\n"))
               invokeRestart("muffleWarning")})

  if (scenario$debugLevel >= 1L) {
    cat("# targetRunner returned:\n")
    print(output, digits = 15L)
  }

  irace_assert(is.null(scenario$targetEvaluator) == is.null(race_state$target_evaluator))
  if (!result) return(FALSE)

  if (!is.null(scenario$targetEvaluator)) {
    irace_note("Executing targetEvaluator...\n")
    output <-  withCallingHandlers(
      tryCatch(execute_evaluator(race_state$target_evaluator, experiments, scenario, output),
                 error = function(e) {
                   cat(sep = "\n",
                       "\n# Error ocurred while executing targetEvaluator:",
                       paste0(conditionMessage(e), collapse="\n"))
                   result <<- FALSE
                   NULL
                 }), warning = function(w) {
                   cat(sep = "\n",
                       "\n# Warning ocurred while executing targetEvaluator:",
                       paste0(conditionMessage(w), collapse="\n"))
                   invokeRestart("muffleWarning")})
    if (scenario$debugLevel >= 1L) {
      cat("# targetEvaluator returned:\n")
      print(output, digits = 15L)
    }
  }
  result
}

withr::with_output_sink("test-target-runner-dummy.Rout", {
  skip_on_cran()

  get_executable <- function(filename) {
    filename <- paste0(filename, if (system_os_is_windows()) ".exe" else "")
    p <- if (.Platform$r_arch == "") file.path( "bin", filename) else file.path("bin", .Platform$r_arch, filename)
    system.file(p, package="irace", mustWork = FALSE)
  }
  runexe <- function(exe, args) {
    err <- NULL
    output <- withCallingHandlers(
      tryCatch(system2(exe, args, stdout = TRUE, stderr = TRUE),
               error = function(e) {
                 err <<- c(err, paste(conditionMessage(e), collapse="\n"))
                 NULL
               }), warning = function(w) {
                 err <<- c(err, paste(conditionMessage(w), collapse="\n"))
                 invokeRestart("muffleWarning")
               })
    if (!is.null(err)) {
      err <- paste(err, collapse = "\n")
      if (!is.null(attr(output, "errmsg")))
        err <- paste(sep = "\n", err, attr(output, "errmsg"))
      stop(err)
    }
    if (is.null(output))
      output <- ""
    return(output)
  }

  test_that("irace exe works", {
    iraceexe <- get_executable("irace")
    skip_if_not(nzchar(iraceexe), "Not run because 'irace' is not installed")
    expect_true(file.exists(iraceexe))
    # FIXME: For some reason, this does not generate any output on Windows
    output <- expect_silent(runexe(iraceexe, "--help"))
    ## cat("irace --help\n")
    ## print(output)
    expected_output <- if (system_os_is_windows()) "^$"
                       else "irace: An implementation in R.*called with: --help"
    expect_match(paste(collapse="", output), expected_output)
  })

  test_that("ablation exe works", {
    ablationexe <- get_executable("ablation")
    skip_if_not(nzchar(ablationexe), "Not run because 'ablation' is not installed")
    expect_true(file.exists(ablationexe))
    # FIXME: For some reason, this does not generate any output on Windows
    output <- expect_silent(runexe(ablationexe, "--help"))
    ## cat("ablation --help\n")
    ## print(output)
    expected_output <- if (system_os_is_windows()) "^$"
                       else "ablation: An implementation in R of Ablation Analysis.*called with: --help"
    expect_match(paste(collapse="", output), expected_output)
  })

  run_cmdline <- function(parameters, args) {
    parameters_file <- tempfile("dummy-parameters", fileext = ".txt")
    withr::local_file(parameters_file)
    cat(parameters, file=parameters_file)
    train_instances_file <- tempfile("dummy-train-instances", fileext = ".txt")
    withr::local_file(train_instances_file)
    cat("1\n", file=train_instances_file)
    irace_cmdline(paste0(args, ' --debug-level 3 --parameter-file=', parameters_file,
                         ' --train-instances-dir=  --train-instances-file=', train_instances_file,
                         ' --target-runner=', target_runner_dummy))
  }

  target_runner_dummy <- get_executable("target-runner-dummy")
  skip_if_not(nzchar(target_runner_dummy), "Not run because 'target-runner-dummy' is not installed")
  expect_true(file.exists(target_runner_dummy))

  test_that("--check", {
    expect_warning(
      run_cmdline(paste0('p_int  "--p_int " i (1, 10)\n',
                         'p_real "--p_real " r (1, 10)\n'),
                  '--check --max-experiments 500'),
      "No scenario file given")
  })
  test_that("--max-time", {
    expect_warning(
      run_cmdline(paste0('p_int  "--p_int " c (1)\n',
        'p_real "--p_real " r (0, 1)\n',
        'dummy1 "--dummy1 " r (0, 1)\n',
        'time   "--time "  c (1)\n'),
        '--max-time 2500'),
      "No scenario file given")
  })
  test_that("boundMax is too large", {
    expect_warning(
      expect_warning(
        run_cmdline(paste0('p_int   "--p_int "    i (1, 10)\n',
                           'p_real  "--p_real "   r (0, 1)\n',
                           'time    "--time "     c (1)\n',
                           'capping "--opt-time " c (1)\n'),
                    '--max-time 1000 --bound-max 100 --capping 1'),
        "is too large"),
      "No scenario file given")
  })
  test_that("--capping", {
    expect_warning(
      expect_warning(
        run_cmdline(paste0('p_int   "--p_int "    i (1, 10)\n',
                           'p_real  "--p_real "   r (1, 10)\n',
                           'dummy1 "--dummy1 " r (1, 10)\n',
                           'dummy2 "--dummy2 " r (1, 10)\n',
                           'dummy3 "--dummy3 " r (1, 10)\n',
                           'dummy4 "--dummy4 " r (1, 10)\n',
                           'dummy5 "--dummy5 " r (1, 10)\n',
                           'dummy6 "--dummy6 " r (1, 10)\n',
                           'dummy7 "--dummy7 " r (1, 10)\n',
                           'dummy8 "--dummy8 " r (1, 10)\n',
                           'dummy9 "--dummy9 " r (1, 10)\n',
                           'dummy11 "--dummy11 " r (1, 10)\n',
                           'dummy12 "--dummy12 " r (1, 10)\n',
                           'dummy13 "--dummy13 " r (1, 10)\n',
                           'dummy14 "--dummy14 " r (1, 10)\n',
                           'dummy15 "--dummy15 " r (1, 10)\n',
                           'dummy16 "--dummy16 " r (1, 10)\n',
                           'dummy17 "--dummy17 " r (1, 10)\n',
                           'dummy18 "--dummy18 " r (1, 10)\n',
                           'dummy19 "--dummy19 " r (1, 10)\n',
                           'time    "--time "     c (1)\n',
                           'capping "--opt-time " c (1)\n'),
                    '--max-time 500 --bound-max 1 --capping 1 '),
        "With the current settings and estimated time per run"),
      "No scenario file given")
  })
  test_that("Error cost time", {
    expect_error(
      expect_warning(
        run_cmdline(paste0('p_int   "--p_int "    i (1, 10)\n',
                           'p_real  "--p_real "   r (1, 10)\n'),
                    '--max-time 100 '),
        "No scenario file given"),
      "The output of targetRunner must be two numbers 'cost time'",
      fixed = TRUE)
  })
  test_that("low --max-experiments", {
    expect_warning(
    expect_error(
      run_cmdline(paste0('p_int  "--p_int " i (1, 10)\n',
                         'p_real "--p_real " r (1, 10)\n',
                         'time   "--time "  c (1)\n'),
                  '--max-experiments 50'),
        "With the current settings"),
      "No scenario file given")
  })
  test_that("--min-experiments", {
    expect_no_warning(
    expect_warning(
      run_cmdline(paste0('p_int  "--p_int " i (1, 10)\n',
                         'p_real "--p_real " r (1, 10)\n',
                         'time   "--time "  c (1)\n'),
                  '--min-experiments 50'),
      "No scenario file given"))
  })

}) # withr::with_output_sink()

withr::with_output_sink("test-maxTime.Rout", {

target_runner <- function(experiment, scenario)
{
  configuration     <- experiment$configuration
  tmax <-  configuration[["tmax"]]
  temp <-  configuration[["temp"]]
  time <- max(1, abs(rnorm(1, mean=(tmax+temp)/10)))
  list(cost = time, time = time)
}

time_irace <- function(...)
{
  args <- list(...)
  parameters <- readParameters(text = '
   tmax "" i (1, 50)
   temp "" r (0, 10)
   dummy "" c ("dummy")
   ')
  scenario <- list(targetRunner = target_runner,
                   instances = 1:10,
                   testInstances = 11:20,
                   seed = 1234567,
                   parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)

  expect_true(irace:::checkTargetFiles(scenario = scenario))
  
  confs <- irace(scenario = scenario)
  final_ids <- sort(as.character(confs$.ID.[1:scenario$testNbElites]))
  expect_gt(nrow(confs), 0L)
  testing_fromlog(scenario$logFile)
  iraceResults <- read_logfile(scenario$logFile)
  if (scenario$testIterationElites) {
    # FIXME: We could test here that the correct configurations are tested.
    expect_gte(ncol(iraceResults$testing$experiments), scenario$testNbElites)
  } else {
    test_ids <- sort(colnames(iraceResults$testing$experiments))
    expect_equal(final_ids, test_ids)
  }
  confs
}


test_that("maxTime=500 testNbElites=2 testIterationElites=FALSE", {
  generate_set_seed()
  time_irace(maxTime = 500, testNbElites=2)
})

test_that("maxTime=1111 testNbElites=3 testIterationElites=TRUE", {
  skip_on_cran()
  generate_set_seed()
  time_irace(maxTime = 1111, testNbElites=3, testIterationElites=TRUE)
})

}) # withr::with_output_sink()

withr::with_output_sink("test-bad_scenario.Rout", {
  test_that("bad scenario: unknown variables", {
    skip_on_cran()
    bad_scenario <- test_path("bad_scenario.txt")
    expect_error(irace_cmdline(paste0("--scenario ", bad_scenario)), "unknown variables")
    good_scenario <- test_path("good_scenario.txt")
    expect_error(irace_cmdline(paste0("--scenario ", good_scenario, " --unknown")), "Unknown command-line options: --unknown")
  })
})

.ablation.params.def <- utils::read.table(header=TRUE, stringsAsFactors = FALSE, text="
name            ab  type short long          default               description
iraceResults     0  p    -l    --log-file    NA                    'Path to the (.Rdata) file created by irace from which the  \"iraceResults\" object will be loaded.'
src              1  i    -S    --src         1                     'Source configuration ID or the path to a file containing the configuration.'
target           1  i    -T    --target      NA                    'Target configuration ID (by default the best configuration found by irace) or the path to a file containing the configuration.'
ab_params        1  s    -P    --params      ''                    'Specific parameter names to be used for the ablation (separated with commas). By default use all'
type             1  s    -t    --type        'full'                'Type of ablation to perform: \"full\" will execute each configuration on all \"--n-instances\" to determine the best-performing one; \"racing\" will apply racing to find the best configurations.'
nrep            1  i    -n    --nrep       1                     'Number of replications per instance used in \"full\" ablation.'
seed             1  i    ''    --seed        1234567               'Integer value to use as seed for the random number generation.'
ablationLogFile  1  p    -o    --output-file 'log-ablation.Rdata'  'Log file to save the ablation log. If \"\", the results are not saved to a file.'
instancesFile    1  s    ''    --instances-file   'train'          'Instances file used for ablation: \"train\", \"test\" or a filename containing the list of instances.'
plot             0  s    -p    --plot        ''                    'Output filename (.pdf) for the plot. If not given, no plot is created.'
plot_type        0  s    -O    --plot-type   'mean'                'Type of plot. Supported values are \"mean\", \"boxplot\", \"rank\" or \"rank,boxplot\".'
old_path         0  p    ''    --old-path    NA                    'Old path found in the log-file (.Rdata) given as input to be replaced by --new-path.'
new_path         0  p    ''    --new-path    NA                    'New path to replace the path found in the log-file (.Rdata) given as input.'
execDir          0  p    -e    --exec-dir    NA                    'Directory where the target runner will be run.'
scenarioFile     0  p    -s    --scenario    NA                    'Scenario file to override the scenario given in the log-file (.Rdata)'
parallel         0  i    ''    --parallel    NA                    'Number of calls to targetRunner to execute in parallel. Values 0 or 1 mean no parallelization.'
")

## __VERSION__ below will be replaced by the version defined in R/version.R
## This avoids constant conflicts within this file.
cat_ablation_license <- function()
{
  ablation_license <-
'#------------------------------------------------------------------------------
# ablation: An implementation in R of Ablation Analysis
# Version: __VERSION__
# Copyright (C) 2020--2025
# Manuel Lopez-Ibanez     <manuel.lopez-ibanez@manchester.ac.uk>
# Leslie Perez Caceres    <leslie.perez.caceres@ulb.ac.be>
#
# This is free software, and you are welcome to redistribute it under certain
# conditions.  See the GNU General Public License for details. There is NO
# WARRANTY; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------
'
  cat(sub("__VERSION__", irace_version, ablation_license, fixed=TRUE))
}

#' Launch ablation with command-line options.
#'
#' Launch [ablation()] with the same command-line options as the command-line
#' executable (`ablation.exe` in Windows).
#'
#' @param argv `character()`\cr The arguments
#' provided on the R command line as a character vector, e.g.,
#' `c("-i", "irace.Rdata", "--src", 1)`.
#'
#' @details The function reads the parameters given on the command line
#' used to invoke R, launches [ablation()] and possibly [plotAblation()].
#'
#' List of command-line options:
#' ```{r echo=FALSE,comment=NA}
#' cmdline_usage(.ablation.params.def)
#' ```
#' @return A list containing the following elements:
#'  \describe{
#'    \item{allConfigurations}{Configurations tested in the ablation.}
#'    \item{state}{State of the ablation process.}
#'    \item{experiments}{A matrix with the results of the experiments (columns are configurations, rows are instances).}
#'    \item{scenario}{Scenario object with the settings used for the experiments.}
#'    \item{trajectory}{IDs of the best configurations at each step of the ablation.}
#'    \item{best}{Best configuration found in the experiments.}
#'    \item{complete}{`TRUE` if the ablation process was completed.}
#'  }
#'
#' @seealso [plotAblation()] [ablation()]
#' @examples
#' ablation_cmdline("--help")
#' # Find the ablation command-line executable:
#' Sys.glob(file.path(system.file(package="irace", "bin"), "ablation*"))
#' @author Manuel López-Ibáñez
#' @concept ablation
#' @export
ablation_cmdline <- function(argv = commandArgs(trailingOnly = TRUE))
{
  withr::local_options(list(width = 9999L)) # Do not wrap the output.

  cat_ablation_license()
  cat ("# installed at: ", system.file(package="irace"), "\n",
       "# called with: ", paste(argv, collapse = " "), "\n", sep = "")
  parser <- CommandArgsParser$new(argv = argv, argsdef = .ablation.params.def)
  if (!is.null(parser$readArg (short = "-h", long = "--help"))) {
    parser$cmdline_usage()
    return(invisible(NULL))
  }

  if (!is.null(parser$readArg (short = "-v", long = "--version"))) {
    print(utils::citation(package="irace"))
    return(invisible(NULL))
  }

  params <- parser$readAll()
  # TODO: Send the other options to the irace command-line parser so the user
  # can override options in scenario via the command-line.
  if (length(parser$argv) > 0L)
    stop("Unknown command-line options: ", paste(parser$argv, collapse = " "))

  if (is.null(params$iraceResults)) {
    irace_error("You must provide the path to the '.Rdata' file that contains the 'iraceResults' object generated by irace.")
    return(invisible(NULL))
  }
  iraceResults <- read_logfile(params$iraceResults)
  if (is.null(params$old_path) != is.null(params$new_path)) {
    irace_error("To update paths you must provide both --old-path and --new-path.")
    return(invisible(NULL))
  } else if (!is.null(params$old_path)) {
    iraceResults$scenario <- scenario_update_paths(iraceResults$scenario, params$old_path, params$new_path)
  }

  scenario <- list()
  if (!is.null(params$scenarioFile)) {
    scenario <- readScenario(params$scenarioFile)
    # We do not want this seed value to override the the command-line.
    scenario$seed <- NULL
  }
  for (p in c("execDir", "parallel")) {
    if (!is.null(params[[p]]))
      scenario[[p]] <- params[[p]]
  }

  if (is_null_or_empty_or_na(trim(params$ablationLogFile))) {
    params$ablationLogFile <- NULL
  } else {
    params$ablationLogFile <- path_rel2abs(params$ablationLogFile)
  }

  if (!is.null(params$ab_params))
    params$ab_params <- trimws(strsplit(params$ab_params, ",", fixed=TRUE)[[1L]])

  # The shell may introduce extra quotes, remove them.
  params$plot_type <- trimws(gsub("[\"']", "", params$plot_type))

  # We want to select elements that actually appear in params, otherwise we get NA names.
  ablation_params <- intersect(.ablation.params.def[.ablation.params.def$ab == 1, "name", drop=TRUE],
                               names(params))
  ablog <- do.call(ablation,
                   args = c(list(iraceResults = iraceResults),
                            params[ablation_params],
                            scenario))
  if (!is.null(params[["plot"]]) || base::interactive()) {
    plotAblation(ablog, pdf_file = params[["plot"]], type = params$plot_type)
  }
  invisible(ablog)
}

fix_dependents <- function(parameters, dep_names, new_conf, final_conf, changed)
{
  # ... check every dependent parameter...
  for (dep  in dep_names) {
    # If the conditions are satisfied and it doesn't have a value but it
    # should, then change it.  If the conditions are not satisfied and it
    # has a value but it shouldn't, then change it.
    ok <- conditionsSatisfied(parameters$conditions[[dep]], new_conf)
    change <- ok == is.na(new_conf[[dep]])
    dep_param <- parameters$get(dep)
    if (change) {
      # If it doesn't have a value but it should, take it from final.
      if (ok) {
        if (dep_param[["is_dependent"]]) {
          # If the parameter is domain-dependent, then:
          if (anyNA(new_conf[, .SD, .SDcols=all.vars(dep_param$domain)])) {
            # If it depends on a parameter that is disabled, then this is disabled, so no need to change.
            next
          } else if (is.na(final_conf[[dep]]) # If the final value is NA, we cannot fix it, so skip it.
            # If parameter X enables dep but dep has domain that depends on
            # parameter Y, then Y in new_conf may have a value that makes
            # final_conf[[dep]] invalid. We cannot fix this, so skip.
            || !is_within_dependent_bound(dep_param, new_conf, value = final_conf[[dep]]))
            return(NULL)
        }
        set(new_conf, j = dep, value = final_conf[[dep]])
      } else {
        set(new_conf, j = dep, value = NA)
      }
      changed <- c(changed, dep)
    } else if (dep_param[["is_dependent"]] && !is.na(new_conf[[dep]])) {
      # The condition may not have changed but the domain may depend on a parameter that has been changed.
      if (anyNA(new_conf[, .SD, .SDcols=all.vars(dep_param$domain)])) {
        set(new_conf, j = dep, value = NA)
        changed <- c(changed, dep)
      } else if (!is_within_dependent_bound(dep_param, new_conf, value = new_conf[[dep]])) {
        # If the current value is not within bound and the final value is NA, we cannot fix it, so skip it.
        if (is.na(final_conf[[dep]])
          # If parameter X enables dep but dep has domain that depends on
          # parameter Y, then Y in new_conf may have a value that makes
          # final_conf[[dep]] invalid. We cannot fix this, so skip.
          || !is_within_dependent_bound(dep_param, new_conf, value = final_conf[[dep]]))
          return(NULL)
        set(new_conf, j = dep, value = final_conf[[dep]])
        changed <- c(changed, dep)
      }
    }
  }
  changed
}

## Function that generates the configurations of the ablation path
## between initial_configuration and final_configuration.
## parameters can be selected by specifying them in para.names.
generate_ablation <- function(initial_configuration, final_configuration,
                              parameters, param_names)
{
  # Only change variable parameters
  param_names <- setdiff(param_names, parameters$names_fixed)
  # Follow the hierarchy order.
  hierarchy <- sort(parameters$hierarchy)
  param_names <- intersect(names(parameters$hierarchy), param_names)
  configurations <- list()
  changed_params <- list()
  # skip if the value is the same or NA (not active)
  skip <- (initial_configuration[param_names] == final_configuration[param_names])
  skip <- skip[1L, ] # Drop to a vector with names
  skip <- skip | is.na(skip)
  for (pname in param_names) {
    # If parameter is not active, it will only change if its condition becomes true.
    if (skip[[pname]]) next
    param <- parameters$get(pname)
    # If the parameter is domain-dependent, then we can only change its
    # value if the new value is within the domain.
    if (param[["is_dependent"]]
      && !is_within_dependent_bound(param, initial_configuration, value = final_configuration[[pname]]))
      next
    new_configuration <- as.data.table(initial_configuration)
    set(new_configuration, j = pname, value = final_configuration[[pname]])
    changed <- pname
    # If other parameters depend on this one then...
    if (any(sapply(parameters$depends, function(x) any(pname %in% x)))) {
      changed <- fix_dependents(parameters, dep_names = names(hierarchy)[hierarchy > hierarchy[[pname]]],
        new_conf = new_configuration, final_conf = final_configuration, changed = changed)
      if (is.null(changed))
        next
    }
    new_configuration <- filter_forbidden(new_configuration, parameters$forbidden)
    if (nrow(new_configuration) == 0L) next
    changed_params <- c(changed_params, list(changed))
    configurations <- c(configurations, list(new_configuration))
  }
  configurations <- rbindlist(configurations)
  set(configurations, j = ".PARENT.", value = configurations[[".ID."]])
  setDF(configurations)
  rownames(configurations) <- NULL
  list(configurations=configurations, changed_params=changed_params)
}

report_duplicated_results <- function(experiments, configurations)
{
  x <- t(experiments)
  x <- x[duplicated(x) | duplicated(x, fromLast = TRUE), , drop=FALSE]
  if (nrow(x) == 0L) return(NULL) # No duplicates
  dups <- split(rownames(x), apply(x, 1L, paste0, collapse=""))
  names(dups) <- NULL
  for (g in dups) {
    cat("Warning: The following configurations are different but produced the same results:\n")
    df <- configurations[configurations[[".ID."]] %in% g, , drop=FALSE]
    print(df)
    cat("Parameters with different values from the above configurations:\n")
    df <- removeConfigurationsMetaData(df)
    print(df[, vapply(df, function(x) length(unique(x)) > 1L, logical(1L)), drop=FALSE])
  }
  dups
}

ab_generate_instances <- function(race_state, scenario, nrep, type, instancesFile)
{
  nrep <- suppressWarnings(as.integer(nrep))
  if (is.na(nrep) || length(nrep) == 0L || nrep <= 0L)
    stop("'nrep' must be an integer larger than zero")
  if (nrep != 1L && type == "racing")
    stop("'nrep' has no effect when type == 'racing'")
  if (nrep > 1L && scenario$deterministic)
    stop("'nrep > 1' does not make sense with a deterministic scenario")

  if (instancesFile == "test") {
    scenario$instances <- scenario$testInstances
  } else if (instancesFile != "train") {
    scenario$instances <- readInstances(instancesFile = path_rel2abs(instancesFile))
  }
  n_inst <- length(scenario$instances)
  if (type == "full" && n_inst * nrep == 1L)
    stop("'nrep' must be larger than 1 when type == 'full' and a single instance")
  generateInstances(race_state, scenario, n_inst * nrep)

  msg <- if (instancesFile %in% c("train", "test"))
           paste0("'", instancesFile, "' instances") else paste0("instances from '", instancesFile, "'")
  if (n_inst > 100L) {
    n_inst <- 100L
    msg <- paste0(msg, " (only showing first 100)")
  }
  cat(sep="", "# Using ", msg, ":\n",
      paste0(collapse="\n", scenario$instances[1L:n_inst]),
      "\n")
  scenario$instances
}

#' Performs ablation between two configurations (from source to target).
#'
#' @description Ablation is a method for analyzing the differences between two configurations.
#'
#' @inheritParams has_testing_data
#' @param src,target `integer(1)|character(1)`\cr Source and target configuration IDs. By default, the first configuration ever evaluated (ID 1) is used as `src` and the best configuration found by irace is used as target. If the argument is a string, it is interpreted as the path to a file, with the format specified by  [readConfigurationsFile()], that contains the configuration.
#' @param ab_params `character()`\cr Specific parameter names to be used for the ablation. They must be in `parameters$names`. By default, use all parameters.
#' @param type `"full"|"racing"`\cr Type of ablation to perform: `"full"` will execute each configuration on all `n_instances` to determine the best-performing one; `"racing"` will apply racing to find the best configurations.
#' @param nrep `integer(1)`\cr Number of replications per instance used in `"full"` ablation. When `nrep > 1`, each configuration will be executed `nrep` times on each instance with different random seeds.
#' @param seed `integer(1)`\cr Integer value to use as seed for the random number generation.
#' @param ablationLogFile  `character(1)`\cr Log file to save the ablation log. If `NULL`, the results are not saved to a file.
#' @param instancesFile  `character(1)`\cr Instances file used for ablation: `'train'`, `'test'` or a filename containing the list of instances.
#' @param ... Further arguments to override scenario settings, e.g., `debugLevel`, `parallel`, etc.
#'
#' @references
#' C. Fawcett and H. H. Hoos. Analysing differences between algorithm
#' configurations through ablation. Journal of Heuristics, 22(4):431–458, 2016.
#'
#' @inherit ablation_cmdline return
#' @seealso [plotAblation()] [ablation_cmdline()]
#' @examples
#' \donttest{
#' logfile <- system.file(package="irace", "exdata", "sann.rda")
#' # Execute ablation between the first and the best configuration found by irace.
#' ablog <- ablation(logfile, ablationLogFile = NULL)
#' plotAblation(ablog)
#' # Execute ablation between two selected configurations, and selecting only a
#' # subset of parameters, directly reading the setup from the irace log file.
#' ablog <- ablation(logfile, src = 1, target = 10,
#'                   ab_params = c("temp"), ablationLogFile = NULL)
#' plotAblation(ablog)
#' }
#'
#' @author Leslie Pérez Cáceres and Manuel López-Ibáñez
#' @concept ablation
#' @export
ablation <- function(iraceResults, src = 1L, target = NULL,
                     ab_params = NULL, type = c("full", "racing"),
                     nrep = 1L, seed = 1234567L,
                     ablationLogFile = "log-ablation.Rdata",
                     instancesFile="train", ...)
{
  # Input check
  if (missing(iraceResults) || is.null(iraceResults))
    stop("You must provide an 'iraceResults' object generated by irace or the path to the '.Rdata' file that contains this object.")

  type <- match.arg(type)

  if (!is.null(ablationLogFile)) {
    file.check(ablationLogFile, writeable = TRUE, text = 'ablationLogFile')
  }

  save_ablog <- function(complete) {
    ablog <- list(changes = changes,
      # This name must match the one used in the irace log.
      allConfigurations = as.data.frame(all_configurations),
      experiments = experiments,
      state = race_state,
      scenario = scenario,
      trajectory = trajectory,
      best = best_configuration,
      complete = complete)
    if (!is.null(ablationLogFile))
      save(ablog, file = ablationLogFile, version = 3L)
    invisible(ablog)
  }

  # Load the data of the log file.
  iraceResults <- read_logfile(iraceResults)
  log_version <- get_log_clean_version(iraceResults)
  if (log_version < "3.9.0")
    irace_error("The version of the logfile (", log_version, ") is too old for this version of ablation")

  if (is.null(iraceResults$state$completed) || length(iraceResults$state$completed) != 1L
    || iraceResults$state$completed == "Incomplete")
    stop("The 'iraceResults' logfile seems to belong to an incomplete run of irace.")
  scenario <- update_scenario(scenario = iraceResults$scenario, ...)
  scenario$logFile <- ""
  old_seed <- get_random_seed()
  on.exit(restore_random_seed(old_seed))
  # FIXME: We need to overwrite the seed because RaceState takes it from the scenario and calls set_random_seed().
  scenario$seed <- seed
  scenario <- checkScenario(scenario)
  race_state <- RaceState$new(scenario)
  # Generate instances
  scenario$instances <- ab_generate_instances(race_state, scenario, nrep, type, instancesFile)
  parameters <- scenario$parameters

  # Process src and target.
  if (is.character(src) && is.na(suppressWarnings(as.integer(src)))) {
    src_configuration <- readConfigurationsFile(src, parameters)
    if (nrow(src_configuration) != 1L) {
      stop("Argument of src=\"", src, "\" contains more than one configuration!")
    }
    src <- 1L + max(iraceResults$allConfigurations$.ID.)
    src_configuration$.ID. <- src
    src_configuration$.PARENT. <- NA_integer_
    iraceResults$allConfigurations <- rbind.data.frame(iraceResults$allConfigurations, src_configuration)
    # FIXME: Check for duplicates
  } else if (src %not_in% iraceResults$allConfigurations[[".ID."]])
    stop("Source configuration ID (", src, ") cannot be found!")

  if (is.null(target))
    target <- iraceResults$iterationElites[length(iraceResults$iterationElites)]
  else if (is.character(target) && is.na(suppressWarnings(as.integer(target)))) {
    target_configuration <- readConfigurationsFile(target, parameters)
    if (nrow(target_configuration) != 1L) {
      stop("Argument of target=\"", target, "\" contains more than one configuration!")
    }
    target <- 1L + max(iraceResults$allConfigurations$.ID.)
    target_configuration$.ID. <- target
    target_configuration$.PARENT. <- NA_integer_
    iraceResults$allConfigurations <- rbind.data.frame(iraceResults$allConfigurations, target_configuration)
    # FIXME: Check for duplicates
  } else if (target %not_in% iraceResults$allConfigurations[[".ID."]])
    stop("Target configuration ID (", target, ") cannot be found!")

  if (src == target)
    stop("Source and target configuration IDs must be different!")

  irace_note ("Starting ablation from ", src, " to ", target, "\n# Seed: ", race_state$seed, "\n")
  cat("# Source configuration (row number is ID):\n")
  src_configuration <- iraceResults$allConfigurations[src, , drop = FALSE]
  configurations_print(src_configuration)
  cat("# Target configuration (row number is ID):\n")
  target_configuration <- iraceResults$allConfigurations[target, , drop = FALSE]
  configurations_print(target_configuration)

  # Select the parameters used for ablation
  if (is.null(ab_params)) {
    ab_params <- parameters$names
  } else if (!all(ab_params %in% parameters$names)) {
    irace_error("Some of the parameters provided (", paste0(setdiff(ab_params, parameters$names), collapse=", "),
                ") are not defined in the parameter space.")
  }
  # Select parameters that are different in both configurations
  neq_params <- which(src_configuration[,ab_params] != target_configuration[,ab_params])
  if (length(neq_params) == 0L)
    irace_error("src and target configurations are equal considering the parameters selected.\n")
  param_names <- colnames(src_configuration[,ab_params])[neq_params]

  # FIXME: Do we really need to override the ID?
  src_configuration[[".ID."]] <- 1L
  target_configuration[[".ID."]] <- 2L
  all_configurations <- rbindlist(list(src_configuration, target_configuration), use.names=TRUE)

  race_state$start_parallel(scenario)
  on.exit(race_state$stop_parallel(), add = TRUE)

  # Execute source and target configurations.
  ## FIXME: We may already have these experiments in the logFile!
  irace_note("Executing source and target configurations on the given instances * nrep (", nrow(race_state$instances_log), ")...\n")
  experiments <- do_experiments(race_state,
    configurations = rbind.data.frame(src_configuration, target_configuration),
    ninstances = nrow(race_state$instances_log), scenario = scenario, iteration = 0L)
  irace_assert(ncol(experiments) == 2L)
  step <- 1L
  # Define variables needed
  trajectory <- 1L
  names(trajectory) <- "source"
  # FIXME: changes should only store the changed parameters.
  changes <- list()
  best_configuration <- src_configuration
  save_ablog(complete = FALSE)
  while (length(param_names) > 1L) {
    # Generate ablation configurations
    cat("# Generating configurations (row number is ID):", param_names, "\n")
    ab_aux <- generate_ablation(best_configuration, target_configuration, parameters,
      param_names)
    aconfigurations <- ab_aux$configurations
    if (nrow(aconfigurations) == 0L) {
      cat("# Stopping ablation, no parameter change possible.\n")
      break
    }
    ## FIXME: We may already have these configurations in the logFile!
    # New configurations IDs
    aconfigurations[[".ID."]] <- max(all_configurations[[".ID."]]) + seq_nrow(aconfigurations)
    configurations_print(aconfigurations, metadata = FALSE)
    all_configurations <- rbindlist(list(all_configurations, aconfigurations), use.names=TRUE)

    irace_note("Ablation (", type, ") of ", nrow(aconfigurations),
      " configurations on ", nrow(race_state$instances_log), " instances (this may take a while ...).\n")
    if (type == "full") {
      step_experiments <- do_experiments(race_state,
        configurations = aconfigurations,
        ninstances = nrow(race_state$instances_log),
        scenario = scenario, iteration = step)
    } else {
      # Set variables for the racing procedure
      # FIXME: what about blockSize?
      race_state$next_instance <- 1L
      race_output <- elitist_race(race_state,
        maxExp = nrow(aconfigurations) * nrow(race_state$instances_log),
        minSurvival = 1L,
        elite_data = NULL,
        configurations = aconfigurations,
        scenario = scenario,
        elitist_new_instances = 0L)
      set(race_output$experiment_log, j = "iteration", value = step)
      race_state$experiment_log <-
        rbindlist(list(race_state$experiment_log, race_output$experiment_log),
          use.names=TRUE)
      step_experiments <- race_output$experiments
    }
    experiments <- merge_matrix(experiments, step_experiments)

    save_ablog(complete = FALSE)
    # Get the best configuration based on the criterion of irace
    cranks <- overall_ranks(experiments[, aconfigurations[[".ID."]], drop=FALSE], test = scenario$testType)
    best_id <- which.min(cranks)[1L]
    changes[[step]] <- ab_aux$changed_params
    best_change <- changes[[step]][[best_id]]
    trajectory <- c(trajectory, aconfigurations[[".ID."]][best_id])

    # Report best.
    cat("# Best changed parameters:\n")
    for (i in seq_along(best_change)) {
      cat("#", best_change[i], ":", best_configuration[,best_change[i]], "->",
          aconfigurations[best_id, best_change[i]], "\n")
    }
    best_configuration <- aconfigurations[best_id,,drop=FALSE]
    best_id <- best_configuration[[".ID."]]
    param_names <- param_names[param_names %not_in% best_change]
    step <- step + 1L
  }

  trajectory <- c(trajectory, target_configuration[[".ID."]])
  # Get the overall best
  cranks <- overall_ranks(experiments[,trajectory, drop=FALSE], test = scenario$testType)
  best_id <- which.min(cranks)[1L]
  setDF(all_configurations, rownames = all_configurations[[".ID."]])
  best_configuration <- all_configurations[trajectory[best_id], , drop=FALSE]
  irace_note("Final best configuration:\n")
  configurations_print(best_configuration)

  # Check for duplicated results:
  report_duplicated_results(experiments, all_configurations)

  save_ablog(complete = TRUE)
}

ablation_labels <- function(trajectory, configurations)
{
  configurations <- removeConfigurationsMetaData(configurations[trajectory, , drop = FALSE])
  labels <- names(trajectory)
  last <- configurations[1L, , drop = FALSE]
  param_names <- colnames(last)
  for (i in 2L:length(trajectory)) {
    current <- configurations[i, , drop = FALSE]
    # Select everything that is NOT NA now and was different or NA before.
    select <- !is.na(current) & (is.na(last) | (current != last))
    irace_assert(!anyNA(select))
    labels[i] <- paste0(param_names[select], "=", current[, select], collapse = "\n")
    last <- current
  }
  labels
}

#' Create plot from an ablation log
#'
#' @param ablog (`list()`|`character(1)`) Ablation log object returned by [ablation()]. Alternatively, the path to an `.Rdata` file, e.g., `"log-ablation.Rdata"`, from which the object will be loaded.

#' @param pdf_file Output filename.
#' @param width Width provided to create the PDF file.
#' @param height Height provided to create the PDF file.
#' @param type Type of plot. Supported values are `"mean"` and `"boxplot"`. Adding `"rank"` will plot rank per instance instead of raw cost value.
#' @param n `integer(1)`\cr Number of parameters included in the plot. By default all parameters are included.
#' @param mar Vector with the margins for the ablation plot.
#' @param ylab Label of y-axis.
#' @param ylim Numeric vector of length 2 giving the y-axis range.
#' @param rename_labels `character()`\cr Renaming table for nicer labels. For example, `c("No value"="NA", "LongParameterName"="LPN")`.
#' @param ... Further graphical parameters may also be supplied as
#'   arguments. See [graphics::plot.default()].
#'
#' @author Leslie Pérez Cáceres and Manuel López-Ibáñez
#' @seealso [ablation()] [ablation_cmdline()]
#' @examples
#' logfile <- file.path(system.file(package="irace"), "exdata", "log-ablation.Rdata")
#' plotAblation(ablog = logfile)
#' plotAblation(ablog = logfile, type = "mean")
#' plotAblation(ablog = logfile, type = c("rank","boxplot"), rename_labels = c(
#'             "localsearch"="ls", algorithm="algo", source="default"))
#' @concept ablation
#' @export
plotAblation <- function (ablog, pdf_file = NULL, width = 20,
                          height = 7,
                          type = c("mean", "boxplot", "rank"), n = 0L,
                          mar = NULL,
                          ylab = "Mean configuration cost", ylim = NULL,
                          rename_labels = NULL,
                          ...)
{
  type <- trimws(unlist(strsplit(type, ",", fixed=TRUE)))
  type <- match.arg(type, several.ok = TRUE)
  if (missing(ylab) && ("rank" %in% type)) ylab <- "Rank per instance"

  if (missing(ablog) || is.null(ablog)) {
    irace_error("You must provide an 'ablog' object generated by ablation() or the path to the '.Rdata' file that contains this object.")
  }
  ablog <- read_ablogfile(ablog)
  if (!ablog$complete)
    stop("The ablog shows that the ablation procedure did not complete cleanly and only contains partial information")


  if (!is.null(pdf_file)) {
    if (!is.file.extension(pdf_file, ".pdf"))
      pdf_file <- paste0(pdf_file, ".pdf")
    cat("Creating PDF file '", pdf_file, "'\n", sep="")
    local_cairo_pdf(pdf_file, width = width, height = height, onefile= TRUE)
  }

  configurations <- ablog$allConfigurations
  # Support irace < 4.2
  if (is.null(configurations))
    configurations <- ablog$configurations

  trajectory <- ablog$trajectory
  if (n > 0L)
    trajectory <- trajectory[seq_len(n+1L)]
  # Generate labels
  labels <- ablation_labels(trajectory, configurations)
  if (!is.null(rename_labels))
    # stringr::str_replace_all() would be better but it has so many dependencies!
    for (i in seq_along(rename_labels))
      labels <- gsub(names(rename_labels)[i], rename_labels[i], labels)

  experiments <- ablog$experiments
  if ("rank" %in% type) {
    experiments <- rowRanks(experiments, ties.method = "average")
    if (is.null(ylim)) ylim <- c(1L, ncol(experiments))
  }
  costs_avg <- colMeans2(experiments, cols = trajectory)

  if (is.null(mar))
    mar <- par("mar")
  inches_to_lines <- (mar / par("mai"))[1L]
  lab_width <- max(strwidth(labels, units = "inches")) * inches_to_lines
  with_par(list(mar = c(lab_width + 2.1, 4.1, 0.1, 0.1), cex.axis = 1), {
    # FIXME: We could also show the other alternatives at each step not just the
    # one selected. See Leonardo Bezerra's thesis.
    if ("boxplot" %in% type) {
      bx <- boxplot(experiments[, trajectory], plot=FALSE)
      if (is.null(ylim)) {
        ylim <- range(bx$stats[is.finite(bx$stats)],
          bx$out[is.finite(bx$out)],
          bx$conf[is.finite(bx$conf)])
      }
    }

    plot(costs_avg, xaxt = "n", xlab = NA, ylab = ylab, ylim = ylim,
      type = "b", pch = 19, ...,
      panel.first = {
        grid(nx = NA, ny = NULL, lwd = 2);
        abline(h = c(costs_avg[1], utils::tail(costs_avg, n = 1L)),
          col = "lightgray", lty = "dotted", lwd = 2) })
    axis(1, at = seq_along(costs_avg), labels = labels, las = 3)
    if ("boxplot" %in% type) {
      bxp(bx, show.names = FALSE, add = TRUE)
    }
  })
  invisible()
}

#' Read the log file (`log-ablation.Rdata`) produced by [irace::ablation()].
#'
#' @param filename `character(1)`\cr Filename that contains the log file saved by [ablation()]. Example: `log-ablation.Rdata`.
#'
#' @return `list()`
#' @concept ablation
#' @export
read_ablogfile <- function(filename) read_logfile(filename, name = "ablog")

Timer <- R6::R6Class("Timer", cloneable = TRUE, lock_class = TRUE, lock_objects = TRUE, public = list(
  start = NULL,
  initialize = function() {
    self$start <- proc.time()
  },
  elapsed = function() {
    x <- proc.time() - self$start
    if (!is.na(x[[4L]])) 
        x[[1L]] <- x[[1L]] + x[[4L]]
    if (!is.na(x[[5L]])) 
        x[[2L]] <- x[[2L]] + x[[5L]]
    c(user=x[[1L]], system=x[[2L]], wallclock=x[[3L]])
  },
  wallclock = function() {
    proc.time()[[3L]] - self$start[[3L]]
  })
)

update_package_version <- function()
{
  dcf <- read.dcf(file = "DESCRIPTION", fields=c("Version", "RemoteSha"))
  if (NROW(dcf) < 1L)
    return(invisible())
  dcf <- as.list(dcf[1L, ])
  git_rev <- dcf$RemoteSha
  if (is.na(git_rev))
    git_rev <- "unknown"
  git <- Sys.which("git")
  if (git != "" && fs::file_exists(".git")
    && grepl("[0-9a-z]+$", system2(git, "describe --first-parent --always", stdout = TRUE), perl=TRUE)) {
    git_rev <- system2(git, "describe --dirty --first-parent --always --exclude '*'", stdout = TRUE)
  }
  if (git_rev != "unknown") {
    realversion <- paste0(dcf$Version, ".", git_rev)
    cat(file='./R/version.R', sep='',
"#' A character string containing the version of `irace` including git SHA.\n#' @export\nirace_version <- '", realversion, "'\n")
  }
  invisible()
}
update_package_version()
# We define this tentatively to avoid: undefined exports: irace_version
irace_version <- "unknown"

.irace_tolerance <- sqrt(.Machine$double.eps)
.get_time_next_save <- function(now) now + 60 # seconds
# Prefix for printing messages to the user.
.irace_msg_prefix <- "== irace == "

# Used in the roxygen2 documentation.
.roxygen_irace_return <- function(invisible = FALSE) {
  paste0(
    if (invisible) "(`invisible(data.frame)`)" else "(`data.frame`)",
    r"(

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:

  *  `.ID.` : Internal id of the candidate configuration.

  * `Parameter names` : One column per parameter name in `parameters`.

  * `.PARENT.` : Internal id of the parent candidate configuration.

  Additionally, this function saves an R data file containing an object called
  `iraceResults`. The path of the file is indicated in `scenario$logFile`.
  The `iraceResults` object is a list with the following structure:


 \describe{

 \item{`scenario`}{The scenario R object containing the \pkg{irace}
 options used for the execution. See [`defaultScenario`]
 for more information. The element `scenario$parameters` contains the
 parameters R object that describes the target algorithm parameters. See
 [`readParameters`].}

 \item{`allConfigurations`}{The target algorithm configurations
 generated by \pkg{irace}. This object is a data frame, each row is a
 candidate configuration, the first column (`.ID.`) indicates the
 internal identifier of the configuration, the following columns
 correspond to the parameter values, each column named as the parameter
 name specified in the parameter object. The final column
 (`.PARENT.`)  is the identifier of the configuration from which
 model the actual configuration was sampled.}

 \item{`allElites`}{A list that contains one element per iteration,
 each element contains the internal identifier of the elite candidate
 configurations of the corresponding iteration (identifiers correspond to
 `allConfigurations$.ID.`).}

 \item{`iterationElites`}{A vector containing the best candidate
 configuration internal identifier of each iteration. The best
 configuration found corresponds to the last one of this vector.}

 \item{`experiments`}{A matrix with configurations as columns and
 instances as rows. Column names correspond to the internal identifier of
 the configuration (`allConfigurations$.ID.`).}

 \item{`experimen_log`}{A `data.table` with columns `iteration`,
 `instance`, `configuration`, `time`.  This matrix contains the log of all the
 experiments that \pkg{irace} performs during its execution.  The
 instance column refers to the index of the `race_state$instances_log`
 data frame. Time is saved ONLY when reported by the `targetRunner`.}

 \item{`softRestart`}{A logical vector that indicates if a soft
 restart was performed on each iteration. If `FALSE`, then no soft
 restart was performed.}

 \item{`state`}{An environment that contains the state of \pkg{irace}, the
 recovery is done using the information contained in this object.}

 \item{`testing`}{A list that contains the testing results. The
 elements of this list are: `experiments` a matrix with the testing
 experiments of the selected configurations in the same format as the
 explained above and `seeds` a vector with the seeds used to execute
 each experiment.}
 }
)")
}

#' Generate a command-line representation of a configuration
#'
#' @description `buildCommandLine` receives two vectors, one containing
#'   the values of the parameters, the other containing the switches of the
#'   parameters. It builds a string with the switches and the values that can
#'   be used as a command line to call the program to be tuned, thus generating
#'   one candidate configuration.
#'
#'
#' @param values A vector containing the value of each parameter for the
#' candidate configuration.
#' @param switches A vector containing the switches of each paramter (in an
#'  order that corresponds to the values vector).
#'
#' @return A string concatenating each element of `switches` and
#'   `values` for all parameters with a space between each pair of
#'   parameters (but none between the switches and the corresponding values).
#'
#' @examples
#' switches <- c("--switch1 ", "--switch2-", "--switch3=")
#' values <- list("value_1", 1L, sqrt(2))
#' buildCommandLine (values, switches)
#' ## Build a command-line from the results produced by a previous run of irace.
#' # First, load the data produced by irace.
#' logfile <- file.path(system.file(package="irace"), "exdata", "irace-acotsp.Rdata")
#' iraceResults <- read_logfile(logfile)
#' allConfigurations <- iraceResults$allConfigurations
#' parameters <- iraceResults$scenario$parameters
#' apply(allConfigurations[1:10, unlist(parameters$names)], 1, buildCommandLine,
#'       unlist(parameters$switches))
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
buildCommandLine <- function(values, switches)
{
  irace_assert(length(values) == length(switches))
  values <- as.list(values)
  sel <- !is.na(values)
  switches <- switches[sel]
  values <- format.default(values[sel], digits=15L, scientific=FALSE)
  paste0(switches, values, collapse=" ")
}

# This function tries to convert a, possibly empty, character vector into a
# numeric vector.
parse_output <- function(outputRaw, verbose)
{
  if (verbose) cat(outputRaw, sep = "\n")
  # Initialize output as raw. If it is empty stays like this.
  output <- outputRaw
  # strsplit crashes if outputRaw == character(0)
  if (length(outputRaw)) {
    outputRaw <- paste0(outputRaw, collapse = "\n")
    output <- strsplit(trim(outputRaw), "[[:space:]]+")[[1L]]
  }
  # suppressWarnings to avoid messages about NAs introduced by coercion
  suppressWarnings(as.numeric(output))
}

target_error <- function(err_msg, output, scenario, target_runner_call,
                         target_evaluator_call = NULL)
{
  if (!is.null(target_evaluator_call)) {
    err_msg <- paste0(err_msg, "\n", .irace_msg_prefix,
                      "The call to targetEvaluator was:\n", target_evaluator_call)
  }
  if (!is.null(target_runner_call)) {
    err_msg <- paste0(err_msg, "\n", .irace_msg_prefix,
                      "The call to targetRunner was:\n", target_runner_call)
  }
  if (is.null(output$outputRaw)) {
    # Message for a function call.
    output$outputRaw <- deparse1(output)
    advice_txt <- paste0(
      "This is not a bug in irace, but means that something failed in ",
      "a call to the targetRunner or targetEvaluator functions provided by the user.",
      " Please check those functions carefully.")
  } else {
    # Message for an external script.
    advice_txt <- paste0(
      "This is not a bug in irace, but means that something failed when",
      " running the command(s) above or they were terminated before completion.",
      " Try to run the command(s) above from the execution directory '",
      scenario$execDir, "' to investigate the issue. See also Appendix B (targetRunner troubleshooting checklist) of the User Guide (https://cran.r-project.org/package=irace/vignettes/irace-package.pdf).")
  }
  irace_error(err_msg, "\n", .irace_msg_prefix,
              "The output was:\n", paste(output$outputRaw, collapse = "\n"),
              "\n", .irace_msg_prefix, advice_txt)
}

check_output_target_evaluator <- function (output, scenario, target_runner_time,
                                           target_runner_call, bound)
{
  if (!is.list(output)) {
    output <- list()
    target_error ("The output of targetEvaluator must be a list",
      output, scenario, target_runner_call = target_runner_call)
    return(output)
  }

  err_msg <- output$error
  if (is.null(err_msg)) {
    if (is.null(output$cost)) {
      err_msg <- "The output of targetEvaluator must contain 'cost'!"
    } else if (is_na_nowarn(output$cost)) {
      err_msg <- "The output of targetEvaluator is not numeric!"
    }
    if (scenario$maxTime > 0 || scenario$capping) {
      if (scenario$batchmode != 0) {
        if (is.null (output$time))
          err_msg <- "When batchmode != 0 and maxTime > 0, the output of targetEvaluator must be two numbers 'cost time'!"
        # With scenario$capping == TRUE, we may have pre-executed targetRunner
        # (which_elite_exe) that already have recorded the time, so when we
        # reach this point, we may not have 'time'.
      } else if (!scenario$capping && is.null(target_runner_time) && is.null(output$time)) {
        err_msg <- "Either targetRunner or targetEvaluator must return 'time' !"
      }
    }
    if (is.null(output$time)) {
      output$time <- target_runner_time
    } else {
      if (!is.null(target_runner_time)) {
        err_msg <- "Both targetRunner and targetEvaluator cannot return 'time' !"
      } else if (is_na_nowarn(output$time)) {
        err_msg <- "The time returned by targetEvaluator is not numeric!"
      } else if (is.infinite(output$time)) {
        err_msg <- "The time returned by targetEvaluator is not finite!"
      } else if (output$time <= 0) {
        err_msg <- paste0("The value of time (", output$time, ") returned by targetEvaluator must be strictly positive!")
      } else {
        # Fix time.
        output$time <- max(output$time, scenario$minMeasurableTime)
        if (!is.null(bound) && !is.na(bound) && bound > 0 && bound + scenario$minMeasurableTime < output$time) {
          err_msg <- paste0("The time returned by targetEvaluator (", output$time, ") does not respect the given bound of ", bound, "!")
        }
      }
    }
  }

  if (is.null(err_msg)) {
    output$error <- NULL
  } else {
    target_error (err_msg, output, scenario,
      target_runner_call = target_runner_call,
      target_evaluator_call = output$call)
  }
  output
}

#' target_evaluator_default
#'
#' `target_evaluator_default` is the default `targetEvaluator` function that is
#'  invoked if `targetEvaluator` is a string (by default
#'  `targetEvaluator` is `NULL` and this function is not invoked). You can use it as
#'  an advanced example of how to create your own `targetEvaluator` function.
#'
#' @param experiment A list describing the experiment. It contains at least:
#'    \describe{
#'     \item{`id_configuration`}{An alphanumeric string that uniquely identifies a configuration;}
#'     \item{`id_instance`}{An alphanumeric string that uniquely identifies an instance;}
#'      \item{`seed`}{Seed for the random number generator to be used for
#'        this evaluation, ignore the seed for deterministic algorithms;}
#'      \item{`instance`}{String giving the instance to be used for this evaluation;}
#'      \item{`bound`}{(only when `capping` is enabled) Time bound for the execution;}
#'      \item{`configuration`}{1-row data frame with a column per parameter
#'        name;}
#'    }
#' @param num_configurations Number of  configurations alive in the race.
#' @param all_conf_id Vector of configuration IDs of the alive configurations.
#' @inheritParams defaultScenario
#' @param target_runner_call String describing the call to `targetRunner` that
#'    corresponds to this call to `targetEvaluator`. This is used for
#'    providing extra information to the user, for example, in case
#'    `targetEvaluator` fails.
#'
#' @return The function `targetEvaluator` must return a list with one element
#'  `"cost"`, the numerical value corresponding to the cost measure of the
#'  given configuration on the given instance.
#'
#'  The return list may also contain the following optional elements that are used
#'  by \pkg{irace} for reporting errors in `targetEvaluator`:
#'  \describe{
#'    \item{`error`}{is a string used to report an error;}
#'    \item{`outputRaw`}{is a string used to report the raw output of calls to
#'      an external program or function;}
#'    \item{`call`}{is a string used to report how `targetRunner` called
#'      an external program or function.}
#'  }
#'
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
target_evaluator_default <- function(experiment, num_configurations, all_conf_id,
                                     scenario, target_runner_call)
{
  configuration_id <- experiment$id_configuration
  instance_id      <- experiment$id_instance
  seed             <- experiment$seed
  instance         <- experiment$instance

  debugLevel <- scenario$debugLevel
  targetEvaluator <- scenario$targetEvaluator
  if (as.logical(file.access(targetEvaluator, mode = 1))) {
    irace_error ("targetEvaluator", shQuote(targetEvaluator),
                 "cannot be found or is not executable!\n")
  }
  all_conf_id <- paste0(all_conf_id, collapse = " ")
  args <- c(configuration_id, instance_id, seed, instance, num_configurations, all_conf_id)
  withr::with_dir(scenario$execDir, {
    output <- runcommand(targetEvaluator, args, configuration_id, debugLevel, timeout = scenario$targetRunnerTimeout)
  })

  cost <- time <- NULL
  err_msg <- output$error
  if (is.null(err_msg)) {
    v_output <- parse_output(output$output, verbose = (scenario$debugLevel >= 2L))
    if (length(v_output) == 1L) {
      cost <- v_output[1L]
    } else if (length(v_output) == 2L) {
      cost <- v_output[1L]
      time <- v_output[2L]
    } else if (length(v_output) == 0L) {
      err_msg <- paste0("The output of targetEvaluator must be at least one number 'cost'!")
    } else {
      err_msg <- paste0("The output of targetEvaluator should not be more than two numbers!")
    }
  }
  list(cost = cost, time = time,
    error = err_msg, outputRaw = output$output,
    call = paste(targetEvaluator, args, collapse=" "))
}

#' Check the output of the target runner and repair it if possible. If the
#' output is incorrect, this function will throw an error.
#'
#' @param output The output from target runner.
#' @inheritParams defaultScenario
#' @param bound Optional time bound that the target runner should have respected.
#'
#' @return The output with its contents repaired.
#'
#' @export
check_output_target_runner <- function(output, scenario, bound = NULL)
{
  if (!is.list(output)) {
    output <- list()
    target_error ("The output of targetRunner must be a list", output, scenario, target_runner_call = NULL)
    return(output)
  }

  err_msg <- output$error
  if (is.null(err_msg)) {
    if (is.null(output$cost)) {
      output$cost <- NULL # make sure to delete it.
    } else if (is_na_or_empty(output$cost)) {
      err_msg <- "The cost returned by targetRunner is not numeric!"
    }

    if (is.null(output$time)) {
      output$time <- NULL # make sure to delete it.
    } else if (is.na(output$time)) {
      err_msg <- paste0("The time returned by targetRunner is not numeric!")
    } else if (is.infinite(output$time)) {
      err_msg <- paste0("The time returned by targetRunner is not finite!")
    } else if (output$time <= 0) {
      err_msg <- paste0("The value of time (", output$time, ") returned by targetRunner must be strictly positive!")
    } else {
      # Fix time.
      output$time <- max(output$time, scenario$minMeasurableTime)
      if (!is.null(bound) && !is.na(bound) && bound > 0 && bound + scenario$minMeasurableTime < output$time) {
        err_msg <- paste0("The time returned by targetRunner (", output$time, ") does not respect the given bound of ", bound, "!")
      }
    }
    if (is.null(err_msg)) {
      # When targetEvaluator is provided, targetRunner must return only the time.
      if (!is.null(scenario$targetEvaluator)) {
        # unless using batchmode, in that case targetRunner returns neither the
        # time nor the cost.
        if (scenario$batchmode != 0) {
          if (!is.null(output$time) || !is.null(output$cost)) {
            err_msg <- "When batchmode != 0, the output of targetRunner must not contain a cost nor a time!"
          }
        } else if (scenario$maxTime > 0 && is.null(output$time)) {
          err_msg <- "The output of targetRunner must be one number 'time'!"
        } else if (!is.null(output$cost)) {
          err_msg <- "The output of targetRunner must be empty or just one number 'time'!"
        }
      } else if (scenario$maxTime > 0 && (is.null(output$cost) || is.null(output$time))) {
        err_msg <- "The output of targetRunner must be two numbers 'cost time'!"
      } else if (scenario$maxExperiments > 0 && is.null(output$cost)) {
        err_msg <- "The output of targetRunner must be one number 'cost'!"
      }
    }
  }

  if (is.null(err_msg)) {
    output$error <- NULL
  } else {
    target_error (err_msg, output, scenario, target_runner_call = output$call)
  }
  output
}

# This function invokes target_runner.  When used on a remote node by Rmpi,
# environments do not seem to be shared and the default value is evaluated too
# late, thus we have to pass race_state$target_runner explicitly.
exec_target_runner <- function(experiment, scenario, target_runner)
{
  doit <- function(experiment, scenario) {
    x <- target_runner(experiment, scenario)
    check_output_target_runner(x, scenario, bound = experiment$bound)
  }

  retries <- scenario$targetRunnerRetries
  while (retries > 0L) {
    output <- try (doit(experiment, scenario))
    if (!inherits(output, "try-error") && is.null(output$error))
      return (output)
    irace_note("Retrying (", retries, " left).\n")
    retries <- retries - 1L
  }
  doit(experiment, scenario)
}

parse.aclib.output <- function(outputRaw)
{
  outputRaw <- paste0(outputRaw, collapse = "\n")
  text <- regmatches(outputRaw,
                     regexec("Result of this algorithm run:\\s*\\{(.+)\\}\\s*\n",
                             outputRaw))[[1]][2]
  aclib.match <- function(text, key, value) {
    pattern <- paste0('"', key, '":\\s*', value)
    return(regmatches(text, regexec(pattern, text))[[1]][2])
  }
  cost <- runtime <- error <- NULL
  # AClib wrappers print:
  # Result of this algorithm run:  {"status": "SUCCESS", "cost": cost, "runtime": time }
  # FIXME: This is not very robust. If we are going to be using jsonlite, then we can simply do:
  # jsonlite::fromJSON('{\"misc\": \"\", \"runtime\": 164.14, \"status\": \"SUCCESS\", \"cost\": \"0.121340\"}')
  status <- aclib.match(text, "status", '"([^"]+)"')
  if (!is.character(status)) {
    error <- paste0("Not valid AClib output")
  } else if (status %in% c("SUCCESS", "TIMEOUT")) {
    cost <- aclib.match(text, "cost", "([^[:space:],}]+)")
    cost <- suppressWarnings(as.numeric(cost))
    runtime <- aclib.match(text, "runtime", "([^[:space:],}]+)")
    runtime <- suppressWarnings(as.numeric(runtime))
    if (is.null.or.na(cost) && is.null.or.na(runtime))
      error <- paste0("Not valid cost or runtime in AClib output")
  } else if (status %in% c("CRASHED", "ABORT")) {
    # FIXME: Implement ABORT semantics of fatal error
    error <- paste0("targetRunner returned status (", status, ")")
  } else {
    error <- paste0("Not valid AClib output status (", status, ")")
  }
  list(status = status, cost = cost, time = runtime, error = error)
}

target_runner_aclib <- function(experiment, scenario)
{
  debugLevel   <- scenario$debugLevel
  res <- run_target_runner(experiment, scenario)
  cmd <- res$cmd
  output <- res$output
  args <- res$args

  err_msg <- output$error
  if (is.null(err_msg)) {
    return(c(parse.aclib.output(output$output),
             list(outputRaw = output$output, call = paste(cmd, args))))
  }

  list(cost = NULL, time = NULL, error = err_msg,
       outputRaw = output$output, call = paste(cmd, args))
}

check_target_cmdline <- function(target_cmdline, launcher, capping)
{
  required <- c("seed", "instance", "targetRunnerArgs")
  if (capping) required <- c(required, "bound")
  for (x in required) {
    if (!grepl(paste0("{", x, "}"), target_cmdline, fixed=TRUE))
      irace_error("targetCmdline '", target_cmdline, "' must contain '{", x, "}'")
  }
}

expand_target_cmdline <- function(target_cmdline, experiment, targetRunner, targetRunnerArgs)
{
  vars <- list(configurationID = experiment$id_configuration,
               instanceID      = experiment$id_instance,
               seed            = experiment$seed,
               instance        = experiment$instance,
               bound           = experiment$bound,
               targetRunner = targetRunner,
               targetRunnerArgs = targetRunnerArgs)
  for (x in names(vars)) {
    value <- vars[[x]]
    if (is.null(value)) value <- ""
    if (x == "targetRunner") value <- shQuote(value)
    target_cmdline <- gsub(paste0("{", x, "}"), value, target_cmdline, fixed=TRUE)
  }
  target_cmdline
}

run_target_runner <- function(experiment, scenario)
{
  configuration_id <- experiment$id_configuration
  instance_id      <- experiment$id_instance
  seed             <- experiment$seed
  configuration    <- experiment$configuration
  instance         <- experiment$instance
  bound            <- experiment$bound

  switches <- scenario$parameters$switches[names(configuration)]

  targetRunner <- scenario[["targetRunner"]]
  debugLevel <- scenario$debugLevel

  if (scenario$aclib) {
    # FIXME: Use targetCmdline for this
    has_value <- !is.na(configuration)
    # <executable> [<arg>] [<arg>] ... [--cutoff <cutoff time>] [--instance <instance name>]
    # [--seed <seed>] --config [-param_name_1 value_1] [-param_name_2 value_2] ...
    args <- paste("--instance", instance, "--seed", seed, "--config",
                  paste0("-", switches[has_value], " ", configuration[has_value],
                         collapse = " "))
    if (!is.null.or.na(bound))
      args <- paste("--cutoff", bound, args)
  } else {
    args <- expand_target_cmdline(scenario$targetCmdline, experiment, targetRunner,
                                  targetRunnerArgs=buildCommandLine(configuration, switches))
  }
  targetRunnerLauncher <- scenario$targetRunnerLauncher
  if (!is.null.or.empty(targetRunnerLauncher))
    targetRunner <- targetRunnerLauncher

  output <- runcommand(targetRunner, args, configuration_id, debugLevel, timeout = scenario$targetRunnerTimeout)
  list(cmd=targetRunner, output=output, args=args)
}

#' Default `targetRunner` function.
#'
#' Use it as an advanced example of how to create your own `targetRunner` function.
#'
#' @param experiment A list describing the experiment. It contains at least:
#'    \describe{
#'     \item{`id_configuration`}{An alphanumeric string that uniquely identifies a configuration;}
#'     \item{`id_instance`}{An alphanumeric string that uniquely identifies an instance;}
#'      \item{`seed`}{Seed for the random number generator to be used for
#'        this evaluation, ignore the seed for deterministic algorithms;}
#'      \item{`instance`}{String giving the instance to be used for this evaluation;}
#'      \item{`bound`}{(only when `capping` is enabled) Time bound for the execution;}
#'      \item{`configuration`}{1-row data frame with a column per parameter
#'        name;}
#'    }
#' @inheritParams defaultScenario
#'
#' @return If `targetEvaluator` is `NULL`, then the `targetRunner`
#'  function must return a list with at least one element `"cost"`,
#'  the numerical value corresponding to the evaluation of the given
#'  configuration on the given instance.
#'
#'  If the scenario option `maxTime` is non-zero or if `capping` is enabled
#'  then the list must contain at least another element `"time"` that reports the
#'  execution time for this call to `targetRunner`.
#'  The return list may also contain the following optional elements that are used
#'  by \pkg{irace} for reporting errors in `targetRunner`:
#'  \describe{
#'    \item{`error`}{is a string used to report an error;}
#'    \item{`outputRaw`}{is a string used to report the raw output of calls to
#'      an external program or function;}
#'    \item{`call`}{is a string used to report how `targetRunner` called
#'      an external program or function.}
#'  }
#'
#'
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @export
target_runner_default <- function(experiment, scenario)
{
  res <- run_target_runner(experiment, scenario)
  cmd <- res$cmd
  output <- res$output
  args <- res$args

  debugLevel <- scenario$debugLevel
  cost <- time <- NULL
  err_msg <- output$error
  if (is.null(err_msg)) {
    v_output <- parse_output(output$output, verbose = (debugLevel >= 2L))
    if (length(v_output) == 1L) {
      if (is.null(scenario$targetEvaluator)) {
        cost <- v_output[1L]
      } else {
        time <- v_output[1L]
      }
    } else if (length(v_output) == 2L) {
      cost <- v_output[1L]
      time <- v_output[2L]
    } else {
      err_msg <- "The output of targetRunner should not be more than two numbers!"
    }
  }
  list(cost = cost, time = time,
       error = err_msg, outputRaw = output$output,
       call = paste(cmd, args, collapse = " "))
}

execute_experiments <- function(race_state, experiments, scenario)
{
  parallel <- scenario$parallel
  mpi <- scenario$mpi
  target_runner <- race_state$target_runner
  execDir <- scenario$execDir
  if (!fs::dir_exists(execDir))
    irace_error ("Execution directory '", execDir, "' is not found or not a directory\n")
  withr::local_dir(execDir)

  if (!is.null(scenario$targetRunnerParallel)) {
    # FIXME: We should remove the exec_target_runner parameter from
    # targetRunnerParallel and do lapply(target_output,
    # check_output_target_runner, scenario=scenario) after it to make sure the output is valid.
    # User-defined parallelization
    target_output <-
      scenario$targetRunnerParallel(experiments, exec_target_runner,
                                    scenario = scenario,
                                    target_runner = target_runner)
    if (length(target_output) != length(experiments)) {
      irace_error("Stopping because the output of targetRunnerParallel is missing elements. The output was:\n",
        paste0(utils::capture.output(utils::str(target_output)), collapse="\n"))
    }
  } else if (scenario$batchmode != 0L) {
    target_output <- cluster_lapply(experiments, scenario = scenario)
  } else if (parallel > 1L) {
    if (mpi) {
      if (scenario$loadBalancing) {
        target_output <- Rmpi::mpi.applyLB(experiments, exec_target_runner,
          scenario = scenario,
          target_runner = target_runner)
      } else {
        # Without load-balancing, we need to split the experiments into chunks
        # of size parallel.
        target_output <- unlist(use.names = FALSE,
          tapply(experiments,
            ceiling(seq_along(experiments) / parallel),
            Rmpi::mpi.apply, exec_target_runner,
            scenario = scenario,
            target_runner = target_runner))
      }
      # FIXME: if stop() is called from mpi.applyLB, it does not
      # terminate the execution of the parent process, so it will
      # continue and give more errors later. We have to terminate
      # here, but is there a nicer way to detect this and terminate?
      if (any(sapply(target_output, inherits, "try-error"))) {
        # FIXME: mclapply has some bugs in case of error. In that
        # case, each element of the list does not keep the output of
        # each configuration and repetitions may occur.
        cat(unique(unlist(target_output[sapply(
          target_output, inherits, "try-error")])), file = stderr(), sep = "")
        irace_error("A slave process terminated with a fatal error")
      }
    } else {
      if (.Platform$OS.type == 'windows') {
        irace_assert(!is.null(race_state$cluster))
        if (scenario$loadBalancing) {
          target_output <-
            parallel::parLapplyLB(race_state$cluster, experiments, exec_target_runner,
              scenario = scenario,
              target_runner = target_runner)
        } else {
          target_output <-
            parallel::parLapply(race_state$cluster, experiments, exec_target_runner,
              scenario = scenario,
              target_runner = target_runner)
        }
        # FIXME: if stop() is called from parLapply, then the parent
        # process also terminates, and we cannot give further errors.
      } else {
        target_output <-
          parallel::mclapply(experiments, exec_target_runner,
            # FALSE means load-balancing.
            mc.preschedule = !scenario$loadBalancing,
            mc.cores = parallel,
            scenario = scenario,
            target_runner = target_runner)
        # FIXME: if stop() is called from mclapply, it does not
        # terminate the execution of the parent process, so it will
        # continue and give more errors later. We have to terminate
        # here, but is there a nicer way to detect this and terminate?
        if (any(sapply(target_output, inherits, "try-error"))
          || any(sapply(target_output, is.null))) {
          # FIXME: mclapply has some bugs in case of error. In that
          # case, each element of the list does not keep the output of
          # each configuration and repetitions may occur.
          cat(unique(unlist(
            target_output[sapply(
              target_output, inherits, "try-error")])), file = stderr())
          irace_error("A child process triggered a fatal error")
        }
      }
    }
  } else {
    # One process, all sequential
    target_output <- lapply(experiments, exec_target_runner,
      scenario = scenario,
      target_runner = target_runner)
  }
  target_output
}

execute_evaluator <- function(target_evaluator, experiments, scenario, target_output)
{
  configurations_id <- unique(unlist_element(experiments, "id_configuration"))
  nconfs <- length(configurations_id)
  # Evaluate configurations sequentially.
  for (k in seq_along(experiments)) {
    experiment <- experiments[[k]]
    target_runner_call <- target_output[[k]]$call
    output <- target_evaluator(experiment = experiment, num_configurations = nconfs,
      all_conf_id = configurations_id, scenario = scenario,
      target_runner_call = target_runner_call)
    output <- check_output_target_evaluator(output, scenario, target_runner_time = target_output[[k]]$time,
      target_runner_call = target_runner_call, bound = experiment$bound)
    target_output[[k]]$cost <- output$cost
    # targetEvaluator may return time, for example for batchmode != 0.
    target_output[[k]]$time <- output$time
    if (is.null(target_output[[k]]$call))
      target_output[[k]]$call <- output$call
  }
  target_output
}

#' Converts a relative path to an absolute path.
#'
#' If the path passed corresponds to an executable, it tries to find its path
#' using [Sys.which()].  Expansion of `'~'` in Windows follows the definition
#' of [fs::path_expand()] rather than [base::path.expand()]. This function
#' tries really hard to create canonical paths.
#'
#' @param path (`character(1)`) Character string representing a relative path.
#' @param cwd (`character(1)`) Current working directory.
#'
#' @return (`character(1)`) Character string representing the absolute path
#'
#' @examples
#' path_rel2abs("..")
#' @export
path_rel2abs <- function (path, cwd = getwd())
{
  if (path == "") return("")
  # FIXME: split this function into two, one for executable files than handles
  # finding executables in the PATH with Sys.which() and another for everything
  # else that doesn't try to be that smart.

  # We need to do this even with the path returned by `getwd()`.
  cwd <- fs::path_norm(fs::path_expand(cwd))

  if (fs::is_absolute_path(path)) {
    # Possibly expand ~/path to /home/user/path.
    path <- fs::path_expand(path)
  } else if (!startsWith(path, ".")) {
    # it may be a command in the path.
    sys_path <- suppressWarnings(Sys.which(path))
    if (sys_path != "" && fs::file_access(sys_path, "execute")
      # fs::is_dir() is broken: https://github.com/r-lib/fs/issues/440
      && !file.info(sys_path)$isdir) {
      path <- as.vector(sys_path)
    }
  }
  as.character(fs::path_abs(path, start = cwd))
}

test_path_rel2abs <- function(testcases)
{
  for (i in 1:nrow(testcases)) {
    orig <- testcases[i,1L]
    cwd <-  testcases[i,2L]
    res <- path_rel2abs(testcases[i,1L], cwd)
    if (testcases[i,3L] == "Sys.which") {
      exp <- fs::path_abs(Sys.which(testcases[i,1L]))
    } else {
      exp <- gsub("\\", "/", fs::path_expand(testcases[i,3L]), fixed = TRUE)
    }
    if (res == exp) {
      #cat("[OK] ", i, ": path_rel2abs(\"", orig, "\", \"", cwd, "\") -> ", res, "\n", sep="")
    } else {
      cat(sep="", "[FAILED] ", i, ": path_rel2abs(\"", orig, "\", \"", cwd, "\") -> ", res, " but expected: ", exp, "\n")
    }
    expect_match(res, exp, fixed = TRUE)
  }
}

test_that("test path_rel2abs", {
  # Try to set wd; otherwise fail silently.
  old.cwd <- getwd()
  skip_if(is.null(old.cwd))
  withr::defer(setwd(old.cwd))
  tryCatch(setwd("/tmp"), error = function(e) { skip(e) })
  
  testcases <- read.table(text='
"."                         "/tmp"  "/tmp"
".."                        "/tmp"  "/"
"../"                       "/tmp"  "/"
"../."                      "/tmp"  "/"
"../.."                     "/tmp"  "/"
"../../"                    "/tmp"  "/"
"../../x.r"                 "/tmp"  "/x.r"
"../leslie/"                "/tmp"  "/leslie"
"../leslie/x.r"             "/tmp"  "/leslie/x.r"
"../x.r"                    "/tmp"  "/x.r"
"..irace"                   "/tmp"  "/tmp/..irace"
"./"                        "/tmp"  "/tmp"
"./."                       "/tmp"  "/tmp"
"./"                        "/tmp/"  "/tmp"
"./."                       "/tmp/"  "/tmp"
"././x.r"                   "/tmp"  "/tmp/x.r"
"./irace/../x.r"            "/tmp"  "/tmp/x.r"
"./x.r"                     "/tmp"  "/tmp/x.r"
".x.R"                      "/tmp"  "/tmp/.x.R"
"/./x.r"                    "/tmp"  "/x.r"
"/home"                     "/tmp"  "/home"
"/home/leslie/././x.r"      "/tmp"  "/home/leslie/x.r"
"/home/leslie/~/x.r"        "/tmp"  "/home/leslie/~/x.r"
"/~/x.r"                    "/tmp"  "/~/x.r"
"leslie/leslie/../../irace" "/tmp"  "/tmp/irace"
"x.r"                       "/tmp"  "/tmp/x.r"
"~/irace/../x.r"            "/tmp"  "~/x.r"
"~/x.r"                     "/tmp"  "~/x.r"
"../../../data"             "./"    "/data"
"../../../data"             "/tmp/a/b/c/" "/tmp/data"
"..//a"                     ".//"   "/a"
', stringsAsFactors=FALSE)
  test_path_rel2abs(testcases)
})

test_that("test path_rel2abs without /tmp", {

  testcases <- read.table(text='
.                         N:\\\\tmp  N:/tmp
..                        N:\\\\tmp  N:/
..\\\\                       N:\\\\tmp  N:/
..\\\\.                      N:\\\\tmp  N:/
..\\\\..                     N:\\\\tmp  N:/
..\\\\..\\\\                    N:\\\\tmp  N:/
..\\\\..\\\\x.r                 N:\\\\tmp  N:/x.r
..\\\\leslie\\\\                N:\\\\tmp  N:/leslie
..\\\\leslie\\\\x.r             N:\\\\tmp  N:/leslie/x.r
..\\\\x.r                    N:\\\\tmp  N:/x.r
..irace                   N:\\\\tmp  N:/tmp/..irace
.\\\\                        N:\\\\tmp  N:/tmp
.\\\\.                       N:\\\\tmp  N:/tmp
.\\\\                        N:\\\\tmp\\\\ N:/tmp
.\\\\.                       N:\\\\tmp\\\\  N:/tmp
.\\\\.\\\\x.r                   N:\\\\tmp  N:/tmp/x.r
.\\\\x.r                     N:\\\\tmp  N:/tmp/x.r
.\\\\irace\\\\..\\\\x.r            N:\\\\tmp  N:/tmp/x.r
.x.R                      N:\\\\tmp  N:/tmp/.x.R
.                         N:\\tmp  N:/tmp
..                        N:\\tmp  N:/
..\\                       N:\\tmp  N:/
..\\.                      N:\\tmp  N:/
..\\..                     N:\\tmp  N:/
..\\..\\                    N:\\tmp  N:/
..\\..\\x.r                 N:\\tmp  N:/x.r
..\\leslie\\                N:\\tmp  N:/leslie
..\\leslie\\x.r             N:\\tmp  N:/leslie/x.r
..\\x.r                    N:\\tmp  N:/x.r
..irace                   N:\\tmp  N:/tmp/..irace
.\\                        N:\\tmp  N:/tmp
.\\.                       N:\\tmp  N:/tmp
.\\                        N:\\tmp\\ N:/tmp
.\\.                       N:\\tmp\\  N:/tmp
.\\.\\x.r                   N:\\tmp  N:/tmp/x.r
.\\irace\\..\\x.r            N:\\tmp  N:/tmp/x.r
.\\x.r                     N:\\tmp  N:/tmp/x.r
.x.R                      N:\\tmp  N:/tmp/.x.R
.                         N:  N:/
..                        N:  N:/
..\\\\                       N:  N:/
..\\\\.                      N:  N:/
..\\\\..                     N:  N:/
..\\\\..\\\\                    N:  N:/
..\\\\..\\\\x.r                 N:  N:/x.r
..\\\\leslie\\\\                N:  N:/leslie
..\\\\leslie\\\\x.r             N:  N:/leslie/x.r
..\\\\x.r                    N:  N:/x.r
..\\                       N:  N:/
..\\.                      N:  N:/
..\\..                     N:  N:/
..\\..\\                    N:  N:/
..\\..\\x.r                 N:  N:/x.r
..\\leslie\\                N:  N:/leslie
..\\leslie\\x.r             N:  N:/leslie/x.r
..\\x.r                    N:  N:/x.r
..irace                   N:  N:/..irace
.\\\\                        N:  N:/
.\\\\.                       N:  N:/
.\\\\                        N:\\\\ N:/
.\\\\.                       N:\\\\  N:/
.\\\\.\\\\x.r                   N:  N:/x.r
.\\\\irace\\\\..\\\\x.r            N:  N:/x.r
.\\\\x.r                     N:  N:/x.r
.\\                        N:  N:/
.\\.                       N:  N:/
.\\                        N:\\ N:/
.\\.                       N:\\  N:/
.\\.\\x.r                   N:  N:/x.r
.\\irace\\..\\x.r            N:  N:/x.r
.\\x.r                     N:  N:/x.r
.x.R                      N:  N:/.x.R
.                         N:/tmp  N:/tmp
..                        N:/tmp  N:/
../                       N:/tmp  N:/
../.                      N:/tmp  N:/
../..                     N:/tmp  N:/
../../                    N:/tmp  N:/
../../x.r                 N:/tmp  N:/x.r
../leslie/                N:/tmp  N:/leslie
../leslie/x.r             N:/tmp  N:/leslie/x.r
../x.r                    N:/tmp  N:/x.r
..irace                   N:/tmp  N:/tmp/..irace
./                        N:/tmp  N:/tmp
./.                       N:/tmp  N:/tmp
./                        N:/tmp/ N:/tmp
./.                       N:/tmp/  N:/tmp
././x.r                   N:/tmp  N:/tmp/x.r
./irace/../x.r            N:/tmp  N:/tmp/x.r
./x.r                     N:/tmp  N:/tmp/x.r
.x.R                      N:/tmp  N:/tmp/.x.R
D:/./x.r                  N:/tmp  D:/x.r
D:\\\\.\\\\x.r            N:/tmp  D:/x.r
D:\\.\\x.r                N:/tmp  D:/x.r
D:                        N:/tmp  D:/
D:\\\\                    N:/tmp  D:/
D:/                       N:/tmp  D:/
D:/leslie/././x.r         N:/tmp  D:/leslie/x.r
D:/leslie/~/x.r           N:/tmp  D:/leslie/~/x.r
e:/home/leslie/x.r        /tmp  E:/home/leslie/x.r
leslie/leslie/../../irace N:/tmp  N:/tmp/irace
x.r                       N:/tmp  N:/tmp/x.r
~/irace/../x.r            N:/tmp  ~/x.r
~/x.r                     N:/tmp  ~/x.r
"R"                       "/tmp/" "Sys.which"
', stringsAsFactors=FALSE)
  test_path_rel2abs(testcases)
})

test_that("test path_rel2abs with symlink", {
  # Try to set wd; otherwise fail silently.
  old.cwd <- getwd()
  skip_if(is.null(old.cwd))
  withr::defer(setwd(old.cwd))
  tryCatch({
    tmp <- withr::local_tempdir()
    setwd(tmp)
    fs::dir_create("a")
    fs::file_create("a/b")
    fs::link_create(fs::path_abs("a"), "c")
  }, error = function(e) { skip(e) })
  testcases <- data.frame(p = "c/b", wd = ".", res = file.path(tmp, "c/b"),
    stringsAsFactors=FALSE)
  test_path_rel2abs(testcases)
})

#' Return the elite configurations of the final iteration.
#'
#' @inheritParams has_testing_data
#' @param n `integer(1)`\cr Number of elite configurations to return, if \code{n} is larger than the
#' number of configurations, then only the existing ones are returned. The default (\code{n=0}) returns all of them.
#' @param drop.metadata `logical(1)`\cr Remove metadata, such as the
#'   configuration ID and the ID of the parent, from the returned
#'   configurations.  See [removeConfigurationsMetaData()].
#'
#' @return A data frame containing the elite configurations required.
#'
#' @examples
#' log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
#' print(removeConfigurationsMetaData(getFinalElites(log_file, n=1)))
#'
#' @author Manuel López-Ibáñez and Leslie Pérez Cáceres
#' @concept analysis
#' @export
getFinalElites <- function(iraceResults, n = 0L, drop.metadata = FALSE)
{
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing")
  iraceResults <- read_logfile(iraceResults)

  last_elites <- iraceResults$allElites[[length(iraceResults$allElites)]]

  if (n == 0L)
    n <- length(last_elites)

  if (length(last_elites) < n) {
    cat("Only", length(last_elites), "configurations available, reducing n,")
    n <- length(last_elites)
  }
  last_elites <- last_elites[seq_len(n)]

  configurations <- subset(iraceResults$allConfigurations,
                           get(".ID.") %in% as.character(last_elites),
                           drop = FALSE)
  if (drop.metadata)
    configurations <- removeConfigurationsMetaData(configurations)
  configurations
}

#' Returns the configurations selected by ID.
#'
#' @param ids `integer()`\cr The id or a vector of ids of the candidates configurations to obtain.
#' @inheritParams getFinalElites
#'
#' @return A data frame containing the elite configurations required, in the
#'   order and with the repetitions given by `ids`.
#' @examples
#' log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
#' getConfigurationById(log_file, ids = c(2,1), drop.metadata = TRUE)
#'
#' @author Manuel López-Ibáñez and Leslie Pérez Cáceres
#' @concept analysis
#' @export
getConfigurationById <- function(iraceResults, ids, drop.metadata = FALSE)
{
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing")
  iraceResults <- read_logfile(iraceResults)

  if (length(ids) < 1L) stop("You must provide at least one configuration id.")

  get_configuration_by_id_helper(iraceResults$allConfigurations, ids, drop_metadata = drop.metadata)
}

#' Returns the configurations by the iteration in which they were executed.
#'
#' @param iterations `integer()`\cr The iteration number or a vector of iteration numbers from where
#'  the configurations should be obtained. Negative values start counting from the last iteration.
#' @inheritParams getFinalElites
#'
#' @return A data frame containing the elite configurations required.
#'
#' @examples
#' log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
#' getConfigurationByIteration(log_file, iterations = c(-2, -1), drop.metadata = TRUE)
#'
#' @author Manuel López-Ibáñez and Leslie Pérez Cáceres
#' @concept analysis
#' @export
getConfigurationByIteration <- function(iraceResults, iterations, drop.metadata = FALSE)
{
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing")
  iraceResults <- read_logfile(iraceResults)

  if (length(iterations) < 1L)
    stop("You must provide at least one iteration number.")

  n_iterations <- length(iraceResults$iterationElites)
  iterations <- as.integer(iterations)
  iterations <- ifelse(iterations >= 0L, iterations, n_iterations + 1L + iterations)
  # To silence warning.
  iteration <- NULL
  if (is.null(iraceResults$state$experiment_log) ||
        nrow(iraceResults$state$experiment_log) == 0L)
    stop("'iraceResults' does not contain experiment_log, maybe the wrong file, an incomplete run or the wrong version of irace?")

  ids <- unique(subset(as.data.frame(iraceResults$state$experiment_log),
                       iteration %in% iterations,
                       select="configuration", drop=TRUE))
  get_configuration_by_id_helper(iraceResults$allConfigurations, ids, drop_metadata = drop.metadata)
}


get_configuration_by_id_helper <- function(allConfigurations, ids, drop_metadata)
{
  configurations <- allConfigurations[match(ids, allConfigurations[[".ID."]]), , drop=FALSE]
  if (nrow(configurations) == 0L)
    stop("No configuration found with ID:", ids, ".")

  if (drop_metadata)
    configurations <- removeConfigurationsMetaData(configurations)
  configurations
}

#' Returns the pairs of instance IDs and seeds used as instances in the race
#' (and optionally the actual instances).
#'
#' @inheritParams getFinalElites
#' @param index  `integer()`\cr Indexes of the (instanceID,seed) pairs to be returned. The default returns everything.
#' @param instances `logical(1)`\cr Whether to add the actual instances as an additional column (only if the instances are of atomic type).
#'
#' @return `data.table()`\cr With default arguments, a `data.table` containing two columns
#'   `"instanceID"` and `"seed"`. With `instances=TRUE` and if the instances
#'   are of atomic type (see [is.atomic()]) type, another column `instance` is
#'   added that contains the actual instance.
#'
#' @examples
#' log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
#' head(get_instanceID_seed_pairs(log_file))
#' # Add the instance names
#' get_instanceID_seed_pairs(log_file, index=1:10, instances=TRUE)
#' @author Manuel López-Ibáñez
#' @concept analysis
#' @export
get_instanceID_seed_pairs <- function(iraceResults, index, instances = FALSE)
{
  if (missing(iraceResults)) stop("argument 'iraceResults' is missing")
  iraceResults <- read_logfile(iraceResults)
  instances_log <- iraceResults$state$instances_log
  if (!missing(index))
    instances_log <- instances_log[index, , drop = FALSE]
  if (!instances)
    return(instances_log)

  instances <- iraceResults$scenario$instances
  if (!is.atomic(instances)) {
    warning("instances=TRUE requested, but instances are not of atomic type")
    return(instances_log)
  }

  instanceID <- instances_log[["instanceID"]]
  cbind(instances_log, instance = instances[instanceID])
}

withr::with_output_sink("test-blocksize.Rout", {
cap_irace <- function(..., targetRunner = force(target_runner_capping_xy))
{
  # Silence Error in `save(iraceResults, file = logfile, version = 3L)`: (converted from warning) 'package:irace' may not be available when loading
  # See https://github.com/r-lib/testthat/issues/2044
  if (!is.null(attr(environment(targetRunner), "name", exact=TRUE))) {
    environment(targetRunner) <- globalenv()
  }

  args <- list(...)
  parameters_table <- '
   x "" r (0, 1.00)
   y "" r (0, 1.00)'

  parameters <- readParameters(text = parameters_table)
  logFile <- withr::local_tempfile(fileext=".Rdata")

  scenario <- list(instances = c("ackley", "goldestein", "matyas", "himmelblau"),
                   targetRunner = targetRunner,
                   capping = TRUE,
                   blockSize = 4,
                   boundMax = 80,
                   logFile = logFile,
                   testType = "t-test",
                   parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)
  confs <- irace(scenario = scenario)
  best_conf <- getFinalElites(scenario$logFile, n = 1L, drop.metadata = TRUE)
  expect_identical(removeConfigurationsMetaData(confs[1L, , drop = FALSE]),
                   best_conf)
  invisible(read_logfile(scenario$logFile))
}

target_runner_time <- function(experiment, scenario)
{
  configuration     <- experiment$configuration
  tmax <-  configuration[["tmax"]]
  temp <-  configuration[["temp"]]
  time <- max(1, abs(rnorm(1, mean=(tmax+temp)/10)))
  list(cost = time, time = time)
}

time_irace <- function(...)
{
  args <- list(...)
  parameters <- readParameters(text = '
   tmax "" i (-10, 10)
   temp "" r (0, 10)
   ')
  scenario <- list(targetRunner = target_runner_time,
                   instances = c("ackley", "goldestein", "matyas"),
                   blockSize=3,
                   parameters = parameters)
  scenario <- modifyList(scenario, args)
  scenario <- checkScenario (scenario)

  confs <- irace(scenario = scenario)
  best_conf <- getFinalElites(scenario$logFile, n = 1L, drop.metadata = TRUE)
  expect_identical(removeConfigurationsMetaData(confs[1L, , drop = FALSE]),
                   best_conf)
  invisible(read_logfile(scenario$logFile))
}

check_blocksize <- function(results)
{
  expect_equal(rowMeans(matrix(get_instanceID_seed_pairs(results)[["instanceID"]],nrow=results$scenario$blockSize)),
               rowMeans(matrix(seq_along(results$scenario$instances), nrow=results$scenario$blockSize)))
  expect_equal(sum(colSums(!is.na(results$experiments)) %% results$scenario$blockSize), 0)
}

test_that("blockSize error", {
  expect_error(cap_irace(maxExperiments = 1000, blockSize=3), "must be a multiple of 'blockSize")
})

test_that("blockSize cap_irace maxExperiments = 1000", {
  generate_set_seed()
  expect_warning(check_blocksize(cap_irace(maxExperiments = 1000, debugLevel = 3)),
                 "Assuming 'mu = firstTest * blockSize' because 'mu' cannot be smaller",
                 fixed = TRUE)
})

test_that("blockSize maxTime=1000", {
  generate_set_seed()
  check_blocksize(time_irace(maxTime = 1000))
})

test_that("blockSize maxTime=1000 elitistNewInstances", {
  skip_on_cran()
  generate_set_seed()
  check_blocksize(time_irace(maxTime = 1000, instances = letters[1:9],
                             elitistNewInstances = 2, elitistLimit = 2))
})

}) # withr::with_output_sink()

# FIXME: Restoring occurs after reading the command-line/scenario file. At
# least for the irace command-line parameters (scenario), it should occur
# before. We would need to:
#
# 1) Read recovery file settings from command-line/scenario file
#
# 2) if set, then recover irace scenario

# 3) then read other settings from command-line/scenario file being
# careful to not override with defaults whatever the recovery has set.
#
# 4) checkScenario()
#
# A work-around is to modify the recovery file (you can load it in R,
# modify scenario then save it again).
recoverFromFile <- function(filename, scenario = list())
{
  iraceResults <- read_logfile(filename)
  if (iraceResults$irace_version != irace::irace_version)
    irace_error("Recovery file '", filename, "' was generated by a version of irace (",
      iraceResults$irace_version, ") different from this version of irace (",
      irace::irace_version, ").")

  # Restore part of scenario but not all.
  for (name in .irace.params.recover)
    scenario[[name]] <- iraceResults$scenario[[name]]
  # We call checkScenario() again to fix any inconsistencies in the recovered data.
  # FIXME: Do not call checkScenario earlier and instead do the minimum to check recoveryFile.
  scenario <- checkScenario(scenario)
  race_state <- iraceResults$state$clone()
  race_state$initialize(scenario, recover = TRUE)
  race_state
}

##
## Numerical configurations similarity function
##
# FIXME: This function is too slow and it shows up in profiles.
numeric.configurations.equal <- function(x, configurations, parameters, threshold, param.names)
{
  d <- numeric(nrow(configurations))
  isSimilar <- matrix(TRUE, nrow = nrow(configurations), ncol = length(param.names))
  selected <- seq_nrow(configurations)
  for (i in seq_along(param.names)) {
    pname <- param.names[i]
    param <- parameters$get(pname)

    is_dep <- param[["is_dependent"]]
    x_domain <- if (is_dep) getDependentBound(param, x) else param[["domain"]]
    x_range <- x_domain[[2L]] - x_domain[[1L]]

    X <- x[[pname]]
    # FIXME: Since at the end we select a subset of configurations, we could use selected here.
    y <- configurations[[pname]]
    ## FIXME: This can probably done much faster by doing a matrix operation that updates
    ## isSimilar[, i] in one step instead of the for-loop.
    ## We would need to handle the NAs first.
    for (j in seq_len(nrow(isSimilar))) { # Configurations loop
      Y <- y[selected[j]]
      if (is.na (X) && is.na(Y)) { # Both NA, just ignore this pname
        next
      } else if (xor(is.na (X), is.na(Y))) { # Distance is 1.0, so not equal
        isSimilar[j,i] <- FALSE
      } else {
        # FIXME: Why is this updating d[j]? It seems that if the difference is
        # large for one configuration, then it will be assumed to be large for
        # the rest.
        if (is_dep) {
          # Compare dependent domains by normalising their values to their own ranges first
          # and calculating the difference. (When possible)
          y_domain <- getDependentBound(param, configurations[selected[j],])
          y_range <- y_domain[[2L]] - y_domain[[1L]]
          dx <- if (x_range == 0) 0 else (as.numeric(X) - x_domain[[1L]]) / x_range
          dy <- if (y_range == 0) 0 else (as.numeric(Y) - y_domain[[1L]]) / y_range
          d[j] <- max(d[j], abs(dx - dy))
        } else {
          # FIXME: We should calculate (X - x.domain[1]) / x.range once for all configurations
          # and all parameters, then calculate the differences using vectorization.
          d[j] <- max(d[j], abs((as.numeric(X) - as.numeric(Y)) / x_range))
        }
        if (d[j] > threshold) isSimilar[j,i] <- FALSE
      }
    }
    index <- which(rowAlls(isSimilar))
    isSimilar <- isSimilar[index, , drop=FALSE]
    d <- d[index]
    selected  <- selected[index]
    if (nrow(isSimilar) == 0L) break
  }

  if (length(selected) == 0L)
    return(NULL)
  c(x[[".ID."]], configurations[selected,".ID."])
}

##
## Identify which configurations are similar.
##
# FIXME: It would be nice to print the minimum similarity found to the user.
similarConfigurations <- function(configurations, parameters, threshold)
{
  # FIXME: Use data.table
  debug.level <- getOption(".irace.debug.level", 0)
  if (debug.level >= 1) irace_note ("Computing similarity of configurations .")

  # Create vectors of categorical and numerical
  p <- parameters$types %in% c("c","o")
  vecCat <- parameters$names[p & !parameters$isFixed]
  vecNum <- parameters$names[!p & !parameters$isFixed]

  irace_assert(all(parameters$types[vecCat] %in% c("c","o")))
  irace_assert(all(parameters$types[vecNum] %not_in% c("c","o")))
  irace_assert(length(intersect(vecCat, vecNum)) == 0)

  nbCater <- length(vecCat)
  nbNumer <- length(vecNum)

  ### Categorical/Ordinal filtering ####
  if (nbCater > 0) {
    ## Build a vector with the categorical appended together in a string
    strings <- do.call(paste, c(configurations[, vecCat, drop=FALSE], sep = " ; "))

    if (nbNumer != 0) configurations <- configurations[, c(".ID.", vecNum)]
    ord.strings <- order(strings)
    configurations <- configurations[ord.strings, ]
    strings <- strings[ord.strings]

    ## keep similar (index i == true means is the same as i + 1)
    similarIdx <- strings[-length(strings)] == strings[-1]

    ## Now let's get just a FALSE if we remove it, TRUE otherwise:
    keepIdx <- c(similarIdx[1],
                 (similarIdx[-1] | similarIdx[-length(similarIdx)]),
                 similarIdx[length(similarIdx)])

    ## filtering them out:
    configurations <- configurations [keepIdx, , drop=FALSE]
    ## filtering their strings out (to use them to define blocks):
    strings <- strings[keepIdx]

    ## if everything is already filtered out, return
    if (nrow(configurations) == 0) {
      if (debug.level >= 1) cat(" DONE\n")
      return(NULL)
    }
  }

  ### Numerical parameters within blocks of the same string ###
  if (nbNumer > 0) {
    similar <- c()
    if (nbCater > 0) {
      ## In this case the object "string" is available to define blocks
      ## Loop over blocks:
      beginBlock <- 1L
      while (beginBlock < nrow(configurations)) {
        ## The current block is made of all configurations that have the same
        ## categorical string as the one of configuration[beginBlock, ]
        blockIds <- which(strings == strings[beginBlock])
        endBlock <- blockIds[length(blockIds)]

        irace_assert (endBlock > beginBlock)
        ## Loop inside blocks:
        for (i in seq(beginBlock, endBlock - 1L)) {
          ## Compare configuration i with all the ones that are after in the block
          similar <- c(similar,
                       numeric.configurations.equal(configurations[i, ], configurations[(i+1L):endBlock,],
                                                parameters, threshold = threshold, param.names = vecNum))
          if (debug.level >= 1) cat(".")
        }
        beginBlock <- endBlock + 1L # Next block starts after the end of the current one
      }
    } else {
      ## No categorical, so no blocks, just do the basic check without blocks
      for (i in seq_len(nrow(configurations) - 1L)) {
        similar <- c(similar,
                     numeric.configurations.equal(configurations[i, ], configurations[(i+1L):nrow(configurations),],
                                              parameters, threshold = threshold, param.names = vecNum))
        if (debug.level >= 1) cat(".")
      }
    }
    # FIXME: We have to use unique because we return the same configuration
    # more than once in different calls to numeric.configurations.equal.
    # Currently, we compare each configuration k=1...n with every configuration
    # k+1...n. Instead, we should compare k=1...n with ((k+1...n) notin
    # similar).  It may happen that A ~ B and A ~ C and B /= C, but this is OK
    # because we still return A, B and C. It may also happen that A ~ B, B ~ C
    # and A /= C, but this is also OK because we will compare A with B,C then B
    # with C.
    similar <- unique(similar)
    configurations <- configurations[configurations[[".ID."]] %in% similar, ]
  }

  if (debug.level >= 1) cat(" DONE\n")
  if (nrow(configurations) == 0L) return(NULL)
  configurations[[".ID."]]
}


## Number of iterations.
computeNbIterations <- function(nbParameters) (2 + log2(nbParameters))

## Computational budget at each iteration.
computeComputationalBudget <- function(remainingBudget, indexIteration,
                                       nbIterations)
  floor (remainingBudget / (nbIterations - indexIteration + 1L))

## The number of configurations
computeNbConfigurations <- function(currentBudget, indexIteration,
                                    mu, eachTest, blockSize,
                                    nElites = 0L, nOldInstances = 0L, newInstances = 0L,
                                    maxConfigurations = 1024L)
{
  # FIXME: This is slightly incorrect, because we may have elites that have not
  # been executed on all nOldInstances. Thus, we need to pass explicitly the
  # budget that we save (that is, number of entries that are not NA).
  savedBudget <- nElites *  nOldInstances
  eachTest <- eachTest * blockSize
  n <- max (mu + eachTest  * min(5L, indexIteration),
            round_to_next_multiple(nOldInstances + newInstances, eachTest))
  min (floor ((currentBudget + savedBudget) / n), maxConfigurations)
}

## Termination of a race at each iteration. The race will stop if the
## number of surviving configurations is equal or less than this number.
computeTerminationOfRace <- function(nbParameters) (2 + log2(nbParameters))

## Compute the minimum budget required, and exit early in case the
## budget given by the user is insufficient.
computeMinimumBudget <- function(scenario, minSurvival, nbIterations, elitist_new_instances)
{
  blockSize <- scenario$blockSize
  eachTest <- blockSize * scenario$eachTest
  Tnew <- elitist_new_instances # This is already multiplied by blockSize.
  mu <- scenario$mu

  # This is computed from the default formulas as follows:
  #  B_1 = B / I
  #  B_2 = B -  (B/I) / (I - 1) = B / I
  #  B_3 = B - 2(B/I) / (I - 2) = B / I
  # thus
  #  B_i = B / I
  # and
  #  C_i = B_i / T_i = B / (I * T_i).
  #
  # We want to enforce that C_i >= min_surv + 1, thus
  #  B / (I * T_i) >= min_surv + 1                             (1)
  # becomes
  #  B >= (min_surv + 1) * I * T_i
  #
  # This is an over-estimation, since actually B_1 = floor(B/I) and if
  # floor(B/I) < B/I, then B_i < B/I, and we could still satisfy Eq. (1)
  # with a smaller budget. However, the exact formula requires computing B_i
  # taking into account the floor() function, which is not obvious.

  minimumBudget <- (minSurvival + 1L) * nbIterations

  # We need to compute T_i:
  if (scenario$elitist) {
    # T_i = max(mu + Teach * min (5, i),
    #           ceiling((T_{i-1} + Tnew) / Teach) * Teach)
    # T_1 = mu + Teach
    # T_2 ~ mu + Teach + max (Teach, Tnew)
    # T_3 ~ max(mu + 3 * Teach,
    #           mu + Teach + max(Teach, Tnew) + T_new)

    #     = mu + Teach + max(Teach + max(Teach, Tnew), 2 * Tnew)

    # if Teach > Tnew then 2*Teach > 2*Tnew then max = 2*Teach
    # if Teach < Tnew then Teach + Tnew < 2*Tnew then max = 2*Tnew
    # hence: T_3 = mu + Teach + 2 * max(Teach, Tnew)

    # T_4 = max(mu + 4 * Teach,
    #           ceiling((mu + Teach + 2 * max(Teach, Tnew)) + Tnew) / Teach) * Teach)
    #     ~ mu + Teach + max(2 * Teach + max(Teach, Tnew), 3 * Tnew)
    #     = mu + Teach + 3 * max(Teach, Tnew)

    # T_i = mu + Teach + (i - 1) * max(Teach, Tnew)

    # T_6 = max (mu + 5*Teach,
    #            mu + Teach + 5 * max(Teach, Tnew) + Tnew)
    #      = mu + Teach + Tnew + 5 * max (Teach, Tnew)

    # T_i = mu + Teach + max(I-5, 0) * Tnew + 5 * max (Teach, Tnew)
    if (nbIterations > 5L) {
      minimumBudget <- minimumBudget *
        (mu + eachTest + (nbIterations - 5L) * Tnew +  5L * max(eachTest, Tnew))
    } else {
      minimumBudget <- minimumBudget *
        (mu + eachTest + (nbIterations - 1L) * max(eachTest, Tnew))
    }
  } else {
    #   T_i = mu + T_each * min (5, i)
    # and the most strict value is for i >= 5, thus
    #   B >= (min_surv + 1) * I * (mu + 5 * T_each)
    minimumBudget <- minimumBudget * (mu + 5L * eachTest)
  }
  minimumBudget
}

checkMinimumBudget <- function(scenario, remainingBudget, minSurvival, nbIterations,
                               boundEstimate, timeUsed, elitist_new_instances)
{
  minimumBudget <- computeMinimumBudget(scenario, minSurvival, nbIterations, elitist_new_instances)

  if (remainingBudget >= minimumBudget)
    return(TRUE)
  if (scenario$maxTime == 0) {
    irace_error("Insufficient budget: ",
                "With the current settings, irace will require a value of ",
                "'maxExperiments' of at least '",  minimumBudget, "'.")
  } else if (nbIterations == 1L) {
    irace_error("Insufficient budget: ",
                "With the current settings and estimated time per run (",
                boundEstimate, ") irace will require a value of 'maxTime' of at least '",
                ceiling((minimumBudget * boundEstimate) + timeUsed), "'.")
  }
  FALSE
}

## Generate instances + seed.
generateInstances <- function(race_state, scenario, n, update = FALSE)
{
  if (!update)
    race_state$instances_log <- NULL
  # If we are adding and the scenario is deterministic, we have already added all instances.
  else if (scenario$deterministic)
    return(race_state$instances_log)

  instances <- scenario$instances
  # Number of times that we need to repeat the set of instances given by the user.
  ntimes <- if (scenario$deterministic) 1L else
            # "Upper bound"" of instances needed
            # FIXME: We could bound it even further if maxExperiments >> nInstances
            ceiling(n / length(instances))

  # Get instances order
  if (scenario$sampleInstances) {
    blockSize <- scenario$blockSize
    n_blocks <- length(instances) / blockSize
    # Sample instances index in groups (ntimes)
    selected_blocks <- unlist(lapply(rep.int(n_blocks, ntimes), sample.int, replace = FALSE))
    sindex <- c(outer(seq_len(blockSize), (selected_blocks - 1L) * blockSize, "+"))
  } else {
    sindex <- rep.int(seq_along(instances), ntimes)
  }
  # Sample seeds.
  race_state$instances_log <- rbindlist(use.names = TRUE, list(
    race_state$instances_log,
    data.table(instanceID = sindex, seed = runif_integer(length(sindex)))))
  race_state$instances_log
}

do_experiments <- function(race_state, configurations, ninstances, scenario, iteration)
{
  instances <- seq_len(ninstances)
  output <- race_wrapper_helper(race_state = race_state,
    configurations = configurations,
    instance_idx = instances,
    bounds = rep(scenario$boundMax, nrow(configurations)),
    is_exe = rep_len(TRUE, nrow(configurations) * ninstances), scenario = scenario)

  set(output, j = "iteration", value = iteration)
  Results <- race_state$update_experiment_log(output, instances = instances,
    scenario = scenario)

  rejected_ids <- configurations[[".ID."]][colAnys(is.infinite(Results))]
  scenario$parameters$forbid_configurations(
    race_state$update_rejected(rejected_ids, configurations)
  )
  Results
}

## Initialize allConfigurations with any initial configurations provided.
allConfigurationsInit <- function(scenario)
{
  initConfigurations <- scenario$initConfigurations

  confs_from_file <- NULL
  if (!is.null.or.empty(scenario$configurationsFile)) {
    confs_from_file <- readConfigurationsFile(scenario$configurationsFile,
                                              scenario$parameters, scenario$debugLevel)
  }
  if (is.null.or.empty(initConfigurations)) {
    initConfigurations <- confs_from_file
  } else {
    if (!is.null.or.empty(scenario$configurationsFile) && !identical(initConfigurations, confs_from_file))
      irace_warning("'initConfigurations' provided in 'scenario',",
                    " thus ignoring configurations from file '",
                    scenario$configurationsFile, "'.")
    cat("# Adding", nrow(initConfigurations), "initial configuration(s)\n")
    initConfigurations <- fix_configurations(initConfigurations, scenario$parameters, debugLevel = scenario$debugLevel)
  }

  if (is.null.or.empty(initConfigurations)) {
    allConfigurations <- configurations_alloc(c(".ID.", scenario$parameters$names, ".PARENT."),
      nrow = 0L, parameters = scenario$parameters)
    setDF(allConfigurations)
  } else {
    allConfigurations <- cbind(.ID. = seq_nrow(initConfigurations),
                               initConfigurations, .PARENT. = NA_integer_)
    rownames(allConfigurations) <- allConfigurations[[".ID."]]
    num <- nrow(allConfigurations)
    allConfigurations <- checkForbidden(allConfigurations, scenario$parameters$forbidden)
    if (nrow(allConfigurations) < num) {
      irace_warning(num - nrow(allConfigurations), " of the ", num,
                    " initial configurations were forbidden",
                    " and, thus, discarded.")
    }
  }
  allConfigurations
}

## extractElites
# Input: the configurations with the .RANK. field filled.
#        the number of elites wished
# Output: nbElites elites, sorted by ranks, with the weights assigned.
extractElites <- function(configurations, nbElites, debugLevel)
{
  irace_assert(nbElites > 0L)
  # Keep only alive configurations.
  elites <- as.data.table(configurations)
  before <- nrow(elites)
  # Remove duplicated. Duplicated configurations may be generated, however, it
  # is too slow to check at generation time. Nevertheless, we can check now
  # since we typically have very few elites.
  elites <- unique(elites, by=which(!startsWith(colnames(elites), ".")))
  after <- nrow(elites)
  if (debugLevel >= 2L && after < before)
    irace_note("Dropped ", before - after, " duplicated elites.\n")

  after <- min(after, nbElites)
  setorderv(elites, cols=".RANK.")
  selected <- seq_len(after)
  elites <- elites[selected, ]
  set(elites, j = ".WEIGHT.", value = ((after + 1L) - selected) / (after * (after + 1L) / 2))
  setDF(elites)
  rownames(elites) <- elites[[".ID."]]
  elites
}

#' Execute one run of the Iterated Racing algorithm.
#'
#' The function `irace` implements the Iterated Racing procedure for parameter
#' tuning. It receives a configuration scenario and a parameter space to be
#' tuned, and returns the best configurations found, namely, the elite
#' configurations obtained from the last iterations. As a first step, it checks
#' the correctness of `scenario` using [checkScenario()] and recovers a
#' previous execution if `scenario$recoveryFile` is set. A R data file log of
#' the execution is created in `scenario$logFile`.
#'
#' The execution of this function is reproducible under some conditions. See
#' the FAQ section in the [User
#' Guide](https://cran.r-project.org/package=irace/vignettes/irace-package.pdf).
#'
#' @inheritParams defaultScenario
#'
#' @template return_irace
#'
#' @examples
#' \dontrun{
#' # In general, there are three steps:
#' scenario <- readScenario(filename = "scenario.txt")
#' irace(scenario = scenario)
#' }
#' #######################################################################
#' # This example illustrates how to tune the parameters of the simulated
#' # annealing algorithm (SANN) provided by the optim() function in the
#' # R base package.  The goal in this example is to optimize instances of
#' # the following family:
#' #      f(x) = lambda * f_rastrigin(x) + (1 - lambda) * f_rosenbrock(x)
#' # where lambda follows a normal distribution whose mean is 0.9 and
#' # standard deviation is 0.02. f_rastrigin and f_rosenbrock are the
#' # well-known Rastrigin and Rosenbrock benchmark functions (taken from
#' # the cmaes package). In this scenario, different instances are given
#' # by different values of lambda.
#' #######################################################################
#' ## First we provide an implementation of the functions to be optimized:
#' f_rosenbrock <- function (x) {
#'   d <- length(x)
#'   z <- x + 1
#'   hz <- z[1L:(d - 1L)]
#'   tz <- z[2L:d]
#'   sum(100 * (hz^2 - tz)^2 + (hz - 1)^2)
#' }
#' f_rastrigin <- function (x) {
#'   sum(x * x - 10 * cos(2 * pi * x) + 10)
#' }
#'
#' ## We generate 20 instances (in this case, weights):
#' weights <- rnorm(20, mean = 0.9, sd = 0.02)
#'
#' ## On this set of instances, we are interested in optimizing two
#' ## parameters of the SANN algorithm: tmax and temp. We setup the
#' ## parameter space as follows:
#' parameters_table <- '
#'   tmax "" i,log (1, 5000)
#'   temp "" r (0, 100)
#'   '
#' ## We use the irace function readParameters to read this table:
#' parameters <- readParameters(text = parameters_table)
#'
#' ## Next, we define the function that will evaluate each candidate
#' ## configuration on a single instance. For simplicity, we restrict to
#' ## three-dimensional functions and we set the maximum number of
#' ## iterations of SANN to 1000.
#' target_runner <- function(experiment, scenario)
#' {
#'     instance <- experiment$instance
#'     configuration <- experiment$configuration
#'
#'     D <- 3
#'     par <- runif(D, min=-1, max=1)
#'     fn <- function(x) {
#'       weight <- instance
#'       return(weight * f_rastrigin(x) + (1 - weight) * f_rosenbrock(x))
#'     }
#'     # For reproducible results, we should use the random seed given by
#'     # experiment$seed to set the random seed of the target algorithm.
#'     res <- withr::with_seed(experiment$seed,
#'                      stats::optim(par,fn, method="SANN",
#'                                   control=list(maxit=1000
#'                                              , tmax = as.numeric(configuration[["tmax"]])
#'                                              , temp = as.numeric(configuration[["temp"]])
#'                                                )))
#'     ## This list may also contain:
#'     ## - 'time' if irace is called with 'maxTime'
#'     ## - 'error' is a string used to report an error
#'     ## - 'outputRaw' is a string used to report the raw output of calls to
#'     ##   an external program or function.
#'     ## - 'call' is a string used to report how target_runner called the
#'     ##   external program or function.
#'     return(list(cost = res$value))
#' }
#'
#' ## We define a configuration scenario by setting targetRunner to the
#' ## function define above, instances to the first 10 random weights, and
#' ## a maximum budget of 'maxExperiments' calls to targetRunner.
#' scenario <- list(targetRunner = target_runner,
#'                  instances = weights[1:10],
#'                  maxExperiments = 500,
#'                  # Do not create a logFile
#'                  logFile = "",
#'                  parameters = parameters)
#'
#' ## We check that the scenario is valid. This will also try to execute
#' ## target_runner.
#' checkIraceScenario(scenario)
#'
#' \donttest{
#' ## We are now ready to launch irace. We do it by means of the irace
#' ## function. The function will print information about its
#' ## progress. This may require a few minutes, so it is not run by default.
#' tuned_confs <- irace(scenario = scenario)
#'
#' ## We can print the best configurations found by irace as follows:
#' configurations_print(tuned_confs)
#'
#' ## We can evaluate the quality of the best configuration found by
#' ## irace versus the default configuration of the SANN algorithm on
#' ## the other 10 instances previously generated.
#' test_index <- 11:20
#' test_seeds <- sample.int(2147483647L, size = length(test_index), replace = TRUE)
#' test <- function(configuration)
#' {
#'   res <- lapply(seq_along(test_index),
#'                 function(x) target_runner(
#'                               experiment = list(instance = weights[test_index[x]],
#'                                                 seed = test_seeds[x],
#'                                                 configuration = configuration),
#'                               scenario = scenario))
#'   return (sapply(res, getElement, name = "cost"))
#' }
#' ## To do so, first we apply the default configuration of the SANN
#' ## algorithm to these instances:
#' default <- test(data.frame(tmax=10, temp=10))
#'
#' ## We extract and apply the winning configuration found by irace
#' ## to these instances:
#' tuned <- test(removeConfigurationsMetaData(tuned_confs[1,]))
#'
#' ## Finally, we can compare using a boxplot the quality obtained with the
#' ## default parametrization of SANN and the quality obtained with the
#' ## best configuration found by irace.
#' boxplot(list(default = default, tuned = tuned))
#' }
#' @seealso
#'  \describe{
#'  \item{[irace_main()]}{a higher-level interface to [irace()].}
#'  \item{[irace_cmdline()]}{a command-line interface to [irace()].}
#'  \item{[readScenario()]}{for reading a configuration scenario from a file.}
#'  \item{[readParameters()]}{read the target algorithm parameters from a file.}
#'  \item{[defaultScenario()]}{returns the default scenario settings of \pkg{irace}.}
#'  \item{[checkScenario()]}{to check that the scenario is valid.}
#' }
#' @author Manuel López-Ibáñez and Jérémie Dubois-Lacoste
#' @concept running
#' @export
irace <- function(scenario)
  irace_common(scenario, simple = TRUE)

irace_common <- function(scenario, simple, output.width = 9999L)
{
  if (!simple) {
    withr::local_options(width = output.width) # Do not wrap the output.
  }
  scenario <- checkScenario(scenario)
  debugLevel <- scenario$debugLevel

  if (debugLevel >= 1L) {
    op <- list(warning.length = 8170L)
    if (!base::interactive())
      op <- c(op, list(error = irace_dump_frames))
    withr::local_options(op)
    printScenario (scenario)
  }

  elite_configurations <- irace_run(scenario = scenario)
  if (simple) return(elite_configurations)

  if (!scenario$quiet) {
    order_str <- test.type.order.str(scenario$testType)
    cat("# Best configurations (first number is the configuration ID;",
        " listed from best to worst according to the ", order_str, "):\n", sep = "")
    configurations_print(elite_configurations)

    cat("# Best configurations as commandlines (first number is the configuration ID;", " listed from best to worst according to the ", order_str, "):\n", sep = "")
    configurations_print_command (elite_configurations, scenario$parameters)
  }
  testing_fromlog(logFile = scenario$logFile)
  invisible(elite_configurations)
}

irace_run <- function(scenario)
{
  # Recover state from file?
  if (is.null.or.empty(scenario$recoveryFile)) {
    race_state <- RaceState$new(scenario)
  } else {
    irace_note ("Recovering from file: '", scenario$recoveryFile,"'\n")
    race_state <- recoverFromFile(scenario$recoveryFile, scenario = scenario)
  }

  quiet <- scenario$quiet
  catInfo <- if (quiet) do_nothing else function(..., verbose = TRUE) {
    irace_note (..., "\n")
    if (verbose) {
      cat ("# Iteration: ", indexIteration, "\n",
           "# nbIterations: ", nbIterations, "\n",
           "# experimentsUsed: ", experimentsUsed, "\n",
           "# timeUsed: ", timeUsed, "\n",
           "# remainingBudget: ", remainingBudget, "\n",
           "# currentBudget: ", currentBudget, "\n",
           "# number of elites: ", nrow(elite_configurations), "\n",
           "# nbConfigurations: ", nbConfigurations, "\n",
           sep = "")
    }
  }
  # FIXME: use this from psrace?
  irace_finish <- function(iraceResults, scenario, reason) {
    elapsed <- race_state$time_elapsed()
    if (!scenario$quiet)
      cat("# Total CPU user time: ", elapsed["user"], ", CPU sys time: ", elapsed["system"],
          ", Wall-clock time: ", elapsed["wallclock"], "\n", sep="")
    # FIXME: Do we need to clone?
    race_state$completed <- reason
    iraceResults$state <- race_state
    save_irace_logfile(iraceResults, logfile = scenario$logFile)
    # FIXME: Handle scenario$maxTime > 0
    if (scenario$postselection && scenario$maxTime == 0 && floor(remainingBudget / max(scenario$blockSize, scenario$eachTest)) > 1L)
      psRace(iraceResults, max_experiments = remainingBudget, iteration_elites = TRUE)
    else
      elite_configurations
  }

  debugLevel <- scenario$debugLevel
  # Set options controlling debug level.
  # FIXME: This should be the other way around, the options set the debugLevel.
  options(.race.debug.level = debugLevel)
  options(.irace.debug.level = debugLevel)

  firstRace <- TRUE
  # Create a data frame of all configurations ever generated.
  allConfigurations <- allConfigurationsInit(scenario)
  irace_assert(is.integer(allConfigurations[[".ID."]]))
  nbUserConfigurations <- nrow(allConfigurations)

  # To save the logs
  iraceResults <- list(
    scenario = scenario,
    irace_version = irace_version,
    iterationElites = c(),
    allElites = list(),
    experiments = matrix(nrow = 0L, ncol = 0L))
  model <- NULL
  nbConfigurations <- 0L
  elite_configurations <- data.frame(stringsAsFactors=FALSE)

  nbIterations <-  if (scenario$nbIterations == 0)
                     computeNbIterations(scenario$parameters$nbVariable)
                   else scenario$nbIterations
  nbIterations <- floor(nbIterations)

  minSurvival <- if (scenario$minNbSurvival == 0)
                   computeTerminationOfRace(scenario$parameters$nbVariable)
                 else scenario$minNbSurvival
  minSurvival <- floor(minSurvival)
  # FIXME: Do this initialization within race_state.
  race_state$minSurvival <- minSurvival

  # Generate initial instance + seed list
  generateInstances(race_state, scenario,
    n = if (scenario$maxExperiments != 0) ceiling(scenario$maxExperiments / minSurvival)
        else max(scenario$firstTest, length(scenario$instances)))
  indexIteration <- 1L
  experimentsUsed <- 0L
  timeUsed <- 0
  boundEstimate <- NA

  race_state$start_parallel(scenario)
  on.exit(race_state$stop_parallel(), add = TRUE)

  if (scenario$maxTime == 0) {
    if (is.na(scenario$minExperiments)) {
      remainingBudget <- scenario$maxExperiments
    } else {
      remainingBudget <- max(scenario$minExperiments,
        computeMinimumBudget(scenario, minSurvival, nbIterations,
          race_state$elitist_new_instances))
    }
  } else { ## Estimate time when maxTime is defined.
    ## IMPORTANT: This is firstTest because these configurations will be
    ## considered elite later, thus preserved up to firstTest, which is
    ## fine. If a larger number of instances is used, it would prevent
    ## discarding these configurations.
    # Get the number of instances to be used.
    ninstances <- scenario$firstTest * scenario$blockSize
    estimationTime <- ceiling(scenario$maxTime * scenario$budgetEstimation)
    irace_note("Estimating execution time using ", 100 * scenario$budgetEstimation,
      "% of ", scenario$maxTime, " = ", estimationTime, "\n")

    # Estimate the number of configurations to be used
    nconfigurations <- max(2L, floor(scenario$parallel / ninstances))
    next_configuration <- 1L
    nruns <- nconfigurations * ninstances
    boundEstimate <- if (is.null(scenario$boundMax)) 1.0 else scenario$boundMax
    if (estimationTime < boundEstimate * nruns) {
      boundEstimate <- max(ceiling_digits(estimationTime / nruns, scenario$boundDigits), scenario$minMeasurableTime)
      if (!is.null(scenario$boundMax)) {
          irace_warning("boundMax=", scenario$boundMax, " is too large, using ", boundEstimate, " instead.\n")
          # FIXME: We should not modify the scenario
          scenario$boundMax <- boundEstimate
      }
    }

    repeat {
      # Sample new configurations if needed
        if (nrow(allConfigurations) < nconfigurations) {
          newConfigurations <- sampleSobol(scenario$parameters,
            nconfigurations - nrow(allConfigurations),
            repair = scenario$repairConfiguration)
          set(newConfigurations, j = ".ID.",
            value = max(0L, vlast(allConfigurations[[".ID."]])) + seq_nrow(newConfigurations))
          setcolorder(newConfigurations, ".ID.", before=1L)
          setDF(newConfigurations)
          # FIXME: use rbindlist(use.names=TRUE)
          allConfigurations <- rbind(allConfigurations, newConfigurations)
          rownames(allConfigurations) <- allConfigurations[[".ID."]]
          # We may have generated less than the number requested if there were duplicates.
          nconfigurations <- nrow(allConfigurations)
        }
      # Estimate the mean execution time.
      # FIXME: Shouldn't we pass the bounds?
      experiments <- do_experiments(race_state,
        configurations = allConfigurations[next_configuration:nconfigurations, ],
        ninstances = ninstances, scenario = scenario,
        # These experiments are assigned iteration 0.
        iteration = 0L)
      # FIXME: Here we should check if everything timed out and increase the bound dynamically.
      iraceResults$experiments <- merge_matrix(iraceResults$experiments, experiments)
      rownames(iraceResults$experiments) <- seq_nrow(iraceResults$experiments)
      # For the used time, we count the time reported in all configurations
      # including rejected ones.
      timeUsed <- sum(race_state$experiment_log[["time"]], na.rm = TRUE)
      experimentsUsed <- nrow(race_state$experiment_log)
      # User should return time zero for rejected_ids.
      boundEstimate <- timeUsed / experimentsUsed
      boundEstimate <- max(ceiling_digits(boundEstimate, scenario$boundDigits), scenario$minMeasurableTime)
      next_configuration <- nconfigurations + 1L

      # Calculate how many new configurations:
      # 1. We do not want to overrun estimationTime
      new_conf <- floor(((estimationTime - timeUsed) / boundEstimate) / ninstances)
      # 2. But there is no point in executing more configurations than those
      # that we can execute in parallel.
      new_conf <- min(new_conf, max(1L, floor(scenario$parallel / ninstances)))
      if (timeUsed >= estimationTime || new_conf == 0L || nconfigurations == 1024L)
        break
      else
        nconfigurations <- min(1024L, nconfigurations + new_conf)
    } # end of repeat

    if (length(race_state$rejected_ids))
      irace_note ("Immediately rejected configurations: ",
        paste0(race_state$rejected_ids, collapse = ", ") , "\n")

    # Update budget
    remainingBudget <- round((scenario$maxTime - timeUsed) / boundEstimate)
    elite_configurations <- allConfigurations[allConfigurations[[".ID."]] %not_in% race_state$rejected_ids, , drop = FALSE]
    irace_assert(is.integer(elite_configurations[[".ID."]]))
    # Without elitist, the racing does not re-use the results computed during
    # the estimation.  This means that the time used during estimation needs
    # to be spent again during racing, thus leaving less time for racing.  We
    # want to avoid having less time for racing, and this is an
    # implementation detail, thus we assume that the time was not actually
    # wasted.
    if (!scenario$elitist) timeUsed <- 0
    irace_note("Estimated execution time is ", boundEstimate, " based on ",
      next_configuration - 1L, " configurations and ",
      ninstances," instances. Used time: ", timeUsed,
      ", remaining time: ", (scenario$maxTime - timeUsed),
      ", remaining budget (experiments): ", remainingBudget, "\n")
    if (!is.null(scenario$boundMax) && 2 * boundEstimate < scenario$boundMax) {
      irace_warning("boundMax=", scenario$boundMax, " is much larger than estimated execution time, using ",
        2 * boundEstimate, " instead.\n")
      scenario$boundMax <- 2 * boundEstimate
    }
  } # end of time estimation

  # Compute the total initial budget, that is, the maximum number of
  # experiments that we can perform.
  currentBudget <- if (scenario$nbExperimentsPerIteration == 0L)
                     computeComputationalBudget(remainingBudget, indexIteration,
                       nbIterations)
                   else scenario$nbExperimentsPerIteration

  # Check that the budget is enough. For the time estimation case we reduce
  # the number of iterations.
  warn_msg <- NULL
  while (!checkMinimumBudget(scenario, remainingBudget, minSurvival, nbIterations,
    boundEstimate, timeUsed, race_state$elitist_new_instances))
  {
    if (is.null(warn_msg))
      warn_msg <-
        paste0("With the current settings and estimated time per run (",
          boundEstimate,
          ") irace will not have enough budget to execute the minimum",
          " number of iterations (", nbIterations, "). ",
          "Execution will continue by assuming that the estimated time",
          " is too high and reducing the minimum number of iterations,",
          " however, if the estimation was correct or too low,",
          " results might not be better than random sampling.\n")
    nbIterations <- nbIterations - 1L
    scenario$nbConfigurations <- if (scenario$nbConfigurations > 0L)
                                   min(minSurvival * 2L, scenario$nbConfigurations)
                                 else minSurvival * 2L
  }
  if (!is.null(warn_msg))
    irace_warning(warn_msg)

  catInfo("Initialization\n",
          if (scenario$elitist)
            paste0("# Elitist race\n",
                   "# Elitist new instances: ", race_state$elitist_new_instances, "\n",
                   "# Elitist limit: ",         scenario$elitistLimit, "\n")
          else paste0("# Non-elitist race\n"),
          "# nbIterations: ", nbIterations, "\n",
          "# minNbSurvival: ", minSurvival, "\n",
          "# nbParameters: ", scenario$parameters$nbVariable, "\n",
          "# seed: ", race_state$seed, "\n",
          "# confidence level: ", scenario$confidence, "\n",
          "# budget: ", remainingBudget, "\n",
          if (scenario$maxTime == 0) ""
          else paste0("# time budget: ", scenario$maxTime - timeUsed, "\n"),
          "# mu: ", scenario$mu, "\n",
          "# deterministic: ", scenario$deterministic, "\n",

          if (scenario$capping)
            paste0("# capping: ", scenario$cappingType, "\n",
                   "# type bound: ", scenario$boundType, "\n",
                   "# boundMax: ", scenario$boundMax, "\n",
                   "# par bound: ", scenario$boundPar, "\n",
                   "# bound digits: ", scenario$boundDigits, "\n")
          else if (!is.null(scenario$boundMax))
            paste0("# boundMax: ", scenario$boundMax, "\n"),
          verbose = FALSE)

  blockSize <- scenario$blockSize

  repeat {
    # FIXME: We could directly use race_state$timeUsed everywhere.
    race_state$timeUsed <- timeUsed
    ## Save to the log file.
    iraceResults$allConfigurations <- allConfigurations
    race_state$save_recovery(iraceResults, logfile = scenario$logFile)


    # With elitist=TRUE and without targetEvaluator we should never re-run the
    # same configuration on the same (instance,seed) pair.
    if (scenario$elitist) {
      irace_assert(sum(!is.na(iraceResults$experiments)) == experimentsUsed)
      if (is.null(scenario$targetEvaluator))
        irace_assert(experimentsUsed == nrow(race_state$experiment_log))
    }

    if (remainingBudget <= 0) {
      catInfo("Stopped because budget is exhausted")
      return(irace_finish(iraceResults, scenario, reason = "Budget exhausted"))
    }
    if (scenario$maxTime > 0 && timeUsed >= scenario$maxTime) {
      catInfo("Stopped because time budget is exhausted")
      return(irace_finish(iraceResults, scenario, reason = "Time budget exhausted"))
    }

    if (indexIteration > nbIterations) {
      if (scenario$nbIterations == 0L) {
        nbIterations <- indexIteration
      } else {
        if (debugLevel >= 1L)
          catInfo("Limit of iterations reached", verbose = FALSE)
        return(irace_finish(iraceResults, scenario, reason = "Limit of iterations reached"))
      }
    }
    # Compute the current budget (nb of experiments for this iteration),
    # or take the value given as parameter.
    currentBudget <- if (scenario$nbExperimentsPerIteration == 0L)
                       computeComputationalBudget(remainingBudget, indexIteration, nbIterations)
                     else scenario$nbExperimentsPerIteration

    # Compute the number of configurations for this race.
    if (scenario$elitist && !firstRace) {
      nbConfigurations <-
        computeNbConfigurations(currentBudget, indexIteration,
                                mu = scenario$mu,
                                eachTest = scenario$eachTest,
                                blockSize = blockSize,
                                nElites = nrow(elite_configurations),
                                nOldInstances = nrow(iraceResults$experiments),
                                newInstances = race_state$elitist_new_instances)
      # If we don't have enough budget, do not evaluate new instances.
      if (nbConfigurations <= minSurvival) {
        race_state$elitist_new_instances <- 0L
        nbConfigurations <- computeNbConfigurations(currentBudget, indexIteration,
          mu = scenario$mu,
          eachTest = scenario$eachTest,
          blockSize = blockSize,
          nElites = nrow(elite_configurations),
          nOldInstances = nrow(iraceResults$experiments),
          newInstances = race_state$elitist_new_instances)
      }
      # If still not enough budget, then try to do at least one test.
      if (nbConfigurations <= minSurvival) {
        nbConfigurations <- computeNbConfigurations(currentBudget, indexIteration = 1L,
          mu = 1L, eachTest = scenario$eachTest, blockSize = blockSize,
          nElites = nrow(elite_configurations),
          nOldInstances = nrow(iraceResults$experiments),
          newInstances = 0L)
      }
    } else {
      nbConfigurations <-
        computeNbConfigurations(currentBudget, indexIteration,
                                mu = scenario$mu,
                                eachTest = scenario$eachTest,
                                blockSize = blockSize,
                                nElites = 0L, nOldInstances = 0L,
                                newInstances = 0L)
    }

    # If a value was given as a parameter, then this value limits the maximum,
    # but if we have budget only for less than this, then we have run out of
    # budget.
    if (scenario$nbConfigurations > 0L) {
      if (scenario$nbConfigurations <= nbConfigurations) {
        nbConfigurations <- scenario$nbConfigurations
      } else if (currentBudget < remainingBudget) {
        # We skip one iteration
        catInfo("Not enough budget for this iteration, skipping to the next one.")
        indexIteration <- indexIteration + 1L
        next
      } else {
        catInfo("Stopped because ",
          "there is not enough budget to enforce the value of nbConfigurations.")
        return(irace_finish(iraceResults, scenario, reason = "Not enough budget to enforce the value of nbConfigurations"))
      }
    }

    # Stop if the number of configurations to test is NOT larger than the minimum.
    if (nbConfigurations <= minSurvival) {
      catInfo("Stopped because there is not enough budget left to race more than ",
              "the minimum (", minSurvival,").\n",
              "# You may either increase the budget or set 'minNbSurvival' to a lower value.")
      return(irace_finish(iraceResults, scenario, reason = "Not enough budget to race more than the minimum configurations"))
    }


    # If we have too many elite_configurations, reduce their number. This can
    # happen before the first race due to the initial budget estimation.
    if (firstRace) {
      if (nbConfigurations < nrow(elite_configurations)) {
        eliteRanks <- overall_ranks(iraceResults$experiments, test = scenario$testType)
        elite_configurations <- elite_configurations[order(eliteRanks), ]
        elite_configurations <- elite_configurations[seq_len(nbConfigurations), ]
      }
    } else if (nbConfigurations <= nrow(elite_configurations)) {
      # Stop if  the number of configurations to produce is not greater than
      # the number of elites.
      catInfo("Stopped because ",
              "there is not enough budget left to race newly sampled configurations.")
      #(number of elites  + 1) * (mu + min(5, indexIteration)) > remainingBudget"
      return(irace_finish(iraceResults, scenario, reason = "Not enough budget left to race newly sampled configurations"))
    }

    if (scenario$elitist) {
      # The non-elite have to run up to the first test. The elites consume
      # budget at most up to the new instances.
      if ((nbConfigurations - nrow(elite_configurations)) * scenario$mu
          + nrow(elite_configurations) * min(race_state$elitist_new_instances, scenario$mu)
          > currentBudget) {
        catInfo("Stopped because there is not enough budget left to race all configurations up to the first test (or mu).")
        return(irace_finish(iraceResults, scenario, reason = "Not enough budget to race all configurations up to the first test (or mu)"))
      }
    } else if (nbConfigurations * scenario$mu > currentBudget) {
      catInfo("Stopped because there is not enough budget left to race all configurations up to the first test (or mu).")
      return(irace_finish(iraceResults, scenario, reason = "Not enough budget to race all configurations up to the first test (or mu)"))
    }

    catInfo("Iteration ", indexIteration, " of ", nbIterations, "\n",
            "# experimentsUsed: ", experimentsUsed, "\n",
            if (scenario$maxTime == 0) ""
            else paste0("# timeUsed: ", timeUsed, "\n",
                        "# boundEstimate: ", boundEstimate, "\n"),
            "# remainingBudget: ", remainingBudget, "\n",
            "# currentBudget: ", currentBudget, "\n",
            "# nbConfigurations: ", nbConfigurations,
            verbose = FALSE)

    iraceResults$softRestart[indexIteration] <- FALSE
    # Sample for the first time.
    if (firstRace) {
      # If we need more configurations, sample uniformly.
      nbNewConfigurations <- nbConfigurations - sum(allConfigurations[[".ID."]] %not_in% race_state$rejected_ids)
      if (nbNewConfigurations > 0L) {
        # Sample new configurations.
        if (debugLevel >= 1L) {
          catInfo("Sample ", nbNewConfigurations,
                  " configurations from Sobol distribution", verbose = FALSE)
        }
        newConfigurations <- sampleSobol(scenario$parameters, nbNewConfigurations,
          repair = scenario$repairConfiguration)
        # We could get fewer than we asked for due to removing duplicates and forbidden.
        nbNewConfigurations <- nrow(newConfigurations)
        set(newConfigurations, j= ".ID.",
          value = max(0L, vlast(allConfigurations[[".ID."]])) + seq_nrow(newConfigurations))
        setcolorder(newConfigurations, ".ID.", before=1L)
        setDF(newConfigurations)
        allConfigurations <- rbind(allConfigurations, newConfigurations)
        rownames(allConfigurations) <- allConfigurations[[".ID."]]
        raceConfigurations <- allConfigurations[allConfigurations[[".ID."]] %not_in% race_state$rejected_ids, , drop = FALSE]
      } else if (nbNewConfigurations <= 0L) {
        # We let the user know that not all configurations will be used.
        if (nbUserConfigurations > nbConfigurations) {
          catInfo("Only ", nbConfigurations, " from the initial configurations will be used",
            verbose = FALSE)
        }

        # This is made only in case that the number of configurations used in
        # the time estimation is more than needed.
        if (nrow(elite_configurations) == nbConfigurations) {
          raceConfigurations <- elite_configurations
        } else {
          raceConfigurations <- allConfigurations[allConfigurations[[".ID."]] %not_in% race_state$rejected_ids, , drop = FALSE]
          raceConfigurations <- raceConfigurations[seq_len(nbConfigurations), , drop = FALSE]
        }
      } # end of indexIteration == 1

    } else {
      # How many new configurations should be sampled?
      nbNewConfigurations <- nbConfigurations - nrow(elite_configurations)

      # Update the model based on elites configurations
      if (debugLevel >= 1L) irace_note("Update model\n")
      model <- updateModel(scenario$parameters, elite_configurations, model, indexIteration,
                           nbIterations, nbNewConfigurations, elitist = scenario$elitist)
      if (debugLevel >= 2L) printModel (model)
      if (debugLevel >= 1L)
        irace_note("Sample ", nbNewConfigurations, " configurations from model\n")
      irace_assert(is.integer(elite_configurations[[".ID."]]))
      newConfigurations <- sampleModel(scenario$parameters, elite_configurations,
                                       model, nbNewConfigurations,
                                       repair = scenario$repairConfiguration)
      # Set ID of the new configurations.
      set(newConfigurations, j = ".ID.",
        value = vlast(allConfigurations[[".ID."]]) + seq_nrow(newConfigurations))
      setcolorder(newConfigurations, ".ID.", before=1L)
      setDF(newConfigurations)
      raceConfigurations <- rbind(elite_configurations[, colnames(newConfigurations)],
        newConfigurations)
      rownames(raceConfigurations) <- raceConfigurations[[".ID."]]

      if (scenario[["softRestart"]]) {
        #          Rprof("profile.out")
        restart_ids <- similarConfigurations (raceConfigurations, scenario$parameters,
                                              threshold = scenario$softRestartThreshold)
        #          Rprof(NULL)
        if (!is.null(restart_ids)) {
          if (debugLevel >= 1L)
            irace_note("Soft restart: ", paste(collapse = " ", restart_ids), " !\n")
          model <- restartModel(model, raceConfigurations, restart_ids,
                                scenario$parameters, nbNewConfigurations)
          iraceResults$softRestart[indexIteration] <- TRUE
          if (debugLevel >= 2L) { printModel (model) }
          # Re-sample after restart like above
          #cat("# ", format(Sys.time(), usetz=TRUE), " sampleModel()\n")
          newConfigurations <- sampleModel(scenario$parameters, elite_configurations,
                                           model, nbNewConfigurations,
                                           repair = scenario$repairConfiguration)
          #cat("# ", format(Sys.time(), usetz=TRUE), " sampleModel() DONE\n")
          # Set ID of the new configurations.
          # Set ID of the new configurations.
          set(newConfigurations, j = ".ID.",
            value = vlast(allConfigurations[[".ID."]]) + seq_nrow(newConfigurations))
          setcolorder(newConfigurations, ".ID.", before=1L)
          setDF(newConfigurations)
          raceConfigurations <- rbind(elite_configurations[, colnames(newConfigurations)],
                                      newConfigurations)
          rownames(raceConfigurations) <- raceConfigurations[[".ID."]]
        }
      }
      # Append these configurations to the global table.
      allConfigurations <- rbind(allConfigurations, newConfigurations)
      rownames(allConfigurations) <- allConfigurations[[".ID."]]
    }

    if (debugLevel >= 2L) {
      irace_note("Configurations for the race n ", indexIteration,
                 " (elite configurations listed first, then new configurations):\n")
      configurations_print(raceConfigurations, metadata = TRUE)
    }

    # Get data from previous races.
    elite_data <- if (scenario$elitist && nrow(elite_configurations))
                    iraceResults$experiments[, as.character(elite_configurations[[".ID."]]), drop=FALSE]
                  else NULL

    race_state$next_instance <- nrow(iraceResults$experiments) + 1L
    # Add instances if needed.
    # Calculate budget needed for old instances assuming non elitist irace.
    if ((nrow(race_state$instances_log) - (race_state$next_instance - 1L))
        < ceiling(remainingBudget / minSurvival)) {
      generateInstances(race_state, scenario, n = ceiling(remainingBudget / minSurvival),
        update = TRUE)
    }

    if (debugLevel >= 1L) irace_note("Launch race\n")
    raceResults <- elitist_race (race_state, scenario = scenario,
                                 configurations = raceConfigurations,
                                 maxExp = currentBudget,
                                 minSurvival = minSurvival,
                                 elite_data = elite_data,
                                 elitist_new_instances = if (firstRace) 0L
                                                         else race_state$elitist_new_instances)
    # We add indexIteration as an additional column.
    set(raceResults$experiment_log, j = "iteration", value = indexIteration)
    # FIXME: There is a chance that the process stops after we remove
    # race_experiment_log in elitist_race(), but before we update
    # race_state$experiment_log here. Doing the two steps in a different order
    # would be more robust but would need a smarter recovery routine that
    # checks for duplicates.
    race_state$experiment_log <- rbindlist(list(race_state$experiment_log, raceResults$experiment_log),
      use.names=TRUE)
    # Merge new results.
    iraceResults$experiments <- merge_matrix(iraceResults$experiments, raceResults$experiments)

    # Update remaining budget.
    experimentsUsed <- experimentsUsed + raceResults$experimentsUsed
    if (scenario$maxTime > 0L) {
      timeUsed <- timeUsed + sum(raceResults$experiment_log[["time"]], na.rm = TRUE)
      boundEstimate <- timeUsed / experimentsUsed
      remainingBudget <- round((scenario$maxTime - timeUsed) / boundEstimate)
    } else {
      remainingBudget <- remainingBudget - raceResults$experimentsUsed
    }

    if (debugLevel >= 3L) {
      irace_note("Results for the race of iteration ", indexIteration,
                 " (from best to worst, according to the ",
                 test.type.order.str(scenario$testType), "):\n")
      configurations_print(raceResults$configurations, metadata = TRUE)
    }

    if (debugLevel >= 1L) irace_note("Extracting elites\n")
    elite_configurations <- extractElites(raceResults$configurations,
      nbElites = minSurvival, debugLevel = scenario$debugLevel)
    irace_note("Elite configurations (first number is the configuration ID;",
               " listed from best to worst according to the ",
               test.type.order.str(scenario$testType), "):\n")
    if (!quiet) configurations_print(elite_configurations, metadata = debugLevel >= 1L)
    iraceResults$iterationElites[indexIteration] <- elite_configurations[[".ID."]][1L]
    iraceResults$allElites[[indexIteration]] <- elite_configurations[[".ID."]]

    if (firstRace) {
      if (debugLevel >= 1L) irace_note("Initialise model\n")
      model <- initialiseModel(scenario$parameters, elite_configurations)
      if (debugLevel >= 2L) printModel (model)
      firstRace <- FALSE
    }

    if (debugLevel >= 1L)  {
      irace_note("End of iteration ", indexIteration, "\n")
      if (debugLevel >= 3L) {
        irace_note("All configurations (sampling order):\n")
        configurations_print(allConfigurations, metadata = TRUE)
        irace_note("Memory used in irace():\n")
        race_state$print_mem_used()
      }
    }
    indexIteration <- indexIteration + 1L
  } # end of repeat
}

#' The irace package: \packageTitle{irace}
#'
#' \packageDescription{irace}
#'
#' @name irace-package
#' @import stats matrixStats withr data.table
#' @importFrom utils str
#' @importFrom R6 R6Class
#' @importFrom graphics abline axis boxplot par plot points strwidth bxp grid
#' @importFrom spacefillr generate_sobol_set
#'
#' @details  License: GPL (>= 2)
#'
#' @author Maintainers: Manuel López-Ibáñez and Leslie Pérez Cáceres
#'         \email{irace-package@googlegroups.com}
#'
#' @keywords package optimize tuning automatic configuration
#'
#' @references
#'   Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres,
#'   Thomas Stützle, and Mauro Birattari. The irace package: Iterated
#'   Racing for Automatic Algorithm Configuration. \emph{Operations Research
#'   Perspectives}, 2016. \doi{10.1016/j.orp.2016.09.002}
#'
#'   Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Thomas Stützle, and Mauro
#'   Birattari. \emph{The irace package, Iterated Race for Automatic
#'   Algorithm Configuration}. Technical Report TR/IRIDIA/2011-004, IRIDIA,
#'   Université Libre de Bruxelles, Belgium, 2011.
#'
#'   Manuel López-Ibáñez and Thomas Stützle. The Automatic Design of
#'   Multi-Objective Ant Colony Optimization Algorithms. \emph{IEEE Transactions
#'   on Evolutionary Computation}, 2012.
#' @seealso
#'  [irace()] for examples and `vignette(package = "irace")` for the user-guide.
"_PACKAGE"


# This silences CRAN NOTE:
#   Namespace in Imports field not imported from: 'codetools'
#     All declared Imports should be used.
.ignore_unused_imports <- function() { # nocov start
  codetools::findGlobals
} # nocov end

#' Execute the given configurations on the testing instances specified in the
#' scenario
#'
#' @inheritParams removeConfigurationsMetaData
#' @inheritParams defaultScenario
#'
#' @return A list with the following elements:
#'   \describe{
#'     \item{\code{experiments}}{Experiments results.}
#'     \item{\code{seeds}}{Array of the instance seeds used in the experiments.}
#'   }
#'
#' @details A test instance set must be provided through `scenario[["testInstances"]]`.
#'
#' @seealso
#'  [testing_fromlog()]
#' 
#' @author Manuel López-Ibáñez
#' @export
testConfigurations <- function(configurations, scenario)
{
  # We need to set up a default scenario (and repeat all checks) in case
  # we are called directly instead of being called after executing irace.
  scenario <- checkScenario(scenario)
  
  testInstances <- scenario[["testInstances"]]
  instances_id <- names(testInstances)
  if (length(testInstances) == 0L) irace_error("No test instances given")
  if (is.null(instances_id)) irace_error("testInstances must have names")
  
  # 2147483647 is the maximum value for a 32-bit signed integer.
  # We use replace = TRUE, because replace = FALSE allocates memory for each possible number.
  ## FIXME: scenario[["testInstances"]] and scenario$instances behave differently,
  ## we should unify them so that the seeds are also saved in scenario.
  instanceSeed <- runif_integer(length(testInstances))
  names(instanceSeed) <- instances_id
  
  # If there is no ID (e.g., after using readConfigurations), then add it.
  if (".ID." %not_in% colnames(configurations))
    configurations[[".ID."]] <- seq_nrow(configurations)
  
  # Create experiment list
  experiments <- createExperimentList(configurations, parameters = scenario$parameters,
    instances = testInstances, instances_ID = instances_id, seeds = instanceSeed,
    # We cannot use rep.int because scenario$boundMax may be NULL.
    bounds = rep(scenario$boundMax, nrow(configurations)))
  race_state <- RaceState$new(scenario)
  if (scenario$debugLevel >= 3L) {
    irace_note ("Memory used before execute_experiments() in testConfigurations():\n")
    race_state$print_mem_used()
  }
  race_state$start_parallel(scenario)
  on.exit(race_state$stop_parallel())
  # We cannot let targetRunner or targetEvaluator modify our random seed, so we save it.
  withr::local_preserve_seed()
  target_output <- execute_experiments(race_state, experiments, scenario)
  # targetEvaluator may be NULL. If so, target_output must contain the right
  # output already.
  if (!is.null(scenario$targetEvaluator))
    target_output <- execute_evaluator(race_state$target_evaluator, experiments,
      scenario, target_output)

  # FIXME: It would be much faster to convert target_output to a data.table like we do in race_wrapper(),
  # then dcast() to a matrix like we do elsewhere.
  testResults <- matrix(NA, ncol = nrow(configurations), nrow = length(testInstances),
                        # dimnames = list(rownames, colnames)
                        dimnames = list(instances_id, configurations$.ID.))

  cost <- unlist_element(target_output, "cost")
  if (scenario$capping)
    cost <- applyPAR(cost, boundMax = scenario$boundMax, boundPar = scenario$boundPar)
  # FIXME: Vectorize this loop
  for (i in seq_along(experiments)) {
    testResults[rownames(testResults) == experiments[[i]]$id_instance,
                colnames(testResults) == experiments[[i]]$id_configuration] <- cost[i]
  }
  if (scenario$debugLevel >= 3L) {
    irace_note ("Memory used at the end of testConfigurations():\n")
    race_state$print_mem_used()
  }

  ## FIXME: Shouldn't we record these experiments in experiment_log ?
  list(experiments = testResults, seeds = instanceSeed)
}

context("mlr")

test_that("mlr", {
  skip_on_cran()
  suppressWarnings(skip_if_not_installed("mlr", minimum_version = "2.15.0"))
  require(mlr, quietly = TRUE)

  ps = ParamHelpers::makeParamSet(
    ParamHelpers::makeNumericParam("cp", lower = 0.1, upper = 1),
    ParamHelpers::makeIntegerParam("minsplit", lower = 1, upper = 10))

  lrn = mlr::makeLearner("classif.rpart")
  
  ctrl = mlr::makeTuneControlIrace(maxExperiments = 30, nbIterations = 1L, minNbSurvival = 1)
  tr = mlr::tuneParams(lrn, mlr::iris.task, mlr::hout, par.set = ps, control = ctrl)
  expect_true(TRUE)
})

withr::with_output_sink("test-argparser.Rout", {
test_that("argparse", {
  params_def <- data.frame(name =".param", type ="s", short = "-p", long=NA, default=NA, domain=NA, description="")
  parser <- irace:::CommandArgsParser$new("-p 'something something'", params_def)
  expect_length(parser$argv, 2L)
})
})

#!/bin/sh
#############################################################################
# Example of how to launch irace in a SLURM computing cluster. See the
# documentation of your computing cluster for the meaning of the SBATCH
# options.
#
# Launch this script with:
#   sbatch job.sh
#
# Contributed by Sarah Lee.
#############################################################################
#SBATCH --account=user_account
# All parallel processes run in the same node.
#SBATCH --nodes=1
# Number of desired cpus (SLURM_NTASKS_PER_NODE):
#SBATCH --ntasks-per-node=30
# Amount of RAM needed for this job:
#SBATCH --mem=0
# The name to show in queue lists for this job:
#SBATCH --job-name=my_irace_job
#SBATCH --mail-user=user@example.com
# The time the job will be running:
#SBATCH --time=12:00:00
# Set output and error files
#SBATCH --output=/mypath/irace-%j.out
#SBATCH --error=/mypath/irace-%j.err

# To load some software (you can show the list with 'module avail'):
module load r
# You may need to load more modules depending on your application.

## Option 1. Call the irace executable:
~/bin/irace --exec-dir=~/local/irace-execdir/ --parallel $((SLURM_NTASKS_PER_NODE))
# I usually add a symbolic link from the location where the irace executable is installed to ~/bin/irace. Otherwise, replace ~/bin/irace with the path to the irace executable. It will be something like: ~/local/R_libs/4.1/irace/bin/irace. How to find this path is explained in the user guide and in the quick-install intro: https://mlopez-ibanez.github.io/irace/#installing-the-irace-package
# You can add additional irace options to the line above (see the user-guide), for example --scenario "tuning-scenario.txt". I added --exec-dir above but you can define the execDir in the scenario file instead. It is important that it is a location that is available for writing from the computing nodes. You may also want to redirect the output of job.sh to some file you can read later. How to do that should be explained in the documentation of your cluster. Also, make sure that you are using the same R version when installing irace than the version that you are loading within job.sh. Otherwise, you may get strange errors.

# MPI: If you want to use mpi, you may need to add something like:
mpirun -np 1 ~/bin/irace --exec-dir=~/local/irace-execdir/ --parallel $((SLURM_NTASKS-1)) --mpi 1
# $SLURM_NTASKS is expanded automatically to the value you give to --ntasks=. We subtract one from that value because irace itself consumes one task.

## Option 2: Using an R script
cd /path_to_irace_R_script
Rscript ./run_irace.R

@echo off
::##############################################################################
:: BAT version of target-runner for Windows.
:: Contributed by Andre de Souza Andrade <andre.andrade@uniriotec.br>.
:: Check other examples in examples/
::
:: This script is run in the execution directory (execDir, --exec-dir).
::
:: PARAMETERS:
:: %%1 is the candidate configuration number
:: %%2 is the instance ID
:: %%3 is the seed
:: %%4 is the instance name
:: The rest are parameters to the target-algorithm
::
:: RETURN VALUE:
:: This script should print one numerical value: the cost that must be minimized.
:: Exit with 0 if no error, with 1 in case of error
::##############################################################################

:: Please change the EXE and FIXED_PARAMS to the correct ones
SET "exe=D:\bin\program"
SET "fixed_params=--time 20 --tries 1 --quiet"

FOR /f "tokens=1-4*" %%a IN ("%*") DO (
	SET candidate=%%a
	SET instance_id=%%b
	SET seed=%%c
	SET instance=%%d
	SET candidate_parameters=%%e
)

SET "stdout=c%candidate%-%instance_id%-%seed%.stdout"
SET "stderr=c%candidate%-%instance_id%-%seed%.stderr"

:: :: FIXME: How to implement this in BAT ?
::if not exist %exe% error "%exe%: not found or not executable (pwd: %(pwd))"

:: Save  the output to a file, and parse the result from it.

%exe% %fixed_params% -i %instance% --seed %seed% %candidate_parameters% 1>%stdout% 2>%stderr%


:: :: This may be used to introduce a delay if there are filesystem
:: :: issues.
:: setlocal EnableDelayedExpansion
:: :loop
:: for /f %%i in ("%stdout%") do set size=%%~zi
:: if "%size%" NEQ "0" (goto endloop)
:: waitfor false /t 10 >nul
:: goto loop
:: :endloop

:: This is an example of reading a number from the output.
:: It assumes that the objective value is the first number in
:: the first column of the last line of the output.
setlocal EnableDelayedExpansion
set "cmd=findstr /R /N "^^" %stdout% | find /C ":""
for /f %%a in ('!cmd!') do set numberlines=%%a
set /a lastline=%numberlines%-1
for /f "tokens=3" %%F in ('more +%lastline% %stdout%') do set COST=%%F
echo %COST%

:: Un-comment this if you want to delete temporary files.
:: del %stdout% %stderr%
exit 0

::##############################################################################
:: In Windows, we cannot execute directly .py files so we use this script as
:: the targetRunner that calls the actual .py file.
:: Contributed by Levi Ribeiro<azalos13@gmail.com>
::
:: This script is run in the execution directory (execDir, --exec-dir).
::
:: PARAMETERS:
:: %%1 is the candidate configuration number
:: %%2 is the instance ID
:: %%3 is the seed
:: %%4 is the instance name
:: The rest are parameters to the target-algorithm
::
:: RETURN VALUE:
:: This script should print one numerical value: the cost that must be minimized.
:: Exit with 0 if no error, with 1 in case of error
::##############################################################################
@echo off
:: If python is in the PATH environment variable, you can call it directly.
SET "pythonexe=python.exe"
:: Otherwise, you need to provide its absolute location
:: SET "pythonexe=C:\bin\python.exe"

:: This script is run in the execution directory (execDir, --exec-dir), which often does not coincide with the directory containing target-runner.py. Either a relative or absolute path should be provided.
SET "targetrunnerpy=..\target-runner.py"
set parameters=%*
%pythonexe% %targetrunnerpy%  %parameters%

% !Rnw weave = knitr
%%% DO NOT EDIT the .tex file directly since it is generated from the .Rnw
%%% sources.
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{irace package: User Guide}
%\VignetteDepends{knitr}
%\VignetteCompiler{knitr}
\synctex=1
\RequirePackage{xparse}
\RequirePackage[dvipsnames]{xcolor}
\documentclass[a4paper,english]{article}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{lmodern}
\usepackage[a4paper]{geometry} % It saves some pages
\usepackage[english]{babel}
\usepackage{ifthen}
\newboolean{Release}
\setboolean{Release}{true}
\usepackage{calc}
\usepackage{afterpage}
\usepackage{algorithm,algorithmic}
\usepackage{booktabs}
\usepackage{tabularx}
\usepackage{xspace}
\usepackage{amsmath,amssymb}
\usepackage{relsize}
\usepackage{fancyvrb}
\usepackage{underscore}
\usepackage{microtype}
% \texttt{test -- test} keeps the "--" as "--" (and does not convert it to an en dash)
\DisableLigatures{encoding = T1, family = tt* }
\usepackage[hyphens]{url}
\usepackage{hyperref}
\usepackage[numbers]{natbib}
\usepackage[nottoc]{tocbibind}
%% For autoref
\hypersetup{
    colorlinks,
    linkcolor={red!50!black},
    citecolor={blue!50!black},
    urlcolor={blue!70!black}
}
\addto\extrasenglish{%
\def\sectionautorefname{Section}
\let\subsectionautorefname\sectionautorefname
\let\subsubsectionautorefname\sectionautorefname
}
\usepackage[titletoc, title]{appendix}
% Fix use with \autoref
\newcommand*{\Appendixautorefname}{Appendix}
\usepackage{tocloft}
\setlength{\cftsubsecnumwidth}{3em}% Set length of number width in ToC for \subsection
\usepackage[inline]{enumitem}
\setlist[enumerate]{leftmargin=*,widest=00}
\setlist[itemize]{leftmargin=1.5em}

%% FIXME: listing is very limited, we should use 'minted'
\usepackage{listings}
\lstdefinestyle{BashInputStyle}{
  language=bash,%
  basicstyle=\ttfamily,%
  numbers=none,%
  frame=tb,%
  rulecolor=\color{lightgray},
%  framesep=1ex,
  framexleftmargin=1ex,
  columns=fullflexible,%
  backgroundcolor=\color{yellow!05},%
  linewidth=\linewidth,%
 % xleftmargin=1\linewidth,%
  identifierstyle=\color{darkgray},%
  keywordstyle=\color{darkgray},%
  keywordstyle={[2]\color{Cyan}},%
  keywordstyle={[3]\color{olive}},%
  stringstyle=\color{MidnightBlue},%
  commentstyle=\color{RedOrange},%
  morestring=[b]',%
  showstringspaces=false
}

\DefineVerbatimEnvironment{Code}{Verbatim}{}
\DefineVerbatimEnvironment{CodeInput}{Verbatim}{fontshape=rm}
\DefineVerbatimEnvironment{CodeOutput}{Verbatim}{}
\newenvironment{CodeChunk}{}{}


\newcommand{\IRACEHOME}[1]{\hyperlink{irace_home}{\path{$IRACE_HOME}}\path{#1}}

\providecommand{\keywords}[1]{\textbf{\textit{Index terms---}} #1}

% Simple font selection is not good enough.  For example, |\texttt{--}|
% gives `\texttt{--}', i.e., an endash in typewriter font.  Hence, we
% need to turn off ligatures, which currently only happens for commands
% |\code| and |\samp| and the ones derived from them.  Hyphenation is
% another issue; it should really be turned off inside |\samp|.  And
% most importantly, \LaTeX{} special characters are a nightmare.  E.g.,
% one needs |\~{}| to produce a tilde in a file name marked by |\file|.
% Perhaps a few years ago, most users would have agreed that this may be
% unfortunate but should not be changed to ensure consistency.  But with
% the advent of the WWW and the need for getting `|~|' and `|#|' into
% URLs, commands which only treat the escape and grouping characters
% specially have gained acceptance
\makeatletter
\DeclareRobustCommand\code{\bgroup\@makeother\_\@makeother\~\@makeother\$\@noligs\@codex}
\def\@codex#1{\texorpdfstring%
{{\text{\normalfont\ttfamily\hyphenchar\font=-1 #1}}}%
{#1}\egroup}
\makeatother

\let\proglang=\textsf
\newcommand{\pkg}[1]{{\fontseries{b}\selectfont #1}}
\newcommand{\aR}{\proglang{R}\xspace}
\newcommand{\MATLAB}{\proglang{MATLAB}\xspace}
\newcommand{\eg}{e.g.,\xspace}
\newcommand{\SoftwarePackage}{\pkg}
\newcommand{\ACOTSP}{\SoftwarePackage{ACOTSP}\xspace}

%% How to use this command:
% Parameter with one short switch: \defparameter[short]{paramName}{long}{default}
% Parameter without short switch:  \defparameter{paramName}{long}{default}
% Parameter without switch:  \defparameter{paramName}{}{default}
\newcommand{\defparameter}[4][]{%
\item[\code{#2}]\hypertarget{opt:#2}{} ~~ %
\ifthenelse{\equal{#3}{}}{}{%
  \emph{flag:} %
  \ifthenelse{\equal{#1}{}}{}{%
    \code{-#1}~~~\emph{or}~~~}%
  \code{--#3} ~~ }%
\emph{default:}~\texttt{#4} \\
}
\newcommand{\parameter}[1]{\hyperlink{opt:#1}{\code{#1}}}
\newcommand{\iracefun}[1]{\href{https://mlopez-ibanez.github.io/irace/reference/#1.html}{\code{#1()}}}
%\usepackage{showlabels}
%\showlabels{hypertarget}

\newcommand{\irace}{\pkg{irace}\xspace}
\newcommand{\Irace}{\pkg{Irace}\xspace}
\newcommand{\race}{\pkg{race}\xspace}
\newcommand{\FRACE}{\text{F-Race}\xspace}
\newcommand{\IFRACE}{\text{I/F-Race}\xspace}
\newcommand{\PyImp}{\pkg{PyImp}\xspace}
\newcommand{\iraceversion}{\Sexpr{packageVersion("irace")}}

\newcommand{\Niter}{\ensuremath{N^\text{iter}}\xspace}
\newcommand{\Nparam}{\ensuremath{{N^\text{param}}}\xspace}
\newcommand{\iter}{\ensuremath{j}\xspace}
\newcommand{\Budget}{\ensuremath{B}\xspace}
\newcommand{\Budgetj}{\ensuremath{\Budget_{\iter}}\xspace}
\newcommand{\Bused}{\ensuremath{\Budget_\text{used}}\xspace}
\newcommand{\Ncand}[1][]{\ensuremath{N_{#1}}\xspace}
\newcommand{\Mui}{\ensuremath{\mu_{\iter}}\xspace}
\newcommand{\Nmin}{\ensuremath{N^\text{min}}\xspace}
\newcommand{\Nsurv}{\ensuremath{N^\text{surv}}\xspace}
\newcommand{\Nelite}{\ensuremath{N^\text{elite}}\xspace}
\newcommand{\Nnew}{\ensuremath{N^\text{new}}\xspace}

\newcommand{\bmax}{\ensuremath{b^\text{max}}\xspace}
\newcommand{\bmin}{\ensuremath{b^\text{min}}\xspace}
\newcommand{\Celite}{\ensuremath{\Theta^\text{elite}}\xspace}

\ifthenelse {\boolean{Release}}{%
\newcommand{\MANUEL}[1]{}
\newcommand{\LESLIE}[1]{}
\newcommand{\THOMAS}[1]{}
}{%
\newcommand{\MANUEL}[1]{{\footnotesize\noindent\textbf{\color{red}[~MANUEL: #1~]}}}
\newcommand{\LESLIE}[1]{\footnote{\noindent\textbf{[ LESLIE: #1 ]}}}
\newcommand{\THOMAS}[1]{\footnote{\noindent\textbf{[ THOMAS: #1 ]}}}
}
\newcommand{\hide}[1]{}

\usepackage{tcolorbox}
\newcommand{\infoicon}{%
\parbox[c]{0.75cm}{\includegraphics[keepaspectratio=true,width=0.75cm]{light-bulb-icon}}%
\hspace{1em}}
\newcommand{\warningicon}{%
\parbox[c]{0.75cm}{\includegraphics[keepaspectratio=true,width=0.75cm]{Warning-icon}}%
\hspace{1em}}

\definecolor{LightGray}{RGB}{193,193,193}
\definecolor{LightYellow}{RGB}{253,247,172}

\newlength\macroiconwidth
\newenvironment{xwarningbox}{%
\setlength{\fboxrule}{3.0\fboxrule}%
\setlength{\fboxsep}{0\fboxsep}%
\begin{tcolorbox}[colback=LightYellow,colframe=LightGray,boxrule=\fboxrule,boxsep=\fboxsep]%
\infoicon%
\settowidth{\macroiconwidth}{\infoicon}%
\begin{minipage}[c]{\columnwidth - \macroiconwidth - 2.0\fboxrule - 2.0\fboxsep}
\raggedright\footnotesize
%
}{%
\end{minipage}
\end{tcolorbox}
%
}

% Workaround for broken knitr: https://github.com/yihui/knitr-examples/blob/master/036-latex-if.tex
\providecommand{\hldef}[1]{\textcolor[rgb]{0.345,0.345,0.345}{#1}}%
\providecommand{\hlsng}[1]{\textcolor[rgb]{0.192,0.494,0.8}{#1}}%

<<setup, include=FALSE>>=
library(knitr)
knit_hooks$set(inline = function(x) {
  if (is.numeric(x)) return(knitr:::format_sci(x, 'latex'))
  highr::hi_latex(x)
})
@

\begin{document}

<<include=FALSE>>=
library(knitr)
@

\author{Manuel L\'opez-Ib\'a\~nez, Leslie P\'erez C\'aceres, J\'er\'emie Dubois-Lacoste,\\
  Thomas St\"utzle and Mauro Birattari
\\IRIDIA, CoDE, Universit\'e Libre de Bruxelles, Brussels, Belgium}

\title{The \irace Package: User Guide}
\date{Version \iraceversion, \today}
%\keywords{automatic
%  algorithm configuration, racing, parameter tuning, \aR}

\maketitle

\tableofcontents

%Load files needed for examples
<<exampleload,eval=TRUE,include=FALSE>>=
library("irace")
load("examples.Rdata") # loads "experiment" and "output"
iraceResults <- irace::read_logfile(system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE))
log_ablation_file <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork = TRUE)
load(log_ablation_file)
options(width = 70)
@
\newpage


%%
%%
%%
%% General info
%%
%%
%%
\section{General information}
\MANUEL{Some things could be taken from the intro of the irace paper and reformulated.}
\MANUEL{It would be good to mention that not only opt algorithms can be configured with irace, we say this in the paper.}

\subsection{Background}
\MANUEL{I would add a paragraph defining what is irace (a bit longer than the abstract above) and references to the literature so people can find more info. The first reference should be the irace TR.} \LESLIE{Here i guess we should say why tune an algorithm is a good idea, and why using irace is a better one.}
The \irace package implements an \emph{iterated racing} procedure,
which is an extension of Iterated F-race (\IFRACE)~\cite{BirYuaBal2010:emaoa}.  The main use of
\irace is the automatic configuration of optimization and decision algorithms,
that is, finding the most appropriate settings of an algorithm given a
set of instances of a problem. However, it may also be useful for
configuring other types of algorithms when performance depends on
the used parameter settings. It builds upon the \pkg{race} package
by Birattari and it is implemented in \aR. The \irace
package is available from CRAN:
%
\begin{center}
\url{https://cran.r-project.org/package=irace}
\end{center}
%
More information about \irace is available at
\url{https://mlopez-ibanez.github.io/irace/}.

\subsection{Version}
The current version of the \irace package is \iraceversion. Previous
versions of the package can also be found in the \href{https://cran.r-project.org/package=irace}{CRAN website}.

The algorithm underlying the current version of \irace and its motivation are
described by \citet{LopDubPerStuBir2016irace}. The \textbf{adaptive capping
  mechanism} available from version $3.0$ is described by
\citet{PerLopHooStu2017:lion}. Details of the implementation before version 2.0
can be found in a previous technical report~\cite{LopDubStu2011irace}.
%
\begin{xwarningbox}
  Versions of \irace before 2.0 are not compatible with the file formats
  detailed in this document.
\end{xwarningbox}


\subsection{License}

The \irace package is Copyright \copyright{} \the\year\ and distributed under the GNU General Public
License version 3.0 (\url{http://www.gnu.org/licenses/gpl-3.0.en.html}).
The \irace package is free software (software libre): You can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

The \irace package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.

Please be aware that the fact that this program is released as Free Software
does not excuse you from scientific propriety, which obligates you to give
appropriate credit! If you write a scientific paper describing research that
made substantive use of this program, it is your obligation as a scientist to
(a) mention the fashion in which this software was used in the Methods section;
(b) mention the algorithm in the References section. The appropriate citation
is:

\begin{itemize}[leftmargin=3em]
\item[] Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres,
  Thomas Stützle, and Mauro Birattari. The \irace package: Iterated Racing for
  Automatic Algorithm Configuration. \emph{Operations Research Perspectives}, 3:43--58, 2016.
  doi:~\href{http://dx.doi.org/10.1016/j.orp.2016.09.002}{10.1016/j.orp.2016.09.002}

\end{itemize}

\section{Before starting}
\MANUEL{I think this could be a bit more detailed by defining what is a parameter, a configuration, an instance, etc. but ok for now.}

The \irace package provides an automatic configuration tool for
 tuning optimization algorithms, that is, automatically finding good configurations for the parameters values
of a (target) algorithm saving the effort that normally requires manual
tuning.

\begin{figure}[t]
  \centering
  \includegraphics[width=0.6\textwidth]{irace-scheme}
  \caption{Scheme of \irace flow of information.}
\label{fig:irace-scheme}
\end{figure}

Figure~\ref{fig:irace-scheme} gives a general scheme of how \irace works.
\Irace receives as input a \emph{parameter space definition} corresponding to the
parameters of the target algorithm that will be tuned, a set of \emph{instances}
for which the parameters must be tuned for and a set of options for \irace that define the \emph{configuration scenario}. Then, \irace searches in the parameter search space for good performing algorithm
configurations by executing the target algorithm on different instances and
with different parameter configurations. A \parameter{targetRunner} must be provided to execute the target algorithm with a
specific parameter configuration ($\theta$) and instance ($i$). The \parameter{targetRunner} function (or program) acts
as an interface between the execution of the target algorithm and \irace: It
receives the instance and configuration as arguments and must return the
evaluation of the execution of the target algorithm.

The following user guide contains guidelines for installing \irace, defining
configuration scenarios, and using \irace to automatically configure your
algorithms.

%%
%%
%%
%% Installation
%%
%%
%%

\section{Installation}

\subsection{System requirements}
\begin{itemize}
\item \aR ($\text{version} \geq 3.2.0$) is required for running irace, but you don't
  need to know the \aR language to use it.  \aR is freely available and you can
  download it from the \aR project website
  (\url{https://www.r-project.org}). See \autoref{sec:installation} for a
  quick installation guide of \aR.

\item For GNU/Linux and OS X, the command-line executable
  \code{parallel-irace} requires GNU Bash.  Individual examples may require additional software.
\end{itemize}

\subsection{\irace installation} \label{sec:irace install}

The \irace package can be installed automatically within \aR or
by manual download and installation. We advise to use the automatic
installation unless particular circumstances do not allow it. The instructions
to install \irace with the two mentioned methods are the following:

\subsubsection[Install automatically within R]{Install automatically within \aR{}}

Execute the following line in the \aR console to install the package:
<<R_irace_install, prompt=FALSE, eval=FALSE>>=
install.packages("irace")
@
Select a mirror close to your location, and test the installation in the \aR console with:
<<R_irace_launch,eval=FALSE, prompt=FALSE>>=
library("irace")
q() # To exit R
@

Alternatively, within the \aR graphical interface, you may use
the \code{Packages and data->Package installer} menu on OS X or the \code{Packages} menu on Windows.

\subsubsection{Manual download and installation}

From the \irace package CRAN website
(\url{https://cran.r-project.org/package=irace}), download one of the three
versions available depending on your operating system:
\begin{itemize}
\item \code{irace_\iraceversion.tar.gz} (Unix/BSD/GNU/Linux)
\item \code{irace_\iraceversion.tgz} (OS X)
\item \code{irace_\iraceversion.zip} (Windows)
\end{itemize}

To install the package on GNU/Linux and OS X, you must execute the following
command at the shell (replace \code{<package>} with the path to the downloaded file, either \code{irace_\iraceversion.tar.gz} or \code{irace_\iraceversion.zip}):
%
\begin{lstlisting}[style=BashInputStyle]
R CMD INSTALL <package>
\end{lstlisting}

To install the package on Windows, open \aR and execute the following line on
the \aR console (replace \code{<package>} with the path to the downloaded file \code{irace_\iraceversion.zip}):
%\LESLIE{Check that this actually works on internet says that  this: \code{Rscript -e "install.packages('foo.zip', repos = NULL)"} also works}
%
<<install_win1,eval=FALSE, prompt=FALSE>>=
install.packages("<package>", repos = NULL)
@

If the previous installation instructions fail because of insufficient
permissions and you do not have sufficient admin rights to install \irace
system-wide, then you need to force a local installation.

\subsubsection{Local installation}

Let's assume you wish to install \irace on a path denoted by
\code{<R_LIBS_USER>}, which is a filesystem path for which you have sufficient
rights. This directory \textbf{must} exist before attempting the
installation. Moreover, you must provide to \aR the path to this
library when loading the package. However, the latter can be avoided by adding
the path to the system variable \code{R_LIBS} or to the \aR internal variable
\code{.libPaths}, as we will see below.\footnote{%
   On Windows, see also
   \url{https://cran.r-project.org/bin/windows/base/rw-FAQ.html\#I-don_0027t-have-permission-to-write-to-the-R_002d3_002e3_002e1_005clibrary-directory}.}


On GNU/Linux or OS X, execute the following commands to install the
package on a local directory:

\begin{lstlisting}[style=BashInputStyle]
export R_LIBS_USER="<R_LIBS_USER>"
# Create R_LIBS_USER if it doesn't exist
mkdir $R_LIBS_USER
# Replace <package> with the path to the downloaded file.
R CMD INSTALL --library=$R_LIBS_USER <package>
# Tell R where to find R_LIBS_USER
export R_LIBS=${R_LIBS_USER}:${R_LIBS}
\end{lstlisting}

On Windows, you can install the package on a local directory by executing the
following lines in the \aR console:

<<install_win2,eval=FALSE, prompt=FALSE>>=
# Replace <package> with the path to the downloaded file.
# Replace <R_LIBS_USER> with the path used for installation.
install.packages("<package>", repos = NULL, lib = "<R_LIBS_USER>")
# Tell R where to find R_LIBS_USER.
# This must be executed for every new session.
.libPaths(c("<R_LIBS_USER>", .libPaths()))
@

\subsubsection{Testing the installation and invoking irace}

Once \irace has been installed, load the package and test that the installation
was successful by opening an \aR console and executing:

<<R_irace_test1, prompt=FALSE, eval=FALSE>>=
# Load the package
library("irace")
# Obtain the installation path
system.file(package = "irace")
@

The last command must print out the filesystem path  where \irace is
installed.  In the remainder of this guide, the variable
\code{\$IRACE_HOME} is used to denote this path. When executing any provided
command that includes the \code{\$IRACE_HOME} variable do not forget to replace
this variable with the installation path of \irace.

On GNU/Linux or OS X, you can let the operating system know where to find \irace
by defining the \code{\$IRACE_HOME} variable and adding it to the system
\code{PATH}. Append the following commands to \path{~/.bash_profile}, \path{~/.bashrc} or
\path{~/.profile}:
%
%<<linux_irace_path1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
# Replace <IRACE_HOME> with the irace installation path
export IRACE_HOME=<IRACE_HOME>
export PATH=${IRACE_HOME}/bin/:$PATH
# Tell R where to find R_LIBS_USER
# Use the following line only if local installation was forced
export R_LIBS=${R_LIBS_USER}:${R_LIBS}
\end{lstlisting}
%@

Then, open a new terminal and launch \irace as follows:

%<<linux_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace --help
\end{lstlisting}
%@

On Windows, you need to add both \aR and the installation path of \irace to the
environment variable \code{PATH}. To edit the \code{PATH}, search for
``Environment variables'' in the control panel, edit \code{PATH} and add
a string similar to \path{C:\R_PATH\bin;C:\IRACE_HOME\bin\x64\} where \code{R_PATH}
is the installation path of \aR and \code{IRACE_HOME} is the installation path
of \irace. If \irace was installed locally, you also need to edit the
environment variable \code{R_LIBS} to add \code{R_LIBS_USER}. Then, open a new
terminal (run program \code{cmd.exe}) and launch \irace as:
%
%<<win_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace.exe --help
\end{lstlisting}
%@

Alternatively, you may directly invoke \irace from within the \aR
console by executing:
<<windows_irace_help ,eval=FALSE, prompt=FALSE>>=
library("irace")
irace_cmdline("--help")
@


\section{Running irace}\label{sec:execution}

Before performing the tuning of your algorithm, it is necessary to define a
tuning scenario that will give \irace all the necessary information to optimize the
parameters of the algorithm. The tuning scenario is composed of the following elements:

\begin{enumerate}
\item Target algorithm parameter description (see \autoref{sec:target parameters}).
\item Target algorithm runner (see \autoref{sec:runner}).
\item Training instances list (see \autoref{sec:training})
\item \irace options (see \autoref{sec:irace options}).
\item \textit{Optional:} Initial configurations (see \autoref{sec:initial}).
\item \textit{Optional:} Target algorithm evaluator (see \autoref{sec:evaluator}).
\end{enumerate}

These scenario elements can be provided as plain text files
or as \aR objects. This user guide provides examples of both types,
but we advise the use of plain text files, which we consider the simpler
option.

For a step-by-step guide to create the scenario elements for your target algorithm
continue to \autoref{sec:step}. For an example execution of \irace using the
\ACOTSP scenario go to \autoref{sec:example}.



\subsection{Step-by-step setup guide}\label{sec:step}
This section provides a guide to setup a basic execution of \irace. The
template files provided in the package (\IRACEHOME{/templates}) will be
used as basis for creating your new scenario. Please follow carefully the
indications provided in each step and in the template files used; if you have
doubts check the the sections that describe each option in detail.

\begin{enumerate}[leftmargin=*]
\item Create a directory (\eg~\path{./tuning/}) for the scenario setup. This directory will contain
all the files that describe the scenario. On GNU/Linux or OS X, you can do this as follows:
%<<dir1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./tuning
cd ./tuning
\end{lstlisting}
%@

\item Initialize the tuning directory with template config files. On GNU/Linux or OS X, you can do this as follows:
%<<dir1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace --init
\end{lstlisting}
%@

\item Define the target algorithm parameters to be tuned by following the
instructions in \code{parameters.txt}.  Available parameter types and other
guidelines can be found in \autoref{sec:target parameters}.


\item \textit{Optional}: Define the initial parameter configuration(s) of your
algorithm, which  allows you to provide good starting configurations (if you know
some) for the tuning. Follow the instructions in \code{configurations.txt} and set \parameter{configurationsFile}\code{="configurations.txt"} in \code{scenario.txt}. More
information in \autoref{sec:initial}. If you do not need to
define initial configurations remove this file from the directory.

\item Place the instances you would like to use for the tuning of your
  algorithm in the folder \path{./tuning/Instances/}. In addition, you can
  create a file (\eg~\code{instances-list.txt}) that specifies which instances
  from that directory should be run and which instance-specific parameters to
  use. To use such an instance file, set the appropriate option in
  \code{scenario.txt}, e.g., \parameter{trainInstancesFile} \code{= "instances-list.txt"}.
  See \autoref{sec:training} for guidelines.

\item Uncomment and assign in \code{scenario.txt} only the options for which you
need a value different from the default. Some common options that you might want to adjust are:
\begin{description}
\item[\parameter{execDir}] (\code{--exec-dir}): the directory in which \irace will execute the
target algorithm; the default value is the current directory.
%\item[\parameter{logFile}] (\code{--log-file}): a file the name of the results \aR data file that produces \irace.
\item[\parameter{maxExperiments}] (\code{--max-experiments}): the maximum number of executions
of the target algorithm that \irace will perform.
\item[\parameter{maxTime}] (\code{--max-time}): maximum total execution time in seconds for the executions of \code{targetRunner}. In this case, \code{targetRunner} must return two values: cost and time. Note that you must provide either \parameter{maxTime} or \parameter{maxExperiments}.
\item[\parameter{trainInstancesDir}] (\code{--train-instances-dir}): set to \path{./Instances} if you put the training instances in that folder as instructed above.
  \end{description}
For setting the tuning budget, see \autoref{sec:budget}. For more information on \irace options and their default values, see \autoref{sec:irace options}.

\item Modify the \code{target-runner} script to run your algorithm. This script
  must execute your algorithm with the parameters and instance specified by
  \irace and return the evaluation of the execution and \textit{optionally} the execution
  time (\code{cost [time]}). When the \parameter{maxTime} option is used, returning \code{time} is mandatory.
  The \code{target-runner} template is written in \proglang{GNU Bash}
  scripting language, which can be executed easily in GNU/Linux and OS X
  systems. However, you may use any other programming language. We provide examples written in \proglang{Python},  \proglang{MATLAB} and other languages in \IRACEHOME{/examples/}. An example using Julia is available at \url{https://github.com/sbomsdorf/An-example-of-irace-using-Julia-code}. %
  Follow these instructions to adjust the given \code{target-runner} template
  to your algorithm:
  \begin{enumerate}
  \item Set the \code{EXE} variable with the path to the executable of the target algorithm.
  \item Set the \code{FIXED_PARAMS} if you need extra arguments in the
    execution line of your algorithm. An example could be the time that your
    algorithm is required to run
    (\code{FIXED_PARAMS}\hspace{0pt}\code{=}\hspace{0pt}\code{"--time 60"}) or
    the number of evaluations required (\code{FIXED_PARAMS}\hspace{0pt}\code{=}\hspace{0pt}\code{"--evaluations 10000"}).

  \item The line provided in the template executes the executable described in the \code{EXE} variable.
    %
    \begin{center}
      \code{\$EXE \$\{FIXED_PARAMS\} -i \$\{INSTANCE\} --seed \$\{SEED\} \$\{CONFIG_PARAMS\}}
    \end{center}
    %
    You must change this line according to the way your algorithm is executed. In this
    example, the algorithm receives the instance to solve with the flag
    \code{-i} and the seed of the random number generator with the flag \code{--seed}. The variable
    \code{CONFIG_PARAMS} adds to the command line the parameters that \irace has
    given for the execution.  You must set the command line execution as
    needed. For example, the instance might not need a flag and might need to be
    the first argument:
    \begin{center}
      \code{\$EXE \$\{INSTANCE\} \$\{FIXED_PARAMS\} --seed \$\{SEED\} \$\{CONFIG_PARAMS\}}
    \end{center}
    The output of your algorithm is saved to the file defined in the
    \code{\$STDOUT} variable, and error output is saved in the file given by
    \code{\$STDERR}. The line:
    \begin{center}
      \code{if [ -s "\${STDOUT}" ]; then}
    \end{center}
    checks if the file containing the output of your algorithm is not
    empty. The example provided in the template assumes that your algorithm
    prints in the last output line the best result found (only a number). The
    line:
    \begin{center}
      \code{COST=\$(cat \$\{STDOUT\} | grep -e '\^{}[[:space:]]*[+-]\textbackslash{}?[0-9]' | cut -f1)}
    \end{center}
    parses the output of your algorithm to obtain the result from the last line. The \code{target-runner}
    script must print \textbf{only} one number. In the template example, the result is printed with
    \code{echo "\$COST"} (assuming \parameter{maxExperiments} is used) and the generated files are deleted (you may remove that line if you wish to keep them).

    \begin{xwarningbox}
      The \code{target-runner} script must be an executable file, unless you specify \parameter{targetRunnerLauncher}.
    \end{xwarningbox}

    You can test the target runner from the \aR console by checking the scenario as
    explained earlier in \autoref{sec:execution}.

If you have problems related to the \code{target-runner} script when executing
\irace, see \autoref{sec:check list} for a check list to help diagnose common
problems. For more information about the \parameter{targetRunner}, please see
\autoref{sec:runner},
        \end{enumerate}

      \item \textit{Optional}: Modify the \code{target-evaluator} file. This is rarely needed and the \code{target-runner} template does not use it. \autoref{sec:evaluator} explains when a \parameter{targetEvaluator} is needed and how to define it.

      \item The \irace executable provides an option (\parameter{--check}) to
        check that the scenario is correctly defined. We recommend to perform a
        check every time you create a new scenario. When performing the check,
        \irace will verify that the scenario and parameter definitions are
        correct and will test the execution of the target algorithm. To check
        your scenario execute the following commands:

\begin{itemize}
\item From the command-line (on Windows, execute \code{irace.bat}):

%<<irace_check,engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
# $IRACE_HOME is the installation directory of irace.
$IRACE_HOME/bin/irace --scenario scenario.txt --check
\end{lstlisting}
%@

\item Or from the \aR console:

<<irace_R_check, eval=FALSE, prompt=FALSE>>=
library("irace")
scenario <- readScenario(filename = "scenario.txt",
                         scenario = defaultScenario())
checkIraceScenario(scenario = scenario)
@

\end{itemize}


\item  Once all the scenario elements are prepared you can execute \irace, either using the command-line wrappers provided by the package
 or directly from the \aR console:

\begin{itemize}
\item
\textbf{From the command-line console}, call the command (on Windows, you should execute
  \code{irace.exe}):
\begin{lstlisting}[style=BashInputStyle]
cd ./tuning/
# $IRACE_HOME is the installation directory of irace
# By default, irace reads scenario.txt, you can specify a different file
# with --scenario.
$IRACE_HOME/bin/irace
\end{lstlisting}
For this example, we assume that the needed scenario files have been set
properly in the \code{scenario.txt} file using the options described in
\autoref{sec:irace options}. Most \irace options can be specified
in the command line or directly in the \code{scenario.txt} file.

\item
\textbf{From the \aR console}, evaluate:

<<irace_R_exe, eval=FALSE, prompt=FALSE>>=
library("irace")
# Go to the directory containing the scenario files
setwd("./tuning")
scenario <- readScenario(filename = "scenario.txt",
                         scenario = defaultScenario())
irace_main(scenario = scenario)
@

\end{itemize}

This will perform one run of \irace. See the output of \code{irace --help} in the command-line or \code{irace_cmdline("--help")} in \aR for quick information on
additional \irace options. For more information about \irace options, see \autoref{sec:irace options}.

\end{enumerate}

\begin{xwarningbox}
Command-line options override the same options specified in the \code{scenario.txt} file.
\end{xwarningbox}

\subsection{Setup example for ACOTSP}\label{sec:example}

The \ACOTSP tuning example can be found in the package installation in the folder \IRACEHOME{/examples/acotsp}.
 %
Other example scenarios can be found in the same folder. More
examples of tuning scenarios can be found in the Algorithm Configuration Library (AClib, \url{https://www.aclib.net/}).

In this section, we describe how to execute the \ACOTSP scenario. If you wish to start setting up
your own scenario, continue to the next section. For this example, we assume
a GNU/Linux system such as Ubuntu with a working \proglang{C} compiler such as \code{gcc}.
To execute this scenario follow these steps:

\begin{enumerate}
\item Create a directory for the tuning (\eg~\path{./tuning/}) and copy the example scenario files
located in the \code{examples} folder to the created directory:
%<<dir0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./tuning
cd ./tuning
# $IRACE_HOME is the installation directory of irace.
cp $IRACE_HOME/examples/acotsp/* ./
ls ./
# Make sure that target-runner is executable
chmod u+x target-runner
\end{lstlisting}
%@

\item Download the training instances from \url{https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/scenarios/acotsp/instances.tar.gz} to the \path{./tuning/} directory and decompress it, which creates create a folder \path{instances}:

%<<instance0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
tar -xvf instances.tar.gz
ls instances/
\end{lstlisting}
%@

If the above gives an error or does not show any files, then the files were not extract correctly. Maybe the \path{instances.tar.gz} file did not download correctly or maybe it is not in the correct place. It should be within the folder \path{tuning}.correctly.

\item Download the \ACOTSP software from \url{https://github.com/MLopez-Ibanez/ACOTSPQAP/archive/refs/heads/master.zip} to the \path{./tuning/}
directory and compile the \path{acotsp} executable using \code{make}.
%<<acotsp0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
unzip master.zip
make -C ACOTSPQAP-master acotsp
./ACOTSPQAP-master/acotsp --help
\end{lstlisting}
%@

If the above gives an error, then the \path{acotsp} executable failed to compile for some reason. Maybe you are missing the C compiler or some files did not extract correctly.

\item Create a directory for executing the experiments and execute \irace:

%<<runexample,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./acotsp-arena/
# $IRACE_HOME is the installation directory of irace.
$IRACE_HOME/bin/irace
\end{lstlisting}
%@

Or you can also execute \irace from the \aR console using:

<<runexample2,prompt=FALSE,eval=FALSE>>=
library("irace")
setwd("./tuning/")
irace_cmdline()
@

\end{enumerate}

The most usual sources of error when running the above commands are:
\begin{itemize}
\item The \irace package is not correctly installed. Please make sure that installing \irace did not give any errors.
\item The location of the files is not correct. Please make sure that you have:
  \begin{itemize}
  \item The folder  \path{tuning} and that it contains the files \path{scenario.txt}, \path{parameters-acotsp.txt}, \path{target-runner}, and the folders \path{instances}, \path{ACOTSPQAP-master} and \path{acotsp-arena}.
  \item The folder \path{instances} should contain the TSP instance files (\path{*.tsp}).
  \item The folder \path{ACOTSPQAP-master} should contain the executable \path{acotsp}. You should be able to invoke \code{./ACOTSPQAP-master/acotsp --help} without an error.
  \item The folder \path{acotsp-arena} should be empty.
  \end{itemize}
\end{itemize}
%%
%%
%%
%% Scenario settings
%%
%%
%%



\section{Defining a configuration scenario}\label{sec:scenario}
\subsection{Target algorithm parameters} \label{sec:target parameters}

The parameters of the target algorithm are defined by a parameter file as
described in \autoref{sec:parameters file}. Optionally, when executing
\irace from the \aR console, the parameters can be specified directly as an \aR
object (see \autoref{sec:parameters object}). For defining your parameters
follow the guidelines provided in the following sections.

\subsubsection{Parameter types}

Each target parameter has an associated type that defines its domain and the
way \irace handles them internally. Understanding the nature of the domains of
the target parameters is important to select appropriate types. The four basic
types supported by \irace are the following:

\begin{itemize}
\item \textit{Real} parameters are numerical parameters that can take
  floating-point values within a given range. The range is
  specified as an interval `\code{(<lower bound>,<upper
    bound>)}'. This interval is closed, that is, the parameter value
  may eventually be one of the bounds. The possible values are rounded
  to a number of \emph{decimal places} specified by
  the global option \code{digits} (Section~\ref{sec:globaloptions}). For example, given the default
  number of digits of $4$, the values $0.12345$ and
  $0.12341$ are both rounded to $0.1234$.
   Selected real-valued parameters can be optionally sampled on
  a logarithmic scale (base $e$).

\item \textit{Integer} parameters are numerical parameters that can take only
  integer values within the given range. Their range is specified as the range
  of real parameters and they can also be optionally sampled on a
  logarithmic scale (base $e$).

\item \textit{Categorical} parameters are defined by a set of possible values
  specified as `\code{(<value 1>,} \code{...,} \code{<value n>)}'. The values
  are quoted or unquoted character strings. Empty strings and strings
  containing commas or spaces must be quoted.

\item \emph{Ordinal} parameters are defined by an \emph{ordered} set
  of possible values in the same format as for categorical
  parameters. They are handled internally as integer parameters, where
  the integers correspond to the indexes of the values.

\end{itemize}

\begin{xwarningbox}
  Boolean (or logical) parameters are best encoded as categorical ones with just two values rather than integer ones with domain $(0,1)$. Some boolean parameters take an explicit value (0/1 or true/false) such as:
\begin{CodeInput}
dlb "--dlb " c (0, 1)
\end{CodeInput}
Others are switches whose presence activates the parameter:
\begin{CodeInput}
dlb "" c ("", "--dlb")
\end{CodeInput}
\end{xwarningbox}




\subsubsection{Parameter domains}

For each target parameter, an interval or a set of values must be defined
according to its type, as described above. There is no limit for the size of
the set or the length of the interval, but keep in mind that larger domains
could increase the difficulty of the tuning task. Choose always values that you
consider relevant for the tuning. In case of doubt, we recommend to choose larger
intervals, as occasionally best parameter settings may be not intuitive a priori.
All intervals are considered as closed intervals.

It is possible to define parameters that will have always the same value.
Such ``\emph{fixed}'' parameters will not be tuned but their values are used
when executing the target algorithm and they are affected by constraints
defined on them. All fixed parameters must be defined as categorical
parameters and have a domain of one element.

\subsubsection{Parameter dependent domains}\label{sec:paramdependant}

Domains that are dependent on the values of other parameters can be specified
only for numerical parameters (both integer and real). To do so, the dependent
domain must be expressed in function of another parameter, which must be a
numerical parameter. The expression that defines a dependency must be written
between quotes: \code{(value, "expression")} or \code{("expression", value)} or
\code{("expression", "expression")}.

The expressions can only use the following operators and \aR functions: \code{+},
\code{-}, \code{*}, \code{/}, \code{\%\%}, \code{min}, \code{max},
\code{round}, \code{floor}, \code{ceiling}, \code{trunc}. If you need to use an
operator or function not listed here, please contact us.

\begin{xwarningbox}
The user must ensure that the defined domain is valid at all times since
\irace currently is not able to detect possible invalid domains based on the expressions
provided.
\end{xwarningbox}

If you have a parameter \code{p2} that is just a transformation of another
\code{p1}, then instead of using a dependent domain (left-hand side of the
following example), it will be better to create a dummy parameter that controls
the transformation (right-hand side) and do the transformation within \code{target-runner}. For example:
%
\begin{center}
\begin{minipage}{0.4\linewidth}
\small\centering%
\begin{CodeInput}[frame=single]
# With dependent domains
p1 "" r (0, 100)
p2 "" r ("p1", "p1 + 10")
\end{CodeInput}
\end{minipage}
\hspace{\stretch{1}} should be \hspace{\stretch{1}}
\begin{minipage}{0.4\linewidth}
\small\centering%
\begin{CodeInput}[frame=single]
# With a dummy parameter
p1    "" r (0, 100)
p2dum "" r (0, 10)
\end{CodeInput}
\end{minipage}
\end{center}
%
and \code{target-runner} will compute $\code{p2} = \code{p2dum} \cdot \code{p1}$.


\subsubsection{Conditional parameters}\label{sec:conditional}

Conditional parameters are active only when others have certain values. These
dependencies define a hierarchical relation between parameters. For example,
the target algorithm may have a parameter \code{localsearch} that takes values
\code{(sa, ts)} and another parameter \code{ts-length} that only needs to be
set if the first parameter takes precisely the value \code{ts}. Thus, parameter
\code{ts-length} is conditional on \code{localsearch == "ts"}.


\subsubsection{Forbidden parameter configurations}\label{sec:forbidden}
A line containing just \texttt{[forbidden]} ends the list of parameters and starts the list of forbidden expressions.
Each line is a logical expression (in \aR syntax) containing
parameter names as defined by the \parameter{parameterFile}
(\autoref{sec:target parameters}), values and logical operators. For a list of \aR logical operators see:
\begin{center}
 \url{https://stat.ethz.ch/R-manual/R-devel/library/base/html/Syntax.html}
\end{center}
If \irace generates a parameter configuration that makes any of the logical
expressions evaluate to \code{TRUE}, then the configuration is considered
forbidden and it is never evaluated. This is useful when some combination
of parameter values could cause the target algorithm to crash, consume
excessive CPU time or memory, or when it is known that they do no produce
satisfactory results.

\begin{xwarningbox}
Initial configuration (\autoref{sec:initial}) that are forbidden will be discarded with a warning.
\end{xwarningbox}

If the forbidden constraints provided are too strict, \irace may produce the following error:
\begin{CodeInput}
  irace tried 100 times to sample from the model a configuration not forbidden
  without success, perhaps your constraints are too strict?
\end{CodeInput}
%
In that case, it may be a good idea to reformulate  the forbidden constraints as conditional parameters (\autoref{sec:conditional}), parameter-dependent domains (\autoref{sec:paramdependant}),  repairing the configurations (\autoref{sec:repairconf}) or post-processing within the target-algorithm (\autoref{sec:complex_domains}).

\subsubsection{Global options}\label{sec:globaloptions}

A line containing just \texttt{[global]} starts the definition of global options. The only global option currently implemented is \code{digits}, which controls the number of decimal digits for real valued parameters. Its default value is 4.

\subsubsection{Parameter file format}\label{sec:parameters file}

For simplicity, the description of the parameters space is given as a
table. Each line of the table defines a configurable parameter
%
\begin{center}
\code{<name>\ <label>\ <type>\ <domain>\ [ | <condition>\ ] }
\end{center}
where each field is defined as follows:
%
\begin{description}[left=5em,itemindent=-0.5em,align=right,itemsep=0pt]
%  \DrawEnumitemLabel
\item[\code{<name>}] The name of the parameter as an unquoted
                  alphanumeric string, e.g., `\code{ants}'.
                  %
\item[\code{<label>}]
A \emph{label} for this parameter. This is a
                  string that will be passed together with the parameter
                  to \parameter{targetRunner}. In the default \parameter{targetRunner}
                  provided with the package (\autoref{sec:runner}), this is the
                  command-line switch used to pass the value of this parameter, for
                  instance `\code{"--ants "}'.\newline
                  The value of the parameter is concatenated \emph{without
                  separator} to the label when
                invoking \parameter{targetRunner}, thus \emph{any whitespace in the label is significant}. Following the same example, when parameter \code{ants} takes value \code{5}, the default targetRunner will pass the parameter as \code{"--ants 5"}.

  \item[\code{{<type>}}] The type of the parameter, either
                    \textit{integer}, \textit{real}, \textit{ordinal} or
                    \textit{categorical}, given as a single letter: `\code{i}',
                    `\code{r}', `\code{o}' or `\code{c}'. Numerical parameters
                    can be sampled using a natural logarithmic scale with '\code{i,log}' and
                    '\code{r,log}' (without spaces) for integer and real
                    parameters, respectively.
                    %
  \item[\code{{<domain>}}] The range or set of values of the parameter delimited by
                     parentheses, \eg~\code{(0,1)} or \code{(a,b,c,d)}.
                     See also parameter dependent domains (\autoref{sec:paramdependant}).
                     %

  \item[\code{{<condition>}}] An optional \emph{condition} that determines whether the parameter is
                         enabled or disabled, thus making the parameter conditional. If the
                         condition evaluates to false, then no value is assigned to this
                         parameter, and neither the parameter value nor the corresponding
                         label are passed to \parameter{targetRunner}. The condition must follow the same syntax as those for specifying forbidden configurations (see below), that is, it must be a
                         valid \aR logical expression\footnote{For a list of \aR operators
                         see: \url{https://stat.ethz.ch/R-manual/R-devel/library/base/html/Syntax.html}}.
                         The condition may contain the name
                         of other parameters as long as the dependency graph does not
                         contain any cycle. Otherwise, \irace will detect the cycle and
                         stop with an error.
\end{description}
\begin{xwarningbox}
Categorical and ordinal parameters are always treated as strings. Given a parameter like:
\begin{CodeInput}
a "" c (0, 5, 10, 20)
\end{CodeInput}
then, a condition like \code{a > 10} will be true when \code{a} is \code{5},
because comparisons between strings are lexicographic and \code{"10"} is sorted
before \code{"5"}. As a work-around, you can convert the string to numeric in
the condition with \code{as.numeric(a)}.
\end{xwarningbox}



As an example, Figure~\ref{fig:acotsp_parameters} shows the parameters file
of the \ACOTSP scenario.

\begin{figure}[!hbt]
\small\centering%
\begin{CodeInput}
# name      switch            type values               [conditions (using R syntax)]
algorithm   "--"              c    (as,mmas,eas,ras,acs)
localsearch "--localsearch "  c    (0, 1, 2, 3)
alpha       "--alpha "        r    (0.00, 5.00)
beta        "--beta "         r    (0.00, 10.00)
rho         "--rho  "         r    (0.01, 1.00)
ants        "--ants "         i    (5, 100)
nnls        "--nnls "         i    (5, 50)              | localsearch %in% c(1, 2, 3)
q0          "--q0 "           r    (0.0, 1.0)           | algorithm == "acs"
dlb         "--dlb "          c    (0, 1)               | localsearch %in% c(1,2,3)
rasrank     "--rasranks "     i    (1, "ants")          | algorithm == "ras"
elitistants "--elitistants "  i    (1, 750)             | algorithm == "eas"

[forbidden]
## Examples of valid logical operators are:
## ==  !=  >=  <=  >  <  &  |  !  %in%
(alpha == 0.0) & (beta == 0.0)
\end{CodeInput}
\caption{Parameter file (\code{parameters.txt}) for tuning \ACOTSP.}\label{fig:acotsp_parameters}
\end{figure}

\subsubsection{Parameters R format}\label{sec:parameters object}
The target parameters are stored in an \aR~\code{list} that you can obtain from the \aR console using the following command:

<<readParameters,prompt=FALSE, eval=FALSE>>=
parameters <- readParameters(file = "parameters.txt")
@

See the help of the \code{readParameters} function (\code{?readParameters}) for more information.
The structure of the parameter list that is created is as follows:

\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{0.98\linewidth}{@{}rX}
    \code{names}        & Vector that contains the names of the parameters. \\
    \code{types}        & Vector that contains the type of each parameter 'i', 'c', 'r', 'o'. \\
    \code{switches}     & Vector that contains the labels of the parameters. e.g., switches to be used for the
    parameters on the command line. \\
    \code{domain}       & List of vectors, where each vector may contain two
    values (minimum, maximum) for real and integer parameters, or
    a set of values for categorical and ordinal parameters.\\
    \code{conditions}   & List of \aR logical expressions, with variables
   corresponding to parameter names.\\
    \code{isFixed}      & Logical vector that specifies which parameter is fixed
   and, thus, it does not need to be tuned.\\
    \code{transform}    & Vector that contains the transformation of each parameter. Currently, it
      can take values \code{``''} (no transformation, default) of \code{``log''}
      (natural logarithmic transformation).\\
    \code{nbParameters} & An integer, the total number of parameters.\\
    \code{nbFixed}      & An integer, the number of parameters with a fixed value.\\
    \code{nbVariable}   & Number of variable (i.e., to be tuned) parameters.\\
    \code{isDependent}  & Logical vector that specifies which parameter defines
                          a dependent domain.\\
    \code{forbidden}  & List of R logical expressions that cannot evaluate to \code{TRUE} for any evaluated configuration.\\
  \end{tabularx}
\end{center}

The following example shows the structure of the \code{parameters} \aR object
for the \code{algorithm}, \code{ants} and \code{q0} parameters of the \ACOTSP scenario:

<<parametersetup,eval=TRUE,include=FALSE>>=
# Setup example
parameters <- readParameters(text='
algorithm       "--"                 c		(as,mmas,eas,ras,acs)
ants            "--ants "            i	        (5, 100)
q0              "--q0 "              r          (0.0, 1.0)              | algorithm %in% c("acs")
')
@

<<parameterlist,eval=TRUE, prompt=TRUE, size='normalsize', comment="">>=
str(parameters, vec.len = 10)
@

\subsection{Target algorithm runner}
\label{sec:runner}

The evaluation of a candidate configuration on a single instance is done by means of a user-given auxiliary
program or, alternatively, a user-given \aR function. The function (or program name) is specified by the option
\parameter{targetRunner}. The \parameter{targetRunner} must return the
cost value (\eg cost of the best solution found) of the evaluation; unless computing the cost requires information from all the configurations evaluated on an instance, \eg when evaluating multi-objective algorithms with unknown normalisation bounds (see \autoref{sec:evaluator} for details).

\begin{xwarningbox}
  The objective of \irace is to minimize the cost value returned by the target
  algorithm.  If you wish to maximize, you can multiply the cost by
  \code{-1} before returning it to \irace.
\end{xwarningbox}

\subsubsection{Target runner as an executable program}

When \parameter{targetRunner} is an auxiliary executable program, it
is invoked for each candidate configuration, passing as arguments:
%
\begin{center}
\small\code{<id_configuration>\ <id_instance>\ <seed>\ <instance>\ [bound]\ <configuration>}
\end{center}
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
    \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
    \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
    \code{seed} & seed for the random number generator to be used for this evaluation, ignore the seed for deterministic algorithms;\\
    \code{instance} & string giving the instance to be used for this evaluation;\\
    \code{bound} & optional execution time bound. Only provided when the \parameter{boundMax} option is set in the scenario, see \autoref{sec:capping};\\
    \code{configuration} & the pairs parameter label-value that
    describe this candidate configuration. Typically given as command-line
    switches to be passed to the executable program.\\
  \end{tabularx}
\end{center}

The experiment list shown in \autoref{sec:runner fx}, would result in the following execution line:

%<<target_runner_ex1, engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
target-runner 1 113 734718556 /home/user/instances/tsp/2000-533.tsp \
 --eas --localsearch 0 --alpha 2.92 --beta 3.06 --rho 0.6 --ants 80
\end{lstlisting}
%@

The command line switches that describe the candidate configuration are
constructed by appending to each parameter label (switch), \emph{without
separator}, the value of the parameter, following the order given in the
parameter table. The program \parameter{targetRunner} must print a real
number, which corresponds to the cost measure of the candidate configuration
for the given instance and optionally its execution time (mandatory when \parameter{maxTime}
is used and/or when the \parameter{capping} option is enabled). The working directory of
\parameter{targetRunner} is set to the execution directory specified by the option
\parameter{execDir}. This allows the user to execute independent runs of \irace
in parallel using different values for \parameter{execDir}, without the runs
interfering with each other.

\subsubsection{Target runner as an R function}
\label{sec:runner fx}

When \parameter{targetRunner} is an \aR function, it is invoked
for each candidate configuration as:
%
<<targetRunner,prompt=FALSE, eval=FALSE>>=
targetRunner(experiment, scenario)
@
%
\noindent where \code{experiment} is a list that contains  information about
configuration and instance to execute one experiment, and \code{scenario} is the scenario
list. The structure of the \code{experiment} list is as follows:

\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
   \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
   \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
   \code{seed} & seed to be used for this evaluation;\\
   \code{instance} & string giving the instance to be used for this evaluation;\\
   \code{bound} &  optional execution time bound;\\
   \code{configuration}    & 1-row data frame with a column per parameter name;\\
  \end{tabularx}
\end{center}

The following is an example of an experiment list for the \ACOTSP scenario:

<<experimentlist,eval=TRUE,size='normalsize', prompt=TRUE,  comment="">>=
print(experiment)
@

If you need the \code{switches} of each parameter (or any other information
about the parameters), you can obtain it from \code{scenario$parameters}.

You can find an example that calls MATLAB from \aR using this approach here: \url{https://github.com/MLopez-Ibanez/irace/blob/master/inst/examples/matlabr/scenario.txt}.

If \parameter{targetEvaluator} is \code{NULL}, then
the \parameter{targetRunner} function must return a list with at least one
element \code{"cost"}, the numerical value corresponding to the evaluation of
the given configuration on the given instance. A cost of \code{Inf} is accepted
and results in the immediate rejection of the configuration (see
\autoref{sec:reject}).

If the scenario option \parameter{maxTime} is non-zero or if the \parameter{capping}
option is enabled, then the list must contain at least another element \code{"time"} that
reports the execution time for this call to \code{targetRunner}.

The return list may also contain the following optional elements that are used
by \irace for reporting errors in \code{targetRunner}:
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
\code{error} & is a string used to report an error;\\
    \code{outputRaw} & is a string used to report the raw output of calls to
      an external program or function;\\
    \code{call} & is a string used to report how \code{targetRunner} called
      an external program or function;\\
  \end{tabularx}
\end{center}



\subsection{Target evaluator} \label{sec:evaluator}

Normally, \parameter{targetRunner} returns the cost of the execution of a
candidate configuration (see \autoref{sec:runner}). However, there are cases
when the cost evaluation must be delayed until all candidate configurations in
a race have been executed on a instance.

The \parameter{targetEvaluator} option defines an auxiliary program (or an \aR
function) that allows postponing the evaluations of the candidate
configurations. For each instance seen, the program \parameter{targetEvaluator} is only invoked after all
the calls to \parameter{targetRunner} for all alive candidate configurations on
the same instance have already finished.

\begin{xwarningbox}
When using \parameter{targetEvaluator}, \parameter{targetRunner} must not return the evaluation
of the configuration. If \parameter{maxTime} is used, \parameter{targetRunner} must return only
execution time.
\end{xwarningbox}

As an example, \parameter{targetEvaluator} may be used to dynamically find
normalization bounds for the output returned by an algorithm for each
individual instance. In this case, \parameter{targetRunner} will save the
output of the algorithm, then the first call to \parameter{targetEvaluator}
will examine the output produced by all calls to \parameter{targetRunner} for
the same instance, update the normalization bounds and return the normalized
output. Subsequent calls to \parameter{targetEvaluator} for the same instance
will simply return the normalized output.

A similar need arises when using quality measures for multi-objective
optimization algorithms, such as the hypervolume, which typically
require specifying reference points or sets. By
using \parameter{targetEvaluator}, it is possible to dynamically compute the
reference points or sets while \irace is running. Examples are provided at
\path{examples/hypervolume}. See also \autoref{sec:multi objective} for
more information on how to tune multi-objective algorithms.

\subsubsection{Target evaluator executable program}

When \parameter{targetEvaluator} is an auxiliary executable program, it
is invoked for each candidate with the following arguments:
%
\begin{center}
\small\code{<id_configuration>\ <id_instance>\ <seed>\ <instance>\ <num_configurations>\ <all_conf_id>}
\end{center}
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
   \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
   \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
   \code{seed} & seed to be used for this evaluation;\\
   \code{instance} & string giving the instance to be used for this evaluation;\\
   \code{num_configurations} & number of alive candidate configurations;\\
   \code{all_conf_id}    & list of IDs of the alive configurations separated by whitespace.\\
  \end{tabularx}
\end{center}

The \parameter{targetEvaluator} executable must print a numerical value
corresponding to the cost measure of the candidate configuration on
the given instance.




\subsubsection{Target evaluator R function}

When \parameter{targetEvaluator} is an \aR function, it is invoked
for each candidate configuration as:%
%
<<targetEvaluator, eval=FALSE>>=
targetEvaluator(experiment, num_configurations, all_conf_id, scenario,
                target_runner_call)
@
%
\noindent where \code{experiment} is a list that contains information about one
experiment (see \autoref{sec:runner fx}), \code{num_configurations}
is the number of configurations alive in the race, \code{all_conf_id} is
the vector of IDs of the alive configurations,  \code{scenario} is
the scenario list and \code{target_runner_call} is the string of the
\parameter{targetRunner} execution line.

The function \parameter{targetEvaluator} must return a list with one element
\code{"cost"}, the numerical value corresponding to the cost measure of the
given configuration on the given instance.

The return list may also contain the following optional elements that are used
by \irace for reporting errors in \code{targetEvaluator}:
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
\code{error} & is a string used to report an error;\\
    \code{outputRaw} & is a string used to report the raw output of calls to
      an external program or function;\\
    \code{call} & is a string used to report how \code{targetEvaluator} called
      an external program or function;\\
  \end{tabularx}
\end{center}

\subsection{Training instances}\label{sec:training}

The \irace options \parameter{trainInstancesDir} and \parameter{trainInstancesFile}
specify where to find the training instances.
%
If you only set the value of \parameter{trainInstancesDir} (e.g., to  \path{./Instances}),  \irace will consider all files within that directory as training instances.

Otherwise, the value of \parameter{trainInstancesFile} may specify a text
file. The format of this file is one instance per line. Within each line, elements separated by white-space will be parsed as separate arguments to be supplied to \parameter{targetRunner}. This allows defining instance-specific parameter settings. Quoted strings will be parsed as a single argument.
%
The following example shows a training instance file for the \ACOTSP scenario:

\begin{figure}[!ht]
\centering%
    \begin{minipage}{0.4\textwidth}
\begin{CodeInput}
# Example training instances file
100/100-1_100-2.tsp --time 1
100/100-1_100-3.tsp --time 2
100/100-1_100-4.tsp --time 3
\end{CodeInput}
\end{minipage}
\caption{Training instances file for tuning \ACOTSP.}\label{fig:acotsp_training}
\end{figure}

The value of \parameter{trainInstancesDir}, if set, is always prefixed to the instance
name, that is, the instances names are treated as relative to this directory.
%
For example, given the above file as \parameter{trainInstancesFile} and   \parameter{trainInstancesDir}\code{="./Instances"}, then a possible invocation of \parameter{targetRunner} would be:
%
%<<target_runner_cmd, engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
target-runner 1 4 5718 ./Instances/100/100-1_100-2.tsp --time 1 --alpha 2.92 ...
\end{lstlisting}
%@


Training instances do not need to be files, \irace just passes the elements of each line as arguments to \parameter{targetRunner}, thus each line may denote the name of a benchmark function or a label, plus instance-specific settings, that the target algorithm understands. Each line may even be the command-line parameters required to call an instance generator within \parameter{targetRunner}. When the instances do not represent actual files, then \parameter{trainInstancesDir} is usually set to the empty string (\code{--train-instances-dir=""}). For example,

\begin{center}
  \begin{minipage}{0.4\linewidth}
\begin{CodeInput}
# Example training instances file
rosenbrock_20 --function=12 --nvar 20
rosenbrock_30 --function=12 --nvar 30
rastrigin_20 --function=15 --nvar 20
rastrigin_30 --function=15 --nvar 30
\end{CodeInput}
\end{minipage}
\end{center}
%\caption{Example of training instances file where instances do not correspond to files.}\label{fig:nofile_training}
%\end{figure}

Optionally, when executing \irace from the \aR console, the list of instances might be provided explicitly by means of the variable \code{scenario$instances}. Thus, the previous example would be equivalent to:

<<instance1, prompt=FALSE, eval=FALSE>>=
scenario$instances <- c("rosenbrock_20 --function=12 --nvar 20",
                        "rosenbrock_40 --function=12 --nvar 30",
                        "rastrigin_20 --function=15 --nvar 20",
                        "rastrigin_40 --function=15 --nvar 30")
@

By default, \irace assumes that the target algorithm is stochastic (the value
of the option \parameter{deterministic} is $0$), thus, the same configuration
can be executed more than once on the same instance and obtain different
results.  In this case, \irace generates pairs \code{(instance, seed)} by
generating a random seed for each instance. In other words, configurations
evaluated on the same instance use the same random seed. This is a well-known
variance reduction technique called \emph{common random
  numbers}~\citep{McG1992vrt}. If all available pairs are used within a run of
\irace, new pairs are generated with different seeds, that is, a configuration
evaluated more than once per instance will use different random seeds.

If \parameter{deterministic} is set to $1$, then each instance will be used at
most once per race. This setting should only be used for target algorithms that
do not have a stochastic behavior and, therefore, executing the target
algorithm on the same instance several times with different seeds does not make
sense.
\begin{xwarningbox}
  If \parameter{deterministic} is active and the number of training instances
  provided to \irace is less than \parameter{firstTest} (default: 5), no
  statistical test will be performed on the race.
\end{xwarningbox}

Finally, \irace randomly re-orders the sequence of instances provided. This
random sampling may be disabled by using the option \parameter{sampleInstances}
(\code{--sample-instances 0}) if keeping the order provided in the instance
file is important.

\begin{xwarningbox}
  We advise to always sample instances to prevent biasing the tuning due to the
  instance order. See also \autoref{sec:het}
\end{xwarningbox}



\subsection{Initial configurations} \label{sec:initial}

The scenario option \parameter{configurationsFile} allows specifying a text
file that contains an initial set of configurations to start the execution of
\irace. If the number of initial configurations supplied in the file is less
than the number of configurations required by \irace in the first iteration,
additional configurations will be sampled uniformly at random.

The format of the configurations file is one configuration per line, and one
parameter value per column. The first line must give the parameter name
corresponding to each column (names must match those given in the parameters
file).  Each configuration must satisfy the parameter conditions (\code{NA}
should be used for those parameters that are not enabled for a given
configuration) and not be forbidden by the constraints that define forbidden
configurations (\autoref{sec:forbidden}), if any.

Figure~\ref{fig:acotsp_default} gives an example file that corresponds to the
\ACOTSP scenario.

\begin{figure}[!ht]
  \centering
  \begin{minipage}{0.7\linewidth}
\footnotesize
\begin{CodeInput}
## Initial candidate configuration for irace
algorithm localsearch alpha beta rho  ants nnls dlb q0 rasrank elitistants
as        0           1.0   1.0  0.95 10   NA   NA  0  NA      NA
\end{CodeInput}
\end{minipage}
\caption{Initial configuration file (\code{default.txt}) for tuning \ACOTSP.}\label{fig:acotsp_default}
\end{figure}

We advise to use this feature when a default configuration of the target algorithm exists or when
different sets of good parameter values are known. This will allow \irace to start the search
from those parameter values and attempt to improve their performance.



\subsection{Repairing configurations}\label{sec:repairconf}

In some problems, the parameter values require complex constraints that cannot
be implemented by constraints defined in the parameter space
(\autoref{sec:forbidden}).
%
The scenario option \parameter{repairConfiguration} can be set to a
user-defined \aR function that takes a single configuration generated by irace
and returns a ``\emph{repaired}'' configuration, thus allowing the
implementation of any rules necessary to satisfy arbitrary constraints on
parameter values.
%
The \parameter{repairConfiguration} function is called after generating a
configuration and before checking for forbidden configurations. The first
argument is a 1-row \code{data.frame} with parameter names as the column names and the second argument is the \code{parameters} list (\autoref{sec:parameters
  object}).  An example that makes all real-valued parameters
sum up to one would be:
%
<<repairEx,prompt=FALSE, eval=FALSE>>=
repairConfiguration = function(configuration, parameters)
{
  isreal <- names(which(parameters$types[colnames(configuration)] == "r"))
  # This ignores 'digits'
  c_real <- unlist(configuration[isreal])
  c_real <- c_real / sum(c_real)
  configuration[isreal] <- c_real
  return(configuration)
}
@
The following example forces three specific parameters to be in increasing order:
%
<<repairEx2,prompt=FALSE, eval=FALSE>>=
repairConfiguration = function(configuration, parameters)
{
 columns <- c("p1","p2","p3")
 # cat("Before"); print(configuration)
 configuration[columns] <- sort(unlist(configuration[columns], use.names=FALSE))
 # cat("After"); print(configuration)
 return(configuration)
}
@

The above code can be specified directly in the \parameter{scenarioFile}, by default \code{scenario.txt}.

%%
%%
%%
%% Parallel processing
%%
%%
%%

\section{Parallelization}\label{sec:parallel}
A single run of \irace can be done much faster by executing the calls
to \parameter{targetRunner} (the runs of the target algorithm) in
parallel. There are four ways to parallelize a single run of \irace:

\begin{enumerate}
\item \textbf{Parallel processes}: The option \parameter{parallel} executes
  multiple calls to \parameter{targetRunner} in parallel within a single
  computer, by means of the \pkg{parallel} \aR package. For example, adding
  \code{--parallel N} to the command line of \irace will launch in parallel up
  to $N$ calls of the target algorithm. When using this option within a
  computing cluster, \irace will be submitted as a \emph{job} in some way  that tells the cluster to   ``reserve'' $N$ CPUs (or tasks depending on the cluster) within a single cluster node (a single machine).

\item \textbf{MPI}: By enabling the option \parameter{mpi}, calls
  to \parameter{targetRunner} will be executed in parallel by using the message
  passing interface (MPI) protocol (requires the \pkg{Rmpi} \aR package). In
  this case, the option \parameter{parallel} controls the number of slave nodes
  used by \irace. For example, adding \code{--mpi 1 --parallel N} to the
  command-line will create $N$ slaves + 1 master, and execute up to $N$ calls of
  \parameter{targetRunner} in parallel.

   The user is responsible for setting up the required MPI environment.  MPI is
   commonly available in computing clusters and requires launching \irace in
   some particular way. An example script for using MPI mode in a SGE cluster
   is given at \IRACEHOME{/bin/parallel-irace-mpi}.

   By default, \irace dynamically balances the load among nodes, however, this
   may significantly increase communication overhead in some parallel
   environments and disabling \parameter{loadBalancing} may be faster.

 \item \textbf{Batch queue mode}: Some computing clusters work by submitting
   jobs to a batch queue and waiting for the jobs to finish. With the
   option \parameter{batchmode} (\code{--batchmode} \code{[sge|pbs|} \code{torque|slurm]}),
   \irace will launch in parallel as many calls of \parameter{targetRunner} as
   possible and use a
   cluster-specific method to wait for jobs to finish. In this mode, \parameter{parallel} controls how many jobs are queued and should be set to the queue limit of your cluster.  If your cluster type is
   not supported or not working as expected, please submit a pull request (\url{https://github.com/MLopez-Ibanez/irace/pulls}) adding support to your cluster type.
See the examples in
    \IRACEHOME{/examples/batchmode-cluster/}.

   \begin{xwarningbox}
     In batchmode, \irace runs in the submission node of the cluster,
     hence, \irace is not submitted to the cluster as a job (that is,
     neither \code{qsub} nor \code{squeue} should be used to invoke \irace
     itself). The user must call the appropriate job submission command (e.g., \code{qsub} or \code{squeue}) from \parameter{targetRunner} with the appropriate settings for their cluster, that is,  \parameter{targetRunner} submits one job to the
     cluster and prints a single string: The job ID that allows
     \irace to determine the status of the running job. Moreover, the use of a
     separate
   \parameter{targetEvaluator} script is required to collect the results of
   \parameter{targetRunner} and return them to \irace.
   \end{xwarningbox}

  \item \parameter{targetRunnerParallel}: This option allows users to fully
    control the parallelization of the execution of \parameter{targetRunner}.
    Its value must be an \aR function that will be invoked by \irace as follows:

   <<targetRunnerParallel,prompt=FALSE, eval=FALSE>>=
   targetRunnerParallel(experiments, exec_target_runner, scenario, target_runner)
   @
   %
   where \code{scenario} is the list describing the configuration scenario
   (\autoref{sec:scenario}); \code{experiments} is a list that describes the
   configurations and instances to be executed (see \autoref{sec:runner} for a
   description); \code{target_runner} is the function that calls the target
   algorithm and it is the same as \parameter{targetRunner}, if the latter is a
   function, or it is a call to \code{target_runner_default}, if
   \parameter{targetRunner} is the path to an executable; and
   \code{exec_target_runner} is an internal function within \irace that takes
   care of executing \code{target_runner}, check its output and, possibly,
   retry in case of error (see \parameter{targetRunnerRetries}). The
   \parameter{targetRunnerParallel} function should call the given
   \code{target_runner} function for each element in the \code{experiments}
   list, possibly using \code{exec_target_runner} as a wrapper. A trivial
   example would be:

   <<targetRunnerParallel2,prompt=FALSE, eval=FALSE>>=
   targetRunnerParallel <- function(experiments, exec_target_runner, scenario,
                                    target_runner)
   {
     lapply(experiments, exec_target_runner, scenario = scenario,
            target_runner = target_runner)
   }
   @
   %
   However, the user is free to set up the calls in any way, perhaps
   implementing their own replacement for \code{target_runner} and/or
   \code{exec_target_runner}. The user may load and call other \aR packages, such as \pkg{batchtools} (\url{https://cran.r-project.org/package=batchtools}).

   The only requirement is that the \parameter{targetRunnerParallel} function
   must return a list of the same length as \code{experiments}, where each
   element is the output expected from the corresponding call
   to \parameter{targetRunner} (see \autoref{sec:runner}).

   The following is an example of the output of a call to
   \parameter{targetRunnerParallel} with 2 experiments, in which the execution
   time is not reported:

  <<targetRunnerParallel3,prompt=FALSE, eval=TRUE>>=
     print(output)
   @
   %

 \end{enumerate}

 The best option will depend on the resources available to you. Option 1 is
 usually the fastest and simplest to setup.  Running on a node (machine) with
 128 CPUs will be faster than running on 8 nodes with 16 CPUs because the
 communication between nodes required by MPI can be very slow depending on the
 cluster.  Option 2 may be faster if irace generates more configurations per
 iteration than the number of CPUs of a single node. However, depending on the
 configuration of your cluster, requesting many CPUs may require waiting in the
 queue a long time. Option 3 may be the slowest since irace has to check the
 queue frequently. However, irace will start running experiments as soon as 1
 CPU is available, thus option 3 may actually finish earlier than the other
 options if there is always some CPUs available in the cluster but the queue
 for requesting many CPUs at once is very long.

 As a rule-of-thumb, if you only have access to a single machine, then you only need option 1. If you have access to a computing cluster with multiple machines, then  use option 1 with the maximum number of CPUs that a single node has in your computing cluster. If that number is 64 or more, it should be enough unless a single run of irace evaluates thousands of configurations. Otherwise, investigate option 2. If option 2 does not work, then investigate option 3.

%%
%%
%%
%% Testing
%%
%%
%%

\section{Testing (Validation) of configurations}\label{sec:testing}

Once the tuning process is finished, \irace returns a set of configurations
corresponding to the elite configurations at the end of the run, ordered from
best to worst. In order to evaluate the generality of these configurations
without looking at their performance on the training set, \irace offers the
possibility of evaluating these configurations on a test instance set,
typically different from the training set used during the tuning phase. These
evaluations will use the same settings for parallel
execution, \parameter{targetRunner} and \parameter{targetEvaluator}.

The test instances can be specified by the
options \parameter{testInstancesDir} and/or \parameter{testInstancesFile}, or by
setting directly the variable \code{scenario\$testInstances}. These options behave similarly to their counterparts for the training instances
(\autoref{sec:training}). In particular, each test instance is assigned a
different seed in the same way as done for the training instances. In
principle, \irace evaluates each configuration on each testing instance just
once, because evaluating one run on $n$ instances is always better than
evaluating $n'$ runs on $n/n'$ instances~\citep{IRIDIA-2004-001}. However, if
the number of instances is limited, one can always duplicate instances as
needed in the \parameter{testInstancesFile}, and \irace will assign a different
random seed to each instance.
%
An example of the output produced by \irace when testing is shown in Fig.~\ref{fig:testing_output}.

\afterpage{\clearpage
  \begin{figure}[p]
\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\scriptsize,
 %
 frame=none,  % top and bottom rule only
 framesep=0pt, % separation between frame and text
 rulecolor=\color{Gray},
}
% FIXME: This doesn't break long lines, find another alternative.
{\ttfamily\scriptsize\ldots
\VerbatimInput{irace-acotsp-testing.txt}}%
\caption{Sample text output of \irace when evaluating on test instances.}\label{fig:testing_output}
\end{figure}}



The options \parameter{testNbElites} and \parameter{testIterationElites}
control which configurations are evaluated during the testing phase. In
particular, setting \code{testIterationElites = 1} will test not only the final
set of elite configurations (those returned at the end of the training phase),
but also the set of elites at the end of each race
(iteration). The option \parameter{testNbElites} limits the maximum number of
configurations considered within each set. Some examples:
\begin{itemize}
\item \code{testIterationElites = 0; testNbElites = 1} means that only the best
  configuration found during the run of \irace, the final best, will be used in
  the testing phase.
\item \code{testIterationElites = 1; testNbElites = 1} will test, in addition to the final best, the best configuration found at each iteration.
\item \code{testIterationElites = 1; testNbElites = 2} will test the two best
  configurations found at each iteration, in addition to the final best and
  second-best configurations.
\end{itemize}

The testing can be also (re-)executed at a later time by using the following
\aR command (but you may need to override \code{testNbElites} and \code{testIterationElites}):
%
<<testing_r, prompt=FALSE, eval=FALSE>>=
testing_fromlog(logFile = "./irace.Rdata", testNbElites = 1)
@
%
The above line will load the scenario setup from \parameter{logFile} to
perform the testing.  The testing results will be stored in the \aR object
\code{iraceResults\$testing}, which is saved in the file specified by
\code{scenario\$logFile}. The structure of the object is described in
\autoref{sec:output r}.  For examples on how to analyse the results see
\autoref{sec:analysis}.

Another alternative is to test a specific set of configurations using the command-line option \parameter{--only-test} as follows:
\begin{lstlisting}[style=BashInputStyle]
irace --only-test configurations.txt
\end{lstlisting}
where \path{configurations.txt} has the same format as the
set of initial configurations (\autoref{sec:initial}).

%%
%%
%%
%% Recovery
%%
%%
%%
\section{Recovering \irace runs}\label{sec:recovery}

Problems like power cuts, hardware malfunction or the need to use computational
power for other tasks may occur during the execution of \irace, terminating a
run before completion. At the end of each iteration, \irace saves an \aR data
file (\parameter{logFile}, by default \code{"./irace.Rdata"}) that not only
contains information about the tuning progress (\autoref{sec:output r}), but
also internal information that allows recovering an incomplete execution.

To recover an incomplete \irace run, set the option \parameter{recoveryFile} to
the log file previously produced, and \irace will continue the execution from
the last saved iteration. The state of the random generator is saved and
loaded, therefore, as long as the execution is continued in the same machine,
the obtained results will be exactly the same as executing \irace in one step
(external factors, such as CPU load and disk caches, may affect the
target algorithm and that may affect the results).  You can specify
the \parameter{recoveryFile} from the command-line or from the scenario file,
and execute \irace as described in \autoref{sec:execution}. For example, from
the command-line use:

\begin{lstlisting}[style=BashInputStyle]
irace --recovery-file "./irace-backup.Rdata"
\end{lstlisting}


\begin{xwarningbox}
  When recovering a previous run, \irace will try to save data on the file
  specified by the \parameter{logFile} option. Thus, you must specify different
  files for \parameter{logFile} and \parameter{recoveryFile}.  Before
  recovering, we strongly advise to rename the saved \aR data file as in the
  example above, which uses \code{"irace-backup.Rdata"}.
\end{xwarningbox}

\hide{%

%\LESLIE{Actually, there other variables we allow to change...what we shoudl say about this?}
% MANUEL: I would say nothing for now until we fix this properly.
When recovery is done you can modify some \irace options from the command-line
or from the scenario file.

\begin{itemize}
\item \parameter{execDir}
\item \parameter{logFile}
\item \parameter{debugLevel}
\item \parameter{parallel}
\item \parameter{loadBalancing}
\item \parameter{mpi}
\item \parameter{batchmode}
\end{itemize}

It is not possible (and in most cases not correct) to change the scenario options, given that the
previous results can become invalid. However there are some cases in which this options must be changed,
for example, if you have instances that are files, it might be the case that the instances are not
available at the same location. To change the location of them you should modify directly the \aR data file,
please be careful not to change the order of the instances, because this would make the results obtained by
\irace invalid. To change the instances path, open the \aR console and use:

<<change_recover, prompt=TRUE, eval=FALSE>>=
iraceResults <- read_logfile("./tuning/irace.Rdata")
new_path <- "./experiments/tuning/instances/"
iraceResults$scenario$instances <-
   paste0(new_path, basename(iraceResults$scenario$instances))
save(iraceResults, file="./tuning/irace.Rdata")
@

This example can also can be applied to the target execution files \parameter{targetRunner} and \parameter{targetEvaluator}.
}


\begin{xwarningbox}
  Do not change anything in the log file or the scenario file before
  recovering, as it may have unexpected effects on the recovered run of
  \irace. In case of doubt, please contact us first (\autoref{sec:contact}). In particular, it is not possible to continue a run of \irace by recovering with a larger budget. Results will \textbf{not} be the same as running \irace from the start with the largest budget.  An alternative is to use the final configurations from one run as the initial configurations of a new run.
\end{xwarningbox}

\begin{xwarningbox}
  If your scenario uses \parameter{targetEvaluator} (\autoref{sec:evaluator})
  and \parameter{targetEvaluator} requires files created
  by \parameter{targetRunner}, then recovery will fail if those files are not
  present in the \parameter{execDir} directory. This can happen, for example,
  if you recover from a different directory than the one from which irace was
  initially executed, or when \parameter{execDir} is set to a temporary
  directory for every irace run. Thus, you need to copy the contents of the
  previous \parameter{execDir} into the new one.
\end{xwarningbox}

%%
%%
%%
%% Output and analysis of results
%%
%%
%%
\section{Output and results}

During its execution, \irace prints information about the progress of the
tuning in the standard output.  Additionally, after each iteration, an \aR data
file is saved (\parameter{logFile} option) containing the state of \irace.

\subsection{Text output}\label{sec:output text}

Figure~\ref{fig:output} shows the output, up to the end of the first iteration,
of a run of elitist \irace applied to the \ACOTSP scenario with 1000
evaluations as budget.

\afterpage{\clearpage
\begin{figure}[p]
\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\scriptsize,
 %
 frame=none,  % top and bottom rule only
 framesep=0pt, % separation between frame and text
 rulecolor=\color{Gray},
 firstline=1,
 lastline=68,
}
% FIXME: This doesn't break long lines, find another alternative.
\VerbatimInput{irace-acotsp-stdout.txt}%
\caption{Sample text output of \irace.}\label{fig:output}
\end{figure}}

First, \irace gives the user a warning informing that it has found a file with
the default scenario filename and it will use it. Then, general information
about the selected \irace options is printed:
\begin{itemize}
 \item \parameter{nbIterations} indicates the minimum number of iterations \irace has calculated
       for the scenario. Depending on the development of the tuning the final iterations that
       are executed can be more.
  \item \parameter{minNbSurvival} indicates the minimum number of alive configurations that
       are required to continue a race. When less configurations are alive the race is
      stopped and a new iteration begins.
  \item \code{nbParameters} is the number of parameters of the scenario.
  \item \parameter{seed} is the number that was used to initialize the random number generator in \irace.
  \item \code{confidence level} is the confidence level of the statistical test.
  \item \code{budget} is the total number of evaluations available for the tuning.
  \item \code{time budget} is the  maximum execution time available for the tuning.
  \item \parameter{mu} is a value used for calculating the minimum number of iterations.
  \item \parameter{deterministic} indicates if the target algorithm is assumed to be deterministic.
\end{itemize}


At each iteration, information about the progress of the execution is printed as follows:
\begin{itemize}
  \item \code{experimentsUsedSoFar} is the number of experiments from the total budget
        that have been used up to the current iteration.
  \item \code{timeUsed} is the execution time used so far in the experiments. Only available when reported
        in the \parameter{targetRunner} (activate it with the \parameter{maxTime} option).
  \item \code{remainingBudget} is the number of experiments that have not been used yet.
  \item \code{timeEstimate} estimation of the mean execution time. This is used to calculate
        the remaining budget when \parameter{maxTime} is used.
  \item \code{currentBudget} is the number of evaluations \irace has allocated to the
        current iteration.
  \item \parameter{nbConfigurations} is the number of configurations \irace will use in the
        current iteration. In the first iteration, this number of configurations include
        the initial configurations provided; in later iterations, it includes the elite
        configurations from the previous iterations.
\end{itemize}

After the iteration information, a table shows the progress of the iteration execution.
Each row of the table gives information about the execution of an instance in the race.
The first column contains a symbol that describes the results of the
statistical test:
%
\begin{description}
\item \code{|x|} No statistical test was performed for this instance. The options \parameter{firstTest}
      and \parameter{eachTest} control on which instances the statistical test is performed.
\item \code{|-|} Statistical test performed and configurations have been discarded. The column \code{Alive} gives an indication of how
      many configurations have been discarded.
\item \code{|=|} Statistical test performed and no configurations have been discarded.
      This means \irace needs to evaluate more instances to identify the best configurations.
\item \code{|!|} This indicator exists only for the elitist version of \irace. It indicates that
      the statistical test was performed and some elite configurations appear to show bad performance and
      could be discarded but they are kept because of the elitist rules. See option \parameter{elitist} in \autoref{sec:irace options} for more information.
\end{description}

Other columns have the following meaning:
\begin{description}
  \item \code{Instance}: Index of \code{(instance, seed)} pair executed. This number
corresponds to the row in the data frame returned by \code{get_instanceID_seed_pairs()}. See \autoref{sec:output r}
for more information. This index is different from the instance ID passed to \parameter{targetRunner}.
\item \code{Bound}: Only when \parameter{capping} is enabled. Execution time used as bound
for the execution of new candidate configurations.
\item \code{Alive}: Number of configurations that have not been
discarded after the statistical test was performed.
\item \code{Best}: ID of the best configuration according to the instances seen
so far in this race (i.e., not including previous iterations).
\item \code{Mean best}: Mean cost value of the best configuration across the instances seen so
far in this race (not globally). Equivalent to the concept of ``iteration-best'' in other algorithms.
\item  \code{Exp so far}: Number of experiments
performed so far.
\item \code{W time}:  Wall-clock time spent on this instance.
\item \code{rho}, \code{KenW}, and \code{Qvar}: Spearman's rank correlation
  coefficient rho, Kendall's concordance coefficient W, and a variance measure
  described in \cite{SchHoo2012quanti}, respectively, of the configurations
  across the instances evaluated so far in this iteration.  These measures
  evaluate how consistent is the performance of the configurations across the
  instances.  Values close to 1 for \code{rho} and \code{KenW} and values close
  to 0 for \code{Qvar} indicate that the scenario is highly homogeneous. For
  heterogeneous scenarios, we provide advice in \autoref{sec:het}.
\end{description}

Finally, \irace outputs the best configuration found and a list of the elite
configurations. The elite configurations are configurations that did not show
statistically significant difference during the race; they are ordered
according to their mean performance on the executed instances.


\subsection[R data file (logFile)]{\aR data file (\code{logFile})}\label{sec:output r}

The \aR data file created by \irace (by default as \code{irace.Rdata}, see option \parameter{logFile}) contains an object called
\code{iraceResults}. You can load this file in the \aR console with:

<<load_rdata, prompt=FALSE, eval=FALSE>>=
logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
iraceResults <- read_logfile(logfile)
@

The \code{iraceResults} object is a list, and the elements of a list can be
accessed in \aR by using the \code{\$} or \code{[[]]} operators:

<<show_version, prompt=TRUE, eval=TRUE, comment="">>=
iraceResults$irace_version
@

\noindent%
The \code{iraceResults} list contains the following elements:

\begin{itemize}
\item \code{scenario}: The scenario \aR object containing the \irace options
  used for the execution. See \autoref{sec:irace options} and the help
  of the \pkg{irace} package; open an \aR console and type: \code{?defaultScenario}.
  See \autoref{sec:irace options} for more information.

\item \code{parameters}: The parameters \aR object containing the description of
  the target algorithm parameters. See \autoref{sec:target parameters}.

\item \code{allConfigurations}: The target algorithm configurations generated by
  \irace. This object is a \code{data frame}, each row is a candidate
   configuration; the first column (\code{.ID.}) indicates the internal identifier of the
   configuration; the final column (\code{.PARENT.}) is the identifier
   of the configuration from which the current configuration was sampled; and the remaining columns correspond to the parameter values; each column is named as
   the parameter name specified in the parameter object.

<<show_configurations, prompt=TRUE, eval=TRUE, comment="">>=
head(iraceResults$allConfigurations)
@

\item \code{allElites}: A list that contains one element per iteration. Each element contains
  the internal identifier of the elite candidate configurations of the corresponding iteration
  (identifiers correspond to \code{allConfigurations$.ID.}).

<<show_idelites, prompt=TRUE, eval=TRUE, comment="">>=
print(iraceResults$allElites)
@

The configurations are ordered by mean performance, that is, the ID of the best
configuration corresponds to the first ID. To obtain the values of the
parameters of all elite configurations found by \irace use:

<<get_elites, prompt=TRUE, eval=TRUE, comment="">>=
logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
getFinalElites(logfile, n = 0)
@

\item \code{iterationElites}: A vector containing the best candidate configuration ID of each iteration. The best configuration found corresponds to the last one
  of this vector.
<<show_iditelites, prompt=TRUE, eval=TRUE, comment="">>=
print(iraceResults$iterationElites)
@

One can obtain the full configuration with:

<<get_elite, prompt=TRUE, eval=TRUE, comment="">>=
last <- length(iraceResults$iterationElites)
id <- iraceResults$iterationElites[last]
getConfigurationById(iraceResults, ids = id)
@

\item \code{rejectedConfigurations}: A vector containing the rejected configurations IDs. These
correspond to configurations that produced failed executions and were ignored by \irace during the
configuration process. See \autoref{sec:reject} to enable the detection of such configurations.

\item \code{experiments}: A matrix with configurations as columns and instances
as rows. Column names correspond to the internal identifier of the configuration
(\code{allConfigurations$.ID.}). The results of a particular
configuration can be obtained using:

<<get_experiments, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we use the best configuration found
best_config <- getFinalElites(iraceResults, n = 1)
best_id <- as.character(best_config$.ID.)
# Obtain the results of the best configuration
all_exp <- iraceResults$experiments[, best_id]
# all_exp is a vector and names(all_exp) is the (instance,seed) index.
all_exp
# Obtain the results of the first and best configurations
all_exp <- iraceResults$experiments[, c("1", best_id)]
# all_exp is a matrix: colnames(all_exp) is configurationID and
# rownames(all_exp) is the (instance,seed) index.
all_exp
@

When a configuration was not executed on an instance, its value is \code{NA}. A
configuration may not be executed on an instance because: (1) it was not created
yet when the instance was used, or (2) it was
discarded by the statistical test and not executed on subsequent instances, or
(3) the race terminated before this instance was considered.

Row names correspond to the row index of the \code{(instance, seed)} pairs
in the data frame returned by \code{get_instanceID_seed_pairs()}. The instanceID and seed used
for a particular experiment can be obtained with:\sloppy

<<get_instance_seed, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we get instanceID, seeds and instances of the experiments
# of the best configuration.
# We could get the indexes of the instances on which at least one
# configuration was executed:
pair_index <- which(apply(!is.na(all_exp), 1L, any))
# or the instances on which all configurations were executed:
pair_index <- which(apply(!is.na(all_exp), 1L, all))
# but in this example we get the indexes of the instances executed for
# the best configuration.
pair_index <- which(!is.na(all_exp[, best_id]))
instanceID <- get_instanceID_seed_pairs(iraceResults)[["instanceID"]][pair_index]
# or get the seeds
get_instanceID_seed_pairs(iraceResults)[["seed"]][pair_index]
# or obtain the actual instances.
iraceResults$scenario$instances[instanceID]
# If the instances are of atomic type (integers, floating-point numbers or
# character strings), the above is similar to:
get_instanceID_seed_pairs(iraceResults, index = pair_index, instances=TRUE)
@
\item \code{experimentLog}: A matrix with columns \code{iteration, instance, configuration}. This matrix contains the log of all the experiments that \irace
performs during its execution. The \code{instance} column refers to the index of the
data frame returned by \code{get_instanceID_seed_pairs()}. When \parameter{capping} is enabled a column \code{bound} is added to log the execution bound applied
for each execution.

\item \code{softRestart}: A logical vector that indicates if a soft restart was
performed on each iteration. If \code{FALSE}, then no soft restart was performed.
See option \parameter{softRestart} in \autoref{sec:irace options}.

\item \code{state}: A list that contains the state of \irace, the recovery
(\autoref{sec:recovery}) is done using the information contained in this object. The
probabilistic model of the last elite configurations can be found here by doing:

<<get_model, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we get the model probabilities for the
# localsearch parameter.
iraceResults$state$model["localsearch"]
# The order of the probabilities corresponds to:
iraceResults$scenario$parameters$domains$localsearch
@

The example shows a list that has one element per elite configuration (ID as
element name). In this case, \code{localsearch} is a categorical parameter and
it has a probability for each of its values.


\item \code{testing}: A list that contains the testing results.
The list contains the following elements:
  \begin{itemize}
    \item \code{experiments}: Matrix of experiments in the same format as the
      \code{iraceResults$experiments} matrix. The column names indicate the candidate
      configuration identifier and the row names contain the name of the instances.
<<get_test_exp, prompt=TRUE, eval=TRUE, comment="">>=
# Get the results of the testing
iraceResults$testing$experiments
@

    \item \code{seeds}: The seeds used for the experiments, each seed corresponds to
      each instance in the rows of the test \code{experiments} matrix.
<<get_test_seeds, prompt=TRUE, eval=TRUE, comment="">>=
# Get the seeds used for testing
iraceResults$testing$seeds
@

In the example, instance \code{1000-1.tsp} is executed with seed \Sexpr{iraceResults$testing$seeds[1]}.

  \end{itemize}

\end{itemize}

\subsection{Analysis of results}\label{sec:analysis}

The final configurations returned by \irace are the elites of the final race.
They are reported in decreasing order of performance, that is, the best
configuration is reported first.

If testing is performed, you can further analyze the resulting best
configurations by performing statistical tests in \aR:
% FIXME: iraceplot
% or just plotting the results:
%<<plot_test, fig.pos="tbp", fig.align="center", fig.height = 4, fig.width = 8, out.width='0.85\\textwidth', fig.cap="Boxplot of the testing results of the best configurations.", prompt=TRUE, eval=TRUE, comment="">>=
<<wilcox_test,prompt=TRUE, eval=TRUE, comment="">>=
results <- iraceResults$testing$experiments
# Wilcoxon paired test
conf <- gl(ncol(results), # number of configurations
           nrow(results), # number of instances
           labels = colnames(results))
pairwise.wilcox.test (as.vector(results), conf, paired = TRUE, p.adj = "bonf")
@
% # Plot the results
% configurationsBoxplot (results, ylab = "Solution cost")

The Kendall concordance coefficient (\code{W}) and the Spearman's rho can be
applied over data that has the characteristics of the data obtained in the testing,
that is a full matrix where all configurations are executed in all instances. \code{W}
can show if the configurations tested have an homogeneous performance on the used instances
set. If evidence of an heterogeneous scenario found we recommend to make some adjustments
in the \irace options as described in \autoref{sec:het}.

<<conc, prompt=TRUE, eval=TRUE, comment="">>=
irace:::concordance(iraceResults$testing$experiments)
@
%

%% FIXME: iraceplot
% During the tuning, \irace iteratively updates the sampling models of the
% parameters to focus on the best regions of the parameter search space. The
% frequency of the sampled configurations can provide insights on the parameter
% search space. We provide a function for plotting the
% frequency of the sampling of a set of configurations. For more information on this function, please see the \aR help, type in the
% \aR console: \code{?parameterFrequency}. The following example
% plots the frequency of the parameters sampled during one \irace run:

% <<freq, fig.pos="tbp", fig.cap="Parameters sampling frequency.", out.width="\\textwidth",prompt=TRUE, eval=TRUE, comment="">>=
% parameterFrequency(iraceResults$allConfigurations, iraceResults$scenario$parameters)
% @

%% FIXME: iraceplot
% By using parallel coordinates plots, it is possible to analyze how the parameters
% interact with each other. For more information on this function, please see the \aR help, type
% in the \aR console: (\code{?parallelCoordinatesPlot}).
%  The following example shows how to create a parallel
% coordinate plot of the configurations in the last two iterations of \irace.

% <<parcord, fig.pos="tbp", fig.align="center", out.width="0.7\\textwidth", fig.cap="Parallel coordinate plots of the parameters of the configurations in the last two iterations of a run of \\irace.", prompt=FALSE, eval=TRUE, comment="">>=
% # Get last iteration number
% last <- length(iraceResults$iterationElites)
% # Get configurations in the last two iterations
% conf <- getConfigurationByIteration(iraceResults,
%                                     iterations = c(last - 1, last))
% parallelCoordinatesPlot (conf, iraceResults$scenario$parameters,
%                          param_names = c("algorithm", "alpha", "beta", "rho",
%                                          "q0", "rasrank"),
%                          hierarchy = FALSE)
% @

%%% FIXME: This plot is completely misleading. We should either calculate the
%%% performance over the whole training set, or find a way to plot the
%%% performance estimated by irace as a confidence interval to account for
%%% different number of instances. Or simply plot a heatmap of instances x
%%% best-of-each iteration.
% It is also possible to plot the performance evolution of the best-so-far configuration over the number of experiments as follows:
% %%<<trainEvo, fig.pos="tbp", fig.align="center", out.width='0.75\\textwidth', fig.cap="Training set performance of the best-so-far configuration over number of experiments.", prompt=FALSE, eval=TRUE, comment="">>=
% # Get number of iterations
% iters <- unique(iraceResults$experimentLog[, "iteration"])
% # Get number of experiments (runs of target-runner) up to each iteration
% fes <- cumsum(table(iraceResults$experimentLog[,"iteration"]))
% # Get the mean value of all experiments executed up to each iteration
% # for the best configuration of that iteration.
% values <- sapply(iters, function(k) {
%   instances <- as.character(
%     unique(iraceResults$experimentLog[
%                           iraceResults$experimentLog[, "iteration"] == k,
%                           "instance"]))
%   return(mean(iraceResults$experiments[
%                              instances,
%                              as.character(iraceResults$iterationElites[k])]))})
% plot(fes, values, type="s", xlab = "Number of runs of the target algorithm",
%      ylab = "Estimated mean value over whole training set")
% points(fes, values)
% @

% FIXME: We should normalize the ranges per instance so since the performance
% on different instances is quite different.
It is also possible, as shown in Fig.~\ref{fig:testEvo}, to plot the performance on the test set of the best-so-far configuration over the number of experiments as follows:
<<testEvo, fig.pos="tb", fig.align="center", out.width='0.7\\textwidth', fig.cap="Testing set performance of the best-so-far configuration over number of experiments. Label of each point is the configuration ID.", prompt=FALSE, eval=TRUE, comment="">>=
# Get summary data from the logfile.
irs <- irace_summarise(iraceResults)
# Get number of iterations
iters <- irs$n_iterations
# Get number of experiments (runs of target-runner) up to each iteration
fes <- cumsum(table(iraceResults$state$experiment_log[["iteration"]]))
# Get the mean value of all experiments executed up to each iteration
# for the best configuration of that iteration.
elites <- as.character(iraceResults$iterationElites)
values <- colMeans(iraceResults$testing$experiments[, elites])
stderr <- function(x) sqrt(var(x)/length(x))
err <- apply(iraceResults$testing$experiments[, elites], 2L, stderr)
plot(fes, values, type = "s",
     xlab = "Number of runs of the target algorithm",
     ylab = "Mean value over testing set", ylim=c(min(values-err),max(values+err)))
points(fes, values, pch=19)
arrows(fes, values - err, fes, values + err, length=0.05, angle=90, code=3)
text(fes, values, elites, pos = 1)
@

The \irace package also provides an implementation of the ablation
method~\cite{FawHoos2015ablation}. See \autoref{sec:ablation}.

\medskip{}
Finally, more advanced visualizations of the behavior of \irace are provided by
the ACVIZ software package~\cite{DesRitLopPer2021acviz}, which is available at \url{https://github.com/souzamarcelo/acviz}. See an example in Fig.~\ref{fig:acviz}.

\begin{figure}[htb]
  \centering
  \includegraphics[width=\textwidth]{fig1u-acotsp-instances}
  \caption{Visualization produced by ACVIZ~\cite{DesRitLopPer2021acviz}.}
  \label{fig:acviz}
\end{figure}



%%
%%
%%
%% Advanced options
%%
%%
%%

\section{Advanced topics}

\subsection{Tuning budget}\label{sec:budget}

Before setting the budget for a run of \irace, please consider the number of
parameters that need to be tuned, available processing power and available
time. The optimal budget depends on the difficulty of the tuning scenario, the
size of the parameter space and the heterogeneity of the instances. Typical
values range from $1\,000$ to $100\,000$ runs of the target algorithm, although
smaller and larger values are also possible. Currently, \irace does not detect
whether the given budget allows generating all possible configurations. In such
a case, the use of \emph{iterated} racing is unnecessary: One can simply
perform a single race of all configurations (see FAQ in
\autoref{faq:allconfs}).

\Irace provides two options for setting the total tuning budget (\parameter{maxExperiments} and
\parameter{maxTime}). The option \parameter{maxExperiments}
limits the number of executions of \parameter{targetRunner} performed by \irace. The option \parameter{maxTime} limits the
total time of the \parameter{targetRunner} executions. When this latter option is used, \parameter{targetRunner}
must return the evaluation cost together with the execution time (\code{"cost time"}).

\begin{xwarningbox}
When the goal is to minimize the computation time of an algorithm, and you wish to use \parameter{maxTime} as the tuning budget, \parameter{targetRunner} must return the time also as the evaluation cost, that is, return the time twice as \code{"time time"}.
\end{xwarningbox}

\begin{xwarningbox}
When using \parameter{targetEvaluator} and using \parameter{maxTime} as tuning budget, \parameter{targetRunner} just returns the time (\code{"time"}) and \parameter{targetEvaluator} returns the cost.
\end{xwarningbox}

When using \parameter{maxTime}, \irace estimates the execution time of each \parameter{targetRunner} execution
before the configuration. The amount of budget used for the estimation is set with the option \parameter{budgetEstimation}
(default is $2\%$). The obtained estimation is adjusted after each iteration using the obtained results and it is used to estimate the number of experiments that can be executed. Internally, \irace uses the number of remaining experiments to
adjust the number of configurations tested in each race.

\subsection{Multi-objective tuning}\label{sec:multi objective}

Currently, \irace only optimizes one cost value at a time, which can be
solution cost, computation time or any other objective that is returned to
\irace by the \parameter{targetRunner}. If the target algorithm is
multi-objective, it will typically return not a single cost value, but a set of
objective vectors (typically, a Pareto front). For tuning such a target
algorithm with \irace, there are two alternatives. If the algorithm returns a
single vector of objective values, they can be aggregated into one single
number by using, for example, a weighted sum. Otherwise, if the target
algorithm returns a set of objective vectors, a unary quality metric (\eg~the
hypervolume) may be used to evaluate the quality of the set.\footnote{Multiple
  metrics are implemented in C, R and Python in the \texttt{moocore} package
  (\url{https://github.com/multi-objective/moocore}).}

% The first option is simple, it requires to devise a formula that can aggregate
% the objectives in a way that balances the importance of all of them. This might
% not be an easy task in some scenarios, and therefore using a more adequate
% indicator to evaluate the performance of a multi-objective optimizer, such as
% the hypervolume, is strongly advised.

%For using the hypervolume as evaluation \irace needs to postpone the evaluation of the
%configurations in an instance until all the executions have been completed. Once
%the executions are finalized, the evaluation starts by obtaining reference points for
%each objective from the configuration results. These points are used to define the
%Pareto front and the hypervolume of each configuration
%is calculated and the tuning process continues normally.

The use of aggregation or quality metrics often requires normalizing the
different objectives. If normalization bounds are known a priori for each
instance, normalized values can be computed by \parameter{targetRunner}.
Otherwise, the bounds may be dynamically computed while running \irace, by
using \parameter{targetEvaluator}. In this case, \parameter{targetRunner} will
save the output of the algorithm, then the first call
to \parameter{targetEvaluator} will examine the output produced by all calls
to \parameter{targetRunner} for the same instance, update the normalization
bounds and return the normalized output. Subsequent calls
to \parameter{targetEvaluator} for the same instance will simply return the
normalized output.
%
A similar approach can be used to dynamically compute the reference points or
sets often required by unary quality metrics.

For more information about defining a \parameter{targetEvaluator}, see
\autoref{sec:evaluator}. Examples of tuning a multi-objective target algorithm
using the hypervolume can be found in the examples at
\IRACEHOME{/examples/hypervolume} and \IRACEHOME{/examples/moaco}.

\subsection{Improving anytime behavior}\label{sec:anytime}

In some scenarios, we may be unsure when the end-user may terminate the
target-algorithm, and we would like to tune the parameters of the algorithm so
that it finds a solution as good as possible for any given termination
criterion. That is, we would like to obtain a parameter configurations with
good \emph{anytime performance}.

Following the approach proposed by \citet{LopStu2013ejor}, \irace can tune an
algorithm to improve its anytime performance. For that purpose, the algorithm
must output a (time, solution cost) pair every time that the best-so-far
solution within a run is improved.  Such pairs form a matrix with two columns
and as many rows as improvements were made. We can define a \emph{reference
  point}, where the first component is value slightly higher than the maximum
runtime of a single run and the second component is a maximum cost value such
that differences above that maximum are irrelevant for evaluating
performance. Using this reference point and the matrix as an input, one can
calculate the hypervolume of the matrix, as described above in
Section~\ref{sec:multi objective}.  The negative of this hypervolume value
becomes the cost returned by the \parameter{targetRunner} to \irace.

It is also possible to specify preferences over the trade-off between runtime
and solution cost by using the weighted hypervolume.  Further details and examples
can be found in \citet{LopStu2013ejor}.

Although there are other ways to measure anytime performance, most of them are
equivalent or an approximation of the measure described
above~\citep{LopVerDreDoe2025}.



\subsection{Tuning for minimizing computation time}
\label{sec:capping}

When using \irace for tuning algorithms that only report computation time to
reach a target, \parameter{targetRunner} should return the execution time of a configuration instead of solution cost. When using \parameter{maxTime} as the budget, this means that \parameter{targetRunner} must return twice the execution time since the first value is the minimization objective and the second value is used to track the budget consumed.

Starting from version $3.0$,  \irace includes an elitist racing procedure that
implements an \textbf{adaptive capping mechanism} \citep{PerLopHooStu2017:lion}.
Adaptive capping \citep{HutHooLeyStu2009jair} is a configuration technique
that avoids the execution of long runs of the target algorithm, focusing the
configuration budget in the evaluation of the best configurations found. This is
done by bounding the execution time of each configuration based on the best
performing candidate configurations.

To use adaptive capping, the \parameter{capping} option must be enabled and the
\parameter{elitist} irace option must be selected. When evaluating candidate
configurations on an instance, \irace calculates an execution bound based on the
execution times of the elite configurations.  The \parameter{boundType} option defines
how the performance of the elite configurations  is defined to obtain the execution bound.
The default value of \parameter{boundType} calculates the performance ($p_i^s$) of each
elite configuration ($s$) as the mean execution time of the  instances already executed
in the race and the currently executed instance ($i$). The \parameter{cappingType} option
specifies the measure used to obtain the elite configurations bound. By default, the execution bound
is calculated as the median of the execution times of the elite configurations:
%
\begin{equation}
  b_i= \text{Median}_{\theta_s \in \Celite} \{ p_i^s\}
\end{equation}

The execution bound for new configurations ($j$) is calculated by multiplying the elite configurations
bound by the number of instances ($i$) in the execution list and subtracting the mean execution time
of the instances executed by the candidate:
%
\begin{equation}
    k_i^{'j} = b_{i} \cdot i + \bmin - p^{j}_{i-1} \cdot (i-1)
\end{equation}

A small constant \bmin is added to account for time measurements errors.
These settings are also used to apply a dominance elimination criterion together with the
statistical test elimination. The domination criterion is defined as:
%
\begin{equation}\label{eq:dom}
b_i + \bmin  < p_i^{j}
\end{equation}

When elite configurations dominate new configurations, these are eliminated from the race.

\begin{xwarningbox}
The default statistical test when \parameter{capping} is enabled is \code{t-test}.
This test is more appropriate to configure algorithms for optimizing runtime (see \autoref{sec:stat test}).
\end{xwarningbox}

The execution bound is constantly adjusted by \irace based on the best configurations
times, nevertheless, a maximum execution time (\bmax) is never  exceeded. This maximum
execution time must be defined in the configuration scenario when \parameter{capping}
is enabled. To specify the maximum execution bound for the target runner executions
use the \parameter{boundMax} option. The final execution bound ($k_i^j$) is calculated by:
%
\begin{equation}\label{eq:cap}
  k_i^j = \begin{cases}
    \bmax &\text{if $k_i^{'j} > \bmax$,} \\
    \min \{b_i, \bmax\}& \text{if $k_i^{'j} \leq 0$,}\\
    k_i^{'j}&\text{otherwise;}
  \end{cases}
\end{equation}

Additionally, the \parameter{boundDigits} option defines the precision of the time bound
provided by \irace, the default setting is 0.

Timed out executions occur when the maximum execution bound
(\parameter{boundMax}) is reached and the algorithm has not achieved successful
termination or a defined quality goal.  In this case, it is a common practice
to apply a penalty known as PARX, in which timeouts are penalized by
multiplying \parameter{boundMax} by a constant $X$. The constant $X$ may be set
using the \parameter{boundPar} option. Bounded executions are executions that
do not achieve successful termination or a defined quality goal in the
execution bound ($k_i^j$) set by \irace, which is smaller than
\parameter{boundMax}. The \parameter{boundAsTimeout} option replaces the
evaluation of bounded executions by the \parameter{boundMax} value. More
details about the implementation of adaptive capping can be found in
\citet{PerLopHooStu2017:lion}.

\begin{xwarningbox}
Note that bounded executions are not timed out executions and thus, they will not be penalized by PARX.
\end{xwarningbox}


\begin{xwarningbox}
  Penalized evaluations of timed out and bounded executions are only used for
  the elimination tests and the comparison between the quality of
  configurations. To calculate execution bounds and computation budget
  consumed, \irace uses only unpenalized execution times. The unpenalized
  execution time must be provided by the target runner or target evaluator as
  described in \autoref{sec:runner} and \autoref{sec:evaluator} .
\end{xwarningbox}

\begin{xwarningbox}
More advanced capping methods that are applicable to minimizing solution cost are available when combining irace with the \texttt{capopt} package described by \citet{SouRitLop2021cap}.
\end{xwarningbox}



\subsection{Hyper-parameter optimization of machine learning methods}

The \irace package can also be used for model selection and hyper-parameter
optimization of machine learning (ML) methods. We will next explain a possible
setup for one given dataset and using $10$-fold cross-validation
(CV). Generalizing to multiple datasets and different resampling strategies,
e.g. leave-one-out, is straightforward.

First, split the dataset into training, to be used by \irace, and testing, to
be used for evaluating the performance of the configuration returned by
\irace. A typical split could be $70\%$ and $30\%$, respectively.

The training set is used by \irace to perform $10$-fold CV, that is, the data
is split into $10$ folds. A single run of the \parameter{targetRunner} will use
$9$ folds for training and the remaining fold for validation. Splitting the
data into folds can be done at each call of \parameter{targetRunner} or before
running \irace, however, it is important that the split is always the same for
every call of the \parameter{targetRunner}, i.e., the content of the folds does
not change, only which folds are used for training and validation will change.

The setup of \irace should be as follows:
%
\begin{itemize}
\item \parameter{trainInstancesFile}\code{="train-instances.txt"}, where this
  file contains one number per line from $1$ to $10$. This number will tell the
  \parameter{targetRunner} which fold should be used for validation.
\item \parameter{trainInstancesDir}\code{=""}, because the folds are the
  ``instances'' and you do not have actual instance files. If you want to pass
  the name of the dataset to the \parameter{targetRunner}, you can specify it
  either at each line of \code{"train-instances.txt"}, directly in the
  \parameter{targetRunner}, or as a fixed parameter in the
  \parameter{parameterFile}.
\item \parameter{deterministic}\code{=1} unless it really makes sense to
  train more than once the same ML model on the same data. If it makes sense,
  then your \parameter{targetRunner} should use the seed passed by \irace to
  seed the ML model before training.
\item \parameter{sampleInstances}\code{=0} because the folds should already be generated by randomly sampling the dataset.
\item \parameter{testType}\code{="t-test} because the performance metrics in
  ML are typically the mean of the CV results, which assumes that the
  performance are close to normally distributed.
\item \parameter{firstTest}\code{=2} because \irace should discard
  configurations very aggressively looking for maximum generality.
\end{itemize}

\noindent%
Finally, your \parameter{targetRunner} needs to be able to do the following:
\begin{itemize}
\item Receive from \irace the hyper-parameter settings, the dataset name and a
  fold number (the ``instance''). Let us use fold 3 as an example.
\item Train the ML model on the whole training set minus fold 3, then validate
  (score) the model on fold 3 and return the score to \irace (negated if the
  score must be maximized, because \irace assumes minimization). Since each
  fold is different, each instance should give a different result. Each row in
  the table printed by \irace should print something different; otherwise,
  something is wrong in your setup.
\end{itemize}

The above is actually 10 times faster than doing 10-fold CV for each call to
\parameter{targetRunner}, thus, you should assign to \irace 10 times the budget
than what would be assigned to other methods that do a complete 10-fold CV at
each step.

%% FIXME: They removed the tutorial
%% For such a task, we recommend the \pkg{mlr} package~\citep{R:mlr}.
%% The following webpage documents how to use \irace for this purpose: \url{https://mlr-org.github.io/mlr-tutorial/devel/html/advanced_tune/index.html}


\subsection{Heterogeneous scenarios}
\label{sec:het}

We classify a scenario as homogeneous when the target algorithm has a
consistent performance regarding the instances; roughly speaking, good
configurations tend to perform well and bad configurations tend to perform
poorly on all instances of the problem. By contrast, in heterogeneous
scenarios, the target algorithm has an inconsistent performance on different
instances, that is, some configurations perform well for a subset of the
instances, while they perform poorly for a different subset.

When facing a heterogeneous scenario, the first question should be whether the
objective of tuning is to find configurations that perform reasonably well over
all instances, even if that configuration is not the best ones in any
particular instance (a generalist). If this is not the goal, then it would be
better to partition instances into more similar subsets and execute \irace
separately on each subset. This will lead to a portfolio of algorithm
configurations, one for each subset, and algorithm selection techniques can be
used to select the best configuration from the portfolio when facing a new
instance.

To make sure \irace is not misled by results on few instances, it may be useful
to increase the number of instances executed before doing a statistical test
using the option \parameter{firstTest}, \eg \code{--first-test 10} (default
value is 5), in order to see more instances before discarding configurations.
The option \parameter{elitistNewInstances} in elitist \irace (option
\parameter{elitist}) can be used to increase the number of new instances
executed in each iteration, \eg \code{--elitist-new-instances 5} (default value
is 1).

If finding an overall good configuration for all the instances is the
objective, then we recommend that instances are randomly sampled (option
\parameter{sampleInstances}), unless one can provide the instances in a
particular order that does not bias the tuning towards any subset.

If instances are easily categorized in different classes, then we recommend to
create ``blocks'' of instances in \parameter{trainInstancesFile}, where each
block should contain one instance from each class.  Then set the option
\parameter{blockSize} to the number of classes within each block, so that
\irace will always see a complete block of instances before eliminating
configurations. The value of \parameter{blockSize} will multiply the values of
\parameter{firstTest}, \parameter{eachTest} and
\parameter{elitistNewInstances}. Randomly sampling instances
(\parameter{sampleInstances}\code{=1}) will randomly sample the blocks but not
break the blocks.

While executing \irace, the homogeneity of the scenario can be observed by
examining the values of Spearman's rank correlation coefficient and Kendall's
concordance coefficient in the text output of \irace.  See \autoref{sec:output
  text} for more information.

\subsection{Choosing the statistical test}
\label{sec:stat test}

The statistical test used in \irace identifies statistically bad performing
configurations that can be discarded from the race in order to save
budget. Different statistical tests use different criteria to compare the
cost of the configurations, which has an effect on the tuning results.

\Irace provides two types of statistical tests
(option \parameter{testType}). Each test has different characteristics that are
beneficial for different goals:
%
\begin{itemize}
\item Friedman test (\code{F-test}): This test uses the ranking of the
  configurations to analyze the differences between their performance. This
  makes the test suitable for scenarios where the scale of the performance
  metric is not as important to assess configurations as their relative
  ranking.  This test is also indicated when the distribution of the mean
  performances deviates greatly from a normal distribution. For example, the
  ranges of the performance metric on different instances may be completely
  difference and comparing the performance of different configurations using
  the mean over multiple instances may be deceiving.  We recommend to use the
  \code{F-test} (default when \parameter{capping} is not enabled) when tuning
  for solution cost and whenever the best performing algorithm should be among
  the best in as many instances as possible.

\item Student's t-test (\code{t-test}): This test uses the mean performance of
  the configurations to analyze the differences between the
  configurations.\footnote{The t-test does not require that the performance
    values follow a normal distribution, only that the distribution of sample
    means does. In practice, the t-test is robust despite large deviations from
    the assumptions.} This makes the test suitable for scenarios where the
  differences between values obtained for different instances are relevant to
  assess good configurations.  We recommend using t-test, in particular, when
  the target algorithm is minimizing computation time and, in general, whenever
  the best configurations should obtain the best average solution cost.
\end{itemize}


The confidence level of the tests may be adjusted by using the
option \parameter{confidence}. Increasing the value of \parameter{confidence}
leads to a more strict statistical test.  Keep in mind that a stricter test
will require more budget to identify which configurations perform worse. A less
strict test discards configurations faster by requiring less evidence against
them and, therefore, it is more likely to discard good configurations.


\subsection{Complex parameter space constraints}\label{sec:complex_domains}

Some parameters may have complex dependencies. Ideally, parameters should be
defined in the way that is more likely to help the search performed by
\irace.  For example, when tuning a branch and bound algorithm, one may have the
following parameters:

\begin{itemize}
\item branching (\code{b}) that takes values in \code{\{0,1,2,3\}}, where 0 indicates no branching will be used
      and the rest are different types of branching.
\item stabilization (\code{s}) that takes values in \code{\{0,1,2,3,4,5,6,7,8,9,10\}}, of which for \code{b=0}
      only \code{\{0,1,2,3,4,5\}} are relevant.
\end{itemize}

In this case, it is not possible to describe the parameter space by defining
only two parameters for \irace. An extra parameter must be introduced as
follows:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range                    condition
b         "-b "   c      (0,1,2,3)
s1        "-s "   c      (0,1,2,3,4,5)            | b == "0"
s2        "-s "   c      (0,1,2,3,4,5,6,7,8,9,10) | b != "0"
  \end{CodeInput}
\end{minipage}
\end{center}

Parameters whose values depend on the value of other parameters may also
require using extra parameters or changing the parameters and processing them
in \parameter{targetRunner}. For example, given the following parameters:
%
\begin{itemize}
\item Population size (\code{p}) takes the integer values $[1,100]$.
\item Selection size (\code{s}) takes the same values but no more than the
  population size, that is $[1,$\code{p}$]$.
\end{itemize}

In this case, it is possible to describe the parameters \code{p} and \code{s}
using surrogate parameters for \irace that represent a ratio of the original
interval as follows:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range
p        "-p "    i      (1,100)
s_f      "-s "   r      (0.0,1.0)
\end{CodeInput}
\end{minipage}
\end{center}
%
and \parameter{targetRunner} must calculate the actual value of \code{s} as $\min(\max(\text{round}(\text{\code{s\_f}}\cdot \text{\code{p}}, 1)), 100)$. For
example, if the parameter \code{p}  has value 50 and the surrogate parameter \code{s\_f} has value 0.3, then \code{s} will have value 15.

The processing within \parameter{targetRunner} can also split and join
parameters. For example, assume the following parameters:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range
m         "-m "   i      (1,250)
e         "-e "   r      (0.0,2.0)
\end{CodeInput}
\end{minipage}
\end{center}

These parameters could be used to define a value
$\texttt{m}\cdot 10^\texttt{e}$ for another parameter
(\code{--strength}) not known by \irace. Then, \parameter{targetRunner} takes care of parsing
\code{-m} and \code{-e}, computing the strength value and passing the parameter
\code{--strength} together with its value to the target algorithm.

More complex parameter space constraints may be implemented by means of the
\parameter{repairConfiguration} function (\autoref{sec:repairconf}).


\subsection{Unreliable target algorithms and immediate rejection}\label{sec:reject}

There are some situations in which the target algorithm may fail to execute
correctly.  By default, \irace stops as soon as a call
to \parameter{targetRunner} or \parameter{targetEvaluator} fails, which helps to
detect bugs in the target algorithm.  Sometimes the failure cannot be fixed
because it is due to system problems, network issues, memory limits, bugs for
which no fix is available, or fixing them is impossible because there is no
access to the source code.

In those cases, if the failure is caused by random errors or transient system
problems, one may wish to ignore the error and try again the same call in the
hope that it succeeds. The option \parameter{targetRunnerRetries} indicates the
number of times a \parameter{targetRunner} execution is repeated if it
fails. Use this option only if you know additional repetitions could be
successful.

If the target algorithm consistently fails for a particular set of
configurations, these configurations may be declared as forbidden
(\autoref{sec:forbidden}) so that \irace avoids them.  On the other hand, if
the configurations that cause the problem are unknown,
the \parameter{targetRunner} should return \code{Inf} so that \irace
immediately rejects the failing configuration. This immediate rejection should
be used with care according to the goals of the tuning. For example, a
configuration that crashes on a particular instance, \eg by running out of
memory, might still be considered acceptable if it gives very good results on
other instances. The configurations which were rejected during the execution of \irace are saved
in the Rdata output file (see \autoref{sec:output r}).
\begin{xwarningbox}
If the configuration budget is specified in total execution time (\parameter{maxTime} option),
immediate rejected executions must provide the cost and time (which must be \code{Inf 0}).
Nevertheless, rejected configurations will be  excluded from the execution time estimation
and the execution bound calculation.
\end{xwarningbox}

\subsection{Ablation Analysis}\label{sec:ablation}

The ablation method~\cite{FawHoos2015ablation} takes two configurations (source
and target) and generates a sequence of configurations that differ between each
other just in one parameter, where parameter values in source are replaced by
values from target. The sequence can be seen as a ``path'' from the source to
the target configuration.  This can be used to find new better ``intermediate''
configurations or to analyse the impact of the parameters in the performance.


To perform ablation, you can use the \code{ablation()} \aR function
%
<<ablation, prompt=FALSE, eval=FALSE>>=
ablog <- ablation("irace.Rdata", src = 1)
plotAblation(ablog)
@
%
\noindent or the \path{ablation} command-line executable (see more details below).
%
%<<linux_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
ablation -l irace.Rdata --src 1 --plot plot-ablation.pdf --plot-type rank,boxplot
\end{lstlisting}
%@

You may specify the IDs of the source and target configurations.  By default, the
source is taken as the first configuration evaluated by \irace and the target
as the best overall configuration found. Use the function \code{plotAblation} to visualize the ablation results
(Fig.~\ref{fig:testAb}).

<<testAb, fig.pos="htb!", fig.align="center", out.width="0.75\\textwidth", fig.cap="Example of plot generated by \\code{plotAblation()}.", prompt=FALSE, eval=TRUE, echo=FALSE>>=
logfile <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork=TRUE)
plotAblation(logfile)
@

The function returns a list containing the following elements:
%
\begin{description}
   \item \code{configurations}: A dataframe of configurations tested during ablation.
   \item \code{instances}: The instances used for the ablation.
   \item \code{scenario}: Scenario options provided by the user.
   \item \code{trajectory}: Best configuration IDs at each step of the ablation.
   \item \code{best}: Best overall configuration found.
\end{description}

We also provide a command-line executable (\path{ablation.exe} in Windows) that allows you to perform ablation without launching \aR. It is installed in the same location as the \irace command-line executable and has the following options:
<<ablation_cmdline, prompt=FALSE, eval=TRUE,echo=FALSE, comment="">>=
ablation_cmdline("--help")
@

\subsection{Post-selection race}\label{sec:postselection}

At the end of the configuration process, \irace will automatically run a
post-selection race \citep{YuaStuMonLauBir13} to try to consume any remaining
budget. This post-selection race will not generate new configurations. It will
only do additional runs on configurations already evaluated in the current run
of \irace.  First, it will try to evaluate the best configurations found in new
instances not seen yet. If this is not possible due to the budget available,
then it will evaluate configurations that have been evaluated on fewer
instances than the best configuration found. Currently, this automatic
post-selection is not supported when the budget is measured in runtime
(\parameter{maxTime}. The automatic post-selection can be disabled using the
option \parameter{postselection}.

It is also possible to manually run a post-selection race after completing a run of \irace using the R function \iracefun{psRace}.

<<postsel, prompt=FALSE, eval=FALSE>>=
# Execute all elite configurations in the iterations
psRace("irace.Rdata", max_experiments = 0.5, iteration_elites=TRUE)
# Execute a set of configurations IDs providing budget
psRace("irace.Rdata", conf_ids = c(34, 87, 102, 172, 293), max_experiments = 500)
@

The execution of the post-selection race adds an element (\code{psrace_log}) to
the \code{iraceResults} object saved in the \irace log file
(\parameter{logFile}). The post-selection log consists of a list with the
following elements:
\begin{description}
\item\code{configurations}: Configurations used in the post-selection race.
\item\code{instances}: Data frame with the instances and seeds used in the post-selection race.
\item \code{max_experiments}: Configuration budget assigned to the post-selection race.
\item \code{experiments}: Matrix of results generated by the post-selection
  race, in the same format as the matrix
  \code{iraceResults$experiments}. Column names are the configuration IDs and
  row names are the instance IDs.
\item \code{elites}: Best configuration(s) found in the post-selection race.
\end{description}


\subsection[Parameter importance analysis using PyImp]{Parameter importance analysis using \PyImp}

The \PyImp\footnote{\url{https://github.com/automl/ParameterImportance}} tool developed by the AutoML group\footnote{\url{https://www.automl.org/}}
supports various parameter importance analysis methods using surrogate models.
Given a performance dataset of an algorithm configuration scenario, a Random Forest is built to predict performance of all algorithm configurations.
Parameter importance analyses are then applied on the prediction model.
The model serves as a surrogate for the original target algorithm, so that the algorithm does not need to be executed during the analyses.
Three analysis methods are supported, namely fANOVA~\citep{HutHooLey2014icml} (functional analysis of variance), forward selection~\cite{HutHooLey2013lion},
and ablation analysis with surrogates~\cite{BieLinEggFraFawHoo2017}.
Note that the \irace package directly supports ablation (without surrogate models) analysis with and without racing (\autoref{sec:ablation}).
Although ablation analysis without surrogates may be more time-consuming, results of the surrogate version may be less accurate than the non-surrogate one.

The \code{.Rdata} dataset generated by \irace can be used as input for \PyImp.
The package \code{irace2pyimp}\footnote{\url{https://github.com/ndangtt}} is able to convert an \code{irace.Rdata} file into the input format required by \PyImp. The conversion can be  accessed either through the \aR console (function \code{irace2pyimp}), or via command line executable \code{irace2pyimp} (or \code{irace2pyimp.exe} in Windows) provided by the package. You can find the location of the executable by running this command in R: \Sexpr{'system.file(file.path("bin", "irace2pyimp"), package="irace2pyimp", mustWork=TRUE)'} or \Sexpr{'system.file(file.path("bin", "irace2pyimp.exe"), package="irace2pyimp", mustWork=TRUE)'} in Windows.

To see the usage of the executable, please run: \code{irace2pyimp --help}.
For more information on the \aR function \code{irace2pyimp}, type in the \aR console: \code{?irace2pyimp}.

Given as input an \code{irace.Rdata} file, the script will generate the following output files:
\begin{itemize}
  \item \code{params.pcs}: a text file containing the parameter space definition.
  \item \code{runhistory.json}: a JSON file containing the list of algorithm configurations evaluated during the tuning and the performance data obtained.
  \item \code{traj_aclib2.json}: a JSON file containing the best configurations after each iteration of \irace. The last configuration will be used as the target configuration in ablation analysis.
  \item \code{scenario.txt}: a text file containing the definition of the tuning scenario.
  \item \code{instances.txt}: a text file containing the list of instances.
  \item \code{features.csv}: a .csv file containing instance features. If no instance features are provided, the index of each instance will be used as a feature.
\end{itemize}

\PyImp can then be called using the files listed above as input. Several
examples on how to use the script and call \PyImp can be found at
\code{system.file("/examples/", package="irace2pyimp")}.

%\emph{\PyImp installation note:} At the time when this document is written (December 2019), the latest \PyImp release version (1.0.6) has some bugs.
%Please check on \url{https://pypi.org/project/PyImp/#history} to see if \PyImp version > 1.0.6 has been released.
%If it has, you can easily install \PyImp using \code{pip install pyimp}.
%Otherwise, please install \PyImp from source (\url{https://github.com/automl/ParameterImportance}), branch \code{development}. Installation instructions can be found at \url{https://automl.github.io/ParameterImportance/installation.html}.
%%
%% irace options
%%

\section{List of command-line and scenario options} \label{sec:irace options}

Most \irace options can be specified in the command line using a flag or in the
\irace scenario file using the option name (or setting their value in the
\code{scenario} list passed to the various \aR functions exported by the
package). This section describes the various \irace options that can be
specified by the user in this way.

\begin{xwarningbox}
  Relative filesystem paths (e.g., \path{../scenario/}) given in the
  command-line are relative to the current working directory (the directory at
  which \irace is invoked). However, paths given in
  the scenario file are relative to the directory containing
  the scenario file. See also Table~\ref{tab:paths}.
\end{xwarningbox}

<<child-demo, child='section/irace-options.Rnw'>>=
@

%%
%%
%%
%% FAQ
%%
%%
%%


\section{FAQ (Frequently Asked Questions)}

\subsection{Is \irace minimizing or maximizing the output of my algorithm?}

By default, \irace considers that the value returned by \parameter{targetRunner} (or by
\parameter{targetEvaluator}, if used) should be \underline{\textbf{minimized}}. In case of a maximization
problem, one can simply multiply the value by -1 before returning it to
irace. This is done, for example, when maximizing the hypervolume (see the last
lines in \IRACEHOME{/examples/hypervolume/target-evaluator}).


\subsection{Are experiments with \irace reproducible?}

Short answer: Yes, under some conditions.\\
Long answer: According to the terminology described by \citet{LopBraPaq2021telo}, we define \emph{repeatability} as ``\emph{exactly repeating the original experiment, generating precisely the same results}''. Following this definition, a run of \irace is repeatable under the following conditions:
\begin{itemize}
\item Same version of \irace.
\item Same version of \aR (different versions of \aR may change the behavior of functions used by \irace).
\item The behavior of \parameter{targetRunner} is deterministic or exactly reproducible for the same instance, parameter configuration and random seed. Make sure that \parameter{targetRunner} uses the seed provided by \irace to initialize all random number generators used. If the result of \parameter{targetRunner} depends on CPU-time, wall-clock time or system load in any way, then \parameter{targetRunner} is not reproducible and neither will be \irace.
\item Same random seed (\parameter{seed}) given to \irace.
\item Same scenario options (\autoref{sec:irace options}). Although some options should not affect reproducibility (e.g., \parameter{debugLevel}), maintaining a list of such options will be a huge effort, thus the safest assumption is that any change may break reproducibility.
\item Same parameter space (\autoref{sec:target parameters}), including types, domains, conditions and forbidden configurations. The order of the parameters may also affect reproducibility (the name of the parameters should not) because it affects the order in which random numbers are used.
\item Same training instances provided and in the same order (\autoref{sec:training}). Even if the instances are sampled randomly (\parameter{sampleInstances}), a different initial order will produce a different sample even with the same random seed.
\item Same initial configurations (\autoref{sec:initial}), if any.
\end{itemize}


\subsection[Is it possible to configure a MATLAB algorithm with irace?]
{Is it possible to configure a \MATLAB algorithm with \irace ?}

Definitely. There are three main ways to achieve this:
\begin{enumerate}
\item Edit the \parameter{targetRunner} script to call \MATLAB in a non-interactive
    way. See the \MATLAB documentation, or the following links.\footnote{See \href{https://stackoverflow.com/questions/1518072/suppress-start-message-of-matlab}{suppress-start-message-of-matlab (stackoverflow.com)} and \href{https://stackoverflow.com/questions/4611195/how-to-call-matlab-from-command-line-and-print-to-stdout-before-exiting}{how-to-call-matlab-from-command-line-and-print-to-stdout-before-exiting (stackoverflow.com)}} %
    You would need to pass the parameter received by \parameter{targetRunner} to your \MATLAB script.\footnote{See \href{https://www.mathworks.com/matlabcentral/answers/97204-how-can-i-pass-input-parameters-when-running-matlab-in-batch-mode-in-windows}{how-can-i-pass-input-parameters-when-running-matlab-in-batch-mode-in-windows (mathworks.com)} and \href{https://stackoverflow.com/questions/3335505/how-can-i-pass-command-line-arguments-to-a-standalone-matlab-executable-running}{how-can-i-pass-command-line-arguments-to-a-standalone-matlab-executable-running (mathworks.com)}}
    %
    There is a minimal example in \IRACEHOME{/examples/matlab/}, where you can find an example \path{target-runner} script calling the \path{matlab} executable.

\item Call \MATLAB code directly from \aR using the
    \pkg{matlabr} package (\url{https://cran.r-project.org/package=matlabr}). This
    is a better option if you are experienced in \aR. Define \parameter{targetRunner} as
    an \aR function instead of a path to a script. The function should
    call your \MATLAB code with appropriate parameters. There is a minimal example in \IRACEHOME{/examples/matlabr/}, where you can find the \parameter{targetRunner} function in \path{scenario.txt} and the \MATLAB code in \path{Main.m}.


  \item Another possibility is calling \MATLAB directly from a different programming language and write \parameter{targetRunner} in that programming language,  for example, in \proglang{Python} (see examples in \IRACEHOME{/examples/target-runner-python/}).\footnote{See \href{https://www.mathworks.com/help/matlab/matlab_external/call-matlab-functions-from-python.html}{call-matlab-functions-from-python (mathworks.com)} and \href{https://www.mathworks.com/help/matlab/matlab_external/call-user-script-and-function-from-python.html}{call-user-script-and-function-from-python (mathworks.com)}}
\end{enumerate}

\subsection{My program works perfectly on its own, but not when running under \irace. Is irace broken?}\label{faq:valgrind}

Every time this was reported, it was a difficult-to-reproduce bug, i.e., a \href{https://en.wikipedia.org/wiki/Heisenbug}{Heisenbug}, in
the program (target algorithm), not in \irace.
%
To detect such bugs, we recommend that you use, within
\parameter{targetRunner}, a memory debugger (e.g.,
\href{https://valgrind.org/}{\code{valgrind}}) to run your program. For example,
if your program is executed by \parameter{targetRunner} as:
  %
\begin{lstlisting}[style=BashInputStyle]
${EXE} ${FIXED_PARAMS} -i ${INSTANCE} ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}
%
then replace that line with:
%
\begin{lstlisting}[style=BashInputStyle]
valgrind --error-exitcode=1 ${EXE} ${FIXED_PARAMS} -i ${INSTANCE} \
  ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}

If there are bugs in your program, they will appear in
\code{\${STDERR}}, thus do not delete those files. Memory debuggers will
significantly slowdown your code, so use them only as a means to find what is
wrong with your target algorithm. Once you have fixed the bugs, you should
remove the use of \code{valgrind}.


\subsection{\irace seems to run forever without any progress, is this a bug?}\label{faq:infloop}

Every time this problem was reported, the issue was in the target algorithm and
not in \irace. Some ideas for debugging this problem:
\begin{itemize}
\item Check that the target algorithm is really not running nor paused nor
  sleeping nor waiting for input-output.
\item Use \parameter{debugLevel}\code{=3} to see how  \irace calls \code{target-runner}, run the same command outside \irace and verify that it terminates.
\item Add some output to your algorithm that reports at the very end the runtime and exit code. Verify that this output is printed when \irace calls your algorithm.
\item In \code{target-runner}, print something to a log file \emph{after} calling your target algorithm. Verify that this output appears in the log file when \irace is running.

\item Set a maximum timeout when calling your target algorithm from \code{target-runner} (see FAQ~\ref{faq:timeout}).

\end{itemize}

\subsection{My program may be buggy and run into an infinite loop. Is it possible to set a maximum timeout?}\label{faq:timeout}

We are not aware of any way to achieve this using \aR. However, in
GNU/Linux, it is easy to implement by using the \code{timeout} command\footnote{\url{https://man7.org/linux/man-pages/man1/timeout.1.html}} in
\code{targetRunner} when invoking your program.

\subsection[When using the mpi option, irace stops with an error message indicating that a
function is not defined. How to fix this?]%
{When using the \parameter{mpi} option, \irace stops with an error message indicating that a
  function is not defined. How to fix this?}\label{sec:mpi-error}

\pkg{Rmpi} does not work the same way when called from within a package and when called from a script or
interactively. When \irace creates the slave nodes, the slaves will load a copy of \irace automatically.
If the slave nodes are on different machines, they must have \irace installed. If \irace is not installed
system-wide, \aR needs to be able to find \irace on the slave nodes. This is usually done by setting \code{R_LIBS},
\code{.libPaths()} or by loading \irace using \code{library()} or \code{require()} with the argument ``\code{lib.loc}''.
The settings on the master are not applied to the slave nodes automatically, thus the slave
nodes may need their own settings. After spawning the slaves, it is too late to modify
those settings, thus modifying the shell variable \code{R_LIBS} seems the only valid
way to tell the slaves where to find \irace.

If the path is set correctly and the problem persists, please check these instructions:
%
\begin{enumerate}[leftmargin=*,widest=3]
\item Test that \irace and \pkg{Rmpi} work. Run \irace on a single machine (submit node), without calling \code{qsub}, \code{mpirun} or a
  similar wrapper around \irace or \aR.
\item Test loading \irace on the slave nodes. However,
  jobs submitted by \code{qsub}/\code{mpirun} may load \aR packages using a different mechanism from the way it happens if you log directly into the node (e.g., with \code{ssh}). Thus, you need to write a little \aR program such as:
  <<faq3, eval=FALSE>>=
  library(Rmpi)
  mpi.spawn.Rslaves(nslaves = 10)
  paths <- mpi.applyLB(1:10, function(x) {
    library(irace); return(path.package("irace")) })
  print(paths)
  @
  Submit this program to the cluster like you would submit \irace (using \code{qsub}, \code{mpirun} or whatever program is used to submit jobs to the cluster).

      \item In the script \code{bin/parallel-irace-mpi}, the function \code{irace_main()} creates an MPI job for our cluster.
      You may need to speak with the admin of your cluster and ask them how to best submit a job for MPI. There may be
      some particular settings that you need. \pkg{Rmpi} normally creates  log files; but \irace suppresses those files unless $\texttt{debugLevel} > 0$.
     \end{enumerate}

     Please contact us (\autoref{sec:contact}) if you have further problems.

\subsection[Error: 4 arguments passed to {.Internal(nchar)} which requires 3]{Error: 4 arguments passed to \code{.Internal(nchar)} which requires 3}

This is a bug in \aR 3.2.0 on Windows. The solution is to update your version of
\aR.

\subsection[Warning: In read.table(filename, header = TRUE, colClasses = "character", : incomplete final line found by \ldots]{Warning: \code{In read.table(filename,}  \code{header = TRUE,} \code{colClasses =} \code{"character",} \code{: incomplete final line found by} \ldots}

This is a warning given by \aR when the last line of an input file does not
finish with the newline character. The warning is harmless and can be
ignored. If you want to suppress it, just open the file and press the
\code{ENTER} key at the end of the last line of the file to end the final line with a newline.


\subsection{How are relative filesystem paths interpreted by \irace?}\label{faq:relpaths}

The answer depends on where the path appears. Relative paths may appear as the
argument of command-line options, as the value of options given in the scenario
file, or within various scripts, functions or instance files. Table
\ref{tab:paths} summarizes how paths are translated from relative to absolute.

\begin{table}[h!]
  \centering
  \caption{Translation of relative to absolute filesystem paths.}
  \label{tab:paths}
  \begin{tabular}{rl}
    \toprule
\bf    Relative path appears as \ldots  & \bf\ldots is relative to \ldots \\
    \midrule
    a string within \parameter{trainInstancesFile} & \parameter{trainInstancesDir}\\
    a string within \parameter{testInstancesFile} & \parameter{testInstancesDir}\\
    code within \parameter{targetRunner} or \parameter{targetEvaluator} & \parameter{execDir}\\
    the value of \parameter{logFile} or \code{--log-file} & \parameter{execDir}\\
    the value of other options in the scenario file  &  the directory containing the scenario file\\
    the value of other command-line options & invocation (working) directory of \irace\\
\bottomrule
  \end{tabular}
\end{table}

\subsection{My parameter space is small enough that \irace could generate all possible configurations; however, \irace generates repeated configurations and/or does not generate some of them. Is this a bug?}\label{faq:allconfs}

Typically, \irace is applied to parameter spaces that are much larger than what
can be explored within the budget given. Thus, \irace does not try to detect
whether all possible configurations can be evaluated for the given budget and
it does not waste computation time to check for repeated configurations. Thus,
if the parameter space is actually very small, the initial random sampling
performed by \irace may generate repeated configurations and/or never generate
some configurations, which is not ideal. If you still want to use
(non-iterated) racing, the recommended approach is to provide all
configurations explicitly to \irace (\autoref{sec:initial}) and execute a
single race (\parameter{nbIterations}\code{=1}) with exactly the number of
configurations provided (e.g., \parameter{nbConfigurations}\code{=240}). A
future version of \irace may automatically detect this case and switch to
non-iterated racing without having to set additional options. Future versions
may also implement computationally cheap checks for repeated
configurations.\footnote{If you are interested in implementing this, please
  contact us!}

\subsection[On Windows and using target-runner.py  (a Python file), I
  get the error target-runner.py is not executable]{On Windows and using \code{target-runner.py}  (a \proglang{Python} file), I
  get the error ``\code{target-runner.py is not executable}''}\label{faq:py-not-exe}

The issue is that \code{.py} files are not executable on their own and you need \code{python.exe} to read the \code{.py} file and execute it. Linux knows how to do this if  the first line of the file is ``\code{\#!/usr/bin/python}'', however, Windows doesn't know how to do it. In Windows you have 2 options:
\begin{itemize}
\item     Create a \code{target-runner.bat} file that contains a line similar to (see \path{templates/windows/target-runner.bat}):
\begin{lstlisting}[style=BashInputStyle]
C:\path\to\python.exe C:\path\to\target-runner.py %instance% %seed% \
  %candidate_parameters% 1>%stdout% 2>%stderr%
\end{lstlisting}

\item Or convert \code{target-runner.py} into an \code{.exe} file, for example, using \code{auto-py-to-exe}\footnote{\url{https://pypi.org/project/auto-py-to-exe/}}, so that you do not need a \code{.bat} file.
\end{itemize}

\subsection[Error in socketConnection(\ldots) : can not open the connection]{Error in \code{socketConnection("localhost", port = port, server = TRUE,} \code{lock = TRUE,} : can not open the connection}

This error may arise if you activate the \parameter{parallel} option of \irace and your \parameter{targetRunner} or \parameter{targetEvaluator} tries to setup a parallel cluster or execute code in parallel in a way that interacts badly with the parallel mechanism in \aR. In this case, you need to either investigate yourself if there is a way for the two parallel mechanisms to co-exist or, if that is not possible, disable parallelism in \irace or in your code. Note that packages or software used by your \parameter{targetRunner} may have a parallel mechanism enabled by default and unknown to you. This is definitely NOT a bug in \irace.

\subsection[irace does not print the call to the target-runner with --debug-level 2 when using the --parallel option]{\irace does not print the call to the \parameter{targetRunner} with \parameter{debugLevel}\code{=2} when using the \parameter{parallel}}

This is a limitation of  Windows or Rstudio. Running without \parameter{parallel} should work. Unfortunately, we cannot fix this limitation in \irace. If you need to understand how \irace calls \parameter{targetRunner} when running in parallel, you can implement a logging mechanism able to handle parallelism directly inside the \parameter{targetRunner}.

\subsection{Why does the column \texttt{Bound} in the output sometimes shows \texttt{NA}?}

This should only happen when the symbol in the first column is `\texttt{.}', which means that all alive configurations are elite and have already been evaluated on this instance, so nothing is evaluated nor discarded.

%%
%%
%%
%% Contact info
%%
%%
%%


%FIXME: complete this section
%\section{Known problems}

\section{Resources and contact information} \label{sec:contact}

More information about the package can be found on the \irace webpage:
\begin{center} \url{https://mlopez-ibanez.github.io/irace/} \end{center}

For questions and suggestions please contact the development team through
the \irace package Google group:
\begin{center}
  \url{https://groups.google.com/d/forum/irace-package}
\end{center}
or by sending an email to:
\begin{center}
  \href{mailto:irace-package@googlegroups.com}{irace-package@googlegroups.com}
\end{center}


\section{Acknowledgements}

We would like to thank all the people that directly or indirectly have
collaborated in the development and improvement of \irace:
%
\newlist{acklist}{itemize*}{1}
\setlist[acklist]{label={}, before=\unskip{}, afterlabel=\unskip{ }, itemjoin={{,}}, itemjoin*={, and }, after={{.}}}
\begin{acklist}
\item Prasanna Balaprakash
\item Zhi (Eric) Yuan
\item Franco Mascia
\item Alberto Franzin
\item Anthony Antoun
\item Esteban Diaz Leiva %MATLAB example
\item Federico Caselli % bug reports with fixes
\item Pablo Valledor Pellicer % testing of --parallel under Windows
\item André de Souza Andrade
\item Nguyen Dang (\texttt{nttd@st-andrews.ac.uk}) % implemented \code{irace2pyimp}
\end{acklist}

\newcommand{\doi}[1]{doi:\hspace{.16667em plus .08333em}\discretionary{}{}{}\href{https://doi.org/#1}{\urlstyle{rm}\nolinkurl{#1}}}
\bibliographystyle{abbrvnat}
\ifthenelse {\boolean{Release}}{%
\bibliography{irace-package}%
}{%
\bibliography{optbib/abbrev,optbib/journals,optbib/authors,optbib/articles,optbib/biblio,optbib/crossref}%
}

\newpage

\begin{appendices}

\section{Installing R} \label{sec:installation}

This section gives a quick \aR installation guide that will work in
most cases. The official instructions are available at
\url{https://cran.r-project.org/doc/manuals/r-release/R-admin.html}

\subsection{GNU/Linux}
You should install \aR from your package manager. On a Debian/Ubuntu system
it will be something like:
\begin{lstlisting}[style=BashInputStyle]
sudo apt-get install r-base
\end{lstlisting}

Once \aR is installed, you can launch \aR from the Terminal and from the \aR prompt
install the \irace package (see \autoref{sec:irace install}).


\subsection{OS X}

You can install \aR directly from a CRAN mirror.\footnote{\url{https://cran.r-project.org/bin/macosx/}} %
Alternatively, if you use
homebrew, you can just brew the \aR formula from the science tap (unfortunately
it does not come already bottled so you need to have Xcode\footnote{Xcode download
webpage: \url{https://developer.apple.com/xcode/download/}} installed to compile it):
%<<R_OS_install,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
brew tap homebrew/science
brew install r
\end{lstlisting}
%@

Once \aR is installed, you can launch \aR from the Terminal (or from your Applications),
and from the \aR prompt install the \irace package (see \autoref{sec:irace install}).

\subsection{Windows}

You can install \aR from a CRAN mirror.\footnote{\url{https://cran.r-project.org/bin/windows/}} %
We recommend that you
install \aR on a filesystem path without spaces, special characters or long
names, such as \path{C:\R}. Once \aR is installed, you can launch the \aR
console and install the \irace package from it (see \autoref{sec:irace
  install}).


\section{targetRunner troubleshooting checklist} \label{sec:check list}

If the \parameter{targetRunner} script fails to return the output expected by
\irace, it can be sometimes difficult to diagnose where the problem lies. The
more descriptive errors provided by your script, the easier it will be to debug
it. If \parameter{targetRunner} enters an infinite loop, irace will wait
indefinitely (see FAQ in \autoref{faq:timeout}).  If you are using temporary
files to redirect the output of your algorithm, check that these files are
properly created. We recommend to follow the structure of the example file
(\code{target-runner}) provided in \IRACEHOME{/templates}. The following
error examples are based on that example file.

In case of failure of \parameter{targetRunner}, \irace will print an error on
its output describing which execution of \parameter{targetRunner} failed.
%
Follow this checklist to detect where the problem is:

\begin{enumerate}[leftmargin=*,widest=9]

\item Make sure that your \code{targetRunner} script or program is at the specified location. If you see this error:
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' does not exist
\end{CodeInput}
%
it means that \irace cannot find the \code{target-runner} file. Check that the file is at
the path specified by the error.

\item Make sure that your \code{targetRunner} script is an executable file and the user running \irace has permission to execute it. The following errors:
%
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' is a directory,
not a file
\end{CodeInput}
or
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' is not executable
\end{CodeInput}
%
mean that your \code{targetRunner} is not an executable file. In the first
case, the script is a folder and therefore there must be a problem with the
name of the script. In the second case, you must make the file executable,
which in GNU/Linux can be done by:
%
\begin{lstlisting}[style=BashInputStyle]
chmod +x ./tuning/target-runner
\end{lstlisting}

\item If your \code{targetRunner} script calls another program, make sure it is at  the location described in the script
(variable \code{EXE} in the examples and templates). A typical output for such an error is:
%
\begin{CodeInput}
Error: == irace == running command ''./tuning/target-runner' 1 8 676651103
./tuning/Instances/1000-16.tsp --ras --localsearch 2 --alpha 4.03 --beta 1.89
--rho  0.02 --ants 37 --nnls 48 --dlb 0 --rasranks 15 2>\&1' had status 1
== irace == The call to target_runner_default was:
./tuning/target-runner 1 8 676651103 ./tuning/Instances/1000-16.tsp --ras
--localsearch 2 --alpha 4.03 --beta 1.89 --rho  0.02 --ants 37 --nnls 48
--dlb 0 --rasranks 15
== irace == The output was:
Tue May  3 19:00:37 UTC 2016: error: ./bin/acotsp: not found or not executable
(pwd: ./tuning/acotsp-arena)
\end{CodeInput}
%
You may test your script by copying the command line shown in the error and
executing \code{target-runner} directly on the execution directory (\parameter{execDir}). In this case, the command line is:
%
\begin{lstlisting}[style=BashInputStyle]
./tuning/target-runner 1 8 676651103 ./tuning/Instances/1000-16.tsp  --ras \
 --localsearch 2 --alpha 4.03 --beta 1.89 --rho  0.02 --ants 37 --nnls 48 \
 --dlb 0 --rasranks 15
\end{lstlisting}

This executes the \code{targetRunner} script as \irace does. The output of this script must be only one number.

\item If your \code{targetRunner} file is an executable script in Python, \aR, Perl, Bash or some other interpreted programming language, make sure that the interpreter specified in the first line of the file exists at the correct location. For example, if the first line of \code{target-runner.py} is:
%
\begin{CodeInput}
#|/usr/bin/python
\end{CodeInput}
%
Then make sure that \path{/usr/bin/python} exists and it is executable. Otherwise, you will get an error such as:
%
\begin{CodeInput}
Error: == irace == error in running command
\end{CodeInput}

\item Check that your \code{targetRunner} script is actually returning one
  number as output. For example:
%
\begin{CodeInput}
Error: == irace == The output of './tuning/target-runner 1 25 365157769
 ./tuning/Instances/1000-31.tsp --ras --localsearch 1 --alpha 0.26 --beta
 6.95 --rho 0.69 --ants 56 --nnls 10 --dlb 0 --rasranks 7' is not numeric!
== irace == The output was:
Solution: 24479793
\end{CodeInput}
% FIXME: Duplicated above.
%% For testing your script, copy the command-line of \code{target-runner} and execute it directly on the execution directory (\parameter{execDir}):
%% %
%% \begin{lstlisting}[style=BashInputStyle]
%% ./tuning/target-runner 1 25 365157769 ./tuning/Instances/1000-31.tsp  --ras \
%%  --localsearch 1 --alpha 0.26 --beta 6.95 --rho  0.69 --ants 56 \
%%  --nnls 10 --dlb 0 --rasranks 7
%% \end{lstlisting}
%% This executes the \code{targetRunner} script as \irace does.
  In the example above, the output of \code{target-runner} is ``\code{Solution:
    24479793}'', which is not a number. If \code{target-runner} is parsing the
  output of the target algorithm, you need to verify that the code only parses
  the solution cost value.

\item Check that your \code{targetRunner} script is creating the output files
  for your algorithm.  If you see an error as:
%
\begin{CodeInput}
== irace == The output was: Tue May  3 19:41:40 UTC 2016:
error: c1-9.stdout: No such file or directory
\end{CodeInput}
The output file of the execution of your algorithm has not been
created (check permissions) or has been deleted before the result can be read.


\item Other errors can produce the following output:
%
\begin{CodeInput}
== irace == The output was: Tue May  3 19:49:06 UTC 2016:
error: c1-23.stdout: Output is not a number
\end{CodeInput}
%
This might be because your \code{targetRunner} script is not executing your
algorithm correctly. To further investigate this issue, comment out the line
that eliminates the temporary files that saves the output of your
algorithm. Similar to this one
%
\begin{lstlisting}[style=BashInputStyle]
rm -f "${STDOUT}" "${STDERR}"
\end{lstlisting}
%
Execute directly the \code{targetRunner} command-line that is provided in the
error message, look in your execution directory for the files that are
created. Check the \code{.stderr} file for errors and the \code{.stdout} file
to see the output that your algorithm produces.

\item Some command within \code{targetRunner} may not be working correctly. In
  that case, you must debug the commands individually exactly as \irace executes them. In
  order to find where the problem is, print the commands to a log file before
  executing them. For example:
  %
\begin{lstlisting}[style=BashInputStyle]
echo "$EXE ${FIXED_PARAMS} -i $INSTANCE ${CONFIG_PARAMS}" >> ${STDERR}.log
$EXE ${FIXED_PARAMS} -i $INSTANCE ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}
%
  then look at the \code{${STDERR}.log} file corresponding to the
  \code{targetRunner} call that failed and execute/debug the last command
  there.

\item If the language of your operating system, the \code{target-runner} or the
  target algorithm is not English, \irace may not be able to recognize the
  numbers generated by \code{target-runner}. We recommend that you run \irace,
  the \code{target-runner} and the target algorithm under an English locale (or
  make sure that their languages and number format are compatible).

\item It is possible that
  \href{https://en.wikipedia.org/wiki/Heisenbug}{transient bugs} in the target
  algorithm are only visible when running within \irace, and
  all commands within \code{targetRunner} appear to work fine when executed directly in the
  command-line outside \irace. See FAQ in \autoref{faq:valgrind}) for suggestions on how to detect such bugs.


\item If your \code{targetRunner} script works when running irace with \parameter{parallel}\code{=0} but it fails when using higher number of cores, this may be due to any number of reasons:
  \begin{itemize}
  \item If you submit jobs through a queuing system, the running environment
    when using the queuing system may not be the same as when you launch \irace
    yourself. The queuing system may also send the job to different machines
    depending on the number of CPUs requested. One way to test this is to
    submit the failing execution of \code{targetRunner} to the queuing system,
    and specifically to any problematic machine.

  \item When using MPI, some calls to \code{targetRunner} may run on different
    computers than the one running the master \irace process. See FAQ in
    \autoref{sec:mpi-error}.

  \item Does \code{targetRunner} read or create intermediate files? These files
    may cause a race condition when two calls to \code{targetRunner} happen at
    the same time. You have to make sure that parallel runs of
    \code{targetRunner} do not interfere with each other's files.

  \item Maybe these files consume too much memory or fill the filesystem when
    there are simultaneous \code{targetRunner} calls? Moreover, queuing systems
    have stricter limits for computing nodes than for the submit/host node.

  \item Does the machine or the queuing system impose any limits on number of
    processes or CPU/memory/filesystem usage per job? Such limits may only
    trigger when more than one process is executed in parallel, killing the
    \code{targetRunner} process before it has a chance to print anything
    useful. In that case, \irace may not detect the the program finished
    unexpectedly, only that the expected output was not printed.
  \end{itemize}

\end{enumerate}

\section{targetEvaluator troubleshooting checklist} \label{sec:evaluator check list}

Even if \parameter{targetRunner} appears to work, the use of \parameter{targetEvaluator} may lead to other problems. The same checklist of \code{targetRunner} can be followed here. In addition, we list here other potential problems unique to \code{targetEvaluator}:

\begin{enumerate}
\item If \parameter{targetEvaluator} fails only in the second or later
  iteration, this may because output files or data generated by a previous call
  to \parameter{targetRunner} are missing. Elite configurations are never
  re-executed on the same instance and seed pair, that is, \irace will call
  only once \parameter{targetRunner} for each pair of configuration ID and
  instance ID. However, \parameter{targetEvaluator} is always re-executed,
  which takes into account any updated information (normalization bounds,
  reference sets/points, best-known values, etc.). Thus, any files or data
  generated by \parameter{targetRunner} for a given configuration must remain
  available to \parameter{targetEvaluator} as long as that configuration is
  alive. The list of alive configurations is passed
  to \parameter{targetEvaluator}, which may decide then which data to keep or
  remove.
\end{enumerate}

\section{Glossary}

\begin{description}
\item[Parameter tuning:] Process of searching good settings for the parameters
  of an algorithm under a particular tuning scenario (instances, execution
  time, etc.).
\item[Scenario:] Settings that define an instance of the tuning problem. These
  settings include the algorithm to be tuned (target), budget for the execution
  of the target algorithm (execution time, evaluations, iterations, etc.), set
  of problem instances and all the information that is required to perform the
  tuning.
\item[Target algorithm:] Algorithm whose parameters will be tuned.
\item[Target parameter:] Parameter of the target algorithm that will be tuned.
\item[\irace option:] Configurable option of \irace.
\item[Elite configurations:] Best configurations found so far by \irace. New
  configurations for the next iteration of \irace are sampled from the
  probabilistic models associated to the elite configurations.  All elite
  configurations are also included in the next iteration.
\item[\hypertarget{irace_home}{\path{$IRACE_HOME}:}] The filesystem path where \irace is
  installed. You can find this information by opening an \aR console and
  executing:

<<R_irace_home2, prompt=FALSE, eval=FALSE>>=
system.file(package = "irace")
@

\end{description}

\section{NEWS}

\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\footnotesize,
 %
 frame=lines,  % top and bottom rule only
 framesep=1em, % separation between frame and text
 rulecolor=\color{Gray},
 %
 label=\fbox{\color{Black}NEWS},
 labelposition=topline,
 %
 % commandchars=\|\(\), % escape character and argument delimiters for
 %                      % commands within the verbatim
 % commentchar=*        % comment character
}

%% NEWS is latin1 but this file is utf8
\begingroup\inputencoding{utf8}%
% FIXME: This doesn't break long lines, find another alternative.
\VerbatimInput{../NEWS.md}%
\endgroup
\end{appendices}

\end{document}

%%% Local Variables:
%%% TeX-master: "irace-package.Rnw"
%%% End:

% LocalWords: iteratively

% !Rnw weave = knitr
%%% DO NOT EDIT the .tex file directly since it is generated from the .Rnw
%%% sources.
%\VignetteEngine{knitr::knitr}
%\VignetteIndexEntry{irace package: User Guide}
%\VignetteDepends{knitr}
%\VignetteCompiler{knitr}
\synctex=1
\RequirePackage{xparse}
\RequirePackage[dvipsnames]{xcolor}
\documentclass[a4paper,english]{article}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{lmodern}
\usepackage[a4paper]{geometry} % It saves some pages
\usepackage[english]{babel}
\usepackage{ifthen}
\newboolean{Release}
\setboolean{Release}{true}
\usepackage{calc}
\usepackage{afterpage}
\usepackage{algorithm,algorithmic}
\usepackage{booktabs}
\usepackage{tabularx}
\usepackage{xspace}
\usepackage{amsmath,amssymb}
\usepackage{relsize}
\usepackage{fancyvrb}
\usepackage{underscore}
\usepackage{microtype}
% \texttt{test -- test} keeps the "--" as "--" (and does not convert it to an en dash)
\DisableLigatures{encoding = T1, family = tt* }
\usepackage[hyphens]{url}
\usepackage{hyperref}
\usepackage[numbers]{natbib}
\usepackage[nottoc]{tocbibind}
%% For autoref
\hypersetup{
    colorlinks,
    linkcolor={red!50!black},
    citecolor={blue!50!black},
    urlcolor={blue!70!black}
}
\addto\extrasenglish{%
\def\sectionautorefname{Section}
\let\subsectionautorefname\sectionautorefname
\let\subsubsectionautorefname\sectionautorefname
}
\usepackage[titletoc, title]{appendix}
% Fix use with \autoref
\newcommand*{\Appendixautorefname}{Appendix}
\usepackage{tocloft}
\setlength{\cftsubsecnumwidth}{3em}% Set length of number width in ToC for \subsection
\usepackage[inline]{enumitem}
\setlist[enumerate]{leftmargin=*,widest=00}
\setlist[itemize]{leftmargin=1.5em}

%% FIXME: listing is very limited, we should use 'minted'
\usepackage{listings}
\lstdefinestyle{BashInputStyle}{
  language=bash,%
  basicstyle=\ttfamily,%
  numbers=none,%
  frame=tb,%
  rulecolor=\color{lightgray},
%  framesep=1ex,
  framexleftmargin=1ex,
  columns=fullflexible,%
  backgroundcolor=\color{yellow!05},%
  linewidth=\linewidth,%
 % xleftmargin=1\linewidth,%
  identifierstyle=\color{darkgray},%
  keywordstyle=\color{darkgray},%
  keywordstyle={[2]\color{Cyan}},%
  keywordstyle={[3]\color{olive}},%
  stringstyle=\color{MidnightBlue},%
  commentstyle=\color{RedOrange},%
  morestring=[b]',%
  showstringspaces=false
}

\DefineVerbatimEnvironment{Code}{Verbatim}{}
\DefineVerbatimEnvironment{CodeInput}{Verbatim}{fontshape=rm}
\DefineVerbatimEnvironment{CodeOutput}{Verbatim}{}
\newenvironment{CodeChunk}{}{}


\newcommand{\IRACEHOME}[1]{\hyperlink{irace_home}{\path{$IRACE_HOME}}\path{#1}}

\providecommand{\keywords}[1]{\textbf{\textit{Index terms---}} #1}

% Simple font selection is not good enough.  For example, |\texttt{--}|
% gives `\texttt{--}', i.e., an endash in typewriter font.  Hence, we
% need to turn off ligatures, which currently only happens for commands
% |\code| and |\samp| and the ones derived from them.  Hyphenation is
% another issue; it should really be turned off inside |\samp|.  And
% most importantly, \LaTeX{} special characters are a nightmare.  E.g.,
% one needs |\~{}| to produce a tilde in a file name marked by |\file|.
% Perhaps a few years ago, most users would have agreed that this may be
% unfortunate but should not be changed to ensure consistency.  But with
% the advent of the WWW and the need for getting `|~|' and `|#|' into
% URLs, commands which only treat the escape and grouping characters
% specially have gained acceptance
\makeatletter
\DeclareRobustCommand\code{\bgroup\@makeother\_\@makeother\~\@makeother\$\@noligs\@codex}
\def\@codex#1{\texorpdfstring%
{{\text{\normalfont\ttfamily\hyphenchar\font=-1 #1}}}%
{#1}\egroup}
\makeatother

\let\proglang=\textsf
\newcommand{\pkg}[1]{{\fontseries{b}\selectfont #1}}
\newcommand{\aR}{\proglang{R}\xspace}
\newcommand{\MATLAB}{\proglang{MATLAB}\xspace}
\newcommand{\eg}{e.g.,\xspace}
\newcommand{\SoftwarePackage}{\pkg}
\newcommand{\ACOTSP}{\SoftwarePackage{ACOTSP}\xspace}

%% How to use this command:
% Parameter with one short switch: \defparameter[short]{paramName}{long}{default}
% Parameter without short switch:  \defparameter{paramName}{long}{default}
% Parameter without switch:  \defparameter{paramName}{}{default}
\newcommand{\defparameter}[4][]{%
\item[\code{#2}]\hypertarget{opt:#2}{} ~~ %
\ifthenelse{\equal{#3}{}}{}{%
  \emph{flag:} %
  \ifthenelse{\equal{#1}{}}{}{%
    \code{-#1}~~~\emph{or}~~~}%
  \code{--#3} ~~ }%
\emph{default:}~\texttt{#4} \\
}
\newcommand{\parameter}[1]{\hyperlink{opt:#1}{\code{#1}}}
\newcommand{\iracefun}[1]{\href{https://mlopez-ibanez.github.io/irace/reference/#1.html}{\code{#1()}}}
%\usepackage{showlabels}
%\showlabels{hypertarget}

\newcommand{\irace}{\pkg{irace}\xspace}
\newcommand{\Irace}{\pkg{Irace}\xspace}
\newcommand{\race}{\pkg{race}\xspace}
\newcommand{\FRACE}{\text{F-Race}\xspace}
\newcommand{\IFRACE}{\text{I/F-Race}\xspace}
\newcommand{\PyImp}{\pkg{PyImp}\xspace}
\newcommand{\iraceversion}{\Sexpr{packageVersion("irace")}}

\newcommand{\Niter}{\ensuremath{N^\text{iter}}\xspace}
\newcommand{\Nparam}{\ensuremath{{N^\text{param}}}\xspace}
\newcommand{\iter}{\ensuremath{j}\xspace}
\newcommand{\Budget}{\ensuremath{B}\xspace}
\newcommand{\Budgetj}{\ensuremath{\Budget_{\iter}}\xspace}
\newcommand{\Bused}{\ensuremath{\Budget_\text{used}}\xspace}
\newcommand{\Ncand}[1][]{\ensuremath{N_{#1}}\xspace}
\newcommand{\Mui}{\ensuremath{\mu_{\iter}}\xspace}
\newcommand{\Nmin}{\ensuremath{N^\text{min}}\xspace}
\newcommand{\Nsurv}{\ensuremath{N^\text{surv}}\xspace}
\newcommand{\Nelite}{\ensuremath{N^\text{elite}}\xspace}
\newcommand{\Nnew}{\ensuremath{N^\text{new}}\xspace}

\newcommand{\bmax}{\ensuremath{b^\text{max}}\xspace}
\newcommand{\bmin}{\ensuremath{b^\text{min}}\xspace}
\newcommand{\Celite}{\ensuremath{\Theta^\text{elite}}\xspace}

\ifthenelse {\boolean{Release}}{%
\newcommand{\MANUEL}[1]{}
\newcommand{\LESLIE}[1]{}
\newcommand{\THOMAS}[1]{}
}{%
\newcommand{\MANUEL}[1]{{\footnotesize\noindent\textbf{\color{red}[~MANUEL: #1~]}}}
\newcommand{\LESLIE}[1]{\footnote{\noindent\textbf{[ LESLIE: #1 ]}}}
\newcommand{\THOMAS}[1]{\footnote{\noindent\textbf{[ THOMAS: #1 ]}}}
}
\newcommand{\hide}[1]{}

\usepackage{tcolorbox}
\newcommand{\infoicon}{%
\parbox[c]{0.75cm}{\includegraphics[keepaspectratio=true,width=0.75cm]{light-bulb-icon}}%
\hspace{1em}}
\newcommand{\warningicon}{%
\parbox[c]{0.75cm}{\includegraphics[keepaspectratio=true,width=0.75cm]{Warning-icon}}%
\hspace{1em}}

\definecolor{LightGray}{RGB}{193,193,193}
\definecolor{LightYellow}{RGB}{253,247,172}

\newlength\macroiconwidth
\newenvironment{xwarningbox}{%
\setlength{\fboxrule}{3.0\fboxrule}%
\setlength{\fboxsep}{0\fboxsep}%
\begin{tcolorbox}[colback=LightYellow,colframe=LightGray,boxrule=\fboxrule,boxsep=\fboxsep]%
\infoicon%
\settowidth{\macroiconwidth}{\infoicon}%
\begin{minipage}[c]{\columnwidth - \macroiconwidth - 2.0\fboxrule - 2.0\fboxsep}
\raggedright\footnotesize
%
}{%
\end{minipage}
\end{tcolorbox}
%
}

% Workaround for broken knitr: https://github.com/yihui/knitr-examples/blob/master/036-latex-if.tex
\providecommand{\hldef}[1]{\textcolor[rgb]{0.345,0.345,0.345}{#1}}%
\providecommand{\hlsng}[1]{\textcolor[rgb]{0.192,0.494,0.8}{#1}}%

<<setup, include=FALSE>>=
library(knitr)
knit_hooks$set(inline = function(x) {
  if (is.numeric(x)) return(knitr:::format_sci(x, 'latex'))
  highr::hi_latex(x)
})
@

\begin{document}

<<include=FALSE>>=
library(knitr)
@

\author{Manuel L\'opez-Ib\'a\~nez, Leslie P\'erez C\'aceres, J\'er\'emie Dubois-Lacoste,\\
  Thomas St\"utzle and Mauro Birattari
\\IRIDIA, CoDE, Universit\'e Libre de Bruxelles, Brussels, Belgium}

\title{The \irace Package: User Guide}
\date{Version \iraceversion, \today}
%\keywords{automatic
%  algorithm configuration, racing, parameter tuning, \aR}

\maketitle

\tableofcontents

%Load files needed for examples
<<exampleload,eval=TRUE,include=FALSE>>=
library("irace")
load("examples.Rdata") # loads "experiment" and "output"
iraceResults <- irace::read_logfile(system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE))
log_ablation_file <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork = TRUE)
load(log_ablation_file)
options(width = 70)
@
\newpage


%%
%%
%%
%% General info
%%
%%
%%
\section{General information}
\MANUEL{Some things could be taken from the intro of the irace paper and reformulated.}
\MANUEL{It would be good to mention that not only opt algorithms can be configured with irace, we say this in the paper.}

\subsection{Background}
\MANUEL{I would add a paragraph defining what is irace (a bit longer than the abstract above) and references to the literature so people can find more info. The first reference should be the irace TR.} \LESLIE{Here i guess we should say why tune an algorithm is a good idea, and why using irace is a better one.}
The \irace package implements an \emph{iterated racing} procedure,
which is an extension of Iterated F-race (\IFRACE)~\cite{BirYuaBal2010:emaoa}.  The main use of
\irace is the automatic configuration of optimization and decision algorithms,
that is, finding the most appropriate settings of an algorithm given a
set of instances of a problem. However, it may also be useful for
configuring other types of algorithms when performance depends on
the used parameter settings. It builds upon the \pkg{race} package
by Birattari and it is implemented in \aR. The \irace
package is available from CRAN:
%
\begin{center}
\url{https://cran.r-project.org/package=irace}
\end{center}
%
More information about \irace is available at
\url{https://mlopez-ibanez.github.io/irace}.

\subsection{Version}
The current version of the \irace package is \iraceversion. Previous
versions of the package can also be found in the \href{https://cran.r-project.org/package=irace}{CRAN website}.

The algorithm underlying the current version of \irace and its motivation are
described by \citet{LopDubPerStuBir2016irace}. The \textbf{adaptive capping
  mechanism} available from version $3.0$ is described by
\citet{PerLopHooStu2017:lion}. Details of the implementation before version 2.0
can be found in a previous technical report~\cite{LopDubStu2011irace}.
%
\begin{xwarningbox}
  Versions of \irace before 2.0 are not compatible with the file formats
  detailed in this document.
\end{xwarningbox}


\subsection{License}

The \irace package is Copyright \copyright{} \the\year\ and distributed under the GNU General Public
License version 3.0 (\url{http://www.gnu.org/licenses/gpl-3.0.en.html}).
The \irace package is free software (software libre): You can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

The \irace package is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.

Please be aware that the fact that this program is released as Free Software
does not excuse you from scientific propriety, which obligates you to give
appropriate credit! If you write a scientific paper describing research that
made substantive use of this program, it is your obligation as a scientist to
(a) mention the fashion in which this software was used in the Methods section;
(b) mention the algorithm in the References section. The appropriate citation
is:

\begin{itemize}[leftmargin=3em]
\item[] Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres,
  Thomas Stützle, and Mauro Birattari. The \irace package: Iterated Racing for
  Automatic Algorithm Configuration. \emph{Operations Research Perspectives}, 3:43--58, 2016.
  doi:~\href{http://dx.doi.org/10.1016/j.orp.2016.09.002}{10.1016/j.orp.2016.09.002}

\end{itemize}

\section{Before starting}
\MANUEL{I think this could be a bit more detailed by defining what is a parameter, a configuration, an instance, etc. but ok for now.}

The \irace package provides an automatic configuration tool for
 tuning optimization algorithms, that is, automatically finding good configurations for the parameters values
of a (target) algorithm saving the effort that normally requires manual
tuning.

\begin{figure}[t]
  \centering
  \includegraphics[width=0.6\textwidth]{irace-scheme}
  \caption{Scheme of \irace flow of information.}
\label{fig:irace-scheme}
\end{figure}

Figure~\ref{fig:irace-scheme} gives a general scheme of how \irace works.
\Irace receives as input a \emph{parameter space definition} corresponding to the
parameters of the target algorithm that will be tuned, a set of \emph{instances}
for which the parameters must be tuned for and a set of options for \irace that define the \emph{configuration scenario}. Then, \irace searches in the parameter search space for good performing algorithm
configurations by executing the target algorithm on different instances and
with different parameter configurations. A \parameter{targetRunner} must be provided to execute the target algorithm with a
specific parameter configuration ($\theta$) and instance ($i$). The \parameter{targetRunner} function (or program) acts
as an interface between the execution of the target algorithm and \irace: It
receives the instance and configuration as arguments and must return the
evaluation of the execution of the target algorithm.

The following user guide contains guidelines for installing \irace, defining
configuration scenarios, and using \irace to automatically configure your
algorithms.

%%
%%
%%
%% Installation
%%
%%
%%

\section{Installation}

\subsection{System requirements}
\begin{itemize}
\item \aR ($\text{version} \geq 3.2.0$) is required for running irace, but you don't
  need to know the \aR language to use it.  \aR is freely available and you can
  download it from the \aR project website
  (\url{https://www.r-project.org}). See \autoref{sec:installation} for a
  quick installation guide of \aR.

\item For GNU/Linux and OS X, the command-line executable
  \code{parallel-irace} requires GNU Bash.  Individual examples may require additional software.
\end{itemize}

\subsection{\irace installation} \label{sec:irace install}

The \irace package can be installed automatically within \aR or
by manual download and installation. We advise to use the automatic
installation unless particular circumstances do not allow it. The instructions
to install \irace with the two mentioned methods are the following:

\subsubsection[Install automatically within R]{Install automatically within \aR{}}

Execute the following line in the \aR console to install the package:
<<R_irace_install, prompt=FALSE, eval=FALSE>>=
install.packages("irace")
@
Select a mirror close to your location, and test the installation in the \aR console with:
<<R_irace_launch,eval=FALSE, prompt=FALSE>>=
library("irace")
q() # To exit R
@

Alternatively, within the \aR graphical interface, you may use
the \code{Packages and data->Package installer} menu on OS X or the \code{Packages} menu on Windows.

\subsubsection{Manual download and installation}

From the \irace package CRAN website
(\url{https://cran.r-project.org/package=irace}), download one of the three
versions available depending on your operating system:
\begin{itemize}
\item \code{irace_\iraceversion.tar.gz} (Unix/BSD/GNU/Linux)
\item \code{irace_\iraceversion.tgz} (OS X)
\item \code{irace_\iraceversion.zip} (Windows)
\end{itemize}

To install the package on GNU/Linux and OS X, you must execute the following
command at the shell (replace \code{<package>} with the path to the downloaded file, either \code{irace_\iraceversion.tar.gz} or \code{irace_\iraceversion.zip}):
%
\begin{lstlisting}[style=BashInputStyle]
R CMD INSTALL <package>
\end{lstlisting}

To install the package on Windows, open \aR and execute the following line on
the \aR console (replace \code{<package>} with the path to the downloaded file \code{irace_\iraceversion.zip}):
%\LESLIE{Check that this actually works on internet says that  this: \code{Rscript -e "install.packages('foo.zip', repos = NULL)"} also works}
%
<<install_win1,eval=FALSE, prompt=FALSE>>=
install.packages("<package>", repos = NULL)
@

If the previous installation instructions fail because of insufficient
permissions and you do not have sufficient admin rights to install \irace
system-wide, then you need to force a local installation.

\subsubsection{Local installation}

Let's assume you wish to install \irace on a path denoted by
\code{<R_LIBS_USER>}, which is a filesystem path for which you have sufficient
rights. This directory \textbf{must} exist before attempting the
installation. Moreover, you must provide to \aR the path to this
library when loading the package. However, the latter can be avoided by adding
the path to the system variable \code{R_LIBS} or to the \aR internal variable
\code{.libPaths}, as we will see below.\footnote{%
   On Windows, see also
   \url{https://cran.r-project.org/bin/windows/base/rw-FAQ.html\#I-don_0027t-have-permission-to-write-to-the-R_002d3_002e3_002e1_005clibrary-directory}.}


On GNU/Linux or OS X, execute the following commands to install the
package on a local directory:

\begin{lstlisting}[style=BashInputStyle]
export R_LIBS_USER="<R_LIBS_USER>"
# Create R_LIBS_USER if it doesn't exist
mkdir $R_LIBS_USER
# Replace <package> with the path to the downloaded file.
R CMD INSTALL --library=$R_LIBS_USER <package>
# Tell R where to find R_LIBS_USER
export R_LIBS=${R_LIBS_USER}:${R_LIBS}
\end{lstlisting}

On Windows, you can install the package on a local directory by executing the
following lines in the \aR console:

<<install_win2,eval=FALSE, prompt=FALSE>>=
# Replace <package> with the path to the downloaded file.
# Replace <R_LIBS_USER> with the path used for installation.
install.packages("<package>", repos = NULL, lib = "<R_LIBS_USER>")
# Tell R where to find R_LIBS_USER.
# This must be executed for every new session.
.libPaths(c("<R_LIBS_USER>", .libPaths()))
@

\subsubsection{Testing the installation and invoking irace}

Once \irace has been installed, load the package and test that the installation
was successful by opening an \aR console and executing:

<<R_irace_test1, prompt=FALSE, eval=FALSE>>=
# Load the package
library("irace")
# Obtain the installation path
system.file(package = "irace")
@

The last command must print out the filesystem path  where \irace is
installed.  In the remainder of this guide, the variable
\code{\$IRACE_HOME} is used to denote this path. When executing any provided
command that includes the \code{\$IRACE_HOME} variable do not forget to replace
this variable with the installation path of \irace.

On GNU/Linux or OS X, you can let the operating system know where to find \irace
by defining the \code{\$IRACE_HOME} variable and adding it to the system
\code{PATH}. Append the following commands to \path{~/.bash_profile}, \path{~/.bashrc} or
\path{~/.profile}:
%
%<<linux_irace_path1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
# Replace <IRACE_HOME> with the irace installation path
export IRACE_HOME=<IRACE_HOME>
export PATH=${IRACE_HOME}/bin/:$PATH
# Tell R where to find R_LIBS_USER
# Use the following line only if local installation was forced
export R_LIBS=${R_LIBS_USER}:${R_LIBS}
\end{lstlisting}
%@

Then, open a new terminal and launch \irace as follows:

%<<linux_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace --help
\end{lstlisting}
%@

On Windows, you need to add both \aR and the installation path of \irace to the
environment variable \code{PATH}. To edit the \code{PATH}, search for
``Environment variables'' in the control panel, edit \code{PATH} and add
a string similar to \path{C:\R_PATH\bin;C:\IRACE_HOME\bin\x64\} where \code{R_PATH}
is the installation path of \aR and \code{IRACE_HOME} is the installation path
of \irace. If \irace was installed locally, you also need to edit the
environment variable \code{R_LIBS} to add \code{R_LIBS_USER}. Then, open a new
terminal (run program \code{cmd.exe}) and launch \irace as:
%
%<<win_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace.exe --help
\end{lstlisting}
%@

Alternatively, you may directly invoke \irace from within the \aR
console by executing:
<<windows_irace_help ,eval=FALSE, prompt=FALSE>>=
library("irace")
irace_cmdline("--help")
@


\section{Running irace}\label{sec:execution}

Before performing the tuning of your algorithm, it is necessary to define a
tuning scenario that will give \irace all the necessary information to optimize the
parameters of the algorithm. The tuning scenario is composed of the following elements:

\begin{enumerate}
\item Target algorithm parameter description (see \autoref{sec:target parameters}).
\item Target algorithm runner (see \autoref{sec:runner}).
\item Training instances list (see \autoref{sec:training})
\item \irace options (see \autoref{sec:irace options}).
\item \textit{Optional:} Initial configurations (see \autoref{sec:initial}).
\item \textit{Optional:} Target algorithm evaluator (see \autoref{sec:evaluator}).
\end{enumerate}

These scenario elements can be provided as plain text files
or as \aR objects. This user guide provides examples of both types,
but we advise the use of plain text files, which we consider the simpler
option.

For a step-by-step guide to create the scenario elements for your target algorithm
continue to \autoref{sec:step}. For an example execution of \irace using the
\ACOTSP scenario go to \autoref{sec:example}.



\subsection{Step-by-step setup guide}\label{sec:step}
This section provides a guide to setup a basic execution of \irace. The
template files provided in the package (\IRACEHOME{/templates}) will be
used as basis for creating your new scenario. Please follow carefully the
indications provided in each step and in the template files used; if you have
doubts check the the sections that describe each option in detail.

\begin{enumerate}[leftmargin=*]
\item Create a directory (\eg~\path{./tuning/}) for the scenario setup. This directory will contain
all the files that describe the scenario. On GNU/Linux or OS X, you can do this as follows:
%<<dir1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./tuning
cd ./tuning
\end{lstlisting}
%@

\item Initialize the tuning directory with template config files. On GNU/Linux or OS X, you can do this as follows:
%<<dir1,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
irace --init
\end{lstlisting}
%@

\item Define the target algorithm parameters to be tuned by following the
instructions in \code{parameters.txt}.  Available parameter types and other
guidelines can be found in \autoref{sec:target parameters}.


\item \textit{Optional}: Define the initial parameter configuration(s) of your
algorithm, which  allows you to provide good starting configurations (if you know
some) for the tuning. Follow the instructions in \code{configurations.txt} and set \parameter{configurationsFile}\code{="configurations.txt"} in \code{scenario.txt}. More
information in \autoref{sec:initial}. If you do not need to
define initial configurations remove this file from the directory.

\item Place the instances you would like to use for the tuning of your
  algorithm in the folder \path{./tuning/Instances/}. In addition, you can
  create a file (\eg~\code{instances-list.txt}) that specifies which instances
  from that directory should be run and which instance-specific parameters to
  use. To use such an instance file, set the appropriate option in
  \code{scenario.txt}, e.g., \parameter{trainInstancesFile} \code{= "instances-list.txt"}.
  See \autoref{sec:training} for guidelines.

\item Uncomment and assign in \code{scenario.txt} only the options for which you
need a value different from the default. Some common options that you might want to adjust are:
\begin{description}
\item[\parameter{execDir}] (\code{--exec-dir}): the directory in which \irace will execute the
target algorithm; the default value is the current directory.
%\item[\parameter{logFile}] (\code{--log-file}): a file the name of the results \aR data file that produces \irace.
\item[\parameter{maxExperiments}] (\code{--max-experiments}): the maximum number of executions
of the target algorithm that \irace will perform.
\item[\parameter{maxTime}] (\code{--max-time}): maximum total execution time in seconds for the executions of \code{targetRunner}. In this case, \code{targetRunner} must return two values: cost and time. Note that you must provide either \parameter{maxTime} or \parameter{maxExperiments}.
\item[\parameter{trainInstancesDir}] (\code{--train-instances-dir}): set to \path{./Instances} if you put the training instances in that folder as instructed above.
  \end{description}
For setting the tuning budget, see \autoref{sec:budget}. For more information on \irace options and their default values, see \autoref{sec:irace options}.

\item Modify the \code{target-runner} script to run your algorithm. This script
  must execute your algorithm with the parameters and instance specified by
  \irace and return the evaluation of the execution and \textit{optionally} the execution
  time (\code{cost [time]}). When the \parameter{maxTime} option is used, returning \code{time} is mandatory.
  The \code{target-runner} template is written in \proglang{GNU Bash}
  scripting language, which can be executed easily in GNU/Linux and OS X
  systems. However, you may use any other programming language. We provide examples written in \proglang{Python},  \proglang{MATLAB} and other languages in \IRACEHOME{/examples/}. An example using Julia is available at \url{https://github.com/sbomsdorf/An-example-of-irace-using-Julia-code}. %
  Follow these instructions to adjust the given \code{target-runner} template
  to your algorithm:
  \begin{enumerate}
  \item Set the \code{EXE} variable with the path to the executable of the target algorithm.
  \item Set the \code{FIXED_PARAMS} if you need extra arguments in the
    execution line of your algorithm. An example could be the time that your
    algorithm is required to run
    (\code{FIXED_PARAMS}\hspace{0pt}\code{=}\hspace{0pt}\code{"--time 60"}) or
    the number of evaluations required (\code{FIXED_PARAMS}\hspace{0pt}\code{=}\hspace{0pt}\code{"--evaluations 10000"}).

  \item The line provided in the template executes the executable described in the \code{EXE} variable.
    %
    \begin{center}
      \code{\$EXE \$\{FIXED_PARAMS\} -i \$\{INSTANCE\} --seed \$\{SEED\} \$\{CONFIG_PARAMS\}}
    \end{center}
    %
    You must change this line according to the way your algorithm is executed. In this
    example, the algorithm receives the instance to solve with the flag
    \code{-i} and the seed of the random number generator with the flag \code{--seed}. The variable
    \code{CONFIG_PARAMS} adds to the command line the parameters that \irace has
    given for the execution.  You must set the command line execution as
    needed. For example, the instance might not need a flag and might need to be
    the first argument:
    \begin{center}
      \code{\$EXE \$\{INSTANCE\} \$\{FIXED_PARAMS\} --seed \$\{SEED\} \$\{CONFIG_PARAMS\}}
    \end{center}
    The output of your algorithm is saved to the file defined in the
    \code{\$STDOUT} variable, and error output is saved in the file given by
    \code{\$STDERR}. The line:
    \begin{center}
      \code{if [ -s "\${STDOUT}" ]; then}
    \end{center}
    checks if the file containing the output of your algorithm is not
    empty. The example provided in the template assumes that your algorithm
    prints in the last output line the best result found (only a number). The
    line:
    \begin{center}
      \code{COST=\$(cat \$\{STDOUT\} | grep -e '\^{}[[:space:]]*[+-]\textbackslash{}?[0-9]' | cut -f1)}
    \end{center}
    parses the output of your algorithm to obtain the result from the last line. The \code{target-runner}
    script must print \textbf{only} one number. In the template example, the result is printed with
    \code{echo "\$COST"} (assuming \parameter{maxExperiments} is used) and the generated files are deleted (you may remove that line if you wish to keep them).

    \begin{xwarningbox}
      The \code{target-runner} script must be an executable file, unless you specify \parameter{targetRunnerLauncher}.
    \end{xwarningbox}

    You can test the target runner from the \aR console by checking the scenario as
    explained earlier in \autoref{sec:execution}.

If you have problems related to the \code{target-runner} script when executing
\irace, see \autoref{sec:check list} for a check list to help diagnose common
problems. For more information about the \parameter{targetRunner}, please see
\autoref{sec:runner},
        \end{enumerate}

      \item \textit{Optional}: Modify the \code{target-evaluator} file. This is rarely needed and the \code{target-runner} template does not use it. \autoref{sec:evaluator} explains when a \parameter{targetEvaluator} is needed and how to define it.

      \item The \irace executable provides an option (\parameter{--check}) to
        check that the scenario is correctly defined. We recommend to perform a
        check every time you create a new scenario. When performing the check,
        \irace will verify that the scenario and parameter definitions are
        correct and will test the execution of the target algorithm. To check
        your scenario execute the following commands:

\begin{itemize}
\item From the command-line (on Windows, execute \code{irace.bat}):

%<<irace_check,engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
# $IRACE_HOME is the installation directory of irace.
$IRACE_HOME/bin/irace --scenario scenario.txt --check
\end{lstlisting}
%@

\item Or from the \aR console:

<<irace_R_check, eval=FALSE, prompt=FALSE>>=
library("irace")
scenario <- readScenario(filename = "scenario.txt",
                         scenario = defaultScenario())
checkIraceScenario(scenario = scenario)
@

\end{itemize}


\item  Once all the scenario elements are prepared you can execute \irace, either using the command-line wrappers provided by the package
 or directly from the \aR console:

\begin{itemize}
\item
\textbf{From the command-line console}, call the command (on Windows, you should execute
  \code{irace.exe}):
\begin{lstlisting}[style=BashInputStyle]
cd ./tuning/
# $IRACE_HOME is the installation directory of irace
# By default, irace reads scenario.txt, you can specify a different file
# with --scenario.
$IRACE_HOME/bin/irace
\end{lstlisting}
For this example, we assume that the needed scenario files have been set
properly in the \code{scenario.txt} file using the options described in
\autoref{sec:irace options}. Most \irace options can be specified
in the command line or directly in the \code{scenario.txt} file.

\item
\textbf{From the \aR console}, evaluate:

<<irace_R_exe, eval=FALSE, prompt=FALSE>>=
library("irace")
# Go to the directory containing the scenario files
setwd("./tuning")
scenario <- readScenario(filename = "scenario.txt",
                         scenario = defaultScenario())
irace_main(scenario = scenario)
@

\end{itemize}

This will perform one run of \irace. See the output of \code{irace --help} in the command-line or \code{irace_cmdline("--help")} in \aR for quick information on
additional \irace options. For more information about \irace options, see \autoref{sec:irace options}.

\end{enumerate}

\begin{xwarningbox}
Command-line options override the same options specified in the \code{scenario.txt} file.
\end{xwarningbox}

\subsection{Setup example for ACOTSP}\label{sec:example}

The \ACOTSP tuning example can be found in the package installation in the folder \IRACEHOME{/examples/acotsp}.
 %
Other example scenarios can be found in the same folder. More
examples of tuning scenarios can be found in the Algorithm Configuration Library (AClib, \url{http://www.aclib.net/}).

In this section, we describe how to execute the \ACOTSP scenario. If you wish to start setting up
your own scenario, continue to the next section. For this example, we assume
a GNU/Linux system such as Ubuntu with a working \proglang{C} compiler such as \code{gcc}.
To execute this scenario follow these steps:

\begin{enumerate}
\item Create a directory for the tuning (\eg~\path{./tuning/}) and copy the example scenario files
located in the \code{examples} folder to the created directory:
%<<dir0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./tuning
cd ./tuning
# $IRACE_HOME is the installation directory of irace.
cp $IRACE_HOME/examples/acotsp/* ./
ls ./
# Make sure that target-runner is executable
chmod u+x target-runner
\end{lstlisting}
%@

\item Download the training instances from \url{https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/scenarios/acotsp/instances.tar.gz} to the \path{./tuning/} directory and decompress it, which creates create a folder \path{instances}:

%<<instance0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
tar -xvf instances.tar.gz
ls instances/
\end{lstlisting}
%@

If the above gives an error or does not show any files, then the files were not extract correctly. Maybe the \path{instances.tar.gz} file did not download correctly or maybe it is not in the correct place. It should be within the folder \path{tuning}.correctly.

\item Download the \ACOTSP software from \url{https://github.com/MLopez-Ibanez/ACOTSPQAP/archive/refs/heads/master.zip} to the \path{./tuning/}
directory and compile the \path{acotsp} executable using \code{make}.
%<<acotsp0,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
unzip master.zip
make -C ACOTSPQAP-master acotsp
./ACOTSPQAP-master/acotsp --version
\end{lstlisting}
%@

If the above gives an error, then the \path{acotsp} executable failed to compile for some reason. Maybe you are missing the C compiler or some files did not extract correctly.

\item Create a directory for executing the experiments and execute \irace:

%<<runexample,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
mkdir ./acotsp-arena/
# $IRACE_HOME is the installation directory of irace.
$IRACE_HOME/bin/irace
\end{lstlisting}
%@

Or you can also execute \irace from the \aR console using:

<<runexample2,prompt=FALSE,eval=FALSE>>=
library("irace")
setwd("./tuning/")
irace_cmdline()
@

\end{enumerate}

The most usual sources of error when running the above commands are:
\begin{itemize}
\item The \irace package is not correctly installed. Please make sure that installing \irace did not give any errors.
\item The location of the files is not correct. Please make sure that you have:
  \begin{itemize}
  \item The folder  \path{tuning} and that it contains the files \path{scenario.txt}, \path{parameters-acotsp.txt}, \path{target-runner}, and the folders \path{instances}, \path{ACOTSPQAP-master} and \path{acotsp-arena}.
  \item The folder \path{instances} should contain the TSP instance files (\path{*.tsp}).
  \item The folder \path{ACOTSPQAP-master} should contain the executable \path{acotsp}. You should be able to invoke \code{./ACOTSPQAP-master/acotsp --version} without an error.
  \item The folder \path{acotsp-arena} should be empty.
  \end{itemize}
\end{itemize}
%%
%%
%%
%% Scenario settings
%%
%%
%%



\section{Defining a configuration scenario}\label{sec:scenario}
\subsection{Target algorithm parameters} \label{sec:target parameters}

The parameters of the target algorithm are defined by a parameter file as
described in \autoref{sec:parameters file}. Optionally, when executing
\irace from the \aR console, the parameters can be specified directly as an \aR
object (see \autoref{sec:parameters object}). For defining your parameters
follow the guidelines provided in the following sections.

\subsubsection{Parameter types}

Each target parameter has an associated type that defines its domain and the
way \irace handles them internally. Understanding the nature of the domains of
the target parameters is important to select appropriate types. The four basic
types supported by \irace are the following:

\begin{itemize}
\item \textit{Real} parameters are numerical parameters that can take
  floating-point values within a given range. The range is
  specified as an interval `\code{(<lower bound>,<upper
    bound>)}'. This interval is closed, that is, the parameter value
  may eventually be one of the bounds. The possible values are rounded
  to a number of \emph{decimal places} specified by
  the global option \code{digits} (Section~\ref{sec:globaloptions}). For example, given the default
  number of digits of $4$, the values $0.12345$ and
  $0.12341$ are both rounded to $0.1234$.
   Selected real-valued parameters can be optionally sampled on
  a logarithmic scale (base $e$).

\item \textit{Integer} parameters are numerical parameters that can take only
  integer values within the given range. Their range is specified as the range
  of real parameters and they can also be optionally sampled on a
  logarithmic scale (base $e$).

\item \textit{Categorical} parameters are defined by a set of possible values
  specified as `\code{(<value 1>,} \code{...,} \code{<value n>)}'. The values
  are quoted or unquoted character strings. Empty strings and strings
  containing commas or spaces must be quoted.

\item \emph{Ordinal} parameters are defined by an \emph{ordered} set
  of possible values in the same format as for categorical
  parameters. They are handled internally as integer parameters, where
  the integers correspond to the indexes of the values.

\end{itemize}

\begin{xwarningbox}
  Boolean (or logical) parameters are best encoded as categorical ones with just two values rather than integer ones with domain $(0,1)$. Some boolean parameters take an explicit value (0/1 or true/false) such as:
\begin{CodeInput}
dlb "--dlb " c (0, 1)
\end{CodeInput}
Others are switches whose presence activates the parameter:
\begin{CodeInput}
dlb "" c ("", "--dlb")
\end{CodeInput}
\end{xwarningbox}




\subsubsection{Parameter domains}

For each target parameter, an interval or a set of values must be defined
according to its type, as described above. There is no limit for the size of
the set or the length of the interval, but keep in mind that larger domains
could increase the difficulty of the tuning task. Choose always values that you
consider relevant for the tuning. In case of doubt, we recommend to choose larger
intervals, as occasionally best parameter settings may be not intuitive a priori.
All intervals are considered as closed intervals.

It is possible to define parameters that will have always the same value.
Such ``\emph{fixed}'' parameters will not be tuned but their values are used
when executing the target algorithm and they are affected by constraints
defined on them. All fixed parameters must be defined as categorical
parameters and have a domain of one element.

\subsubsection{Parameter dependent domains}\label{sec:paramdependant}

Domains that are dependent on the values of other parameters can be specified
only for numerical parameters (both integer and real). To do so, the dependent
domain must be expressed in function of another parameter, which must be a
numerical parameter. The expression that defines a dependency must be written
between quotes: \code{(value, "expression")} or \code{("expression", value)} or
\code{("expression", "expression")}.

The expressions can only use the following operators and \aR functions: \code{+},
\code{-}, \code{*}, \code{/}, \code{\%\%}, \code{min}, \code{max},
\code{round}, \code{floor}, \code{ceiling}, \code{trunc}. If you need to use an
operator or function not listed here, please contact us.

\begin{xwarningbox}
The user must ensure that the defined domain is valid at all times since
\irace currently is not able to detect possible invalid domains based on the expressions
provided.
\end{xwarningbox}

If you have a parameter \code{p2} that is just a transformation of another
\code{p1}, then instead of using a dependent domain (left-hand side of the
following example), it will be better to create a dummy parameter that controls
the transformation (right-hand side) and do the transformation within \code{target-runner}. For example:
%
\begin{center}
\begin{minipage}{0.4\linewidth}
\small\centering%
\begin{CodeInput}[frame=single]
# With dependent domains
p1 "" r (0, 100)
p2 "" r ("p1", "p1 + 10")
\end{CodeInput}
\end{minipage}
\hspace{\stretch{1}} should be \hspace{\stretch{1}}
\begin{minipage}{0.4\linewidth}
\small\centering%
\begin{CodeInput}[frame=single]
# With a dummy parameter
p1    "" r (0, 100)
p2dum "" r (0, 10)
\end{CodeInput}
\end{minipage}
\end{center}
%
and \code{target-runner} will compute $\code{p2} = \code{p2dum} \cdot \code{p1}$.


\subsubsection{Conditional parameters}\label{sec:conditional}

Conditional parameters are active only when others have certain values. These
dependencies define a hierarchical relation between parameters. For example,
the target algorithm may have a parameter \code{localsearch} that takes values
\code{(sa, ts)} and another parameter \code{ts-length} that only needs to be
set if the first parameter takes precisely the value \code{ts}. Thus, parameter
\code{ts-length} is conditional on \code{localsearch == "ts"}.


\subsubsection{Forbidden parameter configurations}\label{sec:forbidden}
A line containing just \texttt{[forbidden]} ends the list of parameters and starts the list of forbidden expressions.
Each line is a logical expression (in \aR syntax) containing
parameter names as defined by the \parameter{parameterFile}
(\autoref{sec:target parameters}), values and logical operators. For a list of \aR logical operators see:
\begin{center}
 \url{https://stat.ethz.ch/R-manual/R-devel/library/base/html/Syntax.html}
\end{center}
If \irace generates a parameter configuration that makes any of the logical
expressions evaluate to \code{TRUE}, then the configuration is considered
forbidden and it is never evaluated. This is useful when some combination
of parameter values could cause the target algorithm to crash, consume
excessive CPU time or memory, or when it is known that they do no produce
satisfactory results.

\begin{xwarningbox}
Initial configuration (\autoref{sec:initial}) that are forbidden will be discarded with a warning.
\end{xwarningbox}

If the forbidden constraints provided are too strict, \irace may produce the following error:
\begin{CodeInput}
  irace tried 100 times to sample from the model a configuration not forbidden
  without success, perhaps your constraints are too strict?
\end{CodeInput}
%
In that case, it may be a good idea to reformulate  the forbidden constraints as conditional parameters (\autoref{sec:conditional}), parameter-dependent domains (\autoref{sec:paramdependant}),  repairing the configurations (\autoref{sec:repairconf}) or post-processing within the target-algorithm (\autoref{sec:complex_domains}).

\subsubsection{Global options}\label{sec:globaloptions}

A line containing just \texttt{[global]} starts the definition of global options. The only global option currently implemented is \code{digits}, which controls the number of decimal digits for real valued parameters. Its default value is 4.

\subsubsection{Parameter file format}\label{sec:parameters file}

For simplicity, the description of the parameters space is given as a
table. Each line of the table defines a configurable parameter
%
\begin{center}
\code{<name>\ <label>\ <type>\ <domain>\ [ | <condition>\ ] }
\end{center}
where each field is defined as follows:
%
\begin{description}[left=5em,itemindent=-0.5em,align=right,itemsep=0pt]
%  \DrawEnumitemLabel
\item[\code{<name>}] The name of the parameter as an unquoted
                  alphanumeric string, e.g., `\code{ants}'.
                  %
\item[\code{<label>}]
A \emph{label} for this parameter. This is a
                  string that will be passed together with the parameter
                  to \parameter{targetRunner}. In the default \parameter{targetRunner}
                  provided with the package (\autoref{sec:runner}), this is the
                  command-line switch used to pass the value of this parameter, for
                  instance `\code{"--ants "}'.\newline
                  The value of the parameter is concatenated \emph{without
                  separator} to the label when
                invoking \parameter{targetRunner}, thus \emph{any whitespace in the label is significant}. Following the same example, when parameter \code{ants} takes value \code{5}, the default targetRunner will pass the parameter as \code{"--ants 5"}.

  \item[\code{{<type>}}] The type of the parameter, either
                    \textit{integer}, \textit{real}, \textit{ordinal} or
                    \textit{categorical}, given as a single letter: `\code{i}',
                    `\code{r}', `\code{o}' or `\code{c}'. Numerical parameters
                    can be sampled using a natural logarithmic scale with '\code{i,log}' and
                    '\code{r,log}' (without spaces) for integer and real
                    parameters, respectively.
                    %
  \item[\code{{<domain>}}] The range or set of values of the parameter delimited by
                     parentheses, \eg~\code{(0,1)} or \code{(a,b,c,d)}.
                     See also parameter dependent domains (\autoref{sec:paramdependant}).
                     %

  \item[\code{{<condition>}}] An optional \emph{condition} that determines whether the parameter is
                         enabled or disabled, thus making the parameter conditional. If the
                         condition evaluates to false, then no value is assigned to this
                         parameter, and neither the parameter value nor the corresponding
                         label are passed to \parameter{targetRunner}. The condition must follow the same syntax as those for specifying forbidden configurations (see below), that is, it must be a
                         valid \aR logical expression\footnote{For a list of \aR operators
                         see: \url{https://stat.ethz.ch/R-manual/R-devel/library/base/html/Syntax.html}}.
                         The condition may contain the name
                         of other parameters as long as the dependency graph does not
                         contain any cycle. Otherwise, \irace will detect the cycle and
                         stop with an error.
\end{description}
\begin{xwarningbox}
Categorical and ordinal parameters are always treated as strings. Given a parameter like:
\begin{CodeInput}
a "" c (0, 5, 10, 20)
\end{CodeInput}
then, a condition like \code{a > 10} will be true when \code{a} is \code{5},
because comparisons between strings are lexicographic and \code{"10"} is sorted
before \code{"5"}. As a work-around, you can convert the string to numeric in
the condition with \code{as.numeric(a)}.
\end{xwarningbox}



As an example, Figure~\ref{fig:acotsp_parameters} shows the parameters file
of the \ACOTSP scenario.

\begin{figure}[!hbt]
\small\centering%
\begin{CodeInput}
# name      switch            type values               [conditions (using R syntax)]
algorithm   "--"              c    (as,mmas,eas,ras,acs)
localsearch "--localsearch "  c    (0, 1, 2, 3)
alpha       "--alpha "        r    (0.00, 5.00)
beta        "--beta "         r    (0.00, 10.00)
rho         "--rho  "         r    (0.01, 1.00)
ants        "--ants "         i    (5, 100)
nnls        "--nnls "         i    (5, 50)              | localsearch %in% c(1, 2, 3)
q0          "--q0 "           r    (0.0, 1.0)           | algorithm == "acs"
dlb         "--dlb "          c    (0, 1)               | localsearch %in% c(1,2,3)
rasrank     "--rasranks "     i    (1, "ants")          | algorithm == "ras"
elitistants "--elitistants "  i    (1, 750)             | algorithm == "eas"

[forbidden]
## Examples of valid logical operators are:
## ==  !=  >=  <=  >  <  &  |  !  %in%
(alpha == 0.0) & (beta == 0.0)
\end{CodeInput}
\caption{Parameter file (\code{parameters.txt}) for tuning \ACOTSP.}\label{fig:acotsp_parameters}
\end{figure}

\subsubsection{Parameters R format}\label{sec:parameters object}
The target parameters are stored in an \aR~\code{list} that you can obtain from the \aR console using the following command:

<<readParameters,prompt=FALSE, eval=FALSE>>=
parameters <- readParameters(file = "parameters.txt")
@

See the help of the \code{readParameters} function (\code{?readParameters}) for more information.
The structure of the parameter list that is created is as follows:

\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{0.98\linewidth}{@{}rX}
    \code{names}        & Vector that contains the names of the parameters. \\
    \code{types}        & Vector that contains the type of each parameter 'i', 'c', 'r', 'o'. \\
    \code{switches}     & Vector that contains the labels of the parameters. e.g., switches to be used for the
    parameters on the command line. \\
    \code{domain}       & List of vectors, where each vector may contain two
    values (minimum, maximum) for real and integer parameters, or
    a set of values for categorical and ordinal parameters.\\
    \code{conditions}   & List of \aR logical expressions, with variables
   corresponding to parameter names.\\
    \code{isFixed}      & Logical vector that specifies which parameter is fixed
   and, thus, it does not need to be tuned.\\
    \code{transform}    & Vector that contains the transformation of each parameter. Currently, it
      can take values \code{``''} (no transformation, default) of \code{``log''}
      (natural logarithmic transformation).\\
    \code{nbParameters} & An integer, the total number of parameters.\\
    \code{nbFixed}      & An integer, the number of parameters with a fixed value.\\
    \code{nbVariable}   & Number of variable (i.e., to be tuned) parameters.\\
    \code{isDependent}  & Logical vector that specifies which parameter defines
                          a dependent domain.\\
    \code{forbidden}  & List of R logical expressions that cannot evaluate to \code{TRUE} for any evaluated configuration.\\
  \end{tabularx}
\end{center}

The following example shows the structure of the \code{parameters} \aR object
for the \code{algorithm}, \code{ants} and \code{q0} parameters of the \ACOTSP scenario:

<<parametersetup,eval=TRUE,include=FALSE>>=
# Setup example
parameters <- readParameters(text='
algorithm       "--"                 c		(as,mmas,eas,ras,acs)
ants            "--ants "            i	        (5, 100)
q0              "--q0 "              r          (0.0, 1.0)              | algorithm %in% c("acs")
')
@

<<parameterlist,eval=TRUE, prompt=TRUE, size='normalsize', comment="">>=
str(parameters, vec.len = 10)
@

\subsection{Target algorithm runner}
\label{sec:runner}

The evaluation of a candidate configuration on a single instance is done by means of a user-given auxiliary
program or, alternatively, a user-given \aR function. The function (or program name) is specified by the option
\parameter{targetRunner}. The \parameter{targetRunner} must return the
cost value (\eg cost of the best solution found) of the evaluation; unless computing the cost requires information from all the configurations evaluated on an instance, \eg when evaluating multi-objective algorithms with unknown normalisation bounds (see \autoref{sec:evaluator} for details).

\begin{xwarningbox}
  The objective of \irace is to minimize the cost value returned by the target
  algorithm.  If you wish to maximize, you can multiply the cost by
  \code{-1} before returning it to \irace.
\end{xwarningbox}

\subsubsection{Target runner as an executable program}

When \parameter{targetRunner} is an auxiliary executable program, it
is invoked for each candidate configuration, passing as arguments:
%
\begin{center}
\small\code{<id_configuration>\ <id_instance>\ <seed>\ <instance>\ [bound]\ <configuration>}
\end{center}
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
    \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
    \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
    \code{seed} & seed for the random number generator to be used for this evaluation, ignore the seed for deterministic algorithms;\\
    \code{instance} & string giving the instance to be used for this evaluation;\\
    \code{bound} & optional execution time bound. Only provided when the \parameter{boundMax} option is set in the scenario, see \autoref{sec:capping};\\
    \code{configuration} & the pairs parameter label-value that
    describe this candidate configuration. Typically given as command-line
    switches to be passed to the executable program.\\
  \end{tabularx}
\end{center}

The experiment list shown in \autoref{sec:runner fx}, would result in the following execution line:

%<<target_runner_ex1, engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
target-runner 1 113 734718556 /home/user/instances/tsp/2000-533.tsp \
 --eas --localsearch 0 --alpha 2.92 --beta 3.06 --rho 0.6 --ants 80
\end{lstlisting}
%@

The command line switches that describe the candidate configuration are
constructed by appending to each parameter label (switch), \emph{without
separator}, the value of the parameter, following the order given in the
parameter table. The program \parameter{targetRunner} must print a real
number, which corresponds to the cost measure of the candidate configuration
for the given instance and optionally its execution time (mandatory when \parameter{maxTime}
is used and/or when the \parameter{capping} option is enabled). The working directory of
\parameter{targetRunner} is set to the execution directory specified by the option
\parameter{execDir}. This allows the user to execute independent runs of \irace
in parallel using different values for \parameter{execDir}, without the runs
interfering with each other.

\subsubsection{Target runner as an R function}
\label{sec:runner fx}

When \parameter{targetRunner} is an \aR function, it is invoked
for each candidate configuration as:
%
<<targetRunner,prompt=FALSE, eval=FALSE>>=
targetRunner(experiment, scenario)
@
%
\noindent where \code{experiment} is a list that contains  information about
configuration and instance to execute one experiment, and \code{scenario} is the scenario
list. The structure of the \code{experiment} list is as follows:

\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
   \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
   \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
   \code{seed} & seed to be used for this evaluation;\\
   \code{instance} & string giving the instance to be used for this evaluation;\\
   \code{bound} &  optional execution time bound;\\
   \code{configuration}    & 1-row data frame with a column per parameter name;\\
  \end{tabularx}
\end{center}

The following is an example of an experiment list for the \ACOTSP scenario:

<<experimentlist,eval=TRUE,size='normalsize', prompt=TRUE,  comment="">>=
print(experiment)
@

If you need the \code{switches} of each parameter (or any other information
about the parameters), you can obtain it from \code{scenario$parameters}.

You can find an example that calls MATLAB from \aR using this approach here: \url{https://github.com/MLopez-Ibanez/irace/blob/master/inst/examples/matlab/scenario.txt}.

If \parameter{targetEvaluator} is \code{NULL}, then
the \parameter{targetRunner} function must return a list with at least one
element \code{"cost"}, the numerical value corresponding to the evaluation of
the given configuration on the given instance. A cost of \code{Inf} is accepted
and results in the immediate rejection of the configuration (see
\autoref{sec:reject}).

If the scenario option \parameter{maxTime} is non-zero or if the \parameter{capping}
option is enabled, then the list must contain at least another element \code{"time"} that
reports the execution time for this call to \code{targetRunner}.

The return list may also contain the following optional elements that are used
by \irace for reporting errors in \code{targetRunner}:
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
\code{error} & is a string used to report an error;\\
    \code{outputRaw} & is a string used to report the raw output of calls to
      an external program or function;\\
    \code{call} & is a string used to report how \code{targetRunner} called
      an external program or function;\\
  \end{tabularx}
\end{center}



\subsection{Target evaluator} \label{sec:evaluator}

Normally, \parameter{targetRunner} returns the cost of the execution of a
candidate configuration (see \autoref{sec:runner}). However, there are cases
when the cost evaluation must be delayed until all candidate configurations in
a race have been executed on a instance.

The \parameter{targetEvaluator} option defines an auxiliary program (or an \aR
function) that allows postponing the evaluations of the candidate
configurations. For each instance seen, the program \parameter{targetEvaluator} is only invoked after all
the calls to \parameter{targetRunner} for all alive candidate configurations on
the same instance have already finished.

\begin{xwarningbox}
When using \parameter{targetEvaluator}, \parameter{targetRunner} must not return the evaluation
of the configuration. If \parameter{maxTime} is used, \parameter{targetRunner} must return only
execution time.
\end{xwarningbox}

As an example, \parameter{targetEvaluator} may be used to dynamically find
normalization bounds for the output returned by an algorithm for each
individual instance. In this case, \parameter{targetRunner} will save the
output of the algorithm, then the first call to \parameter{targetEvaluator}
will examine the output produced by all calls to \parameter{targetRunner} for
the same instance, update the normalization bounds and return the normalized
output. Subsequent calls to \parameter{targetEvaluator} for the same instance
will simply return the normalized output.

A similar need arises when using quality measures for multi-objective
optimization algorithms, such as the hypervolume, which typically
require specifying reference points or sets. By
using \parameter{targetEvaluator}, it is possible to dynamically compute the
reference points or sets while \irace is running. Examples are provided at
\path{examples/hypervolume}. See also \autoref{sec:multi objective} for
more information on how to tune multi-objective algorithms.

\subsubsection{Target evaluator executable program}

When \parameter{targetEvaluator} is an auxiliary executable program, it
is invoked for each candidate with the following arguments:
%
\begin{center}
\small\code{<id_configuration>\ <id_instance>\ <seed>\ <instance>\ <num_configurations>\ <all_conf_id>}
\end{center}
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
   \code{id_configuration}  & an alphanumeric string that uniquely identifies a configuration;\\
   \code{id_instance} & an alphanumeric string that uniquely identifies an instance;\\
   \code{seed} & seed to be used for this evaluation;\\
   \code{instance} & string giving the instance to be used for this evaluation;\\
   \code{num_configurations} & number of alive candidate configurations;\\
   \code{all_conf_id}    & list of IDs of the alive configurations separated by whitespace.\\
  \end{tabularx}
\end{center}

The \parameter{targetEvaluator} executable must print a numerical value
corresponding to the cost measure of the candidate configuration on
the given instance.




\subsubsection{Target evaluator R function}

When \parameter{targetEvaluator} is an \aR function, it is invoked
for each candidate configuration as:%
%
<<targetEvaluator, eval=FALSE>>=
targetEvaluator(experiment, num_configurations, all_conf_id, scenario,
                target_runner_call)
@
%
\noindent where \code{experiment} is a list that contains information about one
experiment (see \autoref{sec:runner fx}), \code{num_configurations}
is the number of configurations alive in the race, \code{all_conf_id} is
the vector of IDs of the alive configurations,  \code{scenario} is
the scenario list and \code{target_runner_call} is the string of the
\parameter{targetRunner} execution line.

The function \parameter{targetEvaluator} must return a list with one element
\code{"cost"}, the numerical value corresponding to the cost measure of the
given configuration on the given instance.

The return list may also contain the following optional elements that are used
by \irace for reporting errors in \code{targetEvaluator}:
%
\begin{center}
\renewcommand{\arraystretch}{1.2}
  \begin{tabularx}{1\linewidth}{@{}rX}
\code{error} & is a string used to report an error;\\
    \code{outputRaw} & is a string used to report the raw output of calls to
      an external program or function;\\
    \code{call} & is a string used to report how \code{targetEvaluator} called
      an external program or function;\\
  \end{tabularx}
\end{center}

\subsection{Training instances}\label{sec:training}

The \irace options \parameter{trainInstancesDir} and \parameter{trainInstancesFile}
specify where to find the training instances.
%
If you only set the value of \parameter{trainInstancesDir} (e.g., to  \path{./Instances}),  \irace will consider all files within that directory as training instances.

Otherwise, the value of \parameter{trainInstancesFile} may specify a text
file. The format of this file is one instance per line. Within each line, elements separated by white-space will be parsed as separate arguments to be supplied to \parameter{targetRunner}. This allows defining instance-specific parameter settings. Quoted strings will be parsed as a single argument.
%
The following example shows a training instance file for the \ACOTSP scenario:

\begin{figure}[!ht]
\centering%
    \begin{minipage}{0.4\textwidth}
\begin{CodeInput}
# Example training instances file
100/100-1_100-2.tsp --time 1
100/100-1_100-3.tsp --time 2
100/100-1_100-4.tsp --time 3
\end{CodeInput}
\end{minipage}
\caption{Training instances file for tuning \ACOTSP.}\label{fig:acotsp_training}
\end{figure}

The value of \parameter{trainInstancesDir}, if set, is always prefixed to the instance
name, that is, the instances names are treated as relative to this directory.
%
For example, given the above file as \parameter{trainInstancesFile} and   \parameter{trainInstancesDir}\code{="./Instances"}, then a possible invocation of \parameter{targetRunner} would be:
%
%<<target_runner_cmd, engine='bash', eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
target-runner 1 4 5718 ./Instances/100/100-1_100-2.tsp --time 1 --alpha 2.92 ...
\end{lstlisting}
%@


Training instances do not need to be files, \irace just passes the elements of each line as arguments to \parameter{targetRunner}, thus each line may denote the name of a benchmark function or a label, plus instance-specific settings, that the target algorithm understands. Each line may even be the command-line parameters required to call an instance generator within \parameter{targetRunner}. When the instances do not represent actual files, then \parameter{trainInstancesDir} is usually set to the empty string (\code{--train-instances-dir=""}). For example,

\begin{center}
  \begin{minipage}{0.4\linewidth}
\begin{CodeInput}
# Example training instances file
rosenbrock_20 --function=12 --nvar 20
rosenbrock_30 --function=12 --nvar 30
rastrigin_20 --function=15 --nvar 20
rastrigin_30 --function=15 --nvar 30
\end{CodeInput}
\end{minipage}
\end{center}
%\caption{Example of training instances file where instances do not correspond to files.}\label{fig:nofile_training}
%\end{figure}

Optionally, when executing \irace from the \aR console, the list of instances might be provided explicitly by means of the variable \code{scenario$instances}. Thus, the previous example would be equivalent to:

<<instance1, prompt=FALSE, eval=FALSE>>=
scenario$instances <- c("rosenbrock_20 --function=12 --nvar 20",
                        "rosenbrock_40 --function=12 --nvar 30",
                        "rastrigin_20 --function=15 --nvar 20",
                        "rastrigin_40 --function=15 --nvar 30")
@

By default, \irace assumes that the target algorithm is stochastic (the value
of the option \parameter{deterministic} is $0$), thus, the same configuration
can be executed more than once on the same instance and obtain different
results.  In this case, \irace generates pairs \code{(instance, seed)} by
generating a random seed for each instance. In other words, configurations
evaluated on the same instance use the same random seed. This is a well-known
variance reduction technique called \emph{common random
  numbers}~\citep{McG1992vrt}. If all available pairs are used within a run of
\irace, new pairs are generated with different seeds, that is, a configuration
evaluated more than once per instance will use different random seeds.

If \parameter{deterministic} is set to $1$, then each instance will be used at
most once per race. This setting should only be used for target algorithms that
do not have a stochastic behavior and, therefore, executing the target
algorithm on the same instance several times with different seeds does not make
sense.
\begin{xwarningbox}
  If \parameter{deterministic} is active and the number of training instances
  provided to \irace is less than \parameter{firstTest} (default: 5), no
  statistical test will be performed on the race.
\end{xwarningbox}

Finally, \irace randomly re-orders the sequence of instances provided. This
random sampling may be disabled by using the option \parameter{sampleInstances}
(\code{--sample-instances 0}) if keeping the order provided in the instance
file is important.

\begin{xwarningbox}
  We advise to always sample instances to prevent biasing the tuning due to the
  instance order. See also \autoref{sec:het}
\end{xwarningbox}



\subsection{Initial configurations} \label{sec:initial}

The scenario option \parameter{configurationsFile} allows specifying a text
file that contains an initial set of configurations to start the execution of
\irace. If the number of initial configurations supplied in the file is less
than the number of configurations required by \irace in the first iteration,
additional configurations will be sampled uniformly at random.

The format of the configurations file is one configuration per line, and one
parameter value per column. The first line must give the parameter name
corresponding to each column (names must match those given in the parameters
file).  Each configuration must satisfy the parameter conditions (\code{NA}
should be used for those parameters that are not enabled for a given
configuration) and not be forbidden by the constraints that define forbidden
configurations (\autoref{sec:forbidden}), if any.

Figure~\ref{fig:acotsp_default} gives an example file that corresponds to the
\ACOTSP scenario.

\begin{figure}[!ht]
  \centering
  \begin{minipage}{0.7\linewidth}
\footnotesize
\begin{CodeInput}
## Initial candidate configuration for irace
algorithm localsearch alpha beta rho  ants nnls dlb q0 rasrank elitistants
as        0           1.0   1.0  0.95 10   NA   NA  0  NA      NA
\end{CodeInput}
\end{minipage}
\caption{Initial configuration file (\code{default.txt}) for tuning \ACOTSP.}\label{fig:acotsp_default}
\end{figure}

We advise to use this feature when a default configuration of the target algorithm exists or when
different sets of good parameter values are known. This will allow \irace to start the search
from those parameter values and attempt to improve their performance.



\subsection{Repairing configurations}\label{sec:repairconf}

In some problems, the parameter values require complex constraints that cannot
be implemented by constraints defined in the parameter space
(\autoref{sec:forbidden}).
%
The scenario option \parameter{repairConfiguration} can be set to a
user-defined \aR function that takes a single configuration generated by irace
and returns a ``\emph{repaired}'' configuration, thus allowing the
implementation of any rules necessary to satisfy arbitrary constraints on
parameter values.
%
The \parameter{repairConfiguration} function is called after generating a
configuration and before checking for forbidden configurations. The first
argument is a 1-row \code{data.frame} with parameter names as the column names and the second argument is the \code{parameters} list (\autoref{sec:parameters
  object}).  An example that makes all real-valued parameters
sum up to one would be:
%
<<repairEx,prompt=FALSE, eval=FALSE>>=
repairConfiguration = function(configuration, parameters)
{
  isreal <- names(which(parameters$types[colnames(configuration)] == "r"))
  # This ignores 'digits'
  c_real <- unlist(configuration[isreal])
  c_real <- c_real / sum(c_real)
  configuration[isreal] <- c_real
  return(configuration)
}
@
The following example forces three specific parameters to be in increasing order:
%
<<repairEx2,prompt=FALSE, eval=FALSE>>=
repairConfiguration = function(configuration, parameters)
{
 columns <- c("p1","p2","p3")
 # cat("Before"); print(configuration)
 configuration[columns] <- sort(unlist(configuration[columns], use.names=FALSE))
 # cat("After"); print(configuration)
 return(configuration)
}
@

The above code can be specified directly in the \parameter{scenarioFile}, by default \code{scenario.txt}.

%%
%%
%%
%% Parallel processing
%%
%%
%%

\section{Parallelization}\label{sec:parallel}
A single run of \irace can be done much faster by executing the calls
to \parameter{targetRunner} (the runs of the target algorithm) in
parallel. There are four ways to parallelize a single run of \irace:

\begin{enumerate}
\item \textbf{Parallel processes}: The option \parameter{parallel} executes
  multiple calls to \parameter{targetRunner} in parallel within a single
  computer, by means of the \pkg{parallel} \aR package. For example, adding
  \code{--parallel N} to the command line of \irace will launch in parallel up
  to $N$ calls of the target algorithm. When using this option within a
  computing cluster, \irace will be submitted as a \emph{job} in some way  that tells the cluster to   ``reserve'' $N$ CPUs (or tasks depending on the cluster) within a single cluster node (a single machine).

\item \textbf{MPI}: By enabling the option \parameter{mpi}, calls
  to \parameter{targetRunner} will be executed in parallel by using the message
  passing interface (MPI) protocol (requires the \pkg{Rmpi} \aR package). In
  this case, the option \parameter{parallel} controls the number of slave nodes
  used by \irace. For example, adding \code{--mpi 1 --parallel N} to the
  command-line will create $N$ slaves + 1 master, and execute up to $N$ calls of
  \parameter{targetRunner} in parallel.

   The user is responsible for setting up the required MPI environment.  MPI is
   commonly available in computing clusters and requires launching \irace in
   some particular way. An example script for using MPI mode in a SGE cluster
   is given at \IRACEHOME{/bin/parallel-irace-mpi}.

   By default, \irace dynamically balances the load among nodes, however, this
   may significantly increase communication overhead in some parallel
   environments and disabling \parameter{loadBalancing} may be faster.

 \item \textbf{Batch queue mode}: Some computing clusters work by submitting
   jobs to a batch queue and waiting for the jobs to finish. With the
   option \parameter{batchmode} (\code{--batchmode} \code{[sge|pbs|} \code{torque|slurm]}),
   \irace will launch in parallel as many calls of \parameter{targetRunner} as
   possible and use a
   cluster-specific method to wait for jobs to finish. In this mode, \parameter{parallel} controls how many jobs are queued and should be set to the queue limit of your cluster.  If your cluster type is
   not supported or not working as expected, please submit a pull request (\url{https://github.com/MLopez-Ibanez/irace/pulls}) adding support to your cluster type.
See the examples in
    \IRACEHOME{/examples/batchmode-cluster/}.

   \begin{xwarningbox}
     In batchmode, \irace runs in the submission node of the cluster,
     hence, \irace is not submitted to the cluster as a job (that is,
     neither \code{qsub} nor \code{squeue} should be used to invoke \irace
     itself). The user must call the appropriate job submission command (e.g., \code{qsub} or \code{squeue}) from \parameter{targetRunner} with the appropriate settings for their cluster, that is,  \parameter{targetRunner} submits one job to the
     cluster and prints a single string: The job ID that allows
     \irace to determine the status of the running job. Moreover, the use of a
     separate
   \parameter{targetEvaluator} script is required to collect the results of
   \parameter{targetRunner} and return them to \irace.
   \end{xwarningbox}

  \item \parameter{targetRunnerParallel}: This option allows users to fully
    control the parallelization of the execution of \parameter{targetRunner}.
    Its value must be an \aR function that will be invoked by \irace as follows:

   <<targetRunnerParallel,prompt=FALSE, eval=FALSE>>=
   targetRunnerParallel(experiments, exec_target_runner, scenario, target_runner)
   @
   %
   where \code{scenario} is the list describing the configuration scenario
   (\autoref{sec:scenario}); \code{experiments} is a list that describes the
   configurations and instances to be executed (see \autoref{sec:runner} for a
   description); \code{target_runner} is the function that calls the target
   algorithm and it is the same as \parameter{targetRunner}, if the latter is a
   function, or it is a call to \code{target_runner_default}, if
   \parameter{targetRunner} is the path to an executable; and
   \code{exec_target_runner} is an internal function within \irace that takes
   care of executing \code{target_runner}, check its output and, possibly,
   retry in case of error (see \parameter{targetRunnerRetries}). The
   \parameter{targetRunnerParallel} function should call the given
   \code{target_runner} function for each element in the \code{experiments}
   list, possibly using \code{exec_target_runner} as a wrapper. A trivial
   example would be:

   <<targetRunnerParallel2,prompt=FALSE, eval=FALSE>>=
   targetRunnerParallel <- function(experiments, exec_target_runner, scenario,
                                    target_runner)
   {
     lapply(experiments, exec_target_runner, scenario = scenario,
            target_runner = target_runner)
   }
   @
   %
   However, the user is free to set up the calls in any way, perhaps
   implementing their own replacement for \code{target_runner} and/or
   \code{exec_target_runner}. The user may load and call other \aR packages, such as \pkg{batchtools} (\url{https://mllg.github.io/batchtools/}).

   The only requirement is that the \parameter{targetRunnerParallel} function
   must return a list of the same length as \code{experiments}, where each
   element is the output expected from the corresponding call
   to \parameter{targetRunner} (see \autoref{sec:runner}).

   The following is an example of the output of a call to
   \parameter{targetRunnerParallel} with 2 experiments, in which the execution
   time is not reported:

  <<targetRunnerParallel3,prompt=FALSE, eval=TRUE>>=
     print(output)
   @
   %

 \end{enumerate}

 The best option will depend on the resources available to you. Option 1 is
 usually the fastest and simplest to setup.  Running on a node (machine) with
 128 CPUs will be faster than running on 8 nodes with 16 CPUs because the
 communication between nodes required by MPI can be very slow depending on the
 cluster.  Option 2 may be faster if irace generates more configurations per
 iteration than the number of CPUs of a single node. However, depending on the
 configuration of your cluster, requesting many CPUs may require waiting in the
 queue a long time. Option 3 may be the slowest since irace has to check the
 queue frequently. However, irace will start running experiments as soon as 1
 CPU is available, thus option 3 may actually finish earlier than the other
 options if there is always some CPUs available in the cluster but the queue
 for requesting many CPUs at once is very long.

 As a rule-of-thumb, if you only have access to a single machine, then you only need option 1. If you have access to a computing cluster with multiple machines, then  use option 1 with the maximum number of CPUs that a single node has in your computing cluster. If that number is 64 or more, it should be enough unless a single run of irace evaluates thousands of configurations. Otherwise, investigate option 2. If option 2 does not work, then investigate option 3.

%%
%%
%%
%% Testing
%%
%%
%%

\section{Testing (Validation) of configurations}\label{sec:testing}

Once the tuning process is finished, \irace returns a set of configurations
corresponding to the elite configurations at the end of the run, ordered from
best to worst. In order to evaluate the generality of these configurations
without looking at their performance on the training set, \irace offers the
possibility of evaluating these configurations on a test instance set,
typically different from the training set used during the tuning phase. These
evaluations will use the same settings for parallel
execution, \parameter{targetRunner} and \parameter{targetEvaluator}.

The test instances can be specified by the
options \parameter{testInstancesDir} and/or \parameter{testInstancesFile}, or by
setting directly the variable \code{scenario\$testInstances}. These options behave similarly to their counterparts for the training instances
(\autoref{sec:training}). In particular, each test instance is assigned a
different seed in the same way as done for the training instances. In
principle, \irace evaluates each configuration on each testing instance just
once, because evaluating one run on $n$ instances is always better than
evaluating $n'$ runs on $n/n'$ instances~\citep{IRIDIA-2004-001}. However, if
the number of instances is limited, one can always duplicate instances as
needed in the \parameter{testInstancesFile}, and \irace will assign a different
random seed to each instance.
%
An example of the output produced by \irace when testing is shown in Fig.~\ref{fig:testing_output}.

\afterpage{\clearpage
  \begin{figure}[p]
\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\scriptsize,
 %
 frame=none,  % top and bottom rule only
 framesep=0pt, % separation between frame and text
 rulecolor=\color{Gray},
}
% FIXME: This doesn't break long lines, find another alternative.
{\ttfamily\scriptsize\ldots
\VerbatimInput{irace-acotsp-testing.txt}}%
\caption{Sample text output of \irace when evaluating on test instances.}\label{fig:testing_output}
\end{figure}}



The options \parameter{testNbElites} and \parameter{testIterationElites}
control which configurations are evaluated during the testing phase. In
particular, setting \code{testIterationElites = 1} will test not only the final
set of elite configurations (those returned at the end of the training phase),
but also the set of elites at the end of each race
(iteration). The option \parameter{testNbElites} limits the maximum number of
configurations considered within each set. Some examples:
\begin{itemize}
\item \code{testIterationElites = 0; testNbElites = 1} means that only the best
  configuration found during the run of \irace, the final best, will be used in
  the testing phase.
\item \code{testIterationElites = 1; testNbElites = 1} will test, in addition to the final best, the best configuration found at each iteration.
\item \code{testIterationElites = 1; testNbElites = 2} will test the two best
  configurations found at each iteration, in addition to the final best and
  second-best configurations.
\end{itemize}

The testing can be also (re-)executed at a later time by using the following
\aR command (but you may need to override \code{testNbElites} and \code{testIterationElites}):
%
<<testing_r, prompt=FALSE, eval=FALSE>>=
testing_fromlog(logFile = "./irace.Rdata", testNbElites = 1)
@
%
The above line will load the scenario setup from \parameter{logFile} to
perform the testing.  The testing results will be stored in the \aR object
\code{iraceResults\$testing}, which is saved in the file specified by
\code{scenario\$logFile}. The structure of the object is described in
\autoref{sec:output r}.  For examples on how to analyse the results see
\autoref{sec:analysis}.

Another alternative is to test a specific set of configurations using the command-line option \parameter{--only-test} as follows:
\begin{lstlisting}[style=BashInputStyle]
irace --only-test configurations.txt
\end{lstlisting}
where \path{configurations.txt} has the same format as the
set of initial configurations (\autoref{sec:initial}).

%%
%%
%%
%% Recovery
%%
%%
%%
\section{Recovering \irace runs}\label{sec:recovery}

Problems like power cuts, hardware malfunction or the need to use computational
power for other tasks may occur during the execution of \irace, terminating a
run before completion. At the end of each iteration, \irace saves an \aR data
file (\parameter{logFile}, by default \code{"./irace.Rdata"}) that not only
contains information about the tuning progress (\autoref{sec:output r}), but
also internal information that allows recovering an incomplete execution.

To recover an incomplete \irace run, set the option \parameter{recoveryFile} to
the log file previously produced, and \irace will continue the execution from
the last saved iteration. The state of the random generator is saved and
loaded, therefore, as long as the execution is continued in the same machine,
the obtained results will be exactly the same as executing \irace in one step
(external factors, such as CPU load and disk caches, may affect the
target algorithm and that may affect the results).  You can specify
the \parameter{recoveryFile} from the command-line or from the scenario file,
and execute \irace as described in \autoref{sec:execution}. For example, from
the command-line use:

\begin{lstlisting}[style=BashInputStyle]
irace --recovery-file "./irace-backup.Rdata"
\end{lstlisting}


\begin{xwarningbox}
  When recovering a previous run, \irace will try to save data on the file
  specified by the \parameter{logFile} option. Thus, you must specify different
  files for \parameter{logFile} and \parameter{recoveryFile}.  Before
  recovering, we strongly advise to rename the saved \aR data file as in the
  example above, which uses \code{"irace-backup.Rdata"}.
\end{xwarningbox}

\hide{%

%\LESLIE{Actually, there other variables we allow to change...what we shoudl say about this?}
% MANUEL: I would say nothing for now until we fix this properly.
When recovery is done you can modify some \irace options from the command-line
or from the scenario file.

\begin{itemize}
\item \parameter{execDir}
\item \parameter{logFile}
\item \parameter{debugLevel}
\item \parameter{parallel}
\item \parameter{loadBalancing}
\item \parameter{mpi}
\item \parameter{batchmode}
\end{itemize}

It is not possible (and in most cases not correct) to change the scenario options, given that the
previous results can become invalid. However there are some cases in which this options must be changed,
for example, if you have instances that are files, it might be the case that the instances are not
available at the same location. To change the location of them you should modify directly the \aR data file,
please be careful not to change the order of the instances, because this would make the results obtained by
\irace invalid. To change the instances path, open the \aR console and use:

<<change_recover, prompt=TRUE, eval=FALSE>>=
iraceResults <- read_logfile("./tuning/irace.Rdata")
new_path <- "./experiments/tuning/instances/"
iraceResults$scenario$instances <-
   paste0(new_path, basename(iraceResults$scenario$instances))
save(iraceResults, file="./tuning/irace.Rdata")
@

This example can also can be applied to the target execution files \parameter{targetRunner} and \parameter{targetEvaluator}.
}


\begin{xwarningbox}
  Do not change anything in the log file or the scenario file before
  recovering, as it may have unexpected effects on the recovered run of
  \irace. In case of doubt, please contact us first (\autoref{sec:contact}). In particular, it is not possible to continue a run of \irace by recovering with a larger budget. Results will \textbf{not} be the same as running \irace from the start with the largest budget.  An alternative is to use the final configurations from one run as the initial configurations of a new run.
\end{xwarningbox}

\begin{xwarningbox}
  If your scenario uses \parameter{targetEvaluator} (\autoref{sec:evaluator})
  and \parameter{targetEvaluator} requires files created
  by \parameter{targetRunner}, then recovery will fail if those files are not
  present in the \parameter{execDir} directory. This can happen, for example,
  if you recover from a different directory than the one from which irace was
  initially executed, or when \parameter{execDir} is set to a temporary
  directory for every irace run. Thus, you need to copy the contents of the
  previous \parameter{execDir} into the new one.
\end{xwarningbox}

%%
%%
%%
%% Output and analysis of results
%%
%%
%%
\section{Output and results}

During its execution, \irace prints information about the progress of the
tuning in the standard output.  Additionally, after each iteration, an \aR data
file is saved (\parameter{logFile} option) containing the state of \irace.

\subsection{Text output}\label{sec:output text}

Figure~\ref{fig:output} shows the output, up to the end of the first iteration,
of a run of elitist \irace applied to the \ACOTSP scenario with 1000
evaluations as budget.

\afterpage{\clearpage
\begin{figure}[p]
\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\scriptsize,
 %
 frame=none,  % top and bottom rule only
 framesep=0pt, % separation between frame and text
 rulecolor=\color{Gray},
 firstline=1,
 lastline=68,
}
% FIXME: This doesn't break long lines, find another alternative.
\VerbatimInput{irace-acotsp-stdout.txt}%
\caption{Sample text output of \irace.}\label{fig:output}
\end{figure}}

First, \irace gives the user a warning informing that it has found a file with
the default scenario filename and it will use it. Then, general information
about the selected \irace options is printed:
\begin{itemize}
 \item \parameter{nbIterations} indicates the minimum number of iterations \irace has calculated
       for the scenario. Depending on the development of the tuning the final iterations that
       are executed can be more.
  \item \parameter{minNbSurvival} indicates the minimum number of alive configurations that
       are required to continue a race. When less configurations are alive the race is
      stopped and a new iteration begins.
  \item \code{nbParameters} is the number of parameters of the scenario.
  \item \parameter{seed} is the number that was used to initialize the random number generator in \irace.
  \item \code{confidence level} is the confidence level of the statistical test.
  \item \code{budget} is the total number of evaluations available for the tuning.
  \item \code{time budget} is the  maximum execution time available for the tuning.
  \item \parameter{mu} is a value used for calculating the minimum number of iterations.
  \item \parameter{deterministic} indicates if the target algorithm is assumed to be deterministic.
\end{itemize}


At each iteration, information about the progress of the execution is printed as follows:
\begin{itemize}
  \item \code{experimentsUsedSoFar} is the number of experiments from the total budget
        that have been used up to the current iteration.
  \item \code{timeUsed} is the execution time used so far in the experiments. Only available when reported
        in the \parameter{targetRunner} (activate it with the \parameter{maxTime} option).
  \item \code{remainingBudget} is the number of experiments that have not been used yet.
  \item \code{timeEstimate} estimation of the mean execution time. This is used to calculate
        the remaining budget when \parameter{maxTime} is used.
  \item \code{currentBudget} is the number of evaluations \irace has allocated to the
        current iteration.
  \item \parameter{nbConfigurations} is the number of configurations \irace will use in the
        current iteration. In the first iteration, this number of configurations include
        the initial configurations provided; in later iterations, it includes the elite
        configurations from the previous iterations.
\end{itemize}

After the iteration information, a table shows the progress of the iteration execution.
Each row of the table gives information about the execution of an instance in the race.
The first column contains a symbol that describes the results of the
statistical test:
%
\begin{description}
\item \code{|x|} No statistical test was performed for this instance. The options \parameter{firstTest}
      and \parameter{eachTest} control on which instances the statistical test is performed.
\item \code{|-|} Statistical test performed and configurations have been discarded. The column \code{Alive} gives an indication of how
      many configurations have been discarded.
\item \code{|=|} Statistical test performed and no configurations have been discarded.
      This means \irace needs to evaluate more instances to identify the best configurations.
\item \code{|!|} This indicator exists only for the elitist version of \irace. It indicates that
      the statistical test was performed and some elite configurations appear to show bad performance and
      could be discarded but they are kept because of the elitist rules. See option \parameter{elitist} in \autoref{sec:irace options} for more information.
\end{description}

Other columns have the following meaning:
\begin{description}
  \item \code{Instance}: Index of \code{(instance, seed)} pair executed. This number
corresponds to the row in the data frame returned by \code{get_instanceID_seed_pairs()}. See \autoref{sec:output r}
for more information. This index is different from the instance ID passed to \parameter{targetRunner}.
\item \code{Bound}: Only when \parameter{capping} is enabled. Execution time used as bound
for the execution of new candidate configurations.
\item \code{Alive}: Number of configurations that have not been
discarded after the statistical test was performed.
\item \code{Best}: ID of the best configuration according to the instances seen
so far in this race (i.e., not including previous iterations).
\item \code{Mean best}: Mean cost value of the best configuration across the instances seen so
far in this race (not globally). Equivalent to the concept of ``iteration-best'' in other algorithms.
\item  \code{Exp so far}: Number of experiments
performed so far.
\item \code{W time}:  Wall-clock time spent on this instance.
\item \code{rho}, \code{KenW}, and \code{Qvar}: Spearman's rank correlation
  coefficient rho, Kendall's concordance coefficient W, and a variance measure
  described in \cite{SchHoo2012quanti}, respectively, of the configurations
  across the instances evaluated so far in this iteration.  These measures
  evaluate how consistent is the performance of the configurations across the
  instances.  Values close to 1 for \code{rho} and \code{KenW} and values close
  to 0 for \code{Qvar} indicate that the scenario is highly homogeneous. For
  heterogeneous scenarios, we provide advice in \autoref{sec:het}.
\end{description}

Finally, \irace outputs the best configuration found and a list of the elite
configurations. The elite configurations are configurations that did not show
statistically significant difference during the race; they are ordered
according to their mean performance on the executed instances.


\subsection[R data file (logFile)]{\aR data file (\code{logFile})}\label{sec:output r}

The \aR data file created by \irace (by default as \code{irace.Rdata}, see option \parameter{logFile}) contains an object called
\code{iraceResults}. You can load this file in the \aR console with:

<<load_rdata, prompt=FALSE, eval=FALSE>>=
logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
iraceResults <- read_logfile(logfile)
@

The \code{iraceResults} object is a list, and the elements of a list can be
accessed in \aR by using the \code{\$} or \code{[[]]} operators:

<<show_version, prompt=TRUE, eval=TRUE, comment="">>=
iraceResults$irace_version
@

\noindent%
The \code{iraceResults} list contains the following elements:

\begin{itemize}
\item \code{scenario}: The scenario \aR object containing the \irace options
  used for the execution. See \autoref{sec:irace options} and the help
  of the \pkg{irace} package; open an \aR console and type: \code{?defaultScenario}.
  See \autoref{sec:irace options} for more information.

\item \code{parameters}: The parameters \aR object containing the description of
  the target algorithm parameters. See \autoref{sec:target parameters}.

\item \code{allConfigurations}: The target algorithm configurations generated by
  \irace. This object is a \code{data frame}, each row is a candidate
   configuration; the first column (\code{.ID.}) indicates the internal identifier of the
   configuration; the final column (\code{.PARENT.}) is the identifier
   of the configuration from which the current configuration was sampled; and the remaining columns correspond to the parameter values; each column is named as
   the parameter name specified in the parameter object.

<<show_configurations, prompt=TRUE, eval=TRUE, comment="">>=
head(iraceResults$allConfigurations)
@

\item \code{allElites}: A list that contains one element per iteration. Each element contains
  the internal identifier of the elite candidate configurations of the corresponding iteration
  (identifiers correspond to \code{allConfigurations$.ID.}).

<<show_idelites, prompt=TRUE, eval=TRUE, comment="">>=
print(iraceResults$allElites)
@

The configurations are ordered by mean performance, that is, the ID of the best
configuration corresponds to the first ID. To obtain the values of the
parameters of all elite configurations found by \irace use:

<<get_elites, prompt=TRUE, eval=TRUE, comment="">>=
logfile <- system.file(package="irace", "exdata", "irace-acotsp.Rdata", mustWork=TRUE)
getFinalElites(logfile, n = 0)
@

\item \code{iterationElites}: A vector containing the best candidate configuration ID of each iteration. The best configuration found corresponds to the last one
  of this vector.
<<show_iditelites, prompt=TRUE, eval=TRUE, comment="">>=
print(iraceResults$iterationElites)
@

One can obtain the full configuration with:

<<get_elite, prompt=TRUE, eval=TRUE, comment="">>=
last <- length(iraceResults$iterationElites)
id <- iraceResults$iterationElites[last]
getConfigurationById(iraceResults, ids = id)
@

\item \code{rejectedConfigurations}: A vector containing the rejected configurations IDs. These
correspond to configurations that produced failed executions and were ignored by \irace during the
configuration process. See \autoref{sec:reject} to enable the detection of such configurations.

\item \code{experiments}: A matrix with configurations as columns and instances
as rows. Column names correspond to the internal identifier of the configuration
(\code{allConfigurations$.ID.}). The results of a particular
configuration can be obtained using:

<<get_experiments, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we use the best configuration found
best_config <- getFinalElites(iraceResults, n = 1)
best_id <- as.character(best_config$.ID.)
# Obtain the results of the best configuration
all_exp <- iraceResults$experiments[, best_id]
# all_exp is a vector and names(all_exp) is the (instance,seed) index.
all_exp
# Obtain the results of the first and best configurations
all_exp <- iraceResults$experiments[, c("1", best_id)]
# all_exp is a matrix: colnames(all_exp) is configurationID and
# rownames(all_exp) is the (instance,seed) index.
all_exp
@

When a configuration was not executed on an instance, its value is \code{NA}. A
configuration may not be executed on an instance because: (1) it was not created
yet when the instance was used, or (2) it was
discarded by the statistical test and not executed on subsequent instances, or
(3) the race terminated before this instance was considered.

Row names correspond to the row index of the \code{(instance, seed)} pairs
in the data frame returned by \code{get_instanceID_seed_pairs()}. The instanceID and seed used
for a particular experiment can be obtained with:\sloppy

<<get_instance_seed, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we get instanceID, seeds and instances of the experiments
# of the best configuration.
# We could get the indexes of the instances on which at least one
# configuration was executed:
pair_index <- which(apply(!is.na(all_exp), 1L, any))
# or the instances on which all configurations were executed:
pair_index <- which(apply(!is.na(all_exp), 1L, all))
# but in this example we get the indexes of the instances executed for
# the best configuration.
pair_index <- which(!is.na(all_exp[, best_id]))
instanceID <- get_instanceID_seed_pairs(iraceResults)[["instanceID"]][pair_index]
# or get the seeds
get_instanceID_seed_pairs(iraceResults)[["seed"]][pair_index]
# or obtain the actual instances.
iraceResults$scenario$instances[instanceID]
# If the instances are of atomic type (integers, floating-point numbers or
# character strings), the above is similar to:
get_instanceID_seed_pairs(iraceResults, index = pair_index, instances=TRUE)
@
\item \code{experimentLog}: A matrix with columns \code{iteration, instance, configuration}. This matrix contains the log of all the experiments that \irace
performs during its execution. The \code{instance} column refers to the index of the
data frame returned by \code{get_instanceID_seed_pairs()}. When \parameter{capping} is enabled a column \code{bound} is added to log the execution bound applied
for each execution.

\item \code{softRestart}: A logical vector that indicates if a soft restart was
performed on each iteration. If \code{FALSE}, then no soft restart was performed.
See option \parameter{softRestart} in \autoref{sec:irace options}.

\item \code{state}: A list that contains the state of \irace, the recovery
(\autoref{sec:recovery}) is done using the information contained in this object. The
probabilistic model of the last elite configurations can be found here by doing:

<<get_model, prompt=TRUE, eval=TRUE, comment="">>=
# As an example, we get the model probabilities for the
# localsearch parameter.
iraceResults$state$model["localsearch"]
# The order of the probabilities corresponds to:
iraceResults$scenario$parameters$domains$localsearch
@

The example shows a list that has one element per elite configuration (ID as
element name). In this case, \code{localsearch} is a categorical parameter and
it has a probability for each of its values.


\item \code{testing}: A list that contains the testing results.
The list contains the following elements:
  \begin{itemize}
    \item \code{experiments}: Matrix of experiments in the same format as the
      \code{iraceResults$experiments} matrix. The column names indicate the candidate
      configuration identifier and the row names contain the name of the instances.
<<get_test_exp, prompt=TRUE, eval=TRUE, comment="">>=
# Get the results of the testing
iraceResults$testing$experiments
@

    \item \code{seeds}: The seeds used for the experiments, each seed corresponds to
      each instance in the rows of the test \code{experiments} matrix.
<<get_test_seeds, prompt=TRUE, eval=TRUE, comment="">>=
# Get the seeds used for testing
iraceResults$testing$seeds
@

In the example, instance \code{1000-1.tsp} is executed with seed \Sexpr{iraceResults$testing$seeds[1]}.

  \end{itemize}

\end{itemize}

\subsection{Analysis of results}\label{sec:analysis}

The final configurations returned by \irace are the elites of the final race.
They are reported in decreasing order of performance, that is, the best
configuration is reported first.

If testing is performed, you can further analyze the resulting best
configurations by performing statistical tests in \aR:
% FIXME: iraceplot
% or just plotting the results:
%<<plot_test, fig.pos="tbp", fig.align="center", fig.height = 4, fig.width = 8, out.width='0.85\\textwidth', fig.cap="Boxplot of the testing results of the best configurations.", prompt=TRUE, eval=TRUE, comment="">>=
<<wilcox_test,prompt=TRUE, eval=TRUE, comment="">>=
results <- iraceResults$testing$experiments
# Wilcoxon paired test
conf <- gl(ncol(results), # number of configurations
           nrow(results), # number of instances
           labels = colnames(results))
pairwise.wilcox.test (as.vector(results), conf, paired = TRUE, p.adj = "bonf")
@
% # Plot the results
% configurationsBoxplot (results, ylab = "Solution cost")

The Kendall concordance coefficient (\code{W}) and the Spearman's rho can be
applied over data that has the characteristics of the data obtained in the testing,
that is a full matrix where all configurations are executed in all instances. \code{W}
can show if the configurations tested have an homogeneous performance on the used instances
set. If evidence of an heterogeneous scenario found we recommend to make some adjustments
in the \irace options as described in \autoref{sec:het}.

<<conc, prompt=TRUE, eval=TRUE, comment="">>=
irace:::concordance(iraceResults$testing$experiments)
@
%

%% FIXME: iraceplot
% During the tuning, \irace iteratively updates the sampling models of the
% parameters to focus on the best regions of the parameter search space. The
% frequency of the sampled configurations can provide insights on the parameter
% search space. We provide a function for plotting the
% frequency of the sampling of a set of configurations. For more information on this function, please see the \aR help, type in the
% \aR console: \code{?parameterFrequency}. The following example
% plots the frequency of the parameters sampled during one \irace run:

% <<freq, fig.pos="tbp", fig.cap="Parameters sampling frequency.", out.width="\\textwidth",prompt=TRUE, eval=TRUE, comment="">>=
% parameterFrequency(iraceResults$allConfigurations, iraceResults$scenario$parameters)
% @

%% FIXME: iraceplot
% By using parallel coordinates plots, it is possible to analyze how the parameters
% interact with each other. For more information on this function, please see the \aR help, type
% in the \aR console: (\code{?parallelCoordinatesPlot}).
%  The following example shows how to create a parallel
% coordinate plot of the configurations in the last two iterations of \irace.

% <<parcord, fig.pos="tbp", fig.align="center", out.width="0.7\\textwidth", fig.cap="Parallel coordinate plots of the parameters of the configurations in the last two iterations of a run of \\irace.", prompt=FALSE, eval=TRUE, comment="">>=
% # Get last iteration number
% last <- length(iraceResults$iterationElites)
% # Get configurations in the last two iterations
% conf <- getConfigurationByIteration(iraceResults,
%                                     iterations = c(last - 1, last))
% parallelCoordinatesPlot (conf, iraceResults$scenario$parameters,
%                          param_names = c("algorithm", "alpha", "beta", "rho",
%                                          "q0", "rasrank"),
%                          hierarchy = FALSE)
% @

%%% FIXME: This plot is completely misleading. We should either calculate the
%%% performance over the whole training set, or find a way to plot the
%%% performance estimated by irace as a confidence interval to account for
%%% different number of instances. Or simply plot a heatmap of instances x
%%% best-of-each iteration.
% It is also possible to plot the performance evolution of the best-so-far configuration over the number of experiments as follows:
% %%<<trainEvo, fig.pos="tbp", fig.align="center", out.width='0.75\\textwidth', fig.cap="Training set performance of the best-so-far configuration over number of experiments.", prompt=FALSE, eval=TRUE, comment="">>=
% # Get number of iterations
% iters <- unique(iraceResults$experimentLog[, "iteration"])
% # Get number of experiments (runs of target-runner) up to each iteration
% fes <- cumsum(table(iraceResults$experimentLog[,"iteration"]))
% # Get the mean value of all experiments executed up to each iteration
% # for the best configuration of that iteration.
% values <- sapply(iters, function(k) {
%   instances <- as.character(
%     unique(iraceResults$experimentLog[
%                           iraceResults$experimentLog[, "iteration"] == k,
%                           "instance"]))
%   return(mean(iraceResults$experiments[
%                              instances,
%                              as.character(iraceResults$iterationElites[k])]))})
% plot(fes, values, type="s", xlab = "Number of runs of the target algorithm",
%      ylab = "Estimated mean value over whole training set")
% points(fes, values)
% @

% FIXME: We should normalize the ranges per instance so since the performance
% on different instances is quite different.
It is also possible, as shown in Fig.~\ref{fig:testEvo}, to plot the performance on the test set of the best-so-far configuration over the number of experiments as follows:
<<testEvo, fig.pos="tb", fig.align="center", out.width='0.7\\textwidth', fig.cap="Testing set performance of the best-so-far configuration over number of experiments. Label of each point is the configuration ID.", prompt=FALSE, eval=TRUE, comment="">>=
# Get summary data from the logfile.
irs <- irace_summarise(iraceResults)
# Get number of iterations
iters <- irs$n_iterations
# Get number of experiments (runs of target-runner) up to each iteration
fes <- cumsum(table(iraceResults$state$experiment_log[["iteration"]]))
# Get the mean value of all experiments executed up to each iteration
# for the best configuration of that iteration.
elites <- as.character(iraceResults$iterationElites)
values <- colMeans(iraceResults$testing$experiments[, elites])
stderr <- function(x) sqrt(var(x)/length(x))
err <- apply(iraceResults$testing$experiments[, elites], 2L, stderr)
plot(fes, values, type = "s",
     xlab = "Number of runs of the target algorithm",
     ylab = "Mean value over testing set", ylim=c(min(values-err),max(values+err)))
points(fes, values, pch=19)
arrows(fes, values - err, fes, values + err, length=0.05, angle=90, code=3)
text(fes, values, elites, pos = 1)
@

The \irace package also provides an implementation of the ablation
method~\cite{FawHoos2015ablation}. See \autoref{sec:ablation}.

\medskip{}
Finally, more advanced visualizations of the behavior of \irace are provided by
the ACVIZ software package~\cite{DesRitLopPer2021acviz}, which is available at \url{https://github.com/souzamarcelo/acviz}. See an example in Fig.~\ref{fig:acviz}.

\begin{figure}[htb]
  \centering
  \includegraphics[width=\textwidth]{fig1u-acotsp-instances}
  \caption{Visualization produced by ACVIZ~\cite{DesRitLopPer2021acviz}.}
  \label{fig:acviz}
\end{figure}



%%
%%
%%
%% Advanced options
%%
%%
%%

\section{Advanced topics}

\subsection{Tuning budget}\label{sec:budget}

Before setting the budget for a run of \irace, please consider the number of
parameters that need to be tuned, available processing power and available
time. The optimal budget depends on the difficulty of the tuning scenario, the
size of the parameter space and the heterogeneity of the instances. Typical
values range from $1\,000$ to $100\,000$ runs of the target algorithm, although
smaller and larger values are also possible. Currently, \irace does not detect
whether the given budget allows generating all possible configurations. In such
a case, the use of \emph{iterated} racing is unnecessary: One can simply
perform a single race of all configurations (see FAQ in
\autoref{faq:allconfs}).

\Irace provides two options for setting the total tuning budget (\parameter{maxExperiments} and
\parameter{maxTime}). The option \parameter{maxExperiments}
limits the number of executions of \parameter{targetRunner} performed by \irace. The option \parameter{maxTime} limits the
total time of the \parameter{targetRunner} executions. When this latter option is used, \parameter{targetRunner}
must return the evaluation cost together with the execution time (\code{"cost time"}).

\begin{xwarningbox}
When the goal is to minimize the computation time of an algorithm, and you wish to use \parameter{maxTime} as the tuning budget, \parameter{targetRunner} must return the time also as the evaluation cost, that is, return the time twice as \code{"time time"}.
\end{xwarningbox}

\begin{xwarningbox}
When using \parameter{targetEvaluator} and using \parameter{maxTime} as tuning budget, \parameter{targetRunner} just returns the time (\code{"time"}) and \parameter{targetEvaluator} returns the cost.
\end{xwarningbox}

When using \parameter{maxTime}, \irace estimates the execution time of each \parameter{targetRunner} execution
before the configuration. The amount of budget used for the estimation is set with the option \parameter{budgetEstimation}
(default is $2\%$). The obtained estimation is adjusted after each iteration using the obtained results and it is used to estimate the number of experiments that can be executed. Internally, \irace uses the number of remaining experiments to
adjust the number of configurations tested in each race.

\subsection{Multi-objective tuning}\label{sec:multi objective}

Currently, \irace only optimizes one cost value at a time, which can be
solution cost, computation time or any other objective that is returned to
\irace by the \parameter{targetRunner}. If the target algorithm is
multi-objective, it will typically return not a single cost value, but a set of
objective vectors (typically, a Pareto front). For tuning such a target
algorithm with \irace, there are two alternatives. If the algorithm returns a
single vector of objective values, they can be aggregated into one single number
by using, for example, a weighted sum. Otherwise, if the target algorithm
returns a set of objective vectors, a unary quality metric (\eg~the
hypervolume) may be used to evaluate the quality of the set.\footnote{An implementation is publicly available at \url{http://lopez-ibanez.eu/hypervolume}~\cite{FonPaqLop06:hypervolume}}

% The first option is simple, it requires to devise a formula that can aggregate
% the objectives in a way that balances the importance of all of them. This might
% not be an easy task in some scenarios, and therefore using a more adequate
% indicator to evaluate the performance of a multi-objective optimizer, such as
% the hypervolume, is strongly advised.

%For using the hypervolume as evaluation \irace needs to postpone the evaluation of the
%configurations in an instance until all the executions have been completed. Once
%the executions are finalized, the evaluation starts by obtaining reference points for
%each objective from the configuration results. These points are used to define the
%Pareto front and the hypervolume of each configuration
%is calculated and the tuning process continues normally.

The use of aggregation or quality metrics often requires normalizing the
different objectives. If normalization bounds are known a priori for each
instance, normalized values can be computed by \parameter{targetRunner}.
Otherwise, the bounds may be dynamically computed while running \irace, by
using \parameter{targetEvaluator}. In this case, \parameter{targetRunner} will
save the output of the algorithm, then the first call
to \parameter{targetEvaluator} will examine the output produced by all calls
to \parameter{targetRunner} for the same instance, update the normalization
bounds and return the normalized output. Subsequent calls
to \parameter{targetEvaluator} for the same instance will simply return the
normalized output.
%
A similar approach can be used to dynamically compute the reference points or
sets often required by unary quality metrics.

For more information about defining a \parameter{targetEvaluator}, see
\autoref{sec:evaluator}. Examples of tuning a multi-objective target algorithm
using the hypervolume can be found in the examples at
\IRACEHOME{/examples/hypervolume} and \IRACEHOME{/examples/moaco}.

\subsection{Tuning for minimizing computation time}
\label{sec:capping}

When using \irace for tuning algorithms that only report computation time to
reach a target, \parameter{targetRunner} should return the execution time of a configuration instead of solution cost. When using \parameter{maxTime} as the budget, this means that \parameter{targetRunner} must return twice the execution time since the first value is the minimization objective and the second value is used to track the budget consumed.

Starting from version $3.0$,  \irace includes an elitist racing procedure that
implements an \textbf{adaptive capping mechanism} \citep{PerLopHooStu2017:lion}.
Adaptive capping \citep{HutHooLeyStu2009jair} is a configuration technique
that avoids the execution of long runs of the target algorithm, focusing the
configuration budget in the evaluation of the best configurations found. This is
done by bounding the execution time of each configuration based on the best
performing candidate configurations.

To use adaptive capping, the \parameter{capping} option must be enabled and the
\parameter{elitist} irace option must be selected. When evaluating candidate
configurations on an instance, \irace calculates an execution bound based on the
execution times of the elite configurations.  The \parameter{boundType} option defines
how the performance of the elite configurations  is defined to obtain the execution bound.
The default value of \parameter{boundType} calculates the performance ($p_i^s$) of each
elite configuration ($s$) as the mean execution time of the  instances already executed
in the race and the currently executed instance ($i$). The \parameter{cappingType} option
specifies the measure used to obtain the elite configurations bound. By default, the execution bound
is calculated as the median of the execution times of the elite configurations:
%
\begin{equation}
  b_i= \text{Median}_{\theta_s \in \Celite} \{ p_i^s\}
\end{equation}

The execution bound for new configurations ($j$) is calculated by multiplying the elite configurations
bound by the number of instances ($i$) in the execution list and subtracting the mean execution time
of the instances executed by the candidate:
%
\begin{equation}
    k_i^{'j} = b_{i} \cdot i + \bmin - p^{j}_{i-1} \cdot (i-1)
\end{equation}

A small constant \bmin is added to account for time measurements errors.
These settings are also used to apply a dominance elimination criterion together with the
statistical test elimination. The domination criterion is defined as:
%
\begin{equation}\label{eq:dom}
b_i + \bmin  < p_i^{j}
\end{equation}

When elite configurations dominate new configurations, these are eliminated from the race.

\begin{xwarningbox}
The default statistical test when \parameter{capping} is enabled is \code{t-test}.
This test is more appropriate to configure algorithms for optimizing runtime (see \autoref{sec:stat test}).
\end{xwarningbox}

The execution bound is constantly adjusted by \irace based on the best configurations
times, nevertheless, a maximum execution time (\bmax) is never  exceeded. This maximum
execution time must be defined in the configuration scenario when \parameter{capping}
is enabled. To specify the maximum execution bound for the target runner executions
use the \parameter{boundMax} option. The final execution bound ($k_i^j$) is calculated by:
%
\begin{equation}\label{eq:cap}
  k_i^j = \begin{cases}
    \bmax &\text{if $k_i^{'j} > \bmax$,} \\
    \min \{b_i, \bmax\}& \text{if $k_i^{'j} \leq 0$,}\\
    k_i^{'j}&\text{otherwise;}
  \end{cases}
\end{equation}

Additionally, the \parameter{boundDigits} option defines the precision of the time bound
provided by \irace, the default setting is 0.

Timed out executions occur when the maximum execution bound
(\parameter{boundMax}) is reached and the algorithm has not achieved successful
termination or a defined quality goal.  In this case, it is a common practice
to apply a penalty known as PARX, in which timeouts are penalized by
multiplying \parameter{boundMax} by a constant $X$. The constant $X$ may be set
using the \parameter{boundPar} option. Bounded executions are executions that
do not achieve successful termination or a defined quality goal in the
execution bound ($k_i^j$) set by \irace, which is smaller than
\parameter{boundMax}. The \parameter{boundAsTimeout} option replaces the
evaluation of bounded executions by the \parameter{boundMax} value. More
details about the implementation of adaptive capping can be found in
\citet{PerLopHooStu2017:lion}.

\begin{xwarningbox}
Note that bounded executions are not timed out executions and thus, they will not be penalized by PARX.
\end{xwarningbox}


\begin{xwarningbox}
  Penalized evaluations of timed out and bounded executions are only used for
  the elimination tests and the comparison between the quality of
  configurations. To calculate execution bounds and computation budget
  consumed, \irace uses only unpenalized execution times. The unpenalized
  execution time must be provided by the target runner or target evaluator as
  described in \autoref{sec:runner} and \autoref{sec:evaluator} .
\end{xwarningbox}

\begin{xwarningbox}
More advanced capping methods that are applicable to minimizing solution cost are available when combining irace with the \texttt{capopt} package described by \citet{SouRitLop2021cap}.
\end{xwarningbox}



\subsection{Hyper-parameter optimization of machine learning methods}

The \irace package can also be used for model selection and hyper-parameter
optimization of machine learning (ML) methods. We will next explain a possible
setup for one given dataset and using $10$-fold cross-validation
(CV). Generalizing to multiple datasets and different resampling strategies,
e.g. leave-one-out, is straightforward.

First, split the dataset into training, to be used by \irace, and testing, to
be used for evaluating the performance of the configuration returned by
\irace. A typical split could be $70\%$ and $30\%$, respectively.

The training set is used by \irace to perform $10$-fold CV, that is, the data
is split into $10$ folds. A single run of the \parameter{targetRunner} will use
$9$ folds for training and the remaining fold for validation. Splitting the
data into folds can be done at each call of \parameter{targetRunner} or before
running \irace, however, it is important that the split is always the same for
every call of the \parameter{targetRunner}, i.e., the content of the folds does
not change, only which folds are used for training and validation will change.

The setup of \irace should be as follows:
%
\begin{itemize}
\item \parameter{trainInstancesFile}\code{="train-instances.txt"}, where this
  file contains one number per line from $1$ to $10$. This number will tell the
  \parameter{targetRunner} which fold should be used for validation.
\item \parameter{trainInstancesDir}\code{=""}, because the folds are the
  ``instances'' and you do not have actual instance files. If you want to pass
  the name of the dataset to the \parameter{targetRunner}, you can specify it
  either at each line of \code{"train-instances.txt"}, directly in the
  \parameter{targetRunner}, or as a fixed parameter in the
  \parameter{parameterFile}.
\item \parameter{deterministic}\code{=1} unless it really makes sense to
  train more than once the same ML model on the same data. If it makes sense,
  then your \parameter{targetRunner} should use the seed passed by \irace to
  seed the ML model before training.
\item \parameter{sampleInstances}\code{=0} because the folds should already be generated by randomly sampling the dataset.
\item \parameter{testType}\code{="t-test} because the performance metrics in
  ML are typically the mean of the CV results, which assumes that the
  performance are close to normally distributed.
\item \parameter{firstTest}\code{=2} because \irace should discard
  configurations very aggressively looking for maximum generality.
\end{itemize}

\noindent%
Finally, your \parameter{targetRunner} needs to be able to do the following:
\begin{itemize}
\item Receive from \irace the hyper-parameter settings, the dataset name and a
  fold number (the ``instance''). Let us use fold 3 as an example.
\item Train the ML model on the whole training set minus fold 3, then validate
  (score) the model on fold 3 and return the score to \irace (negated if the
  score must be maximized, because \irace assumes minimization). Since each
  fold is different, each instance should give a different result. Each row in
  the table printed by \irace should print something different; otherwise,
  something is wrong in your setup.
\end{itemize}

The above is actually 10 times faster than doing 10-fold CV for each call to
\parameter{targetRunner}, thus, you should assign to \irace 10 times the budget
than what would be assigned to other methods that do a complete 10-fold CV at
each step.

%% FIXME: They removed the tutorial
%% For such a task, we recommend the \pkg{mlr} package~\citep{R:mlr}.
%% The following webpage documents how to use \irace for this purpose: \url{https://mlr-org.github.io/mlr-tutorial/devel/html/advanced_tune/index.html}


\subsection{Heterogeneous scenarios}
\label{sec:het}

We classify a scenario as homogeneous when the target algorithm has a
consistent performance regarding the instances; roughly speaking, good
configurations tend to perform well and bad configurations tend to perform
poorly on all instances of the problem. By contrast, in heterogeneous
scenarios, the target algorithm has an inconsistent performance on different
instances, that is, some configurations perform well for a subset of the
instances, while they perform poorly for a different subset.

When facing a heterogeneous scenario, the first question should be whether the
objective of tuning is to find configurations that perform reasonably well over
all instances, even if that configuration is not the best ones in any
particular instance (a generalist). If this is not the goal, then it would be
better to partition instances into more similar subsets and execute \irace
separately on each subset. This will lead to a portfolio of algorithm
configurations, one for each subset, and algorithm selection techniques can be
used to select the best configuration from the portfolio when facing a new
instance.

To make sure \irace is not misled by results on few instances, it may be useful
to increase the number of instances executed before doing a statistical test
using the option \parameter{firstTest}, \eg \code{--first-test 10} (default
value is 5), in order to see more instances before discarding configurations.
The option \parameter{elitistNewInstances} in elitist \irace (option
\parameter{elitist}) can be used to increase the number of new instances
executed in each iteration, \eg \code{--elitist-new-instances 5} (default value
is 1).

If finding an overall good configuration for all the instances is the
objective, then we recommend that instances are randomly sampled (option
\parameter{sampleInstances}), unless one can provide the instances in a
particular order that does not bias the tuning towards any subset.

If instances are easily categorized in different classes, then we recommend to
create ``blocks'' of instances in \parameter{trainInstancesFile}, where each
block should contain one instance from each class.  Then set the option
\parameter{blockSize} to the number of classes within each block, so that
\irace will always see a complete block of instances before eliminating
configurations. The value of \parameter{blockSize} will multiply the values of
\parameter{firstTest}, \parameter{eachTest} and
\parameter{elitistNewInstances}. Randomly sampling instances
(\parameter{sampleInstances}\code{=1}) will randomly sample the blocks but not
break the blocks.

While executing \irace, the homogeneity of the scenario can be observed by
examining the values of Spearman's rank correlation coefficient and Kendall's
concordance coefficient in the text output of \irace.  See \autoref{sec:output
  text} for more information.

\subsection{Choosing the statistical test}
\label{sec:stat test}

The statistical test used in \irace identifies statistically bad performing
configurations that can be discarded from the race in order to save
budget. Different statistical tests use different criteria to compare the
cost of the configurations, which has an effect on the tuning results.

\Irace provides two types of statistical tests
(option \parameter{testType}). Each test has different characteristics that are
beneficial for different goals:
%
\begin{itemize}
\item Friedman test (\code{F-test}): This test uses the ranking of the
  configurations to analyze the differences between their performance. This
  makes the test suitable for scenarios where the scale of the performance
  metric is not as important to assess configurations as their relative
  ranking.  This test is also indicated when the distribution of the mean
  performances deviates greatly from a normal distribution. For example, the
  ranges of the performance metric on different instances may be completely
  difference and comparing the performance of different configurations using
  the mean over multiple instances may be deceiving.  We recommend to use the
  \code{F-test} (default when \parameter{capping} is not enabled) when tuning
  for solution cost and whenever the best performing algorithm should be among
  the best in as many instances as possible.

\item Student's t-test (\code{t-test}): This test uses the mean performance of
  the configurations to analyze the differences between the
  configurations.\footnote{The t-test does not require that the performance
    values follow a normal distribution, only that the distribution of sample
    means does. In practice, the t-test is robust despite large deviations from
    the assumptions.} This makes the test suitable for scenarios where the
  differences between values obtained for different instances are relevant to
  assess good configurations.  We recommend using t-test, in particular, when
  the target algorithm is minimizing computation time and, in general, whenever
  the best configurations should obtain the best average solution cost.
\end{itemize}


The confidence level of the tests may be adjusted by using the
option \parameter{confidence}. Increasing the value of \parameter{confidence}
leads to a more strict statistical test.  Keep in mind that a stricter test
will require more budget to identify which configurations perform worse. A less
strict test discards configurations faster by requiring less evidence against
them and, therefore, it is more likely to discard good configurations.


\subsection{Complex parameter space constraints}\label{sec:complex_domains}

Some parameters may have complex dependencies. Ideally, parameters should be
defined in the way that is more likely to help the search performed by
\irace.  For example, when tuning a branch and bound algorithm, one may have the
following parameters:

\begin{itemize}
\item branching (\code{b}) that takes values in \code{\{0,1,2,3\}}, where 0 indicates no branching will be used
      and the rest are different types of branching.
\item stabilization (\code{s}) that takes values in \code{\{0,1,2,3,4,5,6,7,8,9,10\}}, of which for \code{b=0}
      only \code{\{0,1,2,3,4,5\}} are relevant.
\end{itemize}

In this case, it is not possible to describe the parameter space by defining
only two parameters for \irace. An extra parameter must be introduced as
follows:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range                    condition
b         "-b "   c      (0,1,2,3)
s1        "-s "   c      (0,1,2,3,4,5)            | b == "0"
s2        "-s "   c      (0,1,2,3,4,5,6,7,8,9,10) | b != "0"
  \end{CodeInput}
\end{minipage}
\end{center}

Parameters whose values depend on the value of other parameters may also
require using extra parameters or changing the parameters and processing them
in \parameter{targetRunner}. For example, given the following parameters:
%
\begin{itemize}
\item Population size (\code{p}) takes the integer values $[1,100]$.
\item Selection size (\code{s}) takes the same values but no more than the
  population size, that is $[1,$\code{p}$]$.
\end{itemize}

In this case, it is possible to describe the parameters \code{p} and \code{s}
using surrogate parameters for \irace that represent a ratio of the original
interval as follows:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range
p        "-p "    i      (1,100)
s_f      "-s "   r      (0.0,1.0)
\end{CodeInput}
\end{minipage}
\end{center}
%
and \parameter{targetRunner} must calculate the actual value of \code{s} as $\min(\max(\text{round}(\text{\code{s\_f}}\cdot \text{\code{p}}, 1)), 100)$. For
example, if the parameter \code{p}  has value 50 and the surrogate parameter \code{s\_f} has value 0.3, then \code{s} will have value 15.

The processing within \parameter{targetRunner} can also split and join
parameters. For example, assume the following parameters:

\begin{center}
\begin{minipage}{0.8\linewidth}
\begin{CodeInput}
# name    label   type   range
m         "-m "   i      (1,250)
e         "-e "   r      (0.0,2.0)
\end{CodeInput}
\end{minipage}
\end{center}

These parameters could be used to define a value
$\texttt{m}\cdot 10^\texttt{e}$ for another parameter
(\code{--strength}) not known by \irace. Then, \parameter{targetRunner} takes care of parsing
\code{-m} and \code{-e}, computing the strength value and passing the parameter
\code{--strength} together with its value to the target algorithm.

More complex parameter space constraints may be implemented by means of the
\parameter{repairConfiguration} function (\autoref{sec:repairconf}).


\subsection{Unreliable target algorithms and immediate rejection}\label{sec:reject}

There are some situations in which the target algorithm may fail to execute
correctly.  By default, \irace stops as soon as a call
to \parameter{targetRunner} or \parameter{targetEvaluator} fails, which helps to
detect bugs in the target algorithm.  Sometimes the failure cannot be fixed
because it is due to system problems, network issues, memory limits, bugs for
which no fix is available, or fixing them is impossible because there is no
access to the source code.

In those cases, if the failure is caused by random errors or transient system
problems, one may wish to ignore the error and try again the same call in the
hope that it succeeds. The option \parameter{targetRunnerRetries} indicates the
number of times a \parameter{targetRunner} execution is repeated if it
fails. Use this option only if you know additional repetitions could be
successful.

If the target algorithm consistently fails for a particular set of
configurations, these configurations may be declared as forbidden
(\autoref{sec:forbidden}) so that \irace avoids them.  On the other hand, if
the configurations that cause the problem are unknown,
the \parameter{targetRunner} should return \code{Inf} so that \irace
immediately rejects the failing configuration. This immediate rejection should
be used with care according to the goals of the tuning. For example, a
configuration that crashes on a particular instance, \eg by running out of
memory, might still be considered acceptable if it gives very good results on
other instances. The configurations which were rejected during the execution of \irace are saved
in the Rdata output file (see \autoref{sec:output r}).
\begin{xwarningbox}
If the configuration budget is specified in total execution time (\parameter{maxTime} option),
immediate rejected executions must provide the cost and time (which must be \code{Inf 0}).
Nevertheless, rejected configurations will be  excluded from the execution time estimation
and the execution bound calculation.
\end{xwarningbox}

\subsection{Ablation Analysis}\label{sec:ablation}

The ablation method~\cite{FawHoos2015ablation} takes two configurations (source
and target) and generates a sequence of configurations that differ between each
other just in one parameter, where parameter values in source are replaced by
values from target. The sequence can be seen as a ``path'' from the source to
the target configuration.  This can be used to find new better ``intermediate''
configurations or to analyse the impact of the parameters in the performance.


To perform ablation, you can use the \code{ablation()} \aR function
%
<<ablation, prompt=FALSE, eval=FALSE>>=
ablog <- ablation("irace.Rdata", src = 1)
plotAblation(ablog)
@
%
\noindent or the \path{ablation} command-line executable (see more details below).
%
%<<linux_irace_help,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
ablation -l irace.Rdata --src 1 --plot plot-ablation.pdf --plot-type rank,boxplot
\end{lstlisting}
%@

You may specify the IDs of the source and target configurations.  By default, the
source is taken as the first configuration evaluated by \irace and the target
as the best overall configuration found. Use the function \code{plotAblation} to visualize the ablation results
(Fig.~\ref{fig:testAb}).

<<testAb, fig.pos="htb!", fig.align="center", out.width="0.75\\textwidth", fig.cap="Example of plot generated by \\code{plotAblation()}.", prompt=FALSE, eval=TRUE, echo=FALSE>>=
logfile <- system.file(package="irace", "exdata", "log-ablation.Rdata", mustWork=TRUE)
plotAblation(logfile)
@

The function returns a list containing the following elements:
%
\begin{description}
   \item \code{configurations}: A dataframe of configurations tested during ablation.
   \item \code{instances}: The instances used for the ablation.
   \item \code{scenario}: Scenario options provided by the user.
   \item \code{trajectory}: Best configuration IDs at each step of the ablation.
   \item \code{best}: Best overall configuration found.
\end{description}

We also provide a command-line executable (\path{ablation.exe} in Windows) that allows you to perform ablation without launching \aR. It is installed in the same location as the \irace command-line executable and has the following options:
<<ablation_cmdline, prompt=FALSE, eval=TRUE,echo=FALSE, comment="">>=
ablation_cmdline("--help")
@

\subsection{Post-selection race}\label{sec:postselection}

At the end of the configuration process, \irace will automatically run a
post-selection race \citep{YuaStuMonLauBir13} to try to consume any remaining
budget. This post-selection race will not generate new configurations. It will
only do additional runs on configurations already evaluated in the current run
of \irace.  First, it will try to evaluate the best configurations found in new
instances not seen yet. If this is not possible due to the budget available,
then it will evaluate configurations that have been evaluated on fewer
instances than the best configuration found. Currently, this automatic
post-selection is not supported when the budget is measured in runtime
(\parameter{maxTime}. The automatic post-selection can be disabled using the
option \parameter{postselection}.

It is also possible to manually run a post-selection race after completing a run of \irace using the R function \iracefun{psRace}.

<<postsel, prompt=FALSE, eval=FALSE>>=
# Execute all elite configurations in the iterations
psRace("irace.Rdata", max_experiments = 0.5, iteration_elites=TRUE)
# Execute a set of configurations IDs providing budget
psRace("irace.Rdata", conf_ids = c(34, 87, 102, 172, 293), max_experiments = 500)
@

The execution of the post-selection race adds an element (\code{psrace_log}) to
the \code{iraceResults} object saved in the \irace log file
(\parameter{logFile}). The post-selection log consists of a list with the
following elements:
\begin{description}
\item\code{configurations}: Configurations used in the post-selection race.
\item\code{instances}: Data frame with the instances and seeds used in the post-selection race.
\item \code{max_experiments}: Configuration budget assigned to the post-selection race.
\item \code{experiments}: Matrix of results generated by the post-selection
  race, in the same format as the matrix
  \code{iraceResults$experiments}. Column names are the configuration IDs and
  row names are the instance IDs.
\item \code{elites}: Best configuration(s) found in the post-selection race.
\end{description}


\subsection[Parameter importance analysis using PyImp]{Parameter importance analysis using \PyImp}

The \PyImp\footnote{\url{https://github.com/automl/ParameterImportance}} tool developed by the AutoML group\footnote{\url{https://www.automl.org/}}
supports various parameter importance analysis methods using surrogate models.
Given a performance dataset of an algorithm configuration scenario, a Random Forest is built to predict performance of all algorithm configurations.
Parameter importance analyses are then applied on the prediction model.
The model serves as a surrogate for the original target algorithm, so that the algorithm does not need to be executed during the analyses.
Three analysis methods are supported, namely fANOVA~\citep{HutHooLey2014icml} (functional analysis of variance), forward selection~\cite{HutHooLey2013lion},
and ablation analysis with surrogates~\cite{BieLinEggFraFawHoo2017}.
Note that the \irace package directly supports ablation (without surrogate models) analysis with and without racing (\autoref{sec:ablation}).
Although ablation analysis without surrogates may be more time-consuming, results of the surrogate version may be less accurate than the non-surrogate one.

The \code{.Rdata} dataset generated by \irace can be used as input for \PyImp.
The package \code{irace2pyimp}\footnote{\url{https://github.com/ndangtt/irace2pyimp}} is able to convert an \code{irace.Rdata} file into the input format required by \PyImp. The conversion can be  accessed either through the \aR console (function \code{irace2pyimp}), or via command line executable \code{irace2pyimp} (or \code{irace2pyimp.exe} in Windows) provided by the package. You can find the location of the executable by running this command in R: \Sexpr{'system.file(file.path("bin", "irace2pyimp"), package="irace2pyimp", mustWork=TRUE)'} or \Sexpr{'system.file(file.path("bin", "irace2pyimp.exe"), package="irace2pyimp", mustWork=TRUE)'} in Windows.

To see the usage of the executable, please run: \code{irace2pyimp --help}.
For more information on the \aR function \code{irace2pyimp}, type in the \aR console: \code{?irace2pyimp}.

Given as input an \code{irace.Rdata} file, the script will generate the following output files:
\begin{itemize}
  \item \code{params.pcs}: a text file containing the parameter space definition.
  \item \code{runhistory.json}: a JSON file containing the list of algorithm configurations evaluated during the tuning and the performance data obtained.
  \item \code{traj_aclib2.json}: a JSON file containing the best configurations after each iteration of \irace. The last configuration will be used as the target configuration in ablation analysis.
  \item \code{scenario.txt}: a text file containing the definition of the tuning scenario.
  \item \code{instances.txt}: a text file containing the list of instances.
  \item \code{features.csv}: a .csv file containing instance features. If no instance features are provided, the index of each instance will be used as a feature.
\end{itemize}

\PyImp can then be called using the files listed above as input. Several
examples on how to use the script and call \PyImp can be found at
\code{system.file("/examples/", package="irace2pyimp")}.

%\emph{\PyImp installation note:} At the time when this document is written (December 2019), the latest \PyImp release version (1.0.6) has some bugs.
%Please check on \url{https://pypi.org/project/PyImp/#history} to see if \PyImp version > 1.0.6 has been released.
%If it has, you can easily install \PyImp using \code{pip install pyimp}.
%Otherwise, please install \PyImp from source (\url{https://github.com/automl/ParameterImportance}), branch \code{development}. Installation instructions can be found at \url{https://automl.github.io/ParameterImportance/installation.html}.
%%
%% irace options
%%

\section{List of command-line and scenario options} \label{sec:irace options}

Most \irace options can be specified in the command line using a flag or in the
\irace scenario file using the option name (or setting their value in the
\code{scenario} list passed to the various \aR functions exported by the
package). This section describes the various \irace options that can be
specified by the user in this way.

\begin{xwarningbox}
  Relative filesystem paths (e.g., \path{../scenario/}) given in the
  command-line are relative to the current working directory (the directory at
  which \irace is invoked). However, paths given in
  the scenario file are relative to the directory containing
  the scenario file. See also Table~\ref{tab:paths}.
\end{xwarningbox}

<<child-demo, child='section/irace-options.Rnw'>>=
@

%%
%%
%%
%% FAQ
%%
%%
%%


\section{FAQ (Frequently Asked Questions)}

\subsection{Is \irace minimizing or maximizing the output of my algorithm?}

By default, \irace considers that the value returned by \parameter{targetRunner} (or by
\parameter{targetEvaluator}, if used) should be \underline{\textbf{minimized}}. In case of a maximization
problem, one can simply multiply the value by -1 before returning it to
irace. This is done, for example, when maximizing the hypervolume (see the last
lines in \IRACEHOME{/examples/hypervolume/target-evaluator}).


\subsection{Are experiments with \irace reproducible?}

Short answer: Yes, under some conditions.\\
Long answer: According to the terminology described by \citet{LopBraPaq2021telo}, we define \emph{repeatability} as ``\emph{exactly repeating the original experiment, generating precisely the same results}''. Following this definition, a run of \irace is repeatable under the following conditions:
\begin{itemize}
\item Same version of \irace.
\item Same version of \aR (different versions of \aR may change the behavior of functions used by \irace).
\item The behavior of \parameter{targetRunner} is deterministic or exactly reproducible for the same instance, parameter configuration and random seed. Make sure that \parameter{targetRunner} uses the seed provided by \irace to initialize all random number generators used. If the result of \parameter{targetRunner} depends on CPU-time, wall-clock time or system load in any way, then \parameter{targetRunner} is not reproducible and neither will be \irace.
\item Same random seed (\parameter{seed}) given to \irace.
\item Same scenario options (\autoref{sec:irace options}). Although some options should not affect reproducibility (e.g., \parameter{debugLevel}), maintaining a list of such options will be a huge effort, thus the safest assumption is that any change may break reproducibility.
\item Same parameter space (\autoref{sec:target parameters}), including types, domains, conditions and forbidden configurations. The order of the parameters may also affect reproducibility (the name of the parameters should not) because it affects the order in which random numbers are used.
\item Same training instances provided and in the same order (\autoref{sec:training}). Even if the instances are sampled randomly (\parameter{sampleInstances}), a different initial order will produce a different sample even with the same random seed.
\item Same initial configurations (\autoref{sec:initial}), if any.
\end{itemize}


\subsection[Is it possible to configure a MATLAB algorithm with irace?]
{Is it possible to configure a \MATLAB algorithm with \irace ?}

Definitely. There are three main ways to achieve this:
\begin{enumerate}
\item Edit the \parameter{targetRunner} script to call \MATLAB in a non-interactive
    way. See the \MATLAB documentation, or the following links.\footnote{\url{http://stackoverflow.com/questions/1518072/suppress-start-message-of-matlab}}\footnote{\url{http://stackoverflow.com/questions/4611195/how-to-call-matlab-from-command-line-and-print-to-stdout-before-exiting}} %
    You would need to pass the parameter received by \parameter{targetRunner} to your \MATLAB script.\footnote{\url{https://www.mathworks.com/matlabcentral/answers/97204-how-can-i-pass-input-parameters-when-running-matlab-in-batch-mode-in-windows}}\footnote{\url{https://stackoverflow.com/questions/3335505/how-can-i-pass-command-line-arguments-to-a-standalone-matlab-executable-running}}
    %
    There is a minimal example in \IRACEHOME{/examples/matlab/}.

\item Call \MATLAB code directly from \aR using the
    \pkg{matlabr} package (\url{https://cran.r-project.org/package=matlabr}). This
    is a better option if you are experienced in \aR. Define \parameter{targetRunner} as
    an \aR function instead of a path to a script. The function should
    call your \MATLAB code with appropriate parameters.

  \item Another possibility is calling \MATLAB directly from a different programming language and write \parameter{targetRunner} in that programming language,  for example, in \proglang{Python} (see examples in \IRACEHOME{/examples/target-runner-python/}).\footnote{\url{https://www.mathworks.com/help/matlab/matlab_external/call-matlab-functions-from-python.html}\\
    \url{https://www.mathworks.com/help/matlab/matlab_external/call-user-script-and-function-from-python.html}}
\end{enumerate}

\subsection{My program works perfectly on its own, but not when running under \irace. Is irace broken?}\label{faq:valgrind}

Every time this was reported, it was a difficult-to-reproduce bug, i.e., a \href{https://en.wikipedia.org/wiki/Heisenbug}{Heisenbug}, in
the program (target algorithm), not in \irace.
%
To detect such bugs, we recommend that you use, within
\parameter{targetRunner}, a memory debugger (e.g.,
\href{http://valgrind.org/}{\code{valgrind}}) to run your program. For example,
if your program is executed by \parameter{targetRunner} as:
  %
\begin{lstlisting}[style=BashInputStyle]
${EXE} ${FIXED_PARAMS} -i ${INSTANCE} ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}
%
then replace that line with:
%
\begin{lstlisting}[style=BashInputStyle]
valgrind --error-exitcode=1 ${EXE} ${FIXED_PARAMS} -i ${INSTANCE} \
  ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}

If there are bugs in your program, they will appear in
\code{\${STDERR}}, thus do not delete those files. Memory debuggers will
significantly slowdown your code, so use them only as a means to find what is
wrong with your target algorithm. Once you have fixed the bugs, you should
remove the use of \code{valgrind}.


\subsection{\irace seems to run forever without any progress, is this a bug?}\label{faq:infloop}

Every time this problem was reported, the issue was in the target algorithm and
not in \irace. Some ideas for debugging this problem:
\begin{itemize}
\item Check that the target algorithm is really not running nor paused nor
  sleeping nor waiting for input-output.
\item Use \parameter{debugLevel}\code{=3} to see how  \irace calls \code{target-runner}, run the same command outside \irace and verify that it terminates.
\item Add some output to your algorithm that reports at the very end the runtime and exit code. Verify that this output is printed when \irace calls your algorithm.
\item In \code{target-runner}, print something to a log file \emph{after} calling your target algorithm. Verify that this output appears in the log file when \irace is running.

\item Set a maximum timeout when calling your target algorithm from \code{target-runner} (see FAQ~\ref{faq:timeout}).

\end{itemize}

\subsection{My program may be buggy and run into an infinite loop. Is it possible to set a maximum timeout?}\label{faq:timeout}

We are not aware of any way to achieve this using \aR. However, in
GNU/Linux, it is easy to implement by using the \code{timeout} command\footnote{\url{http://man7.org/linux/man-pages/man1/timeout.1.html}} in
\code{targetRunner} when invoking your program.

\subsection{When using the mpi option, \irace is aborted with an error message indicating that a
  function is not defined. How to fix this?}\label{sec:mpi-error}

\pkg{Rmpi} does not work the same way when called from within a package and when called from a script or
interactively. When \irace creates the slave nodes, the slaves will load a copy of \irace automatically.
If the slave nodes are on different machines, they must have \irace installed. If \irace is not installed
system-wide, \aR needs to be able to find \irace on the slave nodes. This is usually done by setting \code{R_LIBS},
\code{.libPaths()} or by loading \irace using \code{library()} or \code{require()} with the argument ``\code{lib.loc}''.
The settings on the master are not applied to the slave nodes automatically, thus the slave
nodes may need their own settings. After spawning the slaves, it is too late to modify
those settings, thus modifying the shell variable \code{R_LIBS} seems the only valid
way to tell the slaves where to find \irace.

If the path is set correctly and the problem persists, please check these instructions:
%
\begin{enumerate}[leftmargin=*,widest=3]
\item Test that \irace and \pkg{Rmpi} work. Run \irace on a single machine (submit node), without calling \code{qsub}, \code{mpirun} or a
  similar wrapper around \irace or \aR.
\item Test loading \irace on the slave nodes. However,
  jobs submitted by \code{qsub}/\code{mpirun} may load \aR packages using a different mechanism from the way it happens if you log directly into the node (e.g., with \code{ssh}). Thus, you need to write a little \aR program such as:
  <<faq3, eval=FALSE>>=
  library(Rmpi)
  mpi.spawn.Rslaves(nslaves = 10)
  paths <- mpi.applyLB(1:10, function(x) {
    library(irace); return(path.package("irace")) })
  print(paths)
  @
  Submit this program to the cluster like you would submit \irace (using \code{qsub}, \code{mpirun} or whatever program is used to submit jobs to the cluster).

      \item In the script \code{bin/parallel-irace-mpi}, the function \code{irace_main()} creates an MPI job for our cluster.
      You may need to speak with the admin of your cluster and ask them how to best submit a job for MPI. There may be
      some particular settings that you need. \pkg{Rmpi} normally creates  log files; but \irace suppresses those files unless $\texttt{debugLevel} > 0$.
     \end{enumerate}

     Please contact us (\autoref{sec:contact}) if you have further problems.

\subsection[Error: 4 arguments passed to {.Internal(nchar)} which requires 3]{Error: 4 arguments passed to \code{.Internal(nchar)} which requires 3}

This is a bug in \aR 3.2.0 on Windows. The solution is to update your version of
\aR.

\subsection[Warning: In read.table(filename, header = TRUE, colClasses = "character", : incomplete final line found by \ldots]{Warning: \code{In read.table(filename,}  \code{header = TRUE,} \code{colClasses =} \code{"character",} \code{: incomplete final line found by} \ldots}

This is a warning given by \aR when the last line of an input file does not
finish with the newline character. The warning is harmless and can be
ignored. If you want to suppress it, just open the file and press the
\code{ENTER} key at the end of the last line of the file to end the final line with a newline.


\subsection{How are relative filesystem paths interpreted by \irace?}\label{faq:relpaths}

The answer depends on where the path appears. Relative paths may appear as the
argument of command-line options, as the value of options given in the scenario
file, or within various scripts, functions or instance files. Table
\ref{tab:paths} summarizes how paths are translated from relative to absolute.

\begin{table}[h!]
  \centering
  \caption{Translation of relative to absolute filesystem paths.}
  \label{tab:paths}
  \begin{tabular}{rl}
    \toprule
\bf    Relative path appears as \ldots  & \bf\ldots is relative to \ldots \\
    \midrule
    a string within \parameter{trainInstancesFile} & \parameter{trainInstancesDir}\\
    a string within \parameter{testInstancesFile} & \parameter{testInstancesDir}\\
    code within \parameter{targetRunner} or \parameter{targetEvaluator} & \parameter{execDir}\\
    the value of \parameter{logFile} or \code{--log-file} & \parameter{execDir}\\
    the value of other options in the scenario file  &  the directory containing the scenario file\\
    the value of other command-line options & invocation (working) directory of \irace\\
\bottomrule
  \end{tabular}
\end{table}

\subsection{My parameter space is small enough that \irace could generate all possible configurations; however, \irace generates repeated configurations and/or does not generate some of them. Is this a bug?}\label{faq:allconfs}

Typically, \irace is applied to parameter spaces that are much larger than what
can be explored within the budget given. Thus, \irace does not try to detect
whether all possible configurations can be evaluated for the given budget and
it does not waste computation time to check for repeated configurations. Thus,
if the parameter space is actually very small, the initial random sampling
performed by \irace may generate repeated configurations and/or never generate
some configurations, which is not ideal. If you still want to use
(non-iterated) racing, the recommended approach is to provide all
configurations explicitly to \irace (\autoref{sec:initial}) and execute a
single race (\parameter{nbIterations}\code{=1}) with exactly the number of
configurations provided (e.g., \parameter{nbConfigurations}\code{=240}). A
future version of \irace may automatically detect this case and switch to
non-iterated racing without having to set additional options. Future versions
may also implement computationally cheap checks for repeated
configurations.\footnote{If you are interested in implementing this, please
  contact us!}

\subsection[On Windows and using target-runner.py  (a Python file), I
  get the error target-runner.py is not executable]{On Windows and using \code{target-runner.py}  (a \proglang{Python} file), I
  get the error ``\code{target-runner.py is not executable}''}\label{faq:py-not-exe}

The issue is that \code{.py} files are not executable on their own and you need \code{python.exe} to read the \code{.py} file and execute it. Linux knows how to do this if  the first line of the file is ``\code{\#!/usr/bin/python}'', however, Windows doesn't know how to do it. In Windows you have 2 options:
\begin{itemize}
\item     Create a \code{target-runner.bat} file that contains a line similar to (see \path{templates/windows/target-runner.bat}):
\begin{lstlisting}[style=BashInputStyle]
C:\path\to\python.exe C:\path\to\target-runner.py %instance% %seed% \
  %candidate_parameters% 1>%stdout% 2>%stderr%
\end{lstlisting}

\item Or convert \code{target-runner.py} into an \code{.exe} file, for example, using \code{auto-py-to-exe}\footnote{\url{https://pypi.org/project/auto-py-to-exe/}}, so that you do not need a \code{.bat} file.
\end{itemize}

\subsection[Error in socketConnection(\ldots) : can not open the connection]{Error in \code{socketConnection("localhost", port = port, server = TRUE,} \code{lock = TRUE,} : can not open the connection}

This error may arise if you activate the \parameter{parallel} option of \irace and your \parameter{targetRunner} or \parameter{targetEvaluator} tries to setup a parallel cluster or execute code in parallel in a way that interacts badly with the parallel mechanism in \aR. In this case, you need to either investigate yourself if there is a way for the two parallel mechanisms to co-exist or, if that is not possible, disable parallelism in \irace or in your code. Note that packages or software used by your \parameter{targetRunner} may have a parallel mechanism enabled by default and unknown to you. This is definitely NOT a bug in \irace.

\subsection[irace does not print the call to the target-runner with --debug-level 2 when using the --parallel option]{\irace does not print the call to the \parameter{targetRunner} with \parameter{debugLevel}\code{=2} when using the \parameter{parallel}}

This is a limitation of  Windows or Rstudio. Running without \parameter{parallel} should work. Unfortunately, we cannot fix this limitation in \irace. If you need to understand how \irace calls \parameter{targetRunner} when running in parallel, you can implement a logging mechanism able to handle parallelism directly inside the \parameter{targetRunner}.


%%
%%
%%
%% Contact info
%%
%%
%%


%FIXME: complete this section
%\section{Known problems}

\section{Resources and contact information} \label{sec:contact}

More information about the package can be found on the \irace webpage:
\begin{center} \url{https://iridia.ulb.ac.be/supp/IridiaSupp2016-003/index.html} \end{center}

For questions and suggestions please contact the development team through
the \irace package Google group:
\begin{center}
  \url{https://groups.google.com/d/forum/irace-package}
\end{center}
or by sending an email to:
\begin{center}
  \href{mailto:irace-package@googlegroups.com}{irace-package@googlegroups.com}
\end{center}


\section{Acknowledgements}

We would like to thank all the people that directly or indirectly have
collaborated in the development and improvement of \irace:
%
\newlist{acklist}{itemize*}{1}
\setlist[acklist]{label={}, before=\unskip{}, afterlabel=\unskip{ }, itemjoin={{,}}, itemjoin*={, and }, after={{.}}}
\begin{acklist}
\item Prasanna Balaprakash
\item Zhi (Eric) Yuan
\item Franco Mascia
\item Alberto Franzin
\item Anthony Antoun
\item Esteban Diaz Leiva %MATLAB example
\item Federico Caselli % bug reports with fixes
\item Pablo Valledor Pellicer % testing of --parallel under Windows
\item André de Souza Andrade
\item Nguyen Dang (\texttt{nttd@st-andrews.ac.uk}) % implemented \code{irace2pyimp}
\end{acklist}

\newcommand{\doi}[1]{doi:\hspace{.16667em plus .08333em}\discretionary{}{}{}\href{https://doi.org/#1}{\urlstyle{rm}\nolinkurl{#1}}}
\bibliographystyle{abbrvnat}
\ifthenelse {\boolean{Release}}{%
\bibliography{irace-package}%
}{%
\bibliography{optbib/abbrev,optbib/journals,optbib/authors,optbib/articles,optbib/biblio,optbib/crossref}%
}

\newpage

\begin{appendices}

\section{Installing R} \label{sec:installation}

This section gives a quick \aR installation guide that will work in
most cases. The official instructions are available at
\url{https://cran.r-project.org/doc/manuals/r-release/R-admin.html}

\subsection{GNU/Linux}
You should install \aR from your package manager. On a Debian/Ubuntu system
it will be something like:
\begin{lstlisting}[style=BashInputStyle]
sudo apt-get install r-base
\end{lstlisting}

Once \aR is installed, you can launch \aR from the Terminal and from the \aR prompt
install the \irace package (see \autoref{sec:irace install}).


\subsection{OS X}

You can install \aR directly from a CRAN mirror.\footnote{\url{https://cran.r-project.org/bin/macosx/}} %
Alternatively, if you use
homebrew, you can just brew the \aR formula from the science tap (unfortunately
it does not come already bottled so you need to have Xcode\footnote{Xcode download
webpage: \url{https://developer.apple.com/xcode/download/}} installed to compile it):
%<<R_OS_install,engine='bash',eval=FALSE>>=
\begin{lstlisting}[style=BashInputStyle]
brew tap homebrew/science
brew install r
\end{lstlisting}
%@

Once \aR is installed, you can launch \aR from the Terminal (or from your Applications),
and from the \aR prompt install the \irace package (see \autoref{sec:irace install}).

\subsection{Windows}

You can install \aR from a CRAN mirror.\footnote{\url{https://cran.r-project.org/bin/windows/}} %
We recommend that you
install \aR on a filesystem path without spaces, special characters or long
names, such as \path{C:\R}. Once \aR is installed, you can launch the \aR
console and install the \irace package from it (see \autoref{sec:irace
  install}).


\section{targetRunner troubleshooting checklist} \label{sec:check list}

If the \parameter{targetRunner} script fails to return the output expected by
\irace, it can be sometimes difficult to diagnose where the problem lies. The
more descriptive errors provided by your script, the easier it will be to debug
it. If \parameter{targetRunner} enters an infinite loop, irace will wait
indefinitely (see FAQ in \autoref{faq:timeout}).  If you are using temporary
files to redirect the output of your algorithm, check that these files are
properly created. We recommend to follow the structure of the example file
(\code{target-runner}) provided in \IRACEHOME{/templates}. The following
error examples are based on that example file.

In case of failure of \parameter{targetRunner}, \irace will print an error on
its output describing which execution of \parameter{targetRunner} failed.
%
Follow this checklist to detect where the problem is:

\begin{enumerate}[leftmargin=*,widest=9]

\item Make sure that your \code{targetRunner} script or program is at the specified location. If you see this error:
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' does not exist
\end{CodeInput}
%
it means that \irace cannot find the \code{target-runner} file. Check that the file is at
the path specified by the error.

\item Make sure that your \code{targetRunner} script is an executable file and the user running \irace has permission to execute it. The following errors:
%
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' is a directory,
not a file
\end{CodeInput}
or
\begin{CodeInput}
Error: == irace == target runner './tuning/target-runner' is not executable
\end{CodeInput}
%
mean that your \code{targetRunner} is not an executable file. In the first
case, the script is a folder and therefore there must be a problem with the
name of the script. In the second case, you must make the file executable,
which in GNU/Linux can be done by:
%
\begin{lstlisting}[style=BashInputStyle]
chmod +x ./tuning/target-runner
\end{lstlisting}

\item If your \code{targetRunner} script calls another program, make sure it is at  the location described in the script
(variable \code{EXE} in the examples and templates). A typical output for such an error is:
%
\begin{CodeInput}
Error: == irace == running command ''./tuning/target-runner' 1 8 676651103
./tuning/Instances/1000-16.tsp --ras --localsearch 2 --alpha 4.03 --beta 1.89
--rho  0.02 --ants 37 --nnls 48 --dlb 0 --rasranks 15 2>\&1' had status 1
== irace == The call to target_runner_default was:
./tuning/target-runner 1 8 676651103 ./tuning/Instances/1000-16.tsp --ras
--localsearch 2 --alpha 4.03 --beta 1.89 --rho  0.02 --ants 37 --nnls 48
--dlb 0 --rasranks 15
== irace == The output was:
Tue May  3 19:00:37 UTC 2016: error: ./bin/acotsp: not found or not executable
(pwd: ./tuning/acotsp-arena)
\end{CodeInput}
%
You may test your script by copying the command line shown in the error and
executing \code{target-runner} directly on the execution directory (\parameter{execDir}). In this case, the command line is:
%
\begin{lstlisting}[style=BashInputStyle]
./tuning/target-runner 1 8 676651103 ./tuning/Instances/1000-16.tsp  --ras \
 --localsearch 2 --alpha 4.03 --beta 1.89 --rho  0.02 --ants 37 --nnls 48 \
 --dlb 0 --rasranks 15
\end{lstlisting}

This executes the \code{targetRunner} script as \irace does. The output of this script must be only one number.

\item If your \code{targetRunner} file is an executable script in Python, \aR, Perl, Bash or some other interpreted programming language, make sure that the interpreter specified in the first line of the file exists at the correct location. For example, if the first line of \code{target-runner.py} is:
%
\begin{CodeInput}
#|/usr/bin/python
\end{CodeInput}
%
Then make sure that \path{/usr/bin/python} exists and it is executable. Otherwise, you will get an error such as:
%
\begin{CodeInput}
Error: == irace == error in running command
\end{CodeInput}

\item Check that your \code{targetRunner} script is actually returning one
  number as output. For example:
%
\begin{CodeInput}
Error: == irace == The output of './tuning/target-runner 1 25 365157769
 ./tuning/Instances/1000-31.tsp --ras --localsearch 1 --alpha 0.26 --beta
 6.95 --rho 0.69 --ants 56 --nnls 10 --dlb 0 --rasranks 7' is not numeric!
== irace == The output was:
Solution: 24479793
\end{CodeInput}
% FIXME: Duplicated above.
%% For testing your script, copy the command-line of \code{target-runner} and execute it directly on the execution directory (\parameter{execDir}):
%% %
%% \begin{lstlisting}[style=BashInputStyle]
%% ./tuning/target-runner 1 25 365157769 ./tuning/Instances/1000-31.tsp  --ras \
%%  --localsearch 1 --alpha 0.26 --beta 6.95 --rho  0.69 --ants 56 \
%%  --nnls 10 --dlb 0 --rasranks 7
%% \end{lstlisting}
%% This executes the \code{targetRunner} script as \irace does.
  In the example above, the output of \code{target-runner} is ``\code{Solution:
    24479793}'', which is not a number. If \code{target-runner} is parsing the
  output of the target algorithm, you need to verify that the code only parses
  the solution cost value.

\item Check that your \code{targetRunner} script is creating the output files
  for your algorithm.  If you see an error as:
%
\begin{CodeInput}
== irace == The output was: Tue May  3 19:41:40 UTC 2016:
error: c1-9.stdout: No such file or directory
\end{CodeInput}
The output file of the execution of your algorithm has not been
created (check permissions) or has been deleted before the result can be read.


\item Other errors can produce the following output:
%
\begin{CodeInput}
== irace == The output was: Tue May  3 19:49:06 UTC 2016:
error: c1-23.stdout: Output is not a number
\end{CodeInput}
%
This might be because your \code{targetRunner} script is not executing your
algorithm correctly. To further investigate this issue, comment out the line
that eliminates the temporary files that saves the output of your
algorithm. Similar to this one
%
\begin{lstlisting}[style=BashInputStyle]
rm -f "${STDOUT}" "${STDERR}"
\end{lstlisting}
%
Execute directly the \code{targetRunner} command-line that is provided in the
error message, look in your execution directory for the files that are
created. Check the \code{.stderr} file for errors and the \code{.stdout} file
to see the output that your algorithm produces.

\item Some command within \code{targetRunner} may not be working correctly. In
  that case, you must debug the commands individually exactly as \irace executes them. In
  order to find where the problem is, print the commands to a log file before
  executing them. For example:
  %
\begin{lstlisting}[style=BashInputStyle]
echo "$EXE ${FIXED_PARAMS} -i $INSTANCE ${CONFIG_PARAMS}" >> ${STDERR}.log
$EXE ${FIXED_PARAMS} -i $INSTANCE ${CONFIG_PARAMS} 1> ${STDOUT} 2> ${STDERR}
\end{lstlisting}
%
  then look at the \code{${STDERR}.log} file corresponding to the
  \code{targetRunner} call that failed and execute/debug the last command
  there.

\item If the language of your operating system, the \code{target-runner} or the
  target algorithm is not English, \irace may not be able to recognize the
  numbers generated by \code{target-runner}. We recommend that you run \irace,
  the \code{target-runner} and the target algorithm under an English locale (or
  make sure that their languages and number format are compatible).

\item It is possible that
  \href{https://en.wikipedia.org/wiki/Heisenbug}{transient bugs} in the target
  algorithm are only visible when running within \irace, and
  all commands within \code{targetRunner} appear to work fine when executed directly in the
  command-line outside \irace. See FAQ in \autoref{faq:valgrind}) for suggestions on how to detect such bugs.


\item If your \code{targetRunner} script works when running irace with \parameter{parallel}\code{=0} but it fails when using higher number of cores, this may be due to any number of reasons:
  \begin{itemize}
  \item If you submit jobs through a queuing system, the running environment
    when using the queuing system may not be the same as when you launch \irace
    yourself. The queuing system may also send the job to different machines
    depending on the number of CPUs requested. One way to test this is to
    submit the failing execution of \code{targetRunner} to the queuing system,
    and specifically to any problematic machine.

  \item When using MPI, some calls to \code{targetRunner} may run on different
    computers than the one running the master \irace process. See FAQ in
    \autoref{sec:mpi-error}.

  \item Does \code{targetRunner} read or create intermediate files? These files
    may cause a race condition when two calls to \code{targetRunner} happen at
    the same time. You have to make sure that parallel runs of
    \code{targetRunner} do not interfere with each other's files.

  \item Maybe these files consume too much memory or fill the filesystem when
    there are simultaneous \code{targetRunner} calls? Moreover, queuing systems
    have stricter limits for computing nodes than for the submit/host node.

  \item Does the machine or the queuing system impose any limits on number of
    processes or CPU/memory/filesystem usage per job? Such limits may only
    trigger when more than one process is executed in parallel, killing the
    \code{targetRunner} process before it has a chance to print anything
    useful. In that case, \irace may not detect the the program finished
    unexpectedly, only that the expected output was not printed.
  \end{itemize}

\end{enumerate}

\section{targetEvaluator troubleshooting checklist} \label{sec:evaluator check list}

Even if \parameter{targetRunner} appears to work, the use of \parameter{targetEvaluator} may lead to other problems. The same checklist of \code{targetRunner} can be followed here. In addition, we list here other potential problems unique to \code{targetEvaluator}:

\begin{enumerate}
\item If \parameter{targetEvaluator} fails only in the second or later
  iteration, this may because output files or data generated by a previous call
  to \parameter{targetRunner} are missing. Elite configurations are never
  re-executed on the same instance and seed pair, that is, \irace will call
  only once \parameter{targetRunner} for each pair of configuration ID and
  instance ID. However, \parameter{targetEvaluator} is always re-executed,
  which takes into account any updated information (normalization bounds,
  reference sets/points, best-known values, etc.). Thus, any files or data
  generated by \parameter{targetRunner} for a given configuration must remain
  available to \parameter{targetEvaluator} as long as that configuration is
  alive. The list of alive configurations is passed
  to \parameter{targetEvaluator}, which may decide then which data to keep or
  remove.
\end{enumerate}

\section{Glossary}

\begin{description}
\item[Parameter tuning:] Process of searching good settings for the parameters
  of an algorithm under a particular tuning scenario (instances, execution
  time, etc.).
\item[Scenario:] Settings that define an instance of the tuning problem. These
  settings include the algorithm to be tuned (target), budget for the execution
  of the target algorithm (execution time, evaluations, iterations, etc.), set
  of problem instances and all the information that is required to perform the
  tuning.
\item[Target algorithm:] Algorithm whose parameters will be tuned.
\item[Target parameter:] Parameter of the target algorithm that will be tuned.
\item[\irace option:] Configurable option of \irace.
\item[Elite configurations:] Best configurations found so far by \irace. New
  configurations for the next iteration of \irace are sampled from the
  probabilistic models associated to the elite configurations.  All elite
  configurations are also included in the next iteration.
\item[\hypertarget{irace_home}{\path{$IRACE_HOME}:}] The filesystem path where \irace is
  installed. You can find this information by opening an \aR console and
  executing:

<<R_irace_home2, prompt=FALSE, eval=FALSE>>=
system.file(package = "irace")
@

\end{description}

\section{NEWS}

\RecustomVerbatimCommand{\VerbatimInput}{VerbatimInput}%
{fontsize=\footnotesize,
 %
 frame=lines,  % top and bottom rule only
 framesep=1em, % separation between frame and text
 rulecolor=\color{Gray},
 %
 label=\fbox{\color{Black}NEWS},
 labelposition=topline,
 %
 % commandchars=\|\(\), % escape character and argument delimiters for
 %                      % commands within the verbatim
 % commentchar=*        % comment character
}

%% NEWS is latin1 but this file is utf8
\begingroup\inputencoding{utf8}%
% FIXME: This doesn't break long lines, find another alternative.
\VerbatimInput{../NEWS.md}%
\endgroup
\end{appendices}

\end{document}

%%% Local Variables:
%%% TeX-master: "irace-package.Rnw"
%%% End:

% LocalWords: iteratively

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\docType{data}
\name{irace_license}
\alias{irace_license}
\title{irace_license}
\format{
An object of class \code{character} of length 1.
}
\usage{
irace_license
}
\description{
A character string containing the license information of \pkg{irace}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{read_logfile}
\alias{read_logfile}
\title{Read the log file produced by irace (\code{irace.Rdata}).}
\usage{
read_logfile(filename, name = "iraceResults")
}
\arguments{
\item{filename}{Filename that contains the log file saved by irace. Example: \code{irace.Rdata}.}

\item{name}{Optional argument that allows overriding the default name of the object in the file.}
}
\value{
(\code{list()})
}
\description{
Read the log file produced by irace (\code{irace.Rdata}).
}
\examples{
irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata", package="irace",
                                          mustWork=TRUE))
str(irace_results)
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ablation.R
\name{ablation}
\alias{ablation}
\title{Performs ablation between two configurations (from source to target).}
\usage{
ablation(
  iraceResults,
  src = 1L,
  target = NULL,
  ab_params = NULL,
  type = c("full", "racing"),
  nrep = 1L,
  seed = 1234567L,
  ablationLogFile = "log-ablation.Rdata",
  instancesFile = "train",
  ...
)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{src, target}{\code{integer(1)|character(1)}\cr Source and target configuration IDs. By default, the first configuration ever evaluated (ID 1) is used as \code{src} and the best configuration found by irace is used as target. If the argument is a string, it is interpreted as the path to a file, with the format specified by  \code{\link[=readConfigurationsFile]{readConfigurationsFile()}}, that contains the configuration.}

\item{ab_params}{\code{character()}\cr Specific parameter names to be used for the ablation. They must be in \code{parameters$names}. By default, use all parameters.}

\item{type}{\code{"full"|"racing"}\cr Type of ablation to perform: \code{"full"} will execute each configuration on all \code{n_instances} to determine the best-performing one; \code{"racing"} will apply racing to find the best configurations.}

\item{nrep}{\code{integer(1)}\cr Number of replications per instance used in \code{"full"} ablation. When \code{nrep > 1}, each configuration will be executed \code{nrep} times on each instance with different random seeds.}

\item{seed}{\code{integer(1)}\cr Integer value to use as seed for the random number generation.}

\item{ablationLogFile}{\code{character(1)}\cr Log file to save the ablation log. If \code{NULL}, the results are not saved to a file.}

\item{instancesFile}{\code{character(1)}\cr Instances file used for ablation: \code{'train'}, \code{'test'} or a filename containing the list of instances.}

\item{...}{Further arguments to override scenario settings, e.g., \code{debugLevel}, \code{parallel}, etc.}
}
\value{
A list containing the following elements:
\describe{
\item{allConfigurations}{Configurations tested in the ablation.}
\item{state}{State of the ablation process.}
\item{experiments}{A matrix with the results of the experiments (columns are configurations, rows are instances).}
\item{scenario}{Scenario object with the settings used for the experiments.}
\item{trajectory}{IDs of the best configurations at each step of the ablation.}
\item{best}{Best configuration found in the experiments.}
\item{complete}{\code{TRUE} if the ablation process was completed.}
}
}
\description{
Ablation is a method for analyzing the differences between two configurations.
}
\examples{
\donttest{
logfile <- system.file(package="irace", "exdata", "sann.rda")
# Execute ablation between the first and the best configuration found by irace.
ablog <- ablation(logfile, ablationLogFile = NULL)
plotAblation(ablog)
# Execute ablation between two selected configurations, and selecting only a
# subset of parameters, directly reading the setup from the irace log file.
ablog <- ablation(logfile, src = 1, target = 10,
                  ab_params = c("temp"), ablationLogFile = NULL)
plotAblation(ablog)
}

}
\references{
C. Fawcett and H. H. Hoos. Analysing differences between algorithm
configurations through ablation. Journal of Heuristics, 22(4):431–458, 2016.
}
\seealso{
\code{\link[=plotAblation]{plotAblation()}} \code{\link[=ablation_cmdline]{ablation_cmdline()}}
}
\author{
Leslie Pérez Cáceres and Manuel López-Ibáñez
}
\concept{ablation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readParameters.R
\name{readParameters}
\alias{readParameters}
\title{Reads the parameters to be tuned by \pkg{irace} from a file or from a
character string.}
\usage{
readParameters(file, digits = 4L, debugLevel = 0L, text)
}
\arguments{
\item{file}{\code{character(1)}\cr Filename containing the definitions of
the parameters to be tuned.}

\item{digits}{\code{integer(1)}\cr The number of decimal places to be considered for real-valued parameters.}

\item{debugLevel}{\code{integer(1)}\cr Larger values produce more verbose output.}

\item{text}{\code{character(1)}\cr If \code{file} is not supplied and this is,
then parameters are read from the value of \code{text} via a text connection.}
}
\value{
A list containing the definitions of the parameters read. The list is
structured as follows:
\describe{
\item{\code{names}}{Vector that contains the names of the parameters.}
\item{\code{types}}{Vector that contains the type of each parameter 'i', 'c', 'r', 'o'.
Numerical parameters can be sampled in a log-scale with 'i,log' and 'r,log'
(no spaces).}
\item{\code{switches}}{Vector that contains the switches to be used for the
parameters on the command line.}
\item{\code{domain}}{List of vectors, where each vector may contain two
values (minimum, maximum) for real and integer parameters, or
possibly more for categorical parameters.}
\item{\code{conditions}}{List of R logical expressions, with variables
corresponding to parameter names.}
\item{\code{isFixed}}{Logical vector that specifies which parameter is fixed
and, thus, it does not need to be tuned.}
\item{\code{nbParameters}}{An integer, the total number of parameters.}
\item{\code{nbFixed}}{An integer, the number of parameters with a fixed value.}
\item{\code{nbVariable}}{Number of variable (to be tuned) parameters.}
\item{\code{depends}}{List of character vectors, each vector specifies
which parameters depend on this one.}
\item{\code{is_dependent}}{Logical vector that specifies which parameter has
a dependent domain.}
\item{\code{digits}}{Integer vector that specifies the number of digits per parameter.}
\item{\code{forbidden}}{List of expressions that define which parameter configurations are forbidden.}
}
}
\description{
Reads the parameters to be tuned by \pkg{irace} from a file or from a
character string.
}
\details{
Either \code{file} or \code{text} must be given. If \code{file} is given, the
parameters are read from the file \code{file}. If \code{text} is given instead,
the parameters are read directly from the \code{text} character string.
In both cases, the parameters must be given (in \code{text} or in the file
whose name is \code{file}) in the expected form.  See the documentation
for details.  If none of these parameters is given, \pkg{irace}
will stop with an error.

A fixed parameter is a parameter that should not be sampled but
instead should be always set to the only value of its domain.  In this
function we set \code{isFixed} to TRUE only if the parameter is a categorical
and has only one possible value.  If it is an integer and the minimum
and maximum are equal, or it is a real and the minimum and maximum
values satisfy \code{round(minimum, digits) == round(maximum, digits)},
then the parameter description is rejected as invalid to identify
potential user errors.

The order of the parameters determines the order in which parameters are
given to \code{targetRunner}. Changing the order may also change the results
produced by \code{irace}, even with the same random seed.
}
\examples{
 ## Read the parameters directly from text
 parameters_table <- '
 # name       switch           type  values               [conditions (using R syntax)]
 algorithm    "--"             c     (as,mmas,eas,ras,acs)
 localsearch  "--localsearch " o     (0, 1, 2, 3)
 alpha        "--alpha "       r     (0.00, 5.00)
 beta         "--beta "        r     (0.00, 10.00)
 rho          "--rho  "        r     (0.01, 1.00)
 ants         "--ants "        i,log (5, 100)
 q0           "--q0 "          r     (0.0, 1.0)           | algorithm == "acs"
 rasrank      "--rasranks "    i     (1, "min(ants, 10)") | algorithm == "ras"
 elitistants  "--elitistants " i     (1, ants)            | algorithm == "eas"
 nnls         "--nnls "        i     (5, 50)              | localsearch \%in\% c(1,2,3)
 dlb          "--dlb "         c     (0, 1)               | localsearch \%in\% c(1,2,3)

 [forbidden]
 (alpha == 0.0) & (beta == 0.0)
 [global]
 digits = 4
 '
 parameters <- readParameters(text=parameters_table)
 str(parameters)

}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scenario.R
\name{checkScenario}
\alias{checkScenario}
\title{Check and correct the given scenario}
\usage{
checkScenario(scenario = defaultScenario())
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
The scenario received as a parameter, possibly corrected. Unset
scenario settings are set to their default values.
}
\description{
Checks for errors a (possibly incomplete) scenario setup of
\pkg{irace}  and transforms it into a valid scenario.
}
\details{
This function checks that the directories and the file names
provided and required by the \pkg{irace} exist. It also checks that the
settings are of the proper type, e.g. that settings expected to be integers
are really integers. Finally, it also checks that there is no inconsistency
between settings.  If an error is found that prevents \pkg{irace} from
running properly, it will stop with an error.
}
\seealso{
\describe{
\item{\code{\link[=readScenario]{readScenario()}}}{for reading a configuration scenario from a file.}
\item{\code{\link[=printScenario]{printScenario()}}}{prints the given scenario.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link[=checkScenario]{checkScenario()}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ablation.R
\name{ablation_cmdline}
\alias{ablation_cmdline}
\title{Launch ablation with command-line options.}
\usage{
ablation_cmdline(argv = commandArgs(trailingOnly = TRUE))
}
\arguments{
\item{argv}{\code{character()}\cr The arguments
provided on the R command line as a character vector, e.g.,
\code{c("-i", "irace.Rdata", "--src", 1)}.}
}
\value{
A list containing the following elements:
\describe{
\item{allConfigurations}{Configurations tested in the ablation.}
\item{state}{State of the ablation process.}
\item{experiments}{A matrix with the results of the experiments (columns are configurations, rows are instances).}
\item{scenario}{Scenario object with the settings used for the experiments.}
\item{trajectory}{IDs of the best configurations at each step of the ablation.}
\item{best}{Best configuration found in the experiments.}
\item{complete}{\code{TRUE} if the ablation process was completed.}
}
}
\description{
Launch \code{\link[=ablation]{ablation()}} with the same command-line options as the command-line
executable (\code{ablation.exe} in Windows).
}
\details{
The function reads the parameters given on the command line
used to invoke R, launches \code{\link[=ablation]{ablation()}} and possibly \code{\link[=plotAblation]{plotAblation()}}.

List of command-line options:

\if{html}{\out{<div class="sourceCode">}}\preformatted{-l,--log-file            Path to the (.Rdata) file created by irace from which
                         the "iraceResults" object will be loaded.
-S,--src                 Source configuration ID or the path to a file
                         containing the configuration. Default: 1.
-T,--target              Target configuration ID (by default the best
                         configuration found by irace) or the path to a file
                         containing the configuration.
-P,--params              Specific parameter names to be used for the ablation
                         (separated with commas). By default use all
-t,--type                Type of ablation to perform: "full" will execute each
                         configuration on all "--n-instances" to determine the
                         best-performing one; "racing" will apply racing to
                         find the best configurations. Default: full.
-n,--nrep                Number of replications per instance used in "full"
                         ablation. Default: 1.
   --seed                Integer value to use as seed for the random number
                         generation. Default: 1234567.
-o,--output-file         Log file to save the ablation log. If "", the results
                         are not saved to a file. Default: log-ablation.Rdata.
   --instances-file      Instances file used for ablation: "train", "test" or a
                         filename containing the list of instances. Default:
                         train.
-p,--plot                Output filename (.pdf) for the plot. If not given, no
                         plot is created.
-O,--plot-type           Type of plot. Supported values are "mean", "boxplot",
                         "rank" or "rank,boxplot". Default: mean.
   --old-path            Old path found in the log-file (.Rdata) given as input
                         to be replaced by --new-path.
   --new-path            New path to replace the path found in the log-file
                         (.Rdata) given as input.
-e,--exec-dir            Directory where the target runner will be run.
-s,--scenario            Scenario file to override the scenario given in the
                         log-file (.Rdata)
   --parallel            Number of calls to targetRunner to execute in
                         parallel. Values 0 or 1 mean no parallelization.
}\if{html}{\out{</div>}}
}
\examples{
ablation_cmdline("--help")
# Find the ablation command-line executable:
Sys.glob(file.path(system.file(package="irace", "bin"), "ablation*"))
}
\seealso{
\code{\link[=plotAblation]{plotAblation()}} \code{\link[=ablation]{ablation()}}
}
\author{
Manuel López-Ibáñez
}
\concept{ablation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{has_testing_data}
\alias{has_testing_data}
\title{Check if the results object generated by irace has data about the testing phase.}
\usage{
has_testing_data(iraceResults)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}
}
\value{
\code{logical(1)}
}
\description{
Check if the results object generated by irace has data about the testing phase.
}
\examples{
irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata", package="irace",
                                          mustWork=TRUE))
print(has_testing_data(irace_results))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readParameters.R
\name{read_pcs_file}
\alias{read_pcs_file}
\title{Read parameters in PCS (AClib) format and write them in irace format.}
\usage{
read_pcs_file(file, digits = 4L, debugLevel = 0L, text)
}
\arguments{
\item{file}{\code{character(1)}\cr Filename containing the definitions of
the parameters to be tuned.}

\item{digits}{\code{integer(1)}\cr The number of decimal places to be considered for real-valued parameters.}

\item{debugLevel}{\code{integer(1)}\cr Larger values produce more verbose output.}

\item{text}{\code{character(1)}\cr If \code{file} is not supplied and this is,
then parameters are read from the value of \code{text} via a text connection.}
}
\value{
A string representing the parameters in irace format.
}
\description{
Read parameters in PCS (AClib) format and write them in irace format.
}
\details{
Either \code{file} or \code{text} must be given. If \code{file} is given, the
parameters are read from the file \code{file}. If \code{text} is given instead,
the parameters are read directly from the \code{text} character string.
In both cases, the parameters must be given (in \code{text} or in the file
whose name is \code{file}) in the expected form.  See the documentation
for details.  If none of these parameters is given, \pkg{irace}
will stop with an error.

\strong{FIXME:} Multiple conditions and default configuration are currently ignored. See \url{https://github.com/MLopez-Ibanez/irace/issues/31}
}
\examples{
 ## Read the parameters directly from text
 pcs_table <- '
 # name       domain
 algorithm    {as,mmas,eas,ras,acs}[as]
 localsearch  {0, 1, 2, 3}[0]
 alpha        [0.00, 5.00][1]
 beta         [0.00, 10.00][1]
 rho          [0.01, 1.00][0.95]
 ants         [1, 100][10]il
 q0           [0.0, 1.0][0]
 rasrank      [1, 100][1]i
 elitistants  [1, 750][1]i
 nnls         [5, 50][5]i
 dlb          {0, 1}[1]
 Conditionals:
 q0 | algorithm in {acs}
 rasrank | algorithm in {ras}
 elitistants | algorithm in {eas}
 nnls | localsearch in {1,2,3}
 dlb | localsearch in {1,2,3}
 {alpha=0, beta=0}'
 parameters_table <- read_pcs_file(text=pcs_table)
 cat(parameters_table)
 parameters <- readParameters(text=parameters_table)
 str(parameters)

}
\references{
Frank Hutter, Manuel López-Ibáñez, Chris Fawcett, Marius Thomas Lindauer, Holger H. Hoos, Kevin Leyton-Brown, and Thomas Stützle. \strong{AClib: A Benchmark Library for Algorithm Configuration}. In P. M. Pardalos, M. G. C. Resende, C. Vogiatzis, and J. L. Walteros, editors, \emph{Learning and Intelligent Optimization, 8th International Conference, LION 8}, volume 8426 of Lecture Notes in Computer Science, pages 36–40. Springer, Heidelberg, 2014.
}
\seealso{
\code{\link[=readParameters]{readParameters()}}
}
\author{
Manuel López-Ibáñez
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameterAnalysis.R
\name{get_instanceID_seed_pairs}
\alias{get_instanceID_seed_pairs}
\title{Returns the pairs of instance IDs and seeds used as instances in the race
(and optionally the actual instances).}
\usage{
get_instanceID_seed_pairs(iraceResults, index, instances = FALSE)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{index}{\code{integer()}\cr Indexes of the (instanceID,seed) pairs to be returned. The default returns everything.}

\item{instances}{\code{logical(1)}\cr Whether to add the actual instances as an additional column (only if the instances are of atomic type).}
}
\value{
\code{data.table()}\cr With default arguments, a \code{data.table} containing two columns
\code{"instanceID"} and \code{"seed"}. With \code{instances=TRUE} and if the instances
are of atomic type (see \code{\link[=is.atomic]{is.atomic()}}) type, another column \code{instance} is
added that contains the actual instance.
}
\description{
Returns the pairs of instance IDs and seeds used as instances in the race
(and optionally the actual instances).
}
\examples{
log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
head(get_instanceID_seed_pairs(log_file))
# Add the instance names
get_instanceID_seed_pairs(log_file, index=1:10, instances=TRUE)
}
\author{
Manuel López-Ibáñez
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{irace.cmdline}
\alias{irace.cmdline}
\title{Launch \code{irace} with command-line options.}
\usage{
irace.cmdline(argv = commandArgs(trailingOnly = TRUE))
}
\arguments{
\item{argv}{(\code{character()}) \cr The arguments
provided on the R command line as a character vector, e.g.,
\code{c("--scenario", "scenario.txt", "-p", "parameters.txt")}.
Using the  default value (not providing the parameter) is the
easiest way to call \code{irace.cmdline}.}
}
\value{
(\code{invisible(data.frame)})

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:
\itemize{
\item \code{.ID.} : Internal id of the candidate configuration.
\item \verb{Parameter names} : One column per parameter name in \code{parameters}.
\item \code{.PARENT.} : Internal id of the parent candidate configuration.
}

Additionally, this function saves an R data file containing an object called
\code{iraceResults}. The path of the file is indicated in \code{scenario$logFile}.
The \code{iraceResults} object is a list with the following structure:

\describe{

\item{\code{scenario}}{The scenario R object containing the \pkg{irace}
options used for the execution. See \code{\link{defaultScenario}}
for more information.}

\item{\code{parameters}}{The parameters R object containing the
description of the target algorithm parameters. See
\code{\link{readParameters}}.}

\item{\code{allConfigurations}}{The target algorithm configurations
generated by \pkg{irace}. This object is a data frame, each row is a
candidate configuration, the first column (\code{.ID.}) indicates the
internal identifier of the configuration, the following columns
correspond to the parameter values, each column named as the parameter
name specified in the parameter object. The final column
(\code{.PARENT.})  is the identifier of the configuration from which
model the actual configuration was sampled.}

\item{\code{allElites}}{A list that contains one element per iteration,
each element contains the internal identifier of the elite candidate
configurations of the corresponding iteration (identifiers correspond to
\code{allConfigurations$.ID.}).}

\item{\code{iterationElites}}{A vector containing the best candidate
configuration internal identifier of each iteration. The best
configuration found corresponds to the last one of this vector.}

\item{\code{experiments}}{A matrix with configurations as columns and
instances as rows. Column names correspond to the internal identifier of
the configuration (\code{allConfigurations$.ID.}).}

\item{\code{experimentLog}}{A matrix with columns \code{iteration},
\code{instance}, \code{configuration}, \code{time}.  This matrix contains the log of all the
experiments that \pkg{irace} performs during its execution.  The
instance column refers to the index of the \code{scenario$instancesList}
data frame. Time is saved ONLY when reported by the \code{targetRunner}.}

\item{\code{softRestart}}{A logical vector that indicates if a soft
restart was performed on each iteration. If \code{FALSE}, then no soft
restart was performed.}

\item{\code{state}}{A list that contains the state of \pkg{irace}, the
recovery is done using the information contained in this object.}

\item{\code{testing}}{A list that contains the testing results. The
elements of this list are: \code{experiments} a matrix with the testing
experiments of the selected configurations in the same format as the
explained above and \code{seeds} a vector with the seeds used to execute
each experiment.}

}
}
\description{
Calls \code{\link[=irace.main]{irace.main()}} using command-line options, maybe parsed from the
command line used to invoke R.
}
\details{
The function reads the parameters given on the command line
used to invoke R, finds the name of the scenario file,
initializes the scenario from the file (with the function
\code{\link{readScenario}}) and possibly from parameters passed in
the command line. It finally starts \pkg{irace} by calling
\code{\link{irace.main}}.

List of command-line options:\preformatted{-h,--help                Show this help.
-v,--version             Show irace package version.
-c,--check               Check scenario.
-i,--init                Initialize the working directory with template config
                         files.
   --only-test           Only test the configurations given in the file passed
                         as argument.
-s,--scenario            File that describes the configuration scenario setup
                         and other irace settings. Default: ./scenario.txt.
   --exec-dir            Directory where the programs will be run. Default: ./.
-p,--parameter-file      File that contains the description of the parameters
                         of the target algorithm. Default: ./parameters.txt.
   --forbidden-file      File that contains a list of logical expressions that
                         cannot be TRUE for any evaluated configuration. If
                         empty or NULL, do not use forbidden expressions.
   --configurations-file  File that contains a table of initial configurations.
                         If empty or NULL, all initial configurations are
                         randomly generated.
-l,--log-file            File to save tuning results as an R dataset, either
                         absolute path or relative to execDir. Default:
                         ./irace.Rdata.
   --recovery-file       Previously saved log file to recover the execution of
                         irace, either absolute path or relative to the current
                         directory.  If empty or NULL, recovery is not
                         performed.
   --train-instances-dir  Directory where training instances are located;
                         either absolute path or relative to current directory.
                         If no trainInstancesFiles is provided, all the files
                         in trainInstancesDir will be listed as instances.
                         Default: ./Instances.
   --train-instances-file  File that contains a list of training instances and
                         optionally additional parameters for them. If
                         trainInstancesDir is provided, irace will search for
                         the files in this folder.
   --sample-instances    Randomly sample the training instances or use them in
                         the order given. Default: 1.
   --test-instances-dir  Directory where testing instances are located, either
                         absolute or relative to current directory.
   --test-instances-file  File containing a list of test instances and
                         optionally additional parameters for them.
   --test-num-elites     Number of elite configurations returned by irace that
                         will be tested if test instances are provided.
                         Default: 1.
   --test-iteration-elites  Enable/disable testing the elite configurations
                         found at each iteration. Default: 0.
   --test-type           Statistical test used for elimination. The default
                         value selects t-test if capping is enabled or F-test,
                         otherwise. Valid values are: F-test (Friedman test),
                         t-test (pairwise t-tests with no correction),
                         t-test-bonferroni (t-test with Bonferroni's correction
                         for multiple comparisons), t-test-holm (t-test with
                         Holm's correction for multiple comparisons).
   --first-test          Number of instances evaluated before the first
                         elimination test. It must be a multiple of eachTest.
                         Default: 5.
   --each-test           Number of instances evaluated between elimination
                         tests. Default: 1.
   --target-runner       Executable called for each configuration that executes
                         the target algorithm to be tuned. See the templates
                         and examples provided. Default: ./target-runner.
   --target-runner-launcher  Executable that will be used to launch the target
                         runner, when targetRunner cannot be executed directly
                         (.e.g, a Python script in Windows).
   --target-runner-args  Command-line arguments provided to
                         targetRunnerLauncher. The substrings \\\{targetRunner\\\}
                         and \\\{targetRunnerArgs\\\} will be replaced by the value
                         of the option targetRunner and by the arguments
                         usually passed when calling targetRunner,
                         respectively. Example: "-m \{targetRunner --args
                         \{targetRunnerArgs\}"\}. Default: \{targetRunner\}
                         \{targetRunnerArgs\}.
   --target-runner-retries  Number of times to retry a call to targetRunner if
                         the call failed. Default: 0.
   --target-evaluator    Optional script or R function that provides a numeric
                         value for each configuration. See
                         templates/target-evaluator.tmpl
   --deterministic       If the target algorithm is deterministic,
                         configurations will be evaluated only once per
                         instance. Default: 0.
   --max-experiments     Maximum number of runs (invocations of targetRunner)
                         that will be performed. It determines the maximum
                         budget of experiments for the tuning. Default: 0.
   --max-time            Maximum total execution time in seconds for the
                         executions of targetRunner. targetRunner must return
                         two values: cost and time. Default: 0.
   --budget-estimation   Fraction (smaller than 1) of the budget used to
                         estimate the mean computation time of a configuration.
                         Only used when maxTime > 0 Default: 0.02.
   --min-measurable-time  Minimum time unit that is still (significantly)
                         measureable. Default: 0.01.
   --parallel            Number of calls to targetRunner to execute in
                         parallel. Values 0 or 1 mean no parallelization.
                         Default: 0.
   --load-balancing      Enable/disable load-balancing when executing
                         experiments in parallel. Load-balancing makes better
                         use of computing resources, but increases
                         communication overhead. If this overhead is large,
                         disabling load-balancing may be faster. Default: 1.
   --mpi                 Enable/disable MPI. Use Rmpi to execute targetRunner
                         in parallel (parameter parallel is the number of
                         slaves). Default: 0.
   --batchmode           Specify how irace waits for jobs to finish when
                         targetRunner submits jobs to a batch cluster: sge,
                         pbs, torque, slurm or htcondor. targetRunner must
                         submit jobs to the cluster using, for example, qsub.
                         Default: 0.
   --digits              Maximum number of decimal places that are significant
                         for numerical (real) parameters. Default: 4.
-q,--quiet               Reduce the output generated by irace to a minimum.
                         Default: 0.
   --debug-level         Debug level of the output of irace. Set this to 0 to
                         silence all debug messages. Higher values provide more
                         verbose debug messages. Default: 0.
   --seed                Seed of the random number generator (by default,
                         generate a random seed).
   --soft-restart        Enable/disable the soft restart strategy that avoids
                         premature convergence of the probabilistic model.
                         Default: 1.
   --soft-restart-threshold  Soft restart threshold value for numerical
                         parameters. If NA, NULL or "", it is computed as
                         10^-digits.
-e,--elitist             Enable/disable elitist irace. Default: 1.
   --elitist-new-instances  Number of instances added to the execution list
                         before previous instances in elitist irace. Default:
                         1.
   --elitist-limit       In elitist irace, maximum number per race of
                         elimination tests that do not eliminate a
                         configuration. Use 0 for no limit. Default: 2.
   --capping             Enable the use of adaptive capping, a technique
                         designed for minimizing the computation time of
                         configurations. This is only available when elitist is
                         active. Default: 0.
   --capping-type        Measure used to obtain the execution bound from the
                         performance of the elite configurations: median, mean,
                         worst, best. Default: median.
   --bound-type          Method to calculate the mean performance of elite
                         configurations: candidate or instance. Default:
                         candidate.
   --bound-max           Maximum execution bound for targetRunner. It must be
                         specified when capping is enabled. Default: 0.
   --bound-digits        Precision used for calculating the execution time. It
                         must be specified when capping is enabled. Default: 0.
   --bound-par           Penalization constant for timed out executions
                         (executions that reach boundMax execution time).
                         Default: 1.
   --bound-as-timeout    Replace the configuration cost of bounded executions
                         with boundMax. Default: 1.
   --postselection       Percentage of the configuration budget used to perform
                         a postselection race of the best configurations of
                         each iteration after the execution of irace. Default:
                         0.
   --aclib               Enable/disable AClib mode. This option enables
                         compatibility with GenericWrapper4AC as targetRunner
                         script. Default: 0.
   --iterations          Maximum number of iterations. Default: 0.
   --experiments-per-iteration  Number of runs of the target algorithm per
                         iteration. Default: 0.
   --min-survival        Minimum number of configurations needed to continue
                         the execution of each race (iteration). Default: 0.
   --num-configurations  Number of configurations to be sampled and evaluated
                         at each iteration. Default: 0.
   --mu                  Parameter used to define the number of configurations
                         sampled and evaluated at each iteration. Default: 5.
   --confidence          Confidence level for the elimination test. Default:
                         0.95.
}
}
\examples{
irace.cmdline("--version")
}
\seealso{
\code{\link[=irace.main]{irace.main()}} to start \pkg{irace} with a given scenario.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scenario.R
\name{readScenario}
\alias{readScenario}
\title{Reads from a file the scenario settings to be used by \pkg{irace}.}
\usage{
readScenario(filename = "", scenario = list(), params_def = .irace.params.def)
}
\arguments{
\item{filename}{\code{character(1)}\cr Filename from which the scenario will
be read. If empty, the default \code{scenarioFile} is used.  An example
scenario file is provided in \code{system.file(package="irace", "templates/scenario.txt.tmpl")}.}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{params_def}{\code{data.frame()}\cr Definition of the options accepted by
the scenario. This should only be modified by packages that wish to extend
\pkg{irace}.}
}
\value{
The scenario list read from the file. The scenario settings not
present in the file are not present in the list, i.e., they are \code{NULL}.
}
\description{
The scenario argument is an initial scenario that is overwritten for every
setting specified in the file to be read.
}
\seealso{
\describe{
\item{\code{\link[=printScenario]{printScenario()}}}{prints the given scenario.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link[=checkScenario]{checkScenario()}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/multi_irace.R
\name{multi_irace}
\alias{multi_irace}
\title{Execute \code{\link[=irace]{irace()}} multiple times with the same or different scenarios and parameter space definitions.}
\usage{
multi_irace(
  scenarios,
  parameters,
  n = 1L,
  parallel = 1L,
  split_output = parallel > 1L,
  global_seed = NULL
)
}
\arguments{
\item{scenarios}{\code{list()}\cr A list of scenarios.
If only a single scenario is supplied, it is used for all parameters.}

\item{parameters}{\code{list()}\cr A list of parameter space definitions.
If only a single definition is supplied, it is used for all scenarios.}

\item{n}{\code{integer(1)}\cr The number of repetitions.}

\item{parallel}{\code{integer(1)}\cr The number of workers to use.
A value of \code{1} means sequential execution. Note that \code{parallel > 1} is not supported on Windows.}

\item{split_output}{\code{logical(1)}\cr If \code{TRUE}, the output of \code{\link[=irace]{irace()}} is written to \verb{\{execDir\}/run_\{i\}/irace.out}
instead of the standard output.}

\item{global_seed}{\code{integer(1)}\cr The global seed used to seed the individual runs.}
}
\value{
A list of the outputs of \code{\link[=irace]{irace()}}.
}
\description{
There are three modes of operation:
\itemize{
\item One \code{scenarios} and \code{k} \code{parameters}: \code{k} runs with the same scenario and each parameter space definition.
\item One \code{parameters} and \code{k} \code{scenarios}: \code{k} runs with the same parameter space definition and each scenario.
\item \code{k} \code{parameters} and \code{k} scenarios: \code{k} runs with each scenario and parameter space definition.
}
Each of the \code{k} runs can be repeated \code{n} times by supplying a value for \code{n}.
}
\seealso{
\describe{
\item{\code{\link[=irace]{irace()}}}{the main interface for single irace runs.}
}
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path_rel2abs.R
\name{path_rel2abs}
\alias{path_rel2abs}
\title{Converts a relative path to an absolute path.}
\usage{
path_rel2abs(path, cwd = getwd())
}
\arguments{
\item{path}{(\code{character(1)}) Character string representing a relative path.}

\item{cwd}{(\code{character(1)}) Current working directory.}
}
\value{
(\code{character(1)}) Character string representing the absolute path
}
\description{
If the path passed corresponds to an executable, it tries to find its path
using \code{\link[=Sys.which]{Sys.which()}}.  Expansion of \code{'~'} in Windows follows the definition
of \code{\link[fs:path_expand]{fs::path_expand()}} rather than \code{\link[base:path.expand]{base::path.expand()}}. This function
tries really hard to create canonical paths.
}
\examples{
path_rel2abs("..")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/irace.R
\name{irace}
\alias{irace}
\title{Execute one run of the Iterated Racing algorithm.}
\usage{
irace(scenario)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
(\code{data.frame})

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:
\itemize{
\item \code{.ID.} : Internal id of the candidate configuration.
\item \verb{Parameter names} : One column per parameter name in \code{parameters}.
\item \code{.PARENT.} : Internal id of the parent candidate configuration.
}

Additionally, this function saves an R data file containing an object called
\code{iraceResults}. The path of the file is indicated in \code{scenario$logFile}.
The \code{iraceResults} object is a list with the following structure:

\describe{

\item{\code{scenario}}{The scenario R object containing the \pkg{irace}
options used for the execution. See \code{\link{defaultScenario}}
for more information. The element \code{scenario$parameters} contains the parameters R object that describes the target algorithm parameters. See
\code{\link{readParameters}}.}

\item{\code{allConfigurations}}{The target algorithm configurations
generated by \pkg{irace}. This object is a data frame, each row is a
candidate configuration, the first column (\code{.ID.}) indicates the
internal identifier of the configuration, the following columns
correspond to the parameter values, each column named as the parameter
name specified in the parameter object. The final column
(\code{.PARENT.})  is the identifier of the configuration from which
model the actual configuration was sampled.}

\item{\code{allElites}}{A list that contains one element per iteration,
each element contains the internal identifier of the elite candidate
configurations of the corresponding iteration (identifiers correspond to
\code{allConfigurations$.ID.}).}

\item{\code{iterationElites}}{A vector containing the best candidate
configuration internal identifier of each iteration. The best
configuration found corresponds to the last one of this vector.}

\item{\code{experiments}}{A matrix with configurations as columns and
instances as rows. Column names correspond to the internal identifier of
the configuration (\code{allConfigurations$.ID.}).}

\item{\code{experimen_log}}{A \code{data.table} with columns \code{iteration},
\code{instance}, \code{configuration}, \code{time}.  This matrix contains the log of all the
experiments that \pkg{irace} performs during its execution.  The
instance column refers to the index of the \code{race_state$instances_log}
data frame. Time is saved ONLY when reported by the \code{targetRunner}.}

\item{\code{softRestart}}{A logical vector that indicates if a soft
restart was performed on each iteration. If \code{FALSE}, then no soft
restart was performed.}

\item{\code{state}}{An environment that contains the state of \pkg{irace}, the
recovery is done using the information contained in this object.}

\item{\code{testing}}{A list that contains the testing results. The
elements of this list are: \code{experiments} a matrix with the testing
experiments of the selected configurations in the same format as the
explained above and \code{seeds} a vector with the seeds used to execute
each experiment.}

}
}
\description{
The function \code{irace} implements the Iterated Racing procedure for parameter
tuning. It receives a configuration scenario and a parameter space to be
tuned, and returns the best configurations found, namely, the elite
configurations obtained from the last iterations. As a first step, it checks
the correctness of \code{scenario} using \code{\link[=checkScenario]{checkScenario()}} and recovers a
previous execution if \code{scenario$recoveryFile} is set. A R data file log of
the execution is created in \code{scenario$logFile}.
}
\details{
The execution of this function is reproducible under some conditions. See
the FAQ section in the \href{https://cran.r-project.org/package=irace/vignettes/irace-package.pdf}{User Guide}.
}
\examples{
\dontrun{
# In general, there are three steps:
scenario <- readScenario(filename = "scenario.txt")
irace(scenario = scenario)
}
#######################################################################
# This example illustrates how to tune the parameters of the simulated
# annealing algorithm (SANN) provided by the optim() function in the
# R base package.  The goal in this example is to optimize instances of
# the following family:
#      f(x) = lambda * f_rastrigin(x) + (1 - lambda) * f_rosenbrock(x)
# where lambda follows a normal distribution whose mean is 0.9 and
# standard deviation is 0.02. f_rastrigin and f_rosenbrock are the
# well-known Rastrigin and Rosenbrock benchmark functions (taken from
# the cmaes package). In this scenario, different instances are given
# by different values of lambda.
#######################################################################
## First we provide an implementation of the functions to be optimized:
f_rosenbrock <- function (x) {
  d <- length(x)
  z <- x + 1
  hz <- z[1L:(d - 1L)]
  tz <- z[2L:d]
  sum(100 * (hz^2 - tz)^2 + (hz - 1)^2)
}
f_rastrigin <- function (x) {
  sum(x * x - 10 * cos(2 * pi * x) + 10)
}

## We generate 20 instances (in this case, weights):
weights <- rnorm(20, mean = 0.9, sd = 0.02)

## On this set of instances, we are interested in optimizing two
## parameters of the SANN algorithm: tmax and temp. We setup the
## parameter space as follows:
parameters_table <- '
  tmax "" i,log (1, 5000)
  temp "" r (0, 100)
  '
## We use the irace function readParameters to read this table:
parameters <- readParameters(text = parameters_table)

## Next, we define the function that will evaluate each candidate
## configuration on a single instance. For simplicity, we restrict to
## three-dimensional functions and we set the maximum number of
## iterations of SANN to 1000.
target_runner <- function(experiment, scenario)
{
    instance <- experiment$instance
    configuration <- experiment$configuration

    D <- 3
    par <- runif(D, min=-1, max=1)
    fn <- function(x) {
      weight <- instance
      return(weight * f_rastrigin(x) + (1 - weight) * f_rosenbrock(x))
    }
    # For reproducible results, we should use the random seed given by
    # experiment$seed to set the random seed of the target algorithm.
    res <- withr::with_seed(experiment$seed,
                     stats::optim(par,fn, method="SANN",
                                  control=list(maxit=1000
                                             , tmax = as.numeric(configuration[["tmax"]])
                                             , temp = as.numeric(configuration[["temp"]])
                                               )))
    ## This list may also contain:
    ## - 'time' if irace is called with 'maxTime'
    ## - 'error' is a string used to report an error
    ## - 'outputRaw' is a string used to report the raw output of calls to
    ##   an external program or function.
    ## - 'call' is a string used to report how target_runner called the
    ##   external program or function.
    return(list(cost = res$value))
}

## We define a configuration scenario by setting targetRunner to the
## function define above, instances to the first 10 random weights, and
## a maximum budget of 'maxExperiments' calls to targetRunner.
scenario <- list(targetRunner = target_runner,
                 instances = weights[1:10],
                 maxExperiments = 500,
                 # Do not create a logFile
                 logFile = "",
                 parameters = parameters)

## We check that the scenario is valid. This will also try to execute
## target_runner.
checkIraceScenario(scenario)

\donttest{
## We are now ready to launch irace. We do it by means of the irace
## function. The function will print information about its
## progress. This may require a few minutes, so it is not run by default.
tuned_confs <- irace(scenario = scenario)

## We can print the best configurations found by irace as follows:
configurations_print(tuned_confs)

## We can evaluate the quality of the best configuration found by
## irace versus the default configuration of the SANN algorithm on
## the other 10 instances previously generated.
test_index <- 11:20
test_seeds <- sample.int(2147483647L, size = length(test_index), replace = TRUE)
test <- function(configuration)
{
  res <- lapply(seq_along(test_index),
                function(x) target_runner(
                              experiment = list(instance = weights[test_index[x]],
                                                seed = test_seeds[x],
                                                configuration = configuration),
                              scenario = scenario))
  return (sapply(res, getElement, name = "cost"))
}
## To do so, first we apply the default configuration of the SANN
## algorithm to these instances:
default <- test(data.frame(tmax=10, temp=10))

## We extract and apply the winning configuration found by irace
## to these instances:
tuned <- test(removeConfigurationsMetaData(tuned_confs[1,]))

## Finally, we can compare using a boxplot the quality obtained with the
## default parametrization of SANN and the quality obtained with the
## best configuration found by irace.
boxplot(list(default = default, tuned = tuned))
}
}
\seealso{
\describe{
\item{\code{\link[=irace_main]{irace_main()}}}{a higher-level interface to \code{\link[=irace]{irace()}}.}
\item{\code{\link[=irace_cmdline]{irace_cmdline()}}}{a command-line interface to \code{\link[=irace]{irace()}}.}
\item{\code{\link[=readScenario]{readScenario()}}}{for reading a configuration scenario from a file.}
\item{\code{\link[=readParameters]{readParameters()}}}{read the target algorithm parameters from a file.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link[=checkScenario]{checkScenario()}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{irace.main}
\alias{irace.main}
\title{Higher-level interface to launch irace.}
\usage{
irace.main(scenario, output.width = 9999L)
}
\arguments{
\item{scenario}{(\code{list()}) \cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{output.width}{(\code{integer(1)}) The width used for the screen
output.}
}
\value{
(\code{invisible(data.frame)})

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:
\itemize{
\item \code{.ID.} : Internal id of the candidate configuration.
\item \verb{Parameter names} : One column per parameter name in \code{parameters}.
\item \code{.PARENT.} : Internal id of the parent candidate configuration.
}

Additionally, this function saves an R data file containing an object called
\code{iraceResults}. The path of the file is indicated in \code{scenario$logFile}.
The \code{iraceResults} object is a list with the following structure:

\describe{

\item{\code{scenario}}{The scenario R object containing the \pkg{irace}
options used for the execution. See \code{\link{defaultScenario}}
for more information.}

\item{\code{parameters}}{The parameters R object containing the
description of the target algorithm parameters. See
\code{\link{readParameters}}.}

\item{\code{allConfigurations}}{The target algorithm configurations
generated by \pkg{irace}. This object is a data frame, each row is a
candidate configuration, the first column (\code{.ID.}) indicates the
internal identifier of the configuration, the following columns
correspond to the parameter values, each column named as the parameter
name specified in the parameter object. The final column
(\code{.PARENT.})  is the identifier of the configuration from which
model the actual configuration was sampled.}

\item{\code{allElites}}{A list that contains one element per iteration,
each element contains the internal identifier of the elite candidate
configurations of the corresponding iteration (identifiers correspond to
\code{allConfigurations$.ID.}).}

\item{\code{iterationElites}}{A vector containing the best candidate
configuration internal identifier of each iteration. The best
configuration found corresponds to the last one of this vector.}

\item{\code{experiments}}{A matrix with configurations as columns and
instances as rows. Column names correspond to the internal identifier of
the configuration (\code{allConfigurations$.ID.}).}

\item{\code{experimentLog}}{A matrix with columns \code{iteration},
\code{instance}, \code{configuration}, \code{time}.  This matrix contains the log of all the
experiments that \pkg{irace} performs during its execution.  The
instance column refers to the index of the \code{scenario$instancesList}
data frame. Time is saved ONLY when reported by the \code{targetRunner}.}

\item{\code{softRestart}}{A logical vector that indicates if a soft
restart was performed on each iteration. If \code{FALSE}, then no soft
restart was performed.}

\item{\code{state}}{A list that contains the state of \pkg{irace}, the
recovery is done using the information contained in this object.}

\item{\code{testing}}{A list that contains the testing results. The
elements of this list are: \code{experiments} a matrix with the testing
experiments of the selected configurations in the same format as the
explained above and \code{seeds} a vector with the seeds used to execute
each experiment.}

}
}
\description{
Higher-level interface to launch irace.
}
\details{
This function checks the correctness of the scenario, reads the
parameter space from \code{scenario$parameterFile}, invokes \code{\link[=irace]{irace()}},
prints its results in various formatted ways, (optionally) calls
\code{\link[=psRace]{psRace()}} and, finally, evaluates the best configurations on the test
instances (if provided). If you want a lower-level interface that just
runs irace, please see function \code{\link[=irace]{irace()}}.
}
\seealso{
\code{\link[=irace.cmdline]{irace.cmdline()}} a higher-level command-line interface to
\code{\link[=irace]{irace()}}
\code{\link[=readScenario]{readScenario()}} to read the scenario setup from  a file.
\code{\link[=defaultScenario]{defaultScenario()}} to provide a default scenario for \pkg{irace}.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/irace_summarise.R
\name{irace_summarise}
\alias{irace_summarise}
\title{Summarise the results of a run of irace}
\usage{
irace_summarise(iraceResults)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}
}
\value{
\code{list()}
}
\description{
Summarise the results of a run of irace
}
\examples{
irace_results <- read_logfile(system.file("exdata/irace-acotsp.Rdata",
                                          package="irace", mustWork=TRUE))
irace_summarise(irace_results)

}
\author{
Manuel López-Ibáñez
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{configurations.print}
\alias{configurations.print}
\title{Print configurations as a data frame}
\usage{
configurations.print(configurations, metadata = FALSE)
}
\arguments{
\item{configurations}{(\code{data.frame}) \cr Parameter configurations of the
target algorithm (one per row).}

\item{metadata}{A Boolean specifying whether to print the metadata or
not. The metadata are data for the configurations (additionally to the
value of each parameter) used by \pkg{irace}.}
}
\value{
None.
}
\description{
Print configurations as a data frame
}
\seealso{
\code{\link[=configurations.print.command]{configurations.print.command()}} to print the configurations as command-line strings.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{irace_cmdline}
\alias{irace_cmdline}
\alias{irace.cmdline}
\title{Launch \code{irace} with command-line options.}
\usage{
irace_cmdline(argv = commandArgs(trailingOnly = TRUE))

irace.cmdline(argv = commandArgs(trailingOnly = TRUE))
}
\arguments{
\item{argv}{\code{character()}\cr The arguments
provided on the R command line as a character vector, e.g.,
\code{c("--scenario", "scenario.txt", "-p", "parameters.txt")}.
Using the  default value (not providing the parameter) is the
easiest way to call \code{\link[=irace_cmdline]{irace_cmdline()}}.}
}
\value{
(\code{invisible(data.frame)})

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:
\itemize{
\item \code{.ID.} : Internal id of the candidate configuration.
\item \verb{Parameter names} : One column per parameter name in \code{parameters}.
\item \code{.PARENT.} : Internal id of the parent candidate configuration.
}

Additionally, this function saves an R data file containing an object called
\code{iraceResults}. The path of the file is indicated in \code{scenario$logFile}.
The \code{iraceResults} object is a list with the following structure:

\describe{

\item{\code{scenario}}{The scenario R object containing the \pkg{irace}
options used for the execution. See \code{\link{defaultScenario}}
for more information. The element \code{scenario$parameters} contains the parameters R object that describes the target algorithm parameters. See
\code{\link{readParameters}}.}

\item{\code{allConfigurations}}{The target algorithm configurations
generated by \pkg{irace}. This object is a data frame, each row is a
candidate configuration, the first column (\code{.ID.}) indicates the
internal identifier of the configuration, the following columns
correspond to the parameter values, each column named as the parameter
name specified in the parameter object. The final column
(\code{.PARENT.})  is the identifier of the configuration from which
model the actual configuration was sampled.}

\item{\code{allElites}}{A list that contains one element per iteration,
each element contains the internal identifier of the elite candidate
configurations of the corresponding iteration (identifiers correspond to
\code{allConfigurations$.ID.}).}

\item{\code{iterationElites}}{A vector containing the best candidate
configuration internal identifier of each iteration. The best
configuration found corresponds to the last one of this vector.}

\item{\code{experiments}}{A matrix with configurations as columns and
instances as rows. Column names correspond to the internal identifier of
the configuration (\code{allConfigurations$.ID.}).}

\item{\code{experimen_log}}{A \code{data.table} with columns \code{iteration},
\code{instance}, \code{configuration}, \code{time}.  This matrix contains the log of all the
experiments that \pkg{irace} performs during its execution.  The
instance column refers to the index of the \code{race_state$instances_log}
data frame. Time is saved ONLY when reported by the \code{targetRunner}.}

\item{\code{softRestart}}{A logical vector that indicates if a soft
restart was performed on each iteration. If \code{FALSE}, then no soft
restart was performed.}

\item{\code{state}}{An environment that contains the state of \pkg{irace}, the
recovery is done using the information contained in this object.}

\item{\code{testing}}{A list that contains the testing results. The
elements of this list are: \code{experiments} a matrix with the testing
experiments of the selected configurations in the same format as the
explained above and \code{seeds} a vector with the seeds used to execute
each experiment.}

}
}
\description{
Calls \code{\link[=irace_main]{irace_main()}} using command-line options, maybe parsed from the
command line used to invoke R.
}
\details{
The function reads the parameters given on the command line
used to invoke R, finds the name of the scenario file,
initializes the scenario from the file (with the function
\code{\link[=readScenario]{readScenario()}}) and possibly from parameters passed in
the command line. It finally starts \pkg{irace} by calling
\code{\link[=irace_main]{irace_main()}}.

List of command-line options:

\if{html}{\out{<div class="sourceCode">}}\preformatted{-h,--help                Show this help.
-v,--version             Show irace package version.
-c,--check               Check scenario.
-i,--init                Initialize the working directory with template config
                         files.
   --only-test           Only test the configurations given in the file passed
                         as argument.
-s,--scenario            File that describes the configuration scenario setup
                         and other irace settings. Default: ./scenario.txt.
   --exec-dir            Directory where the programs will be run. Default: ./.
-p,--parameter-file      File that contains the description of the parameters
                         of the target algorithm. Default: ./parameters.txt.
   --configurations-file  File that contains a table of initial configurations.
                         If empty or `NULL`, all initial configurations are
                         randomly generated.
-l,--log-file            File to save tuning results as an R dataset, either
                         absolute path or relative to execDir. Default:
                         ./irace.Rdata.
   --recovery-file       Previously saved log file to recover the execution of
                         `irace`, either absolute path or relative to the
                         current directory.  If empty or `NULL`, recovery is
                         not performed.
   --train-instances-dir  Directory where training instances are located;
                         either absolute path or relative to current directory.
                         If no `trainInstancesFiles` is provided, all the files
                         in `trainInstancesDir` will be listed as instances.
   --train-instances-file  File that contains a list of training instances and
                         optionally additional parameters for them. If
                         `trainInstancesDir` is provided, `irace` will search
                         for the files in this folder.
   --sample-instances    Randomly sample the training instances or use them in
                         the order given. Default: 1.
   --test-instances-dir  Directory where testing instances are located, either
                         absolute or relative to current directory.
   --test-instances-file  File containing a list of test instances and
                         optionally additional parameters for them.
   --test-num-elites     Number of elite configurations returned by irace that
                         will be tested if test instances are provided.
                         Default: 1.
   --test-iteration-elites  Enable/disable testing the elite configurations
                         found at each iteration. Default: 0.
   --test-type           Statistical test used for elimination. The default
                         value selects `t-test` if `capping` is enabled or
                         `F-test`, otherwise. Valid values are: F-test
                         (Friedman test), t-test (pairwise t-tests with no
                         correction), t-test-bonferroni (t-test with
                         Bonferroni's correction for multiple comparisons),
                         t-test-holm (t-test with Holm's correction for
                         multiple comparisons).
   --first-test          Number of instances evaluated before the first
                         elimination test. It must be a multiple of `eachTest`.
                         Default: 5.
   --block-size          Number of training instances, that make up a 'block'
                         in `trainInstancesFile`. Elimination of configurations
                         will only be performed after evaluating a complete
                         block and never in the middle of a block. Each block
                         typically contains one instance from each instance
                         class (type or family) and the block size is the
                         number of classes. The value of `blockSize` will
                         multiply `firstTest`, `eachTest` and
                         `elitistNewInstances`. Default: 1.
   --each-test           Number of instances evaluated between elimination
                         tests. Default: 1.
   --target-runner       Executable called for each configuration that executes
                         the target algorithm to be tuned. See the templates
                         and examples provided. Default: ./target-runner.
   --target-runner-launcher  Executable that will be used to launch the target
                         runner, when `targetRunner` cannot be executed
                         directly (e.g., a Python script in Windows).
   --target-cmdline      Command-line arguments provided to `targetRunner` (or
                         `targetRunnerLauncher` if defined). The substrings
                         `\\\{configurationID\\\}`, `\\\{instanceID\\\}`, `\\\{seed\\\}`,
                         `\\\{instance\\\}`, and `\\\{bound\\\}` will be replaced by
                         their corresponding values. The substring
                         `\\\{targetRunnerArgs\\\}` will be replaced by the
                         concatenation of the switch and value of all active
                         parameters of the particular configuration being
                         evaluated.  The substring `\\\{targetRunner\\\}`, if
                         present, will be replaced by the value of
                         `targetRunner` (useful when using
                         `targetRunnerLauncher`). Default: \{configurationID\}
                         \{instanceID\} \{seed\} \{instance\} \{bound\}
                         \{targetRunnerArgs\}.
   --target-runner-retries  Number of times to retry a call to `targetRunner`
                         if the call failed. Default: 0.
   --target-runner-timeout  Timeout in seconds of any `targetRunner` call (only
                         applies to `target-runner` executables not to R
                         functions), ignored if 0. Default: 0.
   --target-evaluator    Optional script or R function that provides a numeric
                         value for each configuration. See
                         templates/target-evaluator.tmpl
   --deterministic       If the target algorithm is deterministic,
                         configurations will be evaluated only once per
                         instance. Default: 0.
   --max-experiments     Maximum number of runs (invocations of `targetRunner`)
                         that will be performed. It determines the maximum
                         budget of experiments for the tuning. Default: 0.
   --min-experiments     Minimum number of runs (invocations of `targetRunner`)
                         that will be performed. It determines the minimum
                         budget of experiments for the tuning. The actual
                         budget depends on the number of parameters and
                         `minSurvival`.
   --max-time            Maximum total execution time for the executions of
                         `targetRunner`. `targetRunner` must return two values:
                         cost and time. This value and the one returned by
                         `targetRunner` must use the same units (seconds,
                         minutes, iterations, evaluations, ...). Default: 0.
   --budget-estimation   Fraction (smaller than 1) of the budget used to
                         estimate the mean computation time of a configuration.
                         Only used when `maxTime` > 0 Default: 0.05.
   --min-measurable-time  Minimum time unit that is still (significantly)
                         measureable. Default: 0.01.
   --parallel            Number of calls to `targetRunner` to execute in
                         parallel. Values `0` or `1` mean no parallelization.
                         Default: 0.
   --load-balancing      Enable/disable load-balancing when executing
                         experiments in parallel. Load-balancing makes better
                         use of computing resources, but increases
                         communication overhead. If this overhead is large,
                         disabling load-balancing may be faster. Default: 1.
   --mpi                 Enable/disable MPI. Use `Rmpi` to execute
                         `targetRunner` in parallel (parameter `parallel` is
                         the number of slaves). Default: 0.
   --batchmode           Specify how irace waits for jobs to finish when
                         `targetRunner` submits jobs to a batch cluster: sge,
                         pbs, torque, slurm or htcondor. `targetRunner` must
                         submit jobs to the cluster using, for example, `qsub`.
                         Default: 0.
-q,--quiet               Reduce the output generated by irace to a minimum.
                         Default: 0.
   --debug-level         Debug level of the output of `irace`. Set this to 0 to
                         silence all debug messages. Higher values provide more
                         verbose debug messages. Default: 0.
   --seed                Seed of the random number generator (by default,
                         generate a random seed).
   --soft-restart        Enable/disable the soft restart strategy that avoids
                         premature convergence of the probabilistic model.
                         Default: 1.
   --soft-restart-threshold  Soft restart threshold value for numerical
                         parameters. Default: 1e-04.
-e,--elitist             Enable/disable elitist irace. Default: 1.
   --elitist-new-instances  Number of instances added to the execution list
                         before previous instances in elitist irace. Default:
                         1.
   --elitist-limit       In elitist irace, maximum number per race of
                         elimination tests that do not eliminate a
                         configuration. Use 0 for no limit. Default: 2.
   --capping             Enable the use of adaptive capping, a technique
                         designed for minimizing the computation time of
                         configurations. Capping is enabled by default if
                         `elitist` is active, `maxTime > 0` and `boundMax > 0`.
   --capping-after-first-test  If set to 1, elimination due to capping only
                         happens after `firstTest` instances are seen. Default:
                         0.
   --capping-type        Measure used to obtain the execution bound from the
                         performance of the elite configurations: median, mean,
                         worst, best. Default: median.
   --bound-type          Method to calculate the mean performance of elite
                         configurations: candidate or instance. Default:
                         candidate.
   --bound-max           Maximum execution bound for `targetRunner`. It must be
                         specified when capping is enabled. Default: 0.
   --bound-digits        Precision used for calculating the execution time. It
                         must be specified when capping is enabled. Default: 0.
   --bound-par           Penalization constant for timed out executions
                         (executions that reach `boundMax` execution time).
                         Default: 1.
   --bound-as-timeout    Replace the configuration cost of bounded executions
                         with `boundMax`. Default: 1.
   --postselection       Perform a postselection race after the execution of
                         irace to consume all remaining budget. Value 0
                         disables the postselection race. Default: 1.
   --aclib               Enable/disable AClib mode. This option enables
                         compatibility with GenericWrapper4AC as targetRunner
                         script. Default: 0.
   --iterations          Maximum number of iterations. Default: 0.
   --experiments-per-iteration  Number of runs of the target algorithm per
                         iteration. Default: 0.
   --min-survival        Minimum number of configurations needed to continue
                         the execution of each race (iteration). Default: 0.
   --num-configurations  Number of configurations to be sampled and evaluated
                         at each iteration. Default: 0.
   --mu                  Parameter used to define the number of configurations
                         sampled and evaluated at each iteration. Default: 5.
   --confidence          Confidence level for the elimination test. Default:
                         0.95.
}\if{html}{\out{</div>}}
}
\examples{
irace_cmdline("--version")
}
\seealso{
\code{\link[=irace_main]{irace_main()}} to start \pkg{irace} with a given scenario.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameterAnalysis.R
\name{getConfigurationByIteration}
\alias{getConfigurationByIteration}
\title{Returns the configurations by the iteration in which they were executed.}
\usage{
getConfigurationByIteration(iraceResults, iterations, drop.metadata = FALSE)
}
\arguments{
\item{iraceResults}{(\code{list()}|\code{character(1)}) \code{iraceResults} object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{iterations}{The iteration number or a vector of iteration numbers from where
the configurations should be obtained.}

\item{drop.metadata}{(\code{FALSE}) Remove metadata, such the configuration ID and
the ID of the parent, from the returned configurations. See
\code{\link{removeConfigurationsMetaData}}.}
}
\value{
A data frame containing the elite configurations required.
}
\description{
Returns the configurations by the iteration in which they were executed.
}
\author{
Manuel López-Ibáñez and Leslie Pérez Cáceres
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{target_runner_default}
\alias{target_runner_default}
\title{Default \code{targetRunner} function.}
\usage{
target_runner_default(experiment, scenario)
}
\arguments{
\item{experiment}{A list describing the experiment. It contains at least:
\describe{
\item{\code{id_configuration}}{An alphanumeric string that uniquely identifies a configuration;}
\item{\code{id_instance}}{An alphanumeric string that uniquely identifies an instance;}
\item{\code{seed}}{Seed for the random number generator to be used for
this evaluation, ignore the seed for deterministic algorithms;}
\item{\code{instance}}{String giving the instance to be used for this evaluation;}
\item{\code{bound}}{(only when \code{capping} is enabled) Time bound for the execution;}
\item{\code{configuration}}{1-row data frame with a column per parameter
name;}
}}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
If \code{targetEvaluator} is \code{NULL}, then the \code{targetRunner}
function must return a list with at least one element \code{"cost"},
the numerical value corresponding to the evaluation of the given
configuration on the given instance.

If the scenario option \code{maxTime} is non-zero or if \code{capping} is enabled
then the list must contain at least another element \code{"time"} that reports the
execution time for this call to \code{targetRunner}.
The return list may also contain the following optional elements that are used
by \pkg{irace} for reporting errors in \code{targetRunner}:
\describe{
\item{\code{error}}{is a string used to report an error;}
\item{\code{outputRaw}}{is a string used to report the raw output of calls to
an external program or function;}
\item{\code{call}}{is a string used to report how \code{targetRunner} called
an external program or function.}
}
}
\description{
Use it as an advanced example of how to create your own \code{targetRunner} function.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{target.evaluator.default}
\alias{target.evaluator.default}
\title{target.evaluator.default}
\usage{
target.evaluator.default(
  experiment,
  num.configurations,
  all.conf.id,
  scenario,
  target.runner.call
)
}
\arguments{
\item{experiment}{A list describing the experiment. It contains at least:
\describe{
\item{\code{id.configuration}}{An alphanumeric string that uniquely identifies a configuration;}
\item{\code{id.instance}}{An alphanumeric string that uniquely identifies an instance;}
\item{\code{seed}}{Seed for the random number generator to be used for
this evaluation, ignore the seed for deterministic algorithms;}
\item{\code{instance}}{String giving the instance to be used for this evaluation;}
\item{\code{bound}}{(only when \code{capping} is enabled) Time bound for the execution;}
\item{\code{configuration}}{1-row data frame with a column per parameter
name;}
\item{\code{switches}}{Vector of parameter switches (labels) in the order
of parameters used in \code{configuration}.}
}}

\item{num.configurations}{Number of  configurations alive in the race.}

\item{all.conf.id}{Vector of configuration IDs of the alive configurations.}

\item{scenario}{(\code{list()}) \cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{target.runner.call}{String describing the call to \code{targetRunner} that
corresponds to this call to \code{targetEvaluator}. This is used for
providing extra information to the user, for example, in case
\code{targetEvaluator} fails.}
}
\value{
The function \code{targetEvaluator} must return a list with one element
\code{"cost"}, the numerical value corresponding to the cost measure of the
given configuration on the given instance.

The return list may also contain the following optional elements that are used
by \pkg{irace} for reporting errors in \code{targetEvaluator}:
\describe{
\item{\code{error}}{is a string used to report an error;}
\item{\code{outputRaw}}{is a string used to report the raw output of calls to
an external program or function;}
\item{\code{call}}{is a string used to report how \code{targetRunner} called
an external program or function.}
}
}
\description{
\code{target.evaluator.default} is the default \code{targetEvaluator} function that is
invoked if \code{targetEvaluator} is a string (by default
\code{targetEvaluator} is \code{NULL} and this function is not invoked). You can use it as
an advanced example of how to create your own \code{targetEvaluator} function.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{target_evaluator_default}
\alias{target_evaluator_default}
\title{target_evaluator_default}
\usage{
target_evaluator_default(
  experiment,
  num_configurations,
  all_conf_id,
  scenario,
  target_runner_call
)
}
\arguments{
\item{experiment}{A list describing the experiment. It contains at least:
\describe{
\item{\code{id_configuration}}{An alphanumeric string that uniquely identifies a configuration;}
\item{\code{id_instance}}{An alphanumeric string that uniquely identifies an instance;}
\item{\code{seed}}{Seed for the random number generator to be used for
this evaluation, ignore the seed for deterministic algorithms;}
\item{\code{instance}}{String giving the instance to be used for this evaluation;}
\item{\code{bound}}{(only when \code{capping} is enabled) Time bound for the execution;}
\item{\code{configuration}}{1-row data frame with a column per parameter
name;}
}}

\item{num_configurations}{Number of  configurations alive in the race.}

\item{all_conf_id}{Vector of configuration IDs of the alive configurations.}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{target_runner_call}{String describing the call to \code{targetRunner} that
corresponds to this call to \code{targetEvaluator}. This is used for
providing extra information to the user, for example, in case
\code{targetEvaluator} fails.}
}
\value{
The function \code{targetEvaluator} must return a list with one element
\code{"cost"}, the numerical value corresponding to the cost measure of the
given configuration on the given instance.

The return list may also contain the following optional elements that are used
by \pkg{irace} for reporting errors in \code{targetEvaluator}:
\describe{
\item{\code{error}}{is a string used to report an error;}
\item{\code{outputRaw}}{is a string used to report the raw output of calls to
an external program or function;}
\item{\code{call}}{is a string used to report how \code{targetRunner} called
an external program or function.}
}
}
\description{
\code{target_evaluator_default} is the default \code{targetEvaluator} function that is
invoked if \code{targetEvaluator} is a string (by default
\code{targetEvaluator} is \code{NULL} and this function is not invoked). You can use it as
an advanced example of how to create your own \code{targetEvaluator} function.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameterAnalysis.R
\name{getConfigurationById}
\alias{getConfigurationById}
\title{Returns the configurations selected by ID.}
\usage{
getConfigurationById(iraceResults, ids, drop.metadata = FALSE)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{ids}{\code{integer()}\cr The id or a vector of ids of the candidates configurations to obtain.}

\item{drop.metadata}{\code{logical(1)}\cr Remove metadata, such as the
configuration ID and the ID of the parent, from the returned
configurations.  See \code{\link[=removeConfigurationsMetaData]{removeConfigurationsMetaData()}}.}
}
\value{
A data frame containing the elite configurations required, in the
order and with the repetitions given by \code{ids}.
}
\description{
Returns the configurations selected by ID.
}
\examples{
log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
getConfigurationById(log_file, ids = c(2,1), drop.metadata = TRUE)

}
\author{
Manuel López-Ibáñez and Leslie Pérez Cáceres
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scenario.R
\name{defaultScenario}
\alias{defaultScenario}
\title{Default scenario settings}
\usage{
defaultScenario(scenario = list(), params_def = .irace.params.def)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{params_def}{\code{data.frame()}\cr Definition of the options accepted by
the scenario. This should only be modified by packages that wish to extend
\pkg{irace}.}
}
\value{
A list indexed by the \pkg{irace} parameter names,
containing the default values for each parameter, except for those
already present in the scenario passed as argument.
The scenario list contains the following elements:
\itemize{
\item General options:
\describe{
\item{\code{scenarioFile}}{Path of the file that describes the configuration scenario setup and other irace settings. (Default: \code{"./scenario.txt"})}
\item{\code{execDir}}{Directory where the programs will be run. (Default: \code{"./"})}
\item{\code{logFile}}{File to save tuning results as an R dataset, either absolute path or relative to execDir. (Default: \code{"./irace.Rdata"})}
\item{\code{quiet}}{Reduce the output generated by irace to a minimum. (Default: \code{0})}
\item{\code{debugLevel}}{Debug level of the output of \code{irace}. Set this to 0 to silence all debug messages. Higher values provide more verbose debug messages. (Default: \code{0})}
\item{\code{seed}}{Seed of the random number generator (by default, generate a random seed). (Default: \code{NA})}
\item{\code{repairConfiguration}}{User-defined R function that takes a configuration generated by irace and repairs it. (Default: \code{""})}
\item{\code{postselection}}{Perform a postselection race after the execution of irace to consume all remaining budget. Value 0 disables the postselection race. (Default: \code{1})}
\item{\code{aclib}}{Enable/disable AClib mode. This option enables compatibility with GenericWrapper4AC as targetRunner script. (Default: \code{0})}
}
\item Elitist \code{irace}:
\describe{
\item{\code{elitist}}{Enable/disable elitist irace. (Default: \code{1})}
\item{\code{elitistNewInstances}}{Number of instances added to the execution list before previous instances in elitist irace. (Default: \code{1})}
\item{\code{elitistLimit}}{In elitist irace, maximum number per race of elimination tests that do not eliminate a configuration. Use 0 for no limit. (Default: \code{2})}
}
\item Internal \code{irace} options:
\describe{
\item{\code{sampleInstances}}{Randomly sample the training instances or use them in the order given. (Default: \code{1})}
\item{\code{softRestart}}{Enable/disable the soft restart strategy that avoids premature convergence of the probabilistic model. (Default: \code{1})}
\item{\code{softRestartThreshold}}{Soft restart threshold value for numerical parameters. (Default: \code{1e-04})}
\item{\code{nbIterations}}{Maximum number of iterations. (Default: \code{0})}
\item{\code{nbExperimentsPerIteration}}{Number of runs of the target algorithm per iteration. (Default: \code{0})}
\item{\code{minNbSurvival}}{Minimum number of configurations needed to continue the execution of each race (iteration). (Default: \code{0})}
\item{\code{nbConfigurations}}{Number of configurations to be sampled and evaluated at each iteration. (Default: \code{0})}
\item{\code{mu}}{Parameter used to define the number of configurations sampled and evaluated at each iteration. (Default: \code{5})}
}
\item Target algorithm parameters:
\describe{
\item{\code{parameterFile}}{File that contains the description of the parameters of the target algorithm. (Default: \code{"./parameters.txt"})}
\item{\code{parameters}}{Parameters space object (usually read from a file using \code{readParameters}). (Default: \code{""})}
}
\item Target algorithm execution:
\describe{
\item{\code{targetRunner}}{Executable called for each configuration that executes the target algorithm to be tuned. See the templates and examples provided. (Default: \code{"./target-runner"})}
\item{\code{targetRunnerLauncher}}{Executable that will be used to launch the target runner, when \code{targetRunner} cannot be executed directly (e.g., a Python script in Windows). (Default: \code{""})}
\item{\code{targetCmdline}}{Command-line arguments provided to \code{targetRunner} (or \code{targetRunnerLauncher} if defined). The substrings \verb{\\\{configurationID\\\}}, \verb{\\\{instanceID\\\}},  \verb{\\\{seed\\\}},  \verb{\\\{instance\\\}}, and \verb{\\\{bound\\\}} will be replaced by their corresponding values. The substring \verb{\\\{targetRunnerArgs\\\}} will be replaced by the concatenation of the switch and value of all active parameters of the particular configuration being evaluated.  The substring \verb{\\\{targetRunner\\\}}, if present, will be replaced by the value of \code{targetRunner} (useful when using \code{targetRunnerLauncher}). (Default: \code{"{configurationID} {instanceID} {seed} {instance} {bound} {targetRunnerArgs}"})}
\item{\code{targetRunnerRetries}}{Number of times to retry a call to \code{targetRunner} if the call failed. (Default: \code{0})}
\item{\code{targetRunnerTimeout}}{Timeout in seconds of any \code{targetRunner} call (only applies to \code{target-runner} executables not to R functions), ignored if 0. (Default: \code{0})}
\item{\code{targetRunnerData}}{Optional data passed to \code{targetRunner}. This is ignored by the default \code{targetRunner} function, but it may be used by custom \code{targetRunner} functions to pass persistent data around. (Default: \code{""})}
\item{\code{targetRunnerParallel}}{Optional R function to provide custom parallelization of \code{targetRunner}. (Default: \code{""})}
\item{\code{targetEvaluator}}{Optional script or R function that provides a numeric value for each configuration. See templates/target-evaluator.tmpl (Default: \code{""})}
\item{\code{deterministic}}{If the target algorithm is deterministic, configurations will be evaluated only once per instance. (Default: \code{0})}
\item{\code{parallel}}{Number of calls to \code{targetRunner} to execute in parallel. Values \code{0} or \code{1} mean no parallelization. (Default: \code{0})}
\item{\code{loadBalancing}}{Enable/disable load-balancing when executing experiments in parallel. Load-balancing makes better use of computing resources, but increases communication overhead. If this overhead is large, disabling load-balancing may be faster. (Default: \code{1})}
\item{\code{mpi}}{Enable/disable MPI. Use \code{Rmpi} to execute \code{targetRunner} in parallel (parameter \code{parallel} is the number of slaves). (Default: \code{0})}
\item{\code{batchmode}}{Specify how irace waits for jobs to finish when \code{targetRunner} submits jobs to a batch cluster: sge, pbs, torque, slurm or htcondor. \code{targetRunner} must submit jobs to the cluster using, for example, \code{qsub}. (Default: \code{0})}
}
\item Initial configurations:
\describe{
\item{\code{initConfigurations}}{Data frame describing initial configurations (usually read from a file using \code{readConfigurations}). (Default: \code{""})}
\item{\code{configurationsFile}}{File that contains a table of initial configurations. If empty or \code{NULL}, all initial configurations are randomly generated. (Default: \code{""})}
}
\item Training instances:
\describe{
\item{\code{instances}}{Character vector of the instances to be used in the \code{targetRunner}. (Default: \code{""})}
\item{\code{trainInstancesDir}}{Directory where training instances are located; either absolute path or relative to current directory. If no \code{trainInstancesFiles} is provided, all the files in \code{trainInstancesDir} will be listed as instances. (Default: \code{""})}
\item{\code{trainInstancesFile}}{File that contains a list of training instances and optionally additional parameters for them. If \code{trainInstancesDir} is provided, \code{irace} will search for the files in this folder. (Default: \code{""})}
\item{\code{blockSize}}{Number of training instances, that make up a 'block' in \code{trainInstancesFile}. Elimination of configurations will only be performed after evaluating a complete block and never in the middle of a block. Each block typically contains one instance from each instance class (type or family) and the block size is the number of classes. The value of \code{blockSize} will multiply \code{firstTest}, \code{eachTest} and \code{elitistNewInstances}. (Default: \code{1})}
}
\item Tuning budget:
\describe{
\item{\code{maxExperiments}}{Maximum number of runs (invocations of \code{targetRunner}) that will be performed. It determines the maximum budget of experiments for the tuning. (Default: \code{0})}
\item{\code{minExperiments}}{Minimum number of runs (invocations of \code{targetRunner}) that will be performed. It determines the minimum budget of experiments for the tuning. The actual budget depends on the number of parameters and \code{minSurvival}. (Default: \code{NA})}
\item{\code{maxTime}}{Maximum total execution time for the executions of \code{targetRunner}. \code{targetRunner} must return two values: cost and time. This value and the one returned by \code{targetRunner} must use the same units (seconds, minutes, iterations, evaluations, ...). (Default: \code{0})}
\item{\code{budgetEstimation}}{Fraction (smaller than 1) of the budget used to estimate the mean computation time of a configuration. Only used when \code{maxTime} > 0 (Default: \code{0.05})}
\item{\code{minMeasurableTime}}{Minimum time unit that is still (significantly) measureable. (Default: \code{0.01})}
}
\item Statistical test:
\describe{
\item{\code{testType}}{Statistical test used for elimination. The default value selects \code{t-test} if \code{capping} is enabled or \code{F-test}, otherwise. Valid values are: F-test (Friedman test), t-test (pairwise t-tests with no correction), t-test-bonferroni (t-test with Bonferroni's correction for multiple comparisons), t-test-holm (t-test with Holm's correction for multiple comparisons). (Default: \code{""})}
\item{\code{firstTest}}{Number of instances evaluated before the first elimination test. It must be a multiple of \code{eachTest}. (Default: \code{5})}
\item{\code{eachTest}}{Number of instances evaluated between elimination tests. (Default: \code{1})}
\item{\code{confidence}}{Confidence level for the elimination test. (Default: \code{0.95})}
}
\item Adaptive capping:
\describe{
\item{\code{capping}}{Enable the use of adaptive capping, a technique designed for minimizing the computation time of configurations. Capping is enabled by default if \code{elitist} is active, \code{maxTime > 0} and  \code{boundMax > 0}. (Default: \code{NA})}
\item{\code{cappingAfterFirstTest}}{If set to 1, elimination due to capping only happens after \code{firstTest} instances are seen. (Default: \code{0})}
\item{\code{cappingType}}{Measure used to obtain the execution bound from the performance of the elite configurations.\itemize{\item median: Median performance of the elite configurations.\item mean: Mean performance of the elite configurations.\item best: Best performance of the elite configurations.\item worst: Worst performance of the elite configurations.} (Default: \code{"median"})}
\item{\code{boundType}}{Method to calculate the mean performance of elite configurations.\itemize{\item candidate: Mean execution times across the executed instances and the current one.\item instance: Execution time of the current instance.} (Default: \code{"candidate"})}
\item{\code{boundMax}}{Maximum execution bound for \code{targetRunner}. It must be specified when capping is enabled. (Default: \code{0})}
\item{\code{boundDigits}}{Precision used for calculating the execution time. It must be specified when capping is enabled. (Default: \code{0})}
\item{\code{boundPar}}{Penalization constant for timed out executions (executions that reach \code{boundMax} execution time). (Default: \code{1})}
\item{\code{boundAsTimeout}}{Replace the configuration cost of bounded executions with \code{boundMax}. (Default: \code{1})}
}
\item Recovery:
\describe{
\item{\code{recoveryFile}}{Previously saved log file to recover the execution of \code{irace}, either absolute path or relative to the current directory.  If empty or \code{NULL}, recovery is not performed. (Default: \code{""})}
}
\item Testing:
\describe{
\item{\code{testInstancesDir}}{Directory where testing instances are located, either absolute or relative to current directory. (Default: \code{""})}
\item{\code{testInstancesFile}}{File containing a list of test instances and optionally additional parameters for them. (Default: \code{""})}
\item{\code{testInstances}}{Character vector of the instances to be used in the \code{targetRunner} when executing the testing. (Default: \code{""})}
\item{\code{testNbElites}}{Number of elite configurations returned by irace that will be tested if test instances are provided. (Default: \code{1})}
\item{\code{testIterationElites}}{Enable/disable testing the elite configurations found at each iteration. (Default: \code{0})}
}
}
}
\description{
Return scenario object with default values.
}
\seealso{
\describe{
\item{\code{\link[=readScenario]{readScenario()}}}{for reading a configuration scenario from a file.}
\item{\code{\link[=printScenario]{printScenario()}}}{prints the given scenario.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link[=checkScenario]{checkScenario()}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{parameters}
\alias{parameters}
\alias{parametersNew}
\alias{param_cat}
\alias{param_ord}
\alias{param_real}
\alias{param_int}
\title{Create a parameter space to be tuned.}
\usage{
parametersNew(..., forbidden = NULL, debugLevel = 0L)

param_cat(name = name, values, label = "", condition = TRUE)

param_ord(name, values, label = "", condition = TRUE)

param_real(
  name,
  lower,
  upper,
  label = "",
  condition = TRUE,
  transf = "",
  digits = 15L
)

param_int(name, lower, upper, label = "", condition = TRUE, transf = "")
}
\arguments{
\item{...}{one or more parameters created by \code{param_int()}, \code{param_real()}, \code{param_ord()}, or \code{param_cat()}.}

\item{forbidden}{\code{expression()|character(1)}\cr String or list of expressions that define forbidden parameter configurations.}

\item{debugLevel}{\code{integer(1)}\cr Larger values produce more verbose output.}

\item{name}{\code{character(1)}\cr Parameter name (must be alphanumeric).}

\item{values}{\code{character()}\cr Domain as a vector of strings.}

\item{label}{\code{character(1)}\cr Label associated to the parameter. Often used to encode a command-line switch that activates the parameter.}

\item{condition}{\code{expression(1)|character(1)}\cr Expression that defines when the parameter is active according to the value of other parameters.}

\item{lower, upper}{Lower and upper limits of the valid domain.}

\item{transf}{\code{character(1)}\cr If \code{"log"}, then the parameter is sampled in a logarithmic scale.}

\item{digits}{\code{integer(1)}\cr The number of decimal places to be considered for real-valued parameters.}
}
\value{
\code{ParameterSpace}
}
\description{
\itemize{
\item \code{param_cat()} creates a categorical parameter.
\item \code{param_ord()} creates an ordinal parameter.
\item \code{param_real()} creates a real-valued parameter.
\item \code{param_int()} creates an integer parameter.
}
}
\examples{
digits <- 4L
parametersNew(
   param_cat(name = "algorithm", values = c("as", "mmas", "eas", "ras", "acs"), label = "--"),
   param_ord(name = "localsearch", values = c("0", "1", "2", "3"), label = "--localsearch "),
   param_real(name = "alpha", lower = 0.0, upper=5.0, label = "--alpha ", digits = digits),
   param_real(name = "beta", lower = 0.0, upper = 10.0, label = "--beta ", digits = digits),
   param_real(name = "rho", lower = 0.01, upper = 1.00, label = "--rho ", digits = digits),
   param_int(name = "ants", lower = 5, upper = 100, transf = "log", label = "--ants "),
   param_real(name = "q0", label = "--q0 ", lower=0.0, upper=1.0,
              condition = expression(algorithm == "acs")),
   param_int(name = "rasrank", label = "--rasranks ", lower=1, upper=quote(min(ants, 10)),
             condition = 'algorithm == "ras"'),
   param_int(name = "elitistants", label = "--elitistants ", lower=1, upper=expression(ants),
             condition = 'algorithm == "eas"'),
   param_int(name = "nnls", label = "--nnls ", lower = 5, upper = 50,
             condition = expression(localsearch \%in\% c(1,2,3))),
   param_cat(name = "dlb",  label = "--dlb ", values = c(0,1),
             condition = "localsearch \%in\% c(1,2,3)"),
   forbidden = "(alpha == 0) & (beta == 0)")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{configurations.print.command}
\alias{configurations.print.command}
\title{Print configurations as command-line strings.}
\usage{
configurations.print.command(configurations, parameters)
}
\arguments{
\item{configurations}{(\code{data.frame}) \cr Parameter configurations of the
target algorithm (one per row).}

\item{parameters}{(\code{list()}) \cr Data structure containing the parameter
space definition. The data structure has to similar to the one returned by the
function \code{\link{readParameters}}.}
}
\value{
None.
}
\description{
Prints configurations after converting them into a representation for the
command-line.
}
\seealso{
\code{\link[=configurations.print]{configurations.print()}} to print the configurations as a data frame.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameters.R
\name{printParameters}
\alias{printParameters}
\title{Print parameter space in the textual format accepted by irace.}
\usage{
printParameters(parameters)
}
\arguments{
\item{parameters}{\code{ParameterSpace}\cr Data structure containing the parameter
space definition. The data structure has to similar to the one returned by the
function \code{\link{readParameters}}.}
}
\value{
\code{character()}
}
\description{
Print parameter space in the textual format accepted by irace.
}
\examples{
parameters_table <- '
 # name       switch           type  values               [conditions (using R syntax)]
 algorithm    "--"             c     (as,mmas,eas,ras,acs)
 localsearch  "--localsearch " c     (0, 1, 2, 3)
 alpha        "--alpha "       r     (0.00, 5.00)
 beta         "--beta "        r     (0.00, 10.00)
 rho          "--rho  "        r     (0.01, 1.00)
 ants         "--ants "        i,log (5, 100)
 q0           "--q0 "          r     (0.0, 1.0)           | algorithm == "acs"
 q0dep       "--q0 "           r     (0.0, q0)            | algorithm != "acs"
 rasrank      "--rasranks "    i     (1, "min(ants, 10)") | algorithm == "ras"
 elitistants  "--elitistants " i     (1, ants)            | algorithm == "eas"
 nnls         "--nnls "        i     (5, 50)              | localsearch \%in\% c(1,2,3)
 dlb          "--dlb "         c     (0, 1)               | localsearch \%in\% c(1,2,3)

 [forbidden]
 (alpha == 0.0) & (beta == 0.0)
'
parameters <- readParameters(text=parameters_table)
printParameters(parameters)
}
\seealso{
\code{\link[=readParameters]{readParameters()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{target.runner.default}
\alias{target.runner.default}
\title{Default \code{targetRunner} function.}
\usage{
target.runner.default(experiment, scenario)
}
\arguments{
\item{experiment}{A list describing the experiment. It contains at least:
\describe{
\item{\code{id.configuration}}{An alphanumeric string that uniquely identifies a configuration;}
\item{\code{id.instance}}{An alphanumeric string that uniquely identifies an instance;}
\item{\code{seed}}{Seed for the random number generator to be used for
this evaluation, ignore the seed for deterministic algorithms;}
\item{\code{instance}}{String giving the instance to be used for this evaluation;}
\item{\code{bound}}{(only when \code{capping} is enabled) Time bound for the execution;}
\item{\code{configuration}}{1-row data frame with a column per parameter
name;}
\item{\code{switches}}{Vector of parameter switches (labels) in the order
of parameters used in \code{configuration}.}
}}

\item{scenario}{(\code{list()}) \cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
If \code{targetEvaluator} is \code{NULL}, then the \code{targetRunner}
function must return a list with at least one element \code{"cost"},
the numerical value corresponding to the evaluation of the given
configuration on the given instance.

If the scenario option \code{maxTime} is non-zero or if \code{capping} is enabled
then the list must contain at least another element \code{"time"} that reports the
execution time for this call to \code{targetRunner}.
The return list may also contain the following optional elements that are used
by \pkg{irace} for reporting errors in \code{targetRunner}:
\describe{
\item{\code{error}}{is a string used to report an error;}
\item{\code{outputRaw}}{is a string used to report the raw output of calls to
an external program or function;}
\item{\code{call}}{is a string used to report how \code{targetRunner} called
an external program or function.}
}
}
\description{
Use it as an advanced example of how to create your own \code{targetRunner} function.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\docType{data}
\name{irace.version}
\alias{irace.version}
\title{irace.version}
\format{
An object of class \code{character} of length 1.
}
\usage{
irace.version
}
\description{
A character string containing the version of \code{irace}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scenario.R
\name{printScenario}
\alias{printScenario}
\title{Prints the given scenario}
\usage{
printScenario(scenario)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\description{
Prints the given scenario
}
\seealso{
\describe{
\item{\code{\link[=readScenario]{readScenario()}}}{for reading a configuration scenario from a file.}
\item{\code{\link[=printScenario]{printScenario()}}}{prints the given scenario.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link[=checkScenario]{checkScenario()}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ablation.R
\name{plotAblation}
\alias{plotAblation}
\title{Create plot from an ablation log}
\usage{
plotAblation(
  ablog,
  pdf_file = NULL,
  width = 20,
  height = 7,
  type = c("mean", "boxplot", "rank"),
  n = 0L,
  mar = NULL,
  ylab = "Mean configuration cost",
  ylim = NULL,
  rename_labels = NULL,
  ...
)
}
\arguments{
\item{ablog}{(\code{list()}|\code{character(1)}) Ablation log object returned by \code{\link[=ablation]{ablation()}}. Alternatively, the path to an \code{.Rdata} file, e.g., \code{"log-ablation.Rdata"}, from which the object will be loaded.}

\item{pdf_file}{Output filename.}

\item{width}{Width provided to create the PDF file.}

\item{height}{Height provided to create the PDF file.}

\item{type}{Type of plot. Supported values are \code{"mean"} and \code{"boxplot"}. Adding \code{"rank"} will plot rank per instance instead of raw cost value.}

\item{n}{\code{integer(1)}\cr Number of parameters included in the plot. By default all parameters are included.}

\item{mar}{Vector with the margins for the ablation plot.}

\item{ylab}{Label of y-axis.}

\item{ylim}{Numeric vector of length 2 giving the y-axis range.}

\item{rename_labels}{\code{character()}\cr Renaming table for nicer labels. For example, \code{c("No value"="NA", "LongParameterName"="LPN")}.}

\item{...}{Further graphical parameters may also be supplied as
arguments. See \code{\link[graphics:plot.default]{graphics::plot.default()}}.}
}
\description{
Create plot from an ablation log
}
\examples{
logfile <- file.path(system.file(package="irace"), "exdata", "log-ablation.Rdata")
plotAblation(ablog = logfile)
plotAblation(ablog = logfile, type = "mean")
plotAblation(ablog = logfile, type = c("rank","boxplot"), rename_labels = c(
            "localsearch"="ls", algorithm="algo", source="default"))
}
\seealso{
\code{\link[=ablation]{ablation()}} \code{\link[=ablation_cmdline]{ablation_cmdline()}}
}
\author{
Leslie Pérez Cáceres and Manuel López-Ibáñez
}
\concept{ablation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readConfiguration.R
\name{readConfigurationsFile}
\alias{readConfigurationsFile}
\title{Read parameter configurations from a file}
\usage{
readConfigurationsFile(filename, parameters, debugLevel = 0L, text)
}
\arguments{
\item{filename}{\code{character(1)}\cr Filename from which the configurations should be read.
The contents should be readable by \code{read.table( , header=TRUE)}.}

\item{parameters}{\code{ParameterSpace}\cr Data structure containing the parameter
space definition. The data structure has to similar to the one returned by the
function \code{\link{readParameters}}.}

\item{debugLevel}{\code{integer(1)}\cr Larger values produce more verbose output.}

\item{text}{\code{character(1)}\cr If \code{file} is not supplied and this is,
then configurations are read from the value of \code{text} via a text connection.}
}
\value{
A data frame containing the obtained configurations.
Each row of the data frame is a candidate configuration,
the columns correspond to the parameter names in \code{parameters}.
}
\description{
Reads a set of target-algorithm configurations from a file and puts them in
\pkg{irace} format. The configurations are checked to match the parameters
description provided.
}
\details{
Example of an input file:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# This is a comment line
param_1    param_2
    0.5  "value_1"
    1.0         NA
    1.2  "value_3"
}\if{html}{\out{</div>}}

The order of the columns does not necessarily have to be the same
as in the file containing the definition of the parameters.
}
\seealso{
\code{\link[=readParameters]{readParameters()}} to obtain a valid parameter structure from a parameters file.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{check_output_target_runner}
\alias{check_output_target_runner}
\title{Check the output of the target runner and repair it if possible. If the
output is incorrect, this function will throw an error.}
\usage{
check_output_target_runner(output, scenario, bound = NULL)
}
\arguments{
\item{output}{The output from target runner.}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{bound}{Optional time bound that the target runner should have respected.}
}
\value{
The output with its contents repaired.
}
\description{
Check the output of the target runner and repair it if possible. If the
output is incorrect, this function will throw an error.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/configurations.R
\name{removeConfigurationsMetaData}
\alias{removeConfigurationsMetaData}
\title{removeConfigurationsMetaData}
\usage{
removeConfigurationsMetaData(configurations)
}
\arguments{
\item{configurations}{\code{data.frame}\cr Parameter configurations of the
target algorithm (one per row).}
}
\value{
The same data frame without "metadata".
}
\description{
Remove the columns with "metadata" of a data frame containing
configurations. Currently, metadata corresponds to column names starting
with a period.  This function should be used before printing the
configurations to output only the values for the parameters of the
configuration without metadata possibly useless to the user.
}
\seealso{
\code{\link[=configurations_print_command]{configurations_print_command()}} to print the configurations as command lines.
\code{\link[=configurations_print]{configurations_print()}} to print the configurations as a data frame.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{testing_fromfile}
\alias{testing_fromfile}
\title{Test configurations given an explicit table of configurations and a scenario file}
\usage{
testing_fromfile(filename, scenario)
}
\arguments{
\item{filename}{\code{character(1)}\cr Path to a file containing configurations: one configuration
per line, one parameter per column, parameter names in header.}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
iraceResults
}
\description{
Executes the testing of an explicit list of configurations given in
\code{filename} (same format as in \code{\link[=readConfigurationsFile]{readConfigurationsFile()}}). A \code{logFile} is
created unless disabled in \code{scenario}. This may overwrite an existing one!
}
\seealso{
\code{\link[=testing_fromlog]{testing_fromlog()}} provides a different interface for testing.
}
\author{
Manuel López-Ibáñez
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{irace_main}
\alias{irace_main}
\title{Higher-level interface to launch irace.}
\usage{
irace_main(scenario, output.width = 9999L)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{output.width}{\code{integer(1)}\cr The width used for the screen
output.}
}
\value{
(\code{invisible(data.frame)})

A data frame with the set of best algorithm configurations found by \pkg{irace}.
The data frame has the following columns:
\itemize{
\item \code{.ID.} : Internal id of the candidate configuration.
\item \verb{Parameter names} : One column per parameter name in \code{parameters}.
\item \code{.PARENT.} : Internal id of the parent candidate configuration.
}

Additionally, this function saves an R data file containing an object called
\code{iraceResults}. The path of the file is indicated in \code{scenario$logFile}.
The \code{iraceResults} object is a list with the following structure:

\describe{

\item{\code{scenario}}{The scenario R object containing the \pkg{irace}
options used for the execution. See \code{\link{defaultScenario}}
for more information. The element \code{scenario$parameters} contains the parameters R object that describes the target algorithm parameters. See
\code{\link{readParameters}}.}

\item{\code{allConfigurations}}{The target algorithm configurations
generated by \pkg{irace}. This object is a data frame, each row is a
candidate configuration, the first column (\code{.ID.}) indicates the
internal identifier of the configuration, the following columns
correspond to the parameter values, each column named as the parameter
name specified in the parameter object. The final column
(\code{.PARENT.})  is the identifier of the configuration from which
model the actual configuration was sampled.}

\item{\code{allElites}}{A list that contains one element per iteration,
each element contains the internal identifier of the elite candidate
configurations of the corresponding iteration (identifiers correspond to
\code{allConfigurations$.ID.}).}

\item{\code{iterationElites}}{A vector containing the best candidate
configuration internal identifier of each iteration. The best
configuration found corresponds to the last one of this vector.}

\item{\code{experiments}}{A matrix with configurations as columns and
instances as rows. Column names correspond to the internal identifier of
the configuration (\code{allConfigurations$.ID.}).}

\item{\code{experimen_log}}{A \code{data.table} with columns \code{iteration},
\code{instance}, \code{configuration}, \code{time}.  This matrix contains the log of all the
experiments that \pkg{irace} performs during its execution.  The
instance column refers to the index of the \code{race_state$instances_log}
data frame. Time is saved ONLY when reported by the \code{targetRunner}.}

\item{\code{softRestart}}{A logical vector that indicates if a soft
restart was performed on each iteration. If \code{FALSE}, then no soft
restart was performed.}

\item{\code{state}}{An environment that contains the state of \pkg{irace}, the
recovery is done using the information contained in this object.}

\item{\code{testing}}{A list that contains the testing results. The
elements of this list are: \code{experiments} a matrix with the testing
experiments of the selected configurations in the same format as the
explained above and \code{seeds} a vector with the seeds used to execute
each experiment.}

}
}
\description{
Higher-level interface to launch irace.
}
\details{
This function checks the correctness of the scenario, reads the
parameter space from \code{scenario$parameterFile}, invokes \code{\link[=irace]{irace()}},
prints its results in various formatted ways, (optionally) calls
\code{\link[=psRace]{psRace()}} and, finally, evaluates the best configurations on the test
instances (if provided). If you want a lower-level interface that just
runs irace, please see function \code{\link[=irace]{irace()}}.
}
\seealso{
\describe{
\item{\code{\link[=irace_cmdline]{irace_cmdline()}}}{a command-line interface to \code{\link[=irace]{irace()}}.}
\item{\code{\link[=readScenario]{readScenario()}}}{for reading a configuration scenario from a file.}
\item{\code{\link[=readParameters]{readParameters()}}}{read the target algorithm parameters from a file.}
\item{\code{\link[=defaultScenario]{defaultScenario()}}}{returns the default scenario settings of \pkg{irace}.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/race-wrapper.R
\name{buildCommandLine}
\alias{buildCommandLine}
\title{Generate a command-line representation of a configuration}
\usage{
buildCommandLine(values, switches)
}
\arguments{
\item{values}{A vector containing the value of each parameter for the
candidate configuration.}

\item{switches}{A vector containing the switches of each paramter (in an
order that corresponds to the values vector).}
}
\value{
A string concatenating each element of \code{switches} and
\code{values} for all parameters with a space between each pair of
parameters (but none between the switches and the corresponding values).
}
\description{
\code{buildCommandLine} receives two vectors, one containing
the values of the parameters, the other containing the switches of the
parameters. It builds a string with the switches and the values that can
be used as a command line to call the program to be tuned, thus generating
one candidate configuration.
}
\examples{
switches <- c("--switch1 ", "--switch2-", "--switch3=")
values <- list("value_1", 1L, sqrt(2))
buildCommandLine (values, switches)
## Build a command-line from the results produced by a previous run of irace.
# First, load the data produced by irace.
logfile <- file.path(system.file(package="irace"), "exdata", "irace-acotsp.Rdata")
iraceResults <- read_logfile(logfile)
allConfigurations <- iraceResults$allConfigurations
parameters <- iraceResults$scenario$parameters
apply(allConfigurations[1:10, unlist(parameters$names)], 1, buildCommandLine,
      unlist(parameters$switches))

}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\docType{data}
\name{irace.license}
\alias{irace.license}
\title{irace.license}
\format{
An object of class \code{character} of length 1.
}
\usage{
irace.license
}
\description{
A character string containing the license information of \pkg{irace}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{testing_fromlog}
\alias{testing_fromlog}
\title{Test configurations given in the logfile (typically \code{irace.Rdata}) produced by \pkg{irace}.}
\usage{
testing_fromlog(
  logFile,
  testNbElites,
  testIterationElites,
  testInstancesDir,
  testInstancesFile,
  testInstances
)
}
\arguments{
\item{logFile}{\code{character(1)}\cr Path to the logfile (typically \code{irace.Rdata}) produced by \pkg{irace}.}

\item{testNbElites}{Number of (final) elite configurations to test. Overrides
the value found in \code{logFile}.}

\item{testIterationElites}{\code{logical(1)}\cr If \code{FALSE}, only the final
\code{testNbElites} configurations are tested; otherwise, also test the best
configurations of each iteration. Overrides the value found in \code{logFile}.}

\item{testInstancesDir}{Directory where testing instances are located, either absolute or relative to current directory.}

\item{testInstancesFile}{File containing a list of test instances and optionally additional parameters for them.}

\item{testInstances}{Character vector of the instances to be used in the \code{targetRunner} when executing the testing.}
}
\value{
\code{logical(1)}\cr \code{TRUE} if the testing ended successfully otherwise, \code{FALSE}.
}
\description{
\code{testing_fromlog} executes the testing of the target algorithm configurations
found by an \pkg{irace} execution.
}
\details{
The function \code{testing_fromlog} loads the \code{logFile} and obtains the
testing setup and configurations to be tested.  Within the \code{logFile}, the
variable \code{scenario$testNbElites} specifies how many final elite
configurations to test and \code{scenario$testIterationElites} indicates
whether test the best configuration of each iteration. The values may be
overridden by setting the corresponding arguments in this function.  The
set of testing instances must appear in \code{scenario[["testInstances"]]}.
}
\seealso{
\code{\link[=defaultScenario]{defaultScenario()}} to provide a default scenario for \pkg{irace}.
\code{\link[=testing_fromfile]{testing_fromfile()}} provides a different interface for testing.
}
\author{
Manuel López-Ibáñez and Leslie Pérez Cáceres
}
\concept{running}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/random_seed.R
\name{random_seed}
\alias{random_seed}
\alias{get_random_seed}
\alias{set_random_seed}
\alias{restore_random_seed}
\title{Get, set and restore the state of the random number generator state.}
\usage{
get_random_seed()

set_random_seed(seed)

restore_random_seed(seed)
}
\arguments{
\item{seed}{(\code{list()}|\code{integer(1)})\cr Either an integer or the list returned gy \code{\link[=get_random_seed]{get_random_seed()}}.}
}
\value{
\code{\link[=get_random_seed]{get_random_seed()}} returns a list with two components \code{random_seed} and \code{rng_kind} or NULL if no seed was set; \code{\link[=set_random_seed]{set_random_seed()}} and \code{\link[=restore_random_seed]{restore_random_seed()}} do not return anything.
}
\description{
Get, set and restore the state of the random number generator state.
}
\details{
These functions originate from the \code{\link[withr:with_seed]{withr}} package.
}
\examples{
old_seed <- get_random_seed()
on.exit(restore_random_seed(old_seed))
set_random_seed(42)
value1 <- runif(1)
set_random_seed(42)
value2 <- runif(1)
stopifnot(all.equal(value1,value2))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/configurations.R
\name{configurations_print}
\alias{configurations_print}
\title{Print configurations as a data frame}
\usage{
configurations_print(configurations, metadata = FALSE)
}
\arguments{
\item{configurations}{\code{data.frame}\cr Parameter configurations of the
target algorithm (one per row).}

\item{metadata}{\code{logical(1)}\cr whether to print the metadata or
not. The metadata are data for the configurations (additionally to the
value of each parameter) used by \pkg{irace}.}
}
\value{
None.
}
\description{
Print configurations as a data frame
}
\seealso{
\code{\link[=configurations_print_command]{configurations_print_command()}} to print the configurations as command-line strings.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/configurations.R
\name{configurations_print_command}
\alias{configurations_print_command}
\title{Print configurations as command-line strings.}
\usage{
configurations_print_command(configurations, parameters)
}
\arguments{
\item{configurations}{\code{data.frame}\cr Parameter configurations of the
target algorithm (one per row).}

\item{parameters}{\code{ParameterSpace}\cr Data structure containing the parameter
space definition. The data structure has to similar to the one returned by the
function \code{\link{readParameters}}.}
}
\value{
None.
}
\description{
Prints configurations after converting them into a representation for the
command-line.
}
\seealso{
\code{\link[=configurations_print]{configurations_print()}} to print the configurations as a data frame.
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/version.R
\docType{data}
\name{irace_version}
\alias{irace_version}
\title{A character string containing the version of \code{irace} including git SHA.}
\format{
An object of class \code{character} of length 1.
}
\usage{
irace_version
}
\description{
A character string containing the version of \code{irace} including git SHA.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psRace.R
\name{psRace}
\alias{psRace}
\title{Post-selection race}
\usage{
psRace(
  iraceResults,
  max_experiments,
  conf_ids = NULL,
  iteration_elites = FALSE,
  psrace_logFile = NULL
)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{max_experiments}{\code{numeric(1)}\cr Number of experiments for the
post-selection race. If it is equal to or smaller than 1, then it is a
fraction of the total budget given by
\code{iraceResults$scenario$maxExperiments} or \code{iraceResults$scenario$maxTime / iraceResults$state$boundEstimate}.}

\item{conf_ids}{IDs of the configurations in \code{iraceResults$allConfigurations} to be used for the post-selection.
If \code{NULL}, then the configurations are automatically selected.}

\item{iteration_elites}{If \code{FALSE}, give priority to selecting the configurations that were elite in the last iteration.
If \code{TRUE}, select from all elite configurations of all iterations. This parameter only has an effect when \code{conf_ids} is not \code{NULL}.}

\item{psrace_logFile}{\code{character(1)}\cr Log file to save the post-selection race log. If \code{NULL}, the log is saved in \code{iraceResults$scenario$logFile}.}
}
\value{
The elite configurations after the post-selection. In addition, if \code{iraceResults$scenario$logFile} is defined,
it saves an updated copy of \code{iraceResults} in that file, where \code{iraceResults$psrace_log} is a list with the following elements:
\describe{
\item{configurations}{Configurations used in the post-selection race.}
\item{instances}{Data frame with the instances used in the post-selection race. First column has the
instances IDs from \code{iraceResults$scenario$instances}, second column the seed assigned to the instance.}
\item{max_experiments}{Configuration budget assigned to the post-selection race.}
\item{experiments}{Matrix of results generated by the post-selection  race, in the same format as the matrix \code{iraceResults$experiments}. Column names are the configuration IDs and  row names are the instance IDs.}
\item{elites}{Best configurations found in the experiments.}
}
}
\description{
\code{psRace} performs a post-selection race of a set of configurations.
}
\examples{
\donttest{
  irace_log <- read_logfile(system.file(package="irace", "exdata", "sann.rda"))
  # Use a temporary file to not change the original "sann.rda".
  psrace_logFile <- withr::local_tempfile(fileext = ".Rdata")
  # Execute the post-selection after the execution of irace. Use 10\% of the total budget.
  psRace(irace_log, max_experiments=0.1, psrace_logFile = psrace_logFile)
  # Print psrace_log
  irace_log <- read_logfile(psrace_logFile)
  str(irace_log$psrace_log)
}
}
\author{
Leslie Pérez Cáceres and Manuel López-Ibáñez
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testing.R
\name{testConfigurations}
\alias{testConfigurations}
\title{Execute the given configurations on the testing instances specified in the
scenario}
\usage{
testConfigurations(configurations, scenario)
}
\arguments{
\item{configurations}{\code{data.frame}\cr Parameter configurations of the
target algorithm (one per row).}

\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
A list with the following elements:
\describe{
\item{\code{experiments}}{Experiments results.}
\item{\code{seeds}}{Array of the instance seeds used in the experiments.}
}
}
\description{
Execute the given configurations on the testing instances specified in the
scenario
}
\details{
A test instance set must be provided through \code{scenario[["testInstances"]]}.
}
\seealso{
\code{\link[=testing_fromlog]{testing_fromlog()}}
}
\author{
Manuel López-Ibáñez
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{checkIraceScenario}
\alias{checkIraceScenario}
\title{Test that the given irace scenario can be run.}
\usage{
checkIraceScenario(scenario)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}
}
\value{
returns \code{TRUE} if successful and gives an error and returns \code{FALSE}
otherwise.
}
\description{
Test that the given irace scenario can be run by checking the scenario
settings provided and trying to run the target-algorithm.
}
\details{
If the \code{parameters} argument is missing, then the parameters
will be read from the file \code{parameterFile}  given by \code{scenario}. If
\code{parameters} is provided, then \code{parameterFile} will not be read.  This function will
try to execute the target-algorithm.
}
\seealso{
\describe{
\item{\code{\link{readScenario}}}{for reading a configuration scenario from a file.}
\item{\code{\link{printScenario}}}{prints the given scenario.}
\item{\code{\link{defaultScenario}}}{returns the default scenario settings of \pkg{irace}.}
\item{\code{\link{checkScenario}}}{to check that the scenario is valid.}
}
}
\author{
Manuel López-Ibáñez and Jérémie Dubois-Lacoste
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parameterAnalysis.R
\name{getFinalElites}
\alias{getFinalElites}
\title{Return the elite configurations of the final iteration.}
\usage{
getFinalElites(iraceResults, n = 0L, drop.metadata = FALSE)
}
\arguments{
\item{iraceResults}{\code{list()}|\code{character(1)}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}. If a character string is given, then it is interpreted as the path to the log file from which the \code{iraceResults} object will be loaded.}

\item{n}{\code{integer(1)}\cr Number of elite configurations to return, if \code{n} is larger than the
number of configurations, then only the existing ones are returned. The default (\code{n=0}) returns all of them.}

\item{drop.metadata}{\code{logical(1)}\cr Remove metadata, such as the
configuration ID and the ID of the parent, from the returned
configurations.  See \code{\link[=removeConfigurationsMetaData]{removeConfigurationsMetaData()}}.}
}
\value{
A data frame containing the elite configurations required.
}
\description{
Return the elite configurations of the final iteration.
}
\examples{
log_file <- system.file("exdata/irace-acotsp.Rdata", package="irace", mustWork=TRUE)
print(removeConfigurationsMetaData(getFinalElites(log_file, n=1)))

}
\author{
Manuel López-Ibáñez and Leslie Pérez Cáceres
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/argparser.R
\name{CommandArgsParser}
\alias{CommandArgsParser}
\alias{cmdline_usage}
\title{R6 Class for parsing command-line arguments}
\usage{
cmdline_usage(cmdline_args)
}
\arguments{
\item{cmdline_args}{Definition of the command-line arguments.}
}
\description{
R6 Class for parsing command-line arguments

R6 Class for parsing command-line arguments

\code{cmdline_usage()} prints the output of \code{--help}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-new}{\code{CommandArgsParser$new()}}
\item \href{#method-readCmdLineParameter}{\code{CommandArgsParser$readCmdLineParameter()}}
\item \href{#method-readArg}{\code{CommandArgsParser$readArg()}}
\item \href{#method-readAll}{\code{CommandArgsParser$readAll()}}
\item \href{#method-cmdline_usage}{\code{CommandArgsParser$cmdline_usage()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-new"></a>}}
\if{latex}{\out{\hypertarget{method-new}{}}}
\subsection{Method \code{new()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CommandArgsParser$new(argv, argsdef)}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-readCmdLineParameter"></a>}}
\if{latex}{\out{\hypertarget{method-readCmdLineParameter}{}}}
\subsection{Method \code{readCmdLineParameter()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CommandArgsParser$readCmdLineParameter(paramName, default = NULL)}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-readArg"></a>}}
\if{latex}{\out{\hypertarget{method-readArg}{}}}
\subsection{Method \code{readArg()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CommandArgsParser$readArg(short = "", long = "")}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-readAll"></a>}}
\if{latex}{\out{\hypertarget{method-readAll}{}}}
\subsection{Method \code{readAll()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CommandArgsParser$readAll()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-cmdline_usage"></a>}}
\if{latex}{\out{\hypertarget{method-cmdline_usage}{}}}
\subsection{Method \code{cmdline_usage()}}{
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{CommandArgsParser$cmdline_usage()}\if{html}{\out{</div>}}
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/irace-package.R
\docType{package}
\name{irace-package}
\alias{irace-package}
\title{The irace package: \packageTitle{irace}}
\description{
\packageDescription{irace}
}
\details{
License: GPL (>= 2)
}
\references{
Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Leslie Pérez Cáceres,
Thomas Stützle, and Mauro Birattari. The irace package: Iterated
Racing for Automatic Algorithm Configuration. \emph{Operations Research
Perspectives}, 2016. \doi{10.1016/j.orp.2016.09.002}

Manuel López-Ibáñez, Jérémie Dubois-Lacoste, Thomas Stützle, and Mauro
Birattari. \emph{The irace package, Iterated Race for Automatic
Algorithm Configuration}. Technical Report TR/IRIDIA/2011-004, IRIDIA,
Université Libre de Bruxelles, Belgium, 2011.

Manuel López-Ibáñez and Thomas Stützle. The Automatic Design of
Multi-Objective Ant Colony Optimization Algorithms. \emph{IEEE Transactions
on Evolutionary Computation}, 2012.
}
\seealso{
\code{\link[=irace]{irace()}} for examples and \code{vignette(package = "irace")} for the user-guide.
}
\author{
Maintainers: Manuel López-Ibáñez and Leslie Pérez Cáceres
\email{irace-package@googlegroups.com}
}
\keyword{automatic}
\keyword{configuration}
\keyword{optimize}
\keyword{package}
\keyword{tuning}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readParameters.R
\name{checkParameters}
\alias{checkParameters}
\title{checkParameters}
\usage{
checkParameters(parameters)
}
\arguments{
\item{parameters}{\code{ParameterSpace}\cr Data structure containing the parameter
space definition. The data structure has to similar to the one returned by the
function \code{\link{readParameters}}.}
}
\description{
FIXME: This is incomplete, for now we only repair inputs from previous irace
versions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ablation.R
\name{read_ablogfile}
\alias{read_ablogfile}
\title{Read the log file (\code{log-ablation.Rdata}) produced by \code{\link[=ablation]{ablation()}}.}
\usage{
read_ablogfile(filename)
}
\arguments{
\item{filename}{\code{character(1)}\cr Filename that contains the log file saved by \code{\link[=ablation]{ablation()}}. Example: \code{log-ablation.Rdata}.}
}
\value{
\code{list()}
}
\description{
Read the log file (\code{log-ablation.Rdata}) produced by \code{\link[=ablation]{ablation()}}.
}
\concept{ablation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{save_irace_logfile}
\alias{save_irace_logfile}
\title{Save the log generated by \pkg{{irace}} to a file (by default \code{irace.Rdata}).}
\usage{
save_irace_logfile(iraceResults, logfile)
}
\arguments{
\item{iraceResults}{\code{list()}\cr Object created by \pkg{irace} and typically saved in the log file \code{irace.Rdata}.}

\item{logfile}{\code{character(1)}\cr Filename to save \code{iraceResults}. Usually, this is given by \code{scenario$logFile}. If \code{NULL} or \code{""}, no data is saved.}
}
\description{
This function may be useful if you are manually editing the log data generated by a run of \pkg{irace}.
}
\seealso{
\code{\link[=read_logfile]{read_logfile()}}
}
\concept{analysis}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{scenario_update_paths}
\alias{scenario_update_paths}
\title{Update filesystem paths of a scenario consistently.}
\usage{
scenario_update_paths(scenario, from, to, fixed = TRUE)
}
\arguments{
\item{scenario}{\code{list()}\cr Data structure containing \pkg{irace}
settings. The data structure has to be the one returned by the function
\code{\link[=defaultScenario]{defaultScenario()}} or \code{\link[=readScenario]{readScenario()}}.}

\item{from}{\code{character(1)}\cr Character string containing a regular expression (or character
string for \code{fixed = TRUE}) to be matched.}

\item{to}{\code{character(1)}\cr  The replacement string.character string. For \code{fixed = FALSE}
this can include backreferences \code{"\\1"} to \code{"\\9"} to
parenthesized subexpressions of \code{from}.}

\item{fixed}{\code{logical(1)}\cr If \code{TRUE}, \code{from} is a string to be matched
as is.}
}
\value{
The updated scenario
}
\description{
This function should be used to change the filesystem paths stored in a
scenario object. Useful when moving a scenario from one computer to another.
}
\examples{
\dontrun{
scenario <- readScenario(filename = "scenario.txt")
scenario <- scenario_update_paths(scenario, from = "/home/manuel/", to = "/home/leslie")
}
}
\seealso{
\code{\link[base:grep]{base::grep()}}
}
